/**************************************************************************

   fotoxx      digital photo edit program

   Copyright 2007, 2008, 2009, 2010  Michael Cornelison
   source URL:  kornelix.squarespace.com
   contact: kornelix@yahoo.de
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program. If not, see http://www.gnu.org/licenses/.

***************************************************************************/

#include <tiffio.h>
#include "zfuncs.h"

#define fversion "Fotoxx v.10.11.1  2010.11.01"                            //  version and date
#define fversioncc 16                                                      //  cc to use for window title bar
#define flicense "Free software - GNU General Public License v.3"
#define fhomepage "http://kornelix.squarespace.com/fotoxx"
#define ftranslators "Translators: \n"                                     \
      "Jie Luo, Eugenio Baldi, Justa, Peter Landgren, \n"                  \
      "Alexander Krasnopolsky, Arthur Kalverboer, \n"                      \
      "André Campos Rodovalho"
#define fcredits "Software used: GTK, libtiff, ufraw-batch, exiftool"
#define fcontact "Bug reports: kornelix@yahoo.de"

//  GTK definitions

#define PXB GdkPixbuf
#define TEXTWIN GTK_TEXT_WINDOW_TEXT                                       //  GDK window of GTK text view
#define NODITHER GDK_RGB_DITHER_NONE
#define ALWAYS GTK_POLICY_ALWAYS
#define NEVER GTK_POLICY_NEVER
#define COLORSPACE GDK_COLORSPACE_RGB
#define LINEATTRIBUTES GDK_LINE_SOLID, GDK_CAP_BUTT, GDK_JOIN_MITER
#define MWIN GTK_WINDOW(mWin)

//  EXIF keys for embedded image data

#define exif_tags_key_old "UserComment"                                    //  tags                obsolete    v.10.0
#define exif_tags_key "Keywords"                                           //  tags                new         v.10.0
#define exif_rating_key "Rating"                                           //  "star" rating       new         v.10.0
#define exif_date_key "DateTimeOriginal"                                   //  date/time
#define exif_width_key "ImageWidth"                                        //  width, pixels
#define exif_height_key "ImageHeight"                                      //  height, pixels
#define exif_orientation_key "Orientation"                                 //  orientation
#define exif_log_key "EditStatus"                                          //  edit log            new         v.10.2
#define exif_comment_key "UserComment"                                     //  user comments       new         v.10.2
#define exif_focal_length_key "FocalLengthIn35mmFormat"                    //  focal length        new         v.10.3

//  fotoxx definitions

#define timer1() double cput0 = CPUtime();                                 //  start CPU timer - diagnostic tool
#define timer2(func) printf("%s %.3f \n",#func,CPUtime()-cput0);           //  print func and timer
#define track(name) printf("track: %s \n",#name);                          //  debug path tracker
#define mwinww 800                                                         //  main window default size
#define mwinhh 600
#define mega (1024 * 1024)
#define ftrash "Desktop/fotoxx-trash"                                      //  trash folder location
#define def_jpeg_quality "90"                                              //  default jpeg quality (high)
#define pi 3.141592654
#define PXMpix(pxm,px,py) ((uint16 *) pxm->bmp+((py)*(pxm->ww)+(px))*3)    //  return PXM pixel[3] at (px,py)
#define pixbright(pix) (0.25*(pix)[0]+0.65*(pix)[1]+0.10*(pix)[2])         //  pixel brightness, 0-64K
#define pixred(pix) (25 * pix[0] / (pixbright(pix)+1))                     //  pixel redness, 0-100%

#define max_images 100000                                                  //  max. image files
#define max_threads 6                                                      //  max. working threads         v.10.10
#define maxedits 100                                                       //  max. edits, undo/redo depth
#define maxtagcats 200                                                     //  max tag categories
#define tagcc  50                                                          //  max cc for one tag or category ID
#define tagFcc 1000                                                        //  max tag cc for one image file
#define tagGcc 10000                                                       //  max tag cc for one category
#define tagntc 1000                                                        //  max no. tags for one category
#define tagTcc 50000                                                       //  max tag cc for all defined tags
#define tagMcc 200                                                         //  max tag cc for batch add tags
#define tagScc 200                                                         //  max tag cc for search tags
#define tagRcc 200                                                         //  max tag cc for recent tags
#define tagrecl 2000                                                       //  tag rec. cc (>maxfcc and >tagFcc)

namespace  image_navi {                                                    //  zfuncs: image_gallery() etc.
   extern int     xwinW, xwinH;                                            //  image gallery window size
   extern int     thumbsize;                                               //  thumbnail image size
}

namespace  zfuncs {
   extern char    zlanguage[8];                                            //  current language lc_RC
   extern cchar   *F1_help_topic;                                          //  F1 context help topic
   extern char    zicondir[200];
}

GtkWidget      *mWin, *drWin, *mVbox;                                      //  main and drawing window
GtkWidget      *mMbar, *mTbar, *STbar;                                     //  menu bar, tool bar, status bar
GdkGC          *gdkgc = 0;                                                 //  GDK graphics context
GdkColor       black, white, gray, red, green;
GdkColormap    *colormap = 0;
uint           maxcolor = 0xffff;

GdkCursor      *arrowcursor = 0;
GdkCursor      *dragcursor = 0;
GdkCursor      *drawcursor = 0;
GdkCursor      *busycursor = 0;
GdkCursor      *currentcursor = 0;

char        command[maxfcc*2+200];                                         //  command, parameters, 2 filespecs   v.10.0
int         ccc = maxfcc*2 + 200;                                          //  command line size

char        PIDstring[12];                                                 //  process PID as string
pthread_t   tid_fmain = 0;                                                 //  thread ID for main()

int         Nwt = 0;                                                       //  working threads to use
int         wtnx[8] = { 0, 1, 2, 3, 4, 5, 6, 7 };                          //  internal thread IDs
int         wthreads_busy = 0;                                             //  active thread count

int         Fexiftool = 0;                                                 //  exiftool program available
int         Fexifwarned = 0;                                               //  missing exiftool was warned
int         Fufraw = 0;                                                    //  ufraw-batch program is available
int         Fxdgopen = 0;                                                  //  xdg-open program available
int         Fshutdown = 0;                                                 //  app shutdown underway
int         Fdebug = 0;                                                    //  debug flag
int         Fkillfunc = 0;                                                 //  signal, kill running function      v.9.1
int         Wrepaint = 0;                                                  //  request window paint
int         Wpainted = 0;                                                  //  window was repainted
int         Fmodified = 0;                                                 //  image was edited/modified
int         Fuprighted = 0;                                                //  image was auto-uprighted
int         Frotated = 0;                                                  //  image was manually rotated
int         Fsaved = 0;                                                    //  image saved since last mod
int         Ftagschanged = 0;                                              //  image tags changed and not saved
int         Fpreview = 0;                                                  //  use window image for edits
int         Fblowup = 0;                                                   //  zoom small images to window size
int         Fshowarea = 0;                                                 //  show select area outline
char        FareaRGB = 'G';                                                //  select area outline color
int         Fslideshow = 0;                                                //  slide show mode is active
int         SS_interval = 3;                                               //  slide show interval
int         SS_timer = 0;                                                  //  slide show timer
int         SBupdate = 0;                                                  //  request to update status bar
int         SB_goal = 0, SB_done = 0;                                      //  status bar progress tracking
char        SB_text[100];                                                  //  optional status bar text

int         Fgrid[2];                                                      //  x and y gridlines on / off 
int         gridspace[2] = { 50, 50 };                                     //  x and y grid line spacing
int         gridcount[2] = { 0, 0 };                                       //  grid line counts (alternative)

char        *image_file = 0;                                               //  current image file
char        *image_dirk = 0;                                               //  current image directory
double      file_MB;                                                       //  disk file size, MB
int         file_bpc;                                                      //  disk file bits/color 8/16
char        file_type[8];                                                  //  "jpeg" "tiff" "png" "other"
char        jpeg_quality[8] = def_jpeg_quality;                            //  jpeg file save quality

char        *recentfiles[30];                                              //  30 most recent image files
int         Nrecentfiles = 30;                                             //  recent file list size

char        *tags_defined_file = 0;                                        //  tags defined file            v.10.5
char        *tags_index_file = 0;                                          //  tags index file              v.10.0
char        *asstags_file = 0;                                             //  tags index file   OBSOLETE from v.10.0

char        *select_area_dirk = 0;                                         //  directory for saved select area files
char        *topdirk = 0;                                                  //  top-level image directory
char        *topmenu;                                                      //  latest top-menu selection

//  fotoxx PXM pixmaps

typedef struct  {                                                          //  PXM pixmap
   char     wmi[8];
   int      ww, hh, bpc;                                                   //  width, height, bits per color
   void     *bmp;                                                          //  uint8*/uint16* (bpc=8/16)
}  PXM;

PXM         *Fpxm8 = 0;                                                    //  input file 1 pixmap, PXM-8
PXM         *Fpxm16 = 0;                                                   //  input file 1 pixmap, PXM-16
PXM         *E1pxm16 = 0;                                                  //  edit pixmap, base image
PXM         *E3pxm16 = 0;                                                  //  edit pixmap, edited image
PXM         *ERpxm16 = 0;                                                  //  edit pixmap, undo/redo image
PXM         *E9pxm16 = 0;                                                  //  scratch image for some functions
PXM         *Dpxm16 = 0;                                                   //  drawing window pixmap, 1x 16-bits
PXM         *Dpxm8 = 0;                                                    //  drawing window pixmap, Mscale 8-bits

int         Fww, Fhh, Gww, Ghh;                                            //  input image dimensions
int         E1ww, E1hh, E3ww, E3hh;                                        //  edit image dimensions
int         ERww, ERhh;                                                    //  undo/redo image dimensions
int         E9ww, E9hh;                                                    //  scratch image dimensions
int         Dww = mwinww, Dhh = mwinhh;                                    //  drawing window size (defaults)

double      Fzoom = 0;                                                     //  image zoom scale (0 = fit window)
int         zoomx = 0, zoomy = 0;                                          //  req. zoom center of window
double      Mscale = 1;                                                    //  scale factor, image to window
int         Iww, Ihh;                                                      //  current image size at 1x
int         iww, ihh;                                                      //  area in drawing window at Mscale
int         dww, dhh;                                                      //  drawing window image, iww * Mscale
int         Iorgx, Iorgy;                                                  //  drawing window origin in image
int         Dorgx, Dorgy;                                                  //  image origin in drawing window
int         Irefresh;                                                      //  full image refresh needed

mutex       pixmaps_lock;                                                  //  lock for accessing PXM pixmaps
int         menu_lock = 0;                                                 //  lock for some menu functions

cchar       *edit_function = 0;                                            //  current edit function
pvlist      *editlog = 0;                                                  //  log of image edits done   v.10.2
char        *undo_files = 0;                                               //  undo/redo stack, image files
int         Pundo = 0;                                                     //  undo/redo stack position
int         Pumax = 0;                                                     //  undo/redo stack depth

int         Mbutton = 0;                                                   //  mouse button, 1/3 = left/right
int         LMclick = 0, RMclick = 0;                                      //  mouse left, right click
int         Mxclick, Myclick;                                              //  mouse click position
int         Mxposn, Myposn;                                                //  mouse move position
int         Mxdown, Mydown, Mxdrag, Mydrag;                                //  mouse drag vector
int         Mdrag = 0;                                                     //  mouse drag underway
int         Mcapture = 0;                                                  //  mouse captured by edit function

int         KBcapture = 0;                                                 //  KB key captured by edit function
int         KBkey = 0;                                                     //  active keyboard key

int         Ntoplines = 0, Nptoplines = 0;                                 //  lines on top of image in window
int         toplinex1[4], topliney1[4], toplinex2[4], topliney2[4];
int         ptoplinex1[4], ptopliney1[4], ptoplinex2[4], ptopliney2[4];

int         Ftoparc = 0, ptoparc = 0;                                      //  arc (circle/ellipse) on top
int         toparcx,toparcy,toparcw,toparch;                               //    of image in window
int         ptoparcx,ptoparcy,ptoparcw,ptoparch;

zdialog     *zdedittags = null;                                            //  edit tags zdialog
zdialog     *zdexifview = null;                                            //  dialog to view EXIF data
zdialog     *zdexifedit = null;                                            //  dialog to edit EXIF data
zdialog     *zdeditucomm = null;                                           //  dialog to edit EXIF user comment
zdialog     *zdrename = null;                                              //  rename file zdialog
zdialog     *zdsela = null;                                                //  select area dialog
zdialog     *zdedit = null;                                                //  image edit dialog

//  GTK functions

int   main(int argc, char * argv[]);                                       //  main program
int   gtkinitfunc(void *data);                                             //  GTK initz. function
int   gtimefunc(void *arg);                                                //  periodic function
int   delete_event();                                                      //  window delete event function
void  destroy_event();                                                     //  window destroy event function
int   mwpaint();                                                           //  window repaint (expose event)
void  mwpaint2();                                                          //  window repaint (thread callable)
void  mwpaint3(int px, int py, int ww, int hh);                            //  paint part of an image being edited
void  update_statusbar();                                                  //  update main window status bar
void  mouse_event(GtkWidget *, GdkEventButton *, void *);                  //  mouse event function
int   KBpress(GtkWidget *, GdkEventKey *, void *);                         //  KB key press event function
int   KBrelease(GtkWidget *, GdkEventKey *, void *);                       //  KB key release event
void  paint_toplines(int arg);                                             //  paint lines on image
void  paint_toparc(int arg);                                               //  paint arc on image
void  paint_gridlines();                                                   //  paint grid lines on image
void  draw_line(int x1, int y1, int x2, int y2);                           //  draw line, image space
void  erase_line(int x1, int y1, int x2, int y2);                          //  erase line

void  topmenufunc(GtkWidget *, cchar *menu);                               //  menu function

typedef void CBfunc();                                                     //  callback function type
CBfunc   *mouseCBfunc = 0;                                                 //  current mouse handler function
CBfunc   *KBkeyCBfunc = 0;                                                 //  current KB handler function

//  file functions

void  m_gallery(GtkWidget *, cchar *);                                     //  show image gallery window
void  m_gallery2(char *);                                                  //  converts function types
void  m_open(GtkWidget *, cchar *);                                        //  open image file (menu)
void  m_open_drag(int x, int y, char *file);                               //  open drag-drop file
void  m_previous(GtkWidget *, cchar *);                                    //  open previously accessed file
void  m_recent(GtkWidget *, cchar *);                                      //  open recently accessed file
void  add_recent_file(cchar *file);                                        //  add file to recent file list
int   f_open(cchar *file, int flock);                                      //  open file helper function
int   make_current_file(cchar *file);                                      //  setup current image file
void  m_prev(GtkWidget *, cchar *);                                        //  open previous file
void  m_next(GtkWidget *, cchar *);                                        //  open next file
void  m_save(GtkWidget *, cchar *);                                        //  save modified image to same file
void  m_saveas(GtkWidget *, cchar *);                                      //  save modified image to another file
void  m_print(GtkWidget *, cchar *);                                       //  print image file
void  m_trash(GtkWidget *, cchar *);                                       //  move image to trash
void  m_rename(GtkWidget *, cchar *);                                      //  rename file
void  m_batchrename(GtkWidget *, cchar *);                                 //  rename many files, base + sequence
void  m_quit(GtkWidget *, cchar *);                                        //  exit application

//  tools functions

void  m_zoom(GtkWidget *, cchar *);                                        //  zoom image +/-
void  m_montest(GtkWidget *, cchar *);                                     //  check monitor
void  m_histogram(GtkWidget *, cchar *);                                   //  start brightness histogram
void  histogram_paint();                                                   //  update brightness histogram
void  histogram_destroy();                                                 //  remove histogram window
void  m_clone(GtkWidget *, cchar *);                                       //  start another fotoxx instance
void  m_slideshow(GtkWidget *, cchar *);                                   //  slideshow mode
void  m_showRGB(GtkWidget *, cchar *);                                     //  show RGB values at mouse click
void  m_gridlines(GtkWidget *, cchar *);                                   //  set up grid lines
void  m_lensparms(GtkWidget *, cchar *);                                   //  edit lens parameters
void  m_lang(GtkWidget *, cchar *);                                        //  change language
void  m_launcher(GtkWidget *, cchar *);                                    //  make desktop icon/launcher
void  m_conv_raw(GtkWidget *, cchar *);                                    //  convert RAW files to tiff
void  m_burn(GtkWidget *, cchar *);                                        //  burn selected images to CD/DVD
void  m_email(GtkWidget *, cchar *);                                       //  e-mail selected images
void  m_rebuild_thumbs(GtkWidget *, cchar *);                              //  rebuild thumbnail files
void  m_rebuild_tags(GtkWidget *, cchar *);                                //  rebuild tags index file
void  m_convert_tags(GtkWidget *, cchar *);                                //  convert tags to new standard  OBSOLETE

//  EXIF functions

void  m_exif_view(GtkWidget *, cchar *);                                   //  view EXIF data in popup window
void  m_exif_edit(GtkWidget *, cchar *);                                   //  add or change EXIF data
void  m_exif_delete(GtkWidget *, cchar *);                                 //  delete EXIF data
char  **exif_get(cchar *file, cchar **keys, int nkeys);                    //  get EXIF data for given key(s)
int   exif_put(cchar *file, cchar **keys, cchar **text, int nkeys);        //  put EXIF data for given key(s)
int   exif_copy(cchar *f1, cchar *f2, cchar **k, cchar **t, int nk);       //  copy EXIF data + opt. updates

//  tags functions

void  m_edittags(GtkWidget *, cchar *);                                    //  edit tags dialog
void  m_batchAddTags(GtkWidget *, cchar *);                                //  add tags to many files at once
void  m_batchDelTag(GtkWidget *, cchar *);                                 //  delete/replace a tag for many files
void  m_edit_ucomm(GtkWidget *, cchar *);                                  //  edit user comments 
void  m_searchtags(GtkWidget *, cchar *);                                  //  search images for matching tags
void  load_filetags(cchar *file);                                          //  EXIF >> tags_date, _stars, _filetags
void  save_filetags(cchar *file);                                          //  tags_date, _stars, _filetags >> EXIF
void  update_tags_index(cchar *file);                                      //  update tags index file
void  delete_tags_index(cchar *file);                                      //  delete entry from tags index file

//  select area functions

void  m_select(GtkWidget *, cchar *);                                      //  select area within image
void  m_select_show(GtkWidget *, cchar *);                                 //  enable area for subsequent edits
void  m_select_hide(GtkWidget *, cchar *);                                 //  show area outline
void  m_select_enable(GtkWidget *, cchar *);                               //  hide area outline
void  m_select_disable(GtkWidget *, cchar *);                              //  disable area
void  m_select_invert(GtkWidget *, cchar *);                               //  invert area 
void  m_select_delete(GtkWidget *, cchar *);                               //  delete area 
void  m_select_copy(GtkWidget *, cchar *);                                 //  copy area, save in memory
void  m_select_paste(GtkWidget *, cchar *);                                //  paste saved area into image
void  m_select_open(GtkWidget *, cchar *);                                 //  open an area file, save in memory
void  m_select_save(GtkWidget *, cchar *);                                 //  save area to a file
   
//  edit functions

void  m_whitebal(GtkWidget *, cchar *);                                    //  adjust white balance
void  m_negate(GtkWidget *, cchar *);                                      //  B+W and color negatives
void  m_flatten(GtkWidget *, cchar *);                                     //  flatten brightness distribution
void  m_tune(GtkWidget *, cchar *);                                        //  brightness / color adjustments
void  m_brightramp(GtkWidget *, cchar *);                                  //  ramp brightness across image
void  m_xbrange(GtkWidget *, cchar *);                                     //  clip brightness, expand rest
void  m_tonemap(GtkWidget *, cchar *);                                     //  tone mapping
void  m_redeye(GtkWidget *, cchar *);                                      //  red-eye removal
void  m_blur(GtkWidget *, cchar *);                                        //  blur image
void  m_sharpen(GtkWidget *, cchar *);                                     //  sharpen image
void  m_denoise(GtkWidget *, cchar *);                                     //  image noise reduction
void  m_resize(GtkWidget *, cchar *);                                      //  resize image
void  m_batchresize(GtkWidget *, cchar *);                                 //  resize many images
void  m_annotate(GtkWidget *, cchar *);                                    //  annotate image
void  m_trim(GtkWidget *, cchar *);                                        //  trim image
void  m_rotate(GtkWidget *, cchar *);                                      //  rotate image
void  m_flip(GtkWidget *, cchar *);                                        //  flip horizontally or vertically
void  m_unbend(GtkWidget *, cchar *);                                      //  fix perspective problems
void  m_warp_area(GtkWidget *, cchar *);                                   //  warp image within an area
void  m_warp_curved(GtkWidget *, cchar *);                                 //  warp image, curved transform
void  m_warp_linear(GtkWidget *, cchar *);                                 //  warp image, linear transform
void  m_warp_affine(GtkWidget *, cchar *);                                 //  warp image, affine transform
void  m_colordep(GtkWidget *, cchar *);                                    //  set color depth 1-16 bits/color
void  m_draw(GtkWidget *, cchar *);                                        //  make simulated drawing
void  m_emboss(GtkWidget *, cchar *);                                      //  make simulated embossing
void  m_tiles(GtkWidget *, cchar *);                                       //  make simulated tiles (pixelate)
void  m_painting(GtkWidget *, cchar *);                                    //  make simulated painting
void  m_pixedit(GtkWidget *, cchar *);                                     //  edit individual pixels
void  m_HDR(GtkWidget *, cchar *);                                         //  make HDR composite image
void  m_HDF(GtkWidget *, cchar *);                                         //  make HDF composite image
void  m_stak(GtkWidget *, cchar *);                                        //  combine images for noise reduction
void  m_pano4(GtkWidget *, cchar *);                                       //  make panorama composite image

//  edit support functions

int   edit_setup(cchar *func, int preview, int delsa);                     //  start new edit transaction
void  edit_cancel();                                                       //  cancel edit
void  edit_done();                                                         //  commit edit, add undo stack
void  edit_undo();                                                         //  undo edit, revert
void  edit_redo();                                                         //  redo the edit after undo
void  edit_reset();                                                        //  reset to initial status
void  edit_zapredo();                                                      //  clear image redo copy
void  edit_fullsize();                                                     //  convert to full-size pixmaps

void  m_undo(GtkWidget *, cchar *);                                        //  toolbar, undo one edit
void  m_redo(GtkWidget *, cchar *);                                        //  toolbar, redo one edit
void  save_undo();                                                         //  undo/redo save function
void  load_undo();                                                         //  undo/redo read function

typedef void * threadfunc(void *);                                         //  edit thread function
void  start_thread(threadfunc func, void *arg);                            //  start edit thread
void  signal_thread();                                                     //  signal thread, work is pending
void  wait_thread_idle();                                                  //  wait for work complete
void  wrapup_thread(int command);                                          //  wait for exit or command exit
int   thread_working();                                                    //  0/1 >> thread is idle/working
void  thread_idle_loop();                                                  //  thread: wait for work or exit command
void  thread_exit();                                                       //  thread: exit unconditionally

void  start_wt(threadfunc func, void *arg);                                //  start working thread (per processor core)
void  exit_wt();                                                           //  exit working thread
void  wait_wts();                                                          //  wait for working threads

//  other support functions

void  m_help(GtkWidget *, cchar *);                                        //  various help functions
void  load_params();                                                       //  load parameters from prior session
void  save_params();                                                       //  save parameters for next session
int   mod_keep();                                                          //  query keep/discard edited image
int   menulock(int lock);                                                  //  lock/unlock menu for some funcs
void  brhood_calc(int radius, char method);                                //  compute neighborhood brightness
float get_brhood(int px, int py);                                          //  get pixel neighborhood brightness
void  free_resources();                                                    //  free all allocated resources
void  set_cursor(GdkCursor *cursor);                                       //  set cursor for main/drawing window
int   vpixel(PXM *pxm, double px, double py, uint16 *vpix);                //  get virtual pixel at (px,py)
int   sigdiff(double d1, double d2, double signf);                         //  test for significant difference

PXM * f_load(cchar *filespec, int bpc);                                    //  load a file into an PXM pixmap
int   f_save(cchar *outfile, cchar *format);                               //  save an PXM pixmap to a file
PXM * TIFFread(cchar *filespec);                                           //  read TIFF file, native bpc
int   TIFFwrite(PXM *pxm, cchar *filespec);                                //  write TIFF file, bpc from pxm
void  tiffwarninghandler(cchar *mod, cchar *fmt, va_list);                 //  intercepts TIFF lib warnings
PXM * PXBread(cchar *filespec);                                            //  read using pixbuf library, bpc = 8
int   PXBwrite(PXM *pxm, cchar *filespec);                                 //  write using pixbuf library, bpc = 8

//  PXM pixmap conversion functions

PXM * PXM_make(int ww, int hh, int bpc);                                   //  initialize PXM pixmap
void  PXM_free(PXM *pxm);                                                  //  free PXM pixmap
PXM * PXM_copy(PXM *pxm);                                                  //  copy PXM pixmap
PXM * PXM_copy_area(PXM *pxm, int orgx, int orgy, int ww, int hh);         //  copy section of PXM pixmap
PXM * PXM_convbpc(PXM *pxm);                                               //  convert from 8/16 to 16/8 bpc
PXM * PXM_rescale(PXM *pxm, int ww, int hh);                               //  rescale PXM pixmap (ww/hh)
void  PXM_update(PXM *p16, PXM *p8, int orgx, int orgy, int ww, int hh);   //  copy section of PXM-16 into PXM-8
PXM * PXM_rotate(PXM *pxm, double angle);                                  //  rotate PXM pixmap

//  translatable strings used in multiple dialogs

cchar  *Baddall;
cchar  *Bamount;
cchar  *Bapply;
cchar  *Bblendwidth;
cchar  *Bblue;
cchar  *Bbrightness;
cchar  *Bbrowse;
cchar  *Bcancel;
cchar  *Bclear;
cchar  *Bcolor;
cchar  *Bdarker;
cchar  *Bdelete;
cchar  *Bdisable;
cchar  *Bdone;
cchar  *Bedit;
cchar  *Benable;
cchar  *Bexiftoolmissing;
cchar  *Bfetch;
cchar  *Bfinish;
cchar  *Bfont;
cchar  *Bgreen;
cchar  *Bheight;
cchar  *Bhide;
cchar  *Binsert;
cchar  *Binvert;
cchar  *Blighter;
cchar  *Blimit;
cchar  *Bnext;
cchar  *BOK;
cchar  *Bopen;
cchar  *Bopenrawfile;
cchar  *Bpause;
cchar  *Bpercent;
cchar  *Bpresets;
cchar  *Bproceed;
cchar  *Bradius;
cchar  *Brange;
cchar  *Bred;
cchar  *Bredo;
cchar  *Breduce;
cchar  *Bresume;
cchar  *Bsave;
cchar  *Bsavetoedit;
cchar  *Bsearch;
cchar  *Bselect;
cchar  *Bselectfiles;
cchar  *Bshow;
cchar  *Bstart;
cchar  *Bsuspend;
cchar  *Bthresh;
cchar  *Bundo;
cchar  *Bundoall;
cchar  *Bundolast;
cchar  *Bwidth;


/**************************************************************************
      main program and GTK/GDK functions
***************************************************************************/

int main(int argc, char *argv[])
{
   using namespace zfuncs;

   char           lang[8] = "";
   
   initz_appfiles("fotoxx",null);                                          //  get app directories

   printf(fversion "\n");                                                  //  print fotoxx version
   if (argc > 1 && strEqu(argv[1],"-v")) return 0;

   gtk_init(&argc,&argv);                                                  //  initz. GTK

   load_params();                                                          //  restore parameters from last session

   for (int ii = 1; ii < argc; ii++)                                       //  command line options
   {
      if (strEqu(argv[ii],"-d"))                                           //  -d (debug flag)
            Fdebug = 1;
      else if (strEqu(argv[ii],"-l") && argc > ii+1)                       //  -l language code
            strncpy0(lang,argv[++ii],7);
      else  image_file = strdupz(argv[ii],0,"initfile");                   //  initial file or directory
   }

   ZTXinit(lang);                                                          //  setup translations

   mWin = gtk_window_new(GTK_WINDOW_TOPLEVEL);                             //  create main window
   gtk_window_set_title(MWIN,fversion);
   gtk_window_set_position(MWIN,GTK_WIN_POS_CENTER);
   gtk_window_set_default_size(MWIN,Dww,Dhh);

   mVbox = gtk_vbox_new(0,0);                                              //  add vert. packing box
   gtk_container_add(GTK_CONTAINER(mWin),mVbox);

   mMbar = create_menubar(mVbox);                                          //  menus / sub-menus

   GtkWidget *mFile = add_menubar_item(mMbar,ZTX("File"),topmenufunc);
      add_submenu_item(mFile,       ZTX("Image Gallery"),         m_gallery);
      add_submenu_item(mFile,       ZTX("Open Image File"),       m_open);
      add_submenu_item(mFile,       ZTX("Open Previous File"),    m_previous);
      add_submenu_item(mFile,       ZTX("Open Recent File"),      m_recent);
      add_submenu_item(mFile,       ZTX("Save to Same File"),     m_save);
      add_submenu_item(mFile,       ZTX("Save to New File"),      m_saveas);
      add_submenu_item(mFile,       ZTX("Print Image File"),      m_print);
      add_submenu_item(mFile,       ZTX("Trash Image File"),      m_trash);
      add_submenu_item(mFile,       ZTX("Rename Image File"),     m_rename);
      add_submenu_item(mFile,       ZTX("Batch Rename Files"),    m_batchrename);
      add_submenu_item(mFile,       ZTX("Quit fotoxx"),           m_quit);

   GtkWidget *mTools = add_menubar_item(mMbar,ZTX("Tools"),topmenufunc);
      add_submenu_item(mTools,      ZTX("Check Monitor"),         m_montest);
      add_submenu_item(mTools,      ZTX("Brightness Graph"),      m_histogram);
      add_submenu_item(mTools,      ZTX("Clone fotoxx"),          m_clone);
      add_submenu_item(mTools,      ZTX("Slide Show"),            m_slideshow);
      add_submenu_item(mTools,      ZTX("Show RGB"),              m_showRGB);
      add_submenu_item(mTools,      ZTX("Grid Lines"),            m_gridlines);
      add_submenu_item(mTools,      ZTX("Lens Parameters"),       m_lensparms);
      add_submenu_item(mTools,      ZTX("Change Language"),       m_lang);
      add_submenu_item(mTools,      ZTX("Create Launcher"),       m_launcher);
      add_submenu_item(mTools,      ZTX("Convert RAW files"),     m_conv_raw);
      add_submenu_item(mTools,      ZTX("Burn Images to CD/DVD"), m_burn);
      add_submenu_item(mTools,      ZTX("E-mail Images"),         m_email);
      add_submenu_item(mTools,      ZTX("Rebuild Thumbnails"),    m_rebuild_thumbs);
      add_submenu_item(mTools,      ZTX("Rebuild Tags Index"),    m_rebuild_tags);
      add_submenu_item(mTools,      ZTX("Convert Tags !!!"),      m_convert_tags);        //  OBSOLETE

   GtkWidget *mExif = add_menubar_item(mMbar,"EXIF",topmenufunc);
      add_submenu_item(mExif,       ZTX("Basic EXIF data"),       m_exif_view);
      add_submenu_item(mExif,       ZTX("All EXIF data"),         m_exif_view);
      add_submenu_item(mExif,       ZTX("Edit EXIF data"),        m_exif_edit);
      add_submenu_item(mExif,       ZTX("Delete EXIF data"),      m_exif_delete);

   GtkWidget *mTags = add_menubar_item(mMbar,ZTX("Tags"),topmenufunc);
      add_submenu_item(mTags,       ZTX("Edit Tags"),             m_edittags);
      add_submenu_item(mTags,       ZTX("Batch Add Tags"),        m_batchAddTags);
      add_submenu_item(mTags,       ZTX("Batch Delete Tag"),      m_batchDelTag);
      add_submenu_item(mTags,       ZTX("Edit User Comments"),    m_edit_ucomm);
      add_submenu_item(mTags,       ZTX("Search Tags"),           m_searchtags);

   GtkWidget *mArea = add_menubar_item(mMbar,ZTX("Area"),topmenufunc);
      add_submenu_item(mArea,       ZTX("Select"),                m_select);
      add_submenu_item(mArea,       ZTX("Show"),                  m_select_show);
      add_submenu_item(mArea,       ZTX("Hide"),                  m_select_hide);
      add_submenu_item(mArea,       ZTX("Enable"),                m_select_enable);
      add_submenu_item(mArea,       ZTX("Disable"),               m_select_disable);
      add_submenu_item(mArea,       ZTX("Invert"),                m_select_invert);
      add_submenu_item(mArea,       ZTX("Delete"),                m_select_delete);
      add_submenu_item(mArea,       ZTX("Copy"),                  m_select_copy);
      add_submenu_item(mArea,       ZTX("Paste"),                 m_select_paste);
      add_submenu_item(mArea,       ZTX("Open"),                  m_select_open);
      add_submenu_item(mArea,       ZTX("Save"),                  m_select_save);

   GtkWidget *mRetouch = add_menubar_item(mMbar,ZTX("Retouch"),topmenufunc);
      add_submenu_item(mRetouch,    ZTX("White Balance"),         m_whitebal);
      add_submenu_item(mRetouch,    ZTX("Make Negative"),         m_negate);
      add_submenu_item(mRetouch,    ZTX("Flatten Brightness"),    m_flatten);
      add_submenu_item(mRetouch,    ZTX("Brightness/Color"),      m_tune);
      add_submenu_item(mRetouch,    ZTX("Brightness Ramp"),       m_brightramp);
      add_submenu_item(mRetouch,    ZTX("Expand Brightness"),     m_xbrange);
      add_submenu_item(mRetouch,    ZTX("Tone Mapping"),          m_tonemap);
      add_submenu_item(mRetouch,    ZTX("Red Eyes"),              m_redeye);
      add_submenu_item(mRetouch,    ZTX("Blur Image"),            m_blur);
      add_submenu_item(mRetouch,    ZTX("Sharpen Image"),         m_sharpen);
      add_submenu_item(mRetouch,    ZTX("Reduce Noise"),          m_denoise);

   GtkWidget *mTransf = add_menubar_item(mMbar,ZTX("Transform"),topmenufunc);
      add_submenu_item(mTransf,     ZTX("Trim Image"),            m_trim);
      add_submenu_item(mTransf,     ZTX("Resize Image"),          m_resize);
      add_submenu_item(mTransf,     ZTX("Batch Resize"),          m_batchresize);
      add_submenu_item(mTransf,     ZTX("Annotate Image"),        m_annotate);
      add_submenu_item(mTransf,     ZTX("Rotate Image"),          m_rotate);
      add_submenu_item(mTransf,     ZTX("Flip Image"),            m_flip);
      add_submenu_item(mTransf,     ZTX("Unbend Image"),          m_unbend);
      add_submenu_item(mTransf,     ZTX("Warp Image (area)"),     m_warp_area);
      add_submenu_item(mTransf,     ZTX("Warp Image (curved)"),   m_warp_curved);
      add_submenu_item(mTransf,     ZTX("Warp Image (linear)"),   m_warp_linear);
      add_submenu_item(mTransf,     ZTX("Warp Image (affine)"),   m_warp_affine);

   GtkWidget *mArt = add_menubar_item(mMbar,ZTX("Art"),topmenufunc);
      add_submenu_item(mArt,        ZTX("Color Depth"),           m_colordep);
      add_submenu_item(mArt,        ZTX("Simulate Drawing"),      m_draw);
      add_submenu_item(mArt,        ZTX("Simulate Embossing"),    m_emboss);
      add_submenu_item(mArt,        ZTX("Simulate Tiles"),        m_tiles);
      add_submenu_item(mArt,        ZTX("Simulate Painting"),     m_painting);
      add_submenu_item(mArt,        ZTX("Edit Pixels"),           m_pixedit);

   GtkWidget *mComb = add_menubar_item(mMbar,ZTX("Combine"),topmenufunc);
      add_submenu_item(mComb,       ZTX("HDR"),                   m_HDR);
      add_submenu_item(mComb,       ZTX("HDF"),                   m_HDF);
      add_submenu_item(mComb,       ZTX("Stack"),                 m_stak);
      add_submenu_item(mComb,       ZTX("Panorama"),              m_pano4);

   GtkWidget *mHelp = add_menubar_item(mMbar,ZTX("Help"),topmenufunc);
      add_submenu_item(mHelp,       ZTX("About"),                 m_help);
      add_submenu_item(mHelp,       ZTX("User Guide"),            m_help);
      add_submenu_item(mHelp,       "README",                     m_help);
      add_submenu_item(mHelp,       ZTX("Change Log"),            m_help);
      add_submenu_item(mHelp,       ZTX("Translate"),             m_help);
      add_submenu_item(mHelp,       ZTX("Home Page"),             m_help);

   mTbar = create_toolbar(mVbox);                                          //  toolbar buttons
      add_toolbar_button(mTbar,  ZTX("Gallery"),   ZTX("Image Gallery"),         "gallery.png", m_gallery);
      add_toolbar_button(mTbar,  ZTX("Open"),      ZTX("Open Image File"),       "open.png",    m_open);
      add_toolbar_button(mTbar,  ZTX("Prev"),      ZTX("Open Previous File"),    "prev.png",    m_prev);
      add_toolbar_button(mTbar,  ZTX("Next"),      ZTX("Open Next File"),        "next.png",    m_next);
      add_toolbar_button(mTbar,  ZTX("Save"),      ZTX("Save to Same File"),     "save.png",    m_save);
      add_toolbar_button(mTbar,  ZTX("Save As"),   ZTX("Save to New File"),      "saveas.png",  m_saveas);
      add_toolbar_button(mTbar,  ZTX("Undo"),      ZTX("Undo One Edit"),         "undo.png",    m_undo);
      add_toolbar_button(mTbar,  ZTX("Redo"),      ZTX("Redo One Edit"),         "redo.png",    m_redo);
      add_toolbar_button(mTbar,  "Zoom+",          ZTX("Zoom-in (bigger)"),      "zoom+.png",   m_zoom);
      add_toolbar_button(mTbar,  "Zoom-",          ZTX("Zoom-out (smaller)"),    "zoom-.png",   m_zoom);
      add_toolbar_button(mTbar,  ZTX("Trash"),     ZTX("Move Image to Trash"),   "trash.png",   m_trash);
      add_toolbar_button(mTbar,  ZTX("Quit"),      ZTX("Quit fotoxx"),           "quit.png",    m_quit);
      add_toolbar_button(mTbar,  ZTX("Help"),      ZTX("Fotoxx Essentials"),     "help.png",    m_help);
   
   drWin = gtk_drawing_area_new();                                         //  add drawing window
   gtk_box_pack_start(GTK_BOX(mVbox),drWin,1,1,0);

   STbar = create_stbar(mVbox);                                            //  add status bar

   G_SIGNAL(mWin,"delete_event",delete_event,0)                            //  connect signals to windows
   G_SIGNAL(mWin,"destroy",destroy_event,0)
   G_SIGNAL(drWin,"expose-event",mwpaint,0)

   gtk_widget_add_events(drWin,GDK_BUTTON_PRESS_MASK);                     //  connect mouse events
   gtk_widget_add_events(drWin,GDK_BUTTON_RELEASE_MASK);
   gtk_widget_add_events(drWin,GDK_BUTTON_MOTION_MASK);                    //  all buttons
   gtk_widget_add_events(drWin,GDK_POINTER_MOTION_MASK);                   //  pointer motion
   G_SIGNAL(drWin,"button-press-event",mouse_event,0)
   G_SIGNAL(drWin,"button-release-event",mouse_event,0)
   G_SIGNAL(drWin,"motion-notify-event",mouse_event,0)

   G_SIGNAL(mWin,"key-press-event",KBpress,0)                           	//  connect KB events
   G_SIGNAL(mWin,"key-release-event",KBrelease,0)
   
   drag_drop_connect(drWin,m_open_drag);                                   //  connect drag-drop event

   gtk_widget_show_all(mWin);                                              //  show all widgets

   gdkgc = gdk_gc_new(drWin->window);                                      //  initz. graphics context

   black.red = black.green = black.blue = 0;                               //  set up colors
   white.red = white.green = white.blue = maxcolor;
   gray.red = gray.green = gray.blue = maxcolor / 3;
   red.red = maxcolor;  red.green = red.blue = 0;
   green.green = maxcolor; green.red = green.blue = 0;

   colormap = gtk_widget_get_colormap(drWin);
   gdk_rgb_find_color(colormap,&black);
   gdk_rgb_find_color(colormap,&white);
   gdk_rgb_find_color(colormap,&gray);
   gdk_rgb_find_color(colormap,&red);
   gdk_rgb_find_color(colormap,&green);
   
   gdk_gc_set_foreground(gdkgc,&black);
   gdk_gc_set_background(gdkgc,&gray);                                     //  does nothing   /////

   gdk_gc_set_line_attributes(gdkgc,1,LINEATTRIBUTES);

   arrowcursor = gdk_cursor_new(GDK_TOP_LEFT_ARROW);                       //  cursor for selection
   dragcursor = gdk_cursor_new(GDK_CROSSHAIR);                             //  cursor for dragging
   drawcursor = gdk_cursor_new(GDK_PENCIL);                                //  cursor for drawing lines

   busycursor = zmakecursor("busy.png");                                   //  cursor for function busy
   if (! busycursor)
      busycursor = gdk_cursor_new(GDK_WATCH);
   
   gtk_init_add((GtkFunction) gtkinitfunc,0);                              //  set initz. call from gtk_main()
   gtk_main();                                                             //  start processing window events
   return 0;
}


/**************************************************************************/

//  initial function called from gtk_main() at startup

int gtkinitfunc(void *data)
{
   int            err, err1, err2, flag, npid;
   char           procfile[20], *ppv;
   cchar          *pp, *pp2;
   struct stat    statb;

   Baddall = ZTX("Add All");
   Bamount = ZTX("Amount");
   Bapply = ZTX("Apply");
   Bblendwidth = ZTX("Blend Width");
   Bblue = ZTX("Blue");
   Bbrightness = ZTX("Brightness");
   Bbrowse = ZTX("Browse");
   Bcancel = ZTX("Cancel");
   Bclear = ZTX("Clear");
   Bcolor = ZTX("Color");
   Bdarker = ZTX("Darker Areas");
   Bdelete = ZTX("Delete");
   Bdisable = ZTX("Disable");
   Bdone = ZTX("Done");
   Bedit = ZTX("Edit");
   Benable = ZTX("Enable");
   Bexiftoolmissing = ZTX("package libimage-exiftool-perl is required");
   Bfetch = ZTX("Fetch");
   Bfinish = ZTX("Finish");
   Bfont = ZTX("Font");
   Bgreen = ZTX("Green");
   Bheight = ZTX("Height");
   Bhide = ZTX("Hide");
   Binsert = ZTX("Insert");
   Binvert = ZTX("Invert");
   Blighter = ZTX("Lighter Areas");
   Blimit = ZTX("limit");
   Bnext = ZTX("Next");
   BOK = ZTX("OK");
   Bopen = ZTX("Open");
   Bopenrawfile = ZTX("Open RAW File");
   Bpause = ZTX("Pause");
   Bpercent = ZTX("Percent");
   Bpresets = ZTX("Presets");
   Bproceed = ZTX("Proceed");
   Bradius = ZTX("Radius");
   Brange = ZTX("range");
   Bredo = ZTX("Redo");
   Breduce = ZTX("Reduce");
   Bred = ZTX("Red");
   Bresume = ZTX("Resume");
   Bsave = ZTX("Save");
   Bsavetoedit = ZTX("Unknown file type, save as tiff/jpeg/png to edit");
   Bsearch = ZTX("Search");
   Bselect = ZTX("Select");
   Bselectfiles = ZTX("Select Files");
   Bshow = ZTX("Show");
   Bstart = ZTX("Start");
   Bsuspend = ZTX("Suspend");
   Bthresh = ZTX("Threshold");
   Bundoall = ZTX("Undo All");
   Bundolast = ZTX("Undo Last");
   Bundo = ZTX("Undo");
   Bwidth = ZTX("Width");

   tid_fmain = pthread_self();                                             //  get main() thread ID
   snprintf(PIDstring,11,"%06d",getpid());                                 //  get fotoxx process PID pppppp

   Nwt = sysconf(_SC_NPROCESSORS_ONLN);                                    //  get SMP CPU count
   if (Nwt <= 0) Nwt = 1;
   if (Nwt > max_threads) Nwt = max_threads;                               //  compile time limit
   printf("using %d threads \n",Nwt);

   err = system("echo -n \"exiftool \"; exiftool -ver");                   //  check for exiftool
   if (! err) Fexiftool = 1;
   err = system("xdg-open --version");                                     //  check for xdg-open
   if (! err) Fxdgopen = 1;
   err = system("ufraw-batch --version");                                  //  check for ufraw-batch   v.10.7
   if (! err) Fufraw = 1;

   tags_defined_file = zmalloc(200,"tagsdefinedfile");                     //  tags defined file      v.10.5
   *tags_defined_file = 0;
   strncatv(tags_defined_file,199,get_zuserdir(),"/tags_defined",null);    //  home/user/.fotoxx/tags_defined

   tags_index_file = zmalloc(200,"tagsindexfile");                         //  tags index file        v.10.0
   *tags_index_file = 0;
   strncatv(tags_index_file,199,get_zuserdir(),"/tags_index",null);        //  home/user/.fotoxx/tags_index

   asstags_file = zmalloc(200,"asstagsfile");                              //  tags index file  OBSOLETE from v.10.0
   *asstags_file = 0;                                                      //  bugfix v.9.9
   strncatv(asstags_file,199,get_zuserdir(),"/assigned_tags",null);        //  home/user/.fotoxx/assigned_tags

   undo_files = zmalloc(200,"undofiles");                                  //  look for orphaned undo files
   *undo_files = 0;
   strncatv(undo_files,199,get_zuserdir(),"/*_undo_*",null);               //  home/user/.fotoxx/*_undo_*
   flag = 1;
   while ((pp = SearchWild(undo_files,flag)))
   {
      pp2 = strstr(pp,".fotoxx/");
      if (! pp2) continue;
      npid = atoi(pp2+8);                                                  //  pid of file owner
      snprintf(procfile,19,"/proc/%d",npid);
      err = stat(procfile,&statb);
      if (! err) continue;                                                 //  pid is active, keep file
      printf("orphaned undo file deleted: %s \n",pp);
      snprintf(command,ccc,"rm -f %s",pp);                                 //  delete orphaned files
      err = system(command);
   }

   *undo_files = 0;                                                        //  setup undo filespec template
   strncatv(undo_files,199,get_zuserdir(),"/",PIDstring,"_undo_nn",null);  //  home/user/.fotoxx/pppppp_undo_nn
   
   editlog = pvlist_create(maxedits);                                      //  history log of image edits done
   for (int ii = 0; ii < maxedits; ii++)                                   //  pre-load all pvlist entries   v.10.2
      pvlist_append(editlog,"nothing");

   select_area_dirk = zmalloc(200,"selectareadirk");                       //  directory for saved areas    v.9.9  
   *select_area_dirk = 0;
   strncatv(select_area_dirk,199,get_zuserdir(),"/saved_areas/",null);     //  home/user/.fotoxx/saved_areas/
   err = stat(select_area_dirk,&statb);
   if (err) mkdir(select_area_dirk,0750);

   mutex_init(&pixmaps_lock,0);                                            //  setup lock for edit pixmaps

   TIFFSetWarningHandler(tiffwarninghandler);                              //  intercept TIFF warning messages
   
   g_timeout_add(20,gtimefunc,0);                                          //  start periodic function (20 ms)  v.10.7

   //  determine if tags conversion is needed                              //  v.10.0   

   err1 = stat(asstags_file,&statb);                                       //  is old tags index present?
   printf("%s",asstags_file);                                              //  OBSOLETE from v.10.0
   if (! err1) printf("  - present \n");
   else printf("  - not present \n");

   err2 = stat(tags_index_file,&statb);                                    //  is new tags index present?
   printf("%s",tags_index_file);
   if (! err2) printf("  - present \n");
   else printf("  - not present \n");

   if (err2 && ! err1) {                                                   //  have old file and no new file
      showz_userguide("convert_tags");                                     //       OBSOLETE
      zmessageACK(MWIN,"WARNING: If you use tags, convert to new standard \n"
                       "before any edits! (menu Tools -> Convert Tags)");
   }
   
   if (image_file) {
      ppv = realpath(image_file,0);                                        //  add cwd if needed            v.10.7
      if (ppv) {
         zfree(image_file);
         image_file = strdupz(ppv,0,"image_file");                         //  convert to zmalloc()
         free(ppv);
      }
   }
      
   if (image_file && *image_file == '/') {
      err = f_open(image_file,1);                                          //  open initial file
      if (err) {
         printf("invalid image file: %s \n",image_file);
         zfree(image_file);
         image_file = 0;
      }
   }

   return 0;
}


/**************************************************************************/

//  Periodic function - runs every few milliseconds.
//  Avoid any thread usage of gtk/gdk functions.

int gtimefunc(void *arg)
{
   double      secs;
   static int  fbusy = 0;

   if (Wrepaint) {                                                         //  update drawing window
      Wrepaint = 0;
      mwpaint();
   }

   if (SBupdate || SB_goal) {                                              //  status bar update
      SBupdate = 0;
      update_statusbar();
   }
   
   if (Fslideshow) {
      secs = get_seconds();                                                //  show next slide
      if (secs > SS_timer) {
         SS_timer = secs + SS_interval;
         m_next(0,0);  
      }
   }
   
   if (! fbusy && thread_working()) {
      if (currentcursor == null) {                                   
         set_cursor(busycursor);                                           //  set function busy cursor
         fbusy = 1;
      }
   }
   
   if (fbusy && ! thread_working()) {
      fbusy = 0;
      set_cursor(0);                                                       //  restore normal cursor
   }

   return 1;
}


/**************************************************************************/

//  functions for main window event signals

int delete_event()                                                         //  main window closed
{
   printf("main window delete event \n");
   if (mod_keep()) return 1;                                               //  allow user bailout
   Fshutdown++;                                                            //  shutdown in progress
   save_params();                                                          //  save state for next session
   free_resources();                                                       //  delete undo files
   return 0;
}

void destroy_event()                                                       //  main window destroyed
{
   printf("main window destroy event \n");
   exit(1);                                                                //  instead of gtk_main_quit();
   return;
}


/**************************************************************************/

//  Cause (modified) output image to get repainted immediately.
//  The entire image is repainted (that part within the window).
//  This function may be called from threads.
//  (window paint is done from periodic function in main() thread)

void mwpaint2()
{
   Irefresh++;
   Wrepaint++;                                                             //  req. repaint by periodic function
   SBupdate++;                                                             //  moved for fvwm window manager   v.9.7
   return;
}


//  Repaint a rectangular part of the image being edited.
//  Called from edit functions. Not thread callable.
//
//  px3, py3, ww3, hh3: modified area within E3pxm16 to be repainted
//  Dpxm16: 1x copy of E3pxm16 area currently visible in main window
//  px2, py2, ww2, hh2: E3pxm16 area copied into Dpxm16
//  Dpxm8: window image, Dpxm16 scaled to window, converted to 8 bits

void mwpaint3(int px3, int py3, int ww3, int hh3)                          //  overhauled v.10.11
{
   int         px2, py2, ww2, hh2, dx, dy;
   uint16      *pix2, *pix3;
   uint8       *pxm8area;
   
   px2 = px3 - Iorgx;                                                      //  map modified area into Dpxm16
   py2 = py3 - Iorgy;
   ww2 = ww3;
   hh2 = hh3;

   if (px2 < 0) {                                                          //  if beyond Dpxm16 edge, reduce
      px3 -= px2;
      ww2 += px2;
      px2 = 0;
   }
   if (px2 + ww2 > iww) ww2 = iww - px2;

   if (py2 < 0) {
      py3 -= py2;
      hh2 += py2;
      py2 = 0;
   }
   if (py2 + hh2 > ihh) hh2 = ihh - py2;

   if (ww2 <= 0 || hh2 <= 0) return;                                       //  completely off Dpxm16 image

   for (dy = 0; dy < hh2; dy++)                                            //  copy pixels from E3pxm16 to Dpxm16
   for (dx = 0; dx < ww2; dx++)
   {
      pix3 = PXMpix(E3pxm16,px3+dx,py3+dy);
      pix2 = PXMpix(Dpxm16,px2+dx,py2+dy);
      pix2[0] = pix3[0];
      pix2[1] = pix3[1];
      pix2[2] = pix3[2];
   }

   PXM_update(Dpxm16,Dpxm8,px2,py2,ww2,hh2);                               //  copy/rescale Dpxm16 area into Dpxm8 area

   px2 = px2 * Mscale - 1;                                                 //  Dpxm8 area + small increase
   py2 = py2 * Mscale - 1;
   ww2 = ww2 * Mscale + 2;
   hh2 = hh2 * Mscale + 2;
   
   if (px2 < 0) px2 = 0;                                                   //  stay within image edge
   if (py2 < 0) py2 = 0;
   if (px2 + ww2 > dww) ww2 = dww - px2;
   if (py2 + hh2 > dhh) hh2 = dhh - py2;
   if (ww2 <= 0 || hh2 <= 0) return;                                       //  completely off Dpxm8 image

   pxm8area = (uint8 *) Dpxm8->bmp + 3 * (dww * py2 + px2);                //  origin of area to copy

   gdk_draw_rgb_image(drWin->window, gdkgc, px2 + Dorgx, py2 + Dorgy,      //  copy into window
                                  ww2, hh2, NODITHER, pxm8area, dww*3);
   return;
}


/**************************************************************************/

//  paint window when created, exposed, resized, or modified (edited)

int mwpaint()
{
   GdkRectangle   wrect;
   int            incrx, incry;                                            //  mouse drag
   static int     pincrx = 0, pincry = 0;                                  //  prior mouse drag
   static int     pdww = 0, pdhh = 0;                                      //  prior window size
   static int     piorgx = 0, piorgy = 0;                                  //  prior image origin in window
   static double  pscale = 1;                                              //  prior scale
   double         wscale, hscale;
   int            refresh = 0;                                             //  flag, image refesh needed
   PXM            *pxmtemp;

   void sa_show(int flag);                                                 //  show area function, defined later

   if (Fshutdown) return 1;                                                //  shutdown underway
   if (! Fpxm8) return 1;                                                  //  no image

   Dww = drWin->allocation.width;                                          //  (new) drawing window size
   Dhh = drWin->allocation.height;
   if (Dww < 20 || Dhh < 20) return 1;                                     //  too small

   if (mutex_trylock(&pixmaps_lock) != 0) {                                //  lock pixmaps
      Wrepaint++;                                                          //  cannot, come back later
      return 1;
   }
   
   if (Irefresh) {                                                         //  image was updated   v.10.10.2
      refresh++;
      Irefresh = 0;
   }

   if (Dww != pdww || Dhh != pdhh) {                                       //  window size changed
      refresh++;                                                           //  image refresh needed
      SBupdate++;                                                          //  trigger status bar update
      pdww = Dww; 
      pdhh = Dhh;
   }

   if (E3pxm16) {                                                          //  get image size
      Iww = E3ww;
      Ihh = E3hh;                                                          //  edit in progress
   }
   else {
      Iww = Fww;                                                           //  no edit
      Ihh = Fhh;
   }

   if (Fzoom == 0) {                                                       //  scale to fit window
      wscale = 1.0 * Dww / Iww;
      hscale = 1.0 * Dhh / Ihh;
      if (wscale < hscale) Mscale = wscale;                                //  use greatest ww/hh ratio
      else  Mscale = hscale;
      if (Iww < Dww && Ihh < Dhh && ! Fblowup) Mscale = 1.0;               //  small image 1x unless Fblowup
      zoomx = zoomy = 0;
   }
   else Mscale = Fzoom;                                                    //  scale to Fzoom level
   
   if (Mscale != pscale) refresh++;                                        //  scale changed, image refresh needed

   if (Mscale > pscale) {                                                  //  zoom increased
      Iorgx += iww * 0.5 * (1.0 - pscale / Mscale);                        //  keep current image center
      Iorgy += ihh * 0.5 * (1.0 - pscale / Mscale);
   }
   pscale = Mscale;
   
   iww = Dww / Mscale;                                                     //  image space fitting in window
   if (iww > Iww) iww = Iww;
   ihh = Dhh / Mscale;
   if (ihh > Ihh) ihh = Ihh;

   if (zoomx || zoomy) {                                                   //  req. zoom center
      Iorgx = zoomx - 0.5 * iww;                                           //  corresp. image origin
      Iorgy = zoomy - 0.5 * ihh;
      zoomx = zoomy = 0;
   }

   if ((Mxdrag || Mydrag) && ! Mcapture) {                                 //  scroll via mouse drag
      incrx = (Mxdrag - Mxdown) * 1.3 * Iww / iww;                         //  scale
      incry = (Mydrag - Mydown) * 1.3 * Ihh / ihh;
      if (pincrx > 0 && incrx < 0) incrx = 0;                              //  stop bounce at extremes
      if (pincrx < 0 && incrx > 0) incrx = 0;
      pincrx = incrx;
      if (pincry > 0 && incry < 0) incry = 0;
      if (pincry < 0 && incry > 0) incry = 0;
      pincry = incry;
      Iorgx += incrx;                                                      //  new image origin after scroll
      Iorgy += incry;
      Mxdown = Mxdrag + incrx;                                             //  new drag origin
      Mydown = Mydrag + incry;
      Mxdrag = Mydrag = 0;
   }

   if (iww == Iww) {                                                       //  scaled image <= window width
      Iorgx = 0;                                                           //  center image in window
      Dorgx = 0.5 * (Dww - Iww * Mscale);
   }
   else Dorgx = 0;                                                         //  image > window, use entire window

   if (ihh == Ihh) {                                                       //  same for image height
      Iorgy = 0;
      Dorgy = 0.5 * (Dhh - Ihh * Mscale);
   }
   else Dorgy = 0;
   
   if (Iorgx + iww > Iww) Iorgx = Iww - iww;                               //  set limits
   if (Iorgy + ihh > Ihh) Iorgy = Ihh - ihh;
   if (Iorgx < 0) Iorgx = 0;
   if (Iorgy < 0) Iorgy = 0;
   
   if (Iorgx != piorgx || Iorgy != piorgy) {                               //  image origin changed
      refresh++;                                                           //  image refresh needed
      piorgx = Iorgx;
      piorgy = Iorgy;
   }
   
   if (refresh)                                                            //  image refresh is needed  v.10.10.2
   {
      if (E3pxm16) {                                                       //  edit in progress
         PXM_free(Dpxm16);
         Dpxm16 = PXM_copy_area(E3pxm16,Iorgx,Iorgy,iww,ihh);              //  copy E3 image area, PXM-16
         pxmtemp = PXM_convbpc(Dpxm16);                                    //  convert to PXM-8
      }
      else pxmtemp = PXM_copy_area(Fpxm8,Iorgx,Iorgy,iww,ihh);             //  no edit, copy PXM-8 image area

      dww = iww * Mscale;                                                  //  scale to window
      dhh = ihh * Mscale;
      PXM_free(Dpxm8);
      Dpxm8 = PXM_rescale(pxmtemp,dww,dhh);
      PXM_free(pxmtemp);
   }

   wrect.x = wrect.y = 0;                                                  //  stop flicker
   wrect.width = Dww;
   wrect.height = Dhh;
   gdk_window_begin_paint_rect(drWin->window,&wrect);

   gdk_draw_rgb_image(drWin->window, gdkgc, Dorgx, Dorgy, dww, dhh,        //  draw scaled image to window
                      NODITHER, (uint8 *) Dpxm8->bmp, dww*3);

   if (Ntoplines) paint_toplines(1);                                       //  draw line overlays
   if (Ftoparc) paint_toparc(1);                                           //  draw arc overlay
   if (Fgrid[0] || Fgrid[1]) paint_gridlines();                            //  draw grid lines           v.10.3.1
   if (Fshowarea && ! Fpreview) sa_show(1);                                //  draw select area outline  v.9.7

   gdk_window_end_paint(drWin->window);                                    //  release all window updates

   mutex_unlock(&pixmaps_lock);                                            //  unlock pixmaps
   Wpainted++;                                                             //  notify edit function of repaint
   histogram_paint();                                                      //  update brightness histogram if exists
   return 1;
}


/**************************************************************************/

//  update status bar with image data and status
//  called periodically from timer function

void update_statusbar()
{
   static char    text1[200], text2[100];
   int            ww, hh, bpc, scale;
   int            percent_done;
   
   if (! image_file) return;
   
   *text1 = *text2 = 0;
   
   if (E3pxm16) {
      ww = E3ww;
      hh = E3hh;
      bpc = 16;
   }
   else if (Fpxm16) {
      ww = Fww;
      hh = Fhh;
      bpc = 16;
   }
   else {
      ww = Fww;
      hh = Fhh;
      bpc = file_bpc;
   }
   
   snprintf(text1,199,"%dx%dx%d",ww,hh,bpc);                               //  2345x1234x16 (preview) 1.56MB 45%
   if (Fpreview) strcat(text1," (preview)");                               //       ... (turned)
   sprintf(text2," %.2fMB",file_MB);
   strcat(text1,text2);
   scale = int(Mscale * 100 + 0.5);
   sprintf(text2," %d%c",scale,'%');
   strcat(text1,text2);
   if (Fuprighted) strcat(text1," (turned)");

   if (Pundo)                                                              //  edit undo stack depth 
   {
      snprintf(text2,99,"  edits: %d",Pundo);
      strcat(text1,text2);
   }
   
   if (thread_working() || SB_goal)                                        //  v.10.10
      strcat(text1,"  BUSY");

   if (SB_goal)                                                            //  progress monitor       v.9.6
   {
      percent_done = 100 * (1.0 * SB_done / SB_goal);
      snprintf(text2,99," %d%c",percent_done,'%');
      strcat(text1,text2);
   }
   
   if (menu_lock) strcat(text1,"  menu locked");
   
   strcat(text1,"  ");                                                     //  application text       v.10.7
   strcat(text1,SB_text);

   stbar_message(STbar,text1);
   
   return;
}


/**************************************************************************/

//  mouse event function - capture buttons and drag movements

void mouse_event(GtkWidget *, GdkEventButton *event, void *)
{
   void mouse_convert(int &xpos, int &ypos);

   static int     bdtime = 0, butime = 0, mbusy = 0;
   int            button, time, type;

   type = event->type;
   button = event->button;                                                 //  button, 1/3 = left/right
   time = event->time;
   Mxposn = int(event->x);                                                 //  mouse position in window
   Myposn = int(event->y);

   mouse_convert(Mxposn,Myposn);                                           //  convert to image space

   if (type == GDK_MOTION_NOTIFY) {
      if (mbusy) return;                                                   //  discard excess motion events
      mbusy++;
      zmainloop();
      mbusy = 0;
   }

   if (type == GDK_BUTTON_PRESS) {                                         //  button down
      bdtime = time;                                                       //  time of button down
      Mxdown = Mxposn;                                                     //  position at button down time
      Mydown = Myposn;
      if (button) {
         Mdrag++;                                                          //  possible drag start
         Mbutton = button;
      }
      Mxdrag = Mydrag = 0;
   }

   if (type == GDK_BUTTON_RELEASE) {                                       //  button up
      Mxclick = Myclick  = 0;                                              //  reset click status
      butime = time;                                                       //  time of button up
      if (butime - bdtime < 400)                                           //  less than 0.4 secs down
         if (Mxposn == Mxdown && Myposn == Mydown) {                       //       and not moving
            if (Mbutton == 1) LMclick++;                                   //  left mouse click
            if (Mbutton == 3) RMclick++;                                   //  right mouse click
            Mxclick = Mxdown;                                              //  click = button down position
            Myclick = Mydown;
         }
      Mxdown = Mydown = Mxdrag = Mydrag = Mdrag = Mbutton = 0;             //  forget buttons and drag
   }
   
   if (type == GDK_MOTION_NOTIFY && Mdrag) {                               //  drag underway
      Mxdrag = Mxposn;
      Mydrag = Myposn;
   }
   
   if (mouseCBfunc) {                                                      //  pass to handler function
      if (mbusy) return;
      mbusy++;                                                             //  stop re-entrance       v.10.8
      (* mouseCBfunc)();
      mbusy = 0;
      return;
   }

   if (LMclick && ! Mcapture) {                                            //  left click = zoom request
      LMclick = 0;
      zoomx = Mxclick;                                                     //  zoom center = mouse
      zoomy = Myclick;
      m_zoom(null, (char *) "+");
   }

   if (RMclick && ! Mcapture) {                                            //  right click = reset zoom
      RMclick = 0;
      zoomx = zoomy = 0;
      m_zoom(null, (char *) "-");
   }

   if ((Mxdrag || Mydrag) && ! Mcapture) mwpaint();                        //  drag = scroll
   return;
}


//  convert mouse position from window space to image space

void mouse_convert(int &xpos, int &ypos)
{
   xpos = int((xpos - Dorgx) / Mscale + Iorgx + 0.5);
   ypos = int((ypos - Dorgy) / Mscale + Iorgy + 0.5);

   if (xpos < 0) xpos = 0;                                                 //  if outside image put at edge
   if (ypos < 0) ypos = 0;

   if (E3pxm16) { 
      if (xpos >= E3ww) xpos = E3ww-1;
      if (ypos >= E3hh) ypos = E3hh-1;
   }
   else {
      if (xpos >= Fww) xpos = Fww-1;
      if (ypos >= Fhh) ypos = Fhh-1;
   }
   
   return;
}


/**************************************************************************/

//  keyboard event function - some toolbar buttons have KB equivalents
//  GDK key symbols: /usr/include/gtk-2.0/gdk/gdkkeysyms.h

int   KBcontrolkey = 0;
int   KBzmalloclog = 0;

int KBpress(GtkWidget *win, GdkEventKey *event, void *)                    //  prevent propagation of key-press
{                                                                          //    events to toolbar buttons
   KBkey = event->keyval;
   if (KBkey == 65507) KBcontrolkey = 1;                                   //  Ctrl key is pressed
   return 1;
}

int KBrelease(GtkWidget *win, GdkEventKey *event, void *)
{
   int      angle = 0;
   PXM      *pxm;

   KBkey = event->keyval;

   if (KBkeyCBfunc) {                                                      //  pass to handler function
      (* KBkeyCBfunc)();
      KBkey = 0;
      return 1;
   }

   if (KBcapture) return 1;                                                //  let function handle it

   if (KBkey == 65507) KBcontrolkey = 0;                                   //  Ctrk key released
   
   if (KBcontrolkey) {
      if (KBkey == GDK_s) m_save(0,0);                                     //  Ctrl-* shortcuts
      if (KBkey == GDK_S) m_saveas(0,0);  
      if (KBkey == GDK_q) m_quit(0,0);  
      if (KBkey == GDK_Q) m_quit(0,0);
   }

   if (KBkey == GDK_G || KBkey == GDK_g) {                                 //  grid lines on/off      v.10.11
      Fgrid[0] = 1 - Fgrid[0];
      Fgrid[1] = 1 - Fgrid[1];
      mwpaint2();
   }

   if (KBkey == GDK_Left) m_prev(0,0);                                     //  arrow keys  >>  prev/next image
   if (KBkey == GDK_Right) m_next(0,0);  

   if (KBkey == GDK_plus) m_zoom(null, (char *) "+");                      //  +/- keys  >>  zoom in/out
   if (KBkey == GDK_equal) m_zoom(null, (char *) "+");                     //  = key: same as +
   if (KBkey == GDK_minus) m_zoom(null, (char *) "-");
   if (KBkey == GDK_KP_Add) m_zoom(null, (char *) "+");                    //  keypad +
   if (KBkey == GDK_KP_Subtract) m_zoom(null, (char *) "-");               //  keypad -
   
   if (KBkey == GDK_Z) m_zoom(null, (char *) "Z");                         //  Z key: zoom to 100%
   if (KBkey == GDK_z) m_zoom(null, (char *) "Z");

   if (KBkey == GDK_Escape) {                                              //  escape
      if (Fslideshow) m_slideshow(0,0);                                    //  exit slideshow mode
      Fslideshow = 0;
   }

   if (KBkey == GDK_Delete) m_trash(0,0);                                  //  delete  >>  trash

   if (KBkey == GDK_F1)                                                    //  F1  >>  user guide
      showz_userguide(zfuncs::F1_help_topic);                              //  show topic if there, or page 1

   if (! E1pxm16) {                                                        //  if no edit in progress,
      if (KBkey == GDK_R || KBkey == GDK_r) angle = 90;                    //    allow manual rotations
      if (KBkey == GDK_L || KBkey == GDK_l) angle = -90;
      if (angle) {
         mutex_lock(&pixmaps_lock);                                        //  v.10.5
         pxm = PXM_rotate(Fpxm8,angle);
         PXM_free(Fpxm8);
         Fpxm8 = pxm;
         Fww = pxm->ww;
         Fhh = pxm->hh;
         mutex_unlock(&pixmaps_lock);
         mwpaint2();
         Frotated += angle;                                                //  track amount of rotation
         if (Frotated == 360) Frotated = 0;
      }
   }

   if (KBcontrolkey && KBkey == GDK_m) {                                   //  zmalloc() activity log
      KBzmalloclog = 1 - KBzmalloclog;                                     //  Ctrl+M = toggle on/off
      if (KBzmalloclog) zmalloc_log(999999);
      else zmalloc_log(0);
      zmalloc_report();                                                    //  v.10.8
   }
   
   if (KBkey == GDK_T || KBkey == GDK_t) m_trim(0,0);                      //  Key T = Trim function        v.10.3.1
   if (KBkey == GDK_B || KBkey == GDK_b) m_tune(0,0);                      //  Key B = brightness/color     v.10.3.1
   
   KBkey = 0;
   return 1;
}


/**************************************************************************/

//  paint a grid of horizontal and vertical lines

void paint_gridlines()                                                     //  revised  v.10.10.2
{
   int      px, py, stepx, stepy;
   
   if (! Fpxm8) return;                                                    //  no image
   
   stepx = gridspace[0];                                                   //  space between grid lines
   stepy = gridspace[1];

   if (gridcount[0]) stepx = dww / (1 + gridcount[0]);                     //  if line counts specified,    v.10.11
   if (gridcount[1]) stepy = dhh / ( 1 + gridcount[1]);                    //    set spacing accordingly
   
   gdk_gc_set_foreground(gdkgc,&white);                                    //  white lines

   if (Fgrid[0])
      for (px = Dorgx+stepx; px < Dorgx+dww; px += stepx)
         gdk_draw_line(drWin->window,gdkgc,px,Dorgy,px,Dorgy+dhh);

   if (Fgrid[1])
      for (py = Dorgy+stepy; py < Dorgy+dhh; py += stepy)
         gdk_draw_line(drWin->window,gdkgc,Dorgx,py,Dorgx+dww,py);

   gdk_gc_set_foreground(gdkgc,&black);                                    //  adjacent black lines

   if (Fgrid[0]) 
      for (px = Dorgx+stepx+1; px < Dorgx+dww; px += stepx)
         gdk_draw_line(drWin->window,gdkgc,px,Dorgy,px,Dorgy+dhh);

   if (Fgrid[1])
      for (py = Dorgy+stepy+1; py < Dorgy+dhh; py += stepy)
         gdk_draw_line(drWin->window,gdkgc,Dorgx,py,Dorgx+dww,py);

   return;
}


/**************************************************************************/

//  refresh overlay lines on top of image
//  arg = 1:   paint lines only (because window repainted)
//        2:   erase lines and forget them
//        3:   erase old lines, paint new lines, save new in old

void paint_toplines(int arg)
{
   int      ii;

   if (arg == 2 || arg == 3)                                               //  erase old lines
      for (ii = 0; ii < Nptoplines; ii++)
         erase_line(ptoplinex1[ii],ptopliney1[ii],ptoplinex2[ii],ptopliney2[ii]);
   
   if (arg == 1 || arg == 3)                                               //  draw new lines
      for (ii = 0; ii < Ntoplines; ii++)
         draw_line(toplinex1[ii],topliney1[ii],toplinex2[ii],topliney2[ii]);

   if (arg == 2) {
      Nptoplines = Ntoplines = 0;                                          //  forget lines
      return;
   }

   for (ii = 0; ii < Ntoplines; ii++)                                      //  save for future erase
   {
      ptoplinex1[ii] = toplinex1[ii];
      ptopliney1[ii] = topliney1[ii];
      ptoplinex2[ii] = toplinex2[ii];
      ptopliney2[ii] = topliney2[ii];
   }

   Nptoplines = Ntoplines;

   return;
}


/**************************************************************************/

//  refresh overlay arc (circle/ellipse) on top of image
//  arg = 1:   paint arc only (because window repainted)
//        2:   erase arc and forget it
//        3:   erase old arc, paint new arc, save new in old

void paint_toparc(int arg)
{
   int      arcx, arcy, arcw, arch;

   if (ptoparc && (arg == 2 || arg == 3)) {                                //  erase old arc
      arcx = int((ptoparcx-Iorgx) * Mscale + Dorgx + 0.5);                 //  image to window space
      arcy = int((ptoparcy-Iorgy) * Mscale + Dorgy + 0.5);
      arcw = int(ptoparcw * Mscale);
      arch = int(ptoparch * Mscale);

      gdk_gc_set_function(gdkgc,GDK_INVERT);                               //  invert pixels
      gdk_draw_arc(drWin->window,gdkgc,0,arcx,arcy,arcw,arch,0,64*360);    //  draw arc
      gdk_gc_set_function(gdkgc,GDK_COPY);
   }
   
   if (Ftoparc && (arg == 1 || arg == 3)) {                                //  draw new arc
      arcx = int((toparcx-Iorgx) * Mscale + Dorgx + 0.5);                  //  image to window space
      arcy = int((toparcy-Iorgy) * Mscale + Dorgy + 0.5);
      arcw = int(toparcw * Mscale);
      arch = int(toparch * Mscale);

      gdk_gc_set_function(gdkgc,GDK_INVERT);                               //  invert pixels
      gdk_draw_arc(drWin->window,gdkgc,0,arcx,arcy,arcw,arch,0,64*360);    //  draw arc
      gdk_gc_set_function(gdkgc,GDK_COPY);
   }

   if (arg == 2) {
      Ftoparc = ptoparc = 0;                                               //  forget arcs
      return;
   }
   
   ptoparc = Ftoparc;                                                      //  save for future erase
   ptoparcx = toparcx;
   ptoparcy = toparcy;
   ptoparcw = toparcw;
   ptoparch = toparch;

   return;
}


/**************************************************************************/

//  draw red/green line. coordinates are in image space.

void draw_line(int ix1, int iy1, int ix2, int iy2)
{
   void draw_pixel(double pxm, double pym);

   double      x1, y1, x2, y2;   
   double      pxm, pym, slope;
   
   x1 = Mscale * (ix1-Iorgx);                                              //  image to window space
   y1 = Mscale * (iy1-Iorgy);
   x2 = Mscale * (ix2-Iorgx);
   y2 = Mscale * (iy2-Iorgy);
   
   if (abs(y2 - y1) > abs(x2 - x1)) {
      slope = 1.0 * (x2 - x1) / (y2 - y1);
      if (y2 > y1) {
         for (pym = y1; pym <= y2; pym++) {
            pxm = round(x1 + slope * (pym - y1));
            draw_pixel(pxm,pym);
         }
      }
      else {
         for (pym = y1; pym >= y2; pym--) {
            pxm = round(x1 + slope * (pym - y1));
            draw_pixel(pxm,pym);
         }
      }
   }
   else {
      slope = 1.0 * (y2 - y1) / (x2 - x1);
      if (x2 > x1) {
         for (pxm = x1; pxm <= x2; pxm++) {
            pym = round(y1 + slope * (pxm - x1));
            draw_pixel(pxm,pym);
         }
      }
      else {
         for (pxm = x1; pxm >= x2; pxm--) {
            pym = round(y1 + slope * (pxm - x1));
            draw_pixel(pxm,pym);
         }
      }
   }

   gdk_gc_set_foreground(gdkgc,&black);
   return;
}

void draw_pixel(double px, double py)
{
   int            pxn, pyn;
   static int     flip = 0;
   
   pxn = int(px);
   pyn = int(py);
   
   if (pxn < 0 || pxn > dww-1) return;
   if (pyn < 0 || pyn > dhh-1) return;
   
   if (++flip > 4) flip = -5;                                              //  black/white line     v.10.10
   if (flip < 0) gdk_gc_set_foreground(gdkgc,&black);
   else gdk_gc_set_foreground(gdkgc,&white);
   gdk_draw_point(drWin->window, gdkgc, pxn + Dorgx, pyn + Dorgy);
   
   return;
}


//  erase line. refresh line path from Dpxm8 pixels.

void erase_line(int ix1, int iy1, int ix2, int iy2)
{
   void erase_pixel(double pxm, double pym);

   double      x1, y1, x2, y2;   
   double      pxm, pym, slope;
   
   if (! Dpxm8) zappcrash("Dpxm8 = 0");                                    //  v.10.3

   x1 = Mscale * (ix1-Iorgx);
   y1 = Mscale * (iy1-Iorgy);
   x2 = Mscale * (ix2-Iorgx);
   y2 = Mscale * (iy2-Iorgy);
   
   if (abs(y2 - y1) > abs(x2 - x1)) {
      slope = 1.0 * (x2 - x1) / (y2 - y1);
      if (y2 > y1) {
         for (pym = y1; pym <= y2; pym++) {
            pxm = x1 + slope * (pym - y1);
            erase_pixel(pxm,pym);
         }
      }
      else {
         for (pym = y1; pym >= y2; pym--) {
            pxm = x1 + slope * (pym - y1);
            erase_pixel(pxm,pym);
         }
      }
   }
   else {
      slope = 1.0 * (y2 - y1) / (x2 - x1);
      if (x2 > x1) {
         for (pxm = x1; pxm <= x2; pxm++) {
            pym = y1 + slope * (pxm - x1);
            erase_pixel(pxm,pym);
         }
      }
      else {
         for (pxm = x1; pxm >= x2; pxm--) {
            pym = y1 + slope * (pxm - x1);
            erase_pixel(pxm,pym);
         }
      }
   }

   return;
}

void erase_pixel(double px, double py)
{
   int            pxn, pyn;
   
   pxn = int(px);
   pyn = int(py);
   
   if (pxn < 0 || pxn > dww-1) return;
   if (pyn < 0 || pyn > dhh-1) return;

   uint8 *pixel = (uint8 *) Dpxm8->bmp + (pyn * dww + pxn) * 3;
   gdk_draw_rgb_image(drWin->window, gdkgc, pxn + Dorgx, pyn + Dorgy, 
                                 1, 1, NODITHER, pixel, dww * 3);         
   return;
}


/**************************************************************************
   spline curve setup and edit functions                                   //  consolidated    v.9.5

   Initialize no. of curves:  Nspc
   Initialize vert/horz flag: vert[]
   Initialize anchor points:  nap[], apx[][], apy[][]
   Generate data for curve N:  curve_generate(N)
   Set up drawing functions:  curve_init(frame, func)
   Nsp curves will now be shown and edited inside frame.
   Callback func(N) will be called when curve N is edited.
   Get data for curve N: yval = curve_yval(N, xval)

***/

namespace  splinecurve                                                     //  up to 10 spline curves
{
   void     curve_init(GtkWidget *frame, void func(int));                  //  initialize spline curves
   int      curve_adjust(void *, GdkEventButton *event);                   //  curve editing function
   int      curve_draw();                                                  //  curve drawing function
   int      curve_generate(int spc);                                       //  generate data from anchor points
   inline 
   double   curve_yval(int spc, double xval);                              //  get curve y-value
   void     (*spcfunc)(int spc);                                           //  user callback function

   GtkWidget   *drawarea = 0;
   int         Nspc = 0;                                                   //  no. active spline curves
   int         vert[10];                                                   //  curve is vert. (1) or horz. (0)
   int         nap[10];                                                    //  no. anchor points per curve
   double      apx[10][50], apy[10][50];                                   //  up to 50 anchor points per curve
   double      yval[10][1000];                                             //  y-values for x = 0 to 1 by 0.001
};


//  initialize for spline curve editing
//  initial anchor points are pre-loaded into splinecurve variables above

void splinecurve::curve_init(GtkWidget *frame, void func(int))
{
   using namespace splinecurve;

   if (func) spcfunc = func;                                               //  set user callback function
   
   if (frame) {
      drawarea = gtk_drawing_area_new();                                   //  connect mouse edit function
      gtk_container_add(GTK_CONTAINER(frame),drawarea);                    //    to drawing area
      gtk_widget_add_events(drawarea,GDK_BUTTON_PRESS_MASK);
      gtk_widget_add_events(drawarea,GDK_BUTTON_RELEASE_MASK);
      gtk_widget_add_events(drawarea,GDK_BUTTON1_MOTION_MASK); 
      G_SIGNAL(drawarea,"motion-notify-event",curve_adjust,0)
      G_SIGNAL(drawarea,"button-press-event",curve_adjust,0)
      G_SIGNAL(drawarea,"expose-event",curve_draw,0)
   }
   
   return;
}


//  modify anchor points in curve using mouse

int splinecurve::curve_adjust(void *, GdkEventButton *event)
{
   using namespace splinecurve;

   int            ww, hh, kk;
   int            mx, my, button, evtype;
   static int     spc, ap, Fdrag = 0;                                      //  drag continuation logic   v.9.7
   int            minspc, minap;
   double         mxval, myval, cxval, cyval;
   double         dist2, mindist2 = 0;
   
   mx = int(event->x);                                                     //  mouse position in drawing area
   my = int(event->y);
   evtype = event->type;
   button = event->button;
   ww = drawarea->allocation.width;                                        //  drawing area size
   hh = drawarea->allocation.height;
   
   if (evtype == GDK_BUTTON_RELEASE) {
      Fdrag = 0;
      return 0;
   }
   
   if (mx < 0) mx = 0;                                                     //  limit edge excursions
   if (mx > ww) mx = ww;
   if (my < 0) my = 0;
   if (my > hh) my = hh;
   
   if (evtype == GDK_BUTTON_PRESS) Fdrag = 0;                              //  left or right click
   
   if (Fdrag)                                                              //  continuation of drag
   {
      if (vert[spc]) {
         mxval = 1.0 * my / hh;                                            //  mouse position in curve space
         myval = 1.0 * mx / ww;
      }
      else {
         mxval = 1.0 * mx / ww;
         myval = 1.0 * (hh - my) / hh;
      }

      if (ap < nap[spc]-1 && apx[spc][ap+1] - mxval < 0.05) return 0;      //  disallow < 0.05 from next
      if (ap > 0 && mxval - apx[spc][ap-1] < 0.05) return 0;               //    or prior anchor point 
   }
   
   else                                                                    //  mouse click or new drag begin
   {
      minspc = minap = -1;                                                 //  find closest curve/anchor point
      mindist2 = 999999;

      for (spc = 0; spc < Nspc; spc++)                                     //  loop curves
      {
         if (vert[spc]) {
            mxval = 1.0 * my / hh;                                         //  mouse position in curve space
            myval = 1.0 * mx / ww;
         }
         else {
            mxval = 1.0 * mx / ww;
            myval = 1.0 * (hh - my) / hh;
         }

         for (ap = 0; ap < nap[spc]; ap++)                                 //  loop anchor points
         {
            cxval = apx[spc][ap];
            cyval = apy[spc][ap];
            dist2 = (mxval-cxval)*(mxval-cxval) 
                  + (myval-cyval)*(myval-cyval);
            if (dist2 < mindist2) {
               mindist2 = dist2;                                           //  remember closest anchor point
               minspc = spc;
               minap = ap;
            }
         }
      }

      if (minspc < 0) return 0;                                            //  impossible
      spc = minspc;
      ap = minap;
   }
   
   if (evtype == GDK_BUTTON_PRESS && button == 3)                          //  right click, remove anchor point
   {
      if (sqrt(mindist2) > 0.05) return 0;                                 //  not close enough
      if (nap[spc] < 3) return 0;                                          //  < 2 anchor points would remain
      for (kk = ap; kk < nap[spc]-1; kk++) {
         apx[spc][kk] = apx[spc][kk+1];
         apy[spc][kk] = apy[spc][kk+1];
      }
      nap[spc]--;
      curve_generate(spc);                                                 //  regenerate data for modified curve
      curve_draw();                                                        //  regen and redraw all curves
      spcfunc(spc);                                                        //  call user function
      return 0;
   }

   if (! Fdrag)                                                            //  new drag or left click
   {
      if (vert[spc]) {
         mxval = 1.0 * my / hh;                                            //  mouse position in curve space
         myval = 1.0 * mx / ww;
      }
      else {
         mxval = 1.0 * mx / ww;
         myval = 1.0 * (hh - my) / hh;
      }

      if (sqrt(mindist2) < 0.05)                                           //  existing point close enough,
      {                                                                    //    move this anchor point to mouse
         if (ap < nap[spc]-1 && apx[spc][ap+1] - mxval < 0.05) return 0;   //  disallow < 0.05 from next
         if (ap > 0 && mxval - apx[spc][ap-1] < 0.05) return 0;            //    or prior anchor point 
      }
   
      else                                                                 //  none close, add new anchor point
      {
         minspc = -1;                                                      //  find closest curve to mouse
         mindist2 = 999999;

         for (spc = 0; spc < Nspc; spc++)                                  //  loop curves
         {
            if (vert[spc]) {
               mxval = 1.0 * my / hh;                                      //  mouse position in curve space
               myval = 1.0 * mx / ww;
            }
            else {
               mxval = 1.0 * mx / ww;
               myval = 1.0 * (hh - my) / hh;
            }

            cyval = curve_yval(spc,mxval);
            dist2 = fabs(myval - cyval);
            if (dist2 < mindist2) {
               mindist2 = dist2;                                           //  remember closest curve
               minspc = spc;
            }
         }

         if (minspc < 0) return 0;                                         //  impossible
         if (mindist2 > 0.05) return 0;                                    //  not close enough to any curve
         spc = minspc;

         if (nap[spc] > 49) {
            zmessageACK(MWIN,ZTX("Exceed 50 anchor points"));
            return 0;
         }

         if (vert[spc]) {
            mxval = 1.0 * my / hh;                                         //  mouse position in curve space
            myval = 1.0 * mx / ww;
         }
         else {
            mxval = 1.0 * mx / ww;
            myval = 1.0 * (hh - my) / hh;
         }

         for (ap = 0; ap < nap[spc]; ap++)                                 //  find anchor point with next higher x
            if (mxval <= apx[spc][ap]) break;                              //    (ap may come out 0 or nap)

         if (ap < nap[spc] && apx[spc][ap] - mxval < 0.05) return 0;       //  disallow < 0.05 from next
         if (ap > 0 && mxval - apx[spc][ap-1] < 0.05) return 0;            //    or prior anchor point

         for (kk = nap[spc]; kk > ap; kk--) {                              //  make hole for new point
            apx[spc][kk] = apx[spc][kk-1];
            apy[spc][kk] = apy[spc][kk-1];
         }

         nap[spc]++;                                                       //  up point count
      }
   }

   apx[spc][ap] = mxval;                                                   //  new or moved anchor point
   apy[spc][ap] = myval;                                                   //    at mouse position

   curve_generate(spc);                                                    //  regenerate data for modified curve
   curve_draw();                                                           //  regen and redraw all curves
   spcfunc(spc);                                                           //  call user function
   
   if (evtype == GDK_MOTION_NOTIFY) Fdrag = 1;                             //  remember drag is underway
   return 0;
}


//  draw all curves based on current anchor points
//  (refresh all curves in drawing area)

int splinecurve::curve_draw()
{
   using namespace splinecurve;

   int         ww, hh, px, py, qx, qy, spc, ap;
   double      xval, yval;

   ww = drawarea->allocation.width;                                        //  drawing area size
   hh = drawarea->allocation.height;
   if (ww < 50 || hh < 50) return 0;

   gdk_window_clear(drawarea->window);                                     //  clear window
   
   for (spc = 0; spc < Nspc; spc++)
   {
      if (vert[spc])                                                       //  vert. curve
      {
         for (py = 0; py < hh; py++)                                       //  generate all points for curve
         {
            xval = 1.0 * py / hh;                                          //  remove anchor point limits   v.10.9
            yval = curve_yval(spc,xval);
            px = ww * yval + 0.49;                                         //  "almost" round - erratic floating point
            gdk_draw_point(drawarea->window,gdkgc,px,py);                  //    causes "bumps" in a flat curve
         }
         
         for (ap = 0; ap < nap[spc]; ap++)                                 //  draw boxes at anchor points
         {
            xval = apx[spc][ap];
            yval = apy[spc][ap];
            px = ww * yval;
            py = hh * xval;
            for (qx = -2; qx < 3; qx++)
            for (qy = -2; qy < 3; qy++) {
               if (px+qx < 0 || px+qx >= ww) continue;
               if (py+qy < 0 || py+qy >= hh) continue;
               gdk_draw_point(drawarea->window,gdkgc,px+qx,py+qy);
            }
         }
      }
      else                                                                 //  horz. curve
      {
         for (px = 0; px < ww; px++)                                       //  generate all points for curve
         {
            xval = 1.0 * px / ww;                                          //  remove anchor point limits   v.10.9
            yval = curve_yval(spc,xval);
            py = hh - hh * yval + 0.49;                                    //  almost round - erratic FP in Intel CPUs
            gdk_draw_point(drawarea->window,gdkgc,px,py);                  //    causes "bumps" in a flat curve
         }
         
         for (ap = 0; ap < nap[spc]; ap++)                                 //  draw boxes at anchor points
         {
            xval = apx[spc][ap];
            yval = apy[spc][ap];
            px = ww * xval;
            py = hh - hh * yval;
            for (qx = -2; qx < 3; qx++)
            for (qy = -2; qy < 3; qy++) {
               if (px+qx < 0 || px+qx >= ww) continue;
               if (py+qy < 0 || py+qy >= hh) continue;
               gdk_draw_point(drawarea->window,gdkgc,px+qx,py+qy);
            }
         }
      }
   }

   return 0;
}


//  generate all curve data points when anchor points are modified

int splinecurve::curve_generate(int spc)
{
   using namespace splinecurve;

   int      kk, kklo, kkhi;
   double   xval, yvalx;

   spline1(nap[spc],apx[spc],apy[spc]);                                    //  compute curve fitting anchor points

   kklo = 1000 * apx[spc][0] - 30;                                         //  xval range = anchor point range
   if (kklo < 0) kklo = 0;                                                 //    + 0.03 extra below/above      v.9.5
   kkhi = 1000 * apx[spc][nap[spc]-1] + 30;
   if (kkhi > 1000) kkhi = 1000;

   for (kk = 0; kk < 1000; kk++)                                           //  generate all points for curve
   {
      xval = 0.001 * kk;                                                   //  remove anchor point limits   v.10.9
      yvalx = spline2(xval);
      if (yvalx < 0) yvalx = 0;                                            //  yval < 0 not allowed, > 1 OK    v.9.5
      yval[spc][kk] = yvalx;
   }

   return 0;
}


//  retrieve curve data using interpolation of saved table of values

inline double splinecurve::curve_yval(int spc, double xval)
{
   using namespace splinecurve;

   int      ii;
   double   x1, x2, y1, y2, y3;
   
   if (xval <= 0) return yval[spc][0];
   if (xval >= 0.999) return yval[spc][999];
   
   x2 = 1000.0 * xval;
   ii = int(x2);
   x1 = ii;
   y1 = yval[spc][ii];
   y2 = yval[spc][ii+1];
   y3 = y1 + (y2 - y1) * (x2 - x1);
   return y3;
}


/**************************************************************************
      begin menu functions
***************************************************************************/

//  process top-level menu entry

void topmenufunc(GtkWidget *, cchar *menu)
{
   topmenu = (char *) menu;                                                //  remember top-level menu in
   return;                                                                 //    case this is needed somewhere
}


/**************************************************************************
      file menu functions
***************************************************************************/

//  display image gallery (thumbnails) in a separate window

void m_gallery(GtkWidget *, cchar *)
{
   if (image_file)
      image_gallery(image_file,"paint1",0,m_gallery2,mWin);                //  overlay main window    v.10.9
   else {
      char *pp = get_current_dir_name();
      if (pp) {
         image_gallery(pp,"paint1",0,m_gallery2,mWin);                     //  use current directory   v.10.9
         free(pp);
      }
   }

   return;
}


//  clicked thumbnail will call this function

void m_gallery2(char *file)
{
   if (strEqu(file,"F1")) {
      showz_userguide("navigation");
      return;
   }

   f_open(file,1);
   return;
}


/**************************************************************************/

//  open file menu function

void m_open(GtkWidget *, cchar *)
{
   f_open(null,1);
   return;
}


/**************************************************************************/

//  open drag-drop file

void m_open_drag(int x, int y, char *file)
{
   f_open(file,1);
   zfree(file);
   return;
}


/**************************************************************************/

//  open the previous file opened (not the same as toolbar [prev] button)

void m_previous(GtkWidget *, cchar *)                                      //  new   v.10.11
{
   char     *file;
   
   zfuncs::F1_help_topic = "open_previous_file";
   if (Nrecentfiles < 2) return;
   file = recentfiles[1];
   f_open(file,1);
   return;
}


/**************************************************************************/

//  open an image file from the list of recent image files

char     recent_selection[300];

void  m_recent(GtkWidget *, cchar *)
{
   int  recent_dialog_event(zdialog *zd, cchar *event);

   zdialog     *zd;
   int         ii, zstat;
   
   zfuncs::F1_help_topic = "open_recent_file";

   if (! recentfiles[0]) return;

   zd = zdialog_new(ZTX("Open Recent File"),mWin,Bcancel,null);
   zdialog_add_widget(zd,"combo","combo","dialog",0);

   for (ii = 0; ii < Nrecentfiles && recentfiles[ii]; ii++)                //  stuff files into combo box list
      zdialog_cb_app(zd,"combo",recentfiles[ii]);
   zdialog_stuff(zd,"combo",recentfiles[0]);
   
   *recent_selection = 0;
   zdialog_run(zd,recent_dialog_event);                                    //  run dialog
   zstat = zdialog_wait(zd);
   zdialog_free(zd);
   
   if (zstat != 1) return;
   if (*recent_selection != '/') return;

   f_open(recent_selection,1);                                             //  open selected file
   return;
}

int  recent_dialog_event(zdialog *zd, cchar *event)                        //  dialog event function
{
   if (strNeq(event,"combo")) return 0;
   zdialog_fetch(zd,"combo",recent_selection,299);                         //  get user selection
   zdialog_send_response(zd,1);                                            //  complete dialog with OK status
   return 1;
}


/**************************************************************************/

//  add a file to the list of recent files

void  add_recent_file(cchar *file)
{
   int      ii;

   for (ii = 0; ii < Nrecentfiles-1 && recentfiles[ii]; ii++)              //  find file in recent list
      if (strEqu(file,recentfiles[ii])) break;                             //    (or find last entry in list)
   if (recentfiles[ii]) zfree(recentfiles[ii]);                            //  free this slot in list
   for (; ii > 0; ii--) recentfiles[ii] = recentfiles[ii-1];               //  move list down to fill hole
   recentfiles[0] = strdupz(file,0,"recentfile");                          //  current file >> first in list
   return;
}


/**************************************************************************/

//  open a file and initialize PXM bitmap
//  if flock and menu is locked, do nothing and return
//  else open the file and display in main window
//  returns: 0 = OK, +N = error

int f_open(cchar *filespec, int flock)
{
   PXM         *temp8;
   int         fcount;
   
   if (flock && menu_lock) return 1;                                       //  v.10.5
   if (mod_keep()) return 2;                                               //  unsaved edits
   
   if (filespec) filespec = strdupz(filespec,0,"image_file");              //  use passed filespec
   else {
      zfuncs::F1_help_topic = "open_image_file";                           //  v.10.9
      filespec = zgetfile1(ZTX("Open Image File"),"open",image_file);      //  dialog to get filespec
   }
   if (! filespec) return 3;
   if (strlen(filespec) >= maxfcc) return 4;                               //  disallow humongous files     v.10.0
   
   temp8 = f_load(filespec,8);                                             //  load image as PXM-8 pixmap
   if (! temp8) return 5;                                                  //  bad image

   if (flock)
      if (! menulock(1)) return 1;                                         //  lock menu     v.10.8.4

   free_resources();                                                       //  free resources for old image file

   Fpxm8 = temp8;                                                          //  pixmap for main window
   Fww = Fpxm8->ww;
   Fhh = Fpxm8->hh;
   
   if (! image_position(filespec,&fcount)) {                               //  image not in gallery list         v.10.9.1
      image_gallery(filespec,"init");                                      //  generate new gallery list
      image_gallery(0,"paint2");                                           //  refresh gallery window if active
   }

   make_current_file(filespec);                                            //  setup new current file    v.10.9
   zfree((char *) filespec);

   if (zdrename) m_rename(0,0);                                            //  update active rename dialog
   if (zdexifview) m_exif_view(0,0);                                       //  update active EXIF view window     v.10.2
   if (zdexifedit) m_exif_edit(0,0);                                       //  update active EXIF edit window     v.10.11
   if (zdedittags) m_edittags(0,0);                                        //  update active edit tags dialog
   if (zdeditucomm) m_edit_ucomm(0,0);                                     //  update active edit comments dialog  v.10.10

   Fzoom = 0;                                                              //  zoom level = fit window
   zoomx = zoomy = 0;                                                      //  no zoom center

   mwpaint2();                                                             //  refresh main window
   zmainloop();

   if (flock) menulock(0);                                                 //  v.10.8.4
   return 0;
}


//  make the given filespec the current file
//  used by f_open() and f_saveas()

int make_current_file(cchar *filespec)                                     //  v.10.9.1
{
   int      err, cc, fposn, fcount;
   char     *pp, titlebar[250], release[fversioncc+2];
   char     fname[100], fdirk[100];

   if (image_file) zfree(image_file);
   image_file = strdupz(filespec,0,"image_file");                          //  current image filespec

   if (image_dirk) zfree(image_dirk);                                      //  set current directory
   image_dirk = strdupz(image_file,0,"image_dirk");                        //    for new current file
   pp = (char *) strrchr(image_dirk,'/');
   *pp = 0;
   err = chdir(image_dirk);

   pp = (char *) strrchr(image_file,'/');
   strncpy0(fname,pp+1,99);                                                //  file name
   cc = pp - image_file;
   if (cc < 99) strncpy0(fdirk,image_file,cc+2);                           //  get dirk/path/ if short enough
   else {
      strncpy(fdirk,image_file,96);                                        //  or use /dirk/path...
      strcpy(fdirk+95,"...");
   }

   fposn = image_position(image_file,&fcount);                             //  position and count in gallery list

   strncpy0(release,fversion,fversioncc+2);                                //  name and release version     v.10.10.2

   snprintf(titlebar,250,"%s  %d/%d  %s  %s",                              //  window title bar
                     release,fposn,fcount,fname,fdirk);
   gtk_window_set_title(MWIN,titlebar);
   gtk_window_present(MWIN);                                               //  bring main window to front

   add_recent_file(image_file);                                            //  first in recent files list

   return 0;
}


/**************************************************************************/

//  open previous or next file in same gallery as last file opened

void m_prev(GtkWidget *, cchar *)
{
   if (! image_file) return;

   char *pp = image_gallery(image_file,"prev");
   if (! pp) return;
   if (image_file_type(pp) == 2) f_open(pp,1);
   zfree(pp);
   return;
}

void m_next(GtkWidget *, cchar *)
{
   if (! image_file) return;

   char *pp = image_gallery(image_file,"next");
   if (! pp) return;
   if (image_file_type(pp) == 2) f_open(pp,1);
   zfree(pp);
   return;
}


/**************************************************************************/

//  save (modified) image to same file - no confirmation of overwrite.

void m_save(GtkWidget *, cchar *)
{   
   char        *outfile, *pext;
   cchar       *format;
   
   if (! image_file) return;

   zfuncs::F1_help_topic = "save_file";
   
   format = "jpeg";                                                        //  use jpeg unless png or tiff
   if (strEqu(file_type,"png")) format = "png";                            //  png file
   if (strEqu(file_type,"tiff")) format = "tiff-8";                        //  8 or 16 bits/color tiff
   if (file_bpc == 16) format = "tiff-16";

   strcpy(jpeg_quality,def_jpeg_quality);                                  //  default jpeg save quality

   outfile = strdupz(image_file,8,"image_file");                           //  use input file name

   pext = (char *) strrchr(outfile,'/');                                   //  force compatible file extension
   if (pext) pext = (char *) strrchr(pext,'.');                            //    if not already
   if (! pext) pext = outfile + strlen(outfile);
   if (strEqu(format,"jpeg") && ! strcmpv(pext,".jpg",".JPG",".jpeg",".JPEG",null))
      strcpy(pext,".jpg");
   if (strEqu(format,"png") && ! strcmpv(pext,".png",".PNG",null))         //  default .jpg and .tif     v.10.9
      strcpy(pext,".png");
   if (strnEqu(format,"tiff",4) && ! strcmpv(pext,".tif",".TIF",".tiff",".TIFF",null))
      strcpy(pext,".tif");

   f_save(outfile,format);
   zfree(outfile);
   return;
}


/**************************************************************************/

//  save (modified) image to new file, confirm if overwrite existing file.

GtkWidget   *saveas_fchooser;

void m_saveas(GtkWidget *, cchar *)
{
   void  saveas_radiobutt(void *, int button);
   void  saveas_kbkey(void *, GdkEventKey *event);

   GtkWidget      *fdialog, *hbox;
   GtkWidget      *tiff8, *tiff16, *jpeg, *png, *jqlab, *jqval, *newvers;
   char           *outfile = 0, *outfile2 = 0, *pext, *pvers;
   cchar          *format, *delim;
   int            ii, err, yn, nvers, status;
   struct stat    fstat;

   if (! image_file) return;

   zfuncs::F1_help_topic = "save_file";

   fdialog = gtk_dialog_new_with_buttons(ZTX("Save File"),                 //  build file save dialog
                           MWIN, GTK_DIALOG_MODAL,
                           GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, 
                           GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT, null);
   gtk_window_set_default_size(GTK_WINDOW(fdialog),600,500);

   saveas_fchooser = gtk_file_chooser_widget_new(GTK_FILE_CHOOSER_ACTION_SAVE);
   gtk_container_add(GTK_CONTAINER(GTK_DIALOG(fdialog)->vbox),saveas_fchooser);

   //  set_filename() should select the file but does nothing      GTK bug ?     v.10.9.1       ////
   //  set_current_name() puts file name in input box, but does not select the file
   gtk_file_chooser_set_filename(GTK_FILE_CHOOSER(saveas_fchooser),image_file);
   char *fname = strrchr(image_file,'/') + 1;
   gtk_file_chooser_set_current_name(GTK_FILE_CHOOSER(saveas_fchooser),fname);
   
   hbox = gtk_hbox_new(0,0);
   gtk_container_add(GTK_CONTAINER(GTK_DIALOG(fdialog)->vbox),hbox);
   gtk_box_set_child_packing(GTK_BOX(GTK_DIALOG(fdialog)->vbox),hbox,0,0,10,GTK_PACK_END);

   tiff8 = gtk_radio_button_new_with_label(null,"tiff-8");
   tiff16 = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(tiff8),"tiff-16");
   png = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(tiff8),"png");
   jpeg = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(tiff8),"jpeg");
   jqlab = gtk_label_new(ZTX("quality"));
   jqval = gtk_entry_new();
   newvers = gtk_check_button_new_with_label(ZTX("make new version"));

   gtk_entry_set_width_chars(GTK_ENTRY(jqval),2);
   gtk_box_pack_start(GTK_BOX(hbox),tiff8,0,0,5);                          //  (o) tiff8  (o) tiff16  (o) png
   gtk_box_pack_start(GTK_BOX(hbox),tiff16,0,0,5);
   gtk_box_pack_start(GTK_BOX(hbox),png,0,0,5);
   gtk_box_pack_start(GTK_BOX(hbox),jpeg,0,0,5);                           //  (o) jpeg  jpeg quality [__]
   gtk_box_pack_start(GTK_BOX(hbox),jqlab,0,0,0);
   gtk_box_pack_start(GTK_BOX(hbox),jqval,0,0,3);
   gtk_box_pack_end(GTK_BOX(hbox),newvers,0,0,10);                         //  [x] new version
   
   G_SIGNAL(tiff8,"pressed",saveas_radiobutt,0)                            //  connect file type radio buttons
   G_SIGNAL(tiff16,"pressed",saveas_radiobutt,1)                           //    to handler function
   G_SIGNAL(png,"pressed",saveas_radiobutt,2)
   G_SIGNAL(jpeg,"pressed",saveas_radiobutt,3)
   G_SIGNAL(fdialog,"key-release-event",saveas_kbkey,0)

   gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(jpeg),1);                //  set default file type = jpeg
   gtk_entry_set_text(GTK_ENTRY(jqval),def_jpeg_quality);                  //  default jpeg save quality
   
   if (strEqu(file_type,"png"))                                            //  default matches file type
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(png),1);              //    if png or tiff
   if (strEqu(file_type,"tiff"))
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tiff8),1);
   if (file_bpc == 16)
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tiff16),1);

dialog_run:

   gtk_widget_show_all(fdialog);                                           //  run dialog
   status = gtk_dialog_run(GTK_DIALOG(fdialog));
   if (status != GTK_RESPONSE_ACCEPT) {                                    //  user cancelled
      gtk_widget_destroy(fdialog);
      return;
   }

   outfile2 = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(saveas_fchooser));
   if (! outfile2) goto dialog_run;
   outfile = strdupz(outfile2,12,"image_file");
   g_free(outfile2);

   format = "jpeg";
   if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(png)))
      format = "png";
   if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tiff8)))
      format = "tiff-8";
   if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tiff16)))
      format = "tiff-16";

   if (strEqu(format,"jpeg")) {                                            //  get save quality
      ii = atoi(gtk_entry_get_text(GTK_ENTRY(jqval)));
      if (ii < 1 || ii > 100) {
         zmessageACK(MWIN,ZTX("jpeg quality must be 1-100"));
         goto dialog_run;
      }
      sprintf(jpeg_quality,"%d",ii);
   }
   
   nvers = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(newvers));       //  new version checked?  v.10.6

   gtk_widget_destroy(fdialog);                                            //  kill dialog

   pext = strrchr(outfile,'/');                                            //  force compatible file extension
   if (pext) pext = strrchr(pext,'.');                                     //    if not already
   if (! pext) pext = outfile + strlen(outfile);
   if (strEqu(format,"jpeg") && ! strcmpv(pext,".jpg",".JPG",".jpeg",".JPEG",null))
      strcpy(pext,".jpg");
   if (strEqu(format,"png") && ! strcmpv(pext,".png",".PNG",null))         //  default .jpg and .tif     v.10.9
      strcpy(pext,".png");
   if (strnEqu(format,"tiff",4) && ! strcmpv(pext,".tif",".TIF",".tiff",".TIFF",null))
      strcpy(pext,".tif");
   
   if (nvers)                                                              //  new version checked    v.10.6
   {
      *pext = 0;
      pvers = strrchr(outfile,'-');                                        //  get version from filename-nn.ext
      *pext = '.';
      if (pvers != pext - 3) nvers = 1;                                    //  nothing
      else {
         err = convSI(pvers+1,nvers,1,98,&delim);                          //  convert -nn. to number 1-98
         if (err > 1) nvers = 1;                                           //  conversion error
         if (*delim != '.') nvers = 1;                                     //  non-numeric characters
      }
      if (nvers == 1) {                                                    //  no prior version in file name
         pvers = pext;
         pext += 3;
         memmove(pext,pvers,6);                                            //  make space for -nn before .ext
      }

      for (ii = 98; ii > nvers; ii--)                                      //  find highest existing version
      {                                                                    //                         v.10.8
         pvers[0] = '-';
         pvers[1] = ii/10 + '0';                                           //  build filename-nn.ext
         pvers[2] = ii - 10 * (ii/10) + '0';
         err = stat(outfile,&fstat);
         if (! err) break;
      }
      ii++;                                                                //  use next version       v.10.8
      pvers[1] = ii/10 + '0';                                              //  build filename-nn.ext
      pvers[2] = ii - 10 * (ii/10) + '0';
   }

   err = stat(outfile,&fstat);                                             //  check if file exists
   if (! err) {
      yn = zmessageYN(MWIN,ZTX("Overwrite file? \n %s"),outfile);          //  confirm overwrite
      if (! yn) {
         zfree(outfile);
         return;
      }
   }
   
   f_save(outfile,format);                                                 //  save the file

   if (samedirk(outfile,image_file)) {                                     //  if no directory change,      v.10.11
      image_gallery(outfile,"insert");                                     //  insert file in gallery list
      image_gallery(0,"paint2");                                           //  refresh gallery window if active
      make_current_file(outfile);                                          //  make it the current file
   }

   zfree(outfile);
   return;
}


//  set dialog file type from user selection of file type radio button

void saveas_radiobutt(void *, int button)                                  //  v.9.4
{
   cchar    *filetypes[4] = { ".tif", ".tif", ".png", ".jpg" };            //  v.10.9
   char     *filespec;
   char     *filename, *pp;

   filespec = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(saveas_fchooser));
   if (! filespec) return;
   filename = strrchr(filespec,'/');
   if (! filename) return;
   filename = strdupz(filename+1,6,"saveas");
   pp = strrchr(filename,'.');
   if (! pp || strlen(pp) > 5) pp = filename + strlen(filename);           //  bugfix        v.10.9.1
   strcpy(pp,filetypes[button]);
   gtk_file_chooser_set_current_name(GTK_FILE_CHOOSER(saveas_fchooser),filename);
   gtk_widget_show_all(saveas_fchooser);
   zfree(filename);
   g_free(filespec);                                                       //  bugfix, leak    v.10.9.1
   return;
}


//  response function for KB key release

void saveas_kbkey(void *, GdkEventKey *event)                              //  v.10.5
{
   if (event->keyval == GDK_F1) 
      showz_userguide(zfuncs::F1_help_topic);
   return;
}


/**************************************************************************/

//  print current image file
//  do not use f_save() (not valid from here) and simplify.                //  v.10.2

void m_print(GtkWidget *, cchar *)                                         //  use GTK print   v.9.4
{
   int      err;
   char     *printfile;

   zfuncs::F1_help_topic = "print";
   
   if (! image_file) return;                                               //  no image file
   
   printfile = strdupz(get_zuserdir(),20,"printfile");                     //  make temp print file:
   strcat(printfile,"/printfile.jpg");                                     //    ~/.fotoxx/printfile.jpg   v.10.9

   set_cursor(busycursor);                                                 //  set function busy cursor
   
   if (Fpxm16) err = PXBwrite(Fpxm16,printfile);                           //  write image to print file
   else  err = PXBwrite(Fpxm8,printfile);

   set_cursor(0);                                                          //  restore normal cursor

   if (err) {
      zfree(printfile);                                                    //  v.10.3
      return;
   }
   
   print_imagefile(printfile);                                             //  GTK print utility in zfuncs.cc
   zfree(printfile);

   return;
}


/**************************************************************************/

//  Delete image file - move image_file to trash.
//  Use new Linux standard trash function.                                 //  v.10.8
//  If not available, revert to Desktop folder for fotoxx trash.

void m_trash(GtkWidget *, cchar *)
{
   int            err, yn;
   char           trashdir[200];
   struct stat    trstat;
   static int     gstat, trashworks = 1;                                   //  assume it works until otherwise
   GError         *gerror = 0;
   GFile          *gfile = 0;
   cchar          *gerrmess = ZTX("Linux standard trash is not supported. \n"
                                  "Desktop trash folder will be created.");
   
   zfuncs::F1_help_topic = "trash";                                        //  v.10.8

   if (! menulock(1)) return;                                              //  check lock but leave unlocked
   menulock(0);

   if (! image_file) return;                                               //  nothing to trash

   err = stat(image_file,&trstat);                                         //  get file status
   if (err) {
      zmessLogACK(MWIN,strerror(errno));
      return;
   }

   if (! (trstat.st_mode & S_IWUSR)) {                                     //  check permission
      yn = zmessageYN(MWIN,ZTX("Move read-only file to trash?"));
      if (! yn) return;
      trstat.st_mode |= S_IWUSR;
      chmod(image_file,trstat.st_mode);
   }
   
   if (trashworks)                                                         //  Linux standard trash implemented?
   {
      gfile = g_file_new_for_path(image_file);
      gstat = g_file_trash(gfile,0,&gerror);
      if (! gstat) {
         printf("g_file_trash() error: %s \n",gerror->message);
         zmessageACK(null,gerrmess);
         trashworks = 0;                                                   //  no
      }
   }

   if (! trashworks)
   {
      snprintf(trashdir,199,"%s/%s",getenv("HOME"),ftrash);                //  fotoxx trash filespec

      trstat.st_mode = 0;
      err = stat(trashdir,&trstat);
      if (! S_ISDIR(trstat.st_mode)) {
         snprintf(command,ccc,"mkdir -m 0750 \"%s\"",trashdir);
         err = system(command);
         if (err) {
            zmessLogACK(MWIN,ZTX("Cannot create trash folder: %s"),wstrerror(err));
            return;
         }
      }

      snprintf(command,ccc,"cp \"%s\" \"%s\" ",image_file,trashdir);       //  copy image file to trash
      err = system(command);
      if (err) {
         zmessLogACK(MWIN,ZTX("error: %s"),wstrerror(err));
         return;
      }

      snprintf(command,ccc,"rm \"%s\"",image_file);                        //  delete image file
      err = system(command);
      if (err) {
         zmessLogACK(MWIN,ZTX("error: %s"),wstrerror(err));
         return;
      }
   }
   
   delete_tags_index(image_file);                                          //  delete in tags index file
   image_gallery(image_file,"delete");                                     //  delete in gallery list   v.10.9.1
   image_gallery(0,"paint2");                                              //  refresh gallery window if active
   m_next(0,0);                                                            //  step to next file if there

   return;
}


/**************************************************************************/

//  rename menu function
//  activate rename dialog, stuff data from current file
//  dialog remains active when new file is opened

char     rename_old[100] = "";
char     rename_new[100] = "";

void  m_rename(GtkWidget *, cchar *)
{
   int rename_dialog_event(zdialog *zd, cchar *event);
   
   char     *pdir, *pfile, *pext;

   if (! image_file) return;
   
   zfuncs::F1_help_topic = "rename";                                       //  v.10.8

   if (! zdrename)                                                         //  restart dialog
   {
      zdrename = zdialog_new(ZTX("Rename Image File"),mWin,Bcancel,null);
      zdialog_add_widget(zdrename,"hbox","hb1","dialog",0,"space=10");
      zdialog_add_widget(zdrename,"vbox","vb1","hb1",0,"homog|space=5");
      zdialog_add_widget(zdrename,"vbox","vb2","hb1",0,"homog|expand");

      zdialog_add_widget(zdrename,"button","Bold","vb1",ZTX("old name"));
      zdialog_add_widget(zdrename,"button","Bnew","vb1",ZTX("rename to"));
      zdialog_add_widget(zdrename,"button","Bprev","vb1",ZTX("previous"));

      zdialog_add_widget(zdrename,"hbox","hb21","vb2",0);                  //  [ old name ] [ oldname  ]
      zdialog_add_widget(zdrename,"hbox","hb22","vb2",0);                  //  [ new name ] [ newname  ] [+1]
      zdialog_add_widget(zdrename,"hbox","hb23","vb2",0);                  //  [ previous ] [ prevname ]

      zdialog_add_widget(zdrename,"label","Lold","hb21");
      zdialog_add_widget(zdrename,"entry","Enew","hb22",0,"expand|scc=30");
      zdialog_add_widget(zdrename,"button","B+1","hb22"," +1 ","space=5");
      zdialog_add_widget(zdrename,"label","Lprev","hb23");

      zdialog_run(zdrename,rename_dialog_event);                           //  run dialog
   }

   parsefile(image_file,&pdir,&pfile,&pext);
   strncpy0(rename_old,pfile,99);
   strncpy0(rename_new,pfile,99);
   zdialog_stuff(zdrename,"Lold",rename_old);                              //  current file name
   zdialog_stuff(zdrename,"Enew",rename_new);                              //  entered file name

   return;
}


//  dialog event and completion callback function

int rename_dialog_event(zdialog *zd, cchar *event)
{
   char           *pp, *pdir, *pfile, *pext, *pnew, *pold;
   int            nseq, digits, ccp, ccn, ccx, err;
   struct stat    statb;
   
   if (zd->zstat) {                                                        //  complete
      zdialog_free(zd);                                                    //  kill dialog
      zdrename = 0;
      return 0;
   }
   
   if (strEqu(event,"Bold"))                                               //  reset to current file name
      zdialog_stuff(zd,"Enew",rename_old);

   if (strEqu(event,"Bprev")) {                                            //  previous name >> new name
      zdialog_fetch(zd,"Lprev",rename_new,99);
      zdialog_stuff(zd,"Enew",rename_new);
   }

   if (strEqu(event,"B+1"))                                                //  increment sequence number
   {
      zdialog_fetch(zd,"Enew",rename_new,94);                              //  get entered filename
      pp = rename_new + strlen(rename_new);
      digits = 0;
      while (pp[-1] >= '0' && pp[-1] <= '9') {
         pp--;                                                             //  look for NNN in filenameNNN
         digits++;
      }
      nseq = 1 + atoi(pp);                                                 //  NNN + 1
      if (nseq > 9999) nseq = 0;
      if (digits < 2) digits = 2;                                          //  keep digit count if enough
      if (nseq > 99 && digits < 3) digits = 3;                             //  use leading zeros
      if (nseq > 999 && digits < 4) digits = 4;
      snprintf(pp,digits+1,"%0*d",digits,nseq);
      zdialog_stuff(zd,"Enew",rename_new);
   }

   if (strEqu(event,"Bnew"))                                               //  [rename to] button
   {
      if (! menulock(1)) return 0;                                         //  check lock but leave unlocked
      menulock(0);

      parsefile(image_file,&pdir,&pfile,&pext);                            //  existing /directories/file.ext

      zdialog_fetch(zd,"Enew",rename_new,94);                              //  new file name from user

      ccp = strlen(pdir);                                                  //  length of /directories/
      ccn = strlen(rename_new);                                            //  length of file
      if (pext) ccx = strlen(pext);                                        //  length of .ext
      else ccx = 0;

      pnew = zmalloc(ccp + ccn + ccx + 1,"rename.new");                    //  put it all together
      strncpy(pnew,image_file,ccp);                                        //   /directories/file.ext
      strcpy(pnew+ccp,rename_new);
      if (ccx) strcpy(pnew+ccp+ccn,pext);
      
      err = stat(pnew,&statb);                                             //  check if new name exists
      if (! err) {
         zmessageACK(MWIN,ZTX("The target file already exists"));
         zfree(pnew);
         return 0;
      }
      
      snprintf(command,ccc,"cp -p \"%s\" \"%s\"",image_file,pnew);         //  copy to new file   -p v.10.3
      err = system(command);
      if (err) {
         zmessageACK(MWIN,ZTX("Rename failed \n %s"),wstrerror(err));
         printf("command: %s \n",command);
         zfree(pnew);
         return 0;
      }

      zdialog_stuff(zd,"Lprev",rename_new);                                //  set previous name in dialog

      load_filetags(pnew);                                                 //  update tags index file
      update_tags_index(pnew);

      pold = strdupz(image_file,0,"image_file");                           //  save file name to be deleted
      delete_tags_index(pold);                                             //  delete in tags index file   v.9.7

      snprintf(command,ccc,"rm \"%s\"",pold);                              //  delete old file
      err = system(command);
      
      image_gallery(pold,"delete");                                        //  update gallery      v.10.9.1
      image_gallery(pnew,"insert");
      image_gallery(0,"paint2");                                           //  refresh gallery window if active

      zfree(pnew);
      zfree(pold);

      m_next(0,0);                                                         //  move to image_file + 1
   }
   
   return 0;
}


/**************************************************************************/

//  menu function - rename many files at once

char     **batchrename_filelist = 0;
int      batchrename_filecount = 0;

void m_batchrename(GtkWidget *, cchar *)                                   //  new v.9.7
{
   int   batchrename_dialog_event(zdialog *zd, cchar *event);
   
   zdialog  *zd;

   if (zdrename) return;                                                   //  interactive rename is active

   if (mod_keep()) return;                                                 //  unsaved edits
   if (! menulock(1)) return;                                              //  lock menus       v.10.8

   zfuncs::F1_help_topic = "batch_rename";                                 //  v.10.8

   zd = zdialog_new(ZTX("Batch Rename"),mWin,Bproceed,Bcancel,null);

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"button","files","hb1",Bselectfiles,"space=10");
   zdialog_add_widget(zd,"label","labcount","hb1","0 files selected","space=10");
   zdialog_add_widget(zd,"hbox","hb2","dialog","space=5");
   zdialog_add_widget(zd,"label","lab2","hb2",ZTX("new base name"),"space=10");
   zdialog_add_widget(zd,"entry","basename","hb2");
   zdialog_add_widget(zd,"hbox","hb3","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","lab31","hb3",ZTX("starting sequence"),"space=10");
   zdialog_add_widget(zd,"entry","sequence","hb3","100","scc=5");
   zdialog_add_widget(zd,"label","lab32","hb3",ZTX("increment"),"space=10");
   zdialog_add_widget(zd,"entry","increment","hb3","01","scc=3");

   batchrename_filelist = 0;
   batchrename_filecount = 0;
   
   zdialog_run(zd,batchrename_dialog_event);                               //  run dialog
   zdialog_wait(zd);                                                       //  wait for completion
   
   zdialog_free(zd);
   menulock(0);
   return;
}


//  dialog event and completion callback function

int batchrename_dialog_event(zdialog *zd, cchar *event)
{
   char           **flist = batchrename_filelist, countmess[32];
   cchar          *selectmess = ZTX("select files to rename");
   int            ii, err, cc, ccp, ccn, ccx;
   int            sequence, increment, adder;
   char           basename[100], filename[120], *oldfile, *newfile;
   char           *pdir, *pfile, *pext;
   cchar          *errmess = ZTX("base name / sequence / increment not reasonable");
   struct stat    statb;
   
   if (strEqu(event,"files"))                                              //  select images to rename
   {
      if (flist) {                                                         //  free prior list
         for (ii = 0; flist[ii]; ii++) 
            zfree(flist[ii]);
         zfree(flist);
      }

      flist = zgetfileN(selectmess,"openN",image_file);                    //  get file list from user
      batchrename_filelist = flist;

      if (flist)                                                           //  count files in list
         for (ii = 0; flist[ii]; ii++);
      else ii = 0;
      batchrename_filecount = ii;

      sprintf(countmess,"%d files selected",batchrename_filecount);
      zdialog_stuff(zd,"labcount",countmess);
   }
   
   if (! zd->zstat) return 0;                                              //  dialog active

   if (zd->zstat != 1) goto cleanup;                                       //  dialog canceled
   if (! batchrename_filecount) goto cleanup;                              //  no files selected

   zdialog_fetch(zd,"basename",basename,99);
   zdialog_fetch(zd,"sequence",sequence);
   zdialog_fetch(zd,"increment",increment);

   if (strlen(basename) < 1 || sequence < 1 || increment < 1) {
      zd->zstat = 0;                                                       //  keep dialog alive      bugfix v.10.7
      zmessageACK(MWIN,errmess);
      return 0;
   }
   
   write_popup_text("open","Renaming files",500,200,mWin);                 //  status monitor popup window    v.10.3

   for (ii = 0; flist[ii]; ii++)
   {
      oldfile = flist[ii];
      parsefile(oldfile,&pdir,&pfile,&pext);
      ccp = strlen(pdir);
      if (pext) ccx = strlen(pext);
      else ccx = 0;

      adder = sequence + ii * increment;
      snprintf(filename,119,"%s%d",basename,adder);                        //  removed "-" between        v.10.7
      ccn = strlen(filename);

      newfile = zmalloc(ccp + ccn + ccx + 1,"newfile");                    //  construct /path/filename.ext
      strcpy(newfile,pdir);
      strcpy(newfile+ccp,filename);
      if (ccx) strcpy(newfile+ccp+ccn,pext);

      err = stat(newfile,&statb);
      if (! err) {
         snprintf(command,ccc,"%s %s",ZTX("new file already exists:"),newfile);
         write_popup_text("write",command);
         zfree(newfile);
         break;
      }

      cc = snprintf(command,ccc,"cp -p \"%s\" \"%s\"",oldfile,newfile);    //  copy to new file   -p v.10.3
      if (cc >= maxfcc*2) {
         snprintf(command,ccc,"%s %s",ZTX("filespec too long:"),oldfile);
         write_popup_text("write",command);
         zfree(newfile);
         break;
      }

      write_popup_text("write",command);

      err = system(command);
      if (err) {
         snprintf(command,ccc,"%s %s",ZTX("Rename failed:"),wstrerror(err));
         write_popup_text("write",command);
         zfree(newfile);
         break;
      }

      load_filetags(newfile);                                              //  update tags index file
      update_tags_index(newfile);
      zfree(newfile);

      snprintf(command,ccc,"rm \"%s\"",oldfile);                           //  delete old file
      err = system(command);
      delete_tags_index(oldfile);                                          //  remove from tags index
   }

   write_popup_text("write","COMPLETED");
   write_popup_text("close",0);

   image_gallery(image_file,"init");                                       //  update gallery file list
   image_gallery(0,"paint2");                                              //  refresh gallery window if active

cleanup:

   if (batchrename_filecount) {
      for (ii = 0; flist[ii]; ii++) 
         zfree(flist[ii]);
      zfree(flist);
   }

   return 0;
}


/**************************************************************************/

//  forced quit - can cause running function to crash

void m_quit(GtkWidget *, cchar *)
{
   if (mod_keep()) return;                                                 //  unsaved edits
   printf("quit \n");
   Fshutdown++;                                                            //  bugfix  v.10.11
   save_params();                                                          //  save state for next session
   free_resources();                                                       //  delete temp files
   if (KBzmalloclog) zmalloc_report();                                     //  v.10.8
   gtk_main_quit();                                                        //  gone forever
   return;
}


/**************************************************************************
      tools menu functions
***************************************************************************/

//  set new image zoom level or magnification

void m_zoom(GtkWidget *, cchar *menu)
{
   int      ii, iww, ihh, Dww, Dhh;
   char     zoom;
   double   scalew, scaleh, fitscale;
   double   scales[9] = { 0.125, 0.176, 0.25, 0.354, 0.5, 0.71, 1.0, 1.41, 2.0 };
   
   if (strnEqu(menu,"Zoom",4)) zoom = menu[4];                             //  get + or -
   else  zoom = *menu;
   
   Dww = drWin->allocation.width;                                          //  drawing window size
   Dhh = drWin->allocation.height;
   
   if (E3pxm16) {                                                          //  bugfix
      iww = E3ww;
      ihh = E3hh;
   }
   else  {
      iww = Fww;
      ihh = Fhh;
   }

   if (iww > Dww || ihh > Dhh) {                                           //  get window fit scale
      scalew = 1.0 * Dww / iww;
      scaleh = 1.0 * Dhh / ihh;
      if (scalew < scaleh) fitscale = scalew;
      else fitscale = scaleh;
   }
   else fitscale = 1.0;                                                    //  if image < window use 100%
   
   if (zoom == '+') {                                                      //  zoom bigger
      if (! Fzoom) Fzoom = fitscale / 1.2;
      Fzoom = Fzoom * sqrt(2.0);                                           //  new scale: 41% bigger
      for (ii = 0; ii < 9; ii++)
         if (Fzoom < 1.01 * scales[ii]) break;                             //  next higher scale in table
      if (ii == 9) ii = 8;
      Fzoom = scales[ii];
      if (Fzoom < fitscale) Fzoom = 0;                                     //  image < window
   }

   if (zoom == '-') Fzoom = 0;                                             //  zoom to fit window

   if (zoom == 'Z') {
      if (Fzoom != 0) Fzoom = 0;                                           //  toggle 100% and fit window
      else  Fzoom = 1;
   }
   
   if (! Fzoom) zoomx = zoomy = 0;                                         //  no req. zoom center
   
   mwpaint2();                                                             //  refresh window
   return;
}


/**************************************************************************/

//  monitor test function

void m_montest(GtkWidget *, cchar *)
{
   uint8       *pixel;
   int         red, green, blue;
   int         row, col, row1, row2;
   int         ww = 800, hh = 500;
   
   if (mod_keep()) return;                                                 //  unsaved edits
   if (! menulock(1)) return;

   zfuncs::F1_help_topic = "check_monitor";                                //  v.10.8

   mutex_lock(&pixmaps_lock);

   PXM_free(Fpxm8);
   Fpxm8 = PXM_make(ww,hh,8);
   Fww = ww;
   Fhh = hh;
   file_bpc = 8;
   file_MB = 0;

   for (red = 0; red <= 1; red++)
   for (green = 0; green <= 1; green++)
   for (blue = 0; blue <= 1; blue++)
   {
      row1 = 4 * red + 2 * green + blue;                                   //  row 0 to 7
      row1 = row1 * hh / 8;                                                //  stripe, 1/8 of image
      row2 = row1 + hh / 8;
      
      for (row = row1; row < row2; row++)
      for (col = 0; col < ww; col++)
      {
         pixel = (uint8 *) Fpxm8->bmp + (row * ww + col) * 3;
         pixel[0] = red * 256 * col / ww;
         pixel[1] = green * 256 * col / ww;
         pixel[2] = blue * 256 * col / ww;
      }
   }

   Fzoom = 0;                                                              //  scale to window
   gtk_window_set_title(MWIN,"monitor check");
   mutex_unlock(&pixmaps_lock);
   mwpaint2();                                                             //  repaint window
   menulock(0);
   return;
}


/**************************************************************************/

//  create or update brightness histogram graph

GtkWidget      *winhisto = 0;                                              //  brightness histogram window
GtkWidget      *winhistoH = 0;                                             //  histogram drawing area
GtkWidget      *winhistoB = 0;                                             //  brightness band underneath


void m_histogram(GtkWidget *, cchar *)                                     //  menu function
{
   zfuncs::F1_help_topic = "brightness_graph";                             //  v.10.8

   if (! Dpxm8) return;

   if (winhisto) {
      histogram_paint();
      return;
   }

   winhisto = gtk_window_new(GTK_WINDOW_TOPLEVEL);
   gtk_window_set_title(GTK_WINDOW(winhisto),ZTX("Brightness Distribution"));
   gtk_window_set_transient_for(GTK_WINDOW(winhisto),MWIN);
   gtk_window_set_default_size(GTK_WINDOW(winhisto),300,200);
   gtk_window_set_position(GTK_WINDOW(winhisto),GTK_WIN_POS_MOUSE);
   GtkWidget * vbox = gtk_vbox_new(0,3);
   gtk_container_add(GTK_CONTAINER(winhisto),vbox);

   winhistoH = gtk_drawing_area_new();                                     //  histogram drawing area
   gtk_box_pack_start(GTK_BOX(vbox),winhistoH,1,1,3);

   winhistoB = gtk_drawing_area_new();                                     //  brightness scale under histogram
   gtk_box_pack_start(GTK_BOX(vbox),winhistoB,0,0,0);                      //  (progresses from black to white)
   gtk_widget_set_size_request(winhistoB,300,12);                          //                   v.9.3

   G_SIGNAL(winhisto,"destroy",histogram_destroy,0)
   G_SIGNAL(winhistoH,"expose-event",histogram_paint,0)
   G_SIGNAL(winhistoB,"expose-event",histogram_paint,0)

   gtk_widget_show_all(winhisto);
   
   return;
}


void histogram_paint()                                                     //  paint graph window
{
   GdkGC          *gdkgc = 0;                                              //  GDK graphics context
   GdkColor       color;
   GdkColormap    *colormap = 0;

   int         brdist[40], bin, nbins = 40;                                //  increased                 v.9.3
   int         dist_maxbin = 0;
   int         winww, winhh;
   int         px, py, ww, hh, orgx, orgy;
   uint8       *pixel;
   double      bright;
   
   if (! winhisto) return;
   if (! Dpxm8) return;

   gdkgc = gdk_gc_new(winhistoH->window);                                  //  use separate graphics context
   colormap = gtk_widget_get_colormap(winhistoH);                          //  (compensate new GDK bug)

   for (bin = 0; bin < nbins; bin++)                                       //  clear brightness distribution
      brdist[bin] = 0;

   mutex_lock(&pixmaps_lock);

   for (py = 0; py < dhh; py++)                                            //  compute brightness distribution
   for (px = 0; px < dww; px++)                                            //    for image in visible window
   {
      pixel = (uint8 *) Dpxm8->bmp + (py * dww + px) * 3;
      bright = pixbright(pixel);                                           //  0 to 255
      brdist[int(bright / 256 * nbins)]++;                                 //  0 to nbins
   }

   for (bin = 0; bin < nbins; bin++)                                       //  find max. bin
      if (brdist[bin] > dist_maxbin) dist_maxbin = brdist[bin];

   mutex_unlock(&pixmaps_lock);

   gdk_window_clear(winhistoH->window);

   winww = winhistoH->allocation.width;                                    //  drawing window size
   winhh = winhistoH->allocation.height;
   ww = winww / nbins;                                                     //  bin width
   bin = -1;

   color.red = 10000;                                                      //  a pleasing color
   color.green = 25000;
   color.blue = 40000;
   gdk_rgb_find_color(colormap,&color);
   gdk_gc_set_foreground(gdkgc,&color);

   for (px = 0; px < winww; px++)                                          //  draw each bin
   {
      if (px * nbins / winww > bin) {
         bin++;
         hh = int(0.9 * winhh * brdist[bin] / dist_maxbin);
         orgx = px;
         orgy = winhh - hh;
         gdk_draw_rectangle(winhistoH->window,gdkgc,1,orgx,orgy,ww+1,hh);
      }
   }

   hh = winhistoB->allocation.height;

   for (px = 0; px < winww; px++)                                          //  draw brightness scale underneath
   {                                                                       //                      v.9.3
      color.red = color.green = color.blue = 65536 * px / winww;
      gdk_rgb_find_color(colormap,&color);
      gdk_gc_set_foreground(gdkgc,&color);
      gdk_draw_line(winhistoB->window,gdkgc,px,0,px,hh-1);
   }

   gdk_gc_set_foreground(gdkgc,&black);
   return;
}


void histogram_destroy()                                                   //  delete window
{
   if (winhisto) gtk_widget_destroy(winhisto);
   winhisto = 0;
   return;
}


/**************************************************************************/

//  start a new parallel instance of fotoxx

void m_clone(GtkWidget *, cchar *)
{
   int      ignore;

   zfuncs::F1_help_topic = "clone";                                        //  v.10.8

   snprintf(command,ccc,"fotoxx -l %s",zfuncs::zlanguage);                 //  keep language
   if (image_file) strncatv(command,ccc," \"",image_file,"\"",null);
   strcat(command," &");
   ignore = system(command);
   return;
}


/**************************************************************************/

//  enter or leave slideshow mode

void m_slideshow(GtkWidget *, cchar *)
{
   static int     ww, hh;
   zdialog        *zd;
   int            zstat, secs;
   cchar          *esc_message = ZTX("press ESC to exit");
   
   zfuncs::F1_help_topic = "slide_show";                                   //  v.10.8

   if (! Fslideshow)
   {
      if (! menulock(1)) return;                                           //  check lock but leave unlocked
      menulock(0);

      gtk_window_get_size(MWIN,&ww,&hh);
      gtk_widget_hide_all(GTK_WIDGET(mMbar));                              //  enter slide show mode
      gtk_widget_hide_all(GTK_WIDGET(mTbar));                              //  (full screen, no extras)
      gtk_widget_hide_all(GTK_WIDGET(STbar));
      gtk_window_fullscreen(MWIN);

      zd = zdialog_new(ZTX("Time Interval"),mWin,Bproceed,Bcancel,null);
      zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=10");
      zdialog_add_widget(zd,"label","lab1","hb1",ZTX("seconds"));
      zdialog_add_widget(zd,"entry","secs","hb1",0,"scc=5");
      zdialog_add_widget(zd,"label","lab2","dialog",esc_message);          //  v.10.8
      zdialog_stuff(zd,"secs",SS_interval);
      zdialog_run(zd);
      zstat = zdialog_wait(zd);
      zdialog_fetch(zd,"secs",secs);
      zdialog_free(zd);
      SS_interval = secs;                                                  //  interval between slides
      if (zstat != 1) secs = 9999;                                         //  cancel, use huge interval
      SS_timer = get_seconds() + secs + 1;                                 //  set timer for next slide
      Fslideshow = 1;
   }

   else
   {
      gtk_window_unfullscreen(MWIN);                                       //  leave slide show mode
      gtk_window_resize(MWIN,ww,hh);
      gtk_widget_show_all(GTK_WIDGET(mMbar));
      gtk_widget_show_all(GTK_WIDGET(mTbar));
      gtk_widget_show_all(GTK_WIDGET(STbar));
      Fslideshow = 0;
   }

   Fzoom = 0;                                                              //  fit image to window
   Fblowup = Fslideshow;                                                   //  blow-up small images if SS mode
   mwpaint2(); 
   return;
}


/**************************************************************************/

//  show RGB values for pixel at mouse click

zdialog     *zdRGB = 0;                                                    //  show RGB dialog

void m_showRGB(GtkWidget *, cchar *)                                       //  menu function
{
   int   RGB_dialog_event(zdialog *zd, cchar *event);
   void  RGB_mousefunc();
   
   cchar  *rgbmess = ZTX("click on window to show RGB");
   cchar  *format = "Pixel: 0 0   RGB: 0.0  0.0  0.0";

   zfuncs::F1_help_topic = "show_RGB";                                     //  v.10.8

   if (! Fpxm8) return;                                                    //  no image

   mouseCBfunc = RGB_mousefunc;                                            //  connect mouse function
   Mcapture = 1;                                                           //  capture mouse clicks

   if (zdRGB) zdialog_free(zdRGB);

   zdRGB = zdialog_new(ZTX("Show RGB"),mWin,Bcancel,null);
   zdialog_add_widget(zdRGB,"label","lab1","dialog",rgbmess,"space=5");
   zdialog_add_widget(zdRGB,"label","labrgb","dialog",format,"space=5");

   zdialog_run(zdRGB,RGB_dialog_event);                                    //  run dialog
   return;
}


//  dialog event function

int RGB_dialog_event(zdialog *zd, cchar *event)
{
   zdialog_free(zdRGB);                                                    //  kill dialog
   zdRGB = null;
   mouseCBfunc = 0;                                                        //  disconnect mouse
   Mcapture = 0;
   return 0;
}


//  mouse function

void RGB_mousefunc()                                                       //  mouse function
{
   int         px, py;
   double      red, green, blue;
   double      fbright, fred;
   char        text[60];
   uint8       *ppix8;
   uint16      *ppix16;
   
   if (LMclick)                                                            //  left mouse click
   {
      LMclick = 0;
      px = Mxclick;                                                        //  click position
      py = Myclick;
      
      if (E3pxm16) {                                                       //  use current image being edited
         if (px < 0 || px > E3ww-1 ||                                      //  outside image area
             py < 0 || py > E3hh-1) return;
         ppix16 = PXMpix(E3pxm16,px,py);
         red = ppix16[0] / 256.0;
         green = ppix16[1] / 256.0;
         blue = ppix16[2] / 256.0;
         fbright = pixbright(ppix16) / 256.0;
         fred = pixred(ppix16);
      }

      else if (Fpxm16) {                                                   //  use edited image
         if (px < 0 || px > Fww-1 || 
             py < 0 || py > Fhh-1) return;
         ppix16 = PXMpix(Fpxm16,px,py);
         red = ppix16[0] / 256.0;
         green = ppix16[1] / 256.0;
         blue = ppix16[2] / 256.0;
         fbright = pixbright(ppix16) / 256.0;
         fred = pixred(ppix16);
      }

      else  {                                                              //  use 8 bpc image
         if (px < 0 || px > Fww-1 || 
             py < 0 || py > Fhh-1) return;
         ppix8 = (uint8 *) Fpxm8->bmp + (py * Fww + px) * 3;
         red = ppix8[0];
         green = ppix8[1];
         blue = ppix8[2];
         fbright = pixbright(ppix8);
         fred = pixred(ppix8);
      }
      
      snprintf(text,59,"Pixel: %d %d  RGB: %6.3f %6.3f %6.3f",             //  show pixel and RGB colors
                                  px, py, red, green, blue);
      zdialog_stuff(zdRGB,"labrgb",text);
   }
   
   return;
}


/**************************************************************************/

//  toggle grid lines on / off                                             //  v.10.3

void m_gridlines(GtkWidget *, cchar *)
{
   int gridlines_dialog_event(zdialog *zd, cchar *event);

   zdialog     *zd;
   
   zfuncs::F1_help_topic = "grid_lines";                                   //  v.10.8

   zd = zdialog_new(ZTX("Grid Lines"),mWin,Bdone,null);
   
   zdialog_add_widget(zd,"hbox","hb0","dialog",0,"space=10");
   zdialog_add_widget(zd,"vbox","vb1","hb0");
   zdialog_add_widget(zd,"vbox","vb2","hb0");

   zdialog_add_widget(zd,"hbox","hb11","vb1",0,"space=5");
   zdialog_add_widget(zd,"hbox","hb12","vb1",0,"space=5");
   zdialog_add_widget(zd,"hbox","hb21","vb2",0,"space=5");
   zdialog_add_widget(zd,"hbox","hb22","vb2",0,"space=5");

   zdialog_add_widget(zd,"label","lab1x","hb11","x-spacing","space=5");
   zdialog_add_widget(zd,"spin","spacex","hb11","10|200|1|50");
   zdialog_add_widget(zd,"label","lab1y","hb21","y-spacing","space=5");
   zdialog_add_widget(zd,"spin","spacey","hb21","10|200|1|50");

   zdialog_add_widget(zd,"label","lab2x","hb12","x-count","space=5");
   zdialog_add_widget(zd,"spin","countx","hb12","0|100|1|0");
   zdialog_add_widget(zd,"label","lab2y","hb22","y-count","space=5");
   zdialog_add_widget(zd,"spin","county","hb22","0|100|1|0");

   zdialog_add_widget(zd,"hbox","hb3","dialog");
   zdialog_add_widget(zd,"label","laben","hb3","enable","space=5");
   zdialog_add_widget(zd,"check","xgrid","hb3","x-grid","space=10");
   zdialog_add_widget(zd,"check","ygrid","hb3","y-grid");

   zdialog_stuff(zd,"spacex",gridspace[0]);
   zdialog_stuff(zd,"spacey",gridspace[1]);
   zdialog_stuff(zd,"countx",gridcount[0]);
   zdialog_stuff(zd,"county",gridcount[1]);
   zdialog_stuff(zd,"xgrid",Fgrid[0]);
   zdialog_stuff(zd,"ygrid",Fgrid[1]);

   zdialog_run(zd,gridlines_dialog_event);
   zdialog_wait(zd);
   zdialog_free(zd);
   return;
}


//  dialog event function

int gridlines_dialog_event(zdialog *zd, cchar *event)
{   
   if (strEqu(event,"xgrid")) {
      zdialog_fetch(zd,"xgrid",Fgrid[0]);
      mwpaint2();
   }
   
   if (strEqu(event,"ygrid")) {
      zdialog_fetch(zd,"ygrid",Fgrid[1]);
      mwpaint2();
   }
   
   if (strEqu(event,"spacex")) {
      zdialog_fetch(zd,"spacex",gridspace[0]);
      if (Fgrid[0]) mwpaint2();
   }

   if (strEqu(event,"spacey")) {
      zdialog_fetch(zd,"spacey",gridspace[1]);
      if (Fgrid[1]) mwpaint2();
   }

   if (strEqu(event,"countx")) {
      zdialog_fetch(zd,"countx",gridcount[0]);
      if (Fgrid[0]) mwpaint2();
   }

   if (strEqu(event,"county")) {
      zdialog_fetch(zd,"county",gridcount[1]);
      if (Fgrid[1]) mwpaint2();
   }

   return 0;
}


/**************************************************************************/

//  choose or set lens parameters for panoramas

char        *lens4_name[4] = { 0, 0, 0, 0 };                               //  names for 4 lenses
double      lens4_mm[4], lens4_bow[4];                                     //  characteristics for 4 lenses
double      lens_mm, lens_bow;                                             //  current lens characteristics
int         lens_cc = 19;                                                  //  lens name cc limit
int         curr_lens = 0;                                                 //  current lens, 0-3

void  m_lensparms(GtkWidget *, cchar *)
{
   int         ii, zstat, radb;
   char        text[20];
   zdialog     *zd;
   
   zfuncs::F1_help_topic = "lens_parms";                                   //  v.10.8
   
   if (! lens4_name[0] || strEqu(lens4_name[0],"undefined")) {
      lens4_name[0] = strdupz("lens_1",lens_cc);                           //  default lens parameters
      lens4_name[1] = strdupz("lens_2",lens_cc);
      lens4_name[2] = strdupz("lens_3",lens_cc);
      lens4_name[3] = strdupz("lens_4",lens_cc);
      lens4_mm[0] = 30;
      lens4_mm[1] = 40;
      lens4_mm[2] = 50;
      lens4_mm[3] = 60;
      lens4_bow[0] = 0;
      lens4_bow[1] = 0;
      lens4_bow[2] = 0;
      lens4_bow[3] = 0;
      curr_lens = 1;
   }

   zd = zdialog_new(ZTX("Lens Parameters"),mWin,Bapply,Bcancel,null);
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"vbox","vb1","hb1",0,"space=5|homog");            //        Lens    mm    bow
   zdialog_add_widget(zd,"vbox","vb2","hb1",0,"space=5|homog");            //   (o)  name1   30    0.33
   zdialog_add_widget(zd,"vbox","vb3","hb1",0,"space=5|homog");            //   (x)  name2   40    0.22
   zdialog_add_widget(zd,"vbox","vb4","hb1",0,"space=5|homog");            //   (o)  name3   45    0.28
   zdialog_add_widget(zd,"label","space","vb1");                           //   (o)  name4   50    0.44
   zdialog_add_widget(zd,"radio","radb0","vb1",0);                         //
   zdialog_add_widget(zd,"radio","radb1","vb1",0);                         //                   [apply]
   zdialog_add_widget(zd,"radio","radb2","vb1",0);
   zdialog_add_widget(zd,"radio","radb3","vb1",0);
   zdialog_add_widget(zd,"label","lname","vb2",ZTX("lens name"));
   zdialog_add_widget(zd,"entry","name0","vb2","scc=10");
   zdialog_add_widget(zd,"entry","name1","vb2","scc=10");
   zdialog_add_widget(zd,"entry","name2","vb2","scc=10");
   zdialog_add_widget(zd,"entry","name3","vb2","scc=10");
   zdialog_add_widget(zd,"label","lmm","vb3",ZTX("lens mm"));
   zdialog_add_widget(zd,"entry","mm0","vb3","0","scc=5");
   zdialog_add_widget(zd,"entry","mm1","vb3","0","scc=5");
   zdialog_add_widget(zd,"entry","mm2","vb3","0","scc=5");
   zdialog_add_widget(zd,"entry","mm3","vb3","0","scc=5");
   zdialog_add_widget(zd,"label","lbow","vb4",ZTX("lens bow"));
   zdialog_add_widget(zd,"entry","bow0","vb4","0.0","scc=6");
   zdialog_add_widget(zd,"entry","bow1","vb4","0.0","scc=6");
   zdialog_add_widget(zd,"entry","bow2","vb4","0.0","scc=6");
   zdialog_add_widget(zd,"entry","bow3","vb4","0.0","scc=6");
   
   for (ii = 0; ii < 4; ii++)                                              //  stuff lens data into dialog
   {
      snprintf(text,20,"name%d",ii);
      zdialog_stuff(zd,text,lens4_name[ii]);
      snprintf(text,20,"mm%d",ii);
      zdialog_stuff(zd,text,lens4_mm[ii]);
      snprintf(text,20,"bow%d",ii);
      zdialog_stuff(zd,text,lens4_bow[ii]);
   }

   snprintf(text,20,"radb%d",curr_lens);                                   //  current lens = selected
   zdialog_stuff(zd,text,1);
   
   zdialog_run(zd,0);                                                      //  run dialog, get inputs
   zstat = zdialog_wait(zd);

   if (zstat != 1) {
      zdialog_free(zd);                                                    //  canceled
      return;
   }
   
   for (ii = 0; ii < 4; ii++)                                              //  fetch lens data (revisions)
   {
      snprintf(text,20,"name%d",ii);
      zdialog_fetch(zd,text,lens4_name[ii],lens_cc);
      repl_1str(lens4_name[ii],lens4_name[ii]," ","_");                    //  replace blank with _
      snprintf(text,20,"mm%d",ii);
      zdialog_fetch(zd,text,lens4_mm[ii]);
      snprintf(text,20,"bow%d",ii);
      zdialog_fetch(zd,text,lens4_bow[ii]);
      snprintf(text,20,"radb%d",ii);                                       //  detect which is selected
      zdialog_fetch(zd,text,radb);
      if (radb) curr_lens = ii;
   }
   
   zdialog_free(zd);
   zd = 0;
   return;
}


/**************************************************************************/

//  set GUI language

void  m_lang(GtkWidget *, cchar *)                                         //  overhauled   v.10.1
{
   zdialog     *zd;
   int         ii, cc, err, val, zstat;
   char        lang1[8], *pp;

   cchar  *langs[12] = { "en English", "de German", "es Spanish",          //  english first
                         "fr French", "gl Galacian", "it Italian", 
                         "nl Dutch", "pt Portuguese", "ru_RU Russian", 
                         "sv Swedish", "zh_CN Chinese", null  };
   
   cchar  *title = ZTX("Available Translations");

   zfuncs::F1_help_topic = "language";                                     //  v.10.8

   zd = zdialog_new(ZTX("Set Language"),mWin,Bapply,Bcancel,null);
   zdialog_add_widget(zd,"label","title","dialog",title,"space=10");
   zdialog_add_widget(zd,"hbox","hb1","dialog");
   zdialog_add_widget(zd,"vbox","vb1","hb1");

   for (ii = 0; langs[ii]; ii++)                                           //  make radio button per language
      zdialog_add_widget(zd,"radio",langs[ii],"vb1",langs[ii]);

   cc = strlen(zfuncs::zlanguage);                                         //  current language
   for (ii = 0; langs[ii]; ii++)                                           //  match on lc_RC
      if (strnEqu(zfuncs::zlanguage,langs[ii],cc)) break;
   if (! langs[ii]) 
      for (ii = 0; langs[ii]; ii++)                                        //  failed, match on lc alone
         if (strnEqu(zfuncs::zlanguage,langs[ii],2)) break;
   if (! langs[ii]) ii = 0;                                                //  failed, default english
   zdialog_stuff(zd,langs[ii],1);

   zdialog_resize(zd,200,0);
   zdialog_run(zd);                                                        //  run dialog
   zstat = zdialog_wait(zd);

   for (ii = 0; langs[ii]; ii++) {                                         //  get active radio button
      zdialog_fetch(zd,langs[ii],val);
      if (val) break;
   }

   zdialog_free(zd);                                                       //  kill dialog

   if (zstat != 1) return;                                                 //  user cancel
   if (! val) return;                                                      //  no selection
   
   strncpy0(lang1,langs[ii],8);
   pp = strchr(lang1,' ');                                                 //  isolate lc_RC part
   *pp = 0;

   sprintf(command,"fotoxx -l %s &",lang1);                                //  start new fotoxx with language
   err = system(command);

   m_quit(0,0);                                                            //  exit
   return;
}


/**************************************************************************/

//  create desktop icon / launcher

void  m_launcher(GtkWidget *, cchar *) 
{
   zfuncs::F1_help_topic = "launcher";                                     //  v.10.8
   zmake_launcher("Graphics","Image Editor");
   return;
}


/**************************************************************************/

//  convert multiple RAW files to tiff

char     **raw_files;
int      raw_filecount;
char     *raw_outfile = 0;

void  m_conv_raw(GtkWidget *, cchar *)                                     //  multi-threaded      v.10.1.1
{
   void * conv_raw_wthread(void *arg);

   int      ii;

   zfuncs::F1_help_topic = "convert_RAW";                                  //  v.10.8

   if (! Fufraw) {
      zmessageACK(MWIN,ZTX("Program ufraw-batch is required"));
      return;
   }

   if (mod_keep()) return;                                                 //  unsaved edits
   if (! menulock(1)) return;                                              //  v.10.5

   raw_files = zgetfileN(ZTX("Select RAW files to convert"),"openN",image_dirk);
   if (! raw_files) goto rawdone;
   
   for (ii = 0; raw_files[ii]; ii++);                                      //  count selected files
   raw_filecount = ii;

   set_cursor(busycursor);                                                 //  set function busy cursor

   for (ii = 0; ii < 2 && ii < raw_filecount; ii++)                        //  start 2 worker threads    v.10.9
      start_wt(conv_raw_wthread,&wtnx[ii]);

   write_popup_text("open","Converting RAW files",500,200,mWin);           //  status monitor popup window
   write_popup_text("write","converting ...");                             //  v.10.9
   
   while (true)
   {
      zmainloop();                                                         //  v.10.9
      zsleep(0.1);
      if (! raw_outfile && ! wthreads_busy) break;                         //  finished
      if (! raw_outfile) continue;                                         //  wait for more output

      write_popup_text("write",raw_outfile);                               //  write output to log window

      if (*raw_outfile != '*')                                             //  if not an error message, show
         f_open(raw_outfile,0);                                            //    converted file in main window

      zfree(raw_outfile);
      raw_outfile = 0;
   }

   write_popup_text("write","COMPLETED");
   write_popup_text("close",0);                                            //  v.10.8

   set_cursor(0);                                                          //  restore normal cursor

rawdone:

   if (raw_files) {
      for (ii = 0; raw_files[ii]; ii++) 
         zfree(raw_files[ii]);
      zfree(raw_files);
   }

   image_gallery(image_file,"init");                                       //  update gallery file list
   image_gallery(0,"paint2");                                              //  refresh gallery window if active

   menulock(0);
   return;
}


//  convert RAW working threads

void * conv_raw_wthread(void *arg)
{
   int      index = *((int *) (arg));
   char     *rawfile, *outfile, *pp;
   int      ii, err, ccc = maxfcc*2;
   char     command[maxfcc*2];

   for (ii = index; ii < raw_filecount; ii += 2)                           //  2 threads     v.10.9
   {
      rawfile = raw_files[ii];
      outfile = strdupz(rawfile,5,"raw_out");
      pp = strrchr(rawfile,'.');
      pp = outfile + (pp - rawfile);
      strcpy(pp,".tiff");
      
      //  try new ufraw command format first, then old command if error    //  --wb  v.10.9

      snprintf(command,ccc,"ufraw-batch --wb=camera --out-type=tiff --out-depth=16"
                  " --overwrite --output=\"%s\" \"%s\" ",outfile, rawfile);
      err = system(command);

      if (err) {
         snprintf(command,ccc,"ufraw-batch  --wb=camera --out-type=tiff16"
                  " --overwrite --output=\"%s\" \"%s\" ",outfile, rawfile);
         err = system(command);
      }
      
      while (raw_outfile) zsleep(0.1);                                     //  wait for prior      bugfix v.10.9

      if (err) {
         zfree(outfile);
         raw_outfile = zmalloc(1000,"raw_errmess");
         snprintf(raw_outfile,999,"*** %s %s",wstrerror(err),rawfile);
      }
      else raw_outfile = outfile;
   }

   exit_wt();                                                              //  exit thread
   return 0;                                                               //  not executed, stop gcc warning
}


/**************************************************************************/

//  burn images to CD/DVD  

void  m_burn(GtkWidget *, cchar *)
{
   int      ii, cc, err;
   char     **filelist, *imagefile, *bcommand;

   if (! menulock(1)) return;                                              //  lock menus
   
   zfuncs::F1_help_topic = "burn";                                         //  v.10.8

   filelist = image_gallery_getfiles(0);                                   //  get list of files to burn   v.10.9
   if (! filelist) {
      menulock(0);
      return;
   }
   
   cc = 0;
   for (ii = 0; filelist[ii]; ii++)                                        //  get memory for brasero command line
      cc += strlen(filelist[ii]) + 4;

   bcommand = zmalloc(cc+20,"brasero-command");
   strcpy(bcommand,"brasero");
   cc = strlen(bcommand);

   for (ii = 0; filelist[ii]; ii++)                                        //  copy files to command line
   {
      imagefile = filelist[ii];
      strcpy(bcommand+cc," \"");
      cc += 2;
      strcpy(bcommand+cc,imagefile);
      cc += strlen(imagefile);
      strcpy(bcommand+cc,"\"");
      cc += 1;
      zfree(imagefile);
   }
   
   zfree(filelist);

   strcat(bcommand," &");                                                  //  brasero command in background
   err = system(bcommand);
   if (err) zmessageACK(0,"error: %s",wstrerror(err));
   zfree(bcommand);

   menulock(0);
   return;
}


/**************************************************************************/

//  menu function
//  resize selected images and send to preferred e-mail program

int         emailsize[2] = { 1000, 600 };                                  //  email resize width, height

void m_email(GtkWidget *, cchar *)                                         //  new v.10.10
{
   int   email_dialog_event(zdialog *zd, cchar *event);

   zdialog     *zd;                                                        //  email dialog

   zfuncs::F1_help_topic = "e-mail";

   if (mod_keep()) return;                                                 //  unsaved edits
   if (! menulock(1)) return;                                              //  lock menus

   //  [select files]  N files selected
   //  max. width   [____]
   //  max. height  [____]
   //
   //                   [proceed]  [cancel]

   zd = zdialog_new(ZTX("E-mail Images"),mWin,Bproceed,Bcancel,null);
   zdialog_add_widget(zd,"hbox","hbf","dialog",0,"space=5");
   zdialog_add_widget(zd,"button","files","hbf",Bselectfiles,"space=5");
   zdialog_add_widget(zd,"label","fcount","hbf","0 files selected","space=10");
   zdialog_add_widget(zd,"hbox","hbwh","dialog","space=10");
   zdialog_add_widget(zd,"vbox","vbwh1","hbwh",0,"homog|space=5");
   zdialog_add_widget(zd,"vbox","vbwh2","hbwh",0,"homog|space=5");
   zdialog_add_widget(zd,"label","labw","vbwh1",ZTX("max. width"));
   zdialog_add_widget(zd,"label","labh","vbwh1",ZTX("max. height"));
   zdialog_add_widget(zd,"entry","maxw","vbwh2","1000","scc=5");
   zdialog_add_widget(zd,"entry","maxh","vbwh2","700","scc=5");

   zdialog_stuff(zd,"maxw",emailsize[0]);
   zdialog_stuff(zd,"maxh",emailsize[1]);
   
   zdialog_run(zd,email_dialog_event);                                     //  run dialog
   zdialog_wait(zd);                                                       //  wait for completion

   menulock(0);
   return;
}


//  dialog event and completion callback function

int email_dialog_event(zdialog *zd, cchar *event)
{
   static char    **flist = 0;
   char           countmess[32], tempdir[100], sequence[8], ecommand[2000];
   int            maxw, maxh, err, fcc, ww, hh;
   int            ii, jj, kk;
   char           *pfile, *pext, *oldfile, *newfile;
   double         scale, wscale, hscale;
   PXM            *pxmout;

   if (strEqu(event,"files"))                                              //  select images to resize
   {
      if (flist) {                                                         //  free prior list
         for (ii = 0; flist[ii]; ii++) 
            zfree(flist[ii]);
         zfree(flist);
      }
      
      flist = image_gallery_getfiles(0);                                   //  get list of files to resize   v.10.9

      if (flist)                                                           //  count files selected
         for (ii = 0; flist[ii]; ii++);
      else ii = 0;

      sprintf(countmess,"%d files selected",ii);                           //  update dialog
      zdialog_stuff(zd,"fcount",countmess);
   }
   
   if (! zd->zstat) return 0;                                              //  dialog still busy

   if (zd->zstat != 1) goto cleanup;                                       //  dialog canceled
   
   if (! flist) {                                                          //  no files selected
      zd->zstat = 0;                                                       //  keep dialog active
      return 0;
   }

   zdialog_fetch(zd,"maxw",maxw);                                          //  new max width
   zdialog_fetch(zd,"maxh",maxh);                                          //  new max height

   if (maxw < 20 || maxh < 20) {
      zmessageACK(MWIN,ZTX("max. size %d x %d is not reasonable"),maxw,maxh);
      zd->zstat = 0;
      return 0;
   }

   *tempdir = 0;                                                           //  temp directory: /tmp/<user>/fotoxx
   strncatv(tempdir,99,"/tmp/",cuserid(null),"/fotoxx",null);

   sprintf(command,"mkdir -p %s",tempdir);                                 //  (re)create directory
   err = system(command);
   if (err) {
      zmessageACK(null,"%s",strerror(err));
      goto cleanup;
   }

   sprintf(command,"rm -f %s/*.jpg",tempdir);                              //  purge prior file list
   err = system(command);
   
   write_popup_text("open","preparing files",500,200,mWin);                //  status monitor popup window

   strcpy(ecommand,"xdg-email");                                           //  e-mail command: xdg-email

   for (ii = 0; flist[ii]; ii++)                                           //  loop selected files
   {
      oldfile = flist[ii];
      pfile = strrchr(oldfile,'/');                                        //  /filename.ext
      if (! pfile) continue;

      fcc = strlen(pfile);
      newfile = strdupz(tempdir,fcc+10,"emailresize.new");                 //  add extra space for .nn and .ext
      strcat(newfile,pfile);                                               //  /tmp/<user>/fotoxx/filename.ext
      pfile = strrchr(newfile,'/');

      for (jj = kk = 1; pfile[jj]; jj++, kk++)                             //  remove special characters
      {                                                                    //   (xdg-email substitutes %nn
         pfile[kk] = pfile[jj];                                            //     and thunderbird falls over)
         if (pfile[jj] < 0) continue;                                      //  UTF-8 multibyte characters OK
         if (pfile[jj] == '.') continue;                                   //  periods OK
         if (pfile[jj] >= 'a' && pfile[jj] <= 'z') continue;               //  keep a-z, A-Z, 0-9
         if (pfile[jj] >= 'A' && pfile[jj] <= 'Z') continue;
         if (pfile[jj] >= '0' && pfile[jj] <= '9') continue;
         kk--;                                                             //  omit others
      }
      
      pfile[kk] = 0;

      pext = strrchr(pfile,'.');                                           //  find .ext if there is one
      if (! pext) pext = pfile + strlen(pfile);
      if (strlen(pext) > 5) pext = pext + strlen(pext);
      
      sprintf(sequence,".%d",ii+1);                                        //  replace with sequence number
      strcpy(pext,sequence);                                               //  filename.nn

      pext = pext + strlen(pext);                                          //  add .jpg extension
      strcpy(pext,".jpg");                                                 //  filename.nn.jpg
      
      write_popup_text("write",newfile);                                   //  log progress

      err = f_open(oldfile,0);                                             //  read old file
      if (err) {
         printf("f_open err %d: %s \n",err,oldfile);
         continue;
      }

      wscale = hscale = 1.0;
      if (Fww > maxw) wscale = 1.0 * maxw / Fww;                           //  compute new size
      if (Fhh > maxh) hscale = 1.0 * maxh / Fhh;
      if (wscale < hscale) scale = wscale;
      else scale = hscale;
      ww = Fww * scale;
      hh = Fhh * scale;
      pxmout = PXM_rescale(Fpxm8,ww,hh);                                   //  rescale file

      PXBwrite(pxmout,newfile);                                            //  write to new file
      PXM_free(pxmout);
      
      err = strncatv(ecommand,1990," --attach ","\"",newfile,"\"",null);   //  --attach /tmp/<user>/fotoxx/filename.jpf
      zfree(newfile);
      
      if (err) {
         zmessageACK(MWIN,ZTX("too many files"));
         goto cleanup;
      }
   }

   write_popup_text("write","COMPLETED");
   write_popup_text("close",0);

   emailsize[0] = maxw;                                                    //  save preferred size
   emailsize[1] = maxh;
   
   strcat(ecommand," &");
   printf("system: %s \n",ecommand);
   err = system(ecommand);

cleanup:

   if (flist) {                                                            //  free memory
      for (ii = 0; flist[ii]; ii++) 
         zfree(flist[ii]);
      zfree(flist);
      flist = 0;
   }

   zdialog_free(zd);                                                       //  kill dialog
   return 0;
}


/**************************************************************************/

//  Rebuild thumbnail files.
//  Process all image files within given top-level directory.

void m_rebuild_thumbs(GtkWidget *, const char *)
{
   int rebuild_thumbs_dialog_event(zdialog *zd, cchar *event);

   zdialog     *zd;
   char        *filespec1, *filespec2, *thumbdirk;
   char        *subdirk, *pp, stbartext[200];
   int         zstat, contx = 0, err, fcount = 0;

   zfuncs::F1_help_topic = "rebuild_thumbs";                               //  v.10.8

   zd = zdialog_new(ZTX("Rebuild Thumbnails"),mWin,Bproceed,Bcancel,null);
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","lab1","hb1",ZTX("Top Image Directory:"));
   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=1");
   zdialog_add_widget(zd,"entry","topdirk","hb2","??","scc=40|space=5");
   zdialog_add_widget(zd,"button","browse","hb2",Bbrowse,"space=5");
   
   if (topdirk) zdialog_stuff(zd,"topdirk",topdirk);

   zdialog_run(zd,rebuild_thumbs_dialog_event);
   zstat = zdialog_wait(zd);

   if (zstat != 1) {
      zdialog_free(zd);
      return;
   }

   pp = zmalloc(maxfcc);   
   zdialog_fetch(zd,"topdirk",pp,maxfcc-1);
   if (topdirk) zfree(topdirk);
   topdirk = strdupz(pp,0,"topdirk");
   zfree(pp);

   zdialog_free(zd);

   if (! menulock(1)) return;
   
   set_cursor(busycursor);                                                 //  v.10.7

   snprintf(command,ccc,"find \"%s\" -type d",topdirk);                    //  find directories under top directory
   contx = 0;

   while ((subdirk = command_output(contx,command)))
   {
      pp = strrchr(subdirk,'/');
      if (pp && strEqu(pp,"/.thumbnails"))                                 //  /.../.thumbnails/ directory
      {
         snprintf(command,ccc,"rm -R \"%s\"",subdirk);                     //  delete thumbnails
         err = system(command);
      }

      zfree(subdirk);
   }

   snprintf(command,ccc,"find \"%s\" -type d",topdirk);                    //  find directories under top directory
   contx = 0;

   while ((subdirk = command_output(contx,command)))
   {
      image_gallery(subdirk,"init");                                       //  get all image files in directory
      filespec1 = image_gallery(subdirk,"first");
      thumbdirk = 0;

      while (filespec1)
      {
         if (image_file_type(filespec1) == 2) 
         {
            if (! thumbdirk) {                                             //  when first image file found,
               thumbdirk = strdupz(subdirk,20);                            //    create .thumbnails if needed
               strcat(thumbdirk,"/.thumbnails");
               err = mkdir(thumbdirk,0751);
               zfree(thumbdirk);
            }

            pp = image_thumbfile(filespec1);                               //  find/update/create thumbnail
            if (pp) zfree(pp);

            snprintf(stbartext,199,"%5d %s",++fcount,filespec1);           //  update status bar
            stbar_message(STbar,stbartext);
         }
         
         filespec2 = image_gallery(filespec1,"next");                      //  next image file
         zfree(filespec1);
         filespec1 = filespec2;
      }

      zfree(subdirk);
   }

   menulock(0);
   set_cursor(0);
   zmessageACK(MWIN,ZTX("completed"));
   return;
}


//  dialog event function - file chooser for top image directory

int rebuild_thumbs_dialog_event(zdialog *zd, cchar *event)
{
   char     *pp;
   
   if (! strEqu(event,"browse")) return 0;
   pp = zgetfile1(ZTX("Select top image directory"),"folder",topdirk);     //  get topmost directory
   if (! pp) return 0;
   zdialog_stuff(zd,"topdirk",pp);
   zfree(pp);
   return 0;
}


/**************************************************************************/

//  Rebuild tags index file.
//  Process all image files within given top-level directory.
//  Works incrementally and is very fast after the first run.              //  overhauled v.10.0

struct tag_index_rec {
   char        *file;                                                      //  image filespec
   char        imagedate[12], filedate[16];                                //  image date, file date
   char        *tags;                                                      //  image tags
   char        stars;                                                      //  image rating, '0' to '5'
   char        *comms;                                                     //  image comments
   char        update;                                                     //  flag, update is needed
   char        fill[2];
};
tag_index_rec   *xrec_old, *xrec_new;


void m_rebuild_tags(GtkWidget *widget, cchar *menu)                        //  menu function
{
   int  rebuild_tags_dialog_event(zdialog *zd, cchar *event);
   int  rebuild_tags_compare(cchar *rec1, cchar *rec2);

   zdialog        *zd;
   FILE           *fid;
   int            zstat, err, contx, fcount;
   char           tagsbuff[tagrecl], stbartext[200];
   char           *subdirk, *pp, **ppv;
   char           *filespec1, *filespec2;
   char           *imagedate, *imagetags, *imagestars, *imagecomms;
   cchar          *exifkeys[4] = { exif_date_key, exif_tags_key, exif_rating_key, exif_comment_key };
   cchar          *ppc, *mode;
   int            Nold, Nnew, orec, nrec, comp;
   struct tm      bdt;
   struct stat    statb;

   zfuncs::F1_help_topic = "rebuild_tags";                                 //  v.10.8

   if (! menu) {                                                           //  called from convert_tags
      mode = "full";                                                       //    OBSOLETE
      goto proceed;
   }

   if (! Fexiftool) {                                                      //  exiftool is required
      zmessageACK(MWIN,Bexiftoolmissing);
      return;
   }
   
   zd = zdialog_new(ZTX("Rebuild Tags Index"),mWin,Bproceed,Bcancel,null);
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","lab1","hb1",ZTX("Top Image Directory:"));
   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=1");
   zdialog_add_widget(zd,"entry","topdirk","hb2","??","scc=40|space=5");
   zdialog_add_widget(zd,"button","browse","hb2",Bbrowse,"space=5");
   zdialog_add_widget(zd,"hbox","hb3","dialog",0,"space=5");
   zdialog_add_widget(zd,"radio","rb1","hb3",ZTX("full rebuild"),"space=5");
   zdialog_add_widget(zd,"radio","rb2","hb3",ZTX("incremental"),"space=10");
   
   if (topdirk) zdialog_stuff(zd,"topdirk",topdirk);
   zdialog_stuff(zd,"rb1",0);
   zdialog_stuff(zd,"rb2",1);

   zdialog_run(zd,rebuild_tags_dialog_event);
   zstat = zdialog_wait(zd);

   if (zstat != 1) {
      zdialog_free(zd);
      return;
   }

   pp = zmalloc(maxfcc);   
   zdialog_fetch(zd,"topdirk",pp,maxfcc-1);
   if (topdirk) zfree(topdirk);
   topdirk = strdupz(pp,0,"topdirk");
   zfree(pp);
   
   zdialog_fetch(zd,"rb1",zstat);
   if (zstat) mode = "full";
   else mode = "incr";

   zdialog_free(zd);

proceed:

   if (! menulock(1)) return;

   set_cursor(busycursor);                                                 //  set function busy cursor
   
   xrec_old = (tag_index_rec *) zmalloc(max_images * sizeof(tag_index_rec),"xrec_old");
   xrec_new = (tag_index_rec *) zmalloc(max_images * sizeof(tag_index_rec),"xrec_new");
   
//  if incremental mode, read current tags index file and build "old list" of tags

   if (strEqu(mode,"incr"))                                                //  bugfix   v.10.9
      fid = fopen(tags_index_file,"r");
   else fid = 0;
   
   orec = Nold = 0;

   if (fid) 
   {
      while (true)                                                         //  read all tags index records
      {
         pp = fgets_trim(tagsbuff,tagrecl,fid);                            //  next record
         if (! pp) break;

         if (strnEqu(pp,"file: ",6))                                       //  start new file entry
         {
            if (++Nold == max_images) 
               zappcrash("more than %d image files: %d",max_images);
            
            orec = Nold - 1;

            xrec_old[orec].file = strdupz(pp+6,0,"xrec_old.file");

            xrec_old[orec].imagedate[0] = 0;
            xrec_old[orec].filedate[0] = 0;
            xrec_old[orec].tags = 0;
            xrec_old[orec].stars = '0';
            xrec_old[orec].comms = 0;
            xrec_old[orec].update = '0';
         }
         
         if (strnEqu(pp,"date: ",6))
         {
            ppc = strField(pp,' ',2);
            if (ppc) strncpy0(xrec_old[orec].imagedate,ppc,12);
            ppc = strField(pp,' ',3);
            if (ppc) strncpy0(xrec_old[orec].filedate,ppc,16);
         }
         
         if (strnEqu(pp,"tags: ",6))
            xrec_old[orec].tags = strdupz(pp+6,0,"xrec_old.tags");
         
         if (strnEqu(pp,"stars: ",7))
            xrec_old[orec].stars = *(pp+7);

         if (strnEqu(pp,"comms: ",7))
            xrec_old[orec].comms = strdupz(pp+7,0,"xrec_old.comms");
      }
      
      fclose(fid);
   }

   printf("%d current tag records found \n",Nold);
   
//  find all image files and create "new list" with no tags

   snprintf(command,ccc,"find \"%s\" -type d",topdirk);                    //  find directories under top directory
   contx = 0;
   Nnew = nrec = 0;

   while ((subdirk = command_output(contx,command)))
   {
      pp = strrchr(subdirk,'/');
      if (pp && strEqu(pp,"/.thumbnails"))                                 //  skip ./thumbnails directories
      {
         zfree(subdirk);
         continue;
      }

      image_gallery(subdirk,"init");                                       //  get all image files in directory
      filespec1 = image_gallery(subdirk,"first");

      while (filespec1)
      {
         if (image_file_type(filespec1) == 2)                              //  construct new tag record
         {
            err = stat(filespec1,&statb);
            if (err) continue;
            if (++Nnew == max_images) 
               zappcrash("more than %d image files: %d",max_images);
            nrec = Nnew - 1;
            xrec_new[nrec].file = strdupz(filespec1,0,"xrec_new.file");    //  filespec
            xrec_new[nrec].imagedate[0] = 0;                               //  image date = empty
            gmtime_r(&statb.st_mtime,&bdt);                                //  file date = yyyymmddhhmmss
            snprintf(xrec_new[nrec].filedate,16,"%04d%02d%02d%02d%02d%02d",
                     bdt.tm_year + 1900, bdt.tm_mon + 1, bdt.tm_mday,
                     bdt.tm_hour, bdt.tm_min, bdt.tm_sec);
            xrec_new[nrec].tags = 0;                                       //  tags = empty
            xrec_new[nrec].stars = '0';                                    //  stars = '0'
            xrec_new[nrec].comms = 0;                                      //  comments = empty
         }
      
         filespec2 = image_gallery(filespec1,"next");                      //  next image file
         zfree(filespec1);
         filespec1 = filespec2;
      }

      zfree(subdirk);
   }
   
   printf("found %d image files \n",Nnew);

//  merge and compare lists
//  if filespecs match and have the same date, then old tags are OK

   HeapSort((char *) xrec_old,sizeof(tag_index_rec),Nold,rebuild_tags_compare);
   HeapSort((char *) xrec_new,sizeof(tag_index_rec),Nnew,rebuild_tags_compare);

   for (orec = nrec = 0; nrec < Nnew; )
   {
      xrec_new[nrec].update = '1';                                         //  assume update is needed

      if (orec == Nold) comp = +1;
      else comp = strcmp(xrec_old[orec].file, xrec_new[nrec].file);        //  compare filespecs
      
      if (comp > 0) nrec++;
      else if (comp < 0) orec++;
      else                                                                 //  matching filespecs
      {
         if (strEqu(xrec_new[nrec].filedate, xrec_old[orec].filedate))     //  and matching file dates
         {                                                                 //  copy data from old to new
            strncpy0(xrec_new[nrec].imagedate,xrec_old[orec].imagedate,12);
            xrec_new[nrec].tags = xrec_old[orec].tags;
            xrec_old[orec].tags = 0;
            xrec_new[nrec].stars = xrec_old[orec].stars;
            xrec_new[nrec].comms = xrec_old[orec].comms;
            xrec_old[orec].comms = 0;
            xrec_new[nrec].update = '0';                                   //  update is not needed
         }
         nrec++;
         orec++;
      }
   }

   for (orec = 0; orec < Nold; orec++)                                     //  release old list memory
   {
      zfree(xrec_old[orec].file);
      if (xrec_old[orec].tags) zfree(xrec_old[orec].tags);
   }

   zfree(xrec_old);
   xrec_old = 0;

//  process entries needing update in new list
//  get updated tags from image file EXIF data

   for (fcount = nrec = 0; nrec < Nnew; nrec++)
   {
      if (xrec_new[nrec].update == '0') continue;

      ppv = exif_get(xrec_new[nrec].file,exifkeys,4);                      //  get exif data
      imagedate = ppv[0];
      imagetags = ppv[1];
      imagestars = ppv[2];
      imagecomms = ppv[3];
      
      if (imagedate && strlen(imagedate)) {                                //  image date
         if (strlen(imagedate) > 9) imagedate[10] = 0;                     //  truncate to yyyy:mm:dd
         strcpy(xrec_new[nrec].imagedate,imagedate);
      }
      else strcpy(xrec_new[nrec].imagedate,"null");
      
      if (imagetags && strlen(imagetags))                                  //  image tags
         xrec_new[nrec].tags = strdupz(imagetags,0,"xrec_new.tags");
      else xrec_new[nrec].tags = strdupz("null; ",0,"xrec_new.tags");

      if (imagestars && strlen(imagestars))                                //  image rating
         xrec_new[nrec].stars = *imagestars;

      if (imagecomms && strlen(imagecomms))                                //  image comments
         xrec_new[nrec].comms = strdupz(imagecomms,0,"xrec_new.comms");
      else xrec_new[nrec].comms = strdupz("null ",0,"xrec_new.comms");

      if (imagedate) zfree(imagedate);
      if (imagetags) zfree(imagetags);
      if (imagestars) zfree(imagestars);
      if (imagecomms) zfree(imagecomms);

      snprintf(stbartext,199,"%5d %s",++fcount,xrec_new[nrec].file);       //  update status bar
      stbar_message(STbar,stbartext);
   }

   fid = fopen(tags_index_file,"w");                                       //  write new tags index file
   if (! fid) zappcrash("cannot write tags file");                         //    with merged data

   for (nrec = 0; nrec < Nnew; nrec++)
   {
      fprintf(fid,"file: %s""\n",xrec_new[nrec].file);
      fprintf(fid,"date: %s  %s""\n", xrec_new[nrec].imagedate, xrec_new[nrec].filedate);
      fprintf(fid,"tags: %s""\n",xrec_new[nrec].tags);
      fprintf(fid,"stars: %c""\n",xrec_new[nrec].stars);
      fprintf(fid,"comms: %s""\n",xrec_new[nrec].comms);
      fprintf(fid,"\n");
   }
   
   fclose(fid);
   
   for (nrec = 0; nrec < Nnew; nrec++)                                     //  release new list memory
   {
      zfree(xrec_new[nrec].file);
      if (xrec_new[nrec].tags) zfree(xrec_new[nrec].tags);
   }
   zfree(xrec_new);
   xrec_new = 0;

   set_cursor(0);                                                          //  restore normal cursor
   menulock(0);
   
   printf("rebuild tags completed \n");
   zmessageACK(MWIN,ZTX("completed"));
   return;
}


//  dialog event function - file chooser for top image directory

int rebuild_tags_dialog_event(zdialog *zd, cchar *event)
{
   char     *pp;
   
   if (! strEqu(event,"browse")) return 0;
   pp = zgetfile1(ZTX("Select top image directory"),"folder",topdirk);     //  get topmost directory
   if (! pp) return 0;
   zdialog_stuff(zd,"topdirk",pp);
   zfree(pp);
   return 0;
}


//  sort compare function - compare tag record filespecs and return
//  filespec is at the start of the record
//   <0 | 0 | >0   for   file1 < | == | > file2

int rebuild_tags_compare(cchar *rec1, cchar *rec2)
{
   char * file1 = ((tag_index_rec *) rec1)->file;
   char * file2 = ((tag_index_rec *) rec2)->file;
   return strcmp(file1,file2);
}


/**************************************************************************/

//  Convert EXIF tags data to new standard:
//     move all tags from EXIF:UserComment to IPTC:Keywords
//     tag delimiter is semicolon instead of blank
//     special tag "stars=N" is converted to EXIF:Rating
//     EXIF:UserComment is not erased, so this function can be repeated

void m_convert_tags(GtkWidget *, cchar *)                                  //  OBSOLETE   v.10.0
{
   int   convert_tags_dialog_event(zdialog *zd, cchar *event);

   zdialog        *zd;
   int            zstat, yn, err, contx, ii, cc, tcc, Nfiles = 0;
   char           *ppv, stbartext[200], *subdirk, sstars[4];
   char           *oldtags, *newtags, **olddata;
   cchar          *filespec1, *filespec2;
   cchar          *oldkeys[1] = { exif_tags_key_old };
   cchar          *newkeys[2] = { exif_tags_key, exif_rating_key };
   cchar          *newdata[2], *pp;
   struct stat    statb;
   
   zfuncs::F1_help_topic = "convert_tags";                                 //  v.10.8

   if (! Fexiftool) {                                                      //  exiftool is required
      zmessageACK(MWIN,Bexiftoolmissing);
      return;
   }

   err = stat(tags_index_file,&statb);                                     //  is new tags index present?
   if (! err) {
      yn = zmessageYN(MWIN,ZTX("New tags file already exists! \n"          //  yes, maybe a user error   v.10.8
                               "Proceed anyway?"));
      if (! yn) return;
   }

   yn = zmessageYN(MWIN,ZTX("Convert tags to new standard now? \n"
                            "Are your image files backed-up?"));
   if (! yn) return;

   zd = zdialog_new(ZTX("Convert tags to new standard"),mWin,Bproceed,Bcancel,null);
   zdialog_add_widget(zd,"hbox","hb1","dialog");
   zdialog_add_widget(zd,"label","lab1","hb1",ZTX("Top Image Directory:"),"space=5");
   zdialog_add_widget(zd,"hbox","hb2","dialog");
   zdialog_add_widget(zd,"entry","topdirk","hb2","??","scc=40|space=5");
   zdialog_add_widget(zd,"button","browse","hb2",Bbrowse,"space=5");
   
   if (topdirk) zdialog_stuff(zd,"topdirk",topdirk);

   zdialog_run(zd,convert_tags_dialog_event);
   zstat = zdialog_wait(zd);

   if (zstat != 1) {
      zdialog_free(zd);
      return;
   }

   ppv = zmalloc(maxfcc);                                                  //  get top image directory from dialog
   zdialog_fetch(zd,"topdirk",ppv,maxfcc-1);
   if (topdirk) zfree(topdirk);
   topdirk = strdupz(ppv,0,"topdirk");
   zfree(ppv);
   
   zdialog_free(zd);

   if (! menulock(1)) return;

   remove(tags_index_file);                                                //  tags index is now invalid

   set_cursor(busycursor);                                                 //  set function busy cursor

   snprintf(command,ccc,"find \"%s\" -type d",topdirk);                    //  find all image files
   contx = 0;

   while ((subdirk = command_output(contx,command)))                       //  find directories under top directory
   {
      pp = strrchr(subdirk,'/');
      if (pp && strEqu(pp,"/.thumbnails")) {                               //  ignore .thumbnails
         zfree(subdirk);
         continue;
      }

      image_gallery(subdirk,"init");                                       //  get all image files in directory
      filespec1 = image_gallery(subdirk,"first");

      while (filespec1)
      {
         if (image_file_type(filespec1) != 2) goto nextfile;

         olddata = exif_get(filespec1,oldkeys,1);                          //  get exif old tags data
         oldtags = olddata[0];
         if (! oldtags) {
            Nfiles++;                                                      //  none
            goto nextfile;
         }
         
         newtags = zmalloc(tagFcc,"newtags");
         *newtags = 0;
         tcc = 0;
         strcpy(sstars,"0");
         
         for (ii = 1; ; ii++)
         {
            pp = strField(oldtags,' ',ii);                                 //  blank delimited tags
            if (! pp) break;

            if (strnEqu(pp,"stars=",6)) {                                  //  special tag, save rating '0' to '5'
               sstars[0] = pp[6];
               continue;
            }
            
            if (strnEqu(pp,"null",4)) continue;                            //  v.10.5

            cc = strlen(pp);                                               //  check enough room
            if (cc + 2 + tcc >= tagFcc) continue;

            strcpy(newtags + tcc, pp);                                     //  copy tag to newtags list
            strcpy(newtags + tcc + cc, "; ");                              //  add semicolon delimiter and blank
            tcc += cc + 2;
         }
         
         newdata[0] = newtags;                                             //  add Keywords tag
         newdata[1] = sstars;                                              //  add Ratings tag
         exif_put(filespec1,newkeys,newdata,2);                            //  update EXIF data

         Nfiles++;                                                         //  count files processed
         zfree(oldtags);
         zfree(newtags);

      nextfile:
         snprintf(stbartext,199,"%5d %s",Nfiles,filespec1);                //  update status bar
         stbar_message(STbar,stbartext);

         filespec2 = image_gallery(filespec1,"next");                      //  next image file
         zfree((char *) filespec1);
         filespec1 = filespec2;
      }

      zfree(subdirk);
   }
   
   printf("processed %d image files \n",Nfiles);
   set_cursor(0);                                                          //  restore normal cursor
   menulock(0);
                                                                           //  create new tags index 
   zmessageACK(MWIN,ZTX("new tags index will now be created"));
   m_rebuild_tags(0,0);

   return;
}


//  dialog event function - file chooser for top image directory

int convert_tags_dialog_event(zdialog *zd, cchar *event)
{
   char     *pp;

   if (! strEqu(event,"browse")) return 0;
   pp = zgetfile1(ZTX("Select top image directory"),"folder",topdirk);     //  get topmost directory
   if (! pp) return 0;
   zdialog_stuff(zd,"topdirk",pp);
   zfree(pp);
   return 0;
}


/**************************************************************************
      Image EXIF functions
***************************************************************************/

//  menu function and popup dialog to show EXIF data
//  window is updated when navigating to another image

void m_exif_view(GtkWidget *, cchar *menu)                                 //  overhauled    v.10.2
{
   int   exif_view_dialog_event(zdialog *zd, cchar *event);

   static int     basic = 1;
   char           *buff;
   int            err, contx = 0;
   GtkWidget      *widget;

   zfuncs::F1_help_topic = "view_exif_data";                               //  v.10.8

   if (! Fexiftool) {                                                      //  exiftool is required
      zmessageACK(MWIN,Bexiftoolmissing);
      return;
   }
   
   if (! image_file) return;

   if (menu) {                                                             //  remember basic or full request
      basic = 0;
      if (strEqu(menu,ZTX("Basic EXIF data"))) basic = 1;
   }

   if (! zdexifview)                                                       //  popup dialog if not already
   {
      zdexifview = zdialog_new(ZTX("EXIF data"),mWin,Bcancel,null);
      zdialog_add_widget(zdexifview,"scrwin","scroll","dialog",0,"expand");
      zdialog_add_widget(zdexifview,"edit","exifdata","scroll",0,"expand");
      zdialog_resize(zdexifview,500,400);
      zdialog_run(zdexifview,exif_view_dialog_event);
   }

   if (basic)                                                              //  set up exiftool command
      snprintf(command,ccc,"exiftool -common -%s -%s -%s -%s -%s \"%s\" ",
         exif_tags_key, exif_rating_key, exif_log_key, 
         exif_comment_key, exif_focal_length_key, image_file);
   else 
      snprintf(command,ccc,"exiftool -e \"%s\" ",image_file);
      
   widget = zdialog_widget(zdexifview,"exifdata");                         //  widget for output
   wclear(widget);

   while ((buff = command_output(contx,command)))                          //  run command, output into window
      wprintf(widget,"%s\n",buff);

   err = command_status(contx);                                            //  free resources

   return;
}


//  dialog event and completion callback function

int exif_view_dialog_event(zdialog *zd, cchar *event)                      //  kill dialog
{
   if (! zd->zstat) return 0;
   zdialog_free(zdexifview);
   zdexifview = null;
   return 0;
}


/**************************************************************************/

//  edit EXIF data - add or change specified EXIF key

void m_exif_edit(GtkWidget *, cchar *menu)                                 //  new v.10.2
{
   char        keyname[40], keydata[1000];
   cchar       *pp1[1];
   char        **pp2;

   int   exif_edit_dialog_event(zdialog *zd, cchar *event);
   
   zfuncs::F1_help_topic = "edit_exif_data";                               //  v.10.8

   if (! Fexiftool) {                                                      //  exiftool is required
      zmessageACK(MWIN,Bexiftoolmissing);
      return;
   }
   
   if (! image_file) return;

   if (! zdexifedit)                                                       //  popup dialog if not already   v.10.8
   {   
      zdexifedit = zdialog_new(ZTX("Edit EXIF data"),mWin,Bfetch,Bsave,Bcancel,null);
      zdialog_add_widget(zdexifedit,"vbox","hb1","dialog");
      zdialog_add_widget(zdexifedit,"hbox","hbkey","dialog",0,"space=5");
      zdialog_add_widget(zdexifedit,"hbox","hbdata","dialog",0,"space=5");
      zdialog_add_widget(zdexifedit,"label","labkey","hbkey","key name");
      zdialog_add_widget(zdexifedit,"entry","keyname","hbkey",0,"scc=20");
      zdialog_add_widget(zdexifedit,"label","labdata","hbdata","key value");
      zdialog_add_widget(zdexifedit,"entry","keydata","hbdata",0,"expand");
      zdialog_run(zdexifedit,exif_edit_dialog_event);
   }

   zdialog_fetch(zdexifedit,"keyname",keyname,40);                         //  get key name from dialog
   strCompress(keyname);                                                   //  v.10.8

   if (*keyname)                                                           //  update live dialog   v.10.8
   {
      pp1[0] = keyname;                                                    //  look for key data 
      pp2 = exif_get(image_file,pp1,1);
      if (pp2[0]) {
         strncpy0(keydata,pp2[0],999);
         zfree(pp2[0]);
      }
      else *keydata = 0;
      zdialog_stuff(zdexifedit,"keydata",keydata);                         //  stuff into dialog
   }

   return;
}


//  dialog event and completion callback function

int  exif_edit_dialog_event(zdialog *zd, cchar *event)
{
   char        keyname[40], keydata[100];
   cchar       *pp1[1], *pp2[1];
   char        **pp3;
   int         err;
   
   if (! zd->zstat) return 0;
   
   zdialog_fetch(zd,"keyname",keyname,40);
   zdialog_fetch(zd,"keydata",keydata,100);
   strCompress(keyname);                                                   //  remove blanks    v.10.8

   if (zd->zstat == 1)                                                     //  fetch
   {
      zd->zstat = 0;
      pp1[0] = keyname;
      pp3 = exif_get(image_file,pp1,1);
      if (pp3[0]) {
         strncpy0(keydata,pp3[0],99);
         zfree(pp3[0]);
      }
      else *keydata = 0;
      zdialog_stuff(zd,"keydata",keydata);
   }

   else if (zd->zstat == 2)                                                //  save
   {
      zd->zstat = 0;
      pp1[0] = keyname;
      pp2[0] = keydata;
      err = exif_put(image_file,pp1,pp2,1);
      if (err) zmessageACK(0,"error: %s",strerror(err));
      if (zdexifview) m_exif_view(0,0);                                    //  update exif view if active
   }

   else zdialog_free(zd);                                                  //  other
   return 1;
}


/**************************************************************************/

//  delete EXIF data, specific key or all data

void m_exif_delete(GtkWidget *, cchar *menu)                               //  new v.10.2
{
   int   exif_delete_dialog_event(zdialog *zd, cchar *event);
   
   zfuncs::F1_help_topic = "delete_exif_data";                             //  v.10.8

   if (! Fexiftool) {                                                      //  exiftool is required
      zmessageACK(MWIN,Bexiftoolmissing);
      return;
   }
   
   if (! image_file) return;

   zdialog *zd = zdialog_new(ZTX("Delete EXIF data"),mWin,Bapply,Bcancel,null);
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"radio","kall","hb1",ZTX("All"),"space=5");
   zdialog_add_widget(zd,"radio","key1","hb1",ZTX("One Key:"));
   zdialog_add_widget(zd,"entry","keyname","hb1",0,"scc=20");
   zdialog_stuff(zd,"key1",1);

   zdialog_run(zd,exif_delete_dialog_event);
   return;
}


//  dialog event and completion callback function

int  exif_delete_dialog_event(zdialog *zd, cchar *event)
{
   int         kall, key1, err;
   char        keyname[40];
   
   if (! zd->zstat) return 0;

   if (zd->zstat != 1) {                                                   //  canceled
      zdialog_free(zd);
      return 1;
   }
   
   zd->zstat = 0;                                                          //  dialog remains active

   zdialog_fetch(zd,"kall",kall);
   zdialog_fetch(zd,"key1",key1);
   zdialog_fetch(zd,"keyname",keyname,40);
   strCompress(keyname);                                                   //  v.10.8
                                                                           //  bugfix - quotes around file  v.10.3
   if (kall)
      snprintf(command,ccc,"exiftool -m -q -overwrite_original -all=  \"%s\"",image_file);
   else if (key1)
      snprintf(command,ccc,"exiftool -m -q -overwrite_original -%s=  \"%s\"",keyname,image_file);
   else return 1;
   
   err = system(command);
   if (err) zmessageACK(0,"%s",wstrerror(err));
   
   if (zdexifview) m_exif_view(0,0);                                       //  update exif view if active

   return 1;
}


/**************************************************************************/

//  get EXIF metadata for given image file and EXIF key(s)
//  returns array of pointers to corresponding key values
//  if a key is missing, corresponding pointer is null
//  returned strings belong to caller, are subject for zfree()
//  up to 9 keynames may be requested per call
//  EXIF command: 
//       exiftool -keyname1 -keyname2 ... "file"
//  command output: 
//       keyname1: keyvalue1
//       keyname2: keyvalue2
//       ...

char ** exif_get(cchar *file, cchar **keys, int nkeys)
{
   char           *pp;
   static char    *rettext[10];
   int            contx = 0, err, ii;
   uint           cc;
   
   if (nkeys < 1 || nkeys > 9) appcrash("exif_get nkeys: %d",nkeys);

   strcpy(command,"exiftool -m -q -S -fast");
   
   for (ii = 0; ii < nkeys; ii++)
   {
      rettext[ii] = null;
      strncatv(command,ccc," -",keys[ii],null);                            //  "-exif:" replaced with "-"  v.10.0
   }

   strncatv(command,ccc," \"",file,"\"",null);

   while (true)
   {
      pp = command_output(contx,command);
      if (! pp) break;

      for (ii = 0; ii < nkeys; ii++)
      {
         cc = strlen(keys[ii]);
         if (strncasecmp(pp,keys[ii],cc) == 0)                             //  ignore case       bugfix v.10.2
            if (strlen(pp) > cc+2) 
               rettext[ii] = strdupz(pp+cc+2,0,"exif_data");               //  check not empty
      }

      zfree(pp);
   }
   
   err = command_status(contx);
   if (err) printf(" exif_get: %s \n",wstrerror(err));                     //  v.10.8

   return rettext;
}


/**************************************************************************/

//  create or change EXIF metadata for given image file and key(s)
//  up to 9 keys may be processed
//  EXIF command: 
//    exiftool -overwrite_original -keyname="keyvalue" ... "file"

int exif_put(cchar *file, cchar **keys, cchar **text, int nkeys)
{
   int      ii, err;
   
   if (nkeys < 1 || nkeys > 9) zappcrash("exif_put nkeys: %d",nkeys);
   
   for (ii = 0; ii < nkeys; ii++)
      if (! text[ii]) text[ii] = "";                                       //  if null pointer use empty string
   
   strcpy(command,"exiftool -m -q -overwrite_original");
   
   for (ii = 0; ii < nkeys; ii++)
      strncatv(command,ccc," -",keys[ii],"=\"",text[ii],"\"",null);        //  "-exif:" replaced with "-"  v.10.0
   strncatv(command,ccc," \"",file,"\"",null);
   
   err = system(command);
   if (err) printf(" exif_put: %s \n",wstrerror(err));
   return err;
}


/**************************************************************************/

//  copy EXIF data from original image file to new (edited) image file
//  if nkeys > 0, up to 9 keys may be replaced with new values
//  exiftool command revised: parameter sequence and -icc_profile                      v.10.8.4

int exif_copy(cchar *file1, cchar *file2, cchar **keys, cchar **text, int nkeys)
{
   int      ii, err;

   strcpy(command,"exiftool -m -q -tagsfromfile");                         //  exiftool -m -q -tagsfromfile "file1"
   strncatv(command,ccc," \"",file1,"\"",null);
   
   strncatv(command,ccc," -all -icc_profile",null);                        //  -all -icc_profile

   for (ii = 0; ii < nkeys; ii++)                                          //  -keyname="keyvalue" ...   (options)
      if (text[ii])
         strncatv(command,ccc," -",keys[ii],"=\"",text[ii],"\"",null);

   strncatv(command,ccc," \"",file2,"\""," -overwrite_original",null);     //  "file2" -overwrite_original
   
   err = system(command);
   if (err) printf(" exiftool: %s \n",wstrerror(err));
   return err;
}


/**************************************************************************
      Image Tags functions
***************************************************************************/

int   get_mouse_tag(GtkTextView *, int px, int py, cchar *);               //  get tag selected by mouse
int   add_tag(char *tag, char *taglist, int maxcc);                        //  add tag if unique and enough space
int   del_tag(char *tag, char *taglist);                                   //  remove tag from tag list
int   add_recentag(char *tag);                                             //  add tag to recent tags, keep most recent
void  load_deftags();                                                      //  tags_defined file >> tags_deftags[]
void  save_deftags();                                                      //  tags_deftags[] >> defined_tags file
int   find_deftag(char *tag);                                              //  find tag in tags_deftags[]
int   add_deftag(char *catg, char *tag);                                   //  add tag to tags_deftags[]
int   del_deftag(char *tag);                                               //  remove tag from tags_deftags[]
void  deftags_stuff(zdialog *zd);                                          //  tags_deftags[] >> dialog widget "deftags"

int      deftags_changed = 0;                                              //  defined tags have been changed
int      admintags = 0;                                                    //  tag edit: do tags admin only

char     tags_date[12] = "";                                               //  image date, yyyymmdd
char     tags_prdate[12] = "";                                             //  previous image date read or set
char     tags_stars = '0';                                                 //  image rating in stars, '0' to '5'
char     tags_cliktag[tagcc] = "";                                         //  tag clicked by mouse
char     tags_clikcatg[tagcc] = "";                                        //  category clicked by mouse
char    *tags_deftags[maxtagcats];                                         //  defined tags: catg: tag1; tag2; ... tagN;
char     tags_deftext[tagTcc] = "";                                        //  defined tags as flat text buffer
char     tags_filetags[tagFcc] = "";                                       //  tags for current image file
char     tags_comments[tagFcc];                                            //  image user comments
char     tags_recentags[tagRcc] = "";                                      //  recently added tags list
char     tags_batchAddTags[tagMcc] = "";                                   //  batch add tags list
char     tags_searchtags[tagScc] = "";                                     //  search tags list
char     tags_searchcomms[tagScc] = "";                                    //  search comments list
char     tags_searchfiles[tagScc] = "";                                    //  search files list


/**************************************************************************/

//  edit tags menu function

void m_edittags(GtkWidget *, cchar *)
{
   void  edittags_fixwidget(zdialog *, cchar * widgetname);                //  fix tags widget for selecting with mouse
   int   edittags_dialog_event(zdialog *zd, cchar *event);

   char        *ppv, starsN[12];

   zfuncs::F1_help_topic = "edit_tags";                                    //  v.10.8

   if (! Fexiftool) {                                                      //  exiftool is required
      zmessageACK(MWIN,Bexiftoolmissing);
      return;
   }
   
   if (! image_file) return;

   if (! zdedittags)                                                       //  (re) start tag edit dialog 
   {
      zdedittags = zdialog_new(ZTX("Edit Tags"),mWin,Bsave,Bcancel,null);

      zdialog_add_widget(zdedittags,"hbox","hb1","dialog",0,"space=5");
      zdialog_add_widget(zdedittags,"label","labfile","hb1",ZTX("file:"),"space=5");
      zdialog_add_widget(zdedittags,"label","file","hb1");

      zdialog_add_widget(zdedittags,"hbox","hb2","dialog",0,"space=5");
      zdialog_add_widget(zdedittags,"label","lab21","hb2",ZTX("image date (yyyymmdd)"),"space=5");
      zdialog_add_widget(zdedittags,"entry","date","hb2",0,"scc=12");
      zdialog_add_widget(zdedittags,"button","prdate","hb2",ZTX("use last"),"space=5");

      zdialog_add_widget(zdedittags,"hbox","hb3","dialog",0,"space=5");
      zdialog_add_widget(zdedittags,"label","labstars","hb3",ZTX("image stars"),"space=5");
      zdialog_add_widget(zdedittags,"vbox","vb3","hb3");
      zdialog_add_widget(zdedittags,"hbox","hb31","vb3",0,"homog");
      zdialog_add_widget(zdedittags,"hbox","hb32","vb3",0,"homog");
      zdialog_add_widget(zdedittags,"label","lab30","hb31","0");
      zdialog_add_widget(zdedittags,"label","lab31","hb31","1");
      zdialog_add_widget(zdedittags,"label","lab32","hb31","2");
      zdialog_add_widget(zdedittags,"label","lab33","hb31","3");
      zdialog_add_widget(zdedittags,"label","lab34","hb31","4");
      zdialog_add_widget(zdedittags,"label","lab35","hb31","5");
      zdialog_add_widget(zdedittags,"radio","stars0","hb32",0);
      zdialog_add_widget(zdedittags,"radio","stars1","hb32",0);
      zdialog_add_widget(zdedittags,"radio","stars2","hb32",0);
      zdialog_add_widget(zdedittags,"radio","stars3","hb32",0);
      zdialog_add_widget(zdedittags,"radio","stars4","hb32",0);
      zdialog_add_widget(zdedittags,"radio","stars5","hb32",0);

      zdialog_add_widget(zdedittags,"hbox","hb4","dialog",0,"space=5");
      zdialog_add_widget(zdedittags,"label","lab4","hb4",ZTX("current tags"),"space=5");
      zdialog_add_widget(zdedittags,"frame","frame4","hb4",0,"expand");
      zdialog_add_widget(zdedittags,"edit","filetags","frame4",0,"expand");

      zdialog_add_widget(zdedittags,"hbox","hb5","dialog",0,"space=5");
      zdialog_add_widget(zdedittags,"label","recent","hb5",ZTX("recent tags"),"space=5");
      zdialog_add_widget(zdedittags,"frame","frame5","hb5",0,"expand");
      zdialog_add_widget(zdedittags,"edit","recentags","frame5",0,"expand");

      zdialog_add_widget(zdedittags,"hsep","sep1","dialog",0,"space=5");   //  v.10.8.1 layout changed
      zdialog_add_widget(zdedittags,"hbox","hb6","dialog");
      zdialog_add_widget(zdedittags,"check","admin","hb6",ZTX("manage tags"),"space=5");
      zdialog_add_widget(zdedittags,"hbox","hb7","dialog",0,"space=5");
      zdialog_add_widget(zdedittags,"label","labcatg","hb7",ZTX("category"),"space=5");
      zdialog_add_widget(zdedittags,"entry","catg","hb7",0,"scc=10");
      zdialog_add_widget(zdedittags,"label","space","hb7",0,"space=5");
      zdialog_add_widget(zdedittags,"label","labtag","hb7",ZTX("tag"),"space=5");
      zdialog_add_widget(zdedittags,"entry","tag","hb7",0,"scc=10");
      zdialog_add_widget(zdedittags,"label","space","hb7",0,"space=5");
      zdialog_add_widget(zdedittags,"button","newtag","hb7",ZTX("create"));
      zdialog_add_widget(zdedittags,"button","deltag","hb7",ZTX("delete"));

      zdialog_add_widget(zdedittags,"hbox","hb8","dialog");
      zdialog_add_widget(zdedittags,"label","labdeftags","hb8",ZTX("defined tags"),"space=5");
      zdialog_add_widget(zdedittags,"hbox","hb9","dialog",0,"expand");
      zdialog_add_widget(zdedittags,"frame","frame8","hb9",0,"space=5|expand");
      zdialog_add_widget(zdedittags,"scrwin","scrwin8","frame8",0,"expand");
      zdialog_add_widget(zdedittags,"edit","deftags","scrwin8",0,"expand");

      zdialog_resize(zdedittags,0,500);                                    //  run dialog
      zdialog_run(zdedittags,edittags_dialog_event);
      
      edittags_fixwidget(zdedittags,"filetags");                           //  setup for mouse tag selection
      edittags_fixwidget(zdedittags,"deftags");
      edittags_fixwidget(zdedittags,"recentags");

      zdialog_stuff(zdedittags,"admin",0);                                 //  reset admin mode    bugfix v.10.6
      admintags = 0;

      load_deftags();                                                      //  stuff defined tags into dialog
      deftags_stuff(zdedittags);
   }

   load_filetags(image_file);                                              //  get file tags from EXIF data

   ppv = (char *) strrchr(image_file,'/');
   zdialog_stuff(zdedittags,"file",ppv+1);                                 //  stuff dialog file name

   zdialog_stuff(zdedittags,"date",tags_date);                             //  stuff dialog data
   sprintf(starsN,"stars%c",tags_stars);
   zdialog_stuff(zdedittags,starsN,1);
   zdialog_stuff(zdedittags,"filetags",tags_filetags);
   zdialog_stuff(zdedittags,"recentags",tags_recentags);

   return;
}


//  setup tag display widget for tag selection using mouse clicks

void edittags_fixwidget(zdialog *zd, cchar * widgetname)
{
   void  edittags_mouse(GtkTextView *, GdkEventButton *, cchar *);         //  select tag via mouse click

   GtkWidget         *widget;
   GdkWindow         *gdkwin;

   widget = zdialog_widget(zd,widgetname);                                 //  make widget wrap text
   gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(widget),GTK_WRAP_WORD);
   gtk_text_view_set_editable(GTK_TEXT_VIEW(widget),0);                    //  disable widget editing

   gdkwin = gtk_text_view_get_window(GTK_TEXT_VIEW(widget),TEXTWIN);       //  cursor for tag selection
   gdk_window_set_cursor(gdkwin,arrowcursor);

   gtk_widget_add_events(widget,GDK_BUTTON_PRESS_MASK);                    //  connect mouse-click event
   G_SIGNAL(widget,"button-press-event",edittags_mouse,widgetname)
}


//  edit tags mouse-click event function
//  get clicked tag and add to or remove from file tags and recent tags

void edittags_mouse(GtkTextView *widget, GdkEventButton *event, cchar *widgetname)
{
   int      mpx, mpy, cc;

   if (event->type != GDK_BUTTON_PRESS) return;
   mpx = int(event->x);                                                    //  mouse click position
   mpy = int(event->y);
   
   cc = get_mouse_tag(widget,mpx,mpy,widgetname);                          //  tags_cliktag = clicked tag in list
   
   if (! cc) {
      if (*tags_clikcatg)                                                  //  selected category >> dialog widget
         zdialog_stuff(zdedittags,"catg",tags_clikcatg);
      return;
   }
   
   zdialog_stuff(zdedittags,"tag",tags_cliktag);                           //  selected tag >> dialog widget

   if (strEqu(widgetname,"filetags")) {                                    //  remove tag from file tags
      del_tag(tags_cliktag,tags_filetags);
      zdialog_stuff(zdedittags,"filetags",tags_filetags);                  //  update dialog widgets
   }
   
   if (strEqu(widgetname,"deftags")) {                                     //  add defined tag to file tags
      if (! admintags) {
         add_tag(tags_cliktag,tags_filetags,tagFcc);                       //  add defined tag to file tags
         zdialog_stuff(zdedittags,"filetags",tags_filetags);
         add_recentag(tags_cliktag);                                       //  and recent tags
         zdialog_stuff(zdedittags,"recentags",tags_recentags);
      }
   }

   if (strEqu(widgetname,"recentags")) {                                   //  add recent tag to file tags
      add_tag(tags_cliktag,tags_filetags,tagFcc);
      zdialog_stuff(zdedittags,"filetags",tags_filetags);
   }

   return;
}


//  dialog event and completion callback function

int edittags_dialog_event(zdialog *zd, cchar *event)
{
   int         err;
   char        tag[tagcc], catg[tagcc];
   
   if (zd->zstat) {
      if (zd->zstat == 1) {                                                //  [save]                 v.10.11
         save_filetags(image_file);                                        //  save tag changes
         zd->zstat = 0;                                                    //  keep dialog active
         return 0;
      }

      zdialog_free(zdedittags);                                            //  cancel - kill dialog
      zdedittags = null;
      save_deftags();                                                      //  save changes if any    v.10.6
      return 0;
   }
   
   if (strEqu(event,"date")) {                                             //  image date revised
      err = zdialog_fetch(zd,"date",tags_date,11);
      if (err) return 1;
      if (strlen(tags_date) == 4) strcat(tags_date,"0101");                //  yyyy >> yyyy0101
      if (strlen(tags_date) == 6) strcat(tags_date,"01");                  //  yyyymm >> yyyymm01
      Ftagschanged++;
   }
   
   if (strEqu(event,"prdate")) {                                           //  repeat last date used
      if (*tags_prdate) {
         zdialog_stuff(zd,"date",tags_prdate);
         strcpy(tags_date,tags_prdate);
         Ftagschanged++;
      }
   }

   if (strnEqu(event,"stars",5)) {                                         //  event = stars0 to stars5
      tags_stars = event[5];                                               //  '0' to '5'
      Ftagschanged++;
   }

   if (strEqu(event,"newtag")) {                                           //  add new tag to defined tags
      zdialog_fetch(zd,"catg",catg,tagcc);
      zdialog_fetch(zd,"tag",tag,tagcc);
      add_deftag(catg,tag);
      deftags_stuff(zdedittags);
   }

   if (strEqu(event,"deltag")) {                                           //  remove tag from defined tags
      zdialog_fetch(zd,"tag",tag,tagcc);
      del_deftag(tag);
      deftags_stuff(zdedittags);
   }
   
   if (strEqu(event,"admin"))                                              //  set tag admin mode
      zdialog_fetch(zd,"admin",admintags);

   return 0;
}


/**************************************************************************/

//  Convert mouse click position in a tag list into the selected tag.
//  cc of selected tag is returned, or zero if no tag clicked.
//  Selected tag is returned in tags_cliktag.
//  If a category is clicked, it is returned in tags_clikcatg.

int get_mouse_tag(GtkTextView *widget, int mpx, int mpy, cchar *widgetname)
{
   GtkTextIter    iter;
   int            tbx, tby, offset, cc;
   char           *ptext, *pp1, *pp2;
   cchar          *delims = ";:", *skips = ";: \n";

   gtk_text_view_window_to_buffer_coords(widget,GTK_TEXT_WINDOW_TEXT,mpx,mpy,&tbx,&tby);
   gtk_text_view_get_iter_at_location(widget,&iter,tbx,tby);
   offset = gtk_text_iter_get_offset(&iter);                               //  graphic position in widget text

   ptext = 0;   
   if (strEqu(widgetname,"filetags")) ptext = tags_filetags;               //  get corresponding text
   if (strEqu(widgetname,"recentags")) ptext = tags_recentags;
   if (strEqu(widgetname,"batchAddTags")) ptext = tags_batchAddTags;
   if (strEqu(widgetname,"searchtags")) ptext = tags_searchtags;
   if (strEqu(widgetname,"deftags")) ptext = tags_deftext;
   if (! ptext) return 0;

   pp1 = ptext + utf8_position(ptext,offset);                              //  graphic position to byte position

   if (! *pp1 || strchr(delims,*pp1)) return 0;                            //  reject ambiguity
   while (pp1 > ptext && ! strchr(delims,*pp1)) pp1--;                     //  find preceding delimiter
   while (strchr(skips,*pp1)) pp1++;                                       //  remove leading stuff
   pp2 = pp1;
   while (*pp2 && ! strchr(delims,*pp2)) pp2++;                            //  find following delimiter
   cc = pp2 - pp1;
   if (cc >= tagcc) return 0;                                              //  reject tag too big
   if (*pp2 == ';') {
      strncpy0(tags_cliktag,pp1,cc+1);                                     //  tags_cliktag = selected tag
      return cc;                                                           //  tag cc
   }
   if (*pp2 == ':') strncpy0(tags_clikcatg,pp1,cc+1);                      //  tags_clikcatg = selected category
   return 0;                                                               //  not a tag
}


/**************************************************************************/

//  add input tag to output tag list if not already there and enough room
//  returns:   0 = added OK     1 = not unique (case ignored)
//             2 = overflow     3 = bad utf8 characters     4 = null tag

int add_tag(char *tag, char *taglist, int maxcc)
{
   char     *pp1, *pp2, tag1[tagcc], tag2[tagcc];
   int      cc, cc1, cc2;
   
   strncpy0(tag1,tag,tagcc);                                               //  remove leading and trailing blanks
   cc = strTrim2(tag2,tag1);
   if (! cc) return 4;

   if (utf8_check(tag2)) {                                                 //  check for valid utf8 encoding
      printf("bad utf8 characters: %s \n",tag2);
      return 3;
   }
   
   while ((pp1 = strchr(tag2,';'))) *pp1 = '-';                            //  replace problem characters
   while ((pp1 = strchr(tag2,':'))) *pp1 = '-';

   strcpy(tag,tag2);                                                       //  replace tag with sanitized version
   
   pp1 = taglist;
   cc1 = strlen(tag);

   while (true)                                                            //  check if already in tag list
   {
      while (*pp1 == ' ' || *pp1 == ';') pp1++;
      if (! *pp1) break;
      pp2 = pp1 + 1;
      while (*pp2 && *pp2 != ';') pp2++;
      cc2 = pp2 - pp1;
      if (cc2 == cc1 && strncaseEqu(tag,pp1,cc1)) return 1;
      pp1 = pp2;
   }

   cc2 = strlen(taglist);                                                  //  append to tag list if space enough
   if (cc1 + cc2 + 3 > maxcc) return 2;
   strcpy(taglist + cc2,tag);
   strcpy(taglist + cc2 + cc1,"; ");                                       //  add semicolon delimiter + space

   if (taglist == tags_filetags)                                           //  image tags were changed
      Ftagschanged++;
   return 0;
}


//  remove tag from taglist, if present
//  returns: 0 if found and deleted, otherwise 1

int del_tag(char *tag, char *taglist)
{
   int         ii, ftcc, atcc, found;
   char        *temptags;
   cchar       *pp;
   
   temptags = strdupz(taglist,0,"temptags");
   
   *taglist = 0;
   ftcc = found = 0;
   
   for (ii = 1; ; ii++)
   {
      pp = strField(temptags,';',ii);                                      //  next tag
      if (! pp) {
         zfree(temptags);
         if (found && taglist == tags_filetags)                            //  image tags were changed
            Ftagschanged++;
         return 1-found;
      }
      if (*pp == ' ') continue;
      
      if (strcaseEqu(pp,tag)) {                                            //  skip matching tag
         found = 1;
         continue;
      }

      atcc = strlen(pp);                                                   //  copy non-matching tag
      strcpy(taglist + ftcc, pp);
      ftcc += atcc;
      strcpy(taglist + ftcc,"; ");
      ftcc += 2;
   }
}


//  add new tag to recent tags, if not already.
//  remove oldest to make space if needed.

int add_recentag(char *tag)
{
   int         err;
   char        *pp, temptags[tagRcc];

   err = add_tag(tag,tags_recentags,tagRcc);                               //  add tag to recent tags

   while (err == 2)                                                        //  overflow
   {
      strncpy0(temptags,tags_recentags,tagRcc);                            //  remove oldest to make room
      pp = strstr(temptags,"; ");
      if (! pp) return 0;
      strcpy(tags_recentags,pp+2);
      err = add_tag(tag,tags_recentags,tagRcc);
   }

   return 0;
}


/**************************************************************************/

//  Load tags_defined file into tags_deftags[ii] => category: tag1; tag2; ...
//  Read tags_index file and add unmatched tags: => nocatg: tag1; tag2; ...

void load_deftags()
{
   int tags_Ucomp(cchar *tag1, cchar *tag2);

   FILE        *fid;
   int         ii, jj, ntags, err, cc, tcc;
   int         ncats, catoverflow;
   int         nocat, nocatcc;
   char        tag[tagcc], catg[tagcc];
   char        tagsbuff[tagGcc+tagcc+4];
   char        *pp1, *pp2;
   char        ptags[tagntc][tagcc];
   
   for (ii = 0; ii < maxtagcats; ii++)                                     //  clean memory
      tags_deftags[ii] = 0;

   ncats = catoverflow = 0;

   fid = fopen(tags_defined_file,"r");                                     //  read tags_defined file
   if (fid) {
      while (true) {
         pp1 = fgets_trim(tagsbuff,tagGcc+tagcc+4,fid);
         if (! pp1) break;
         pp2 = strchr(pp1,':');                                            //  isolate "category:"
         if (! pp2) continue;                                              //  reject bad data
         cc = pp2 - pp1;
         if (cc > tagcc-1) continue;
         strncpy0(catg,pp1,cc+1);
         pp2++;
         while (*pp2 == ' ') pp2++;
         if (strlen(pp2) < 3) continue;
         if (ncats == maxtagcats-1) goto toomanycats;
         if (strlen(pp1) > tagGcc-2) goto cattoobig;
         tags_deftags[ncats] = strdupz(pp1,0,"deftags");                   //  tags_deftags[ii]
         ncats++;                                                          //     = category: tag1; tag2; ... tagN;
      }
      err = fclose(fid);
      if (err) goto deftagserr;
   }
   
   nocat = ncats;                                                          //  last category, "nocatg"
   ncats++;                                                                //   for non-matching tags in 
   tags_deftags[nocat] = zmalloc(tagGcc,"deftags");                        //    tags_index file
   strcpy(tags_deftags[nocat],"nocatg: ");
   strcpy(catg,"nocatg");
   nocatcc = 8;

   fid = fopen(tags_index_file,"r");                                       //  read tags_index file
   if (! fid) return;
   
   while (! catoverflow)
   {
      pp1 = fgets_trim(tagsbuff,tagGcc,fid);                               //  next image file tags record
      if (! pp1) break;
      if (strnNeq(tagsbuff,"tags: ",6)) continue;
      pp1 = pp1 + 6;
      
      
      while (true)
      {
         while (*pp1 == ';' || *pp1 == ' ') pp1++;                         //  next image tag
         if (! *pp1) break;
         pp2 = pp1;
         while (*pp2 && *pp2 != ';') pp2++;
         cc = pp2 - pp1;
         if (cc > tagcc-1) {
            pp1 = pp2;                                                     //  bugfix    v.10.9
            continue;                                                      //  ignore huge tag
         }
         
         strncpy0(tag,pp1,cc+1);                                           //  look for tag in defined tags
         err = find_deftag(tag);
         if (! err) {                                                      //  found
            pp1 = pp2;
            continue;
         }

         if (nocatcc + cc > tagGcc-3) {
            catoverflow = 1;                                               //  nocatg: length limit reached
            break;
         }
         else {
            strcpy(tags_deftags[nocat] + nocatcc, tag);                    //  append "tag; " to list
            nocatcc += cc;
            strcpy(tags_deftags[nocat] + nocatcc, "; ");
            nocatcc += 2;
         }
         
         pp1 = pp2;
      }
   }
            
   err = fclose(fid);
   if (err) goto filetagserr;
   if (catoverflow) goto cattoobig;

//  parse all the tags in each category and sort in ascending order

   for (ii = 0; ii < ncats; ii++)
   {
      pp1 = tags_deftags[ii];
      pp2 = strchr(pp1,':');
      cc = pp2 - pp1;
      strncpy0(catg,pp1,cc+1);
      pp1 = pp2 + 1;
      while (*pp1 == ' ') pp1++;
      tcc = 0;

      for (jj = 0; jj < tagntc; jj++)
      {
         if (! *pp1) break;
         pp2 = strchr(pp1,';');
         if (pp2) cc = pp2 - pp1;
         else cc = strlen(pp1);
         if (cc > tagcc-1) cc = tagcc-1;
         strncpy0(ptags[jj],pp1,cc+1);
         pp1 += cc + 1;
         tcc += cc;
         while (*pp1 == ' ') pp1++;
      }
      
      ntags = jj;
      if (ntags == tagntc) goto cattoobig;
      HeapSort((char *) ptags,tagcc,ntags,tags_Ucomp);

      pp1 = tags_deftags[ii];
      tcc += strlen(catg) + 2 + 2 * ntags + 2;                             //  category, all tags, delimiters
      pp2 = zmalloc(tcc,"deftags");

      tags_deftags[ii] = pp2;                                              //  swap memory
      zfree(pp1);

      strcpy(pp2,catg);
      pp2 += strlen(catg);
      strcpy(pp2,": ");                                                    //  pp2 = "category: "
      pp2 += 2;
      
      for (jj = 0; jj < ntags; jj++)                                       //  add the sorted tags
      {
         strcpy(pp2,ptags[jj]);
         pp2 += strlen(pp2);
         strcpy(pp2,"; ");
         pp2 += 2;                                                         //  "tag; " added to pp2
      }
      
      *pp2 = 0;
   }
   
//  sort the categories in ascending order
//  leave "nocatg" at the end

   for (ii = 0; ii < ncats-1; ii++)
   for (jj = ii+1; jj < ncats-2; jj++)
   {
      pp1 = tags_deftags[ii];
      pp2 = tags_deftags[jj];
      if (strcasecmp(pp1,pp2) > 0) {
         tags_deftags[ii] = pp2;
         tags_deftags[jj] = pp1;
      }
   }

   return;

toomanycats:
   zmessLogACK(MWIN,"more than %d categories",maxtagcats);
   return;
   
cattoobig:
   zmessLogACK(MWIN,"category %s is too big",catg);
   return;

deftagserr:
   zmessLogACK(MWIN,"tags_defined file error: %s",strerror(errno));
   return;

filetagserr:
   zmessLogACK(MWIN,"tags_index file error: %s",strerror(errno));
   return;
}


//  compare function for tag sorting
//  wcscasecmp does not work for plain ascii
//  so what happens to Chinese, etc?

int tags_Ucomp(cchar *tag1, cchar *tag2)
{
   return strcasecmp(tag1,tag2);
}


//  write tags_deftags[] memory data to the defined tags file if any changes were made

void save_deftags()
{
   int         ii, err;
   FILE        *fid;

   if (! deftags_changed) return;                                          //  do nothing if nothing changed
   deftags_changed = 0;

   fid = fopen(tags_defined_file,"w");                                     //  write tags_defined file
   if (! fid) goto deftagserr;

   for (ii = 0; ii < maxtagcats-1; ii++)
   {
      if (! tags_deftags[ii+1]) break;                                     //  omit last category, "nocatg"
      err = fprintf(fid,"%s\n",tags_deftags[ii]);                          //  each record: 
      if (err < 0) goto deftagserr;                                        //    category: tag1; tag2; ... tagN;
   }

   err = fclose(fid);
   if (err) goto deftagserr;
   return;
   
deftagserr:
   zmessLogACK(MWIN,"tags_defined file error: %s",strerror(errno));
   return;
}


//  find a given tag in tags_deftags[]
//  return: 0 = found, 1 = not found

int find_deftag(char *tag)
{
   int      ii, cc;
   char     tag2[tagcc+4];
   char     *pp;

   if (! tag || *tag <= ' ') return 0;                                     //  bad tag

   strncpy0(tag2,tag,tagcc);                                               //  construct "tag;"
   cc = strlen(tag2);
   strcpy(tag2+cc,"; ");
   cc += 2;

   for (ii = 0; ii < maxtagcats; ii++)
   {
      pp = tags_deftags[ii];
      if (! pp) return 1;                                                  //  not found

      while (pp)
      {
         pp = strcasestr(pp,tag2);                                         //  look for match with "; tag; "
         if (! pp) break;
         if (pp[-2] == ';' || pp[-2] == ':') return 0;                     //    or ": tag; "
         pp += cc;
      }
   }

   return 1;
}


//  add new tag to tags_deftags[] >> category: tag1; tag2; ... newtag;
//  returns:   0 = added OK     1 = not unique (case ignored)
//             2 = overflow     3 = bad utf8 characters     4 = null tag
//  if tag present under another category, it is moved to new category

int add_deftag(char *catg, char *tag)
{
   int         ii, cc, cc1, cc2;
   char        tag1[tagcc], tag2[tagcc];
   char        *pp1, *pp2;
   
   strncpy0(tag1,tag,tagcc);                                               //  remove leading and trailing blanks
   cc = strTrim2(tag2,tag1);
   if (! cc) return 4;

   if (utf8_check(tag2)) {                                                 //  check for valid utf8 encoding
      printf("bad utf8 characters: %s \n",tag2);
      return 3;
   }

   while ((pp1 = strchr(tag2,';'))) *pp1 = '-';                            //  replace problem characters
   while ((pp1 = strchr(tag2,':'))) *pp1 = '-';
   
   strcpy(tag,tag2);                                                       //  replace tag with sanitized version

   del_deftag(tag);                                                        //  delete if already there
   deftags_changed++;

   if (! catg || *catg <= ' ') catg = (char *) "nocatg";

   cc1 = strlen(catg);

   for (ii = 0; ii < maxtagcats; ii++)                                     //  look for given category
   {
      pp1 = tags_deftags[ii];
      if (! pp1) goto newcatg;
      if (! strnEqu(catg,pp1,cc1)) continue;
      if (pp1[cc1] == ':') goto oldcatg;
   }

newcatg:
   if (ii == maxtagcats) goto toomanycats;
   cc1 = strlen(catg) + strlen(tag) + 6;
   pp1 = zmalloc(cc1,"deftags");
   *pp1 = 0;
   strncatv(pp1,cc1,catg,": ",tag,"; ",null);                              //  record = "category: tag; "
   tags_deftags[ii] = tags_deftags[ii-1];                                  //  move "nocatg" record to next slot
   tags_deftags[ii-1] = pp1;                                               //  insert new record before
   return 0;

oldcatg:
   cc1 = strlen(tag);
   pp2 = pp1 + 2;
   while (true)
   {
      pp2 = strcasestr(pp2,tag);                                           //  look for match with "; tag;"
      if (! pp2) break;                                                    //                   or ": tag;"
      if (pp2 && pp2[cc] == ';' && (pp2[-2] == ';' || pp2[-2] == ':'))
         return 1;                                                         //  tag not unique
      pp2 += cc1;
   }
   cc2 = strlen(pp1);                                                      //  add new tag to old record
   if (cc1 + cc2 + 4 > tagGcc) goto cattoobig;
   pp2 = strdupz(pp1,cc2+4,"deftags");                                     //  expand string
   zfree(pp1);
   tags_deftags[ii] = pp2;
   strcpy(pp2+cc2,tag);                                                    //  old record + "tag; "
   strcpy(pp2+cc2+cc1,"; ");
   return 0;

toomanycats:
   zmessLogACK(MWIN,"more than %d categories",maxtagcats);
   return 2;

cattoobig:
   zmessLogACK(MWIN,"category is too big");
   return 2;
}


//  delete tag from defined tags list, tags_deftags[] 
//  return: 0 = found and deleted, 1 = not found

int del_deftag(char *tag)
{
   int      ii, cc;
   char     tag2[tagcc+4];
   char     *pp, *pp1, *pp2;

   if (! tag || *tag <= ' ') return 1;                                     //  bad tag

   strncpy0(tag2,tag,tagcc);                                               //  construct "tag; "
   cc = strlen(tag2);
   strcpy(tag2+cc,"; ");
   cc += 2;

   for (ii = 0; ii < maxtagcats; ii++)
   {
      pp = tags_deftags[ii];
      if (! pp) return 1;                                                  //  not found

      while (pp)
      {
         pp = strcasestr(pp,tag2);                                         //  look for match with "; tag; "
         if (! pp) break;
         if (pp && (pp[-2] == ';' || pp[-2] == ':')) goto found;           //                   or ": tag; "
         pp += cc;
      }
   }
   
found:
   for (pp1 = pp, pp2 = pp+cc; *pp2; pp1++, pp2++)                         //  eliminate "tag; "
      *pp1 = *pp2;
   *pp1 = 0;
   
   deftags_changed++;
   return 0;
}


//  stuff tags_deftags[] into given text widget and format by category
//  create tags_deftext with flat list of tags for mouse clicking

void deftags_stuff(zdialog *zd)
{
   GtkWidget      *widget;
   GtkTextBuffer  *textbuff;
   GtkTextIter    iter1, iter2;
   int            ii, cc;
   char           catgname[tagcc+3];
   char           *pp1, *pp2;

   widget = zdialog_widget(zd,"deftags");
   wclear(widget);

   for (ii = 0; ii < maxtagcats; ii++)
   {
      pp1 = tags_deftags[ii];
      if (! pp1) break;
      pp2 = strchr(pp1,':');
      if (! pp2) continue;
      if (pp2 > pp1 + tagcc-3) continue;
      pp2 += 2;
      cc = pp2 - pp1 + 1;
      strncpy0(catgname,pp1,cc);
      wprintx(widget,0,catgname,"monospace bold 8");
      if (*pp2) wprintx(widget,0,pp2,"monospace 8");
      wprintx(widget,0,"\n");
   }
   
   textbuff = gtk_text_view_get_buffer(GTK_TEXT_VIEW(widget));
   gtk_text_buffer_get_bounds(textbuff,&iter1,&iter2);
   pp1 = gtk_text_buffer_get_text(textbuff,&iter1,&iter2,0);
   strncpy0(tags_deftext,pp1,tagTcc);

   return;
}


/**************************************************************************/

//  image file EXIF data >> tags_date, tags_stars, tags_filetags, tags_comments  in memory

void load_filetags(cchar *file)
{
   void exif_tagdate(cchar *exifdate, char *tagdate);

   int         ii, jj, cc;
   char        *pp;
   cchar       *exifkeys[4] = { exif_date_key, exif_tags_key, exif_rating_key, exif_comment_key };
   char        **ppv, *imagedate, *imagetags, *imagestars, *imagecomms;

   *tags_filetags = *tags_date = *tags_comments = 0;
   tags_stars = '0';
   
   ppv = exif_get(file,exifkeys,4);                                        //  stars rating added        v.10.0
   imagedate = ppv[0];
   imagetags = ppv[1];
   imagestars = ppv[2];
   imagecomms = ppv[3];

   if (imagedate) {
      exif_tagdate(imagedate,tags_date);                                   //  EXIF date/time >> yyyymmdd
      strcpy(tags_prdate,tags_date);
      zfree(imagedate);
   }

   if (imagetags)
   {
      for (ii = 1; ; ii++)
      {
         pp = (char *) strField(imagetags,';',ii);                         //  assume colon delimited tags
         if (! pp) break;
         if (*pp == ' ') continue;
         cc = strlen(pp);
         if (cc >= tagcc) continue;                                        //  reject tags too big
         for (jj = 0; jj < cc; jj++)
            if (pp[jj] > 0 && pp[jj] < ' ') break;                         //  reject tags with control characters
         if (jj < cc) continue;
         add_tag(pp,tags_filetags,tagFcc);                                 //  add to file tags if unique
      }

      zfree(imagetags);
   }

   if (imagestars) {                                                       //  v.10.0
      tags_stars = *imagestars;
      if (tags_stars < '0' || tags_stars > '5') tags_stars = '0';
      zfree(imagestars);
   }
   
   if (imagecomms) {                                                       //  v.10.11
      strncpy0(tags_comments,imagecomms,tagFcc);
      zfree(imagecomms);
   }

   Ftagschanged = 0;
   return;
}


//  tags_date, tags_stars, tags_filetags in memory >> image file EXIF data
//  update defined tags file if any changes

void save_filetags(cchar *file)
{
   void tag_exifdate(cchar *tagdate, char *exifdate);
   
   cchar       *exifkeys[4] = { exif_date_key, exif_tags_key, exif_rating_key, exif_comment_key };
   cchar       *exifdata[4];
   char        imagedate[24], sstars[4];

   save_deftags();                                                         //  update defined tags if changed

   if (! Ftagschanged) return;                                             //  no changes
   Ftagschanged = 0;

   *imagedate = 0;
   if (*tags_date) {
      tag_exifdate(tags_date,imagedate);                                   //  yyyymmdd >> EXIF date/time
      strcpy(tags_prdate,tags_date);
   }
   
   sstars[0] = tags_stars;                                                 //  string for stars rating
   sstars[1] = 0;

   exifdata[0] = imagedate;                                                //  update file EXIF data
   exifdata[1] = tags_filetags;
   exifdata[2] = sstars;
   exifdata[3] = tags_comments;                                            //  v.10.11
   exif_put(file,exifkeys,exifdata,4);

   update_tags_index(file);                                                //  update tags index file

   if (zdexifview) m_exif_view(0,0);                                       //  live EXIF update   v.10.2
   return;
}


//  update tags index file (replace updated file data)                     //  overhauled    v.10.11

void update_tags_index(cchar *file)
{
   char           *ppv, temp_tags_index_file[200];
   char           imagedate[12], filedate[16];
   char           buff[tagrecl];
   int            err, fcopy = 0;
   FILE           *fidr, *fidw;
   struct stat    statb;
   struct tm      bdt;

   strcpy(temp_tags_index_file,tags_index_file);                           //  temp tag file
   strcat(temp_tags_index_file,"_temp");

   fidw = fopen(temp_tags_index_file,"w");                                 //  write temp tag file (new)
   if (! fidw) goto tagserror;
   
   fidr = fopen(tags_index_file,"r");                                      //  read tag file (old)
   
   if (fidr)
   {   
      while (true)                                                         //  copy tags index file to temp
      {                                                                    //    file, omitting this image file
         ppv = fgets_trim(buff,tagrecl,fidr);
         if (! ppv) break;
         
         if (strnEqu(buff,"file: ",6)) {                                   //  start of next file entry
            if (strEqu(buff+6,file)) fcopy = 0;                            //  my file, skip this entry
            else fcopy = 1;
         }
         
         if (fcopy) fprintf(fidw,"%s\n",buff);                             //  copy data for other file entries
      }

      err = fclose(fidr);
      if (err) goto tagserror;
   }

//  append input file and tags data to end of tags index file

   err = fprintf(fidw,"file: %s\n",file);                                  //  output filespec
   if (err <= 0) goto tagserror;

   if (*tags_date) {                                                       //  image date
      strncpy(imagedate,tags_date,4);
      strncpy(imagedate+5,tags_date+4,2);                                  //  yyyymmdd >> yyyy:mm:dd
      strncpy(imagedate+8,tags_date+6,2);
      imagedate[4] = imagedate[7] = ':';
      imagedate[10] = 0;
   }
   else strcpy(imagedate,"null");                                          //  missing image date

   err = stat(file,&statb);
   gmtime_r(&statb.st_mtime,&bdt);
   sprintf(filedate,"%04d%02d%02d%02d%02d%02d",                            //  file date = yyyymmddhhmmss
            bdt.tm_year + 1900, bdt.tm_mon + 1, bdt.tm_mday,
            bdt.tm_hour, bdt.tm_min, bdt.tm_sec);

   err = fprintf(fidw,"date: %s  %s\n",imagedate,filedate);                //  output image and file date

   if (*tags_filetags)                                                     //  output filetags
      err = fprintf(fidw,"tags: %s\n",tags_filetags);
   else  err = fprintf(fidw,"tags: null; \n");                             //  "null" tag if none        v.10.5
   
   err = fprintf(fidw,"stars: %c\n",tags_stars);                           //  output stars rating
   
   if (*tags_comments)                                                     //  output user comments
      err = fprintf(fidw,"comms: %s\n",tags_comments);
   else err = fprintf(fidw,"comms: null \n");                              //  "null" if none

   err = fprintf(fidw,"\n");                                               //  EOL

   err = fclose(fidw);                                                     //  close temp file
   if (err) goto tagserror;
   
   err = rename(temp_tags_index_file,tags_index_file);                     //  replace tags file with temp file
   if (err) goto tagserror;

   return;
   
tagserror:
   zmessLogACK(MWIN,ZTX("tags index file error: %s"),strerror(errno));
   return;
}


//  file is deleted from tags index file

void delete_tags_index(cchar *file)                                        //  new v.10.11
{
   char           *ppv, temp_tags_index_file[200];
   char           filebuff[tagrecl], datebuff[tagrecl];
   char           tagsbuff[tagrecl], starsbuff[tagrecl];
   int            err;
   FILE           *fidr, *fidw;

   strcpy(temp_tags_index_file,tags_index_file);                           //  temp tag file
   strcat(temp_tags_index_file,"_temp");

   fidr = fopen(tags_index_file,"r");                                      //  read tag file
   if (! fidr) return;
   
   fidw = fopen(temp_tags_index_file,"w");                                 //  write temp tag file
   if (! fidw) goto tagserror;
   
   while (true)                                                            //  copy tags index file to temp
   {                                                                       //    file, omitting this image file
      ppv = fgets_trim(filebuff,tagrecl,fidr);
      if (! ppv) break;
      if (strnNeq(filebuff,"file: ",6)) continue;

      if (strEqu(filebuff+6,file)) continue;                               //  if my file, skip copy

      ppv = fgets_trim(datebuff,tagrecl,fidr);
      if (! ppv) break;
      if (strnNeq(datebuff,"date: ",6)) continue;

      ppv = fgets_trim(tagsbuff,tagrecl,fidr);
      if (! ppv) break;
      if (strnNeq(tagsbuff,"tags: ",6)) continue;

      ppv = fgets_trim(starsbuff,tagrecl,fidr);
      if (! ppv) break;
      if (strnNeq(starsbuff,"stars: ",7)) continue;

      fprintf(fidw,"%s\n",filebuff);
      fprintf(fidw,"%s\n",datebuff);                                       //  copy to temp file
      fprintf(fidw,"%s\n",tagsbuff);
      fprintf(fidw,"%s\n",starsbuff);
      fprintf(fidw,"\n");
   }

   err = fclose(fidr);
   if (err) goto tagserror;

   err = fclose(fidw);
   if (err) goto tagserror;
   
   err = rename(temp_tags_index_file,tags_index_file);                     //  replace tags file with temp file
   if (err) goto tagserror;

   return;
   
tagserror:
   zmessLogACK(MWIN,ZTX("tags index file error: %s"),strerror(errno));
   return;
}


/**************************************************************************/

//  menu function - add tags to many files at once

char        **batchAddTags_filelist = 0;
int         batchAddTags_filecount = 0;
zdialog     *zdbatchAddTags = null;                                        //  batch add tags dialog

void m_batchAddTags(GtkWidget *, cchar *)                                  //  new v.9.7
{
   void  batchAddTags_fixwidget(zdialog *zd, cchar * widgetname);
   int   batchAddTags_dialog_event(zdialog *zd, cchar *event);
   
   char       *ptag, **flist, *file;
   int         ii, jj, err;
   
   zfuncs::F1_help_topic = "batch_add_tags";                               //  v.10.8
   
   if (! Fexiftool) {                                                      //  exiftool is required
      zmessageACK(MWIN,Bexiftoolmissing);
      return;
   }

   if (mod_keep()) return;                                                 //  unsaved edits
   if (! menulock(1)) return;                                              //  v.10.5

   zdbatchAddTags = zdialog_new(ZTX("Batch Add Tags"),mWin,Bproceed,Bcancel,null);

   zdialog_add_widget(zdbatchAddTags,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zdbatchAddTags,"label","lab1","hb1",ZTX("tags to add"),"space=10");
   zdialog_add_widget(zdbatchAddTags,"frame","frame1","hb1",0,"expand");
   zdialog_add_widget(zdbatchAddTags,"edit","batchAddTags","frame1",0,"expand");

   zdialog_add_widget(zdbatchAddTags,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(zdbatchAddTags,"button","createtag","hb2",ZTX("create tag"),"space=10");
   zdialog_add_widget(zdbatchAddTags,"entry","ctag","hb2",0);

   zdialog_add_widget(zdbatchAddTags,"hbox","hb3","dialog",0,"space=5");
   zdialog_add_widget(zdbatchAddTags,"button","files","hb3",Bselectfiles,"space=10");
   zdialog_add_widget(zdbatchAddTags,"label","labcount","hb3","0 files selected","space=10");

   zdialog_add_widget(zdbatchAddTags,"hbox","space","dialog",0,"space=3");
   zdialog_add_widget(zdbatchAddTags,"hbox","hb5","dialog");
   zdialog_add_widget(zdbatchAddTags,"label","labdeftags","hb5",ZTX("defined tags"));
   zdialog_add_widget(zdbatchAddTags,"frame","frame5","dialog",0,"space=5|expand");
   zdialog_add_widget(zdbatchAddTags,"scrwin","scrwin5","frame5",0,"expand");
   zdialog_add_widget(zdbatchAddTags,"edit","deftags","scrwin5",0,"expand");

   load_deftags();                                                         //  stuff defined tags into dialog
   deftags_stuff(zdbatchAddTags);

   batchAddTags_filelist = 0;
   batchAddTags_filecount = 0;
   *tags_batchAddTags = 0;

   zdialog_resize(zdbatchAddTags,400,400);                                 //  run dialog
   zdialog_run(zdbatchAddTags,batchAddTags_dialog_event);
   
   batchAddTags_fixwidget(zdbatchAddTags,"batchAddTags");                  //  setup for mouse tag selection
   batchAddTags_fixwidget(zdbatchAddTags,"deftags");

   zdialog_wait(zdbatchAddTags);                                           //  wait for dialog completion

   flist = batchAddTags_filelist;

   if (zdbatchAddTags->zstat != 1) goto cleanup;
   if (! batchAddTags_filecount) goto cleanup;
   if (*tags_batchAddTags <= ' ') goto cleanup;                            //  v.10.5

   SB_goal = batchAddTags_filecount;
   SB_done = 0;

   for (ii = 0; flist[ii]; ii++)                                           //  loop all selected files
   {
      file = flist[ii];                                                    //  display image

      err = f_open(file,0);
      if (err) continue;

      load_filetags(file);                                                 //  load current file tags

      for (jj = 1; ; jj++)                                                 //  add new tags unless already
      {
         ptag = (char *) strField(tags_batchAddTags,';',jj);
         if (! ptag) break;
         if (*ptag == ' ') continue;
         err = add_tag(ptag,tags_filetags,tagFcc);
         if (err == 2) 
            zmessageACK(MWIN,ZTX("%s \n too many tags"),file);
      }

      save_filetags(file);                                                 //  save tag changes
      SB_done++;
   }

cleanup:
   SB_goal = SB_done = 0;

   zdialog_free(zdbatchAddTags);
   zdbatchAddTags = 0;

   if (batchAddTags_filecount) {
      for (ii = 0; flist[ii]; ii++) 
         zfree(flist[ii]);
      zfree(flist);
   }

   menulock(0);
   return;
}


//  setup tag display widget for tag selection using mouse clicks

void batchAddTags_fixwidget(zdialog *zd, cchar * widgetname)
{
   void batchAddTags_mouse(GtkTextView *widget, GdkEventButton *event, cchar *widgetname);

   GtkWidget         *widget;
   GdkWindow         *gdkwin;

   widget = zdialog_widget(zd,widgetname);                                 //  make widget wrap text
   gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(widget),GTK_WRAP_WORD);
   gtk_text_view_set_editable(GTK_TEXT_VIEW(widget),0);                    //  disable widget editing

   gdkwin = gtk_text_view_get_window(GTK_TEXT_VIEW(widget),TEXTWIN);       //  cursor for tag selection
   gdk_window_set_cursor(gdkwin,arrowcursor);

   gtk_widget_add_events(widget,GDK_BUTTON_PRESS_MASK);                    //  connect mouse-click event
   G_SIGNAL(widget,"button-press-event",batchAddTags_mouse,widgetname)
}


//  batchAddTags mouse-click event function
//  get clicked tag and add to or remove from tags_batchAddTags

void batchAddTags_mouse(GtkTextView *widget, GdkEventButton *event, cchar *widgetname)
{
   int            mpx, mpy, cc;

   if (event->type != GDK_BUTTON_PRESS) return;
   mpx = int(event->x);                                                    //  mouse click position
   mpy = int(event->y);

   cc = get_mouse_tag(widget,mpx,mpy,widgetname);                          //  tags_cliktag = clicked tag in list
   if (! cc) return;

   if (strEqu(widgetname,"batchAddTags")) {
      del_tag(tags_cliktag,tags_batchAddTags);                             //  remove tag from tags_batchAddTags
      zdialog_stuff(zdbatchAddTags,"batchAddTags",tags_batchAddTags);      //  update dialog widgets
   }
   
   if (strEqu(widgetname,"deftags")) {
      add_tag(tags_cliktag,tags_batchAddTags,tagMcc);                      //  add defined tag to tags_batchAddTags
      zdialog_stuff(zdbatchAddTags,"batchAddTags",tags_batchAddTags);
   }

   return;
}


//  batchAddTags dialog event function

int batchAddTags_dialog_event(zdialog *zd, cchar *event)
{
   int      ii, err;
   char     tag[tagcc];
   char     **flist = batchAddTags_filelist, countmess[32];

   if (strEqu(event,"createtag")) {
      err = zdialog_fetch(zd,"ctag",tag,tagcc);                            //  add new tag to list
      if (err) return 0;                                                   //  reject too big tag
      add_tag(tag,tags_batchAddTags,tagMcc);                               //  add tag to tags_batchAddTags
      zdialog_stuff(zd,"batchAddTags",tags_batchAddTags);                  //  update dialog widgets
      zdialog_stuff(zd,"ctag","");
   }

   if (strEqu(event,"files"))                                              //  select images to add tags
   {
      if (flist) {                                                         //  free prior list
         for (ii = 0; flist[ii]; ii++) 
            zfree(flist[ii]);
         zfree(flist);
      }

      flist = image_gallery_getfiles(0);                                   //  get file list from user
      batchAddTags_filelist = flist;

      if (flist)                                                           //  count files in list
         for (ii = 0; flist[ii]; ii++);
      else ii = 0;
      batchAddTags_filecount = ii;
      
      sprintf(countmess,"%d files selected",batchAddTags_filecount);
      zdialog_stuff(zd,"labcount",countmess);
   }
   
   return 0;
}
   

/**************************************************************************/

//  menu function - delete or replace a tag for many files at once

char        **batchDelTag_filelist = 0;

void m_batchDelTag(GtkWidget *, cchar *)                                   //  new v.10.9
{
   int   batchDelTag_dialog_event(zdialog *zd, cchar *event);
   
   char           **flist, *file;
   char           deltag[tagcc], reptag[tagcc];
   int            ii, err;
   zdialog        *zd;

   zfuncs::F1_help_topic = "batch_del_tag";
   
   if (! Fexiftool) {                                                      //  exiftool is required
      zmessageACK(MWIN,Bexiftoolmissing);
      return;
   }

   if (mod_keep()) return;                                                 //  unsaved edits
   if (! menulock(1)) return;                                              //  v.10.5

   zd = zdialog_new(ZTX("Batch Delete Tag"),mWin,Bproceed,Bcancel,null);

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","lab1","hb1",ZTX("tag to remove"),"space=10");
   zdialog_add_widget(zd,"entry","deltag","hb1",0,"scc=20");

   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","lab2","hb2",ZTX("optional replacement"),"space=10");
   zdialog_add_widget(zd,"entry","reptag","hb2",0,"scc=20");

   zdialog_add_widget(zd,"hbox","hb3","dialog",0,"space=5");
   zdialog_add_widget(zd,"button","files","hb3",Bselectfiles,"space=10");
   zdialog_add_widget(zd,"label","labcount","hb3","0 files selected","space=10");
   
   zdialog_add_widget(zd,"hbox","hb4","dialog",0,"space=5");
   zdialog_add_widget(zd,"check","allfiles","hb4",ZTX("search all files"),"space=10");

   batchDelTag_filelist = flist = 0;

   zdialog_run(zd,batchDelTag_dialog_event);                               //  run dialog
   zdialog_wait(zd);                                                       //  wait for completion

   if (zd->zstat != 1) goto cleanup;                                       //  canceled

   flist = batchDelTag_filelist;                                           //  get selected files
   if (! flist) goto cleanup;                                              //  none
   
   zdialog_fetch(zd,"deltag",deltag,tagcc);                                //  get tag to delete and opt. replacement
   zdialog_fetch(zd,"reptag",reptag,tagcc);
   if (! *deltag) goto cleanup;                                            //  no tag

   strTrim2(deltag);                                                       //  remove leading and trailing blanks
   strTrim2(reptag);

   for (ii = 0; flist[ii]; ii++)                                           //  get file count
   SB_goal = ii;
   SB_done = 0;

   for (ii = 0; flist[ii]; ii++)                                           //  loop all selected files
   {
      file = flist[ii];

      err = f_open(file,0);                                                //  display image
      if (err) continue;

      load_filetags(file);                                                 //  load current file tags

      err = del_tag(deltag,tags_filetags);                                 //  remove tag, if present
      if (err) continue;                                                   //  not present

      if (*reptag) {                                                       //  add replacement tag, if defined
         err = add_tag(reptag,tags_filetags,tagFcc);
         if (err == 2) 
            zmessageACK(MWIN,ZTX("%s \n too many tags"),file);
      }

      save_filetags(file);                                                 //  save tag changes
      SB_done++;
   }

cleanup:

   SB_goal = SB_done = 0;
   zdialog_free(zd);
   zd = 0;

   if (flist) {
      for (ii = 0; flist[ii]; ii++) 
         zfree(flist[ii]);
      zfree(flist);
   }

   menulock(0);
   return;
}


//  batchDelTag dialog event function

int batchDelTag_dialog_event(zdialog *zd, cchar *event)
{
   FILE           *fidr;
   struct stat    statbuf;
   char           deltag[tagcc];
   char           **flist = batchDelTag_filelist;
   int            err, nfiles, ii, allfiles;
   char           filebuff[tagrecl], tagsbuff[tagrecl];
   char           *ppv, *file, *tags, countmess[32];
   cchar          *ppf;

   if (strEqu(event,"files"))                                              //  select images to add tags
   {
      if (flist) {                                                         //  free prior list
         for (ii = 0; flist[ii]; ii++) 
            zfree(flist[ii]);
         zfree(flist);
      }

      flist = image_gallery_getfiles(0);                                   //  get file list from user
      goto finish;
   }
   
   if (strEqu(event,"allfiles"))
   {
      if (flist) {                                                         //  free prior list
         for (ii = 0; flist[ii]; ii++) 
            zfree(flist[ii]);
         zfree(flist);
      }

      flist = 0;

      zdialog_fetch(zd,"allfiles",allfiles);                               //  get checkbox "search all files"
      if (! allfiles) goto finish;                                         //  not selected

      zdialog_fetch(zd,"deltag",deltag,tagcc);                             //  get tag to delete
      strTrim2(deltag);                                                    //  remove leading and trailing blanks
      strToLower(deltag);                                                  //  use lower case for matching
      if (! *deltag) goto finish;                                          //  tag not specified
      
      fidr = fopen(tags_index_file,"r");                                   //  read tags index file
      if (! fidr) goto finish;                                             //  no file

      flist = (char **) zmalloc(max_images*sizeof(char *),"batchDelTag");  //  list for files found
      nfiles = 0;                                                          //  count files found
      flist[0] = 0;

      while (true)
      {
         ppv = fgets_trim(filebuff,tagrecl,fidr,1);                        //  next record
         if (! ppv) break;                                                 //  EOF

         if (! strnEqu(ppv,"file: ",6)) continue;                          //  file: /dir.../filename.jpg

         file = ppv+6;
         err=stat(file,&statbuf);                                          //  check file exists
         if (err) continue;
         if (! S_ISREG(statbuf.st_mode)) continue;

         ppv = fgets_trim(tagsbuff,tagrecl,fidr);                          //  next record
         if (! ppv) break;
         if (! strnEqu(ppv,"date: ",6)) continue;                          //  date: yyyy:mm:dd 

         ppv = fgets_trim(tagsbuff,tagrecl,fidr);                          //  next record
         if (! ppv) break;
         if (! strnEqu(ppv,"tags: ",6)) continue;                          //  tags: xxxx; xxxxx; ...

         tags = ppv + 6;
         strToLower(tags);                                                 //  use lower case for matching
         
         for (ii = 1; ; ii++)                                              //  step thru file tags
         {
            ppf = strField(tags,';',ii);
            if (! ppf) break;
            if (*ppf == ' ') continue;
            if (strEqu(ppf,deltag)) break;                                 //  look for my tag
         }
         
         if (! ppf) continue;                                              //  tag not found
         
         flist[nfiles] = strdupz(file,0,"batchDelTag");
         nfiles++;
         if (nfiles == max_images-1) break;
      }

      fclose(fidr);                                                        //  close tag index file
      flist[nfiles] = 0;                                                   //  EOL marker
   }

finish:

   batchDelTag_filelist = flist;                                           //  file list to process

   if (flist)                                                              //  count files in list
      for (ii = 0; flist[ii]; ii++);
   else ii = 0;
   
   sprintf(countmess,"%d files selected",ii);                              //  stuff file count into dialog
   zdialog_stuff(zd,"labcount",countmess);
   
   return 0;
}
   

/**************************************************************************/

//  view and edit EXIF User Comment data

void m_edit_ucomm(GtkWidget *, cchar *menu)                                //  new v.10.10.2
{
   char     text[tagFcc];

   int   edit_ucomm_dialog_event(zdialog *zd, cchar *event);
   
   zfuncs::F1_help_topic = "edit_ucomm";

   if (! Fexiftool) {                                                      //  exiftool is required
      zmessageACK(MWIN,Bexiftoolmissing);
      return;
   }
   
   if (! image_file) return;

   if (! zdeditucomm)                                                      //  popup dialog if not already   v.10.8
   {   
      zdeditucomm = zdialog_new(ZTX("Edit User Comments"),mWin,Bsave,Bcancel,null);
      zdialog_add_widget(zdeditucomm,"hbox","hbtext","dialog");
      zdialog_add_widget(zdeditucomm,"frame","frame","hbtext",0,"expand");
      zdialog_add_widget(zdeditucomm,"edit","text","frame",0,"space=5");
      zdialog_run(zdeditucomm,edit_ucomm_dialog_event);
   }
   
   load_filetags(image_file);

   if (*tags_comments)
      repl_1str(tags_comments,text,"\\n","\n");                            //  replace "\n" with real newlines
   else *text = 0;

   zdialog_stuff(zdeditucomm,"text",text);                                 //  stuff into dialog

   return;
}


//  dialog event and completion callback function

int  edit_ucomm_dialog_event(zdialog *zd, cchar *event)
{
   char        text[tagFcc];
   
   if (! zd->zstat) return 0;
   
   else if (zd->zstat == 1)                                                //  save text
   {
      zd->zstat = 0;
      zdialog_fetch(zd,"text",text, 0.9 * tagFcc);                         //  space for newline expansion
      repl_1str(text,tags_comments,"\n","\\n");                            //  replace newlines with "\n"
      Ftagschanged = 1;
      save_filetags(image_file);                                           //  save new comments in file
   }

   else {
      zdialog_free(zd);                                                    //  other
      zdeditucomm = 0;
   }

   return 1;
}


/**************************************************************************/

//  search image tags for matching images

char     searchDateFrom[12] = "";                                          //  image search date range
char     searchDateTo[12] = "";
char     searchStarsFrom[4] = "";                                          //  image search stars range
char     searchStarsTo[4] = "";
zdialog  *zdsearchtags = 0;                                                //  search tags dialog


void m_searchtags(GtkWidget *, cchar *)
{
   void  searchtags_fixwidget(zdialog *zd, cchar * widgetname);
   int   searchtags_dialog_event(zdialog*, cchar *event);

   zfuncs::F1_help_topic = "search_tags";                                  //  v.10.8
   
/***
         date range    [_______] [_______] (yyyymmdd)
         stars range   [__] [__]
         search tags   [______________________________]
         user comments [______________________] [clear]
         /path*file*   [______________________] [clear]

         defined tags
          --------------------------------------------
         |                                            |
         |                                            |
         |                                            |
         |                                            |
         |                                            |
          --------------------------------------------
                                     [search] [cancel]
***/

   zdsearchtags = zdialog_new(ZTX("Search Tags, Comments, File Names"),mWin,Bsearch,Bcancel,null);
   
   zdialog_add_widget(zdsearchtags,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zdsearchtags,"vbox","vb1","hb1",0,"homog|space=3");
   zdialog_add_widget(zdsearchtags,"vbox","vb2","hb1",0,"homog|space=3|expand");

   zdialog_add_widget(zdsearchtags,"label","labD","vb1",ZTX("date range"));
   zdialog_add_widget(zdsearchtags,"label","labS","vb1",ZTX("stars range"));
   zdialog_add_widget(zdsearchtags,"label","labT","vb1",ZTX("search tags"));
   zdialog_add_widget(zdsearchtags,"label","labT","vb1",ZTX("comments"));
   zdialog_add_widget(zdsearchtags,"label","labF","vb1",ZTX("/path*/file*"));

   zdialog_add_widget(zdsearchtags,"hbox","hbD","vb2",0,"space=1");
   zdialog_add_widget(zdsearchtags,"entry","datefrom","hbD",0,"scc=10");
   zdialog_add_widget(zdsearchtags,"entry","dateto","hbD",0,"scc=10");
   zdialog_add_widget(zdsearchtags,"label","labD","hbD"," (yyyymmdd)");

   zdialog_add_widget(zdsearchtags,"hbox","hbS","vb2",0,"space=1|expand");
   zdialog_add_widget(zdsearchtags,"entry","starsfrom","hbS",0,"scc=2");
   zdialog_add_widget(zdsearchtags,"entry","starsto","hbS",0,"scc=2");

   zdialog_add_widget(zdsearchtags,"hbox","hbT","vb2",0,"space=1");
   zdialog_add_widget(zdsearchtags,"frame","frameT","hbT",0,"expand");
   zdialog_add_widget(zdsearchtags,"edit","searchtags","frameT",0,"expand");

   zdialog_add_widget(zdsearchtags,"hbox","hbC","vb2",0,"space=1|expand");
   zdialog_add_widget(zdsearchtags,"entry","searchcomms","hbC",0,"expand");

   zdialog_add_widget(zdsearchtags,"hbox","hbF","vb2",0,"space=1|expand");
   zdialog_add_widget(zdsearchtags,"entry","searchfiles","hbF",0,"expand");

   zdialog_add_widget(zdsearchtags,"hbox","hbM","dialog",0,"space=5");
   zdialog_add_widget(zdsearchtags,"radio","rmall","hbM",ZTX("match all tags"),"space=5");
   zdialog_add_widget(zdsearchtags,"label","lspace","hbM","","space=10");
   zdialog_add_widget(zdsearchtags,"radio","rmany","hbM",ZTX("match any tag"));

   zdialog_add_widget(zdsearchtags,"hbox","space","dialog",0,"space=3");
   zdialog_add_widget(zdsearchtags,"hbox","hbA1","dialog");
   zdialog_add_widget(zdsearchtags,"label","labdeftags","hbA1",ZTX("defined tags"),"space=5");
   zdialog_add_widget(zdsearchtags,"hbox","hbA2","dialog",0,"expand");
   zdialog_add_widget(zdsearchtags,"frame","frameA","hbA2",0,"space=5|expand");
   zdialog_add_widget(zdsearchtags,"scrwin","scrwinA","frameA",0,"expand");
   zdialog_add_widget(zdsearchtags,"edit","deftags","scrwinA",0,"expand");

   zdialog_resize(zdsearchtags,400,500);                                   //  start dialog
   zdialog_run(zdsearchtags,searchtags_dialog_event);

   searchtags_fixwidget(zdsearchtags,"deftags");                           //  setup tag selection via mouse
   searchtags_fixwidget(zdsearchtags,"searchtags");
   
   zdialog_stuff(zdsearchtags,"datefrom",searchDateFrom);                  //  stuff previous date range
   zdialog_stuff(zdsearchtags,"dateto",searchDateTo);
   zdialog_stuff(zdsearchtags,"starsfrom",searchStarsFrom);
   zdialog_stuff(zdsearchtags,"starsto",searchStarsTo);
   zdialog_stuff(zdsearchtags,"searchtags",tags_searchtags);               //  stuff previous search tags
   zdialog_stuff(zdsearchtags,"searchcomms",tags_searchcomms);             //  stuff previous search comments
   zdialog_stuff(zdsearchtags,"searchfiles",tags_searchfiles);             //  stuff previous search files
   zdialog_stuff(zdsearchtags,"rmall",0);
   zdialog_stuff(zdsearchtags,"rmany",1);                                  //  default is match any tags

   load_deftags();                                                         //  stuff defined tags into dialog
   deftags_stuff(zdsearchtags);
   
   return;
}


//  setup tag display widget for tag selection using mouse clicks

void searchtags_fixwidget(zdialog *zd, cchar * widgetname)
{
   void searchtags_mouse(GtkTextView *widget, GdkEventButton *event, cchar *widgetname);

   GtkWidget         *widget;
   GdkWindow         *gdkwin;

   widget = zdialog_widget(zd,widgetname);                                 //  make widget wrap text
   gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(widget),GTK_WRAP_WORD);
   gtk_text_view_set_editable(GTK_TEXT_VIEW(widget),0);                    //  disable widget editing

   gdkwin = gtk_text_view_get_window(GTK_TEXT_VIEW(widget),TEXTWIN);       //  cursor for tag selection
   gdk_window_set_cursor(gdkwin,arrowcursor);

   gtk_widget_add_events(widget,GDK_BUTTON_PRESS_MASK);                    //  connect mouse-click event
   G_SIGNAL(widget,"button-press-event",searchtags_mouse,widgetname)
}


//  search tags mouse-click event function
//  get clicked tag and add to or remove from tags_searchtags

void searchtags_mouse(GtkTextView *widget, GdkEventButton *event, cchar *widgetname)
{
   int      mpx, mpy, cc;
   
   if (event->type != GDK_BUTTON_PRESS) return;
   mpx = int(event->x);                                                    //  mouse click position
   mpy = int(event->y);

   cc = get_mouse_tag(widget,mpx,mpy,widgetname);                          //  tags_cliktag = clicked tag in list
   if (! cc) return;

   if (strEqu(widgetname,"deftags")) {
      add_tag(tags_cliktag,tags_searchtags,tagScc);                        //  add defined tag to search tags
      zdialog_stuff(zdsearchtags,"searchtags",tags_searchtags);
   }

   if (strEqu(widgetname,"searchtags")) {                                  //  v.10.6
      del_tag(tags_cliktag,tags_searchtags);                               //  remove tag from searchtags
      zdialog_stuff(zdsearchtags,"searchtags",tags_searchtags);            //  update dialog widgets
   }
   
   return;
}


//  dialog event and completion callback function

int searchtags_dialog_event(zdialog *zd, cchar *event)
{
   cchar          *pps, *ppf;
   char           resultsfile[tagrecl];
   char           *ppv, *file, *tags, *comms;
   char           filebuff[tagrecl], tagsbuff[tagrecl];
   char           date1[12], date2[12], lcfile[tagrecl];
   cchar          *comment_delims = " ,.!@#$%^&()_-+={}[]:;<>?/";          //  non-western languages?
   int            err, nfiles, iis, iif, stars, nmatch, nfail;
   int            date1cc, date2cc;
   int            Fmall, Fdates, Fcomms, Ffiles, Ftags, Fstars, Faccept;
   FILE           *fidr, *fidw;
   struct stat    statbuf;

   if (! zd->zstat) return 0;                                              //  wait for completion

   if (zd->zstat == 1)                                                     //  dialog complete
   {
      zdialog_fetch(zd,"datefrom",searchDateFrom,10);                      //  get search date range
      zdialog_fetch(zd,"dateto",searchDateTo,10);
      zdialog_fetch(zd,"starsfrom",searchStarsFrom,2);                     //  get search stars range
      zdialog_fetch(zd,"starsto",searchStarsTo,2);
      zdialog_fetch(zd,"searchtags",tags_searchtags,tagScc);               //  get search tags
      zdialog_fetch(zd,"searchcomms",tags_searchcomms,tagScc);             //  get search user* comments*
      zdialog_fetch(zd,"searchfiles",tags_searchfiles,tagScc);             //  get search /path*/file*
      zdialog_fetch(zd,"rmall",Fmall);                                     //  get match all/any option
   }
   
   zdialog_free(zdsearchtags);                                             //  kill dialog
   zdsearchtags = null;
   if (zd->zstat != 1) return 0;                                           //  cancelled

   strcpy(date1,"0000");                                                   //  defaults for missing dates
   strcpy(date2,"9999");                                                   //  (year 0000 to year 9999)
   date1cc = date2cc = 4;

   Fdates = 0;

   if (*searchDateFrom) {                                                  //  date from is given
      Fdates++;
      strncpy(date1,searchDateFrom,4);                                     //  convert format
      strncpy(date1+5,searchDateFrom+4,2);                                 //    yyyymmdd >> yyyy:mm:dd
      strncpy(date1+8,searchDateFrom+6,2);
      date1[4] = date1[7] = ':';
      date1[10] = 0;
      date1cc = strlen(date1);
   }
   if (*searchDateTo) {                                                    //  date to is given
      Fdates++;
      strncpy(date2,searchDateTo,4);
      strncpy(date2+5,searchDateTo+4,2);
      strncpy(date2+8,searchDateTo+6,2);
      date2[4] = date2[7] = ':';
      date2[10] = 0;
      date2cc = strlen(date2);
   }

   Fstars = 0;
   if (*searchStarsFrom || *searchStarsTo) Fstars = 1;                     //  stars given
   
   Ffiles = 0;
   if (! blank_null(tags_searchfiles)) Ffiles = 1;                         //  search /path*/file* given

   Fcomms = 0;
   if (! blank_null(tags_searchcomms)) Fcomms = 1;                         //  search user*comments* given

   Ftags = 0;
   if (! blank_null(tags_searchtags)) Ftags = 1;                           //  search tags given

   if (Ffiles) strToLower(tags_searchfiles);
   if (Ftags) strToLower(tags_searchtags);
   if (Fcomms) strToLower(tags_searchcomms);
   
   fidr = fopen(tags_index_file,"r");                                      //  read tags index file
   if (! fidr) goto nodeftags;
   
   snprintf(resultsfile,199,"%s/search_results",get_zuserdir());
   fidw = fopen(resultsfile,"w");                                          //  search results output file
   if (! fidw) goto writerror;

   nfiles = 0;                                                             //  count matching files found

next_file:                                                                 //  find next file record

   ppv = fgets_trim(filebuff,tagrecl,fidr,1);
   if (! ppv) goto file_EOF;
   if (! strnEqu(ppv,"file: ",6)) goto next_file;                          //  file: /dir.../filename.jpg

   file = ppv+6;
   err=stat(file,&statbuf);                                                //  check file exists
   if (err) goto next_file;
   if (! S_ISREG(statbuf.st_mode)) goto next_file;
   
   Faccept = 0;                                                            //  file passed no test yet

   if (! Ffiles) goto next_rec;

   strToLower(lcfile,file);                                                //  test for path*/file* match
   nmatch = nfail = 0;

   for (iis = 1; ; iis++)
   {
      pps = strField(tags_searchfiles,' ',iis);                            //  step thru search files
      if (! pps) break;
      if (MatchWild(pps,lcfile) == 0) nmatch++;
      else nfail++;
   }
   if (nmatch == 0) goto next_file;                                        //  no match
   if (Fmall && nfail) goto next_file;                                     //  no match to all files 
   Faccept++;                                                              //  accept
   
next_rec:

   ppv = fgets_trim(tagsbuff,tagrecl,fidr);                                //  get data record for this file
   if (! ppv) goto file_EOF;

   if (*tagsbuff < ' ') {                                                  //  EOL for this file
      if (Faccept) {
         fprintf(fidw,"%s\n",file);                                        //  output matching image file
         nfiles++;
      }
      goto next_file;
   }

   if (strnEqu(ppv,"date: ",6))                                            //  date: yyyy:mm:dd
   {
      if (! Fdates) goto next_rec;
      if (strncmp(ppv+6,date1,date1cc) < 0) goto next_file;
      if (strncmp(ppv+6,date2,date2cc) > 0) goto next_file;
      Faccept++;                                                           //  accept
      goto next_rec;
   }

   if (strnEqu(ppv,"comms: ",7))                                           //  comms: user comments ...   v.10.11
   {
      if (! Fcomms) goto next_rec;

      comms = ppv + 6;
      strToLower(comms);
      nmatch = nfail = 0;

      for (iis = 1; ; iis++)
      {
         pps = strField(tags_searchcomms,' ',iis);                         //  step thru search comments
         if (! pps) break;                                                 //  (blank delimited, wildcards)
         if (*pps == ' ') continue;

         for (iif = 1; ; iif++)                                            //  step thru user comments (words)
         {                                                                 //  (delimiters: blank comma period ";")
            ppf = strField(comms,comment_delims,iif);
            if (! ppf) { nfail++; break; }
            if (*ppf == ' ') continue;
            if (MatchWild(pps,ppf) == 0) { nmatch++; break; }              //  wildcard match 
         }
      }
      if (nmatch == 0) goto next_file;                                     //  no match to any word
      if (Fmall && nfail) goto next_file;                                  //  no match to all words
      Faccept++;                                                           //  accept
      goto next_rec;
   }

   if (strnEqu(ppv,"tags: ",6))                                            //  tags: xxxx; xxxxx; ...
   {
      if (! Ftags) goto next_rec;

      tags = ppv + 6;
      strToLower(tags);
      nmatch = nfail = 0;

      for (iis = 1; ; iis++)
      {
         pps = strField(tags_searchtags,';',iis);                          //  step thru search tags
         if (! pps) break;                                                 //  (semicolon delimited, wildcards)
         if (*pps == ' ') continue;

         for (iif = 1; ; iif++)                                            //  step thru file tags
         {                                                                 //  (semicolon delimited)
            ppf = strField(tags,';',iif);
            if (! ppf) { nfail++; break; }                                 //  count matches and fails
            if (*ppf == ' ') continue;
            if (MatchWild(pps,ppf) == 0) { nmatch++; break; }              //  wildcard match 
         }
      }

      if (nmatch == 0) goto next_file;                                     //  no match to any tag
      if (Fmall && nfail) goto next_file;                                  //  no match to all tags
      Faccept++;
      goto next_rec;
   }
   
   if (strnEqu(ppv,"stars: ",7))                                           //  stars: N
   {
      if (! Fstars) goto next_rec;
      stars = ppv[7];
      if (*searchStarsFrom && stars < *searchStarsFrom) goto next_file;
      if (*searchStarsTo && stars > *searchStarsTo) goto next_file;
      Faccept++;
      goto next_rec;
   }

   printf("unknown tags record: %s \n",tagsbuff);   
   goto next_rec;


file_EOF:

   fclose(fidr);
   
   err = fclose(fidw);
   if (err) goto writerror;
   
   if (! nfiles) {
      zmessageACK(MWIN,ZTX("No matching images found"));
      return 0;
   }
   
   image_gallery(resultsfile,"initF",0,m_gallery2);                        //  generate gallery of matching files
   image_gallery(0,"paint1");                                              //  show new image gallery window

   return 0;

nodeftags:
   zmessageACK(MWIN,ZTX("No tags index file"));
   return 0;

writerror:
   zmessLogACK(MWIN,ZTX("Search results file error %s"),strerror(errno));
   return 0;
}


/**************************************************************************/

//  convert between EXIF and fotoxx tag date formats
//  EXIF date: yyyy:mm:dd hh:mm:ss[.ss]
//  tag date: yyyymmdd
//  

void exif_tagdate(cchar *exifdate, char *tagdate)
{
   time_t      tnow;
   struct tm   *snow;
   
   if (! exifdate || strlen(exifdate) < 10) {                              //  bad EXIF date, use current date
      tnow = time(0);
      snow = localtime(&tnow);
      snprintf(tagdate,8,"%4d%02d%02d",
               snow->tm_year+1900, snow->tm_mon+1, snow->tm_mday);
      tagdate[8] = 0;
      return;
   }

   strncpy(tagdate,exifdate,4);                                            //  convert 
   strncpy(tagdate+4,exifdate+5,2);
   strncpy(tagdate+6,exifdate+8,2);
   tagdate[8] = 0;
   return;
}

void tag_exifdate(cchar *tagdate, char *exifdate)
{
   int         cc;
   time_t      tnow;
   struct tm   *snow;
   
   if (! tagdate || strlen(tagdate) < 4) {
      tnow = time(0);
      snow = localtime(&tnow);
      snprintf(exifdate,20,"%4d:%02d:%02d %02d:%02d:%02d",
               snow->tm_year+1900, snow->tm_mon+1, snow->tm_mday,
               snow->tm_hour, snow->tm_min, snow->tm_sec);
      exifdate[19] = 0;
      return;
   }

   strncpy(exifdate,tagdate,4);
   strcpy(exifdate+4,":01:01 00:00:00");
   cc = strlen(tagdate);
   if (cc >= 6) strncpy(exifdate+5,tagdate+4,2);
   if (cc >= 8) strncpy(exifdate+8,tagdate+6,2);
   exifdate[19] = 0;
   return;
}


/**************************************************************************
   Select an area within the current image.
   Subsequent edit functions are carried out within the area.
   Otherwise, edit functions apply to the entire image.
***************************************************************************/

//  functions and data for area selection using the mouse

void  sa_draw_mousefunc();                                                 //  line drawing functions
int   sa_nearpix(int mx, int my, int rad, int &npx, int &npy);             //  find nearest line-end
void  sa_draw_line(int px1, int py1, int px2, int py2);                    //  draw a connected line
void  sa_drawpixel(int px, int py);                                        //  draw 1 pixel unconditionally
void  sa_color_mousefunc();                                                //  color range select function
void  sa_radius_mousefunc();                                               //  radius select function
void  sa_nextseq();                                                        //  start next sequence number
void  sa_unselect_pixels();                                                //  remove current selection

//  other functions

void  sa_finish();                                                         //  finish - find inside pixels
void  sa_show(int flag);                                                   //  show or hide area
void  sa_enable();                                                         //  enable area
void  sa_disable();                                                        //  disable area
void  sa_invert();                                                         //  invert area
void  sa_delete();                                                         //  delete area
void  sa_edgecalc();                                                       //  calculate edge distances

int         sa_stat = 0;                                                   //  0/1/2/3 = none/edit/pause/complete
int         sa_mode = 0;                                                   //  1/2/3/4 = draw/follow/color/mouse
int         sa_active = 0;                                                 //  select area is complete and enabled

uint16      *sa_pixseq = 0;                                                //  mark pixels by select sequence no.
int         sa_currseq = 0;                                                //  current select sequence no.
int         sa_Ncurrseq = 0;                                               //  current sequence pixel count
char        *sa_pixselc = 0;                                               //  maps pixels (re)selected in current cycle

uint16      *sa_pixisin = 0;                                               //  mark pixels, 1=edge, 2+=inside (edge dist.)
int         sa_Npixel = 0;                                                 //  total select_area pixel count

int         sa_calced = 0;                                                 //  edge calculation done
int         sa_blend = 0;                                                  //  edge blend width

char        *sa_stackdirec = 0;                                            //  pixel search stack
int         *sa_stackii = 0;
int         sa_maxstack;
int         sa_Nstack;

int         sa_minx, sa_maxx;                                              //  enclosing rectangle for area
int         sa_miny, sa_maxy;

uint16      sa_endpx[10000], sa_endpy[10000];                              //  last pixel drawn per seqence no.
int         sa_initseq = 10;                                               //  initial sequence number (0/1/2 reserved)
int         sa_maxseq = 9999;                                              //  could go to 16 bits = 64K
int         sa_thresh;                                                     //  mouse pixel distance threshold
int         sa_mouseradius;                                                //  mouse selection radius
int         sa_mousex, sa_mousey;                                          //  mouse position in image
double      sa_colormatch;                                                 //  color range to match (0.001 to 1.0)
int         sa_colorradius;                                                //  color selection radius
uint16      sa_matchRGB[1000][3];                                          //  match colors, up to 1000
int         sa_Nmatch = 0;                                                 //  count of match colors
int         sa_firewall = 0;                                               //  selected pixels block propagation


/**************************************************************************/

//  user select area dialog
//  line drawing and selection by color range are combined                 //  v.9.7

void m_select(GtkWidget *, cchar *)                                        //  menu function
{
   int   select_dialog_event(zdialog *, cchar *event);                     //  dialog event and completion funcs

   int      cc;
   cchar    *title = ZTX("Select Area for Edits");
   cchar    *helptext = ZTX("Use F1 for context help");

   zfuncs::F1_help_topic = "select_area";                                  //  v.10.8

   if (! image_file) return;                                               //  no image
   if (zdsela) return;                                                     //  already active
   
   if (Fpreview) edit_fullsize();                                          //  use full-size pixmaps
   
   if (! Fpxm16) {                                                         //  create Fpxm16 if not already
      mutex_lock(&pixmaps_lock);
      Fpxm16 = f_load(image_file,16);
      mutex_unlock(&pixmaps_lock);
      if (! Fpxm16) return;
      SBupdate++;
   }

   //  (o) draw: freehand
   //  (o) draw: follow edge
   //  (o) select by color:  radius [__]  match [__]
   //  (o) select by mouse:  radius [__]
   //   Blend Width   [__]
   //  [Pause] [Resume] [Show] [Hide] [Color] [Finish]
   //  [Enable] [Disable] [Invert] [Delete] [Done]

   zdsela = zdialog_new(title,mWin,null);
   zdialog_add_widget(zdsela,"label","labhelp","dialog",helptext,"space=5");
   zdialog_add_widget(zdsela,"hbox","hb1","dialog");
   zdialog_add_widget(zdsela,"vbox","vb1","hb1",0,"homog");
   zdialog_add_widget(zdsela,"vbox","vb2","hb1",0,"homog");
   zdialog_add_widget(zdsela,"radio","rbdraw","vb1",ZTX("draw: freehand"));
   zdialog_add_widget(zdsela,"radio","rbfollow","vb1",ZTX("draw: follow edge"));
   zdialog_add_widget(zdsela,"radio","rbcolor","vb1",ZTX("select by color:"));
   zdialog_add_widget(zdsela,"radio","rbmouse","vb1",ZTX("select by mouse:"));

   zdialog_add_widget(zdsela,"label","space1","vb2");
   zdialog_add_widget(zdsela,"label","space2","vb2");

   zdialog_add_widget(zdsela,"hbox","hbcolor","vb2");
   zdialog_add_widget(zdsela,"label","labcolrad","hbcolor",ZTX("radius"),"space=5");
   zdialog_add_widget(zdsela,"spin","colorradius","hbcolor","0|100|1|10");
   zdialog_add_widget(zdsela,"label","space3","hbcolor",0,"space=5");
   zdialog_add_widget(zdsela,"label","labcolmatch","hbcolor",ZTX("match"),"space=5");
   zdialog_add_widget(zdsela,"spin","colormatch","hbcolor","0|100|1|90");

   zdialog_add_widget(zdsela,"hbox","hbmouserad","vb2");
   zdialog_add_widget(zdsela,"label","labmouserad","hbmouserad",ZTX("radius"),"space=5");
   zdialog_add_widget(zdsela,"spin","mouseradius","hbmouserad","0|300|1|20");

   zdialog_add_widget(zdsela,"hbox","hbblend","dialog",0,"space=5");
   zdialog_add_widget(zdsela,"label","labblend","hbblend",Bblendwidth,"space=9");
   zdialog_add_widget(zdsela,"spin","blendwidth","hbblend","0|500|1|0");
   zdialog_add_widget(zdsela,"check","firewall","hbblend",ZTX("color select firewall"),"space=20");

   zdialog_add_widget(zdsela,"hbox","hbb1","dialog",0,"space=8");
   zdialog_add_widget(zdsela,"button","pause","hbb1",Bpause);
   zdialog_add_widget(zdsela,"button","resume","hbb1",Bresume);
   zdialog_add_widget(zdsela,"button","show","hbb1",Bshow);
   zdialog_add_widget(zdsela,"button","hide","hbb1",Bhide);
   zdialog_add_widget(zdsela,"button","color","hbb1",Bcolor);
   zdialog_add_widget(zdsela,"button","finish","hbb1",Bfinish);

   zdialog_add_widget(zdsela,"hbox","hbb2","dialog",0,"space=8");          //  removed from menu         v.10.9
   zdialog_add_widget(zdsela,"button","enable","hbb2",Benable);            //    and added to dialog
   zdialog_add_widget(zdsela,"button","disable","hbb2",Bdisable);
   zdialog_add_widget(zdsela,"button","invert","hbb2",Binvert);
   zdialog_add_widget(zdsela,"button","delete","hbb2",Bdelete);
   zdialog_add_widget(zdsela,"button","done","hbb2",Bdone);

   zdialog_run(zdsela,select_dialog_event);                                //  run dialog - parallel

   sa_colormatch = 90;                                                     //  initial color match level
   sa_colorradius = 10;                                                    //  initial color match select radius
   sa_mouseradius = 20;                                                    //  initial mouse select radius
   sa_blend = 0;
   sa_firewall = 0;

   if (! sa_pixseq) {                                                      //  start new area if none    v.10.8
      cc = Fww * Fhh;
      sa_pixseq = (uint16 *) zmalloc(2*cc,"pixseq");
      memset(sa_pixseq,0,2*cc);
      sa_currseq = sa_Ncurrseq = 0;                                        //  v.10.8
   }

   sa_mode = 1;                                                            //  default mode = freehand draw    v.10.8
   sa_stat = 1;                                                            //  status = active edit
   sa_show(1);                                                             //  show existing area
   mouseCBfunc = sa_draw_mousefunc;                                        //  connect mouse function
   Mcapture++;
   set_cursor(drawcursor);                                                 //  set draw cursor

   return;
}


//  dialog event and completion callback function

int select_dialog_event(zdialog *zd, cchar *event)
{
   static int  ignore1 = 0;
   int         ii, cc;
   
   if (strEqu(event,"done") || zd->zstat)                                  //  done or cancel
   {
      mouseCBfunc = 0;                                                     //  disconnect mouse function
      Mcapture = 0;
      set_cursor(0);                                                       //  restore normal cursor
      paint_toparc(2);                                                     //  erase radius circle
      zdialog_free(zdsela);                                                //  kill dialog
      zdsela = null;
      return 0;
   }
   
   if (! sa_pixseq) {                                                      //  start new area if none    v.10.8
      cc = Fww * Fhh;
      sa_pixseq = (uint16 *) zmalloc(2*cc,"pixseq");
      memset(sa_pixseq,0,cc*sizeof(uint16));
      sa_currseq = sa_Ncurrseq = 0;                                        //  v.10.8
   }

   ii = strcmpv(event,"rbdraw","rbfollow","rbcolor","rbmouse",null);       //  which radio button
   if (ii) {
      sa_mode = ii;
      sa_stat = 1;                                                         //  status = active edit   v.10.8
      sa_show(1);                                                          //  show area
      sa_Npixel = sa_blend = sa_calced = sa_active = 0;
   }

   if (strEqu(event,"mouseradius"))                                        //  mouse selection radius
      zdialog_fetch(zd,"mouseradius",sa_mouseradius);

   if (strEqu(event,"colorradius"))                                        //  color match selection radius
      zdialog_fetch(zd,"colorradius",sa_colorradius);

   if (strEqu(event,"colormatch"))                                         //  color match range, 0 to 99.9
      zdialog_fetch(zdsela,"colormatch",sa_colormatch);
   
   if (strEqu(event,"firewall"))                                           //  color select firewall on/off   v.10.11
      zdialog_fetch(zdsela,"firewall",sa_firewall);

   if (strEqu(event,"pause")) sa_stat = 2;                                 //  pause edit

   if (strEqu(event,"resume")) {
      sa_stat = 1;                                                         //  resume edit
      sa_Npixel = sa_blend = sa_calced = sa_active = 0;
      sa_show(1);
   }

   if (strEqu(event,"show")) sa_show(1);                                   //  show area

   if (strEqu(event,"hide")) {
      sa_show(0);                                                          //  hide area
      sa_stat = 2;                                                         //  pause mode       v.10.8
   }

   if (strEqu(event,"color")) {                                            //  switch colors    v.9.8
      if (FareaRGB == 'R') FareaRGB = 'G';
      else if (FareaRGB == 'G') FareaRGB = 'B';
      else FareaRGB = 'R';
      sa_show(1);
   }

   if (strEqu(event,"finish")) sa_finish();                                //  finish (finalize) area
   if (strEqu(event,"delete")) sa_delete();                                //  delete area
   if (strEqu(event,"enable")) sa_enable();                                //  enable area
   if (strEqu(event,"disable")) sa_disable();                              //  disable area
   if (strEqu(event,"invert")) sa_invert();                                //  invert area
   
   if (strEqu(event,"blendwidth") && sa_active) {                          //  blend width changed
      if (ignore1) ignore1 = 0;                                            //  ignore once      v.9.4
      else if (sa_Npixel) {
         sa_edgecalc();                                                    //  edge calc. unless already
         if (sa_calced && zdedit) {
            zdialog_fetch(zd,"blendwidth",sa_blend);                       //  update sa_blend
            zdialog_send_event(zdedit,event);                              //  notify active edit dialog
         }
         else ignore1 = 1;                                                 //  edge calc failed or killed   v.9.4
      }                                                                    //  (gtk bug workaround, repeated event)
      return 0;                                                            //  don't change mouse   v.10.11
   }

   if (sa_stat == 1) {                                                     //  active edit mode
      if (sa_mode == 1) mouseCBfunc = sa_draw_mousefunc;                   //  connect mouse function
      if (sa_mode == 2) mouseCBfunc = sa_draw_mousefunc;
      if (sa_mode == 3) mouseCBfunc = sa_color_mousefunc;
      if (sa_mode == 4) mouseCBfunc = sa_radius_mousefunc;
      Mcapture++;
      set_cursor(drawcursor);                                              //  set draw cursor
      if (sa_mode < 3) paint_toparc(2);                                    //  erase radius circle
   }
   else {
      mouseCBfunc = 0;                                                     //  disconnect mouse function
      Mcapture = 0;
      set_cursor(0);                                                       //  normal cursor
      paint_toparc(2);                                                     //  erase radius circle
   }

   return 0;
}


//  select area mouse function - draw lines

void sa_draw_mousefunc()
{
   void sa_follow_edge(int mx1, int my1, int mx2, int my2);

   int         mx1, my1, mx2, my2;
   int         npdist, npx, npy;
   int         ii, click, newseq, thresh;
   static int  drag = 0, mdx0, mdy0, mdx1, mdy1;

   sa_thresh = 4.0 / Mscale + 1;                                           //  mouse pixel distance threshold
   click = newseq = 0;
   
   if (LMclick || Mxdrag || Mydrag)                                        //  left mouse click or mouse drag
   {
      if (LMclick)                                                         //  left mouse click
      {
         LMclick = 0;
         mx1 = mx2 = Mxclick;                                              //  click position
         my1 = my2 = Myclick;
         newseq++;
         click++;
         drag = 0;
      }
      else                                                                 //  drag motion
      {
         if (Mxdown != mdx0 || Mydown != mdy0) {                           //  new drag initiated
            mdx0 = mdx1 = Mxdown;
            mdy0 = mdy1 = Mydown;
            newseq++;
         }
         mx1 = mdx1;                                                       //  drag start
         my1 = mdy1;
         mx2 = Mxdrag;                                                     //  drag position
         my2 = Mydrag;
         mdx1 = mx2;                                                       //  next drag start
         mdy1 = my2;
         drag++;
         click = 0;
      }
      
      if (Mbutton == 3)                                                    //  right mouse >> erase
      {
         while (true) {
            thresh = sa_thresh;
            npdist = sa_nearpix(mx2,my2,thresh,npx,npy);
            if (! npdist) break;
            ii = npy * Fww + npx;
            sa_pixseq[ii] = 0;
         }
         mwpaint2();
         return;
      }

      if (sa_currseq > sa_maxseq-2) {
         zmessageACK(MWIN,ZTX("exceed %d edits"),sa_maxseq);               //  cannot continue
         return;
      }
      
      if (sa_currseq == 0 && newseq)                                       //  1st pixel(s) of 1st sequence
      {
         sa_nextseq();                                                     //  set next (1st) sequence no.  v.10.8
         sa_draw_line(mx1,my1,mx2,my2);                                    //  draw initial pixel or line
         sa_endpx[sa_currseq] = mx2;
         sa_endpy[sa_currseq] = my2;
         return;
      }
      
      if (click) {
         mx1 = sa_endpx[sa_currseq];                                       //  prior sequence end pixel
         my1 = sa_endpy[sa_currseq];                                       //  (before this click)
      }
      
      if (drag) {
         if (newseq) thresh = 2 * sa_thresh;                               //  new drag threshold
         else thresh = 5 * sa_thresh;                                      //  continuation drag threshold
         npx = sa_endpx[sa_currseq];                                       //  distance from prior end pixel
         npy = sa_endpy[sa_currseq];                                       //    (before this drag)
         if (abs(mx1-npx) < thresh && abs(my1-npy) < thresh) {
            mx1 = sa_endpx[sa_currseq];                                    //  if < threshold, connect this
            my1 = sa_endpy[sa_currseq];                                    //    drag to prior drag or click
         }
      }

      if (newseq || drag > 50) {
         sa_nextseq();                                                     //  set next sequence no.     v.10.8
         drag = 1;                                                         //  drag length within sequence
      }
      
      if (sa_mode == 2) sa_follow_edge(mx1,my1,mx2,my2);                   //  follow edge or draw line
      else sa_draw_line(mx1,my1,mx2,my2);                                  //    from end pixel to mouse
      
      sa_endpx[sa_currseq] = mx2;                                          //  set end pixel for this sequence
      sa_endpy[sa_currseq] = my2;
   }

   else if (RMclick)                                                       //  right mouse click
   {
      RMclick = 0;
      sa_unselect_pixels();                                                //  remove latest selection   v.10.8
      mwpaint2();
   }
   
   return;
}


//  Find the nearest drawn pixel within a radius of a given pixel.
//  Returns distance to pixel, or zero if nothing found.
//  Returns 1 for adjacent or diagonally adjacent pixel.

int sa_nearpix(int mx, int my, int rad2, int &npx, int &npy)
{
   int      ii, rad, qx, qy, dx, dy;
   int      seq, mindist, dist;

   npx = npy = 0;
   mindist = (rad2+1) * (rad2+1);

   for (rad = 1; rad <= rad2; rad++)                                       //  seek neighbors within range
   {
      if (rad * rad > mindist) break;                                      //  can stop searching now

      for (qx = mx-rad; qx <= mx+rad; qx++)                                //  search within rad
      for (qy = my-rad; qy <= my+rad; qy++)
      {
         if (qx != mx-rad && qx != mx+rad &&                               //  exclude within rad-1
             qy != my-rad && qy != my+rad) continue;                       //  (already searched)
         if (qx < 0 || qx >= Fww) continue;
         if (qy < 0 || qy >= Fhh) continue;
         ii = qy * Fww + qx;
         seq = sa_pixseq[ii];
         if (! seq) continue;
         dx = (mx - qx) * (mx - qx);                                       //  found pixel
         dy = (my - qy) * (my - qy);
         dist = dx + dy;                                                   //  distance**2
         if (dist < mindist) {
            mindist = dist;
            npx = qx;                                                      //  save nearest pixel found
            npy = qy;
         }
      }
   }
   
   if (npx + npy) return sqrt(mindist) + 0.5;
   return 0;
}


//  draw a line between two given pixels
//  add all in-line pixels to sa_pixseq[]
  
void sa_draw_line(int px1, int py1, int px2, int py2)
{
   void  sa_draw1pix(int px, int py);

   int      pxm, pym;
   double   slope;
   
   if (px1 == px2 && py1 == py2) {                                         //  only one pixel
      sa_draw1pix(px1,py1);
      return;
   }
   
   if (abs(py2 - py1) > abs(px2 - px1)) {
      slope = 1.0 * (px2 - px1) / (py2 - py1);
      if (py2 > py1) {
         for (pym = py1; pym <= py2; pym++) {
            pxm = round(px1 + slope * (pym - py1));
            sa_draw1pix(pxm,pym);
         }
      }
      else {
         for (pym = py1; pym >= py2; pym--) {
            pxm = round(px1 + slope * (pym - py1));
            sa_draw1pix(pxm,pym);
         }
      }
   }
   else {
      slope = 1.0 * (py2 - py1) / (px2 - px1);
      if (px2 > px1) {
         for (pxm = px1; pxm <= px2; pxm++) {
            pym = round(py1 + slope * (pxm - px1));
            sa_draw1pix(pxm,pym);
         }
      }
      else {
         for (pxm = px1; pxm >= px2; pxm--) {
            pym = round(py1 + slope * (pxm - px1));
            sa_draw1pix(pxm,pym);
         }
      }
   }

   return;
}


//  draw one pixel only if not already drawn

void sa_draw1pix(int px, int py)
{
   int ii = Fww * py + px;
   if (sa_pixseq[ii]) return;
   sa_pixseq[ii] = sa_currseq;
   sa_Ncurrseq++;                                                          //  v.10.8
   sa_drawpixel(px,py);
   return;
}


//  draw one pixel using current FareaRGB color

void sa_drawpixel(int px, int py)
{
   int            qx, qy;
   static int     pqx, pqy;

   if (FareaRGB == 'G') gdk_gc_set_foreground(gdkgc,&green);               //  v.9.8
   else if (FareaRGB == 'B') gdk_gc_set_foreground(gdkgc,&black);
   else gdk_gc_set_foreground(gdkgc,&red);

   qx = Mscale * (px-Iorgx) + 0.5;                                         //  image to window space
   qy = Mscale * (py-Iorgy) + 0.5;

   if (qx == pqx && qy == pqy) {                                           //  avoid redundant points   v.9.7
      gdk_gc_set_foreground(gdkgc,&black);                                 //  bugfix   v.10.8
      return;
   }

   pqx = qx;
   pqy = qy;

   gdk_draw_point(drWin->window,gdkgc,qx+Dorgx,qy+Dorgy);                  //  draw pixel

   gdk_gc_set_foreground(gdkgc,&black);
   return;
}


//  Find series of edge pixels from px1/py1 to px2/py2 and connect them together.

void sa_follow_edge(int px1, int py1, int px2, int py2)                    //  v.10.8
{
   double   sa_get_contrast(int px, int py);
   
   double   px3, py3, px4, py4, px5, py5, px6, py6;
   double   dx, dy, dist, contrast, maxcontrast;
   
   px3 = px1;                                                              //  p3 progresses from p1 to p2
   py3 = py1;
   
   while (true)
   {
      dx = px2 - px3;
      dy = py2 - py3;
      
      dist = sqrt(dx * dx + dy * dy);                                      //  last segment
      if (dist < 3) break;
      
      px4 = px3 + dx / dist;                                               //  p4 = p3 moved toward p2
      py4 = py3 + dy / dist;
      
      maxcontrast = 0;
      px6 = px4;
      py6 = py4;

      for (int ii = -2; ii <= +2; ii++)                                    //  p5 points are in a line through p4
      {                                                                    //    and perpendicular to p4 - p2
         px5 = px4 + ii * dy / dist;
         py5 = py4 - ii * dx / dist;
         contrast = sa_get_contrast(px5,py5);
         contrast *= (7 - abs(ii));                                        //  favor points closer together   v.10.9
         if (contrast > maxcontrast) {
            px6 = px5;                                                     //  p6 = highest contrast point in p5
            py6 = py5;
            maxcontrast = contrast;
         }
      }
      
      sa_draw_line(px3,py3,px6,py6);                                       //  draw p3 to p6

      px3 = px6;                                                           //  next p3
      py3 = py6;
   }
   
   sa_draw_line(px3,py3,px2,py2);
   return;
}


//  Find max. contrast between neighbors on opposite sides of given pixel

double sa_get_contrast(int px, int py)                                     //  v.10.8
{
   int         map[4][2] = { {1, 0}, {1, 1}, {0, 1}, {-1, 1} };
   int         ii, qx, qy;
   uint16      *pix1, *pix2;
   double      red, green, blue;
   double      contrast, maxcontrast = 0;
   double      f65k = 1.0 / 65535.0;
   
   if (px < 1 || px > Fww-2) return 0;                                     //  avoid edge pixels
   if (py < 1 || py > Fhh-2) return 0;
   
   for (ii = 0; ii < 4; ii++)                                              //  compare pixels around target
   {                                                                       //  e.g. (px-1,py) to (px+1,py)
      qx = map[ii][0];
      qy = map[ii][1];
      pix1 = PXMpix(Fpxm16,px+qx,py+qy);
      pix2 = PXMpix(Fpxm16,px-qx,py-qy);
      red = f65k * abs(pix1[0] - pix2[0]);
      green = f65k * abs(pix1[1] - pix2[1]);
      blue = f65k * abs(pix1[2] - pix2[2]);
      contrast = (1.0 - red) * (1.0 - green) * (1.0 - blue);               //  no contrast = 1.0
      contrast = 1.0 - contrast;                                           //  max. contrast = 1.0
      if (contrast > maxcontrast) maxcontrast = contrast;
   }
   
   return maxcontrast;
}


//  select area by color range - mouse function

void sa_color_mousefunc()
{
   void sa_select_color_pixels();

   int         radius, cc;
   static int  mxdown, mydown, drag = 0;

   radius = sa_colorradius;                                                //  color match selection radius

   if (! radius) paint_toparc(2);                                          //  no radius, no circle
   else {
      toparcx = Mxposn - radius;                                           //  draw radius outline circle
      toparcy = Myposn - radius;
      toparcw = toparch = 2 * radius;
      Ftoparc = 1;
      paint_toparc(3);
   }
   
   if (sa_stackdirec) zfree(sa_stackdirec);                                //  allocate pixel search stack
   if (sa_stackii) zfree(sa_stackii);
   cc = Fww * Fhh;   
   sa_stackdirec = zmalloc(cc,"stack.direc");
   sa_stackii = (int *) zmalloc(4*cc,"stack.ii");
   sa_maxstack = cc;
   sa_Nstack = 0;

   sa_mousex = sa_mousey = 0;

   if (LMclick) {                                                          //  get mouse position at click
      sa_mousex = Mxclick;
      sa_mousey = Myclick;
      LMclick = 0;
      sa_nextseq();                                                        //  set next sequence no.     v.10.8
      drag = 1;
   }
      
   if (Mxdrag || Mydrag) {                                                 //  get mouse drag position
      sa_mousex = Mxdrag;
      sa_mousey = Mydrag;
      Mxdrag = Mydrag = 0;

      if (Mxdown != mxdown || Mydown != mydown) {                          //  detect if new drag started
         mxdown = Mxdown;
         mydown = Mydown;
         sa_nextseq();                                                     //  set next sequence no.     v.10.8
         drag = 1;
      }
      else if (++drag > 50) {                                              //  limit work per sequence no.
         sa_nextseq();                                                     //  set next sequence no.     v.10.8
         drag = 1;
      }
   }
   
   if (sa_mousex || sa_mousey) {
      sa_select_color_pixels();                                            //  accumulate pixels
      mwpaint2();
   }

   if (RMclick) {
      RMclick = 0;
      sa_unselect_pixels();                                                //  remove latest selection   v.10.8
      mwpaint2();
   }

   return;
}


//  find all contiguous pixels within the specified range of colors to match
//  overhauled v.10.8: 
//    - test against all colors within radius
//    - propagate through previously selected pixels

void sa_select_color_pixels()
{
   void  sa_select_color_pushstack(int px, int py, char direc);

   int         ii, kk, cc, px, py, rx, ry;
   int         radius, radius2;
   uint16      *matchpix;
   double      match1, match2, ff = 1.0 / 65536.0;
   double      dred, dgreen, dblue;
   char        direc;
   
   match1 = 0.01 * sa_colormatch;                                          //  color match level, 0.01 to 1.0

   px = sa_mousex;
   py = sa_mousey;
   if (px < 0 || px >= Fww) return;                                        //  mouse outside image
   if (py < 0 || py >= Fhh) return;

   cc = Fww * Fhh;                                                         //  allocate selection map
   sa_pixselc = zmalloc(cc,"pixels.selected");
   memset(sa_pixselc,0,cc);

   radius = sa_colorradius;                                                //  use radius from dialog     v.10.8
   if (! radius) radius = 1 + 1.0 / Mscale;                                //  or small default radius
   radius2 = radius * radius;

   sa_Nmatch = 0;                                                          //  match color count

   for (rx = -radius; rx <= radius; rx++)                                  //  loop every pixel in radius of mouse
   for (ry = -radius; ry <= radius; ry++)
   {
      if (rx * rx + ry * ry > radius2) continue;                           //  outside radius
      px = sa_mousex + rx;
      py = sa_mousey + ry;
      if (px < 0 || px >= Fww) continue;                                   //  off the image edge
      if (py < 0 || py >= Fhh) continue;

      matchpix = PXMpix(Fpxm16,px,py);                                     //  get color at mouse position
      
      for (ii = 0; ii < sa_Nmatch; ii++)                                   //  see if color is already included
      {
         dred =   ff * abs(sa_matchRGB[ii][0] - matchpix[0]);              //  0 = perfect match
         dgreen = ff * abs(sa_matchRGB[ii][1] - matchpix[1]);
         dblue =  ff * abs(sa_matchRGB[ii][2] - matchpix[2]);
         match2 = (1.0 - dred) * (1.0 - dgreen) * (1.0 - dblue);           //  1 = perfect match
         if (match2 >= match1) break;                                      //  matches close enough
      }
      
      if (ii == sa_Nmatch) {                                               //  no close match 
         sa_matchRGB[ii][0] = matchpix[0];                                 //  add new match color to list
         sa_matchRGB[ii][1] = matchpix[1];
         sa_matchRGB[ii][2] = matchpix[2];
         sa_Nmatch++;
         if (sa_Nmatch == 1000) goto matchfull;                            //  capacity limit
      }
   }
   
matchfull:

   px = sa_mousex;                                                         //  pixel at mouse
   py = sa_mousey;
   ii = Fww * py + px;

   sa_pixselc[ii] = 1;                                                     //  pixel is in current selection

   sa_stackii[0] = ii;                                                     //  put 1st pixel into stack
   sa_stackdirec[0] = 'r';                                                 //  direction = right
   sa_Nstack = 1;                                                          //  stack count

   while (sa_Nstack)
   {
      kk = sa_Nstack - 1;                                                  //  get last pixel in stack
      ii = sa_stackii[kk];
      direc = sa_stackdirec[kk];
      
      py = ii / Fww;                                                       //  reconstruct px, py
      px = ii - Fww * py;

      if (direc == 'x') {                                                  //  no neighbors left to check
         sa_Nstack--;
         continue;
      }
      
      if (direc == 'r') {                                                  //  push next right pixel into stack
         sa_select_color_pushstack(px,py,'r');                             //    if color within range
         sa_stackdirec[kk] = 'l';                                          //  this pixel next direction to look
         continue;
      }

      if (direc == 'l') {                                                  //  or next left pixel
         sa_select_color_pushstack(px,py,'l');
         sa_stackdirec[kk] = 'a';
         continue;
      }

      if (direc == 'a') {                                                  //  or next ahead pixel
         sa_select_color_pushstack(px,py,'a');
         sa_stackdirec[kk] = 'x';
         continue;
      }
   }

   sa_Ncurrseq = 0;

   for (ii = 0; ii < Fww * Fhh; ii++)                                      //  find all selected pixels
   {
      if (sa_pixselc[ii] && ! sa_pixseq[ii]) {                             //  if selected for the first time,
         sa_pixseq[ii] = sa_currseq;                                       //    map pixel to current sequence
         sa_Ncurrseq++;                                                    //  current sequence pixel count
      }
   }

   zfree(sa_pixselc);                                                      //  free memory
   return;
}      


//  push pixel into stack memory if its color is within range
//  and not already mapped to a prior sequence number

void sa_select_color_pushstack(int px, int py, char direc)
{
   int         ii, kk, ppx, ppy, npx, npy;
   uint16      *matchpix;
   double      match1, match2, ff = 1.0 / 65536.0;
   double      dred, dgreen, dblue;

   match1 = 0.01 * sa_colormatch;                                          //  color match level, 0.01 to 1.0

   if (sa_Nstack == sa_maxstack) return;                                   //  stack is full

   if (sa_Nstack > 1) {
      kk = sa_Nstack - 2;                                                  //  get prior pixel in stack
      ii = sa_stackii[kk];
      ppy = ii / Fww;
      ppx = ii - ppy * Fww;
   }
   else {
      ppx = px - 1;                                                        //  if only one, assume prior = left
      ppy = py;
   }
   
   if (direc == 'r') {                                                     //  get pixel in direction right
      npx = px + ppy - py;
      npy = py + px - ppx;
   }
   else if (direc == 'l') {                                                //  or left
      npx = px + py - ppy;
      npy = py + ppx - px;
   }
   else if (direc == 'a') {                                                //  or ahead
      npx = px + px - ppx;
      npy = py + py - ppy;
   }
   else npx = npy = -1;                                                    //  stop warning
   
   if (npx < 0 || npx >= Fww) return;                                      //  pixel off the edge
   if (npy < 0 || npy >= Fhh) return;
   
   ii = npy * Fww + npx;
   if (sa_pixselc[ii]) return;                                             //  already in current selection   v.10.8

   if (sa_firewall && sa_pixseq[ii]) return;                               //  aleady selected, firewall mode  v.10.11

   matchpix = PXMpix(Fpxm16,npx,npy);
   for (kk = 0; kk < sa_Nmatch; kk++) {                                    //  compare pixel RGB to match colors
      dred =   ff * abs(sa_matchRGB[kk][0] - matchpix[0]);                 //  v.10.8
      dgreen = ff * abs(sa_matchRGB[kk][1] - matchpix[1]);
      dblue =  ff * abs(sa_matchRGB[kk][2] - matchpix[2]);
      match2 = (1.0 - dred) * (1.0 - dgreen) * (1.0 - dblue);              //  1 = perfect match
      if (match2 >= match1) break;                                         //  within range
   }
   if (kk == sa_Nmatch) return;                                            //  not within range of any color

   sa_pixselc[ii] = 1;                                                     //  map pixel to current selection   v.10.8
   
   kk = sa_Nstack++;                                                       //  put pixel into stack
   sa_stackii[kk] = ii;
   sa_stackdirec[kk] = 'r';                                                //  direction = right

   return;
}


//  select or un-select all pixels within radius - mouse function

void sa_radius_mousefunc()
{
   int      ii, radius, radius2;
   int      px, py, rx, ry;
   
   radius = sa_mouseradius;                                                //  pixel selection radius
   radius2 = radius * radius;

   if (! radius) {
      paint_toparc(2);                                                     //  no radius, no circle
      return;
   }

   toparcx = Mxposn - radius;                                              //  draw radius outline circle
   toparcy = Myposn - radius;
   toparcw = toparch = 2 * radius;
   Ftoparc = 1;
   paint_toparc(3);

   if (LMclick || RMclick) {                                               //  mouse click
      sa_nextseq();                                                        //  set next sequence no.     v.10.8
      LMclick = RMclick = 0;
   }

   if (Mbutton != 1 && Mbutton != 3) {                                     //  button released
      sa_nextseq();                                                        //  set next sequence no.     v.10.8
      return;                                                              //  (if some pixels mapped)
   }
   
   for (rx = -radius; rx <= radius; rx++)                                  //  loop every pixel in radius
   for (ry = -radius; ry <= radius; ry++)
   {
      if (rx * rx + ry * ry > radius2) continue;                           //  outside radius
      px = Mxposn + rx;
      py = Myposn + ry;
      if (px < 0 || px >= Fww) continue;                                   //  off the image edge
      if (py < 0 || py >= Fhh) continue;

      ii = Fww * py + px;
      
      if (Mbutton == 3)                                                    //  right mouse button
         sa_pixseq[ii] = 0;                                                //  remove pixel from select area

      if (Mbutton == 1)                                                    //  left mouse button
      {
         if (sa_pixseq[ii]) continue;                                      //  pixel already selected

         if (sa_Ncurrseq > 1000)                                           //  start new sequence no.
            sa_nextseq();                                                  //  set next sequence no.     v.10.8

         sa_pixseq[ii] = sa_currseq;                                       //  map pixel to current sequence
         sa_Ncurrseq++;
      }
   }

   mwpaint2();
}


//  set next sequence number for pixels about to be selected

void sa_nextseq()                                                          //  v.10.8
{
   if (sa_Ncurrseq > 0) sa_currseq++;                                      //  increase only if some pixels mapped
   if (sa_currseq < sa_initseq) sa_currseq = sa_initseq;                   //  start at initial value
   sa_Ncurrseq = 0;
   return;
}


//  un-select all pixels mapped to current sequence number
//  reduce sequence number and set pixel count = 1

void sa_unselect_pixels()
{
   if (! sa_currseq) return;                                               //  no pixels mapped, do nothing
   
   for (int ii = 0; ii < Fww * Fhh; ii++)
   {
      if (sa_pixseq[ii] != sa_currseq) continue;                           //  unmap current selection
      sa_pixseq[ii] = 0;
   }
   
   if (sa_currseq > sa_initseq) {                                          //  reduce sequence no.    v.10.8
      sa_currseq--;
      sa_Ncurrseq = 1;                                                     //  unknown but > 0
   }
   else  sa_Ncurrseq = 0;                                                  //  initial sequence no. reached
   
   return;
}


//  Finish select area - map pixels enclosed by edge pixels 
//    into sa_pixisin[]: 1=edge, 2=inside, [ii]=py*Fww+px
//  total count = sa_Npixel

namespace sa_finish_names
{
   zdialog     *finzd = 0;
   int         finish_fails;
   int         pxmin, pxmax, pymin, pymax;
}

void sa_finish()                                                           //  overhauled          v.9.7
{
   using namespace sa_finish_names;

   void sa_finish_mousefunc();
   int  sa_finish_dialog_event(zdialog *, cchar *event);

   cchar  *fmess = ZTX("Search all areas for edge and inside pixels. \n"
                       "Click inside each enclosed area in sequence.");
   int         ii, kk, cc, px, py, npix;

   if (! sa_pixseq) return;                                                //  nothing selected
   sa_Npixel = sa_active = 0;                                              //  area disabled, unfinished
   set_cursor(0);                                                          //  normal cursor
   
   sa_show(1);                                                             //  v.10.8

   if (sa_pixisin) zfree(sa_pixisin);                                      //  allocate pixisin[]
   cc = Fww * Fhh * 2;
   sa_pixisin = (uint16 *) zmalloc(cc,"pixisin");
   memset(sa_pixisin,0,cc);                                                //  clear included pixel map

   pxmin = Fww;
   pxmax = 0;
   pymin = Fhh;
   pymax = 0;

   for (ii = 0; ii < Fww * Fhh; ii++)                                      //  get enclosing rectangle
   {                                                                       //    for selected area
      if (! sa_pixseq[ii]) continue;
      py = ii / Fww;
      px = ii - Fww * py;
      if (px > pxmax) pxmax = px;
      if (px < pxmin) pxmin = px;
      if (py > pymax) pymax = py;
      if (py < pymin) pymin = py;
   }
   
   sa_minx = pxmin;                                                        //  save enclosing rectangle     v.10.1
   sa_maxx = pxmax;
   sa_miny = pymin;
   sa_maxy = pymax;
   
   pxmin -= 10;                                                            //  add margins if possible
   if (pxmin < 0) pxmin = 0;                                               //  (enhance gap visibility if
   pxmax += 10;                                                            //    finish algorithm fails)
   if (pxmax > Fww-1) pxmax = Fww - 1;
   pymin -= 10;
   if (pymin < 0) pymin = 0;
   pymax += 10;
   if (pymax > Fhh-1) pymax = Fhh - 1;
   
   cc = (pxmax-pxmin) * (pymax-pymin);                                     //  allocate stack memory
   if (sa_stackdirec) zfree(sa_stackdirec);
   sa_stackdirec = zmalloc(cc,"stack.direc");
   if (sa_stackii) zfree(sa_stackii);
   sa_stackii = (int *) zmalloc(cc*4,"stack.ii");
   sa_maxstack = cc;

   for (ii = npix = 0; ii < Fww * Fhh; ii++)                               //  find pixels already included in area
   {
      if (! sa_pixseq[ii]) continue;
      sa_pixisin[ii] = 1;                                                  //  populate pixisin[]
      npix++;
   }

   if (npix < 20) return;                                                  //  ridiculous

   for (py = 0; py < Fhh; py++)                                            //  find edge pixels
   for (px = 0; px < Fww; px++)
   {
      ii = py * Fww + px;
      if (! sa_pixisin[ii]) continue;                                      //  outside of selected area

      if (py < 1 || py > Fhh-1) goto edgepixel;                            //  bugfix   v.10.8
      if (px < 1 || px > Fww-1) goto edgepixel;
      if (! sa_pixisin[ii-1] || ! sa_pixisin[ii+1]) goto edgepixel;        //  check 8 neighbor pixels
      kk = ii - Fww;
      if (! sa_pixisin[kk] || ! sa_pixisin[kk-1] || ! sa_pixisin[kk+1]) goto edgepixel;
      kk = ii + Fww;
      if (! sa_pixisin[kk] || ! sa_pixisin[kk-1] || ! sa_pixisin[kk+1]) goto edgepixel;

      sa_pixisin[ii] = 2;                                                  //  interior pixel
      continue;
   edgepixel:
      sa_pixisin[ii] = 1;                                                  //  edge pixel
   }

   for (ii = 0; ii < Fww * Fhh; ii++)                                      //  eliminate interior pixels
      if (sa_pixisin[ii] == 2) sa_pixisin[ii] = 0;

   mouseCBfunc = sa_finish_mousefunc;                                      //  connect mouse function
   Mcapture++;
   finish_fails = 0;

   finzd = zdialog_new(ZTX("finish area"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(finzd,"label","fmess","dialog",fmess);
   zdialog_add_widget(finzd,"label","smess","dialog","status:");

   zdialog_run(finzd,sa_finish_dialog_event);                              //  run dialog, parallel
   zdialog_wait(finzd);
   return;
}


//  dialog event and completion callback function

int  sa_finish_dialog_event(zdialog *zd, cchar *event)
{
   using namespace sa_finish_names;

   int         zstat, ii, kk, npix, px, py;
   
   mouseCBfunc = 0;                                                        //  disconnect mouse
   Mcapture = 0;

   zstat = zd->zstat;

   zdialog_free(finzd);                                                    //  kill dialog
   finzd = 0;

   if (zstat != 1 || finish_fails) {                                       //  user cancel or pixel search failure
      sa_stat = 1;                                                         //  resume edit      v.10.8
      mwpaint2();
      return 0;
   }

   npix = 0;

   for (px = 0; px < Fww; px += 1)                                         //  image top & bottom edges
   for (py = 0; py < Fhh; py += Fhh-1)
   {
      ii = py * Fww + px;
      if (! sa_pixisin[ii]) continue;
      sa_pixisin[ii] = 2;                                                  //  if selected, force non-edge
      npix++;
   }

   for (px = 0; px < Fww; px += Fww-1)                                     //  image left and right edges
   for (py = 0; py < Fhh; py += 1)
   {
      ii = py * Fww + px;
      if (! sa_pixisin[ii]) continue;
      sa_pixisin[ii] = 2;                                                  //  if selected, force non-edge
      npix++;
   }

   for (py = 1; py < Fhh-1; py++)                                          //  check all other pixels
   for (px = 1; px < Fww-1; px++)
   {
      ii = py * Fww + px;
      if (! sa_pixisin[ii]) continue;                                      //  outside of selected area
      npix++;

      if (! sa_pixisin[ii-1] || ! sa_pixisin[ii+1]) goto edgepixel;        //  check 8 neighbor pixels
      kk = ii - Fww;
      if (! sa_pixisin[kk] || ! sa_pixisin[kk-1] || ! sa_pixisin[kk+1]) goto edgepixel;
      kk = ii + Fww;
      if (! sa_pixisin[kk] || ! sa_pixisin[kk-1] || ! sa_pixisin[kk+1]) goto edgepixel;

      sa_pixseq[ii] = sa_pixisin[ii] = 2;                                  //  interior pixel      v.10.8
      continue;
   edgepixel:
      sa_pixseq[ii] = sa_pixisin[ii] = 1;                                  //  edge pixel          v.10.8
   }

   sa_Npixel = npix;                                                       //  total pixel count
   sa_stat = 3;                                                            //  area is finished
   sa_active = 1;                                                          //  area is active by default
   sa_calced = sa_blend = 0;                                               //  edge calculation is missing
   mwpaint2();
   return 0;
}


//  mouse function - get user clicks and perform pixel searches

void sa_finish_mousefunc()
{
   using namespace sa_finish_names;

   void  sa_finish_pushstack(int px, int py, char direc);
   
   int         seedpx, seedpy;
   int         px, py, ii, kk;
   char        direc;

   if (! LMclick) return;
   LMclick = 0;

   seedpx = Mxclick;
   seedpy = Myclick;
   
   zdialog_stuff(finzd,"smess",ZTX("searching"));

   ii = seedpy * Fww + seedpx;                                             //  seed pixel
   sa_stackii[0] = ii;                                                     //  put 1st pixel into stack
   sa_stackdirec[0] = 'r';                                                 //  direction = right
   sa_Nstack = 1;                                                          //  stack count

   while (sa_Nstack)                                                       //  find all pixels inside drawn area
   {
      kk = sa_Nstack - 1;                                                  //  get last pixel in stack
      ii = sa_stackii[kk];
      direc = sa_stackdirec[kk];
      
      py = ii / Fww;                                                       //  reconstruct px, py
      px = ii - Fww * py;

      if (px < pxmin || px > pxmax || py < pymin || py > pymax) {          //  ran off the edge, seed pixel
         zdialog_stuff(finzd,"smess",ZTX("area outline has a hole"));      //    was not inside an area or
         finish_fails++;                                                   //      area boundary has a hole
         return;      
      }

      if (direc == 'x') {                                                  //  no neighbors left to check
         sa_Nstack--;
         continue;
      }
      
      if (direc == 'r') {                                                  //  push next right pixel into stack
         sa_finish_pushstack(px,py,'r');
         sa_stackdirec[kk] = 'l';                                          //  this pixel next direction to look
         continue;
      }

      if (direc == 'l') {                                                  //  or next left pixel
         sa_finish_pushstack(px,py,'l');
         sa_stackdirec[kk] = 'a';
         continue;
      }

      if (direc == 'a') {                                                  //  or next ahead pixel
         sa_finish_pushstack(px,py,'a');
         sa_stackdirec[kk] = 'x';
         continue;
      }
   }
   
   zdialog_stuff(finzd,"smess",ZTX("success"));
   return;
}


//  push pixel into stack memory if not already mapped

void sa_finish_pushstack(int px, int py, char direc)
{
   int      ii, kk, ppx, ppy, npx, npy;

   if (sa_Nstack > 1) {
      kk = sa_Nstack - 2;                                                  //  get prior pixel in stack
      ii = sa_stackii[kk];
      ppy = ii / Fww;
      ppx = ii - ppy * Fww;
   }
   else {
      ppx = px - 1;                                                        //  if only one, assume prior = left
      ppy = py;
   }
   
   if (direc == 'a') {                                                     //  get pixel in direction ahead
      npx = px + ppy - py;                                                 //  leak >> straight line out
      npy = py + px - ppx;
   }
   else if (direc == 'l') {                                                //  or left
      npx = px + py - ppy;
      npy = py + ppx - px;
   }
   else if (direc == 'r') {                                                //  or right
      npx = px + px - ppx;
      npy = py + py - ppy;
   }
   else npx = npy = -1;                                                    //  stop warning
   
   if (npx < 0 || npx >= Fww) return;                                      //  pixel off the edge
   if (npy < 0 || npy >= Fhh) return;
   
   ii = npy * Fww + npx;
   if (sa_pixisin[ii]) return;                                             //  pixel already mapped
   if (sa_Nstack == sa_maxstack) return;                                   //  stack is full (impossible)
   
   sa_pixisin[ii] = 1;                                                     //  pixel is in area
   kk = sa_Nstack++;                                                       //  put pixel into stack
   sa_stackii[kk] = ii;
   sa_stackdirec[kk] = 'r';                                                //  direction = right
   sa_drawpixel(npx,npy);                                                  //  feedback to see holes     v.9.7
   
   return;
}


//  menu function for show, hide, enable, disable, invert, delete
//  (also implemented as buttons in select area dialog)

void m_select_show(GtkWidget *, cchar *menu)                               //  added to area menu   v.10.11
{
   zfuncs::F1_help_topic = "area_show_hide";
   sa_show(1);
   return;
}


void m_select_hide(GtkWidget *, cchar *menu)                               //  added to area menu   v.10.11
{
   zfuncs::F1_help_topic = "area_show_hide";
   sa_show(0);
   return;
}


void m_select_enable(GtkWidget *, cchar *menu)                             //  added to area menu   v.10.11
{
   zfuncs::F1_help_topic = "area_enable_disable";
   sa_enable();
   return;
}


void m_select_disable(GtkWidget *, cchar *menu)                            //  added to area menu   v.10.11
{
   zfuncs::F1_help_topic = "area_enable_disable";
   sa_disable();
   return;
}


void m_select_invert(GtkWidget *, cchar *menu)                             //  added to area menu   v.10.11
{
   zfuncs::F1_help_topic = "area_invert";
   sa_invert();
   return;
}


void m_select_delete(GtkWidget *, cchar *menu)                             //  added to area menu   v.10.11
{
   zfuncs::F1_help_topic = "area_delete";
   sa_delete();
   return;
}


//  show or hide outline of select area
//  also called from mwpaint() if Fshowarea = 1

void sa_show(int flag)
{
   int      px, py, ii, kk;
   
   Fshowarea = flag;
   
   if (! sa_pixseq) return;

   if (! flag) {
      mwpaint2();                                                          //  erase area outline     v.10.8
      return;
   }
   
   for (py = 1; py < Fhh-1; py++)                                          //  find pixels in area    bugfix  v.10.9
   for (px = 1; px < Fww-1; px++)
   {
      ii = py * Fww + px;
      if (! sa_pixseq[ii]) continue;                                       //  outside of area

      if (! sa_pixseq[ii-1] || ! sa_pixseq[ii+1]) goto edgepixel;          //  check 8 neighbor pixels
      kk = ii - Fww;
      if (! sa_pixseq[kk] || ! sa_pixseq[kk-1] || ! sa_pixseq[kk+1]) goto edgepixel;
      kk = ii + Fww;
      if (! sa_pixseq[kk] || ! sa_pixseq[kk-1] || ! sa_pixseq[kk+1]) goto edgepixel;
      continue;

   edgepixel:
      sa_drawpixel(px,py);
   }

   gdk_gc_set_foreground(gdkgc,&black);
   return;
}


//  enable select area that was disabled

void sa_enable()
{
   if (! sa_Npixel) return;                                                //  v.9.7
   sa_active = 1;
   sa_show(1);                                                             //  v.10.11
   return;
}


//  disable select area

void sa_disable()
{
   sa_active = 0;                                                          //  v.9.7
   sa_show(0);                                                             //  v.10.11
   return;
}


//  invert a selected area

void sa_invert()
{
   int      ii, jj, kk;

   if (! sa_Npixel) {
      zmessageACK(MWIN,ZTX("the area is not finished"));
      return;
   }

   for (ii = kk = 0; ii < Fww * Fhh; ii++)                                 //  reverse member pixels
   {
      jj = sa_pixisin[ii];                                                 //  0/1/2 = outside/edge/inside

      if (jj > 1)                                                          //  inside pixel >> outside
         sa_pixisin[ii] = sa_pixseq[ii] = 0;
      else if (jj == 1)                                                    //  edge >> edge
         kk++;                                                             //  count
      else  {
         sa_pixisin[ii] = sa_pixseq[ii] = 2;                               //  outside pixel >> inside
         kk++;                                                             //  count
      }
   }

   sa_Npixel = kk;                                                         //  new select area pixel count
   sa_calced = sa_blend = 0;                                               //  edge calculation missing
   return;
}


//  delete current area

void sa_delete()
{
   sa_stat = sa_Npixel = sa_blend = sa_calced = sa_active = 0;
   sa_currseq = sa_Ncurrseq = 0;
   if (sa_pixisin) zfree(sa_pixisin);
   if (sa_pixseq) zfree(sa_pixseq);
   if (sa_stackii) zfree(sa_stackii);
   if (sa_stackdirec) zfree(sa_stackdirec);
   sa_pixisin = 0;
   sa_pixseq = 0;
   sa_stackii = 0;
   sa_stackdirec = 0;
   mwpaint2();
   return;
}


//  compute distance from all pixels in area to nearest edge

namespace sa_edgecalc_names
{
   uint16      *sa_edgepx, *sa_edgepy, *sa_edgedist;
   int         sa_Nedge;
}

void sa_edgecalc()
{
   using namespace sa_edgecalc_names;
   
   int    edgecalc_dialog_event(zdialog*, cchar *event);
   void * edgecalc_wthread(void *);

   int         ii, nn, cc, px, py;
   zdialog     *zd = 0;
   cchar       *zectext = ZTX("Edge calculation in progress");

   if (sa_calced) return;                                                  //  done already
   if (! sa_active) sa_finish();                                           //  finish if needed
   if (! sa_active) return;                                                //  no finished area
   
   for (ii = nn = 0; ii < Fww * Fhh; ii++)                                 //  count edge pixels in select area
      if (sa_pixisin[ii] == 1) nn++;
   if (! nn) return;                                                       //  none, area is entire image?   v.10.9

   zd = zdialog_new(ZTX("Area Edge Calc"),mWin,Bcancel,null);
   zdialog_add_widget(zd,"label","lab1","dialog",zectext,"space=10");
   zdialog_run(zd,edgecalc_dialog_event);

   cc = Fww * Fhh * 2;                                                     //  allocate memory for calculations
   sa_edgedist = (uint16 *) zmalloc(cc,"edge.dist");
   memset(sa_edgedist,0,cc);
   
   sa_edgepx = (uint16 *) zmalloc(nn*2,"edge.px");                         //  allocate memory
   sa_edgepy = (uint16 *) zmalloc(nn*2,"edge.py");

   for (ii = nn = 0; ii < Fww * Fhh; ii++)                                 //  build list of edge pixels
   {                                                                       //  v.9.6
      if (sa_pixisin[ii] != 1) continue;
      py = ii / Fww;
      px = ii - py * Fww;
      sa_edgepx[nn] = px;
      sa_edgepy[nn] = py;
      nn++;
   }

   sa_Nedge = nn;
   SB_goal = sa_Npixel;
   SB_done = 0;
   
   for (int ii = 0; ii < Nwt; ii++)                                        //  start worker threads to calculate
      start_wt(edgecalc_wthread,&wtnx[ii]);                                //    sa_pixisin[] edge distances
   wait_wts();                                                             //  wait for completion

   SB_goal = SB_done = 0;                                                  //  v.9.6
   mwpaint2();                                                             //  v.10.8
   
   zdialog_free(zd);                                                       //  kill dialog
   sa_calced = 1;                                                          //  edge calculation available

   zfree(sa_edgedist);                                                     //  free memory
   zfree(sa_edgepx);
   zfree(sa_edgepy);

   if (Fkillfunc) {
      Fkillfunc = 0;
      sa_calced = 0;
   }

   return;
}


//  dialog event and completion callback function

int edgecalc_dialog_event(zdialog *zd, cchar *event)                       //  respond to user cancel
{
   Fkillfunc = 1;
   printf("edge calc killed \n");
   return 0;
}


void * edgecalc_wthread(void *arg)                                         //  worker thread function
{
   using namespace sa_edgecalc_names;
   
   void  edgecalc_f1(int px, int py);
   void  edgecalc_f2(int px, int py);

   int      index = *((int *) (arg));
   int      ii, nn, px, py;
   int64    seed = 34567 * index;

   for (nn = 0; nn < Fww * Fhh * 2; nn++)                                  //  do random pixels    v.9.6
   {
      ii = drandz(&seed) * Fww * Fhh;
      if (sa_pixisin[ii] < 2) continue;                                    //  ignore outside and edge pixels
      if (sa_edgedist[ii]) continue;                                       //  already calculated
      py = ii / Fww;
      px = ii - py * Fww;
      edgecalc_f1(px,py);                                                  //  calculate edge distance
      if (Fkillfunc) exit_wt();
   }
   
   for (ii = index; ii < Fww * Fhh; ii += Nwt)                             //  do all pixels       v.9.6
   {
      if (sa_pixisin[ii] < 2) continue;
      if (sa_edgedist[ii]) continue;
      py = ii / Fww;
      px = ii - py * Fww;
      edgecalc_f2(px,py);
      if (Fkillfunc) exit_wt();
   }

   for (ii = index; ii < Fww * Fhh; ii += Nwt)                             //  copy data from sa_edgedist[]
   {                                                                       //    to sa_pixisin[]
      if (sa_pixisin[ii] < 2) continue;
      sa_pixisin[ii] = sa_edgedist[ii];
   }

   exit_wt();
   return 0;                                                               //  not executed, stop gcc warning
}


//  Find the nearest edge pixel for a given pixel.
//  For all pixels in a line from the given pixel to the edge pixel, 
//  the same edge pixel can be used to compute edge distance.

void edgecalc_f1(int px1, int py1)
{
   using namespace sa_edgecalc_names;
   
   int      ii, px2, py2;
   uint     dist2, mindist2;
   int      epx, epy, pxm, pym, dx, dy, inc;
   double   slope;

   mindist2 = 2000000000;
   epx = epy = 0;

   for (ii = 0; ii < sa_Nedge; ii++)                                       //  loop all edge pixels
   {
      px2 = sa_edgepx[ii];
      py2 = sa_edgepy[ii];
      if (px2 < 4 || px2 > Fww-5) continue;                                //  omit pixels < 4 from image edge
      if (py2 < 4 || py2 > Fhh-5) continue;                                //                              v.9.2
      dx = px2 - px1;
      dy = py2 - py1;
      dist2 = dx*dx + dy*dy;                                               //  avoid sqrt()
      if (dist2 < mindist2) {
         mindist2 = dist2;                                                 //  remember minimum
         epx = px2;                                                        //  remember nearest edge pixel
         epy = py2;
      }
   }
   
   if (abs(epy - py1) > abs(epx - px1)) {                                  //  find all pixels along a line
      slope = 1.0 * (epx - px1) / (epy - py1);                             //    to the edge pixel
      if (epy > py1) inc = 1;
      else inc = -1;
      for (pym = py1; pym != epy; pym += inc) {
         pxm = px1 + slope * (pym - py1);
         ii = pym * Fww + pxm;
         if (sa_edgedist[ii]) continue;
         dx = epx - pxm;                                                   //  calculate distance to edge
         dy = epy - pym;
         dist2 = sqrt(dx*dx + dy*dy) + 0.5;
         sa_edgedist[ii] = dist2;                                          //  save
         SB_done++;
      }
   }

   else {
      slope = 1.0 * (epy - py1) / (epx - px1);
      if (epx > px1) inc = 1;
      else inc = -1;
      for (pxm = px1; pxm != epx; pxm += inc) {
         pym = py1 + slope * (pxm - px1);
         ii = pym * Fww + pxm;
         if (sa_edgedist[ii]) continue;
         dx = epx - pxm;
         dy = epy - pym;
         dist2 = sqrt(dx*dx + dy*dy) + 0.5;
         sa_edgedist[ii] = dist2;
         SB_done++;
      }
   }

   return;
}


void edgecalc_f2(int px1, int py1)                                         //  calculate 1 pixel
{
   using namespace sa_edgecalc_names;
   
   int      ii, kk, px2, py2;
   uint     dist2, mindist2;
   int      epx, epy, dx, dy;

   mindist2 = 2000000000;
   epx = epy = 0;

   for (ii = 0; ii < Fww * Fhh; ii++)
   {
      if (sa_pixisin[ii] != 1) continue;                                   //  find all edge pixels
      py2 = ii / Fww;
      px2 = ii - py2 * Fww;
      if (px2 < 4 || px2 > Fww-5) continue;                                //  < pixels 4 from image edge   v.9.2
      if (py2 < 4 || py2 > Fhh-5) continue;
      dx = px2 - px1;                                                      //  calculate distance to edge pixel
      dy = py2 - py1;
      dist2 = dx*dx + dy*dy;                                               //  avoid sqrt()
      if (dist2 < mindist2) {
         mindist2 = dist2;                                                 //  remember minimum
         epx = px2;                                                        //  remember nearest edge pixel
         epy = py2;
      }
   }
   
   kk = py1 * Fww + px1;
   sa_edgedist[kk] = sqrt(mindist2) + 0.5;

   SB_done++;
   return;
}


/**************************************************************************
   select area copy and paste menu functions
***************************************************************************/

PXM      *sacp_pxm16 = 0;                                                  //  select area pixmap copy
int      sacp_ww, sacp_hh;                                                 //  dimensions
uint16   *sacp_dist = 0;                                                   //  pixel edge distances

PXM      *sacpR_pxm16 = 0;                                                 //  resized pixmap
int      sacpR_ww, sacpR_hh;                                               //  dimensions
double   sacp_resize;                                                      //  size, 1.0 = original size
int      sacp_orgx, sacp_orgy;                                             //  origin in target image
int      sacp_blend;                                                       //  edge blend with target image
int      sacp_susp = 0;                                                    //  dialog is suspended

void select_paste_pixmap();                                                //  copy saved area into image


//  copy selected area, save in memory

void m_select_copy(GtkWidget *, cchar *)                                   //  menu function
{
   int      ii, jj, cc, px, py, qx, qy, dist;
   int      pxmin, pxmax, pymin, pymax;
   uint16   *pix1, *pix2;

   zfuncs::F1_help_topic = "area_copy_paste";                              //  v.10.8

   if (! sa_active) sa_finish();                                           //  finish area if not already
   if (! sa_active) return;
   sa_edgecalc();                                                          //  do edge calc if not already
   
   pxmin = Fww;
   pxmax = 0;
   pymin = Fhh;
   pymax = 0;

   for (ii = 0; ii < Fww * Fhh; ii++)                                      //  find pixels in select area
   {                                                                       //  v.9.6
      if (! sa_pixisin[ii]) continue;
      py = ii / Fww;
      px = ii - py * Fww;
      if (px > pxmax) pxmax = px;                                          //  find enclosing rectangle
      if (px < pxmin) pxmin = px;
      if (py > pymax) pymax = py;
      if (py < pymin) pymin = py;
   }
   
   PXM_free(sacp_pxm16);
   sacp_ww = pxmax - pxmin + 11;                                           //  create new PXM pixmap for area
   sacp_hh = pymax - pymin + 11;                                           //  add 5-pixel margins
   sacp_pxm16 = PXM_make(sacp_ww,sacp_hh,16);
   
   for (py = 0; py < sacp_hh; py++)                                        //  clear all pixels to black
   for (px = 0; px < sacp_ww; px++)
   {
      pix2 = PXMpix(sacp_pxm16,px,py);
      pix2[0] = pix2[1] = pix2[2] = 0;
   }
   
   if (sacp_dist) zfree(sacp_dist);
   cc = 2 * sacp_ww * sacp_hh;                                             //  set all edge distances = 0
   sacp_dist = (uint16 *) zmalloc(cc+100,"sacp.dist");
   memset(sacp_dist,0,cc);
   
   for (ii = 0; ii < Fww * Fhh; ii++)                                      //  find pixels in select area
   {                                                                       //  v.9.6
      if (! sa_pixisin[ii]) continue;
      py = ii / Fww;
      px = ii - py * Fww;
      dist = sa_pixisin[ii];
      pix1 = PXMpix(Fpxm16,px,py);                                         //  copy pixel into pixmap
      px = px - pxmin + 5;                                                 //  leave margin
      py = py - pymin + 5;
      pix2 = PXMpix(sacp_pxm16,px,py);
      pix2[0] = pix1[0];
      pix2[1] = pix1[1];
      pix2[2] = pix1[2];
      jj = py * sacp_ww + px;
      sacp_dist[jj] = dist + 1;                                            //  0/1/2+ = outside/edge/inside
   }

   for (ii = 0; ii < sacp_ww * sacp_hh; ii++)                              //  propagate edge pixels outward,
   {                                                                       //    prevent mixing black pixels
      if (sacp_dist[ii] != 1) continue;                                    //      when pixmap is rescaled
      py = ii / sacp_ww;
      px = ii - sacp_ww * py;
      pix1 = PXMpix(sacp_pxm16,px,py);
      for (qx = px-4; qx <= px+4; qx++)
      for (qy = py-4; qy <= py+4; qy++)
      {
         pix2 = PXMpix(sacp_pxm16,qx,qy);
         if (pix2[0] || pix2[1] || pix2[2]) continue;
         pix2[0] = pix1[0];
         pix2[1] = pix1[1];
         pix2[2] = pix1[2];
      }
   }  

   return;
}


//  paste selected area into current image
//  this is an edit function - select area image is copied into main image

int      sacp_porg = 0;                                                    //  pasted area is present
int      sacp_porgx, sacp_porgy;                                           //  pasted area origin in image
int      sacp_pww, sacp_phh;                                               //  pasted area dimensions

void m_select_paste(GtkWidget *, cchar *)                                  //  menu function
{
   int   select_paste_dialog_event(zdialog *, cchar *event);
   void  select_paste_mousefunc();

   cchar  *dragmess = ZTX("position with mouse click/drag");

   zfuncs::F1_help_topic = "area_copy_paste";                              //  v.10.8

   if (! sacp_pxm16) return;                                               //  nothing to paste
   sa_delete();                                                            //  delete select area if present   v.9.7
   if (! edit_setup("paste",0,0)) return;                                  //  setup edit
   
   sacp_resize = 1.0;                                                      //  size = 1x
   sacp_blend = 1;                                                         //  edge blend = 1
   sacp_susp = 0;                                                          //  not suspended

   sacpR_ww = sacp_ww;                                                     //  setup resized paste image
   sacpR_hh = sacp_hh;                                                     //  (initially 1x)
   sacpR_pxm16 = PXM_copy(sacp_pxm16);
   
   sacp_porg = 0;                                                          //  no image paste location yet
   
   zdedit = zdialog_new(ZTX("Paste Image"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"hbox","hb0","dialog",0,"space=8");
   zdialog_add_widget(zdedit,"label","lab1","hb0",dragmess,"space=8");
   zdialog_add_widget(zdedit,"button","susp-resm","hb0",Bsuspend);
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"label","lab1","hb1","resize");
   zdialog_add_widget(zdedit,"button","+.1%","hb1","+.1%");
   zdialog_add_widget(zdedit,"button","+1%","hb1","+1%");
   zdialog_add_widget(zdedit,"button","+10%","hb1","+10%");
   zdialog_add_widget(zdedit,"button","-.1%","hb1","-.1%");
   zdialog_add_widget(zdedit,"button","-1%","hb1","-1%");
   zdialog_add_widget(zdedit,"button","-10%","hb1","-10%");
   zdialog_add_widget(zdedit,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"label","lab2","hb2","edge blend");
   zdialog_add_widget(zdedit,"hscale","blend","hb2","1|50|1|1","expand");

   zdialog_run(zdedit,select_paste_dialog_event);

   mouseCBfunc = select_paste_mousefunc;                                   //  connect mouse function
   Mcapture = 1;
   return;
}


//  Dialog event and completion callback function
//  Get dialog values and convert image. When done, commit edited image 
//  (with pasted pixels) and set up a new select area for the pasted 
//  pixels, allowing further editing of the area.

int select_paste_dialog_event(zdialog *zd, cchar *event)
{
   void  select_paste_mousefunc();

   int      cc, ii;
   int      px1, py1, px2, py2;
   uint16   *pix1;
   
   if (zd->zstat)                                                          //  dialog completed
   {
      mouseCBfunc = 0;                                                     //  disconnect mouse
      Mcapture = 0;

      if (zd->zstat != 1) {                                                //  cancel paste
         edit_cancel();                                                    //  cancel edit, restore image
         return 0;
      }

      edit_done();                                                         //  commit the edit (pasted image)
      
      if (sa_pixseq) zfree(sa_pixseq);                                     //  allocate sa_pixseq[]
      cc = 2 * Fww * Fhh;
      sa_pixseq = (uint16 *) zmalloc(cc,"pixseq");
      memset(sa_pixseq,0,cc);
      
      for (py1 = 0; py1 < sacpR_hh; py1++)                                 //  map non-transparent pixels
      for (px1 = 0; px1 < sacpR_ww; px1++)                                 //    into sa_pixseq[]
      {
         px2 = px1 / sacp_resize + 0.5;
         py2 = py1 / sacp_resize + 0.5;
         ii = sacp_ww * py2 + px2;                                         //  corresponding sacp_dist[]
         if (sacp_dist[ii] == 0) continue;                                 //  outside select area
         pix1 = PXMpix(sacpR_pxm16,px1,py1);
         px2 = px1 + sacp_orgx;
         py2 = py1 + sacp_orgy;
         if (px2 < 0 || px2 >= Fww) continue;                              //  parts may be beyond edges
         if (py2 < 0 || py2 >= Fhh) continue;
         ii = py2 * Fww + px2;
         sa_pixseq[ii] = 1;                                                //  result is equivalent to a
      }                                                                    //    1-click select-color area

      sa_finish();                                                         //  finish the area

      PXM_free(sacpR_pxm16);                                               //  free memory
      return 0;
   }

   if (strEqu(event,"susp-resm")) {                                        //  toggle suspend/resume
      if (! sacp_susp) {
         mouseCBfunc = 0;                                                  //  disconnect mouse function
         Mcapture = 0;
         set_cursor(0);                                                    //  restore normal cursor
         zdialog_stuff(zdedit,"susp-resm",Bresume);
         sacp_susp = 1;
      }
      else {
         mouseCBfunc = select_paste_mousefunc;                             //  connect mouse function
         Mcapture = 1;
         zdialog_stuff(zdedit,"susp-resm",Bsuspend);
         sacp_susp = 0;
      }
   }

   if (strchr(event,'%')) {                                                //  resize by 1% or 10%
      if (strEqu(event,"+.1%")) sacp_resize = sacp_resize * 1.001;
      if (strEqu(event,"+1%")) sacp_resize = sacp_resize * 1.01;
      if (strEqu(event,"+10%")) sacp_resize = sacp_resize * 1.10;
      if (strEqu(event,"-.1%")) sacp_resize = sacp_resize * 0.999;
      if (strEqu(event,"-1%")) sacp_resize = sacp_resize * 0.99;
      if (strEqu(event,"-10%")) sacp_resize = sacp_resize * 0.90;

      PXM_free(sacpR_pxm16);
      sacpR_ww = sacp_resize * sacp_ww;                                    //  make new rescaled paste image
      sacpR_hh = sacp_resize * sacp_hh;
      sacpR_pxm16 = PXM_make(sacpR_ww,sacpR_hh,16);
      sacpR_pxm16 = PXM_rescale(sacp_pxm16,sacpR_ww,sacpR_hh);
      select_paste_pixmap();                                               //  copy onto image
   }

   if (strEqu(event,"blend")) {
      zdialog_fetch(zd,"blend",sacp_blend);                                //  blend, 0-50 pixels
      select_paste_pixmap();                                               //  copy onto image
   }

   return 0;
}


//  mouse function - follow mouse drags and move pasted area accordingly

void select_paste_mousefunc()
{
   int            mx1, my1, mx2, my2;
   static int     mdx0, mdy0, mdx1, mdy1;
   
   if (LMclick) {                                                          //  left mouse click
      LMclick = 0;
      sacp_orgx = Mxclick - sacpR_ww / 2;                                  //  position image at mouse   v.10.11
      sacp_orgy = Myclick - sacpR_hh / 2;
      select_paste_pixmap();
   }
   
   if (Mxposn < sacp_orgx || Mxposn > sacp_orgx + sacpR_ww ||              //  mouse outside select area
      Myposn < sacp_orgy || Myposn > sacp_orgy + sacpR_hh)
      set_cursor(0);                                                       //  set normal cursor
   else 
      set_cursor(dragcursor);                                              //  set drag cursor

   if (Mxdrag + Mydrag == 0) return;                                       //  no drag underway

   if (Mxdown != mdx0 || Mydown != mdy0) {                                 //  new drag initiated
      mdx0 = mdx1 = Mxdown;
      mdy0 = mdy1 = Mydown;
   }

   mx1 = mdx1;                                                             //  drag start
   my1 = mdy1;
   mx2 = Mxdrag;                                                           //  drag position
   my2 = Mydrag;
   mdx1 = mx2;                                                             //  next drag start
   mdy1 = my2;
   
   sacp_orgx += (mx2 - mx1);                                               //  move position of select area
   sacp_orgy += (my2 - my1);                                               //    by mouse drag amount
   select_paste_pixmap();                                                  //  re-copy area to new position

   return;      
}


//  copy select area into edit image, starting at sacp_orgx/y

void select_paste_pixmap()
{
   int      ii, px1, py1, px3, py3;
   uint16   *pix1, *pix3;
   double   f1, f2, dist;
   
   if (sacp_porg)                                                          //  prior area overlap rectangle
   {
      for (py1 = 4; py1 < sacp_phh-5; py1++)                               //  restore original image pixels   v.10.11
      for (px1 = 4; px1 < sacp_pww-5; px1++)
      {
         px3 = px1 + sacp_porgx;
         py3 = py1 + sacp_porgy;
         if (px3 < 0 || px3 >= E3ww) continue;                             //  parts may be beyond edges
         if (py3 < 0 || py3 >= E3hh) continue;
         pix1 = PXMpix(E1pxm16,px3,py3);
         pix3 = PXMpix(E3pxm16,px3,py3);
         pix3[0] = pix1[0];
         pix3[1] = pix1[1];
         pix3[2] = pix1[2];
      }
   }
   
   for (py1 = 4; py1 < sacpR_hh-5; py1++)                                  //  copy area pixels less margins
   for (px1 = 4; px1 < sacpR_ww-5; px1++)                                  //    to new image overlap rectangle
   {
      px3 = px1 / sacp_resize + 0.5;
      py3 = py1 / sacp_resize + 0.5;
      ii = sacp_ww * py3 + px3;                                            //  corresp. sacp_dist[]
      dist = sacp_dist[ii];
      if (dist == 0) continue;                                             //  pixel is outside select area
      pix1 = PXMpix(sacpR_pxm16,px1,py1);
      px3 = px1 + sacp_orgx;
      py3 = py1 + sacp_orgy;
      if (px3 < 0 || px3 >= E3ww) continue;                                //  parts may be beyond edges
      if (py3 < 0 || py3 >= E3hh) continue;
      pix3 = PXMpix(E3pxm16,px3,py3);                                      //  target image pixel
      f1 = (dist-1) * sacp_resize / sacp_blend;
      if (f1 > 1.0) f1 = 1.0;                                              //  blend area pixel and target pixel
      f2 = 1.0 - f1;
      pix3[0] = f2 * pix3[0] + f1 * pix1[0];
      pix3[1] = f2 * pix3[1] + f1 * pix1[1];
      pix3[2] = f2 * pix3[2] + f1 * pix1[2];
   }
                                                                           //  v.10.11
   mwpaint3(sacp_porgx+4,sacp_porgy+4,sacp_pww-8,sacp_phh-8);              //  update window for old overlap area
   mwpaint3(sacp_orgx+4,sacp_orgy+4,sacpR_ww-8,sacpR_hh-8);                //  update window for new overlap area

   sacp_porgx = sacp_orgx;                                                 //  remember location for next call
   sacp_porgy = sacp_orgy;
   sacp_pww = sacpR_ww;
   sacp_phh = sacpR_hh;
   sacp_porg = 1;

   Fmodified = 1;
   return;
}


/**************************************************************************
   select area load from file and save to file functions           v.9.9
***************************************************************************/

//  Load a select area from a disk file and paste into current image.

void m_select_open(GtkWidget *, cchar *)
{
   char        safile[200], *pfile;
   char        *pp, buff[24], fotoxx[8];
   int         fid, ww, hh, cc, cc2;
   PXM         *temppxm;
   uint16      *tempdist;

   zfuncs::F1_help_topic = "area_open_save";                               //  v.10.8

   *safile = 0;                                                            //  get area name from user
   strncatv(safile,199,select_area_dirk,"/",null);
   pfile = zgetfile1(ZTX("load select area from a file"),"open",safile);
   if (! pfile) return;
   
   temppxm = TIFFread(pfile);                                              //  TIFF file --> PXM
   if (! temppxm) return;

   strncpy0(safile,pfile,195);                                             //  replace .tiff with .dist
   pp = strrchr(safile,'/');
   if (pp) pp = strrchr(pp,'.');
   if (pp) strcpy(pp,".dist");
   else strcat(safile,".dist");
   
   fid = open(safile,O_RDONLY);                                            //  open .dist file
   if (! fid) {
      zmessageACK(MWIN,ZTX("cannot read .dist file"));
      PXM_free(temppxm);
      return;
   }
   
   *fotoxx = 0;                                                            //  .dist file --> area ww/hh
   cc = read(fid,buff,20);
   sscanf(buff," %d %d %8s ",&ww, &hh, fotoxx);
   if (! strEqu(fotoxx,"fotoxx")) {
      zmessageACK(MWIN,ZTX("cannot read .dist file"));
      PXM_free(temppxm);
      close(fid);
      return;
   }

   cc = 2 * ww * hh;                                                       //  allocate memory for edge distance data
   tempdist = (uint16 *) zmalloc(cc+100,"select-distfile");

   cc2 = read(fid,tempdist,cc);                                            //  .dist file --> edge distance data
   if (cc2 != cc) {
      zmessageACK(MWIN,ZTX("cannot read .dist file"));
      PXM_free(temppxm);
      zfree(tempdist);
      close(fid);
      return;
   }

   PXM_free(sacp_pxm16);                                                   //  replace select area in memory
   sacp_pxm16 = temppxm;
   sacp_ww = ww;
   sacp_hh = hh;

   if (sacp_dist) zfree(sacp_dist);
   sacp_dist = tempdist;

   m_select_paste(0,0);                                                    //  paste it into current image
   return;
}


//  save a select area as a disk file

void m_select_save(GtkWidget *, cchar *)
{
   char        safile[200], *pfile, *pp, buff[24];
   int         fid, err, cc, cc2;

   if (! sa_active) sa_finish();                                           //  finish select area if not already
   if (sa_active) {
      sa_edgecalc();                                                       //  do edge calc if not already
      m_select_copy(0,0);                                                  //  copy select area to memory
   }

   zfuncs::F1_help_topic = "area_open_save";                               //  v.10.11

   if (! sacp_pxm16) return;                                               //  no active area, none in memory
   
   *safile = 0;                                                            //  get area name from user
   strncatv(safile,199,select_area_dirk,"/",null);
   pfile = zgetfile1(ZTX("save select area as a file"),"save",safile);
   if (! pfile) return;

   strncpy0(safile,pfile,195);                                             //  insure extension is .tiff
   pp = strrchr(safile,'/');
   if (pp) pp = strrchr(pp,'.');
   if (pp) strcpy(pp,".tiff");
   else strcat(safile,".tiff");
   
   err = TIFFwrite(sacp_pxm16,safile);                                     //  select area PXM --> .tiff file
   if (err) return;
   
   pp = strrchr(safile,'.');
   strcpy(pp,".dist");
   fid = open(safile,O_WRONLY|O_CREAT|O_TRUNC,0640);                       //  edge distance data --> .dist file
   if (! fid) zappcrash("cannot open .dist file");

   snprintf(buff,24," %05d %05d fotoxx ",sacp_ww,sacp_hh);
   cc = write(fid,buff,20);
   if (cc != 20) zappcrash("cannot write .dist file");
   
   cc = 2 * sacp_ww * sacp_hh;
   cc2 = write(fid,sacp_dist,cc);
   if (cc2 != cc) zappcrash("cannot write .dist file");

   close(fid);
   return;
}


/**************************************************************************
      begin image edit functions
***************************************************************************/

//  adjust white balance

double   whitebal_red, whitebal_green, whitebal_blue;

void m_whitebal(GtkWidget *, cchar *)
{
   int    whitebal_dialog_event(zdialog* zd, cchar *event);
   void   whitebal_mousefunc();
   void * whitebal_thread(void *);

   cchar  *wbtitle = ZTX("Adjust White Balance");
   cchar  *wbhelp = ZTX("Click white or gray image location");

   zfuncs::F1_help_topic = "white_balance";                                //  v.10.8
   
   if (! edit_setup("whitebal",1,2)) return;                               //  setup edit: preview

   zdedit = zdialog_new(wbtitle,mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"space=10");
   zdialog_add_widget(zdedit,"label","labwbh","hb1",wbhelp,"space=5");
   zdialog_add_widget(zdedit,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"label","labpix","hb2","pixel:");
   zdialog_add_widget(zdedit,"label","pixel","hb2","0000 0000");
   zdialog_add_widget(zdedit,"label","labrgb","hb2","   RGB:");
   zdialog_add_widget(zdedit,"label","rgb","hb2","000 000 000");

   zdialog_run(zdedit,whitebal_dialog_event);                              //  run dialog - parallel

   whitebal_red = whitebal_green = whitebal_blue = 1.0;
   start_thread(whitebal_thread,0);                                        //  start working thread

   mouseCBfunc = whitebal_mousefunc;                                       //  connect mouse function
   Mcapture = 1;                                                           //  capture mouse clicks
   return;
}


//  dialog event and completion callback function

int whitebal_dialog_event(zdialog *zd, cchar *event)                       //  dialog event function     v.10.2
{
   if (zd->zstat)
   {
      mouseCBfunc = 0;                                                     //  disconnect mouse
      Mcapture = 0;
      if (zd->zstat == 1) edit_done();                                     //  done
      else edit_cancel();                                                  //  cancel or destroy
      return 0;
   }
   
   if (strEqu(event,"undo")) edit_undo();                                  //  v.10.2
   if (strEqu(event,"redo")) edit_redo();                                  //  v.10.3
   if (strEqu(event,"blendwidth")) signal_thread();                        //  v.10.3
   return 1;
}


void whitebal_mousefunc()                                                  //  mouse function
{
   int         px, py, dx, dy;
   double      red, green, blue, rgbmean;
   char        work[40];
   uint16      *ppix16;
   
   if (! LMclick) return;
   
   LMclick = 0;
   px = Mxclick;                                                           //  mouse click position
   py = Myclick;
   
   if (px < 2) px = 2;                                                     //  pull back from edge
   if (px > E3ww-3) px = E3ww-3;
   if (py < 2) py = 2;
   if (py > E3hh-3) py = E3hh-3;
   
   red = green = blue = 0;

   for (dy = -2; dy <= 2; dy++)                                            //  5x5 block around mouse position
   for (dx = -2; dx <= 2; dx++)
   {
      ppix16 = PXMpix(E1pxm16,px+dx,py+dy);                                //  input image
      red += ppix16[0];
      green += ppix16[1];
      blue += ppix16[2];
   }
   
   red = red / 25.0;                                                       //  mean RGB levels 
   green = green / 25.0;
   blue = blue / 25.0;
   rgbmean = (red + green + blue) / 3.0;

   whitebal_red = rgbmean / red;
   whitebal_green = rgbmean / green;
   whitebal_blue = rgbmean / blue;

   signal_thread();                                                        //  trigger image update

   snprintf(work,40,"%d %d",px,py);
   zdialog_stuff(zdedit,"pixel",work);

   snprintf(work,40,"%7.3f %7.3f %7.3f",red/256,green/256,blue/256);
   zdialog_stuff(zdedit,"rgb",work);
   
   return;
}


//  Update image based on neutral pixel that was clicked

void * whitebal_thread(void *)
{
   void * whitebal_wthread(void *arg);

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wt(whitebal_wthread,&wtnx[ii]);
      wait_wts();                                                          //  wait for completion

      Fmodified = 1;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}


void * whitebal_wthread(void *arg)                                         //  worker thread function
{
   int         index = *((int *) arg);
   int         px, py, ii, dist = 0;
   uint16      *pix1, *pix3;
   double      red1, green1, blue1;
   double      red3, green3, blue3;
   double      brmax, dold, dnew;

   for (py = index; py < E1hh; py += Nwt)
   for (px = 0; px < E1ww; px++)
   {
      if (sa_active) {                                                     //  select area active
         ii = py * Fww + px;
         dist = sa_pixisin[ii];                                            //  distance from edge
         if (! dist) continue;                                             //  pixel outside area
      }

      pix1 = PXMpix(E1pxm16,px,py);                                        //  input pixel
      pix3 = PXMpix(E3pxm16,px,py);                                        //  output pixel
      
      red1 = pix1[0];
      green1 = pix1[1];
      blue1 = pix1[2];
      
      red3 = whitebal_red * red1;                                          //  change color ratios
      green3 = whitebal_green * green1;
      blue3 = whitebal_blue * blue1;

      if (sa_active && dist < sa_blend) {                                  //  select area is active,
         dnew = 1.0 * dist / sa_blend;                                     //    blend changes over sa_blend
         dold = 1.0 - dnew;
         red3 = dnew * red3 + dold * red1;
         green3 = dnew * green3 + dold * green1;
         blue3 = dnew * blue3 + dold * blue1;
      }
      
      brmax = red3;                                                        //  brmax = brightest color
      if (green3 > brmax) brmax = green3;
      if (blue3 > brmax) brmax = blue3;
      
      if (brmax > 65535) {                                                 //  if overflow, reduce
         brmax = 65535 / brmax;
         red3 = red3 * brmax;
         green3 = green3 * brmax;
         blue3 = blue3 * brmax;
      }

      pix3[0] = int(red3);
      pix3[1] = int(green3);
      pix3[2] = int(blue3);
   }
   
   exit_wt();
   return 0;                                                               //  not executed, stop gcc warning
}


/**************************************************************************/

//  make a black & white or color negative image

void m_negate(GtkWidget *, cchar *)                                        //  v.10.9
{
   int negate_dialog_event(zdialog *zd, cchar *event);

   zfuncs::F1_help_topic = "make_negative";

   if (! edit_setup("negate",0,0)) return;                                 //  setup edit: no preview

   zdedit = zdialog_new(ZTX("Make Negative"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"radio","b&wpos","dialog","black/white positive");
   zdialog_add_widget(zdedit,"radio","b&wneg","dialog","black/white negative");
   zdialog_add_widget(zdedit,"radio","colpos","dialog","color positive");
   zdialog_add_widget(zdedit,"radio","colneg","dialog","color negatice");

   zdialog_stuff(zdedit,"colpos",1);

   zdialog_resize(zdedit,200,0);
   zdialog_run(zdedit,negate_dialog_event);                                //  run dialog - parallel
   return;
}


//  dialog event and completion callback function

int negate_dialog_event(zdialog *zd, cchar *event)
{
   int         mode, px, py;
   int         red, green, blue;
   uint16      *pix1, *pix3;

   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) edit_done();
      else edit_cancel();
      return 0;
   }
   
   if (strEqu(event,"b&wpos")) mode = 1;
   if (strEqu(event,"b&wneg")) mode = 2;
   if (strEqu(event,"colpos")) mode = 3;
   if (strEqu(event,"colneg")) mode = 4;
   
   edit_zapredo();                                                         //  delete redo copy

   for (py = 0; py < E3hh; py++)
   for (px = 0; px < E3ww; px++)
   {
      pix1 = PXMpix(E1pxm16,px,py);

      red = pix1[0];
      green = pix1[1];
      blue = pix1[2];
      
      if (mode == 1)                                                       //  black and white positive
         red = green = blue = (red + green + blue) / 3;
      
      else if (mode == 2)                                                  //  black and white negative
         red = green = blue = 65535 - (red + green + blue) / 3;
      
      if (mode == 3) {  /** do nothing  **/  }                             //  color positive
      
      if (mode == 4) {                                                     //  color negative
         red = 65535 - red;
         green = 65535 - green;
         blue = 65535 - blue; 
      }
      
      pix3 = PXMpix(E3pxm16,px,py);

      pix3[0] = red;
      pix3[1] = green;
      pix3[2] = blue;
   }
   
   Fmodified = 1;
   mwpaint2();

   return 0;
}


/**************************************************************************/

//  flatten brightness distribution

double   flatten_value = 0;                                                //  flatten value, 0 - 100%
double   flatten_brdist[65536];


void m_flatten(GtkWidget *, cchar *)
{
   int    flatten_dialog_event(zdialog* zd, cchar *event);
   void * flatten_thread(void *);

   cchar  *title = ZTX("Flatten Brightness Distribution");

   zfuncs::F1_help_topic = "flatten";                                      //  v.10.8
   
   if (! edit_setup("flatten",1,2)) return;                                //  setup edit: preview, select area OK

   zdedit = zdialog_new(title,mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"space=15");
   zdialog_add_widget(zdedit,"label","labfd","hb1",ZTX("Flatten"),"space=5");
   zdialog_add_widget(zdedit,"hscale","flatten","hb1","0|100|1|0","expand");

   zdialog_resize(zdedit,300,0);
   zdialog_run(zdedit,flatten_dialog_event);                               //  run dialog - parallel
   
   flatten_value = 0;
   start_thread(flatten_thread,0);                                         //  start working thread
   return;
}


//  dialog event and completion callback function

int flatten_dialog_event(zdialog *zd, cchar *event)                        //  flatten dialog event function
{
   if (zd->zstat)
   {
      if (zd->zstat == 1) edit_done();                                     //  done
      else edit_cancel();                                                  //  cancel or destroy
      return 0;
   }
   
   if (strEqu(event,"undo")) edit_undo();                                  //  v.10.2
   if (strEqu(event,"redo")) edit_redo();                                  //  v.10.3
   if (strEqu(event,"blendwidth")) signal_thread();                        //  v.10.3

   if (strEqu(event,"flatten")) {
      zdialog_fetch(zd,"flatten",flatten_value);                           //  get slider value
      signal_thread();                                                     //  trigger update thread
   }
   
   return 1;
}


//  thread function - use multiple working threads

void * flatten_thread(void *)
{
   void  * flatten_wthread(void *arg);

   int         px, py, ii;
   double      bright1;
   uint16      *pix1;

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      for (ii = 0; ii < 65536; ii++)                                       //  clear brightness distribution data
         flatten_brdist[ii] = 0;

      if (sa_active)                                                       //  process selected area
      {
         for (ii = 0; ii < Fww * Fhh; ii++)                                //  v.9.6
         {
            if (! sa_pixisin[ii]) continue;
            py = ii / Fww;
            px = ii - py * Fww;
            pix1 = PXMpix(E1pxm16,px,py);
            bright1 = pixbright(pix1);
            flatten_brdist[int(bright1)]++;
         }
         
         for (ii = 1; ii < 65536; ii++)                                    //  cumulative brightness distribution
            flatten_brdist[ii] += flatten_brdist[ii-1];                    //   0 ... sa_Npixel

         for (ii = 0; ii < 65536; ii++)
            flatten_brdist[ii] = flatten_brdist[ii]                        //  multiplier per brightness level
                               / sa_Npixel * 65536.0 / (ii + 1);
      }

      else                                                                 //  process whole image
      {
         for (py = 0; py < E1hh; py++)                                     //  compute brightness distribution
         for (px = 0; px < E1ww; px++)
         {
            pix1 = PXMpix(E1pxm16,px,py);
            bright1 = pixbright(pix1);
            flatten_brdist[int(bright1)]++;
         }
         
         for (ii = 1; ii < 65536; ii++)                                    //  cumulative brightness distribution
            flatten_brdist[ii] += flatten_brdist[ii-1];                    //   0 ... (ww1 * hh1)

         for (ii = 0; ii < 65536; ii++)
            flatten_brdist[ii] = flatten_brdist[ii]                        //  multiplier per brightness level 
                               / (E1ww * E1hh) * 65536.0 / (ii + 1);
      }
      
      for (ii = 0; ii < Nwt; ii++)                                         //  start worker threads
         start_wt(flatten_wthread,&wtnx[ii]);
      wait_wts();                                                          //  wait for completion

      Fmodified = 1;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}


void * flatten_wthread(void *arg)                                          //  worker thread function
{
   int         index = *((int *) (arg));
   int         px, py, ii, dist = 0;
   uint16      *pix1, *pix3;
   double      fold, fnew, dold, dnew, cmax;
   double      red1, green1, blue1, red3, green3, blue3;
   double      bright1, bright2;

   for (py = index; py < E1hh; py += Nwt)                                  //  flatten brightness distribution
   for (px = 0; px < E1ww; px++)
   {
      if (sa_active) {                                                     //  select area active
         ii = py * Fww + px;
         dist = sa_pixisin[ii];                                            //  distance from edge
         if (! dist) continue;                                             //  outside pixel
      }

      pix1 = PXMpix(E1pxm16,px,py);                                        //  input pixel
      pix3 = PXMpix(E3pxm16,px,py);                                        //  output pixel
         
      fnew = 0.01 * flatten_value;                                         //  0.0 - 1.0  how much to flatten
      fold = 1.0 - fnew;                                                   //  1.0 - 0.0  how much to retain

      red1 = pix1[0];
      green1 = pix1[1];
      blue1 = pix1[2];

      bright1 = pixbright(pix1);                                           //  input brightness
      bright2 = flatten_brdist[int(bright1)];                              //  output brightness adjustment

      red3 = bright2 * red1;                                               //  flattened brightness
      green3 = bright2 * green1;
      blue3 = bright2 * blue1;

      red3 = fnew * red3 + fold * red1;                                    //  blend new and old brightness
      green3 = fnew * green3 + fold * green1;
      blue3 = fnew * blue3 + fold * blue1;

      if (sa_active && dist < sa_blend) {                                  //  select area is active,
         dnew = 1.0 * dist / sa_blend;                                     //    blend changes over sa_blend
         dold = 1.0 - dnew;
         red3 = dnew * red3 + dold * red1;
         green3 = dnew * green3 + dold * green1;
         blue3 = dnew * blue3 + dold * blue1;
      }

      cmax = red3;                                                         //  stop overflow, keep color balance
      if (green3 > cmax) cmax = green3;
      if (blue3 > cmax) cmax = blue3;
      if (cmax > 65535) {
         cmax = 65535 / cmax;
         red3 = red3 * cmax;
         green3 = green3 * cmax;
         blue3 = blue3 * cmax;
      }

      pix3[0] = int(red3 + 0.5);
      pix3[1] = int(green3 + 0.5);
      pix3[2] = int(blue3 + 0.5);
   }

   exit_wt();
   return 0;                                                               //  not executed, avoid gcc warning
}


/**************************************************************************/

//  brightness / color / contrast adjustment

void  tune_curve_update(int spc);                                          //  curve update callback function
int   tune_spc;                                                            //  current spline curve 1-5
int   tune_spc_moved[6];                                                   //  tracks which curves changed


void m_tune(GtkWidget *, cchar *)                                          //  menu function
{
   using namespace splinecurve;

   int   tune_dialog_event(zdialog *zd, cchar *event);
   void  *tune_thread(void *);

   cchar  *title = ZTX("Adjust Brightness and Color");
   static char    *Bbrightness2 = 0;

   if (! Bbrightness2) {
      Bbrightness2 = strdupz(Bbrightness,4,"tune");
      strcat(Bbrightness2,"-2");
   }

   zfuncs::F1_help_topic = "tune";                                         //  v.10.8
   
   if (! edit_setup("britecolor",1,2)) return;                             //  setup edit: preview
   
   for (int spc = 0; spc < 6; spc++)                                       //  setup 6 spline curves
   {                                                                       //  no. 0 is active curve
      vert[spc] = 0;                                                       //  all curves are horizontal
      nap[spc] = 3;                                                        //  curves 1-6 are copied to 0 when active
      apx[spc][0] = 0.01;
      apy[spc][0] = 0.5;
      apx[spc][1] = 0.5;
      apy[spc][1] = 0.5;
      apx[spc][2] = 0.98;
      apy[spc][2] = 0.5;
      curve_generate(spc);
      tune_spc_moved[spc] = 0;
   }

   Nspc = 1;                                                               //  only one at a time is active
   tune_spc = 1;                                                           //  default curve = brightness

/***
       --------------------------------------------
      |                                            |
      |                                            |
      |           curve drawing area               |
      |                                            |
      |                                            |
       --------------------------------------------
       darker areas                   lighter areas

      [+++] [---] [+ -] [- +] [+-+] [-+-]

      (o) brightness         [reset 1]  [reset all]
      (o) color saturation   [histogram]
      (o) red
      (o) green
      (o) blue
                                 [done] [cancel]
***/

   zdedit = zdialog_new(title,mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"frame","fr1","dialog",0,"expand");
   zdialog_add_widget(zdedit,"hbox","hba","dialog");
   zdialog_add_widget(zdedit,"label","labda","hba",Bdarker,"space=5");
   zdialog_add_widget(zdedit,"label","space","hba",0,"expand");
   zdialog_add_widget(zdedit,"label","labba","hba",Blighter,"space=5");
   zdialog_add_widget(zdedit,"hbox","hbb","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"button","b +++","hbb","+++");
   zdialog_add_widget(zdedit,"button","b ---","hbb","‒ ‒ ‒");
   zdialog_add_widget(zdedit,"button","b +-", "hbb"," + ‒ ");
   zdialog_add_widget(zdedit,"button","b -+", "hbb"," ‒ + ");
   zdialog_add_widget(zdedit,"button","b +-+","hbb","+ ‒ +");
   zdialog_add_widget(zdedit,"button","b -+-","hbb","‒ + ‒");
   zdialog_add_widget(zdedit,"hbox","hb2","dialog");
   zdialog_add_widget(zdedit,"vbox","vb21","hb2");
   zdialog_add_widget(zdedit,"vbox","vb22","hb2");
   zdialog_add_widget(zdedit,"radio","radbri","vb21",Bbrightness);
   zdialog_add_widget(zdedit,"radio","radsat","vb21",ZTX("color saturation"));
   zdialog_add_widget(zdedit,"radio","radR","vb21",Bred);
   zdialog_add_widget(zdedit,"radio","radG","vb21",Bgreen);
   zdialog_add_widget(zdedit,"radio","radB","vb21",Bblue);
   zdialog_add_widget(zdedit,"hbox","hbrs","vb22",0,"space=5");
   zdialog_add_widget(zdedit,"label","space","hbrs",0,"space=10");
   zdialog_add_widget(zdedit,"button","reset1","hbrs",ZTX(" reset 1 "));
   zdialog_add_widget(zdedit,"button","resetA","hbrs",ZTX("reset all"));
   zdialog_add_widget(zdedit,"hbox","hbhist","vb22",0,"space=6");
   zdialog_add_widget(zdedit,"label","space","hbhist",0,"space=10");
   zdialog_add_widget(zdedit,"button","histo","hbhist",ZTX("histogram"));
   
   GtkWidget *frame = zdialog_widget(zdedit,"fr1");                        //  setup for curve editing
   curve_init(frame,tune_curve_update);

   zdialog_stuff(zdedit,"radbri",1);                                       //  stuff default selection
   
   zdialog_resize(zdedit,0,380);
   zdialog_run(zdedit,tune_dialog_event);                                  //  run dialog - parallel
   start_thread(tune_thread,0);                                            //  start working thread
   return;
}


//  dialog event and completion callback function

int tune_dialog_event(zdialog *zd, cchar *event)
{
   using namespace splinecurve;

   int         Fupdate = 0;
   int         ii, jj;
   double      px, py;
   
   if (zd->zstat)
   {
      if (zd->zstat == 1) edit_done();                                     //  done
      else edit_cancel();                                                  //  cancel or destroy
      return 0;
   }

   if (strEqu(event,"undo")) edit_undo();                                  //  v.10.2
   if (strEqu(event,"redo")) edit_redo();                                  //  v.10.3
   if (strEqu(event,"blendwidth")) signal_thread();                        //  v.10.3
   if (strEqu(event,"histo")) m_histogram(0,0);                            //  popup brightness histogram

   if (strnEqu(event,"rad",3)) {                                           //  new choice of curve
      ii = strcmpv(event,"radbri","radsat","radR","radG","radB",null);
      tune_spc = ii;
      nap[0] = nap[ii];                                                    //  copy active curve to curve 0
      for (jj = 0; jj < nap[0]; jj++) {
         apx[0][jj] = apx[ii][jj];
         apy[0][jj] = apy[ii][jj];
      }
      Fupdate++;
   }
   
   ii = tune_spc;                                                          //  current active curve

   if (strnEqu(event,"b ",2)) {                                            //  button to move entire curve
      for (jj = 0; jj < nap[ii]; jj++) {
         px = apx[0][jj];
         py = apy[0][jj];
         if (strEqu(event,"b +++")) py += 0.1;
         if (strEqu(event,"b ---")) py -= 0.1;
         if (strEqu(event,"b +-"))  py += 0.1 - 0.2 * px;
         if (strEqu(event,"b -+"))  py -= 0.1 - 0.2 * px;
         if (strEqu(event,"b +-+")) py -= 0.05 - 0.2 * fabs(px-0.5);
         if (strEqu(event,"b -+-")) py += 0.05 - 0.2 * fabs(px-0.5);
         if (py > 1) py = 1;
         if (py < 0) py = 0;
         apy[0][jj] = py;
      }
      tune_spc_moved[ii] = 1;
      Fupdate++;
   }
   
   if (strEqu(event,"reset1")) {                                           //  reset current curve
      nap[0] = 3;
      apx[0][0] = 0.01;                                                    //  3 anchor points, flatline
      apy[0][0] = 0.5;
      apx[0][1] = 0.5;
      apy[0][1] = 0.5;
      apx[0][2] = 0.99;
      apy[0][2] = 0.5;
      Fupdate++;
      tune_spc_moved[ii] = 0;
   }
   
   if (strEqu(event,"resetA")) 
   {
      for (jj = 0; jj < 6; jj++) {                                         //  reset all curves
         nap[jj] = 3;
         apx[jj][0] = 0.01;
         apy[jj][0] = 0.5;
         apx[jj][1] = 0.5;
         apy[jj][1] = 0.5;
         apx[jj][2] = 0.99;
         apy[jj][2] = 0.5;
         curve_generate(jj);                                               //  regenerate all
         tune_spc_moved[jj] = 0;
      }
      Fupdate++;
   }
   
   if (Fupdate)                                                            //  curve has changed
   {
      curve_generate(0);                                                   //  regenerate curve 0

      nap[ii] = nap[0];                                                    //  copy curve 0 to active curve
      for (jj = 0; jj < nap[0]; jj++) {
         apx[ii][jj] = apx[0][jj];
         apy[ii][jj] = apy[0][jj];
      }
      for (jj = 0; jj < 1000; jj++)
         yval[ii][jj] = yval[0][jj];

      curve_draw();                                                        //  draw curve
      signal_thread();                                                     //  trigger image update
   }
   
   return 1;
}


//  this function is called when curve 0 is edited using mouse

void  tune_curve_update(int)
{
   using namespace splinecurve;
   
   int   ii = tune_spc, jj;

   nap[ii] = nap[0];                                                       //  copy curve 0 to current curve
   for (jj = 0; jj < nap[0]; jj++) {
      apx[ii][jj] = apx[0][jj];
      apy[ii][jj] = apy[0][jj];
   }
   for (jj = 0; jj < 1000; jj++)
      yval[ii][jj] = yval[0][jj];

   tune_spc_moved[ii] = 1;
   
   signal_thread();                                                        //  trigger image update
   return;
}


//  Update image based on latest settings of all dialog controls.

void * tune_thread(void *)
{
   void * tune_wthread(void *arg);

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wt(tune_wthread,&wtnx[ii]);
      wait_wts();                                                          //  wait for completion

      Fmodified = 1;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}


void * tune_wthread(void *arg)                                             //  worker thread function
{
   void  tune1pix(int px, int py);

   int      px, py;
   int      index = *((int *) arg);

   for (py = index; py < E1hh; py += Nwt)
   for (px = 0; px < E1ww; px++)
      tune1pix(px,py);

   exit_wt();
   return 0;                                                               //  not executed, avoid gcc warning
}


void tune1pix(int px, int py)                                              //  process one pixel
{
   using namespace splinecurve;

   int         ii, dist = 0;
   uint16      *pix1, *pix3;
   double      red1, green1, blue1, red3, green3, blue3;
   double      xval, brmax, brmean, brout, brightness;
   
   if (sa_active) {                                                        //  select area active
      ii = py * Fww + px;
      dist = sa_pixisin[ii];                                               //  distance from edge
      if (! dist) return;                                                  //  outside pixel
   }

   pix1 = PXMpix(E1pxm16,px,py);                                           //  input pixel
   pix3 = PXMpix(E3pxm16,px,py);                                           //  output pixel
   
   red1 = red3 = pix1[0];                                                  //  input and output RGB values
   green1 = green3 = pix1[1];
   blue1 = blue3 = pix1[2];

   brightness = 0.25 * red3 + 0.65 * green3 + 0.10 * blue3;                //  perceived brightness   v.10.9
   xval = brightness / 65536.0;                                            //  curve x-value, 0 to 0.999

/* ------------------------------------------------------------------------

      brightness adjustment

      brightness curve values:
           0 = dark
         0.5 = normal, unchanged
         1.0 = 200% brightness, clipped
*/

   if (tune_spc_moved[1])                                                  //  curve has been edited
   {
      brout = curve_yval(1,xval);                                          //  brightness factor, 0 - 1
      
      brmax = red3;                                                        //  brmax = brightest color
      if (green3 > brmax) brmax = green3;
      if (blue3 > brmax) brmax = blue3;

      brout = brout / 0.5;                                                 //  0 - 2.0
      red3 = red3 * brout;                                                 //  apply to all colors
      green3 = green3 * brout;                                             //  remove limit check     v.10.9
      blue3 = blue3 * brout;                                               //  (result is better)
   }
      
/* ------------------------------------------------------------------------

      color saturation curve values:
           0 = no color saturation (gray scale)
         0.5 = normal (initial unmodified RGB)
         1.0 = max. color saturation

      0.5 >> 0:  move all RGB values to their mean: (R+G+B)/3
      0.5 >> 1.0:  increase RGB spread to 2x original level
*/

   if (tune_spc_moved[2])                                                  //  curve has been edited
   {
      brout = curve_yval(2,xval);                                          //  saturation factor, 0 - 1
      brout = 2.0 * brout - 1.0;                                           //  -1.0 .. +1.0
      brmean = 0.333 * (red3 + green3 + blue3);
      red3 = red3 + brout * (red3 - brmean);                               //  simplified       v.10.9
      green3 = green3 + brout * (green3 - brmean);
      blue3 = blue3 + brout * (blue3 - brmean);
      if (red3 < 0) red3 = 0;
      if (green3 < 0) green3 = 0;
      if (blue3 < 0) blue3 = 0;
   }

/* ------------------------------------------------------------------------

      color balance curve values:
           0 = 0.5 * original color
         0.5 = unmodified
         1.0 = 1.5 * original color, clipped
*/

   if (tune_spc_moved[3]) {                                                //  curve has been edited
      brout = curve_yval(3,xval);
      red3 = red3 * (brout + 0.5);
   }

   if (tune_spc_moved[4]) {                                                //  curve has been edited
      brout = curve_yval(4,xval);
      green3 = green3 * (brout + 0.5);
   }

   if (tune_spc_moved[5]) {                                                //  curve has been edited
      brout = curve_yval(5,xval);
      blue3 = blue3 * (brout + 0.5);
   }

/* ------------------------------------------------------------------------
   
   if working within a select area, blend changes over distance from edge

*/
      
   double      dold, dnew;

   if (sa_active && dist < sa_blend) {
      dnew = 1.0 * dist / sa_blend;
      dold = 1.0 - dnew;
      red3 = dnew * red3 + dold * red1;
      green3 = dnew * green3 + dold * green1;
      blue3 = dnew * blue3 + dold * blue1;
   }

/* ------------------------------------------------------------------------
   
   prevent clipping and set output RGB values

*/

   if (red3 > 65535) red3 = 65535;
   if (green3 > 65535) green3 = 65535;
   if (blue3 > 65535) blue3 = 65535;

   pix3[0] = int(red3);
   pix3[1] = int(green3);
   pix3[2] = int(blue3);
   
   return;
}


/**************************************************************************/

//  ramp brightness across image, vertical and horizontal gradients        //  new  v.9.3

void m_brightramp(GtkWidget *, cchar *)
{
   using namespace splinecurve;
   
   int    brramp_dialog_event(zdialog* zd, cchar *event);
   void   brramp_curvedit(int spc);
   void * brramp_thread(void *);
   void   brramp_mousefunc();

   zfuncs::F1_help_topic = "brightness_ramp";                              //  v.10.8

   if (! edit_setup("briteramp",1,2)) return;                              //  setup edit: preview, select area OK

   zdedit = zdialog_new(ZTX("Ramp brightness across image"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"space=5|expand");
   zdialog_add_widget(zdedit,"vbox","vb1","hb1");
   zdialog_add_widget(zdedit,"vbox","vb2","hb1",0,"expand");
   zdialog_add_widget(zdedit,"label","lmax","vb1","+","space=3");
   zdialog_add_widget(zdedit,"label","lspace","vb1",0,"expand");
   zdialog_add_widget(zdedit,"label","lmin","vb1","‒","space=3");
   zdialog_add_widget(zdedit,"label","lspace","vb1");
   zdialog_add_widget(zdedit,"frame","frame","vb2",0,"expand");
   zdialog_add_widget(zdedit,"hbox","hb2","vb2");
   zdialog_add_widget(zdedit,"label","lmin","hb2","‒","space=3");
   zdialog_add_widget(zdedit,"label","lspace","hb2",0,"expand");
   zdialog_add_widget(zdedit,"label","lmax","hb2","+","space=3");

   Nspc = 2;

   nap[0] = 4;
   vert[0] = 0;
   apx[0][0] = 0.01;
   apx[0][1] = 0.4;
   apx[0][2] = 0.6;
   apx[0][3] = 0.99;
   apy[0][0] = apy[0][1] = apy[0][2] = apy[0][3] = 0.5;

   nap[1] = 4;
   vert[1] = 1;
   apx[1][0] = 0.01;
   apx[1][1] = 0.4;
   apx[1][2] = 0.6;
   apx[1][3] = 0.99;
   apy[1][0] = apy[1][1] = apy[1][2] = apy[1][3] = 0.5;

   curve_generate(0);
   curve_generate(1);

   GtkWidget *frame = zdialog_widget(zdedit,"frame");
   curve_init(frame,brramp_curvedit);

   zdialog_resize(zdedit,260,280);
   zdialog_run(zdedit,brramp_dialog_event);                                //  run dialog, parallel

   start_thread(brramp_thread,0);                                          //  start working thread
   signal_thread();
   return;
}


//  dialog event and completion callback function

int brramp_dialog_event(zdialog *zd, cchar *event)
{
   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) edit_done();
      else edit_cancel();
      return 1;
   }
   
   if (strEqu(event,"undo")) edit_undo();                                  //  v.10.2
   if (strEqu(event,"redo")) edit_redo();                                  //  v.10.3
   if (strEqu(event,"blendwidth")) signal_thread();                        //  v.10.3
   return 1;
}


//  this function is called when a curve is edited

void brramp_curvedit(int spc)
{
   signal_thread();
   return;
}


//  brramp thread function

void * brramp_thread(void *arg)
{
   void * brramp_wthread(void *);
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wt(brramp_wthread,&wtnx[ii]);
      wait_wts();                                                          //  wait for completion

      mwpaint2();                                                          //  update window
      Fmodified = 1;                                                       //  image3 modified
   }

   return 0;                                                               //  not executed, stop g++ warning
}


void * brramp_wthread(void *arg)                                           //  worker thread function
{
   using namespace splinecurve;

   int         index = *((int *) arg);
   int         ii, dist = 0, px3, py3;
   int         red1, green1, blue1, maxrgb;
   int         red3, green3, blue3;
   uint16      *pix1, *pix3;
   double      dispx, dispy;
   double      hramp, vramp, tramp;
   double      spanw, spanh, dold, dnew;

   if (sa_active) {                                                        //  if select area active, ramp
      spanw = sa_maxx - sa_minx;                                           //    brightness over enclosing rectangle
      spanh = sa_maxy - sa_miny;
   }
   else {
      spanw = E3ww;                                                        //  else over entire image
      spanh = E3hh;
   }   

   for (py3 = index; py3 < E3hh; py3 += Nwt)                               //  loop output pixels
   for (px3 = 0; px3 < E3ww; px3++)
   {
      if (sa_active) {                                                     //  select area active        v.10.1
         ii = py3 * E3ww + px3;
         dist = sa_pixisin[ii];                                            //  distance from edge
         if (! dist) continue;                                             //  pixel outside area
         dispx = (px3 - sa_minx) / spanw;
         dispy = (py3 - sa_miny) / spanh;
      }
      else {
         dispx = px3 / spanw;                                              //  left > right = 0 to 1
         dispy = py3 / spanh;                                              //  top > bottom = 0 to 1
      }
      
      hramp = curve_yval(0,dispx) - 0.5;                                   //  -0.5 to +0.5
      vramp = curve_yval(1,dispy) - 0.5;                                   //  -0.5 to +0.5
      tramp = 1.0 + hramp + vramp;

      pix1 = PXMpix(E1pxm16,px3,py3);                                      //  input pixel
      pix3 = PXMpix(E3pxm16,px3,py3);                                      //  output pixel
      
      red1 = pix1[0];
      green1 = pix1[1];
      blue1 = pix1[2];

      maxrgb = red1;
      if (green1 > maxrgb) maxrgb = green1;
      if (blue1 > maxrgb) maxrgb = blue1;

      if (tramp * maxrgb > 65535) tramp = 65535.0 / maxrgb;

      red3 = tramp * red1;
      green3 = tramp * green1;
      blue3 = tramp * blue1;

      if (sa_active && dist < sa_blend) {                                  //  blend changes over blendwidth   v.10.1
         dnew = 1.0 * dist / sa_blend;
         dold = 1.0 - dnew;
         red3 = dnew * red3 + dold * red1;
         green3 = dnew * green3 + dold * green1;
         blue3 = dnew * blue3 + dold * blue1;
      }

      pix3[0] = red3;
      pix3[1] = green3;
      pix3[2] = blue3;
   }

   exit_wt();
   return 0;                                                               //  not executed, avoid gcc warning
}


/**************************************************************************/

//  Clip pixels outside the given low/high range and expand the remaining
//  brightness range back to the full range of 0 to 65535.

double   xbrangeD, xbrangeB;

void m_xbrange(GtkWidget *, cchar *)                                       //  new v.10.1
{
   int      xbrange_dialog_event(zdialog *zd, cchar *event);
   void *   xbrange_thread(void *);
   
   zfuncs::F1_help_topic = "expand_brightness";                            //  v.10.8

   cchar  *title = ZTX("Expand Brightness Range");
   cchar  *labP = ZTX("brightness to clip (percent)");
   cchar  *labD = ZTX("dark pixels");
   cchar  *labB = ZTX("bright pixels");

   if (! edit_setup("expand",1,2)) return;                                 //  setup: preview, select area OK

   zdedit = zdialog_new(title,mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"label","labP","dialog",labP,"space=5");      //  brightness to clip (percent)
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"space=5");           //  dark pixels [__]  bright pixels [__]
   zdialog_add_widget(zdedit,"label","labD","hb1",labD,"space=4");         //
   zdialog_add_widget(zdedit,"spin","clipD","hb1","0|49|0.1|0");           //  [done] [cancel]
   zdialog_add_widget(zdedit,"label","labS","hb1","","space=10");
   zdialog_add_widget(zdedit,"label","labB","hb1",labB,"space=4");
   zdialog_add_widget(zdedit,"spin","clipB","hb1","0|49|0.1|0");

   xbrangeD = xbrangeB = 0;                                                //  initial clip = 0
   
   zdialog_run(zdedit,xbrange_dialog_event);                               //  run dialog, parallel

   start_thread(xbrange_thread,0);                                         //  start working thread
   return;
}


//  dialog event and completion callback function

int xbrange_dialog_event(zdialog *zd, cchar *event)
{
   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) edit_done();
      else edit_cancel();
      return 1;
   }

   if (strEqu(event,"undo")) edit_undo();                                  //  v.10.2
   if (strEqu(event,"redo")) edit_redo();                                  //  v.10.3
   if (strEqu(event,"blendwidth")) signal_thread();                        //  v.10.3

   if (strEqu(event,"clipD")) {
      zdialog_fetch(zd,"clipD",xbrangeD);
      signal_thread();
   }

   if (strEqu(event,"clipB")) {
      zdialog_fetch(zd,"clipB",xbrangeB);
      signal_thread();
   }

   return 0;
}


//  thread function

void * xbrange_thread(void *)
{
   int         ii, px, py, rgb, dist = 0;
   double      dark, bright, b1, b3, bf;
   double      pval1, pval3, f1, f2;
   uint16      *pix1, *pix3;
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      dark = 0.01 * xbrangeD * 65536;                                      //  clipping brightness levels
      bright = (1.0 - 0.01 * xbrangeB) * 65536;

      for (py = 0; py < E3hh; py++)                                        //  loop all image pixels
      for (px = 0; px < E3ww; px++)
      {
         ii = py * E3ww + px;

         if (sa_active) {                                                  //  select area active     bugfix v.9.9
            dist = sa_pixisin[ii];                                         //  distance from edge
            if (! dist) continue;                                          //  pixel is outside area
         }

         pix1 = PXMpix(E1pxm16,px,py);                                     //  input pixel
         pix3 = PXMpix(E3pxm16,px,py);                                     //  output pixel

         b1 = pixbright(pix1);

         if (b1 < dark)                                                    //  clip dark pixels
            b3 = 0;
         else if (b1 > bright)                                             //  clip bright pixels
            b3 = bright;
         else b3 = b1;
         
         if (b3 > dark) b3 = b3 - dark;                                    //  expand the rest
         b3 = b3 * (65535.0 / (bright - dark));

         bf = b3 / (b1 + 1);                                               //  brightness ratio
         
         for (rgb = 0; rgb < 3; rgb++)                                     //  loop 3 RGB colors
         {
            pval1 = pix1[rgb];
            pval3 = bf * pval1;                                            //  apply ratio
            if (pval3 > 65535) pval3 = 65535;

            if (sa_active && dist < sa_blend) {                            //  select area is active,
               f1 = 1.0 * dist / sa_blend;                                 //    blend changes over sa_blend
               f2 = 1.0 - f1;
               pval3 = int(f1 * pval3 + f2 * pval1);
            }

            pix3[rgb] = pval3;
         }
      }

      Fmodified = 1;
      mwpaint2();
   }

   return 0;                                                               //  not executed, stop g++ warning
}


/**************************************************************************

   Image Tone Mapping function
   enhance local contrast as opposed to overall contrast

   methodology:
   get brightness gradients for each pixel in 4 directions: SE SW NE NW
   amplify gradients using the edit curve (x-axis range 0-max. gradient)
   integrate 4 new brightness surfaces from the amplified gradients:
     - pixel brightness = prior pixel brightness + amplified gradient
     - use constraint factor to push brightness toward original value
   new pixel brightness = average from 4 calculated brightness surfaces

***************************************************************************/

float    *Tmap_brmap1;
float    *Tmap_brmap3[4];
int      Tmap_contrast99;
double   Tmap_constraint;

void m_tonemap(GtkWidget *, cchar *)                                       //  new v.9.8
{
   using namespace splinecurve;
   
   int      Tmap_dialog_event(zdialog *zd, cchar *event);
   void     Tmap_curvedit(int);
   void *   Tmap_thread(void *);

   zfuncs::F1_help_topic = "tone_mapping";                                 //  v.10.8

   int         ii, cc, px, py;
   uint16      *pix1;
   cchar       *title = ZTX("Tone Mapping");
   cchar       *Bcons = ZTX("Constrain");
   int         jj, sum, limit, condist[100];

   if (! edit_setup("tonemap",0,2)) return;                                //  setup: no preview, select area OK

   zdedit = zdialog_new(title,mWin,Bdone,Bcancel,null);                       //   _____________________________
   zdialog_add_widget(zdedit,"frame","frame","dialog",0,"expand");            //  |                             |
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"space=5");              //  |    curve drawing area       |
   zdialog_add_widget(zdedit,"label","labcL","hb1","low","space=4");          //  |                             |
   zdialog_add_widget(zdedit,"label","labcM","hb1","contrast","expand");      //  |_____________________________| 
   zdialog_add_widget(zdedit,"label","labcH","hb1","high","space=5");         //    low      contrast      high
   zdialog_add_widget(zdedit,"hbox","hb2","dialog",0,"space=5");              //
   zdialog_add_widget(zdedit,"label","labcon","hb2",Bcons,"space=5");         //    constrain ======[]======
   zdialog_add_widget(zdedit,"hscale","cons","hb2","0|1|0.01|0.4","expand");
   
   GtkWidget *frame = zdialog_widget(zdedit,"frame");                      //  set up curve edit
   curve_init(frame,Tmap_curvedit);

   Nspc = 1;
   vert[0] = 0;
   nap[0] = 3;                                                             //  initial curve anchor points
   apx[0][0] = 0.01;
   apy[0][0] = 0.2;
   apx[0][1] = 0.50;
   apy[0][1] = 0.4;
   apx[0][2] = 0.99;
   apy[0][2] = 0.0;

   curve_generate(0);                                                      //  generate curve data

   cc = Fww * Fhh * sizeof(float);                                         //  allocate brightness map memory
   Tmap_brmap1 = (float *) zmalloc(cc,"tmap.br1");
   for (ii = 0; ii < 4; ii++)
      Tmap_brmap3[ii] = (float *) zmalloc(cc,"tmap.br3");

   for (py = 0; py < Fhh; py++)                                            //  map initial image brightness
   for (px = 0; px < Fww; px++)
   {
      ii = py * Fww + px;
      pix1 = PXMpix(E1pxm16,px,py);
      Tmap_brmap1[ii] = pixbright(pix1);
   }

   for (ii = 0; ii < 100; ii++) 
      condist[ii] = 0;

   for (py = 1; py < Fhh; py++)                                            //  map contrast distribution
   for (px = 1; px < Fww; px++)
   {
      ii = py * Fww + px;
      jj = 0.00152 * fabsf(Tmap_brmap1[ii] - Tmap_brmap1[ii-1]);           //  contrast, ranged 0 - 99
      condist[jj]++;
      jj = 0.00152 * fabsf(Tmap_brmap1[ii] - Tmap_brmap1[ii-Fww]);
      condist[jj]++;
   }
   
   sum = 0;
   limit = 0.99 * 2 * (Fww-1) * (Fhh-1);                                   //  find 99th percentile contrast

   for (ii = 0; ii < 100; ii++) {
      sum += condist[ii];
      if (sum > limit) break;
   }
   
   Tmap_contrast99 = 65535.0 * ii / 100.0;                                 //  0 to 65535
   if (Tmap_contrast99 < 1000) Tmap_contrast99 = 1000;                     //  rescale low-contrast image   v.10.9

   zdialog_resize(zdedit,300,250);
   zdialog_run(zdedit,Tmap_dialog_event);                                  //  run dialog, parallel

   Tmap_constraint = 0.5;
   start_thread(Tmap_thread,0);                                            //  start working thread
   signal_thread();
   return;
}


//  dialog event and completion callback function

int Tmap_dialog_event(zdialog *zd, cchar *event)
{
   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) edit_done();
      else edit_cancel();
      zfree(Tmap_brmap1);                                                  //  free memory
      for (int ii = 0; ii < 4; ii++)
      zfree(Tmap_brmap3[ii]);
      return 1;
   }

   if (strEqu(event,"undo")) edit_undo();                                  //  v.10.2
   if (strEqu(event,"redo")) edit_redo();                                  //  v.10.3
   if (strEqu(event,"blendwidth")) signal_thread();                        //  v.10.3

   if (strEqu(event,"cons")) {
      zdialog_fetch(zd,"cons",Tmap_constraint);
      signal_thread();
   }
   
   return 0;
}


//  this function is called when the curve is edited

void Tmap_curvedit(int)
{
   signal_thread();
   return;
}


//  thread function

void * Tmap_thread(void *)
{
   void * Tmap_wthread1(void *arg);
   void * Tmap_wthread2(void *arg);

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      if (sa_active) SB_goal = sa_Npixel;                                  //  setup progress tracking   v.10.2
      else  SB_goal = E3ww * E3hh;
      SB_goal = SB_goal * 2;                                               //  2 passes through all pixels
      SB_done = 0;

      for (int ii = 0; ii < 4; ii++)                                       //  start working threads1 (must be 4)
         start_wt(Tmap_wthread1,&wtnx[ii]);
      wait_wts();                                                          //  wait for completion

      for (int ii = 0; ii < Nwt; ii++)                                     //  start working threads2
         start_wt(Tmap_wthread2,&wtnx[ii]);
      wait_wts();                                                          //  wait for completion

      SB_goal = SB_done = 0;

      Fmodified = 1;
      mwpaint2(); 
   }

   return 0;                                                               //  not executed, stop g++ warning
}


//  working threads

void * Tmap_wthread1(void *arg)                                            //  hardened   v.9.9
{
   using namespace splinecurve;

   int         ii, bii, pii, dist = 0;
   int         px, px1, px2, pxinc;
   int         py, py1, py2, pyinc;
   float       b1, b3, xval, yval, grad;
   float       cons1, cons2, contrast99;

   bii = *((int *) arg);

   if (bii == 0) {                                                         //  direction SE
      px1 = 1; px2 = E3ww; pxinc = 1;
      py1 = 1; py2 = E3hh; pyinc = 1;
      pii = - 1 - E3ww;
   }
   
   else if (bii == 1) {                                                    //  direction SW
      px1 = E3ww-2; px2 = 0; pxinc = -1;
      py1 = 1; py2 = E3hh; pyinc = 1;
      pii = + 1 - E3ww;
   }
   
   else if (bii == 2) {                                                    //  direction NE
      px1 = 1; px2 = E3ww; pxinc = 1;
      py1 = E3hh-2; py2 = 0; pyinc = -1;
      pii = - 1 + E3ww;
   }
   
   else {   /* bii == 3 */                                                 //  direction NW
      px1 = E3ww-2; px2 = 0; pxinc = -1;
      py1 = E3hh-2; py2 = 0; pyinc = -1;
      pii = + 1 + E3ww;
   }

   contrast99 = Tmap_contrast99;                                           //  99th percentile contrast
   contrast99 = 1.0 / contrast99;                                          //  inverted

   cons1 = pow(Tmap_constraint,2);                                         //  get constraint, 0 to 1
   cons2 = 1.0 - cons1;

   for (ii = 0; ii < E3ww * E3hh; ii++)                                    //  initial brightness map
      Tmap_brmap3[bii][ii] = Tmap_brmap1[ii];

   for (py = py1; py != py2; py += pyinc)                                  //  loop all image pixels
   for (px = px1; px != px2; px += pxinc)
   {
      ii = py * E3ww + px;

      if (sa_active) {                                                     //  select area active
         dist = sa_pixisin[ii];                                            //  distance from edge
         if (! dist) continue;                                             //  outside pixel
      }

      b1 = Tmap_brmap1[ii];                                                //  this pixel brightness
      grad = b1 - Tmap_brmap1[ii+pii];                                     //  - prior pixel --> gradient

      xval = fabsf(grad) * contrast99;                                     //  gradient scaled 0 to 1+
      yval = 1.0 + 5 * curve_yval(0,xval);                                 //  1 to 6
      grad = grad * yval;                                                  //  magnified gradient

      b3 = Tmap_brmap3[bii][ii+pii] + grad;                                //  pixel brightness = prior + gradient
      b3 = cons1 * b1 + cons2 * b3;                                        //  constrain: push b3 toward b1
      
      if (b3 > 65535) b3 = 65535;                                          //  constrain
      if (b3 < 10) b3 = 10;
      
      Tmap_brmap3[bii][ii] = b3;                                           //  new pixel brightness
      
      SB_done++;                                                           //  track progress         v.10.2
   }

   exit_wt();
   return 0;                                                               //  not executed, avoid gcc warning
}


void * Tmap_wthread2(void *arg)                                            //  speedup   v.10.2
{
   uint16      *pix1, *pix3;
   int         index, ii, px, py, dist = 0;
   int         rgb, pval1, pval3;
   float       b1, b3, bf, f1, f2;

   index = *((int *) arg);

   for (py = index; py < E3hh; py += Nwt)                                  //  loop image pixels
   for (px = 0; px < E3ww; px++)
   {
      ii = py * E3ww + px;

      if (sa_active) {                                                     //  select area active     bugfix v.9.9
         dist = sa_pixisin[ii];                                            //  distance from edge
         if (! dist) continue;                                             //  pixel is outside area
      }

      pix1 = PXMpix(E1pxm16,px,py);                                        //  input pixel
      pix3 = PXMpix(E3pxm16,px,py);                                        //  output pixel

      b1 = Tmap_brmap1[ii];                                                //  initial pixel brightness
      b3 = Tmap_brmap3[0][ii] + Tmap_brmap3[1][ii]                         //  new brightness = average of four
         + Tmap_brmap3[2][ii] + Tmap_brmap3[3][ii];                        //    calculated brightness surfaces
      
      bf = 0.25 * b3 / (b1 + 1);                                           //  brightness ratio
      
      for (rgb = 0; rgb < 3; rgb++)                                        //  loop 3 RGB colors
      {
         pval1 = pix1[rgb];
         pval3 = bf * pval1;                                               //  apply ratio
         if (pval3 > 65535) pval3 = 65535;

         if (sa_active && dist < sa_blend) {                               //  select area is active,
            f1 = 1.0 * dist / sa_blend;                                    //    blend changes over sa_blend
            f2 = 1.0 - f1;
            pval3 = int(f1 * pval3 + f2 * pval1);
         }

         pix3[rgb] = pval3;
      }

      SB_done++;                                                           //  track progress         v.10.2
   }

   exit_wt();
   return 0;                                                               //  not executed, stop g++ warning
}


/**************************************************************************/

//  red eye removal function

struct sredmem {                                                           //  red-eye struct in memory
   char        type, space[3];
   int         cx, cy, ww, hh, rad, clicks;
   double      thresh, tstep;
};
sredmem  redmem[100];                                                      //  store up to 100 red-eyes

int      Nredmem = 0, maxredmem = 100;


void m_redeye(GtkWidget *, cchar *)
{
   void     redeye_mousefunc();
   int      redeye_dialog_event(zdialog *zd, cchar *event);

   cchar    *redeye_message = ZTX(
               "Method 1:\n"
               "  Left-click on red-eye to darken.\n"
               "Method 2:\n"
               "  Drag down and right to enclose red-eye.\n"
               "  Left-click on red-eye to darken.\n"
               "Undo red-eye:\n"
               "  Right-click on red-eye.");

   zfuncs::F1_help_topic = "red_eye";                                      //  v.10.8

   if (! edit_setup("redeye",0,1)) return;                                 //  setup edit: no preview

   zdedit = zdialog_new(ZTX("Red Eye Reduction"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"label","lab1","dialog",redeye_message);
   zdialog_run(zdedit,redeye_dialog_event);                                //  run dialog, parallel mode

   Nredmem = 0;
   mouseCBfunc = redeye_mousefunc;                                         //  connect mouse function
   Mcapture = 1;                                                           //  capture mouse clicks
   return;
}


//  dialog event and completion callback function

int redeye_dialog_event(zdialog *zd, cchar *event)
{
   if (! zd->zstat) return 0;

   mouseCBfunc = 0;                                                        //  disconnect mouse
   Mcapture = 0;
   if (Nredmem > 0) Fmodified = 1;
   Ftoparc = ptoparc = 0;
   if (zd->zstat == 1) edit_done();
   else edit_cancel();
   return 0;
}


//  mouse functions to define, darken, and undo red-eyes

int      redeye_createF(int px, int py);                                   //  create 1-click red-eye (type F)
int      redeye_createR(int px, int py, int ww, int hh);                   //  create robust red-eye (type R)
void     redeye_darken(int ii);                                            //  darken red-eye
void     redeye_distr(int ii);                                             //  build pixel redness distribution
int      redeye_find(int px, int py);                                      //  find red-eye at mouse position
void     redeye_remove(int ii);                                            //  remove red-eye at mouse position
int      redeye_radlim(int cx, int cy);                                    //  compute red-eye radius limit


void redeye_mousefunc()
{
   int         ii, px, py, ww, hh;

   if (Nredmem == maxredmem) {
      zmessageACK(MWIN,"%d red-eye limit reached",maxredmem);              //  too many red-eyes
      return;
   }

   if (LMclick)                                                            //  left mouse click
   {
      LMclick = 0;

      px = Mxclick;                                                        //  click position
      py = Myclick;
      if (px < 0 || px > E3ww-1 || py < 0 || py > E3hh-1) return;          //  outside image area

      ii = redeye_find(px,py);                                             //  find existing red-eye
      if (ii < 0) ii = redeye_createF(px,py);                              //  or create new type F
      redeye_darken(ii);                                                   //  darken red-eye
   }
   
   if (RMclick)                                                            //  right mouse click
   {
      RMclick = 0;
      px = Mxclick;                                                        //  click position
      py = Myclick;
      ii = redeye_find(px,py);                                             //  find red-eye
      if (ii >= 0) redeye_remove(ii);                                      //  if found, remove
   }

   if (Mxdrag || Mydrag)                                                   //  mouse drag underway
   {
      px = Mxdown;                                                         //  initial position
      py = Mydown;
      ww = Mxdrag - Mxdown;                                                //  increment
      hh = Mydrag - Mydown;
      if (ww < 2 && hh < 2) return;
      if (ww < 2) ww = 2;
      if (hh < 2) hh = 2;
      if (px < 1) px = 1;                                                  //  keep within image area
      if (py < 1) py = 1;      
      if (px + ww > E3ww-1) ww = E3ww-1 - px;
      if (py + hh > E3hh-1) hh = E3hh-1 - py;
      ii = redeye_find(px,py);                                             //  find existing red-eye
      if (ii >= 0) redeye_remove(ii);                                      //  remove it
      ii = redeye_createR(px,py,ww,hh);                                    //  create new red-eye type R
   }

   mwpaint2();
   return;
}


//  create type F redeye (1-click automatic)

int redeye_createF(int cx, int cy)
{
   int         cx0, cy0, cx1, cy1, px, py, rad, radlim;
   int         loops, ii;
   int         Tnpix, Rnpix, R2npix;
   double      rd, rcx, rcy, redpart;
   double      Tsum, Rsum, R2sum, Tavg, Ravg, R2avg;
   double      sumx, sumy, sumr;
   uint16      *ppix;
   
   cx0 = cx;
   cy0 = cy;
   
   for (loops = 0; loops < 8; loops++)
   {
      cx1 = cx;
      cy1 = cy;

      radlim = redeye_radlim(cx,cy);                                       //  radius limit (image edge)
      Tsum = Tavg = Ravg = Tnpix = 0;

      for (rad = 0; rad < radlim-2; rad++)                                 //  find red-eye radius from (cx,cy)
      {
         Rsum = Rnpix = 0;
         R2sum = R2npix = 0;

         for (py = cy-rad-2; py <= cy+rad+2; py++)
         for (px = cx-rad-2; px <= cx+rad+2; px++)
         {
            rd = sqrt((px-cx)*(px-cx) + (py-cy)*(py-cy));
            ppix = PXMpix(E3pxm16,px,py);
            redpart = pixred(ppix);

            if (rd <= rad + 0.5 && rd > rad - 0.5) {                       //  accum. redness at rad
               Rsum += redpart;
               Rnpix++;
            }
            else if (rd <= rad + 2.5 && rd > rad + 1.5) {                  //  accum. redness at rad+2
               R2sum += redpart;
               R2npix++;
            }
         }
         
         Tsum += Rsum;
         Tnpix += Rnpix;
         Tavg = Tsum / Tnpix;                                              //  avg. redness over 0-rad
         Ravg = Rsum / Rnpix;                                              //  avg. redness at rad
         R2avg = R2sum / R2npix;                                           //  avg. redness at rad+2
         if (R2avg > Ravg || Ravg > Tavg) continue;
         if ((Ravg - R2avg) < 0.2 * (Tavg - Ravg)) break;                  //  0.1 --> 0.2
      }
      
      sumx = sumy = sumr = 0;
      rad = int(1.2 * rad + 1);
      if (rad > radlim) rad = radlim;
      
      for (py = cy-rad; py <= cy+rad; py++)                                //  compute center of gravity for
      for (px = cx-rad; px <= cx+rad; px++)                                //   pixels within rad of (cx,cy)
      {
         rd = sqrt((px-cx)*(px-cx) + (py-cy)*(py-cy));
         if (rd > rad + 0.5) continue;
         ppix = PXMpix(E3pxm16,px,py);
         redpart = pixred(ppix);                                           //  weight by redness
         sumx += redpart * (px - cx);
         sumy += redpart * (py - cy);
         sumr += redpart;
      }

      rcx = cx + 1.0 * sumx / sumr;                                        //  new center of red-eye
      rcy = cy + 1.0 * sumy / sumr;
      if (fabs(cx0 - rcx) > 0.6 * rad) break;                              //  give up if big movement
      if (fabs(cy0 - rcy) > 0.6 * rad) break;
      cx = int(rcx + 0.5);
      cy = int(rcy + 0.5);
      if (cx == cx1 && cy == cy1) break;                                   //  done if no change
   }

   radlim = redeye_radlim(cx,cy);
   if (rad > radlim) rad = radlim;

   ii = Nredmem++;                                                         //  add red-eye to memory
   redmem[ii].type = 'F';
   redmem[ii].cx = cx;
   redmem[ii].cy = cy;
   redmem[ii].rad = rad;
   redmem[ii].clicks = 0;
   redmem[ii].thresh = 0;
   return ii;
}


//  create type R red-eye (drag an ellipse over red-eye area)

int redeye_createR(int cx, int cy, int ww, int hh)
{
   int      rad, radlim;

   Ftoparc = 1;                                                            //  paint ellipse over image
   toparcx = cx - ww;
   toparcy = cy - hh;
   toparcw = 2 * ww;
   toparch = 2 * hh;

   if (ww > hh) rad = ww;
   else rad = hh;
   radlim = redeye_radlim(cx,cy);
   if (rad > radlim) rad = radlim;

   int ii = Nredmem++;                                                     //  add red-eye to memory
   redmem[ii].type = 'R';
   redmem[ii].cx = cx;
   redmem[ii].cy = cy;
   redmem[ii].ww = 2 * ww;
   redmem[ii].hh = 2 * hh;
   redmem[ii].rad = rad;
   redmem[ii].clicks = 0;
   redmem[ii].thresh = 0;
   return ii;
}


//  darken a red-eye and increase click count

void redeye_darken(int ii)
{
   int         cx, cy, ww, hh, px, py, rad, clicks;
   double      rd, thresh, tstep;
   char        type;
   uint16      *ppix;

   type = redmem[ii].type;
   cx = redmem[ii].cx;
   cy = redmem[ii].cy;
   ww = redmem[ii].ww;
   hh = redmem[ii].hh;
   rad = redmem[ii].rad;
   thresh = redmem[ii].thresh;
   tstep = redmem[ii].tstep;
   clicks = redmem[ii].clicks++;
   
   if (thresh == 0)                                                        //  1st click 
   {
      redeye_distr(ii);                                                    //  get pixel redness distribution
      thresh = redmem[ii].thresh;                                          //  initial redness threshhold
      tstep = redmem[ii].tstep;                                            //  redness step size
      Ftoparc = 0;
   }

   tstep = (thresh - tstep) / thresh;                                      //  convert to reduction factor
   thresh = thresh * pow(tstep,clicks);                                    //  reduce threshhold by total clicks

   for (py = cy-rad; py <= cy+rad; py++)                                   //  darken pixels over threshhold
   for (px = cx-rad; px <= cx+rad; px++)
   {
      if (type == 'R') {
         if (px < cx - ww/2) continue;
         if (px > cx + ww/2) continue;
         if (py < cy - hh/2) continue;
         if (py > cy + hh/2) continue;
      }
      rd = sqrt((px-cx)*(px-cx) + (py-cy)*(py-cy));
      if (rd > rad + 0.5) continue;
      ppix = PXMpix(E3pxm16,px,py);                                        //  set redness = threshhold
      if (pixred(ppix) > thresh)
         ppix[0] = int(thresh * (0.65 * ppix[1] + 0.10 * ppix[2] + 1) / (25 - 0.25 * thresh));
   }

   return;
}


//  Build a distribution of redness for a red-eye. Use this information 
//  to set initial threshhold and step size for stepwise darkening.

void redeye_distr(int ii)
{
   int         cx, cy, ww, hh, rad, px, py;
   int         bin, npix, dbins[20], bsum, blim;
   double      rd, maxred, minred, redpart, dbase, dstep;
   char        type;
   uint16      *ppix;
   
   type = redmem[ii].type;
   cx = redmem[ii].cx;
   cy = redmem[ii].cy;
   ww = redmem[ii].ww;
   hh = redmem[ii].hh;
   rad = redmem[ii].rad;
   
   maxred = 0;
   minred = 100;

   for (py = cy-rad; py <= cy+rad; py++)
   for (px = cx-rad; px <= cx+rad; px++)
   {
      if (type == 'R') {
         if (px < cx - ww/2) continue;
         if (px > cx + ww/2) continue;
         if (py < cy - hh/2) continue;
         if (py > cy + hh/2) continue;
      }
      rd = sqrt((px-cx)*(px-cx) + (py-cy)*(py-cy));
      if (rd > rad + 0.5) continue;
      ppix = PXMpix(E3pxm16,px,py);
      redpart = pixred(ppix);
      if (redpart > maxred) maxred = redpart;
      if (redpart < minred) minred = redpart;
   }
   
   dbase = minred;
   dstep = (maxred - minred) / 19.99;

   for (bin = 0; bin < 20; bin++) dbins[bin] = 0;
   npix = 0;

   for (py = cy-rad; py <= cy+rad; py++)
   for (px = cx-rad; px <= cx+rad; px++)
   {
      if (type == 'R') {
         if (px < cx - ww/2) continue;
         if (px > cx + ww/2) continue;
         if (py < cy - hh/2) continue;
         if (py > cy + hh/2) continue;
      }
      rd = sqrt((px-cx)*(px-cx) + (py-cy)*(py-cy));
      if (rd > rad + 0.5) continue;
      ppix = PXMpix(E3pxm16,px,py);
      redpart = pixred(ppix);
      bin = int((redpart - dbase) / dstep);
      ++dbins[bin];
      ++npix;
   }
   
   bsum = 0;
   blim = int(0.5 * npix);

   for (bin = 0; bin < 20; bin++)                                          //  find redness level for 50% of
   {                                                                       //    pixels within red-eye radius
      bsum += dbins[bin];
      if (bsum > blim) break;
   }

   redmem[ii].thresh = dbase + dstep * bin;                                //  initial redness threshhold
   redmem[ii].tstep = dstep;                                               //  redness step (5% of range)

   return;
}


//  find a red-eye (nearly) overlapping the mouse click position

int redeye_find(int cx, int cy)
{
   for (int ii = 0; ii < Nredmem; ii++)
   {
      if (cx > redmem[ii].cx - 2 * redmem[ii].rad && 
          cx < redmem[ii].cx + 2 * redmem[ii].rad &&
          cy > redmem[ii].cy - 2 * redmem[ii].rad && 
          cy < redmem[ii].cy + 2 * redmem[ii].rad) 
            return ii;                                                     //  found
   }
   return -1;                                                              //  not found
}


//  remove a red-eye from memory

void redeye_remove(int ii)
{
   int      cx, cy, rad, px, py;
   uint16   *pix1, *pix3;

   cx = redmem[ii].cx;
   cy = redmem[ii].cy;
   rad = redmem[ii].rad;

   for (px = cx-rad; px <= cx+rad; px++)
   for (py = cy-rad; py <= cy+rad; py++)
   {
      pix1 = PXMpix(E1pxm16,px,py);
      pix3 = PXMpix(E3pxm16,px,py);
      pix3[0] = pix1[0];
      pix3[1] = pix1[1];
      pix3[2] = pix1[2];
   }
   
   for (ii++; ii < Nredmem; ii++) 
      redmem[ii-1] = redmem[ii];
   Nredmem--;
   
   Ftoparc = 0;
   return;
}


//  compute red-eye radius limit: smaller of 100 and nearest image edge

int redeye_radlim(int cx, int cy)
{
   int radlim = 100;
   if (cx < 100) radlim = cx;
   if (E3ww-1 - cx < 100) radlim = E3ww-1 - cx;
   if (cy < 100) radlim = cy;
   if (E3hh-1 - cy < 100) radlim = E3hh-1 - cy;
   return radlim;
}


/**************************************************************************/

//  image blur function                                                    //  radius steps of 0.5   v.9.2

double      blur_radius;
double      blur_weight[101][101];                                         //  up to blur radius = 99   v.9.2
int         blur_Npixels, blur_pixdone;


void m_blur(GtkWidget *, cchar *)
{
   int    blur_dialog_event(zdialog *zd, cchar *event);
   void * blur_thread(void *);

   zfuncs::F1_help_topic = "blur";                                         //  v.10.8

   if (! edit_setup("blur",0,2)) return;                                   //  setup edit: no preview

   zdedit = zdialog_new(ZTX("Set Blur Radius"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"hbox","hb2","dialog",0,"space=10");
   zdialog_add_widget(zdedit,"label","labrad","hb2",Bradius,"space=5");
   zdialog_add_widget(zdedit,"spin","radius","hb2","0|99|0.5|0.5","space=5");
   zdialog_add_widget(zdedit,"button","apply","hb2",Bapply,"space=5");

   zdialog_run(zdedit,blur_dialog_event);                                  //  start dialog
   
   blur_radius = 0.5;
   start_thread(blur_thread,0);                                            //  start working thread
   return;
}


//  dialog event and completion callback function

int blur_dialog_event(zdialog * zd, cchar *event)
{
   if (zd->zstat)
   {
      if (zd->zstat == 1) edit_done();                                     //  done
      else edit_cancel();                                                  //  cancel or destroy
      return 1;
   }
   
   if (strEqu(event,"undo")) edit_undo();                                  //  v.10.2
   if (strEqu(event,"redo")) edit_redo();                                  //  v.10.3
   if (strEqu(event,"blendwidth")) signal_thread();                        //  v.10.3

   if (strEqu(event,"apply")) {
      zdialog_fetch(zd,"radius",blur_radius);                              //  get blur radius
      if (blur_radius == 0) edit_reset();
      else signal_thread();                                                //  trigger working thread
   }

   return 1;
}


//  image blur thread function

void * blur_thread(void *)
{
   void * blur_wthread(void *arg);

   int         dx, dy;
   double      rad, rad2;
   double      m, d, w, sum;
  
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      rad = blur_radius - 0.2;                                             //  v.9.2
      rad2 = rad * rad;

      for (dx = 0; dx <= rad+1; dx++)                                      //  clear weights array
      for (dy = 0; dy <= rad+1; dy++)
         blur_weight[dx][dy] = 0;

      for (dx = -rad-1; dx <= rad+1; dx++)                                 //  blur_weight[dx][dy] = no. of pixels
      for (dy = -rad-1; dy <= rad+1; dy++)                                 //    at distance (dx,dy) from center
         ++blur_weight[abs(dx)][abs(dy)];

      m = sqrt(rad2 + rad2);                                               //  corner pixel distance from center
      sum = 0;

      for (dx = 0; dx <= rad+1; dx++)                                      //  compute weight of pixel
      for (dy = 0; dy <= rad+1; dy++)                                      //    at distance dx, dy
      {
         d = sqrt(dx*dx + dy*dy);
         w = (m + 1.2 - d) / m;                                            //  v.9.2
         w = w * w;
         sum += blur_weight[dx][dy] * w;
         blur_weight[dx][dy] = w;
      }

      for (dx = 0; dx <= rad+1; dx++)                                      //  make weights add up to 1.0
      for (dy = 0; dy <= rad+1; dy++)
         blur_weight[dx][dy] = blur_weight[dx][dy] / sum;
      
      if (sa_active) SB_goal = sa_Npixel;
      else  SB_goal = E3ww * E3hh;
      SB_done = 0;

      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wt(blur_wthread,&wtnx[ii]);
      wait_wts();                                                          //  wait for completion

      SB_goal = SB_done = 0;                                               //  v.9.6
      Fmodified = 1;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}


void * blur_wthread(void *arg)                                             //  worker thread function
{
   void  blur_pixel(int px, int py);

   int      index = *((int *) arg);
   int      px, py;
   
   for (py = index; py < E3hh-1; py += Nwt)                                //  loop all image pixels
   for (px = 0; px < E3ww-1; px++)
   {
      blur_pixel(px,py);
      if (Fkillfunc) break;
   }

   exit_wt();
   return 0;                                                               //  not executed, avoid gcc warning
}


void blur_pixel(int px, int py)
{
   int         ii, dist = 0;
   int         jj, dx, dy, adx, ady, rad;
   double      red, green, blue;
   double      weight1, weight2, f1, f2;
   uint16      *pix1, *pix3, *pixN;

   if (sa_active) {                                                        //  select area active
      ii = py * Fww + px;
      dist = sa_pixisin[ii];                                               //  distance from edge
      if (! dist) return;                                                  //  outside pixel
   }

   pix1 = PXMpix(E1pxm16,px,py);                                           //  source pixel
   pix3 = PXMpix(E3pxm16,px,py);                                           //  target pixel
   
   rad = blur_radius;
   red = green = blue = 0;
   weight2 = 0.0;
   
   if (sa_active)                                                          //  select area active
   {
      for (dy = -rad-1; dy <= rad+1; dy++)                                 //  loop neighbor pixels within radius
      for (dx = -rad-1; dx <= rad+1; dx++)
      {
         if (px+dx < 0 || px+dx > E3ww-1) continue;                        //  omit pixels off edge
         if (py+dy < 0 || py+dy > E3hh-1) continue;
         jj = (py+dy) * E3ww + (px+dx);
         if (! sa_pixisin[jj]) continue;                                   //  omit pixels outside area
         adx = abs(dx);
         ady = abs(dy);
         pixN = pix1 + (dy * E3ww + dx) * 3;
         weight1 = blur_weight[adx][ady];                                  //  weight at distance (dx,dy)
         weight2 += weight1;
         red += pixN[0] * weight1;                                         //  accumulate contributions
         green += pixN[1] * weight1;
         blue += pixN[2] * weight1;
      }
      
      red = red / weight2;                                                 //  weighted average
      green = green / weight2;
      blue = blue / weight2;

      if (dist < sa_blend) {                                               //  select area is active,
         f1 = 1.0 * dist / sa_blend;                                       //    blend changes over sa_blend
         f2 = 1.0 - f1;
         red = f1 * red + f2 * pix1[0];
         green = f1 * green + f2 * pix1[1];
         blue = f1 * blue + f2 * pix1[2];
      }

      pix3[0] = int(red);
      pix3[1] = int(green);
      pix3[2] = int(blue);
   }

   else
   {
      for (dy = -rad-1; dy <= rad+1; dy++)                                 //  loop neighbor pixels within radius
      for (dx = -rad-1; dx <= rad+1; dx++)
      {
         if (px+dx < 0 || px+dx > E3ww-1) continue;                        //  omit pixels off edge
         if (py+dy < 0 || py+dy > E3hh-1) continue;
         adx = abs(dx);
         ady = abs(dy);
         pixN = pix1 + (dy * E3ww + dx) * 3;
         weight1 = blur_weight[adx][ady];                                  //  weight at distance (dx,dy)
         weight2 += weight1;
         red += pixN[0] * weight1;                                         //  accumulate contributions
         green += pixN[1] * weight1;
         blue += pixN[2] * weight1;
      }

      red = red / weight2;                                                 //  weighted average
      green = green / weight2;
      blue = blue / weight2;

      pix3[0] = int(red);
      pix3[1] = int(green);
      pix3[2] = int(blue);
   }
   
   SB_done++;
   return;
}


/**************************************************************************/

//  image sharpening function

int      sharp_ED_cycles;
int      sharp_ED_reduce;
int      sharp_ED_thresh;
int      sharp_UM_radius;
int      sharp_UM_amount;
int      sharp_UM_thresh;
int      sharp_UM_Fcalc;
int      sharp_GR_amount;
int      sharp_GR_thresh;
char     sharp_function[4];


void m_sharpen(GtkWidget *, cchar *)
{
   int    sharp_dialog_event(zdialog *zd, cchar *event);
   void * sharp_thread(void *);

   zfuncs::F1_help_topic = "sharpen";                                      //  v.10.8

   if (! edit_setup("sharp",0,2)) return;                                  //  setup edit: no preview

   zdedit = zdialog_new(ZTX("Sharpen Image"),mWin,Bdone,Bcancel,null); 

   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"vbox","vb11","hb1",0,"space=5");
   zdialog_add_widget(zdedit,"vbox","vb12","hb1",0,"homog|space=5");
   zdialog_add_widget(zdedit,"vbox","vb13","hb1",0,"homog|space=5");
   zdialog_add_widget(zdedit,"button","ED","vb11",ZTX("edge detection"),"space=5");
   zdialog_add_widget(zdedit,"label","lab11","vb12",ZTX("cycles"));
   zdialog_add_widget(zdedit,"label","lab12","vb12",ZTX("reduce"));
   zdialog_add_widget(zdedit,"label","lab13","vb12",Bthresh);
   zdialog_add_widget(zdedit,"spin","cyclesED","vb13","1|30|1|10");
   zdialog_add_widget(zdedit,"spin","reduceED","vb13","50|95|1|80");
   zdialog_add_widget(zdedit,"spin","threshED","vb13","1|99|1|1");

   zdialog_add_widget(zdedit,"hsep","sep2","dialog");
   zdialog_add_widget(zdedit,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"vbox","vb21","hb2",0,"space=5");
   zdialog_add_widget(zdedit,"vbox","vb22","hb2",0,"homog|space=5");
   zdialog_add_widget(zdedit,"vbox","vb23","hb2",0,"homog|space=5");
   zdialog_add_widget(zdedit,"button","UM","vb21",ZTX("unsharp mask"),"space=5");
   zdialog_add_widget(zdedit,"label","lab21","vb22",Bradius);
   zdialog_add_widget(zdedit,"label","lab22","vb22",Bamount);
   zdialog_add_widget(zdedit,"label","lab23","vb22",Bthresh);
   zdialog_add_widget(zdedit,"spin","radiusUM","vb23","1|20|1|2");
   zdialog_add_widget(zdedit,"spin","amountUM","vb23","0|200|1|100");
   zdialog_add_widget(zdedit,"spin","threshUM","vb23","0|100|1|0");

   zdialog_add_widget(zdedit,"hsep","sep3","dialog");
   zdialog_add_widget(zdedit,"hbox","hb3","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"vbox","vb31","hb3",0,"space=5");
   zdialog_add_widget(zdedit,"vbox","vb32","hb3",0,"homog|space=5");
   zdialog_add_widget(zdedit,"vbox","vb33","hb3",0,"homog|space=5");
   zdialog_add_widget(zdedit,"button","GR","vb31",ZTX("brightness gradient"),"space=5");
   zdialog_add_widget(zdedit,"label","lab32","vb32",Bamount);
   zdialog_add_widget(zdedit,"label","lab33","vb32",Bthresh);
   zdialog_add_widget(zdedit,"spin","amountGR","vb33","0|400|1|100");
   zdialog_add_widget(zdedit,"spin","threshGR","vb33","0|100|1|0");

   zdialog_run(zdedit,sharp_dialog_event);                                 //  run dialog, parallel

   *sharp_function = 0;
   sharp_UM_Fcalc = 1;
   start_thread(sharp_thread,0);                                           //  start working thread
   return;
}


//  dialog event and completion callback function

int sharp_dialog_event(zdialog *zd, cchar *event)
{
   if (zd->zstat)
   {
      if (zd->zstat == 1) edit_done();
      else edit_cancel();
      return 0;
   }
   
   if (strEqu(event,"undo")) edit_undo();                                  //  v.10.2
   if (strEqu(event,"redo")) edit_redo();                                  //  v.10.3
   if (strEqu(event,"blendwidth")) signal_thread();                        //  v.10.3
   if (strEqu(event,"radiusUM")) sharp_UM_Fcalc = 1;                       //  must recalculate

   if (strcmpv(event,"ED","UM","GR",null))
   {
      edit_reset();                                                        //  restore original image

      zdialog_fetch(zd,"cyclesED",sharp_ED_cycles);                        //  get all input values
      zdialog_fetch(zd,"reduceED",sharp_ED_reduce);
      zdialog_fetch(zd,"threshED",sharp_ED_thresh);
      zdialog_fetch(zd,"radiusUM",sharp_UM_radius);
      zdialog_fetch(zd,"amountUM",sharp_UM_amount);
      zdialog_fetch(zd,"threshUM",sharp_UM_thresh);
      zdialog_fetch(zd,"amountGR",sharp_GR_amount);
      zdialog_fetch(zd,"threshGR",sharp_GR_thresh);

      strcpy(sharp_function,event);                                        //  pass to working thread
      signal_thread();
   }

   return 0;
}


//  sharpen image thread function

void * sharp_thread(void *)
{
   int    sharp_ED();
   int    sharp_UM();
   int    sharp_GR();
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      if (strEqu(sharp_function,"ED")) sharp_ED();                         //  do requested function
      if (strEqu(sharp_function,"UM")) sharp_UM();
      if (strEqu(sharp_function,"GR")) sharp_GR();

      Fmodified = 1;
      mwpaint2(); 
   }

   return 0;                                                               //  not executed, stop g++ warning
}


//  image sharpen function by edge detection and compression

int sharp_ED()
{
   void  sharp_pixel_ED(int px, int py, int thresh);

   int      sharp_thresh1 = 100;                                           //  initial threshold
   double   sharp_thresh2 = 0.01 * sharp_ED_reduce;                        //  decline rate
   int      px, py, thresh, cycles;
   
   thresh = sharp_thresh1;
    
   if (sa_active) SB_goal = sa_Npixel;                                     //  v.9.6
   else  SB_goal = E3ww * E3hh;
   SB_goal *= sharp_ED_cycles;
   SB_done = 0;

   for (cycles = 0; cycles < sharp_ED_cycles; cycles++)
   {
      if (cycles > 0) thresh = int(thresh * sharp_thresh2);

      for (py = 2; py < E3hh-2; py++)
      for (px = 2; px < E3ww-2; px++)                                      //  loop all pixels
      {
         sharp_pixel_ED(px,py,thresh);
         if (Fkillfunc) break;                                             //  v.9.1
      }

      if (Fkillfunc) break;
   }

   SB_goal = SB_done = 0;                                                  //  v.9.6
   return 1;
}


void sharp_pixel_ED(int px, int py, int thresh)
{
   uint16   *pix1, *pix1u, *pix1d;
   uint16   *pix3, *pix3u, *pix3d, *pix3uu, *pix3dd;
   int      ii, dist = 0;
   int      dd, rgb, pthresh;
   int      dx[4] = { -1, 0, 1, 1 };                                       //  4 directions: NW N NE E
   int      dy[4] = { -1, -1, -1, 0 };
   int      pv2, pv2u, pv2d, pv2uu, pv2dd, pvdiff;
   double   f1, f2;
   
   if (sa_active) {                                                        //  select area active
      ii = py * Fww + px;
      dist = sa_pixisin[ii];                                               //  distance from edge
      if (! dist) return;                                                  //  outside pixel
   }

   pthresh = sharp_ED_thresh;                                              //  pthresh = larger
   if (thresh > pthresh) pthresh = thresh;

   pix1 = PXMpix(E1pxm16,px,py);                                           //  input pixel
   pix3 = PXMpix(E3pxm16,px,py);                                           //  output pixel

   for (dd = 0; dd < 4; dd++)                                              //  4 directions
   {
      pix3u = pix3 + (dy[dd] * E3ww + dx[dd]) * 3;                         //  upstream pixel
      pix3d = pix3 - (dy[dd] * E3ww - dx[dd]) * 3;                         //  downstream pixel

      for (rgb = 0; rgb < 3; rgb++)                                        //  loop 3 RGB colors
      {
         pv2 = pix3[rgb];
         pv2u = pix3u[rgb];                                                //  brightness difference
         pv2d = pix3d[rgb];                                                //    across target pixel

         pvdiff = pv2d - pv2u;
         if (pvdiff < 0) pvdiff = -pvdiff;
         if (pvdiff < 256 * pthresh) continue;                             //  brightness slope < threshold

         if (pv2u < pv2 && pv2 < pv2d)                                     //  slope up, monotone
         {
            pix3uu = pix3u + (dy[dd] * E3ww + dx[dd]) * 3;                 //  upstream of upstream pixel
            pix3dd = pix3d - (dy[dd] * E3ww - dx[dd]) * 3;                 //  downstream of downstream
            pv2uu = pix3uu[rgb];
            pv2dd = pix3dd[rgb];

            if (pv2uu >= pv2u) {                                           //  shift focus of changes to
               pix3u = pix3;                                               //    avoid up/down/up jaggies
               pv2u = pv2;
            }
            
            if (pv2dd <= pv2d) {
               pix3d = pix3;
               pv2d = pv2;
            }
               
            if (pv2u > 256) pv2u -= 256;
            if (pv2d < 65279) pv2d += 256;
         }
         
         else if (pv2u > pv2 && pv2 > pv2d)                                //  slope down, monotone
         {
            pix3uu = pix3u + (dy[dd] * E3ww + dx[dd]) * 3;
            pix3dd = pix3d - (dy[dd] * E3ww - dx[dd]) * 3;
            pv2uu = pix3uu[rgb];
            pv2dd = pix3dd[rgb];

            if (pv2uu <= pv2u) {
               pix3u = pix3;
               pv2u = pv2;
            }
            
            if (pv2dd >= pv2d) {
               pix3d = pix3;
               pv2d = pv2;
            }

            if (pv2d > 256) pv2d -= 256;
            if (pv2u < 65279) pv2u += 256;
         }

         else continue;                                                    //  slope too small

         if (sa_active && dist < sa_blend) {                               //  select area is active,
            f1 = 1.0 * dist / sa_blend;                                    //    blend changes over sa_blend
            f2 = 1.0 - f1;
            pix1u = pix1 + (dy[dd] * E1ww + dx[dd]) * 3;                   //  upstream input pixel
            pix1d = pix1 - (dy[dd] * E1ww - dx[dd]) * 3;                   //  downstream input pixel
            pv2u = int(f1 * pv2u + f2 * pix1u[rgb]);
            pv2d = int(f1 * pv2d + f2 * pix1d[rgb]);
         }

         pix3u[rgb] = pv2u;                                                //  modified brightness values
         pix3d[rgb] = pv2d;                                                //    >> image3 pixel
      }
   }

   SB_done++;                                                              //  v.9.6
   return;
}


//  image sharpen function using unsharp mask

int sharp_UM()
{
   void * sharp_UM_wthread(void *arg);

   if (sharp_UM_Fcalc) {                                                   //  speedup   v.9.6 
      sharp_UM_Fcalc = 0;
      brhood_calc(sharp_UM_radius,'f');
   }   

   if (sa_active) SB_goal = sa_Npixel;                                     //  v.9.6
   else  SB_goal = E3ww * E3hh;
   SB_done = 0;

   for (int ii = 0; ii < Nwt; ii++)                                        //  start worker threads
      start_wt(sharp_UM_wthread,&wtnx[ii]);
   wait_wts();                                                             //  wait for completion

   SB_goal = SB_done = 0;                                                  //  v.9.6
   return 1;
}


void * sharp_UM_wthread(void *arg)                                         //  worker thread function
{
   void  sharp_pixel_UM(int px, int py);

   int      index = *((int *) arg);
   int      px, py;
   
   for (py = index; py < E3hh; py += Nwt)                                  //  loop all image3 pixels
   for (px = 0; px < E3ww; px++)
   {
      sharp_pixel_UM(px,py);
      if (Fkillfunc) break;                                                //  v.9.1
   }

   exit_wt();
   return 0;                                                               //  not executed, avoid gcc warning
}


void sharp_pixel_UM(int px, int py)                                        //  process one pixel
{                                                                          //  revised  v.9.6
   int         ii, dist = 0;
   double      amount, thresh, bright;
   double      mean, incr, ratio, f1, f2;
   int         rgb, cval1, cval3;
   uint16      *pix1, *pix3;
   
   if (sa_active) {                                                        //  select area active
      ii = py * Fww + px;
      dist = sa_pixisin[ii];                                               //  distance from edge
      if (! dist) return;                                                  //  outside pixel
   }

   amount = 0.01 * sharp_UM_amount;                                        //  0.0 to 2.0
   thresh = 100 * sharp_UM_thresh;                                         //  0 to 10K (64K max. possible)

   pix1 = PXMpix(E1pxm16,px,py);                                           //  input pixel
   pix3 = PXMpix(E3pxm16,px,py);                                           //  output pixel

   bright = pixbright(pix1);
   if (bright < 100) return;                                               //  effectively black
   mean = get_brhood(px,py);
   incr = (bright - mean);
   if (fabs(incr) < thresh) return;                                        //  omit low-contrast pixels

   incr = incr * amount;                                                   //  0.0 to 2.0
   if (bright + incr > 65535) incr = 65535 - bright;
   ratio = (bright + incr) / bright;

   for (rgb = 0; rgb < 3; rgb++)                                           //  loop 3 RGB colors
   {
      cval1 = pix1[rgb];
      cval3 = ratio * cval1;
      if (cval3 < 0) cval3 = 0;
      if (cval3 > 65535) cval3 = 65535;
      
      if (sa_active && dist < sa_blend) {                                  //  select area is active,
         f1 = 1.0 * dist / sa_blend;                                       //    blend changes over sa_blend
         f2 = 1.0 - f1;
         cval3 = f1 * cval3 + f2 * cval1;
      }
      
      pix3[rgb] = cval3;
   }

   SB_done++;   
   return;
}


//  sharpen image by increasing brightness gradient                        //  new v.9.8

int sharp_GR()
{
   uint16      *pix1, *pix3;
   int         ii, px, py, dist = 0, rgb;   
   double      amount, thresh;
   double      b1, b1x, b1y, b3x, b3y, b3, bf;
   double      pval, f1, f2;

   amount = 1 + 0.01 * sharp_GR_amount;                                    //  1.0 - 5.0
   thresh = 655.35 * sharp_GR_thresh;                                      //  0 - 64K

   if (sa_active) SB_goal = sa_Npixel;
   else  SB_goal = E3ww * E3hh;
   SB_done = 0;

   for (py = 1; py < E1hh; py++)                                           //  loop all image pixels
   for (px = 1; px < E1ww; px++)
   {
      if (sa_active) {                                                     //  select area active
         ii = py * Fww + px;
         dist = sa_pixisin[ii];                                            //  distance from edge
         if (! dist) continue;                                             //  pixel is outside area
      }

      pix1 = PXMpix(E1pxm16,px,py);                                        //  input pixel
      pix3 = PXMpix(E3pxm16,px,py);                                        //  output pixel

      b1 = pixbright(pix1);                                                //  pixel brightness, 0 - 64K
      b1x = b1 - pixbright(pix1 - 3);                                      //  brightness gradient (x,y)
      b1y = b1 - pixbright(pix1 - 3 * E1ww);
      
      if (abs(b1x + b1y) < thresh)                                         //  moderate brightness change for
         f1 = abs(b1x + b1y) / thresh;                                     //    pixels below threshold gradient
      else  f1 = 1.0;                                                      //  v.10.9
      f2 = 1.0 - f1;

      b1x = b1x * amount;                                                  //  amplified gradient
      b1y = b1y * amount;

      b3x = pixbright(pix1 - 3) + b1x;                                     //  + prior pixel brightness
      b3y = pixbright(pix1 - 3 * E3ww) + b1y;                              //  = new brightness
      b3 = 0.5 * (b3x + b3y);

      b3 = f1 * b3 + f2 * b1;                                              //  possibly moderated     v.10.9

      bf = b3 / b1;                                                        //  ratio of brightness change
      if (bf < 0) bf = 0;
      if (bf > 4) bf = 4;
      
      for (rgb = 0; rgb < 3; rgb++)                                        //  loop 3 RGB colors
      {
         pval = bf * pix3[rgb];                                            //  apply factor
         if (pval > 65535) pval = 65535;
         
         if (sa_active && dist < sa_blend) {                               //  select area is active,
            f1 = 1.0 * dist / sa_blend;                                    //    blend changes over sa_blend
            f2 = 1.0 - f1;
            pval = int(f1 * pval + f2 * pix1[rgb]);
         }
         
         pix3[rgb] = pval;
      }

      SB_done++;
   }

   SB_goal = SB_done = 0;
   return 1;
}


/**************************************************************************/

//  image noise reduction

int      denoise_method = 5;                                               //  default algorithm
int      denoise_radius = 4;


void m_denoise(GtkWidget *, cchar *)
{
   int    denoise_dialog_event(zdialog *zd, cchar *event);                 //  dialog event function
   void * denoise_thread(void *);

   cchar  *denoise_message = ZTX(" Press the reduce button to \n"
                                 " reduce noise in small steps. \n"
                                 " Use undo to start over.");

   zfuncs::F1_help_topic = "reduce_noise";                                 //  v.10.8

   if (! edit_setup("noise",0,2)) return;                                  //  setup edit: no preview

   zdedit = zdialog_new(ZTX("Noise Reduction"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"label","lab1","dialog",denoise_message,"space=5");
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"label","labalg","hb1",ZTX("algorithm"),"space=5");
   zdialog_add_widget(zdedit,"combo","method","hb1",0,"space=5|expand");
   zdialog_add_widget(zdedit,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"label","labrad","hb2",Bradius,"space=5");
   zdialog_add_widget(zdedit,"spin","radius","hb2","1|9|1|4","space=5");
   zdialog_add_widget(zdedit,"button","reduce","hb2",Breduce,"space=5");
   
   zdialog_cb_app(zdedit,"method",ZTX("flatten outliers by color (1)"));
   zdialog_cb_app(zdedit,"method",ZTX("flatten outliers by color (2)"));
   zdialog_cb_app(zdedit,"method",ZTX("set median brightness by color"));
   zdialog_cb_app(zdedit,"method",ZTX("top hat filter by color"));
   zdialog_stuff(zdedit,"method",ZTX("top hat filter by color"));          //  default

   zdialog_run(zdedit,denoise_dialog_event);                               //  run dialog
   start_thread(denoise_thread,0);                                         //  start working thread
   return;
}


//  dialog event and completion callback function

int denoise_dialog_event(zdialog * zd, cchar *event)
{
   char     method[40];

   if (zd->zstat)
   {
      if (zd->zstat == 1) edit_done();
      else edit_cancel();
      return 0;
   }

   if (strEqu(event,"undo")) edit_undo();                                  //  v.10.2
   if (strEqu(event,"redo")) edit_redo();                                  //  v.10.3
   if (strEqu(event,"blendwidth")) signal_thread();                        //  trigger update thread

   if (strEqu(event,"radius")) 
      zdialog_fetch(zd,"radius",denoise_radius);

   if (strEqu(event,"method")) 
   {
      zdialog_fetch(zd,"method",method,39);

      if (strEqu(method,"flatten outliers by color (1)")) {
         denoise_method = 1;
         denoise_radius = 1;
      }

      if (strEqu(method,"flatten outliers by color (2)")) {
         denoise_method = 2;
         denoise_radius = 3;
      }

      if (strEqu(method,"set median brightness by color")) {
         denoise_method = 4;
         denoise_radius = 2;
      }

      if (strEqu(method,"top hat filter by color")) {
         denoise_method = 5;
         denoise_radius = 4;
      }
      
      zdialog_stuff(zd,"radius",denoise_radius);
   }
   
   if (strEqu(event,"reduce")) signal_thread();                            //  trigger update thread

   return 1;
}


//  image noise reduction thread

void * denoise_thread(void *)
{
   void * denoise_wthread(void *arg);

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      E9pxm16 = PXM_copy(E3pxm16);                                         //  image3 is reference source
                                                                           //  image9 will be modified
      if (sa_active) SB_goal = sa_Npixel;
      else  SB_goal = E3ww * E3hh;
      SB_done = 0;

      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wt(denoise_wthread,&wtnx[ii]);
      wait_wts();                                                          //  wait for completion

      SB_goal = SB_done = 0;                                               //  v.9.6

      mutex_lock(&pixmaps_lock);
      PXM_free(E3pxm16);                                                   //  image9 >> image3
      E3pxm16 = E9pxm16;
      E9pxm16 = 0;
      mutex_unlock(&pixmaps_lock);

      Fmodified = 1;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}


void * denoise_wthread(void *arg)                                          //  worker thread function
{
   void  denoise_func1(uint16 *pix3, uint16 *pix9);
   void  denoise_func2(uint16 *pix3, uint16 *pix9);
   void  denoise_func4(uint16 *pix3, uint16 *pix9);
   void  denoise_func5(uint16 *pix3, uint16 *pix9);
   
   int         index = *((int *) arg);
   int         ii, px, py, rad, dist = 0;
   double      f1, f2;
   uint16      *pix1, *pix3, *pix9;

   rad = denoise_radius;

   for (py = index+rad; py < E3hh-rad; py += Nwt)                          //  loop all image3 pixels
   for (px = rad; px < E3ww-rad; px++)
   {
      if (sa_active) {                                                     //  select area active
         ii = py * Fww + px;
         dist = sa_pixisin[ii];                                            //  distance from edge
         if (! dist) continue;                                             //  outside pixel
      }

      pix3 = PXMpix(E3pxm16,px,py);                                        //  source pixel
      pix9 = PXMpix(E9pxm16,px,py);                                        //  target pixel

      if (denoise_method == 1) denoise_func1(pix3,pix9);
      if (denoise_method == 2) denoise_func2(pix3,pix9);
      if (denoise_method == 4) denoise_func4(pix3,pix9);
      if (denoise_method == 5) denoise_func5(pix3,pix9);

      if (sa_active && dist < sa_blend) {                                  //  select area is active,
         f1 = 1.0 * dist / sa_blend;                                       //    blend changes over sa_blend
         f2 = 1.0 - f1;
         pix1 = PXMpix(E1pxm16,px,py);                                     //  source pixel
         pix9[0] = int(f1 * pix9[0] + f2 * pix1[0]);
         pix9[1] = int(f1 * pix9[1] + f2 * pix1[1]);
         pix9[2] = int(f1 * pix9[2] + f2 * pix1[2]);
      }

      SB_done++;
      if (Fkillfunc) break;                                                //  v.9.1
   }

   exit_wt();
   return 0;                                                               //  not executed, avoid gcc warning
}


//  flatten outliers within radius, by color 
//  an outlier is the max or min value within a radius

void denoise_func1(uint16 *pix3, uint16 *pix9)
{
   int         dy, dx, rad;
   int         min0, min1, min2, max0, max1, max2;
   uint16      *pixN;

   min0 = min1 = min2 = 65535;
   max0 = max1 = max2 = 0;
   rad = denoise_radius;

   for (dy = -rad; dy <= rad; dy++)                                        //  loop surrounding pixels
   for (dx = -rad; dx <= rad; dx++)
   {
      if (dy == 0 && dx == 0) continue;                                    //  skip self

      pixN = pix3 + (dy * E3ww + dx) * 3;
      if (pixN[0] < min0) min0 = pixN[0];                                  //  find min and max per color
      if (pixN[0] > max0) max0 = pixN[0];
      if (pixN[1] < min1) min1 = pixN[1];
      if (pixN[1] > max1) max1 = pixN[1];
      if (pixN[2] < min2) min2 = pixN[2];
      if (pixN[2] > max2) max2 = pixN[2];
   }
   
   if (pix3[0] <= min0 && min0 < 65279) pix9[0] = min0 + 256;              //  if outlier, flatten a little
   if (pix3[0] >= max0 && max0 > 256) pix9[0] = max0 - 256;
   if (pix3[1] <= min1 && min1 < 65279) pix9[1] = min1 + 256;
   if (pix3[1] >= max1 && max1 > 256) pix9[1] = max1 - 256;
   if (pix3[2] <= min2 && min2 < 65279) pix9[2] = min2 + 256;
   if (pix3[2] >= max2 && max2 > 256) pix9[2] = max2 - 256;
   
   return;
}


//  flatten outliers
//  An outlier pixel has an RGB value outside one sigma of 
//  the mean for all pixels within a given radius of the pixel.

void denoise_func2(uint16 *pix3, uint16 *pix9)
{
   int         rgb, dy, dx, rad, nn;
   double      nn1, val, sum, sum2, mean, variance, sigma;
   uint16      *pixN;

   rad = denoise_radius;
   nn = (rad * 2 + 1);
   nn = nn * nn - 1;
   nn1 = 1.0 / nn;

   for (rgb = 0; rgb < 3; rgb++)                                           //  loop RGB color
   {
      sum = sum2 = 0;

      for (dy = -rad; dy <= rad; dy++)                                     //  loop surrounding pixels
      for (dx = -rad; dx <= rad; dx++)
      {
         if (dy == 0 && dx == 0) continue;                                 //  skip self
         pixN = pix3 + (dy * E3ww + dx) * 3;
         val = pixN[rgb];
         sum += val;
         sum2 += val * val;
      }
      
      mean = nn1 * sum;
      variance = nn1 * (sum2 - 2.0 * mean * sum) + mean * mean;
      sigma = sqrt(variance);

      val = pix3[rgb];      
      if (val > mean + sigma) {                                            //  move value to mean +/- sigma
         val = mean + sigma;
         pix9[rgb] = val;
      }
      else if (val < mean - sigma) {
         val = mean - sigma;
         pix9[rgb] = val;
      }
   }
   
   return;
}


//  use median brightness for pixels within radius

void denoise_func4(uint16 *pix3, uint16 *pix9)
{
   int         dy, dx, rad;
   int         ns, rgb, bsortN[400];
   uint16      *pixN;

   rad = denoise_radius;

   for (rgb = 0; rgb < 3; rgb++)                                           //  loop all RGB colors
   {
      ns = 0;

      for (dy = -rad; dy <= rad; dy++)                                     //  loop surrounding pixels
      for (dx = -rad; dx <= rad; dx++)                                     //  get brightness values
      {
         pixN = pix3 + (dy * E3ww + dx) * 3;
         bsortN[ns] = pixN[rgb];
         ns++;
      }

      HeapSort(bsortN,ns);
      pix9[rgb] = bsortN[ns/2];                                            //  median brightness of ns pixels
   }

   return;
}


//  modified top hat filter: execute with increasing radius from 1 to limit
//  detect outlier by comparing with pixels in outer radius

void denoise_func5(uint16 *pix3, uint16 *pix9)
{
   int         dy, dx, rad;
   int         min0, min1, min2, max0, max1, max2;
   uint16      *pixN;

   for (rad = 1; rad <= denoise_radius; rad++)
   for (int loops = 0; loops < 2; loops++)
   {
      min0 = min1 = min2 = 65535;
      max0 = max1 = max2 = 0;

      for (dy = -rad; dy <= rad; dy++)                                     //  loop all pixels within rad
      for (dx = -rad; dx <= rad; dx++)
      {
         if (dx > -rad && dx < rad) continue;                              //  skip inner pixels
         if (dy > -rad && dy < rad) continue;

         pixN = pix3 + (dy * E3ww + dx) * 3;
         if (pixN[0] < min0) min0 = pixN[0];                               //  find min and max per color
         if (pixN[0] > max0) max0 = pixN[0];                               //    among outermost pixels
         if (pixN[1] < min1) min1 = pixN[1];
         if (pixN[1] > max1) max1 = pixN[1];
         if (pixN[2] < min2) min2 = pixN[2];
         if (pixN[2] > max2) max2 = pixN[2];
      }
      
      if (pix3[0] < min0 && pix9[0] < 65279) pix9[0] += 256;               //  if central pixel is outlier,
      if (pix3[0] > max0 && pix9[0] > 256) pix9[0] -= 256;                 //    moderate its values
      if (pix3[1] < min1 && pix9[1] < 65279) pix9[1] += 256;
      if (pix3[1] > max1 && pix9[1] > 256) pix9[1] -= 256;
      if (pix3[2] < min2 && pix9[2] < 65279) pix9[2] += 256;
      if (pix3[2] > max2 && pix9[2] > 256) pix9[2] -= 256;
   }

   return;
}


/**************************************************************************/

//  trim image - use mouse to select image region to retain

int         trimsize[2] = { 1600, 1000 };                                  //  trim (crop) width, height
char        *trimbuttons[6] = { 0,0,0,0,0,0 };                             //  trim dialog button labels
char        *trimratios[6];                                                //  corresponding aspect ratios
int         trimx1, trimy1, trimx2, trimy2;                                //  trim rectangle
double      trimR;                                                         //  trim aspect ratio
int         trim_status = 0;


void m_trim(GtkWidget *, cchar *)
{
   void   trim_dialog();
   void * trim_thread(void *);

   zfuncs::F1_help_topic = "trim_image";                                   //  v.10.8

   if (! edit_setup("trim",0,0)) return;                                   //  setup edit
   
   if (! trimbuttons[0] || strEqu(trimbuttons[0],"undefined")) {
      trimbuttons[0] = strdupz("1:1",8);                                   //  default trim buttons   v.10.10.2
      trimbuttons[1] = strdupz("2:1",8);
      trimbuttons[2] = strdupz("3:2",8);
      trimbuttons[3] = strdupz("4:3",8);
      trimbuttons[4] = strdupz("16:9",8);
      trimbuttons[5] = strdupz("gold",8);

      trimratios[0] = strdupz("1:1",8);                                    //  default trim ratios    v.10.10.2
      trimratios[1] = strdupz("2:1",8);
      trimratios[2] = strdupz("3:2",8);
      trimratios[3] = strdupz("4:3",8);
      trimratios[4] = strdupz("16:9",8);
      trimratios[5] = strdupz("1.618:1",8);                                //  fix inverted ratio     v.10.10.3
   }

   if (trimsize[0] < 0.9 * E3ww) {                                         //  start with prior trim size used,
      trimx1 = 0.5 * (E3ww - trimsize[0]);                                 //    unless larger than 90% image size
      trimx2 = trimx1 + trimsize[0];                                       //  v.10.10.2
   }
   else {
      trimx1 = 0.05 * E3ww;
      trimx2 = 0.95 * E3ww;
   }
   
   if (trimsize[1] < 0.9 * E3hh) {
      trimy1 = 0.5 * (E3hh - trimsize[1]);
      trimy2 = trimy1 + trimsize[1];
   }
   else {
      trimy1 = 0.05 * E3hh;
      trimy2 = 0.95 * E3hh;
   }
   
   trimsize[0] = (trimx2 - trimx1);
   trimsize[1] = (trimy2 - trimy1);
   trimR = 1.0 * trimsize[0] / trimsize[1];
   
   trim_dialog();                                                          //  start trim dialog
   
   trim_status = 0;
   start_thread(trim_thread,0);                                            //  start working thread
   signal_thread();
   return;
}


//  start trim dialog

void trim_dialog()
{
   void   trim_mousefunc();
   int    trim_dialog_event(zdialog *zd, cchar *event);

   cchar       *trim_message = ZTX("Drag middle to move, drag corners to resize.");
   char        text[40];
   int         ii;

   zdedit = zdialog_new(ZTX("Trim Image"),mWin,ZTX("customize"),Bapply,Bcancel,null);
   zdialog_add_widget(zdedit,"label","lab1","dialog",trim_message,"space=10");
   zdialog_add_widget(zdedit,"hbox","hb1","dialog");
   zdialog_add_widget(zdedit,"label","space","hb1",0,"space=4");
   zdialog_add_widget(zdedit,"label","labwhr","hb1","2345 x 1234  (R=1.90)");
   zdialog_add_widget(zdedit,"check","lock","hb1",ZTX("Lock Ratio"),"space=10");
   zdialog_add_widget(zdedit,"hbox","hb2","dialog",0,"space=5");

   for (ii = 0; ii < 6; ii++)                                              //  6 custom buttons    v.10.10.2
      zdialog_add_widget(zdedit,"button",trimbuttons[ii],"hb2",trimbuttons[ii]);

   zdialog_add_widget(zdedit,"button","invert","hb2",ZTX("invert"));       //  [invert] button

   snprintf(text,39,"%d x %d  (R=%.2f)",trimsize[0],trimsize[1],trimR);    //  stuff starting trim size >> dialog
   zdialog_stuff(zdedit,"labwhr",text);

   zdialog_run(zdedit,trim_dialog_event);                                  //  run dialog, parallel

   mouseCBfunc = trim_mousefunc;                                           //  connect mouse function
   Mcapture++;
   
   return;
}


//  dialog event and completion callback function

int trim_dialog_event(zdialog *zd, cchar *event)
{
   void  trim_customize();

   double   r1, r2, ratio = 0;
   char     text[40];
   cchar    *pp;
   int      ii;
   
   if (zd->zstat)                                                          //  dialog complete
   {
      mouseCBfunc = 0;                                                     //  disconnect mouse
      Mcapture = 0;
      
      if (zd->zstat == 1) {                                                //  customize buttons
         zdialog_free(zdedit);                                             //  kill trim dialog
         trim_customize();                                                 //  do customize dialog
         trim_dialog();                                                    //  start trim dialog
         return 0;
      }

      if (zd->zstat != 2) {                                                //  cancel
         edit_cancel();
         return 0;
      }

      trimsize[0] = trimx2 - trimx1;                                       //  apply
      trimsize[1] = trimy2 - trimy1;

      Fmodified = 1;
      trim_status = 1;
      signal_thread();                                                     //  v.10.10.2
      edit_done();
      return 0;
   }

   for (ii = 0; ii < 6; ii++)                                              //  trim ratio buttons
      if (strEqu(event,trimbuttons[ii])) break;
   if (ii < 6) {
      r1 = r2 = ratio = 0;
      pp = strField(trimratios[ii],':',1);
      if (pp) r1 = atof(pp);
      pp = strField(trimratios[ii],':',2);
      if (pp) r2 = atof(pp);
      if (r1 > 0 && r2 > 0) ratio = r1/r2;
      if (ratio < 0.1 || ratio > 10) ratio = 1.0;
      if (! ratio) return 0;
      zdialog_stuff(zd,"lock",1);                                          //  assume lock is wanted
      trimR = ratio;
   }
   
   if (strEqu(event,"invert"))                                             //  invert ratio button
      if (trimR) ratio = 1.0 / trimR;
   
   if (! ratio) return 0;
   trimR = ratio;                                                          //  bugfix   v.10.10.3

   if (trimx2 - trimx1 > trimy2 - trimy1)
      trimy2 = trimy1 + (trimx2 - trimx1) / trimR;                         //  adjust smaller dimension
   else 
      trimx2 = trimx1 + (trimy2 - trimy1) * trimR;

   if (trimy2 >= E3hh) {                                                   //  if off the bottom edge,
      trimy2 = E3hh-1;                                                     //  adjust width
      trimx2 = trimx1 + (trimy2 - trimy1) * trimR;
   }

   if (trimx2 >= E3ww) {                                                   //  if off the right edge,
      trimx2 = E3ww-1;                                                     //  adjust height
      trimy2 = trimy1 + (trimx2 - trimx1) / trimR;
   }

   trimsize[0] = trimx2 - trimx1;                                          //  new rectangle dimensions
   trimsize[1] = trimy2 - trimy1;

   snprintf(text,39,"%d x %d  (R=%.2f)",trimsize[0],trimsize[1],trimR);    //  stuff trim size >> dialog
   zdialog_stuff(zdedit,"labwhr",text);

   signal_thread();
   return 0;
}


//  dialog to get custom trim button names and corresponding ratios

void trim_customize()
{
   char        text[40], blab[8], rlab[8];
   double      r1, r2, ratio;
   cchar       *pp;
   int         ii, zstat;
   
   zdedit = zdialog_new(ZTX("Trim Buttons"),mWin,Bapply,Bcancel,null);     //  start dialog
   zdialog_add_widget(zdedit,"hbox","hbb","dialog",0,"homog|space=5");
   zdialog_add_widget(zdedit,"hbox","hbr","dialog",0,"homog|space=5");
   
   strcpy(blab,"butt-0");
   strcpy(rlab,"ratio-0");
   
   for (ii = 0; ii < 6; ii++) 
   {
      blab[5] = '0' + ii;
      rlab[6] = '0' + ii;
      zdialog_add_widget(zdedit,"entry",blab,"hbb",trimbuttons[ii],"scc=6");
      zdialog_add_widget(zdedit,"entry",rlab,"hbr",trimratios[ii],"scc=6");
   }
   
   zdialog_run(zdedit);
   zstat = zdialog_wait(zdedit);

   if (zstat == 1)                                                         //  apply
   {
      for (ii = 0; ii < 6; ii++)                                           //  get custom button names
      {
         blab[5] = '0' + ii;
         zdialog_fetch(zdedit,blab,text,12);
         strTrim2(text);
         if (! *text) continue;
         zfree(trimbuttons[ii]);
         trimbuttons[ii] = strdupz(text);
      }

      for (ii = 0; ii < 6; ii++)                                           //  get custom ratios
      {
         rlab[6] = '0' + ii;
         zdialog_fetch(zdedit,rlab,text,12);
         strTrim2(text);
         r1 = r2 = ratio = 0;
         pp = strField(text,':',1);
         if (! pp) continue;
         r1 = atof(pp);
         pp = strField(text,':',2);
         if (! pp) continue;
         r2 = atof(pp);
         if (r1 > 0 && r2 > 0) ratio = r1/r2;
         if (ratio < 0.1 || ratio > 10) continue;
         zfree(trimratios[ii]);
         trimratios[ii] = strdupz(text);
      }
   }

   zdialog_free(zdedit);                                                   //  kill dialog
   return;
}


//  trim mouse function

void trim_mousefunc()
{
   int         mpx, mpy, xdrag, ydrag, rlock;
   int         corner, chop, moveall = 0;
   int         dx, dy, dd, d1, d2, d3, d4;
   char        text[40];
   double      drr;
   
   if (LMclick || Mxdrag || Mydrag)                                        //  mouse click or drag
   {
      if (LMclick) {
         mpx = Mxclick;                                                    //  click
         mpy = Myclick;
         xdrag = ydrag = 0;
         LMclick = 0;
      }
      else {
         mpx = Mxdrag;                                                     //  drag
         mpy = Mydrag;
         xdrag = Mxdrag - Mxdown;
         ydrag = Mydrag - Mydown;
         Mxdown = Mxdrag;                                                  //  reset drag origin
         Mydown = Mydrag;
      }
      
      if (Mxdrag || Mydrag) {
         moveall = 1;
         dd = 0.1 * (trimx2 - trimx1);                                     //  test if mouse is in the broad
         if (mpx < trimx1 + dd) moveall = 0;                               //    middle of the rectangle
         if (mpx > trimx2 - dd) moveall = 0;
         dd = 0.1 * (trimy2 - trimy1);
         if (mpy < trimy1 + dd) moveall = 0;
         if (mpy > trimy2 - dd) moveall = 0;
      }

      if (moveall) {                                                       //  yes, move the whole rectangle
         trimx1 += xdrag;
         trimx2 += xdrag;
         trimy1 += ydrag;
         trimy2 += ydrag;
         corner = 0;
      }

      else {                                                               //  no, find closest corner
         dx = mpx - trimx1;
         dy = mpy - trimy1;
         d1 = sqrt(dx*dx + dy*dy);
         
         dx = mpx - trimx2;
         dy = mpy - trimy1;
         d2 = sqrt(dx*dx + dy*dy);
         
         dx = mpx - trimx2;
         dy = mpy - trimy2;
         d3 = sqrt(dx*dx + dy*dy);
         
         dx = mpx - trimx1;
         dy = mpy - trimy2;
         d4 = sqrt(dx*dx + dy*dy);
         
         corner = 1;                                                       //  NW
         dd = d1;
         if (d2 < dd) { corner = 2; dd = d2; }                             //  NE
         if (d3 < dd) { corner = 3; dd = d3; }                             //  SE
         if (d4 < dd) { corner = 4; dd = d4; }                             //  SW
         
         if (corner == 1) { trimx1 = mpx; trimy1 = mpy; }                  //  move this corner to mouse
         if (corner == 2) { trimx2 = mpx; trimy1 = mpy; }
         if (corner == 3) { trimx2 = mpx; trimy2 = mpy; }
         if (corner == 4) { trimx1 = mpx; trimy2 = mpy; }
      }

      if (trimx1 > trimx2-10) trimx1 = trimx2-10;                          //  sanity limits
      if (trimy1 > trimy2-10) trimy1 = trimy2-10;

      zdialog_fetch(zdedit,"lock",rlock);                                  //  w/h ratio locked
      if (rlock && corner) {
         if (corner < 3)                                                   //  bugfix   v.10.3.1
            trimy2 = trimy1 + 1.0 * (trimx2 - trimx1) / trimR;
         else
            trimy1 = trimy2 - 1.0 * (trimx2 - trimx1) / trimR;
      }

      chop = 0;
      if (trimx1 < 0) { trimx1 = 0; chop = 1; }                            //  look for off the edge   v.10.1
      if (trimx2 > E3ww) { trimx2 = E3ww; chop = 2; }                      //  bugfix, after corner move    v.10.3.1
      if (trimy1 < 0) { trimy1 = 0; chop = 3; }
      if (trimy2 > E3hh) { trimy2 = E3hh; chop = 4; }

      if (rlock && chop) {                                                 //  keep ratio if off the edge   v.10.1
         if (chop < 3)
            trimy2 = trimy1 + 1.0 * (trimx2 - trimx1) / trimR;
         else
            trimx2 = trimx1 + 1.0 * (trimy2 - trimy1) * trimR;
      }

      if (trimx1 > trimx2-10) trimx1 = trimx2-10;                          //  sanity limits
      if (trimy1 > trimy2-10) trimy1 = trimy2-10;

      if (trimx1 < 0) trimx1 = 0;
      if (trimy1 < 0) trimy1 = 0;
      if (trimx2 > E3ww) trimx2 = E3ww;
      if (trimy2 > E3hh) trimy2 = E3hh;

      trimsize[0] = trimx2 - trimx1;                                       //  new rectangle dimensions
      trimsize[1] = trimy2 - trimy1;

      drr = 1.0 * trimsize[0] / trimsize[1];                               //  new w/h ratio
      if (! rlock) trimR = drr;

      snprintf(text,39,"%d x %d  (R=%.2f)",trimsize[0],trimsize[1],trimR); //  stuff starting trim size >> dialog
      zdialog_stuff(zdedit,"labwhr",text);
      
      signal_thread();                                                     //  trigger update thread
   }

   return;
}


//  trim thread function

void * trim_thread(void *)
{
   int      px1, py1, px2, py2;
   uint16   *pix1, *pix3;
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      if (trim_status == 0)                                                //  darken margins
      {
         for (py1 = 0; py1 < E3hh; py1++)                                  //  copy pixels E1 >> E3
         for (px1 = 0; px1 < E3ww; px1++)
         {
            pix1 = PXMpix(E1pxm16,px1,py1);
            pix3 = PXMpix(E3pxm16,px1,py1);

            if (px1 < trimx1 || px1 > trimx2 || py1 < trimy1 || py1 > trimy2)
            {
               pix3[0] = pix1[0] / 2;
               pix3[1] = pix1[1] / 2;
               pix3[2] = pix1[2] / 2;
            }

            else
            {
               pix3[0] = pix1[0];
               pix3[1] = pix1[1];
               pix3[2] = pix1[2];
            }
         }
         
         mwpaint2();                                                       //  update window
      }
      
      if (trim_status == 1)                                                //  do the trim
      {
         mutex_lock(&pixmaps_lock);
         PXM_free(E3pxm16);
         E3pxm16 = PXM_make(trimsize[0],trimsize[1],16);                   //  new pixmap with requested size
         E3ww = trimsize[0];
         E3hh = trimsize[1];
         
         for (py1 = trimy1; py1 < trimy2; py1++)                           //  copy pixels
         for (px1 = trimx1; px1 < trimx2; px1++)
         {
            px2 = px1 - trimx1;
            py2 = py1 - trimy1;
            pix1 = PXMpix(E1pxm16,px1,py1);
            pix3 = PXMpix(E3pxm16,px2,py2);
            pix3[0] = pix1[0];
            pix3[1] = pix1[1];
            pix3[2] = pix1[2];
         }

         Fmodified = 1;
         mutex_unlock(&pixmaps_lock);
         mwpaint2();                                                       //  update window
      }
   }

   return 0;                                                               //  not executed, stop g++ warning
}


/**************************************************************************/

//  Resize (rescale) image
//
//  Output pixels are composites of input pixels, e.g. 2/3 size means 
//  that 3x3 input pixels are mapped into 2x2 output pixels, and an 
//  image size of 1000 x 600 becomes 667 x 400.

int      editresize[2] = { 1600, 1200 };                                   //  edit resize width, height
int      resize_ww0, resize_hh0;                                           //  original size
int      resize_ww1, resize_hh1;                                           //  new size


void m_resize(GtkWidget *, cchar *)
{
   int      resize_dialog_event(zdialog *zd, cchar *event);

   cchar  *lockmess = ZTX("Lock aspect ratio");

   zfuncs::F1_help_topic = "resize";                                       //  v.10.8

   if (! edit_setup("resize",0,0)) return;                                 //  setup edit: no preview

   zdedit = zdialog_new(ZTX("Resize Image"),mWin,Bapply,Bcancel,null);
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"space=10");
   zdialog_add_widget(zdedit,"vbox","vb11","hb1",0,"homog|space=5");
   zdialog_add_widget(zdedit,"vbox","vb12","hb1",0,"homog|space=5");
   zdialog_add_widget(zdedit,"vbox","vb13","hb1",0,"homog|space=5");
   zdialog_add_widget(zdedit,"label","placeholder","vb11",0);              //            pixels     percent
   zdialog_add_widget(zdedit,"label","labw","vb11",Bwidth);                //    width   [______]   [______]
   zdialog_add_widget(zdedit,"label","labh","vb11",Bheight);               //    height  [______]   [______]
   zdialog_add_widget(zdedit,"label","labpix","vb12","pixels");            //
   zdialog_add_widget(zdedit,"spin","wpix","vb12","20|9999|1|0");          //    presets [2/3] [1/2] [1/3] [1/4] [Prev]
   zdialog_add_widget(zdedit,"spin","hpix","vb12","20|9999|1|0");          //
   zdialog_add_widget(zdedit,"label","labpct","vb13",Bpercent);            //    [_] lock width/height ratio
   zdialog_add_widget(zdedit,"spin","wpct","vb13","1|500|0.1|100");        //
   zdialog_add_widget(zdedit,"spin","hpct","vb13","1|500|0.1|100");        //       [  done  ]  [ cancel ]
   zdialog_add_widget(zdedit,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"label","preset","hb2",Bpresets,"space=5");
   zdialog_add_widget(zdedit,"button","b 3/4","hb2","3/4");
   zdialog_add_widget(zdedit,"button","b 2/3","hb2","2/3");
   zdialog_add_widget(zdedit,"button","b 1/2","hb2","1/2");
   zdialog_add_widget(zdedit,"button","b 1/3","hb2","1/3");
   zdialog_add_widget(zdedit,"button","b 1/4","hb2","1/4");
   zdialog_add_widget(zdedit,"button","prev","hb2",ZTX("Prev"));
   zdialog_add_widget(zdedit,"check","lock","dialog",lockmess);

   resize_ww0 = Fpxm16->ww;                                                //  original width, height
   resize_hh0 = Fpxm16->hh;

   zdialog_stuff(zdedit,"wpix",resize_ww0);
   zdialog_stuff(zdedit,"hpix",resize_hh0);
   zdialog_stuff(zdedit,"lock",1);
   
   zdialog_run(zdedit,resize_dialog_event);                                //  run dialog
   zdialog_wait(zdedit);
   return;
}


//  dialog event and completion callback function

int resize_dialog_event(zdialog *zd, cchar * event)
{
   int         lock;
   double      wpct1, hpct1;
   
   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat != 1) {                                                //  user cancel
         edit_cancel();
         return 0;
      }

      if (resize_ww1 != resize_ww0 || resize_hh1 != resize_hh0)            //  thread removed   v.10.8
      {                                                                    //  rescale to target size
         mutex_lock(&pixmaps_lock);
         PXM_free(E3pxm16);
         E3pxm16 = PXM_rescale(Fpxm16,resize_ww1,resize_hh1);
         E3ww = resize_ww1;
         E3hh = resize_hh1;
         Fmodified = 1;
         mutex_unlock(&pixmaps_lock);
         editresize[0] = resize_ww1;                                       //  remember size used     v.10.10
         editresize[1] = resize_hh1;
      }

      edit_done();
      return 0;
   }

   zdialog_fetch(zd,"wpix",resize_ww1);                                    //  get all widget values
   zdialog_fetch(zd,"hpix",resize_hh1);
   zdialog_fetch(zd,"wpct",wpct1);
   zdialog_fetch(zd,"hpct",hpct1);
   zdialog_fetch(zd,"lock",lock);
   
   if (strEqu(event,"b 3/4")) {
      resize_ww1 = (3 * resize_ww0 + 3) / 4;
      resize_hh1 = (3 * resize_hh0 + 3) / 4;
   }
   
   if (strEqu(event,"b 2/3")) {
      resize_ww1 = (2 * resize_ww0 + 2) / 3;
      resize_hh1 = (2 * resize_hh0 + 2) / 3;
   }
   
   if (strEqu(event,"b 1/2")) {
      resize_ww1 = (resize_ww0 + 1) / 2;
      resize_hh1 = (resize_hh0 + 1) / 2;
   }
   
   if (strEqu(event,"b 1/3")) {
      resize_ww1 = (resize_ww0 + 2) / 3;
      resize_hh1 = (resize_hh0 + 2) / 3;
   }
   
   if (strEqu(event,"b 1/4")) {
      resize_ww1 = (resize_ww0 + 3) / 4;
      resize_hh1 = (resize_hh0 + 3) / 4;
   }
   
   if (strEqu(event,"prev")) {                                             //  use previous resize values   v.10.10
      resize_ww1 = editresize[0];
      resize_hh1 = editresize[1];
   }

   if (strEqu(event,"wpct"))                                               //  width % - set pixel width
      resize_ww1 = int(wpct1 / 100.0 * resize_ww0 + 0.5);

   if (strEqu(event,"hpct"))                                               //  height % - set pixel height
      resize_hh1 = int(hpct1 / 100.0 * resize_hh0 + 0.5);
   
   if (lock && event[0] == 'w')                                            //  preserve width/height ratio
      resize_hh1 = int(resize_ww1 * (1.0 * resize_hh0 / resize_ww0) + 0.5);
   if (lock && event[0] == 'h') 
      resize_ww1 = int(resize_hh1 * (1.0 * resize_ww0 / resize_hh0) + 0.5);
   
   hpct1 = 100.0 * resize_hh1 / resize_hh0;                                //  set percents to match pixels
   wpct1 = 100.0 * resize_ww1 / resize_ww0;
   
   zdialog_stuff(zd,"wpix",resize_ww1);                                    //  index all widget values
   zdialog_stuff(zd,"hpix",resize_hh1);
   zdialog_stuff(zd,"wpct",wpct1);
   zdialog_stuff(zd,"hpct",hpct1);
   
   return 1;
}


/**************************************************************************/

//  menu function - batch resize - resize many files at once
//  resized files are .jpg files

int      batchresize[2] = { 1600, 1200 };                                  //  batch resize width, height

void m_batchresize(GtkWidget *, cchar *)                                   //  new v.10.8
{
   int   batchresize_dialog_event(zdialog *zd, cchar *event);

   zdialog     *zd;                                                        //  batch resize files dialog

   zfuncs::F1_help_topic = "batch_resize";                                 //  v.10.8

   if (mod_keep()) return;                                                 //  unsaved edits
   if (! menulock(1)) return;                                              //  lock menus

   //  [select files]  N files selected
   //  new max. width   [____]
   //  new max. height  [____]
   //  (o) replace originals
   //  (o) export to location [browse] [x] copy EXIF
   //  [___________________________________________]
   //
   //                   [resize]  [cancel]

   zd = zdialog_new(ZTX("Batch Resize"),mWin,Bproceed,Bcancel,null);
   zdialog_add_widget(zd,"hbox","hbf","dialog",0,"space=5");
   zdialog_add_widget(zd,"button","files","hbf",Bselectfiles,"space=5");
   zdialog_add_widget(zd,"label","fcount","hbf","0 files selected","space=10");
   zdialog_add_widget(zd,"hbox","hbwh","dialog","space=10");
   zdialog_add_widget(zd,"vbox","vbwh1","hbwh",0,"homog|space=5");
   zdialog_add_widget(zd,"vbox","vbwh2","hbwh",0,"homog|space=5");
   zdialog_add_widget(zd,"label","labw","vbwh1",ZTX("new max. width"));
   zdialog_add_widget(zd,"label","labh","vbwh1",ZTX("new max. height"));
   zdialog_add_widget(zd,"entry","maxw","vbwh2","1000","scc=5");
   zdialog_add_widget(zd,"entry","maxh","vbwh2","700","scc=5");
   zdialog_add_widget(zd,"hbox","hbsep","dialog",0,"space=5");
   zdialog_add_widget(zd,"hbox","hbloc","dialog");
   zdialog_add_widget(zd,"vbox","vbloc1","hbloc",0,"homog");
   zdialog_add_widget(zd,"vbox","vbloc2","hbloc",0,"homog");
   zdialog_add_widget(zd,"radio","replace","vbloc1",ZTX("replace originals"));
   zdialog_add_widget(zd,"radio","export","vbloc1",ZTX("export to location"));
   zdialog_add_widget(zd,"label","space","vbloc2");
   zdialog_add_widget(zd,"hbox","hbloc2","vbloc2","space=5");
   zdialog_add_widget(zd,"button","browse","hbloc2",Bbrowse,"space=5");
   zdialog_add_widget(zd,"check","copyexif","hbloc2",ZTX("copy EXIF"),"space=5");
   zdialog_add_widget(zd,"hbox","hbloc3","dialog",0,"expand");
   zdialog_add_widget(zd,"entry","location","hbloc3",0,"expand|space=5");

   zdialog_stuff(zd,"replace",0);                                          //  defaults: export, no EXIF
   zdialog_stuff(zd,"export",1);
   zdialog_stuff(zd,"copyexif",0);
   zdialog_stuff(zd,"maxw",batchresize[0]);                                //  v.10.9
   zdialog_stuff(zd,"maxh",batchresize[1]);
   
   zdialog_run(zd,batchresize_dialog_event);                               //  run dialog
   zdialog_wait(zd);                                                       //  wait for completion

   menulock(0);
   return;
}


//  dialog event and completion callback function

int batchresize_dialog_event(zdialog *zd, cchar *event)
{
   static char    **flist = 0;
   char           countmess[32];
   int            Freplace, maxw, maxh, Fcopyexif, yn;
   int            ii, err, lcc, fcc, ww, hh;
   char           location[maxfcc], *ploc, *pfile, *pext;
   char           *oldfile, *newfile, *tempfile;
   double         scale, wscale, hscale;
   PXM            *pxmout;
   struct stat    statb;

   if (strEqu(event,"files"))                                              //  select images to resize
   {
      if (flist) {                                                         //  free prior list
         for (ii = 0; flist[ii]; ii++) 
            zfree(flist[ii]);
         zfree(flist);
      }
      
      flist = image_gallery_getfiles(0);                                   //  get list of files to resize   v.10.9

      if (flist)                                                           //  count files selected
         for (ii = 0; flist[ii]; ii++);
      else ii = 0;

      sprintf(countmess,"%d files selected",ii);                           //  update dialog
      zdialog_stuff(zd,"fcount",countmess);
   }
   
   if (strEqu(event,"browse"))
   {
      ploc = zgetfile1(ZTX("Select directory"),"folder",image_dirk);       //  get directory          v.10.9
      if (! ploc) return 0;
      zdialog_stuff(zd,"location",ploc);
      zfree(ploc);
      zdialog_stuff(zd,"export",1);
   }

   if (! zd->zstat) return 0;                                              //  dialog still busy

   if (zd->zstat != 1) goto cleanup;                                       //  dialog canceled
   
   if (! flist) {                                                          //  no files selected
      zd->zstat = 0;                                                       //  keep dialog active
      return 0;
   }

   zdialog_fetch(zd,"maxw",maxw);                                          //  new max width
   zdialog_fetch(zd,"maxh",maxh);                                          //  new max height
   zdialog_fetch(zd,"replace",Freplace);                                   //  replace originals y/n
   zdialog_fetch(zd,"copyexif",Fcopyexif);                                 //  copy EXIF y/n
   zdialog_fetch(zd,"location",location,maxfcc);                           //  output location (Freplace = 0)
   strTrim2(location,location);                                            //  trim leading and trailing blanks

   if (Freplace) {
      yn = zmessageYN(MWIN,ZTX("replace original files? (max. %d x %d)"),maxw,maxh);
      if (! yn) {
         zd->zstat = 0;
         return 0;
      }
   }
   else {
      yn = zmessageYN(MWIN,ZTX("copy files? (max. %d x %d) \n"
                               " to location %s"),maxw,maxh,location);
      if (! yn) {
         zd->zstat = 0;
         return 0;
      }
   }

   if (! Freplace) {
      err = stat(location,&statb);
      if (err || ! S_ISDIR(statb.st_mode)) {
         zmessageACK(MWIN,ZTX("location is not a valid directory"));
         zd->zstat = 0;
         return 0;
      }
   }

   if (maxw < 20 || maxh < 20) {
      zmessageACK(MWIN,ZTX("max. size %d x %d is not reasonable"),maxw,maxh);
      zd->zstat = 0;
      return 0;
   }
   
   write_popup_text("open","Resizing files",500,200,mWin);                 //  status monitor popup window

   lcc = strlen(location);
   if (! Freplace && location[lcc-1] == '/') lcc--;                        //  remove trailing '/'

   for (ii = 0; flist[ii]; ii++)                                           //  loop selected files
   {
      oldfile = flist[ii];

      if (Freplace)
         newfile = strdupz(oldfile,8,"batchresize.old");                   //  new file = old file
      else {
         pfile = strrchr(oldfile,'/');
         if (! pfile) continue;
         fcc = strlen(pfile);
         newfile = strdupz(location,fcc+9,"batchresize.new");
         strcpy(newfile+lcc,pfile);                                        //  new file at location
      }
      
      pfile = strrchr(newfile,'/');                                        //  force .jpg extension      v.10.9
      pext = strrchr(pfile,'.');
      if (! pext) pext = pfile + strlen(pfile);
      if (strlen(pext) > 5) pext = pext + strlen(pext);
      strcpy(pext,".jpg");
      
      write_popup_text("write",newfile);

      if (! Freplace) {
         err = stat(newfile,&statb);                                       //  if export, check if file exists
         if (! err) {
            snprintf(command,ccc,"%s", ZTX("new file already exists"));
            write_popup_text("write",command);
            zfree(newfile);
            continue;
         }
      }
      
      tempfile = strdupz(newfile,12,"batchresize.temp");                   //  temp file needed for EXIF copy
      strcat(tempfile,"-temp.jpg");                                        //  v.10.9
      
      f_open(oldfile,0);                                                   //  read old file

      wscale = hscale = 1.0;
      if (Fww > maxw) wscale = 1.0 * maxw / Fww;                           //  compute new size
      if (Fhh > maxh) hscale = 1.0 * maxh / Fhh;
      if (wscale < hscale) scale = wscale;
      else scale = hscale;
      ww = Fww * scale;
      hh = Fhh * scale;
      pxmout = PXM_rescale(Fpxm8,ww,hh);                                   //  rescale file

      PXBwrite(pxmout,tempfile);                                           //  write to temp file

      if (Freplace || Fcopyexif)                                           //  copy EXIF if files replaced
         exif_copy(oldfile,tempfile,0,0,0);                                //    or if requested for export
      
      snprintf(command,ccc,"cp -p \"%s\" \"%s\" ",tempfile,newfile);       //  copy tempfile to newfile
      err = system(command);
      if (err) write_popup_text("write",wstrerror(err));

      remove(tempfile);                                                    //  remove tempfile

      if (Freplace && ! err && strNeq(oldfile,newfile))                    //  remove oldfile if not
         remove(oldfile);                                                  //    already replaced

      zfree(newfile);
      zfree(tempfile);
      PXM_free(pxmout);
   }

   write_popup_text("write","COMPLETED");
   write_popup_text("close",0);
   
   batchresize[0] = maxw;                                                  //  save preferred size    v.10.9
   batchresize[1] = maxh;

cleanup:

   if (flist) {                                                            //  free memory
      for (ii = 0; flist[ii]; ii++) 
         zfree(flist[ii]);
      zfree(flist);
      flist = 0;
   }

   zdialog_free(zd);                                                       //  kill dialog
   return 0;
}


/**************************************************************************/

//  edit image annotation - write text on the image itself

char        *annotate_text = 0;                                            //  annotation text
char        *annotate_font = 0;                                            //  font and size, e.g. "Sans 20"
char        *annotate_color[2] = { 0, 0 };                                 //  foreground (text) and background color
int         annotate_trans[2] = { 0, 0 };                                  //  foreground and background transparency
double      annotate_angle = 0;                                            //  angle of text
char        annotate_file[1000] = "";                                      //  file for annotation data

int         annotate_write = 0;                                            //  0/1/2 = 1st write / re-write / erase
GdkPixbuf   *annotate_pixbuf = 0;                                          //  pixbuf with annotation text
int         annotate_px, annotate_py;                                      //  pixbuf position on image

void  annotate_dialog_stuff(zdialog *zd);                                  //  stuff dialog widgets from memory data
int   annotate_dialog_event(zdialog *zd, cchar *event);                    //  dialog event function
void  annotate_mousefunc();                                                //  mouse event function
void  annotate_image();                                                    //  write text on image
void  annotate_gettext();                                                  //  make pixbuf with text
void  annotate_load();                                                     //  load annotation data from file
void  annotate_save();                                                     //  save annotation data to file


void m_annotate(GtkWidget *, cchar *menu)                                  //  overhauled  v.10.11
{
   char     *pp;
   cchar    *intro = ZTX("Enter text, click/drag on image.\n"
                         "Right click to remove");

   if (! Fexiftool) {                                                      //  exiftool is required
      zmessageACK(MWIN,Bexiftoolmissing);
      return;
   }

   zfuncs::F1_help_topic = "annotate";                                     //  user guide topic

   if (! edit_setup("annotate",0,1)) return;                               //  setup edit

   if (! annotate_text) 
      annotate_text = strdupz("enter text",0,"annotate");
   if (! annotate_font || strEqu(annotate_font,"undefined"))
      annotate_font = strdupz("Sans 20",4,"annotate");                     //  default font and size
   if (! annotate_color[0] || strEqu(annotate_color[0],"undefined")) {
      annotate_color[0] = strdupz("0|0|0",0,"annotate");                   //  default text and background colors
      annotate_color[1] = strdupz("255|255|255",0,"annotate");             //   (black on white)
   }
   
   pp = annotate_color[0];                                                 //  insure at least 20 bytes
   annotate_color[0] = strdupz(pp,20,"annotate");
   zfree(pp);
   pp = annotate_color[1];
   annotate_color[1] = strdupz(pp,20,"annotate");
   zfree(pp);

   if (! *annotate_file) {                                                 //  directory for annotation files
      strncatv(annotate_file,999,get_zuserdir(),"/annotations",null);
      mkdir(annotate_file,0740);
   }


//      Annotate Image
//      text [__________________________________]                          //  setup dialog
//      [font]  size [___|v]  angle [___|v]
//      foreground [color]  transparency [___|v]
//      background [color]  transparency [___|v]
//      annotation file: [open] [save]
//                              [done]  [cancel]

   zdedit = zdialog_new(ZTX("Annotate Image"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"label","intro","dialog",intro,"space=3");
   zdialog_add_widget(zdedit,"hbox","hbtext","dialog",0,"space=3");
   zdialog_add_widget(zdedit,"label","labtext","hbtext",ZTX("text"),"space=5");
   zdialog_add_widget(zdedit,"frame","frtext","hbtext",0,"expand");
   zdialog_add_widget(zdedit,"edit","text","frtext",0,"expand");

   zdialog_add_widget(zdedit,"hbox","hbfont","dialog",0,"space=3");
   zdialog_add_widget(zdedit,"button","font","hbfont",Bfont,"space=5");
   zdialog_add_widget(zdedit,"label","space","hbfont","","space=5");
   zdialog_add_widget(zdedit,"label","labsize","hbfont","size","space=3");
   zdialog_add_widget(zdedit,"spin","size","hbfont","6|80|1|20","space=3");
   zdialog_add_widget(zdedit,"label","space","hbfont","","space=5");
   zdialog_add_widget(zdedit,"label","labangle","hbfont","angle","space=3");
   zdialog_add_widget(zdedit,"spin","angle","hbfont","-180|180|0.2|0","space=3");
   
   zdialog_add_widget(zdedit,"hbox","hbfg","dialog",0,"space=3");
   zdialog_add_widget(zdedit,"label","labfg","hbfg",ZTX("foreground"),"space=5");
   zdialog_add_widget(zdedit,"colorbutt","fgcolor","hbfg","0|0|0","space=3");
   zdialog_add_widget(zdedit,"label","space","hbfg","","space=5");
   zdialog_add_widget(zdedit,"label","labfgt","hbfg",ZTX("transparency"),"space=3");
   zdialog_add_widget(zdedit,"spin","fgtrans","hbfg","0|100|1|0","space=3");
   
   zdialog_add_widget(zdedit,"hbox","hbbg","dialog",0,"space=3");
   zdialog_add_widget(zdedit,"label","labbg","hbbg",ZTX("background"),"space=5");
   zdialog_add_widget(zdedit,"colorbutt","bgcolor","hbbg","255|255|255","space=3");
   zdialog_add_widget(zdedit,"label","space","hbbg","","space=5");
   zdialog_add_widget(zdedit,"label","labbgt","hbbg",ZTX("transparency"),"space=3");
   zdialog_add_widget(zdedit,"spin","bgtrans","hbbg","0|100|1|0","space=3");

   zdialog_add_widget(zdedit,"hbox","hbaf","dialog",0,"space=3");
   zdialog_add_widget(zdedit,"label","labbg","hbaf",ZTX("annotation file:"),"space=5");
   zdialog_add_widget(zdedit,"button","load","hbaf",Bopen,"space=5");
   zdialog_add_widget(zdedit,"button","save","hbaf",Bsave,"space=5");

   annotate_dialog_stuff(zdedit);                                          //  stuff dialog widgets from memory data

   zdialog_run(zdedit,annotate_dialog_event);                              //  start dialog

   mouseCBfunc = annotate_mousefunc;                                       //  connect mouse function
   Mcapture++;

   annotate_write = 0;                                                     //  no write on image (yet)
   annotate_px = annotate_py = -1;                                         //  no location on image (yet)
   return;
}


//  stuff all dialog widgets from annotation data in memory

void  annotate_dialog_stuff(zdialog *zd)
{
   int      size;
   char     *pp;

   zdialog_stuff(zd,"text",annotate_text);
   zdialog_stuff(zd,"angle",annotate_angle);
   zdialog_stuff(zd,"fgtrans",annotate_trans[0]);
   zdialog_stuff(zd,"bgtrans",annotate_trans[1]);
   zdialog_stuff(zd,"fgcolor",annotate_color[0]);
   zdialog_stuff(zd,"bgcolor",annotate_color[1]);

   pp = annotate_font + strlen(annotate_font);
   while (*pp != ' ') pp--;
   if (pp > annotate_font) {
      size = atoi(pp);
      if (size >= 6 && size <= 80) 
         zdialog_stuff(zd,"size",size);
   }
   
   return;
}


//  dialog event and completion callback function

int annotate_dialog_event(zdialog *zd, cchar *event)
{
   GtkWidget   *font_dialog;
   int         size;
   char        *pp, text[1000];
   
   if (zd->zstat)
   {
      if (zd->zstat == 1 && Fmodified)                                     //  Done
         edit_done();                                                      //  complete pending edit
      else edit_cancel();                                                  //  Cancel or kill

      mouseCBfunc = 0;                                                     //  disconnect mouse
      Mcapture = 0;
      return 0;
   }
   
   if (strEqu(event,"text")) {
      zdialog_fetch(zd,"text",text,999);                                   //  get text from dialog
      if (annotate_text) zfree(annotate_text);
      annotate_text = 0;
      if (*text) annotate_text = strdupz(text,0,"annotate");
   }
   
   if (strEqu(event,"font")) {                                             //  new font
      font_dialog = gtk_font_selection_dialog_new(ZTX("select font"));
      gtk_font_selection_dialog_set_font_name(GTK_FONT_SELECTION_DIALOG(font_dialog),annotate_font);
      gtk_dialog_run(GTK_DIALOG(font_dialog));
      pp = gtk_font_selection_dialog_get_font_name(GTK_FONT_SELECTION_DIALOG(font_dialog));
      gtk_widget_destroy(GTK_WIDGET(font_dialog));

      if (pp) {
         zfree(annotate_font);
         annotate_font = strdupz(pp,4,"annotate");                         //  update font and size
         g_free(pp);
         pp = annotate_font + strlen(annotate_font);
         while (*pp != ' ') pp--;
         if (pp > annotate_font) {
            size = atoi(pp);
            if (size >= 6 && size <= 80) zdialog_stuff(zd,"size",size);
         }
      }
   }

   if (strEqu(event,"load")) {                                             //  load annotate data from file
      annotate_load();
      annotate_dialog_stuff(zd);
      return 1;
   }

   if (strEqu(event,"save")) {                                             //  save annotate data to file
      annotate_save();
      return 1;
   }
   
   if (strEqu(event,"size")) {                                             //  new font size
      zdialog_fetch(zd,"size",size);
      pp = annotate_font + strlen(annotate_font);                          //  "fontname NN"
      while (*pp != ' ') pp--;                                             //  back-up to " NN"
      if (pp > annotate_font) sprintf(pp," %d",size);                      //  replace NN with new size
   }

   if (strEqu(event,"angle"))
      zdialog_fetch(zd,"angle",annotate_angle);
   
   if (strEqu(event,"fgcolor"))                                            //  foreground (text) color
      zdialog_fetch(zd,"fgcolor",annotate_color[0],20);
   
   if (strEqu(event,"bgcolor"))                                            //  background color
      zdialog_fetch(zd,"bgcolor",annotate_color[1],20);
   
   if (strEqu(event,"fgtrans"))                                            //  foreground transparency
      zdialog_fetch(zd,"fgtrans",annotate_trans[0]);
   
   if (strEqu(event,"bgtrans"))                                            //  background transparency
      zdialog_fetch(zd,"bgtrans",annotate_trans[1]);

   annotate_image();                                                       //  write on image

   return 1;
}


//  mouse function, set position for annotation text on image

void annotate_mousefunc()
{
   static int     dragging = 0;

   if (LMclick) {                                                          //  left mouse click
      LMclick = 0;
      annotate_px = Mxclick;                                               //  new pixbuf position on image
      annotate_py = Myclick;
      annotate_image();                                                    //  write text on image
   }
   
   if (RMclick) {                                                          //  right mouse click
      RMclick = 0;
      annotate_write = 2;                                                  //  erase text on image
      annotate_image();
      annotate_px = annotate_py = -1;                                      //  no location on image
   }
   
   if (Mxdrag || Mydrag)                                                   //  mouse dragged
   {
      if (! dragging && annotate_write) dragging = 1;

      if (dragging) {
         annotate_px = Mxdrag;                                             //  new pixbuf position on image
         annotate_py = Mydrag;
         annotate_image();                                                 //  write text on image
      }
   }
   else  dragging = 0;                                                     //  no drag underway

   return;
}


//  write annotation text on image at designated place, if wanted
//  annotate_write: 0/1/2 = 1st write / re-write / erase

void annotate_image()
{
   GdkPixbuf   *pixbuf;
   uint8       *ppix2, *pix2;
   uint16      *pix3, *pix31, *pix33;
   int         px2, py2, rs2, px3, py3;
   int         fgred, fggreen, fgblue;
   int         bgred, bggreen, bgblue;
   int         red, green, blue;
   double      fgtrans, bgtrans, fgpart, bgpart, e3part;
   double      f256 = 1.0 / 256.0;
   cchar       *pp;

   static int  orgx1 = 0, orgy1 = 0, ww1 = 0, hh1 = 0;                     //  old pixbuf overlap rectangle
   int         orgx2, orgy2, ww2, hh2;                                     //  new overlap rectangle
   
   fgred = fggreen = fgblue = 0;                                           //  black text on white
   bgred = bggreen = bgblue = 255;

   if (annotate_write)                                                     //  re-write or erase
   {
      for (py3 = orgy1; py3 < orgy1 + hh1; py3++)                          //  erase prior pixbuf
      for (px3 = orgx1; px3 < orgx1 + ww1; px3++)                          //  replace E3 pixels with E1 pixels
      {                                                                    //    in prior overlap rectangle
         if (px3 < 0 || px3 >= E3ww) continue;
         if (py3 < 0 || py3 >= E3hh) continue;
         pix31 = PXMpix(E1pxm16,px3,py3);
         pix33 = PXMpix(E3pxm16,px3,py3);
         pix33[0] = pix31[0];
         pix33[1] = pix31[1];
         pix33[2] = pix31[2];
      }
      
      Fmodified = 0;
   }

   if (annotate_write == 2) {                                              //  erase only
      annotate_write = 0;                                                  //  next time is a 1st write
      mwpaint3(orgx1,orgy1,ww1,hh1);                                       //  update window
      return;
   }   

   if (! annotate_text) {                                                  //  no text to write
      annotate_write = 0;                                                  //  next time is a 1st write
      mwpaint3(orgx1,orgy1,ww1,hh1);                                       //  update window
      return;
   }
   
   if (annotate_px < 0 || annotate_py < 0) {                               //  no image position to write
      annotate_write = 0;                                                  //  next time is a 1st write
      mwpaint3(orgx1,orgy1,ww1,hh1);                                       //  update window
      return;
   }

   annotate_write = 1;                                                     //  next time is a re-write

   annotate_gettext();                                                     //  get pixbuf with formatted text
   pixbuf = annotate_pixbuf;                                               //  (white on black)

   ww2 = gdk_pixbuf_get_width(pixbuf);                                     //  copy-from pixbuf
   hh2 = gdk_pixbuf_get_height(pixbuf);
   rs2 = gdk_pixbuf_get_rowstride(pixbuf);
   ppix2 = gdk_pixbuf_get_pixels(pixbuf);
   
   orgx2 = annotate_px - ww2/2;                                            //  copy-to image3 location
   orgy2 = annotate_py - hh2/2;

   pp = strField(annotate_color[0],'|',1);                                 //  get text foreground color
   if (pp) fgred = atoi(pp);                                               //  0 - 255 per RGB color
   pp = strField(annotate_color[0],'|',2);
   if (pp) fggreen = atoi(pp);
   pp = strField(annotate_color[0],'|',3);
   if (pp) fgblue = atoi(pp);

   pp = strField(annotate_color[1],'|',1);                                 //  get text background color
   if (pp) bgred = atoi(pp);
   pp = strField(annotate_color[1],'|',2);
   if (pp) bggreen = atoi(pp);
   pp = strField(annotate_color[1],'|',3);
   if (pp) bgblue = atoi(pp);
   
   fgtrans = 0.01 * annotate_trans[0];                                     //  get foreground and background 
   bgtrans = 0.01 * annotate_trans[1];                                     //    transparencies, range 0 - 1

   for (py2 = 0; py2 < hh2; py2++)                                         //  loop all pixels in pixbuf
   for (px2 = 0; px2 < ww2; px2++)
   {
      px3 = orgx2 + px2;                                                   //  copy-to image3 pixel
      py3 = orgy2 + py2;
      pix3 = PXMpix(E3pxm16,px3,py3);

      if (px3 < 0 || px3 >= E3ww) continue;                                //  omit parts beyond edges
      if (py3 < 0 || py3 >= E3hh) continue;

      pix2 = ppix2 + rs2 * py2 + 3 * px2;                                  //  copy-from pixbuf pixel
      
      fgpart = pix2[1] * f256;                                             //  text foreground part, 0 - 1
      bgpart = 1.0 - fgpart;                                               //  text background part, 1 - 0
      
      fgpart = fgpart * (1.0 - fgtrans);                                   //  reduce for transparencies
      bgpart = bgpart * (1.0 - bgtrans);
      e3part = (1.0 - fgpart - bgpart);                                    //  image part visible through text
      
      red = fgred * fgpart + bgred * bgpart;                               //  red part for text + text background
      red = red + f256 * pix3[0] * e3part;                                 //  + red part for image
      pix3[0] = red << 8;                                                  //  output total red

      green = fggreen * fgpart + bggreen * bgpart;                         //  same for green
      green = green + f256 * pix3[1] * e3part;
      pix3[1] = green << 8;

      blue = fgblue * fgpart + bgblue * bgpart;                            //  same for blue
      blue = blue + f256 * pix3[2] * e3part;
      pix3[2] = blue << 8;
   }
   
   mwpaint3(orgx1,orgy1,ww1,hh1);                                          //  update prior overlap rectangle
   mwpaint3(orgx2,orgy2,ww2,hh2);                                          //  update new pixbuf overlap

   Fmodified = 1;
   
   orgx1 = orgx2;                                                          //  remember overlap rectangle
   orgy1 = orgy2;                                                          //    for next call
   ww1 = ww2 + 1;
   hh1 = hh2 + 1;                                                          //  +1 stops incomplete erasure (??)
   
   return;
}


//  Create a pixbuf with text, white on black, any font, any angle.

void annotate_gettext()
{
   PangoFontDescription  *pfont;
   static GdkColormap    *colormap = 0;
   static GdkColor       black, white;

   PangoLayout    *playout;
   GdkPixmap      *pixmap;
   GdkGC          *gdkgc;
   GdkPixbuf      *pixbuf, *pixbufR;
   
   char           *text = annotate_text;
   char           *font = annotate_font;
   double         angle = annotate_angle;
   
   int            ww, hh;
   
   if (! colormap) {
      colormap = gtk_widget_get_colormap(drWin);
      black.red = black.green = black.blue = 0;
      white.red = white.green = white.blue = 0xffff;
      gdk_rgb_find_color(colormap,&black);
      gdk_rgb_find_color(colormap,&white);
   }
   
   pfont = pango_font_description_from_string(font);                       //  make layout with text
   playout = gtk_widget_create_pango_layout(drWin,null);
   pango_layout_set_font_description(playout,pfont);
   pango_layout_set_text(playout,text,-1);
   pango_layout_get_pixel_size(playout,&ww,&hh);

   pixmap = gdk_pixmap_new(drWin->window,ww,hh,-1);                        //  then make pixmap
   gdkgc = gdk_gc_new(pixmap);
   gdk_draw_rectangle(pixmap,gdkgc,1,0,0,ww,hh);
   gdk_draw_layout_with_colors(pixmap,gdkgc,0,0,playout,&white,&black);

   pixbuf = gdk_pixbuf_get_from_drawable(null,pixmap,0,0,0,0,0,ww,hh);     //  then make pixbuf
   
   if (angle) {                                                            //  rotate pixbuf if wanted
      pixbufR = gdk_pixbuf_rotate(pixbuf,angle,0);
      g_object_unref(pixbuf);
      pixbuf = pixbufR;
   }
   
   g_object_unref(playout);
   g_object_unref(pixmap);
   g_object_unref(gdkgc);
   
   if (annotate_pixbuf) g_object_unref(annotate_pixbuf);
   annotate_pixbuf = pixbuf;
   return;
}


//  load search criteria from a file

void annotate_load()                                                       //  v.10.11
{
   FILE     *fid;
   int      cc, err;
   char     *pp, *file, buff[1200];
   cchar    *dialogtitle = "load annotation data from a file";

   file = zgetfile1(dialogtitle,"open",annotate_file);                     //  get input file from user
   if (! file) return;
   
   fid = fopen(file,"r");                                                  //  open for read
   if (! fid) {
      zmessageACK(GTK_WINDOW(mWin),"%s",strerror(errno));
      zfree(file);
      return;
   }

   while (true)
   {
      pp = fgets_trim(buff,1200,fid,1);
      if (! pp) break;
      
      if (strnEqu(pp,"annotate_text  ",15)) {
         if (annotate_text) zfree(annotate_text);
         cc = strlen(pp+15) + 100;
         annotate_text = zmalloc(cc,"annotate");
         repl_1str(pp+15,annotate_text,"\\n","\n");
      }
      
      if (strnEqu(pp,"annotate_font  ",15))
         strcpy(annotate_font,pp+15);
         
      if (strnEqu(pp,"annotate_angle  ",16))
         convSD(pp+16,annotate_angle,-180,+180);
      
      if (strnEqu(pp,"annotate_fgcolor  ",18))
         strcpy(annotate_color[0],pp+18);

      if (strnEqu(pp,"annotate_bgcolor  ",18))
         strcpy(annotate_color[1],pp+18);

      if (strnEqu(pp,"annotate_fgtrans  ",18))
         convSI(pp+18,annotate_trans[0],0,100);

      if (strnEqu(pp,"annotate_bgtrans  ",18))
         convSI(pp+18,annotate_trans[1],0,100);
   }

   err = fclose(fid);
   if (err) {
      zmessageACK(GTK_WINDOW(mWin),"%s",strerror(errno));
      zfree(file);
      return;
   }
   
   strcpy(annotate_file,file);                                             //  update current file
   zfree(file);

   return;
}


//  save annotation data to a file

void annotate_save()                                                       //  v.10.11
{
   FILE     *fid;
   char     *file, text[1100];
   cchar    *dialogtitle = "save annotation data to a file";
   int      err;

   file = zgetfile1(dialogtitle,"save",annotate_file);                     //  get output file from user
   if (! file) return;
   
   fid = fopen(file,"w");                                                  //  open for write
   if (! fid) {
      zmessageACK(GTK_WINDOW(mWin),"%s",strerror(errno));
      zfree(file);
      return;
   }
   
   repl_1str(annotate_text,text,"\n","\\n");                               //  replace newlines with "\n"

   fprintf(fid,"annotate_text  %s \n", text);
   fprintf(fid,"annotate_font  %s \n", annotate_font);
   fprintf(fid,"annotate_angle  %.1f \n", annotate_angle);
   fprintf(fid,"annotate_fgcolor  %s \n", annotate_color[0]);
   fprintf(fid,"annotate_bgcolor  %s \n", annotate_color[1]);
   fprintf(fid,"annotate_fgtrans  %d \n", annotate_trans[0]);
   fprintf(fid,"annotate_bgtrans  %d \n", annotate_trans[1]);

   fprintf(fid,"\n");
   
   err = fclose(fid);
   if (err) {
      zmessageACK(GTK_WINDOW(mWin),"file I/O error %s",file);
      zfree(file);
      return;
   }
   
   strcpy(annotate_file,file);                                             //  update current file
   zfree(file);

   return;
}


/**************************************************************************/

//  rotate image through any arbitrary angle

double      rotate_angle = 0;                                              //  E3 rotatation vs. F
double      rotate_delta = 0;
int         rotate_trim = 0;


void m_rotate(GtkWidget *, cchar *menu)                                    //  menu function
{
   int    rotate_dialog_event(zdialog *zd, cchar *event);
   void * rotate_thread(void *);
   void   rotate_mousefunc();
   
   cchar  *rotmess = ZTX("Use buttons or drag right edge with mouse");

   zfuncs::F1_help_topic = "rotate";                                       //  v.10.8

   if (! edit_setup("rotate",1,0)) return;                                 //  setup edit: use preview

   zdedit = zdialog_new(ZTX("Rotate Image"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"label","labrot","dialog",ZTX(rotmess),"space=5");
   zdialog_add_widget(zdedit,"label","labdeg","dialog",ZTX("degrees"),"space=5");
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"homog|space=5");
   zdialog_add_widget(zdedit,"vbox","vb1","hb1",0,"space=5");
   zdialog_add_widget(zdedit,"vbox","vb2","hb1",0,"space=5");
   zdialog_add_widget(zdedit,"vbox","vb3","hb1",0,"space=5");
   zdialog_add_widget(zdedit,"vbox","vb4","hb1",0,"space=5");
   zdialog_add_widget(zdedit,"button"," +0.1  ","vb1"," + 0.1 ");          //  button name is increment to use
   zdialog_add_widget(zdedit,"button"," -0.1  ","vb1"," - 0.1 ");
   zdialog_add_widget(zdedit,"button"," +1.0  ","vb2"," + 1   ");
   zdialog_add_widget(zdedit,"button"," -1.0  ","vb2"," - 1   ");
   zdialog_add_widget(zdedit,"button"," +10.0 ","vb3"," + 10  ");
   zdialog_add_widget(zdedit,"button"," -10.0 ","vb3"," - 10  ");
   zdialog_add_widget(zdedit,"button"," +90.0 ","vb4"," + 90  ");
   zdialog_add_widget(zdedit,"button"," -90.0 ","vb4"," - 90  ");
   zdialog_add_widget(zdedit,"hbox","hb2","dialog",0,"space=10");
   zdialog_add_widget(zdedit,"button","trim","hb2",ZTX("Trim"),"space=10");

   zdialog_run(zdedit,rotate_dialog_event);                                //  run dialog - parallel

   mouseCBfunc = rotate_mousefunc;                                         //  connect mouse function
   Mcapture++;
   set_cursor(dragcursor);                                                 //  set drag cursor

   rotate_angle = rotate_delta = rotate_trim = 0;
   start_thread(rotate_thread,0);                                          //  start working thread
   return;
}


//  dialog event and completion callback function

int rotate_dialog_event(zdialog *zd, cchar * event)
{
   int         err, trim = 0;
   double      incr;
   char        text[20];
   
   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) {
         rotate_delta = rotate_angle;                                      //  rotate main image
         rotate_angle = 0;
         edit_done();
      }
      else edit_cancel();

      rotate_angle = rotate_delta = rotate_trim = 0;
      mouseCBfunc = 0;                                                     //  disconnect mouse
      Mcapture = 0;
      set_cursor(0);                                                       //  restore normal cursor
      return 0;
   }
   
   if (strEqu(event,"undo")) edit_undo();                                  //  v.10.2
   if (strEqu(event,"redo")) edit_redo();                                  //  v.10.3

   if (strEqu(event,"trim")) {
      rotate_trim = 1 - rotate_trim;                                       //  toggle trim button
      if (rotate_trim) zdialog_stuff(zd,"trim",ZTX("Undo Trim"));
      else zdialog_stuff(zd,"trim",ZTX("Trim"));
      trim = 1;                                                            //  v.10.3
   }
   
   if (strpbrk(event,"+-")) {
      err = convSD(event,incr);                                            //  button name is increment to use
      if (err) return 0;
      rotate_delta += incr;
   }

   if (rotate_delta || trim) {
      trim = 0;
      zdialog_stuff(zd,"labdeg","computing");
      signal_thread();                                                     //  do rotate in thread
      wait_thread_idle();
      sprintf(text,ZTX("degrees: %.1f"),rotate_angle);                     //  update dialog angle display
      zdialog_stuff(zd,"labdeg",text);
   }

   return 1;
}


//  rotate mouse function - drag right edge of image up/down for rotation

void rotate_mousefunc()
{
   static int     mpx0 = 0, mpy0 = 0;
   static int     mpy1, mpy2, dist;

   if (! Mxdrag && ! Mydrag) return;                                       //  no drag underway
   if (Mxdrag < 0.8 * E3ww) return;                                        //  not right edge of image

   if (Mxdown != mpx0 || Mydown != mpy0) {
      mpx0 = Mxdown;                                                       //  new drag started
      mpy0 = mpy1 = Mydown;
   }
   
   mpy2 = Mydrag;
   dist = mpy2 - mpy1;                                                     //  drag distance
   mpy1 = mpy2;                                                            //  reset origin for next time
   if (! dist) return;

   rotate_delta = 30.0 * dist / E3ww;                                      //  convert to angle
   rotate_dialog_event(zdedit,"mouse");
   return;
}


//  rotate thread function

void * rotate_thread(void *)
{
   int         px3, py3, px9, py9;
   int         wwcut, hhcut, ww, hh;
   double      trim_angle, radians;
   uint16      *pix3, *pix9;
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      mutex_lock(&pixmaps_lock);

      rotate_angle += rotate_delta;                                        //  accum. net rotation   
      rotate_delta = 0;                                                    //    from dialog widget
      
      if (rotate_angle >= 360) rotate_angle -=360;
      if (rotate_angle <= -360) rotate_angle +=360;
      if (fabs(rotate_angle) < 0.01) rotate_angle = 0;

      if (! rotate_angle) {
         PXM_free(E3pxm16);                                                //  E1 >> E3
         E3pxm16 = PXM_copy(E1pxm16);
         E3ww = E1ww;
         E3hh = E1hh;
         Fmodified = 0;
      }
      
      if (rotate_angle) {
         PXM_free(E3pxm16);
         E3pxm16 = PXM_rotate(E1pxm16,rotate_angle);                       //  E3 is rotated E1
         E3ww = E3pxm16->ww;
         E3hh = E3pxm16->hh;
         Fmodified = 1;
      }

      if (rotate_trim)
      {                                                                    //  auto trim
         trim_angle = fabs(rotate_angle);
         while (trim_angle > 45) trim_angle -= 90;
         radians = fabs(trim_angle / 57.296);
         wwcut = int(E3pxm16->hh * sin(radians) + 1);                      //  amount to trim
         hhcut = int(E3pxm16->ww * sin(radians) + 1);
         ww = E3pxm16->ww - 2 * wwcut;
         hh = E3pxm16->hh - 2 * hhcut;
         if (ww > 0 && hh > 0) {
            E9pxm16 = PXM_make(ww,hh,16);
            
            for (py3 = hhcut; py3 < E3hh-hhcut; py3++)                     //  E9 = trimmed E3
            for (px3 = wwcut; px3 < E3ww-wwcut; px3++)
            {
               px9 = px3 - wwcut;
               py9 = py3 - hhcut;
               pix3 = PXMpix(E3pxm16,px3,py3);
               pix9 = PXMpix(E9pxm16,px9,py9);
               pix9[0] = pix3[0];
               pix9[1] = pix3[1];
               pix9[2] = pix3[2];
            }

            PXM_free(E3pxm16);                                             //  E3 = E9
            E3pxm16 = E9pxm16;
            E9pxm16 = 0;
            E3ww = ww;
            E3hh = hh;
         }
      }
      
      mutex_unlock(&pixmaps_lock);
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}


/**************************************************************************/

//  flip an image horizontally or vertically

void m_flip(GtkWidget *, cchar *)
{
   int flip_dialog_event(zdialog *zd, cchar *event);

   zfuncs::F1_help_topic = "flip_image";                                   //  v.10.8

   if (! edit_setup("flip",0,0)) return;                                   //  setup edit: no preview

   zdedit = zdialog_new(ZTX("Flip Image"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"space=10");
   zdialog_add_widget(zdedit,"button","horz","hb1",ZTX("horizontal"),"space=5");
   zdialog_add_widget(zdedit,"button","vert","hb1",ZTX("vertical"),"space=5");

   zdialog_run(zdedit,flip_dialog_event);                                  //  run dialog - parallel
   return;
}


//  dialog event and completion callback function

int flip_dialog_event(zdialog *zd, cchar *event)
{
   int flip_horz();
   int flip_vert();

   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) edit_done();
      else edit_cancel();
      return 0;
   }

   if (strEqu(event,"undo")) edit_undo();                                  //  v.10.2
   if (strEqu(event,"redo")) edit_redo();                                  //  v.10.3
   if (strEqu(event,"horz")) flip_horz();
   if (strEqu(event,"vert")) flip_vert();
   return 0;
}


int flip_horz()
{
   int      px, py;
   uint16   *pix3, *pix9;
   
   edit_zapredo();                                                         //  delete redo copy    v.10.3

   E9pxm16 = PXM_copy(E3pxm16);
   
   for (py = 0; py < E3hh; py++)
   for (px = 0; px < E3ww; px++)
   {
      pix3 = PXMpix(E3pxm16,px,py);                                        //  image9 = flipped image3
      pix9 = PXMpix(E9pxm16,E3ww-1-px,py);
      pix9[0] = pix3[0];
      pix9[1] = pix3[1];
      pix9[2] = pix3[2];
   }
   
   mutex_lock(&pixmaps_lock);
   PXM_free(E3pxm16);                                                      //  image9 >> image3
   E3pxm16 = E9pxm16;
   E9pxm16 = 0;
   mutex_unlock(&pixmaps_lock);

   Fmodified = 1;
   mwpaint2();
   return 0;
}


int flip_vert()
{
   int      px, py;
   uint16   *pix3, *pix9;
   
   edit_zapredo();                                                         //  delete redo copy    v.10.3

   E9pxm16 = PXM_copy(E3pxm16);
   
   for (py = 0; py < E3hh; py++)
   for (px = 0; px < E3ww; px++)
   {
      pix3 = PXMpix(E3pxm16,px,py);                                        //  image9 = flipped image3
      pix9 = PXMpix(E9pxm16,px,E3hh-1-py);
      pix9[0] = pix3[0];
      pix9[1] = pix3[1];
      pix9[2] = pix3[2];
   }
   
   mutex_lock(&pixmaps_lock);
   PXM_free(E3pxm16);                                                      //  image9 >> image3
   E3pxm16 = E9pxm16;
   E9pxm16 = 0;
   mutex_unlock(&pixmaps_lock);

   Fmodified = 1;
   mwpaint2();
   return 0;
}


/**************************************************************************/

//  unbend an image
//  straighten curvature added by pano or improve perspective

int      unbend_horz, unbend_vert;                                         //  unbend values from dialog
int      unbend_vert2, unbend_horz2;
double   unbend_x1, unbend_x2, unbend_y1, unbend_y2;                       //  unbend axes scaled 0 to 1
int      unbend_hx1, unbend_hy1, unbend_hx2, unbend_hy2;
int      unbend_vx1, unbend_vy1, unbend_vx2, unbend_vy2;


void m_unbend(GtkWidget *, cchar *)                                        //  overhaul for preview
{
   int    unbend_dialog_event(zdialog* zd, cchar *event);
   void * unbend_thread(void *);
   void   unbend_mousefunc();

   zfuncs::F1_help_topic = "unbend";                                       //  v.10.8

   if (! edit_setup("unbend",1,0)) return;                                 //  setup edit: preview

   zdedit = zdialog_new(ZTX("Unbend Image"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"space=10");
   zdialog_add_widget(zdedit,"vbox","vb1","hb1",0,"homog|space=10");
   zdialog_add_widget(zdedit,"vbox","vb2","hb1",0,"homog|space=10");
   zdialog_add_widget(zdedit,"spin","spvert","vb1","-30|30|1|0");
   zdialog_add_widget(zdedit,"spin","sphorz","vb1","-20|20|1|0");
   zdialog_add_widget(zdedit,"label","labvert","vb2",ZTX("vertical unbend"));
   zdialog_add_widget(zdedit,"label","labhorz","vb2",ZTX("horizontal unbend"));
   
   zdialog_resize(zdedit,260,0);
   zdialog_run(zdedit,unbend_dialog_event);                                //  run dialog, parallel

   unbend_x1 = unbend_x2 = unbend_y1 = unbend_y2 = 0.5;                    //  initial axes thru image middle
   unbend_horz = unbend_vert = 0;

   Mcapture = 1;
   mouseCBfunc = unbend_mousefunc;                                         //  connect mouse function
   
   start_thread(unbend_thread,0);                                          //  start working thread
   signal_thread();

   return;
}


//  dialog event and completion callback function

int unbend_dialog_event(zdialog *zd, cchar *event)
{
   if (zd->zstat)                                                          //  dialog complete
   {
      paint_toplines(2);                                                   //  erase axes-lines
      mouseCBfunc = 0;                                                     //  disconnect mouse
      Mcapture = 0;

      if (zd->zstat != 1) {
         edit_cancel();                                                    //  canceled
         return 0;
      }

      if (unbend_vert || unbend_horz) Fmodified = 1;                       //  image3 modified
      else Fmodified = 0;
      edit_done();                                                         //  commit changes to image3
      return 1;
   }
   
   if (strEqu(event,"undo")) edit_undo();                                  //  v.10.2
   if (strEqu(event,"redo")) edit_redo();                                  //  v.10.3

   if (strstr(event,"sp")) {
      zdialog_fetch(zd,"spvert",unbend_vert);                              //  get new unbend values
      zdialog_fetch(zd,"sphorz",unbend_horz);
      signal_thread();                                                     //  trigger thread
   }

   return 1;
}


//  unbend mouse function                                                  //  adjustable axes

void unbend_mousefunc()   
{
   cchar       *close;
   double      dist1, dist2;
   double      mpx = 0, mpy = 0;
   
   if (LMclick) {                                                          //  left mouse click
      LMclick = 0;
      mpx = Mxclick;
      mpy = Myclick;
   }
   
   if (Mxdrag || Mydrag) {                                                 //  mouse dragged
      mpx = Mxdrag;
      mpy = Mydrag;
   }
   
   if (! mpx && ! mpy) return;

   mpx = 1.0 * mpx / E3ww;                                                 //  scale mouse position 0 to 1
   mpy = 1.0 * mpy / E3hh;

   if (mpx < 0.2 || mpx > 0.8 ) {                                          //  check reasonable position
      if (mpy < 0.1 || mpy > 0.9) return;
   }
   else if (mpy < 0.2 || mpy > 0.8) {
      if (mpx < 0.1 || mpx > 0.9) return;
   }
   else return;

   close = "?";                                                            //  find closest axis end-point
   dist1 = 2;

   dist2 = mpx * mpx + (mpy-unbend_y1) * (mpy-unbend_y1);
   if (dist2 < dist1) {
      dist1 = dist2;
      close = "left";
   }

   dist2 = (1-mpx) * (1-mpx) + (mpy-unbend_y2) * (mpy-unbend_y2);
   if (dist2 < dist1) {
      dist1 = dist2;
      close = "right";
   }

   dist2 = (mpx-unbend_x1) * (mpx-unbend_x1) + mpy * mpy;
   if (dist2 < dist1) {
      dist1 = dist2;
      close = "top";
   }

   dist2 = (mpx-unbend_x2) * (mpx-unbend_x2) + (1-mpy) * (1-mpy);
   if (dist2 < dist1) {
      dist1 = dist2;
      close = "bottom";
   }
   
   if (strEqu(close,"left")) unbend_y1 = mpy;                              //  set new axis end-point
   if (strEqu(close,"right")) unbend_y2 = mpy;
   if (strEqu(close,"top")) unbend_x1 = mpx;
   if (strEqu(close,"bottom")) unbend_x2 = mpx;

   signal_thread();                                                        //  trigger thread 

   return ;
}


//  unbend thread function

void * unbend_thread(void *arg)
{
   void * unbend_wthread(void *);

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      unbend_vert2 = int(unbend_vert * 0.01 * E3hh);                       //  convert % to pixels
      unbend_horz2 = int(unbend_horz * 0.005 * E3ww);

      unbend_hx1 = 0;                                                      //  scale axes to E3ww/hh
      unbend_hy1 = unbend_y1 * E3hh;
      unbend_hx2 = E3ww;
      unbend_hy2 = unbend_y2 * E3hh;

      unbend_vx1 = unbend_x1 * E3ww;
      unbend_vy1 = 0;
      unbend_vx2 = unbend_x2 * E3ww;
      unbend_vy2 = E3hh;

      if (Fpreview) {                                                      //  omit for final unbend
         Ntoplines = 2;
         toplinex1[0] = unbend_hx1;                                        //  lines on window
         topliney1[0] = unbend_hy1;
         toplinex2[0] = unbend_hx2;
         topliney2[0] = unbend_hy2;
         toplinex1[1] = unbend_vx1;
         topliney1[1] = unbend_vy1;
         toplinex2[1] = unbend_vx2;
         topliney2[1] = unbend_vy2;
      }
      
      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wt(unbend_wthread,&wtnx[ii]);
      wait_wts();                                                          //  wait for completion

      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}

   
void * unbend_wthread(void *arg)                                           //  worker thread function
{
   int         index = *((int *) arg);
   int         vstat, px3, py3, cx3, cy3;
   double      px1, py1, dispx, dispx2, dispy;
   uint16      vpix[3], *pix3;

   for (py3 = index; py3 < E3hh; py3 += Nwt)                               //  step through F3 pixels
   for (px3 = 0; px3 < E3ww; px3++)
   {
      pix3 = PXMpix(E3pxm16,px3,py3);                                      //  output pixel

      cx3 = unbend_vx1 + (unbend_vx2 - unbend_vx1) * py3 / E3hh;           //  center of unbend
      cy3 = unbend_hy1 + (unbend_hy2 - unbend_hy1) * px3 / E3ww;
      dispx = 2.0 * (px3 - cx3) / E3ww;                                    //  -1.0 ..  0.0 .. +1.0 (roughly)
      dispy = 2.0 * (py3 - cy3) / E3hh;                                    //  -1.0 ..  0.0 .. +1.0
      dispx2 = dispx * dispx - 0.5;                                        //  +0.5 .. -0.5 .. +0.5  curved

      px1 = px3 + dispx * dispy * unbend_horz2;                            //  input virtual pixel, x
      py1 = py3 - dispy * dispx2 * unbend_vert2;                           //  input virtual pixel, y
      vstat = vpixel(E1pxm16,px1,py1,vpix);                                //  input virtual pixel

      if (vstat) {
         pix3[0] = vpix[0];                                                //  input pixel >> output pixel
         pix3[1] = vpix[1];
         pix3[2] = vpix[2];
      }
      else pix3[0] = pix3[1] = pix3[2] = 0;
   }
      
   exit_wt();
   return 0;                                                               //  not executed, avoid gcc warning
}


/**************************************************************************/

//  warp/distort area - select image area and pull with mouse

float       *WarpAx, *WarpAy;                                              //  memory of all displaced pixels
float       WarpAmem[4][100];                                              //  undo memory, last 100 warps
int         NWarpA;                                                        //  WarpA mem count

void  WarpA_warpfunc(float wdx, float wdy, float wdw, float wdh, int acc);


void m_warp_area(GtkWidget *, cchar *)
{
   int      WarpA_dialog_event(zdialog *zd, cchar *event);
   
   cchar  *WarpA_message = ZTX(
             " Select an area to warp using select area function. \n"
             " Press [start warp] and pull area with mouse. \n"
             " Make multiple mouse pulls until satisfied. \n"
             " When finished, select another area or press [done]."); 
   
   int         px, py, ii;

   zfuncs::F1_help_topic = "warp_area";

   if (! edit_setup("warp_area",0,2)) return;                        //  setup edit: no preview

   zdedit = zdialog_new(ZTX("Warp Image (area)"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"label","lab1","dialog",WarpA_message,"space=5");
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"space=10");
   zdialog_add_widget(zdedit,"button","swarp","hb1",ZTX("start warp"),"space=5");
   zdialog_add_widget(zdedit,"button","undlast","hb1",Bundolast,"space=5");
   zdialog_add_widget(zdedit,"button","undall","hb1",Bundoall,"space=5");

   zdialog_run(zdedit,WarpA_dialog_event);                                 //  run dialog

   WarpAx = (float *) zmalloc(E3ww * E3hh * sizeof(float),"warp.ax");      //  get memory for pixel displacements
   WarpAy = (float *) zmalloc(E3ww * E3hh * sizeof(float),"warp.ay");
   
   NWarpA = 0;                                                             //  no warp data

   for (py = 0; py < E3hh; py++)                                           //  no pixel displacements
   for (px = 0; px < E3ww; px++)
   {
      ii = py * E3ww + px;
      WarpAx[ii] = WarpAy[ii] = 0.0;
   }
   
   return;
}


//  dialog event and completion callback function

int WarpA_dialog_event(zdialog * zd, cchar *event)
{
   void     WarpA_mousefunc(void);

   int         px, py, ii;
   float       wdx, wdy, wdw, wdh;

   if (zd->zstat)                                                          //  dialog complete
   {
      if (NWarpA) Fmodified = 1;
      else Fmodified = 0;

      if (zd->zstat == 1) edit_done();
      else edit_cancel();

      mouseCBfunc = 0;                                                     //  disconnect mouse
      Mcapture = 0;
      set_cursor(0);                                                       //  restore normal cursor

      zfree(WarpAx);                                                       //  release undo memory
      zfree(WarpAy);
      return 0;
   }
   
   if (strEqu(event,"undo")) edit_undo();                                  //  v.10.2
   if (strEqu(event,"redo")) edit_redo();                                  //  v.10.3

   if (strEqu(event,"swarp"))                                              //  start warp
   {
      if (! sa_active) {                                                   //  no select area active
      zmessageACK(MWIN,ZTX("Select area first"));
         return 0;
      }

      sa_edgecalc();                                                       //  calculate area edge distances

      mouseCBfunc = WarpA_mousefunc;                                       //  connect mouse function
      Mcapture++;
      set_cursor(dragcursor);                                              //  set drag cursor
   }

   if (strEqu(event,"undlast")) {
      if (NWarpA) {                                                        //  undo most recent warp
         ii = --NWarpA;
         wdx = WarpAmem[0][ii];
         wdy = WarpAmem[1][ii];
         wdw = WarpAmem[2][ii];
         wdh = WarpAmem[3][ii];
         WarpA_warpfunc(wdx,wdy,-wdw,-wdh,0);                              //  unwarp image
         WarpA_warpfunc(wdx,wdy,-wdw,-wdh,1);                              //  unwarp memory
      }
   }

   if (strEqu(event,"undall"))                                             //  undo all warps
   {
      edit_reset();                                                        //  v.10.3

      for (py = 0; py < E3hh; py++)                                        //  reset pixel displacements
      for (px = 0; px < E3ww; px++)
      {
         ii = py * E3ww + px;
         WarpAx[ii] = WarpAy[ii] = 0.0;
      }

      NWarpA = 0;                                                          //  erase undo memory
      mwpaint2(); 
   }

   return 1;
}


//  warp mouse function

void  WarpA_mousefunc(void)
{
   static float   wdx, wdy, wdw, wdh;
   static int     ii, warped = 0;

   if (Mxdrag || Mydrag)                                                   //  mouse drag underway
   {
      wdx = Mxdown;                                                        //  drag origin, image coordinates
      wdy = Mydown;
      wdw = Mxdrag - Mxdown;                                               //  drag increment
      wdh = Mydrag - Mydown;
      WarpA_warpfunc(wdx,wdy,wdw,wdh,0);                                   //  warp image
      warped = 1;
      return;
   }
   
   else if (warped) 
   {
      warped = 0;
      WarpA_warpfunc(wdx,wdy,wdw,wdh,1);                                   //  drag done, add to warp memory

      if (NWarpA == 100)                                                   //  if full, throw away oldest
      {
         NWarpA = 99;
         for (ii = 0; ii < NWarpA; ii++)
         {
            WarpAmem[0][ii] = WarpAmem[0][ii+1];
            WarpAmem[1][ii] = WarpAmem[1][ii+1];
            WarpAmem[2][ii] = WarpAmem[2][ii+1];
            WarpAmem[3][ii] = WarpAmem[3][ii+1];
         }
      }

      ii = NWarpA;
      WarpAmem[0][ii] = wdx;                                               //  save warp for undo
      WarpAmem[1][ii] = wdy;
      WarpAmem[2][ii] = wdw;
      WarpAmem[3][ii] = wdh;
      NWarpA++;
   }
   
   return;
}


//  warp image and accumulate warp memory

void  WarpA_warpfunc(float wdx, float wdy, float wdw, float wdh, int acc)
{
   int            ii, jj, px, py, ww, hh, vstat;
   double         ddx, ddy, dpe, dpm, mag, dispx, dispy;
   uint16         vpix[3], *pix3;

   edit_zapredo();                                                         //  delete redo copy    v.10.3
   
   for (py = sa_miny; py <= sa_maxy; py++)                                 //  loop all pixels in area   v.10.11
   for (px = sa_minx; px <= sa_maxx; px++)
   {
      ii = py * Fww + px;
      dpe = sa_pixisin[ii];                                                //  distance from area edge
      if (! dpe) continue;

      ddx = (px - wdx);
      ddy = (py - wdy);
      dpm = sqrt(ddx*ddx + ddy*ddy);                                       //  distance from drag origin

      if (dpm < 1.0) mag = 1.0;                                            //  revised   v.10.8
      else  mag = 1.0 - dpm / (dpm + dpe);
      mag = mag * mag;

      dispx = -wdw * mag;                                                  //  warp = drag * magnification
      dispy = -wdh * mag;
      
      jj = py * E3ww + px;

      if (acc) {                                                           //  mouse drag done,
         WarpAx[jj] += dispx;                                              //    accumulate warp memory
         WarpAy[jj] += dispy;
         continue;
      }

      dispx += WarpAx[jj];                                                 //  add this warp to prior
      dispy += WarpAy[jj];

      vstat = vpixel(E1pxm16,px+dispx,py+dispy,vpix);                      //  input virtual pixel
      if (vstat) {
         pix3 = PXMpix(E3pxm16,px,py);                                     //  output pixel
         pix3[0] = vpix[0];
         pix3[1] = vpix[1];
         pix3[2] = vpix[2];
      }
   }

   ww = sa_maxx - sa_minx;                                                 //  update window    v.10.11
   hh = sa_maxy - sa_miny;
   mwpaint3(sa_minx,sa_miny,ww,hh);

   Fmodified = 1;                                                          //  v.10.2
   return;
}


/**************************************************************************/

//  warp/distort whole image with a curved transform
//  fix perspective problems (e.g. curved walls, leaning buildings)

float       *WarpCx, *WarpCy;                                              //  memory of all dragged pixels
float       WarpCmem[4][100];                                              //  undo memory, last 100 drags
int         NWarpC;                                                        //  WarpCmem count
int         WarpCdrag;
int         WarpCww, WarpChh;

void  WarpC_warpfunc(float wdx, float wdy, float wdw, float wdh, int acc);


void m_warp_curved(GtkWidget *, cchar *)
{
   int      WarpC_dialog_event(zdialog *zd, cchar *event);
   void     WarpC_mousefunc(void);

   cchar  *WarpC_message = ZTX(
             " Pull an image position using the mouse. \n"
             " Make multiple mouse pulls until satisfied. \n"
             " When finished, press [done]."); 
   
   int         px, py, ii;

   zfuncs::F1_help_topic = "warp_curved";

   if (! edit_setup("warp_curved",1,0)) return;                      //  setup edit: use preview

   zdedit = zdialog_new(ZTX("Warp Image (curved)"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"label","lab1","dialog",WarpC_message,"space=5");
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"space=10");
   zdialog_add_widget(zdedit,"button","undlast","hb1",Bundolast,"space=5");
   zdialog_add_widget(zdedit,"button","undall","hb1",Bundoall,"space=5");

   zdialog_run(zdedit,WarpC_dialog_event);                                 //  run dialog

   NWarpC = WarpCdrag = 0;                                                 //  no drag data

   WarpCx = (float *) zmalloc(E3ww * E3hh * sizeof(float),"warp.ix");      //  get memory for pixel displacements
   WarpCy = (float *) zmalloc(E3ww * E3hh * sizeof(float),"warp.iy");
   
   for (py = 0; py < E3hh; py++)                                           //  no pixel displacements
   for (px = 0; px < E3ww; px++)
   {
      ii = py * E3ww + px;
      WarpCx[ii] = WarpCy[ii] = 0.0;
   }
   
   WarpCww = E3ww;                                                         //  preview dimensions
   WarpChh = E3hh;
   
   mouseCBfunc = WarpC_mousefunc;                                          //  connect mouse function
   Mcapture++;
   set_cursor(dragcursor);                                                 //  set drag cursor

   return;
}


//  dialog event and completion callback function

int WarpC_dialog_event(zdialog * zd, cchar *event)
{
   int         px, py, ii;
   float       wdx, wdy, wdw, wdh;
   int         fpx, fpy, epx, epy, vstat;
   double      scale, dispx, dispy;
   uint16      vpix[3], *pix3;

   if (zd->zstat) goto complete;

   if (strEqu(event,"undo")) edit_undo();                                  //  v.10.2
   if (strEqu(event,"redo")) edit_redo();                                  //  v.10.3

   if (strEqu(event,"undlast")) 
   {
      if (NWarpC == 1) event = "undall";
      else if (NWarpC) {                                                   //  undo most recent drag
         ii = --NWarpC;
         wdx = WarpCmem[0][ii];
         wdy = WarpCmem[1][ii];
         wdw = WarpCmem[2][ii];
         wdh = WarpCmem[3][ii];
         WarpC_warpfunc(wdx,wdy,-wdw,-wdh,0);                              //  undrag image
         WarpC_warpfunc(wdx,wdy,-wdw,-wdh,1);                              //  undrag memory
      }
   }

   if (strEqu(event,"undall"))                                             //  undo all drags
   {
      NWarpC = 0;                                                          //  erase undo memory

      for (py = 0; py < E3hh; py++)                                        //  reset pixel displacements
      for (px = 0; px < E3ww; px++)
      {
         ii = py * E3ww + px;
         WarpCx[ii] = WarpCy[ii] = 0.0;
      }
      edit_reset();                                                        //  restore image 1  v.10.3
   }

   return 1;

complete:

   if (zd->zstat != 1) edit_cancel();
   else if (NWarpC == 0) edit_cancel();
   else 
   {
      edit_fullsize();                                                     //  get full-size E1/E3

      scale = 1.0 * (E3ww + E3hh) / (WarpCww + WarpChh);

      for (fpy = 0; fpy < E3hh; fpy++)                                     //  scale net pixel displacements
      for (fpx = 0; fpx < E3ww; fpx++)                                     //    to full image size
      {
         epx = WarpCww * fpx / E3ww;
         epy = WarpChh * fpy / E3hh;
         ii = epy * WarpCww + epx;
         dispx = WarpCx[ii] * scale;
         dispy = WarpCy[ii] * scale;

         vstat = vpixel(E1pxm16,fpx+dispx,fpy+dispy,vpix);                 //  input virtual pixel
         pix3 = PXMpix(E3pxm16,fpx,fpy);                                   //  output pixel
         if (vstat) {
            pix3[0] = vpix[0];
            pix3[1] = vpix[1];
            pix3[2] = vpix[2];
         }
         else pix3[0] = pix3[1] = pix3[2] = 0;
      }

      edit_done();
   }

   mouseCBfunc = 0;                                                        //  disconnect mouse
   Mcapture = 0;
   set_cursor(0);                                                          //  restore normal cursor

   zfree(WarpCx);                                                          //  release memory
   zfree(WarpCy);
   return 0;
}


//  WarpC mouse function

void  WarpC_mousefunc(void)
{
   static float   wdx, wdy, wdw, wdh;
   int            ii;

   if (Mxdrag || Mydrag)                                                   //  mouse drag underway
   {
      wdx = Mxdown;                                                        //  drag origin, window coordinates
      wdy = Mydown;
      wdw = Mxdrag - Mxdown;                                               //  drag increment
      wdh = Mydrag - Mydown;
      WarpC_warpfunc(wdx,wdy,wdw,wdh,0);                                   //  drag image
      WarpCdrag = 1;
      return;
   }
   
   else if (WarpCdrag) 
   {
      WarpCdrag = 0;
      WarpC_warpfunc(wdx,wdy,wdw,wdh,1);                                   //  drag done, add to memory

      if (NWarpC == 100)                                                   //  if full, throw away oldest
      {
         NWarpC = 99;
         for (ii = 0; ii < NWarpC; ii++)
         {
            WarpCmem[0][ii] = WarpCmem[0][ii+1];
            WarpCmem[1][ii] = WarpCmem[1][ii+1];
            WarpCmem[2][ii] = WarpCmem[2][ii+1];
            WarpCmem[3][ii] = WarpCmem[3][ii+1];
         }
      }

      ii = NWarpC;
      WarpCmem[0][ii] = wdx;                                               //  save drag for undo
      WarpCmem[1][ii] = wdy;
      WarpCmem[2][ii] = wdw;
      WarpCmem[3][ii] = wdh;
      NWarpC++;
   }
   
   return;
}


//  warp image and accumulate warp memory
//  mouse at (mx,my) is moved (mw,mh) pixels

void  WarpC_warpfunc(float mx, float my, float mw, float mh, int acc)
{
   int         ii, px, py, vstat;
   double      mag, dispx, dispy;
   double      d1, d2, d3, d4;
   uint16      vpix[3], *pix3;
   
   edit_zapredo();                                                         //  delete redo copy    v.10.3
   
   d1 = (mx-0) * (mx-0) + (my-0) * (my-0);                                 //  distance, mouse to 4 corners
   d2 = (E3ww-mx) * (E3ww-mx) + (my-0) * (my-0);
   d3 = (E3ww-mx) * (E3ww-mx) + (E3hh-my) * (E3hh-my);
   d4 = (mx-0) * (mx-0) + (E3hh-my) * (E3hh-my);
   
   if (d2 > d1) d1 = d2;                                                   //  d1 = greatest       v.10.11
   if (d3 > d1) d1 = d3;
   if (d4 > d1) d1 = d4;

   for (py = 0; py < E3hh; py++)                                           //  process all pixels
   for (px = 0; px < E3ww; px++)
   {
      d2 = (px-mx)*(px-mx) + (py-my)*(py-my);
      mag = (1.0 - d2 / d1);
      mag = mag * mag;                                                     //  faster than pow(mag,16);
      mag = mag * mag;
      //mag = mag * mag;

      dispx = -mw * mag;                                                   //  displacement = drag * mag
      dispy = -mh * mag;
      
      ii = py * E3ww + px;

      if (acc) {                                                           //  drag done, accumulate drag sum
         WarpCx[ii] += dispx;
         WarpCy[ii] += dispy;
         continue;
      }

      dispx += WarpCx[ii];                                                 //  add this drag to prior sum
      dispy += WarpCy[ii];

      vstat = vpixel(E1pxm16,px+dispx,py+dispy,vpix);                      //  input virtual pixel
      pix3 = PXMpix(E3pxm16,px,py);                                        //  output pixel
      if (vstat) {
         pix3[0] = vpix[0];
         pix3[1] = vpix[1];
         pix3[2] = vpix[2];
      }
      else pix3[0] = pix3[1] = pix3[2] = 0;
   }

   Fmodified = 1;
   mwpaint2();                                                             //  update window
   return;
}


/**************************************************************************/

//  warp/distort whole image with a linear transform
//  fix perspective problems (e.g. curved walls, leaning buildings)

float       *WarpLx, *WarpLy;                                              //  memory of all dragged pixels
float       WarpLmem[4][100];                                              //  undo memory, last 100 drags
int         NWarpL;                                                        //  WarpLmem count
int         WarpLdrag;
int         WarpLww, WarpLhh;

void  WarpL_warpfunc(float wdx, float wdy, float wdw, float wdh, int acc);


void m_warp_linear(GtkWidget *, cchar *)                                   //  new v.10.11
{
   int      WarpL_dialog_event(zdialog *zd, cchar *event);
   void     WarpL_mousefunc(void);

   cchar  *WarpL_message = ZTX(
             " Pull an image position using the mouse. \n"
             " Make multiple mouse pulls until satisfied. \n"
             " When finished, press [done]."); 
   
   int         px, py, ii;

   zfuncs::F1_help_topic = "warp_linear";

   if (! edit_setup("warp_linear",1,0)) return;                            //  setup edit: use preview

   zdedit = zdialog_new(ZTX("Warp Image (linear)"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"label","lab1","dialog",WarpL_message,"space=5");
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"space=10");
   zdialog_add_widget(zdedit,"button","undlast","hb1",Bundolast,"space=5");
   zdialog_add_widget(zdedit,"button","undall","hb1",Bundoall,"space=5");

   zdialog_run(zdedit,WarpL_dialog_event);                                 //  run dialog

   NWarpL = WarpLdrag = 0;                                                 //  no drag data

   WarpLx = (float *) zmalloc(E3ww * E3hh * sizeof(float),"warp.ix");      //  get memory for pixel displacements
   WarpLy = (float *) zmalloc(E3ww * E3hh * sizeof(float),"warp.iy");
   
   for (py = 0; py < E3hh; py++)                                           //  no pixel displacements
   for (px = 0; px < E3ww; px++)
   {
      ii = py * E3ww + px;
      WarpLx[ii] = WarpLy[ii] = 0.0;
   }
   
   WarpLww = E3ww;                                                         //  preview dimensions
   WarpLhh = E3hh;
   
   mouseCBfunc = WarpL_mousefunc;                                          //  connect mouse function
   Mcapture++;
   set_cursor(dragcursor);                                                 //  set drag cursor

   return;
}


//  dialog event and completion callback function

int WarpL_dialog_event(zdialog * zd, cchar *event)
{
   int         px, py, ii;
   float       wdx, wdy, wdw, wdh;
   int         fpx, fpy, epx, epy, vstat;
   double      scale, dispx, dispy;
   uint16      vpix[3], *pix3;

   if (zd->zstat) goto complete;

   if (strEqu(event,"undo")) edit_undo();                                  //  v.10.2
   if (strEqu(event,"redo")) edit_redo();                                  //  v.10.3

   if (strEqu(event,"undlast")) 
   {
      if (NWarpL == 1) event = "undall";
      else if (NWarpL) {                                                   //  undo most recent drag
         ii = --NWarpL;
         wdx = WarpLmem[0][ii];
         wdy = WarpLmem[1][ii];
         wdw = WarpLmem[2][ii];
         wdh = WarpLmem[3][ii];
         WarpL_warpfunc(wdx,wdy,-wdw,-wdh,0);                              //  undrag image
         WarpL_warpfunc(wdx,wdy,-wdw,-wdh,1);                              //  undrag memory
      }
   }

   if (strEqu(event,"undall"))                                             //  undo all drags
   {
      NWarpL = 0;                                                          //  erase undo memory

      for (py = 0; py < E3hh; py++)                                        //  reset pixel displacements
      for (px = 0; px < E3ww; px++)
      {
         ii = py * E3ww + px;
         WarpLx[ii] = WarpLy[ii] = 0.0;
      }
      edit_reset();                                                        //  restore image 1  v.10.3
   }

   return 1;

complete:

   if (zd->zstat != 1) edit_cancel();
   else if (NWarpL == 0) edit_cancel();
   else 
   {
      edit_fullsize();                                                     //  get full-size E1/E3

      scale = 1.0 * (E3ww + E3hh) / (WarpLww + WarpLhh);

      for (fpy = 0; fpy < E3hh; fpy++)                                     //  scale net pixel displacements
      for (fpx = 0; fpx < E3ww; fpx++)                                     //    to full image size
      {
         epx = WarpLww * fpx / E3ww;
         epy = WarpLhh * fpy / E3hh;
         ii = epy * WarpLww + epx;
         dispx = WarpLx[ii] * scale;
         dispy = WarpLy[ii] * scale;

         vstat = vpixel(E1pxm16,fpx+dispx,fpy+dispy,vpix);                 //  input virtual pixel
         pix3 = PXMpix(E3pxm16,fpx,fpy);                                   //  output pixel
         if (vstat) {
            pix3[0] = vpix[0];
            pix3[1] = vpix[1];
            pix3[2] = vpix[2];
         }
         else pix3[0] = pix3[1] = pix3[2] = 0;
      }

      edit_done();
   }

   mouseCBfunc = 0;                                                        //  disconnect mouse
   Mcapture = 0;
   set_cursor(0);                                                          //  restore normal cursor

   zfree(WarpLx);                                                          //  release memory
   zfree(WarpLy);
   return 0;
}


//  WarpL mouse function

void  WarpL_mousefunc(void)
{
   static float   wdx, wdy, wdw, wdh;
   int            ii;

   if (Mxdrag || Mydrag)                                                   //  mouse drag underway
   {
      wdx = Mxdown;                                                        //  drag origin, window coordinates
      wdy = Mydown;
      wdw = Mxdrag - Mxdown;                                               //  drag increment
      wdh = Mydrag - Mydown;
      WarpL_warpfunc(wdx,wdy,wdw,wdh,0);                                   //  drag image
      WarpLdrag = 1;
      return;
   }
   
   else if (WarpLdrag) 
   {
      WarpLdrag = 0;
      WarpL_warpfunc(wdx,wdy,wdw,wdh,1);                                   //  drag done, add to memory

      if (NWarpL == 100)                                                   //  if full, throw away oldest
      {
         NWarpL = 99;
         for (ii = 0; ii < NWarpL; ii++)
         {
            WarpLmem[0][ii] = WarpLmem[0][ii+1];
            WarpLmem[1][ii] = WarpLmem[1][ii+1];
            WarpLmem[2][ii] = WarpLmem[2][ii+1];
            WarpLmem[3][ii] = WarpLmem[3][ii+1];
         }
      }

      ii = NWarpL;
      WarpLmem[0][ii] = wdx;                                               //  save drag for undo
      WarpLmem[1][ii] = wdy;
      WarpLmem[2][ii] = wdw;
      WarpLmem[3][ii] = wdh;
      NWarpL++;
   }
   
   return;
}


//  warp image and accumulate warp memory
//  mouse at (mx,my) is moved (mw,mh) pixels

void  WarpL_warpfunc(float mx, float my, float mw, float mh, int acc)
{
   int         ii, px, py, vstat;
   double      mag, dispx, dispy;
   double      d1, d2, d3, d4;
   uint16      vpix[3], *pix3;
   
   edit_zapredo();                                                         //  delete redo copy    v.10.3
   
   d1 = (mx-0) * (mx-0) + (my-0) * (my-0);                                 //  distance, mouse to 4 corners
   d2 = (E3ww-mx) * (E3ww-mx) + (my-0) * (my-0);
   d3 = (E3ww-mx) * (E3ww-mx) + (E3hh-my) * (E3hh-my);
   d4 = (mx-0) * (mx-0) + (E3hh-my) * (E3hh-my);
   
   if (d2 > d1) d1 = d2;                                                   //  d1 = greatest       v.10.11
   if (d3 > d1) d1 = d3;
   if (d4 > d1) d1 = d4;
   
   d1 = sqrt(d1);

   for (py = 0; py < E3hh; py++)                                           //  process all pixels
   for (px = 0; px < E3ww; px++)
   {
      d2 = (px-mx)*(px-mx) + (py-my)*(py-my);
      d2 = sqrt(d2);
      mag = (1.0 - d2 / d1);

      dispx = -mw * mag;                                                   //  displacement = drag * mag
      dispy = -mh * mag;
      
      ii = py * E3ww + px;

      if (acc) {                                                           //  drag done, accumulate drag sum
         WarpLx[ii] += dispx;
         WarpLy[ii] += dispy;
         continue;
      }

      dispx += WarpLx[ii];                                                 //  add this drag to prior sum
      dispy += WarpLy[ii];

      vstat = vpixel(E1pxm16,px+dispx,py+dispy,vpix);                      //  input virtual pixel
      pix3 = PXMpix(E3pxm16,px,py);                                        //  output pixel
      if (vstat) {
         pix3[0] = vpix[0];
         pix3[1] = vpix[1];
         pix3[2] = vpix[2];
      }
      else pix3[0] = pix3[1] = pix3[2] = 0;
   }

   Fmodified = 1;
   mwpaint2();                                                             //  update window
   return;
}


/**************************************************************************/

//  warp/distort whole image using affine transform 
//  (straight lines remain straight)

double      WarpF_old[3][2];                                               //  3 original image points 
double      WarpF_new[3][2];                                               //  corresponding warped points
double      WarpF_coeff[6];                                                //  transform coefficients
double      WarpF_Icoeff[6];                                               //  inverse transform coefficients
int         WarpF_ftf;                                                     //  first time flag

void  WarpF_warpfunc();                                                    //  image warp function
void  WarpF_affine(double po[3][2], double pn[3][2], double coeff[6]);     //  compute affine transform coefficients
void  WarpF_invert(double coeff[6], double Icoeff[6]);                     //  compute reverse transform coefficients


void m_warp_affine(GtkWidget *, cchar *)
{
   int      WarpF_dialog_event(zdialog *zd, cchar *event);
   void     WarpF_mousefunc(void);

   cchar  *WarpF_message = ZTX(
             " Pull on an image corner using the mouse. \n"
             " Make multiple mouse pulls until satisfied. \n"
             " When finished, press [done]."); 
   
   zfuncs::F1_help_topic = "warp_affine";

   if (! edit_setup("warp_affine",1,0)) return;                      //  setup edit: use preview

   zdedit = zdialog_new(ZTX("Warp Image (affine)"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"label","lab1","dialog",WarpF_message,"space=5");

   zdialog_run(zdedit,WarpF_dialog_event);                                 //  run dialog, parallel

   WarpF_ftf = 1;                                                          //  1st warp flag

   mouseCBfunc = WarpF_mousefunc;                                          //  connect mouse function
   Mcapture++;
   set_cursor(dragcursor);                                                 //  set drag cursor

   return;
}


//  dialog event and completion callback function

int WarpF_dialog_event(zdialog *zd, cchar *event)
{
   double      scale;
   int         ww, hh;

   if (zd->zstat) goto complete;

   if (strEqu(event,"undo")) edit_undo();                                  //  v.10.2
   if (strEqu(event,"redo")) edit_redo();                                  //  v.10.3
   return 1;

complete:

   if (zd->zstat != 1) edit_cancel();
   
   else 
   {
      ww = E3ww;                                                           //  preview image dimensions
      hh = E3hh;

      edit_fullsize();                                                     //  get full-size images

      scale = 1.0 * (E3ww + E3hh) / (ww + hh);                             //  preview to full-size scale factor
      
      WarpF_old[0][0] = WarpF_old[0][0] * scale;                           //  re-scale new and old points
      WarpF_old[0][1] = WarpF_old[0][1] * scale;
      WarpF_old[1][0] = WarpF_old[1][0] * scale;
      WarpF_old[1][1] = WarpF_old[1][1] * scale;
      WarpF_old[2][0] = WarpF_old[2][0] * scale;
      WarpF_old[2][1] = WarpF_old[2][1] * scale;

      WarpF_new[0][0] = WarpF_new[0][0] * scale;
      WarpF_new[0][1] = WarpF_new[0][1] * scale;
      WarpF_new[1][0] = WarpF_new[1][0] * scale;
      WarpF_new[1][1] = WarpF_new[1][1] * scale;
      WarpF_new[2][0] = WarpF_new[2][0] * scale;
      WarpF_new[2][1] = WarpF_new[2][1] * scale;
      
      WarpF_warpfunc();                                                    //  warp full-size image
      edit_done();
   }

   mouseCBfunc = 0;                                                        //  disconnect mouse
   Mcapture = 0;
   set_cursor(0);                                                          //  restore normal cursor
   return 0;
}


//  WarpF mouse function

void  WarpF_mousefunc(void)
{
   int      mdx1, mdy1, mdx2, mdy2;
   double   x1o, y1o, x2o, y2o, x3o, y3o;
   double   x1n, y1n, x2n, y2n, x3n, y3n;
   double   a, b, c, d, e, f;
   
   if (Mxdrag + Mydrag == 0) return;

   mdx1 = Mxdown;                                                          //  mouse drag origin
   mdy1 = Mydown;
   mdx2 = Mxdrag;                                                          //  mouse drag position
   mdy2 = Mydrag;

   Mxdown = Mxdrag;                                                        //  reset origin for next time
   Mydown = Mydrag;
   
   x1n = mdx1;                                                             //  point 1 = drag origin
   y1n = mdy1;
   x2n = E3ww - x1n;                                                       //  point 2 = mirror of point1
   y2n = E3hh - y1n;
   x3n = E3ww * (y2n / E3hh);
   y3n = E3hh * (1.0 - (x2n / E3ww));

   if (WarpF_ftf)                                                          //  first warp
   {
      WarpF_ftf = 0;
      x1o = x1n;                                                           //  old = current positions
      y1o = y1n;
      x2o = x2n;
      y2o = y2n;
      x3o = x3n;
      y3o = y3n;
   }
   else
   {
      WarpF_invert(WarpF_coeff,WarpF_Icoeff);                              //  get inverse coefficients
      a = WarpF_Icoeff[0];
      b = WarpF_Icoeff[1];
      c = WarpF_Icoeff[2];
      d = WarpF_Icoeff[3];
      e = WarpF_Icoeff[4];
      f = WarpF_Icoeff[5];
      
      x1o = a * x1n + b * y1n + c;                                         //  compute old from current positions
      y1o = d * x1n + e * y1n + f;
      x2o = a * x2n + b * y2n + c;
      y2o = d * x2n + e * y2n + f;
      x3o = a * x3n + b * y3n + c;
      y3o = d * x3n + e * y3n + f;
   }
      
   WarpF_old[0][0] = x1o;                                                  //  set up 3 old points and corresponding
   WarpF_old[0][1] = y1o;                                                  //    new points for affine translation
   WarpF_old[1][0] = x2o;
   WarpF_old[1][1] = y2o;
   WarpF_old[2][0] = x3o;
   WarpF_old[2][1] = y3o;

   x1n = mdx2;                                                             //  point 1 new position = drag position
   y1n = mdy2;
   x2n = E3ww - x1n;                                                       //  point 2 new = mirror of point1 new
   y2n = E3hh - y1n;

   WarpF_new[0][0] = x1n;                                                  //  3 new points 
   WarpF_new[0][1] = y1n;
   WarpF_new[1][0] = x2n;
   WarpF_new[1][1] = y2n;
   WarpF_new[2][0] = x3n;
   WarpF_new[2][1] = y3n;
   
   WarpF_warpfunc();                                                       //  do the warp

   return;
}


//  warp image and accumulate warp memory

void  WarpF_warpfunc()
{
   double      a, b, c, d, e, f;
   int         px3, py3, vstat;
   double      px1, py1;
   uint16      vpix1[3], *pix3;
   
   edit_zapredo();                                                         //  delete redo copy    v.10.3

   WarpF_affine(WarpF_old, WarpF_new, WarpF_coeff);                        //  get coefficients for forward transform
   WarpF_invert(WarpF_coeff, WarpF_Icoeff);                                //  get coefficients for reverse transform
   
   a = WarpF_Icoeff[0];                                                    //  coefficients to map output pixels
   b = WarpF_Icoeff[1];                                                    //    to corresponding input pixels
   c = WarpF_Icoeff[2];
   d = WarpF_Icoeff[3];
   e = WarpF_Icoeff[4];
   f = WarpF_Icoeff[5];
   
   for (py3 = 0; py3 < E3hh; py3++)                                        //  loop all output pixels
   for (px3 = 0; px3 < E3ww; px3++)
   {
      px1 = a * px3 + b * py3 + c;                                         //  corresponding input pixel
      py1 = d * px3 + e * py3 + f;

      vstat = vpixel(E1pxm16,px1,py1,vpix1);                               //  input virtual pixel
      pix3 = PXMpix(E3pxm16,px3,py3);                                      //  output pixel

      if (vstat) {
         pix3[0] = vpix1[0];
         pix3[1] = vpix1[1];
         pix3[2] = vpix1[2];
      }
      else pix3[0] = pix3[1] = pix3[2] = 0;
   }

   Fmodified = 1;
   mwpaint2();                                                             //  update window
   return;
}


/**************************************************************************

   Compute affine transformation of an image (warp image).

   Given 3 new (warped) positions for 3 image points, derive the 
   coefficients of the translation function to warp the entire image.

   Inputs:
      pold[3][2]  (x,y) coordinates for 3 points in original image
      pnew[3][2]  (x,y) coordinates for same points in warped image
   
   Output: 
      coeff[6]  coefficients of translation function which can be used
                to convert all image points to their warped positions

   If coeff[6] = (a, b, c, d, e, f) then the following formula
   can be used to convert an image point to its warped position:

      Xnew = a * Xold + b * Yold + c
      Ynew = d * Xold + e * Yold + f

***************************************************************************/

void WarpF_affine(double pold[3][2], double pnew[3][2], double coeff[6])
{
   double   x11, y11, x12, y12, x13, y13;                                  //  original points
   double   x21, y21, x22, y22, x23, y23;                                  //  moved points
   double   a, b, c, d, e, f;                                              //  coefficients
   double   A1, A2, B1, B2, C1, C2;
   
   x11 = pold[0][0];
   y11 = pold[0][1];
   x12 = pold[1][0];
   y12 = pold[1][1];
   x13 = pold[2][0];
   y13 = pold[2][1];

   x21 = pnew[0][0];
   y21 = pnew[0][1];
   x22 = pnew[1][0];
   y22 = pnew[1][1];
   x23 = pnew[2][0];
   y23 = pnew[2][1];
   
   A1 = x11 - x12;
   A2 = x12 - x13;
   B1 = y11 - y12;
   B2 = y12 - y13;
   C1 = x21 - x22;
   C2 = x22 - x23;
   
   a = (B1 * C2 - B2 * C1) / (A2 * B1 - A1 * B2);
   b = (A1 * C2 - A2 * C1) / (A1 * B2 - A2 * B1);
   c = x23 - a * x13 - b * y13;
   
   C1 = y21 - y22;
   C2 = y22 - y23;
   
   d = (B1 * C2 - B2 * C1) / (A2 * B1 - A1 * B2);
   e = (A1 * C2 - A2 * C1) / (A1 * B2 - A2 * B1);
   f = y23 - d * x13 - e * y13;
   
   coeff[0] = a;
   coeff[1] = b;
   coeff[2] = c;
   coeff[3] = d;
   coeff[4] = e;
   coeff[5] = f;

   return;
}   


/**************************************************************************

   Invert affine transform

   Input:
      coeff[6]  coefficients of translation function to convert
                image points to their warped positions
   Output: 
      Icoeff[6]  coefficients of translation function to convert
                 warped image points to their original positions

   If Icoeff[6] = (a, b, c, d, e, f) then the following formula can be
      used to translate a warped image point to its original position:

      Xold = a * Xnew + b * Ynew + c
      Yold = d * Xnew + e * Ynew + f

***************************************************************************/

void WarpF_invert(double coeff[6], double Icoeff[6])
{
   double   a, b, c, d, e, f, Z;
   
   a = coeff[0];
   b = coeff[1];
   c = coeff[2];
   d = coeff[3];
   e = coeff[4];
   f = coeff[5];
   
   Z = 1.0 / (a * e - b * d);
   
   Icoeff[0] = e * Z;
   Icoeff[1] = - b * Z;
   Icoeff[2] = Z * (b * f - c * e);
   Icoeff[3] = - d * Z;
   Icoeff[4] = a * Z;
   Icoeff[5] = Z * (c * d - a * f);
   
   return;
}


/**************************************************************************/

//  image color-depth reduction

int      colordep_depth = 16;                                              //  bits per RGB color


void m_colordep(GtkWidget *, cchar *)
{
   int    colordep_dialog_event(zdialog *zd, cchar *event);
   void * colordep_thread(void *);

   cchar  *colmess = ZTX("Set color depth to 1-16 bits");
   
   zfuncs::F1_help_topic = "color_depth";                                  //  v.10.8

   if (! edit_setup("colordepth",1,2)) return;                             //  setup edit: preview

   zdedit = zdialog_new(ZTX("Set Color Depth"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"label","lab1","dialog",colmess,"space=5");
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"space=10");
   zdialog_add_widget(zdedit,"spin","colors","hb1","1|16|1|16","space=5");

   zdialog_run(zdedit,colordep_dialog_event);                              //  run dialog - parallel
   
   colordep_depth = 16;
   start_thread(colordep_thread,0);                                        //  start working thread
   return;
}


//  dialog event and completion callback function

int colordep_dialog_event(zdialog *zd, cchar *event)
{
   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) edit_done();
      else edit_cancel();
      return 0;
   }

   if (strEqu(event,"undo")) edit_undo();                                  //  v.10.2
   if (strEqu(event,"redo")) edit_redo();                                  //  v.10.3
   if (strEqu(event,"blendwidth")) signal_thread();                        //  v.10.3

   if (strEqu(event,"colors")) {
      zdialog_fetch(zd,"colors",colordep_depth);
      signal_thread();
   }
   
   return 0;
}


//  image color depth thread function

void * colordep_thread(void *)
{
   int         ii, px, py, rgb, dist = 0;
   uint16      m1, m2, val1, val3;
   uint16      *pix1, *pix3;
   double      fmag, f1, f2;
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
   
      m1 = 0xFFFF << (16 - colordep_depth);                                //  5 > 1111100000000000
      m2 = 0x8000 >> colordep_depth;                                       //  5 > 0000010000000000
      
      fmag = 65535.0 / m1;

      for (py = 0; py < E3hh; py++)
      for (px = 0; px < E3ww; px++)
      {
         if (sa_active) {                                                  //  select area active
            ii = py * Fww + px;
            dist = sa_pixisin[ii];                                         //  distance from edge
            if (! dist) continue;                                          //  outside pixel
         }

         pix1 = PXMpix(E1pxm16,px,py);                                     //  input pixel
         pix3 = PXMpix(E3pxm16,px,py);                                     //  output pixel
         
         for (rgb = 0; rgb < 3; rgb++)
         {
            val1 = pix1[rgb];
            if (val1 < m1) val3 = (val1 + m2) & m1;
            else val3 = m1;
            val3 = uint(val3 * fmag);

            if (sa_active && dist < sa_blend) {                            //  select area is active,
               f2 = 1.0 * dist / sa_blend;                                 //    blend changes over sa_blend
               f1 = 1.0 - f2;
               val3 = int(f1 * val1 + f2 * val3);
            }

            pix3[rgb] = val3;
         }
      }

      Fmodified = 1;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}


/**************************************************************************/

//  convert image to simulate a drawing

int      draw_contrast;
int      draw_threshold;
int      draw_pixcon;
int      draw_reverse;
double   draw_trfunc[256];
double   draw_pixcon3;
uint8    *draw_pixcon_map = 0;

void m_draw(GtkWidget *, cchar *)
{
   int    draw_dialog_event(zdialog* zd, cchar *event);
   void * draw_thread(void *);

   cchar       *title = ZTX("Simulate Drawing");
   uint16      *pix1, *pix2;
   int         ii, px, py, qx, qy;
   int         red, green, blue, con, maxcon;

   zfuncs::F1_help_topic = "drawing";                                      //  v.10.8

   if (! edit_setup("draw",0,2)) return;                                   //  setup edit: no preview

   draw_pixcon_map = (uint8 *) zmalloc(E1ww*E1hh,"pixconmap");             //  set up pixel contrast map
   memset(draw_pixcon_map,0,E1ww*E1hh);

   for (py = 1; py < E1hh-1; py++)                                         //  scan image pixels
   for (px = 1; px < E1ww-1; px++)
   {
      pix1 = PXMpix(E1pxm16,px,py);                                        //  pixel at (px,py)
      red = pix1[0];                                                       //  pixel RGB levels
      green = pix1[1];
      blue = pix1[2];
      maxcon = 0;

      for (qy = py-1; qy < py+2; qy++)                                     //  loop 3x3 block of neighbor
      for (qx = px-1; qx < px+2; qx++)                                     //    pixels around pix1
      {
         pix2 = PXMpix(E1pxm16,qx,qy);                                     //  find max. contrast with
         con = abs(red-pix2[0]) + abs(green-pix2[1]) + abs(blue-pix2[2]);  //    neighbor pixel
         if (con > maxcon) maxcon = con;
      }

      ii = py * E1ww + px;
      draw_pixcon_map[ii] = (maxcon/3) >> 8;                               //  contrast for (px,py) 0-255
   }

   zdedit = zdialog_new(title,mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"space=10");
   zdialog_add_widget(zdedit,"vbox","vb1","hb1",0,"homog");
   zdialog_add_widget(zdedit,"vbox","vb2","hb1",0,"homog|expand");
   zdialog_add_widget(zdedit,"label","lab1","vb1",ZTX("contrast"));
   zdialog_add_widget(zdedit,"label","lab2","vb1",Bthresh);
   zdialog_add_widget(zdedit,"label","lab3","vb1",ZTX("outlines"));
   zdialog_add_widget(zdedit,"hscale","contrast","vb2","0|100|1|0","expand");
   zdialog_add_widget(zdedit,"hscale","threshold","vb2","0|100|1|0","expand");
   zdialog_add_widget(zdedit,"hscale","pixcon","vb2","0|255|1|0","expand");
   zdialog_add_widget(zdedit,"hbox","hb4","dialog");
   zdialog_add_widget(zdedit,"radio","pencil","hb4",ZTX("pencil"),"space=10");
   zdialog_add_widget(zdedit,"radio","chalk","hb4",ZTX("chalk"),"space=10");

   zdialog_resize(zdedit,300,0);
   zdialog_run(zdedit,draw_dialog_event);                                  //  run dialog - parallel
   
   start_thread(draw_thread,0);                                            //  start working thread
   return;
}


//  dialog event and completion callback function

int draw_dialog_event(zdialog *zd, cchar *event)                           //  draw dialog event function
{
   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) edit_done();
      else edit_cancel();
      zfree(draw_pixcon_map);
      return 0;
   }

   if (strEqu(event,"undo")) edit_undo();                                  //  v.10.2
   if (strEqu(event,"redo")) edit_redo();                                  //  v.10.3
   if (strEqu(event,"blendwidth")) signal_thread();                        //  v.10.3

   if (strcmpv(event,"contrast","threshold","pixcon","chalk",null)) 
   {
      zdialog_fetch(zd,"contrast",draw_contrast);                          //  get slider values
      zdialog_fetch(zd,"threshold",draw_threshold);
      zdialog_fetch(zd,"pixcon",draw_pixcon);
      zdialog_fetch(zd,"chalk",draw_reverse);
      signal_thread();                                                     //  trigger update thread
   }

   return 1;
}


//  thread function - use multiple working threads

void * draw_thread(void *)
{
   void  * draw_wthread(void *arg);

   int         ii;
   double      threshold, contrast, trf;
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      threshold = 0.01 * draw_threshold;                                   //  range 0 to 1
      contrast = 0.01 * draw_contrast;                                     //  range 0 to 1
      
      for (ii = 0; ii < 256; ii++)                                         //  brightness transfer function
      {
         trf = 1.0 - 0.003906 * (256 - ii) * contrast;                     //  ramp-up from 0-1 to 1
         if (ii < 256 * threshold) trf = 0;                                //  0 if below threshold
         draw_trfunc[ii] = trf;
      }
      
      for (ii = 0; ii < Nwt; ii++)                                         //  start worker threads
         start_wt(draw_wthread,&wtnx[ii]);
      wait_wts();                                                          //  wait for completion

      Fmodified = 1;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}


void * draw_wthread(void *arg)                                             //  worker thread function
{
   void  draw_1pix(int px, int py);

   int         index = *((int *) (arg));
   int         px, py;
   double      pixcon;

   pixcon = draw_pixcon / 255.0;                                           //  0-1 linear ramp
   draw_pixcon3 = 255 * pixcon * pixcon * pixcon;                          //  0-255 cubic ramp

   for (py = index+1; py < E1hh-1; py += Nwt)                              //  process all pixels
   for (px = 1; px < E1ww-1; px++)
      draw_1pix(px,py);

   exit_wt();
   return 0;                                                               //  not executed, avoid gcc warning
}


void draw_1pix(int px, int py)                                             //  process one pixel
{
   uint16      *pix1, *pix3;
   int         ii, dist = 0;
   int         bright1, bright2;
   int         red1, green1, blue1;
   int         red3, green3, blue3;
   double      dold, dnew;
   double      pixcon = draw_pixcon3;

   if (sa_active) {                                                        //  select area active
      ii = py * Fww + px;
      dist = sa_pixisin[ii];                                               //  distance from edge
      if (! dist) return;                                                  //  outside pixel
   }
   
   pix1 = PXMpix(E1pxm16,px,py);                                           //  input pixel
   pix3 = PXMpix(E3pxm16,px,py);                                           //  output pixel
      
   red1 = pix1[0];
   green1 = pix1[1];
   blue1 = pix1[2];
   
   bright1 = ((red1 + green1 + blue1) / 3) >> 8;                           //  old brightness  0-255
   bright2 = bright1 * draw_trfunc[bright1];                               //  new brightness  0-255
   
   ii = py * E1ww + px;
   if (draw_pixcon_map[ii] < pixcon) bright2 = 255;

   if (pixcon > 1 && bright2 > draw_threshold) bright2 = 255;              //  empirical !!!

   if (draw_reverse) bright2 = 255 - bright2;                              //  negate if "chalk"

   red3 = green3 = blue3 = bright2 << 8;                                   //  gray scale, new brightness

   if (sa_active && dist < sa_blend) {                                     //  select area is active,
      dnew = 1.0 * dist / sa_blend;
      dold = 1.0 - dnew;
      red3 = dnew * red3 + dold * red1;
      green3 = dnew * green3 + dold * green1;
      blue3 = dnew * blue3 + dold * blue1;
   }

   pix3[0] = red3;
   pix3[1] = green3;
   pix3[2] = blue3;
   
   return;
}


/**************************************************************************/

//  convert image to simulate an embossing

int      emboss_radius, emboss_color;
double   emboss_depth;
double   emboss_kernel[20][20];                                            //  support radius <= 9

void m_emboss(GtkWidget *, cchar *)
{
   int    emboss_dialog_event(zdialog* zd, cchar *event);
   void * emboss_thread(void *);

   cchar  *title = ZTX("Simulate Embossing");

   zfuncs::F1_help_topic = "embossing";                                    //  v.10.8

   if (! edit_setup("emboss",0,2)) return;                                 //  setup edit: no preview

   zdedit = zdialog_new(title,mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"space=10");
   zdialog_add_widget(zdedit,"label","lab1","hb1",Bradius,"space=5");
   zdialog_add_widget(zdedit,"spin","radius","hb1","0|9|1|0");
   zdialog_add_widget(zdedit,"label","lab2","hb1",ZTX("depth"),"space=5");
   zdialog_add_widget(zdedit,"spin","depth","hb1","0|99|1|0");
   zdialog_add_widget(zdedit,"check","color","hb1",ZTX("color"),"space=8");

   zdialog_run(zdedit,emboss_dialog_event);                                //  run dialog - parallel
   
   start_thread(emboss_thread,0);                                          //  start working thread
   return;
}


//  dialog event and completion callback function

int emboss_dialog_event(zdialog *zd, cchar *event)                         //  emboss dialog event function
{
   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) edit_done();
      else edit_cancel();
      return 0;
   }

   if (strEqu(event,"undo")) edit_undo();                                  //  v.10.2
   if (strEqu(event,"redo")) edit_redo();                                  //  v.10.3
   if (strEqu(event,"blendwidth")) signal_thread();                        //  v.10.3

   if (strcmpv(event,"radius","depth","color",null))
   {
      zdialog_fetch(zd,"radius",emboss_radius);                            //  get user inputs
      zdialog_fetch(zd,"depth",emboss_depth);
      zdialog_fetch(zd,"color",emboss_color);
      signal_thread();                                                     //  trigger update thread
   }

   return 1;
}


//  thread function - use multiple working threads

void * emboss_thread(void *)
{
   void  * emboss_wthread(void *arg);

   int         ii, dx, dy, rad;
   double      depth, kern, coeff;
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      rad = emboss_radius;
      depth = emboss_depth;

      coeff = 0.1 * depth / (rad * rad + 1);

      for (dy = -rad; dy <= rad; dy++)                                     //  build kernel with radius and depth
      for (dx = -rad; dx <= rad; dx++)
      {
         kern = coeff * (dx + dy);
         emboss_kernel[dx+rad][dy+rad] = kern;
      }
      
      emboss_kernel[rad][rad] = 1;                                         //  kernel center cell = 1

      for (ii = 0; ii < Nwt; ii++)                                         //  start worker threads
         start_wt(emboss_wthread,&wtnx[ii]);
      wait_wts();                                                          //  wait for comletion

      Fmodified = 1;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}


void * emboss_wthread(void *arg)                                           //  worker thread function
{
   void  emboss_1pix(int px, int py);

   int         index = *((int *) (arg));
   int         px, py;

   for (py = index; py < E1hh; py += Nwt)                                  //  process all pixels
   for (px = 0; px < E1ww; px++)
      emboss_1pix(px,py);

   exit_wt();
   return 0;                                                               //  not executed, avoid gcc warning
}


void emboss_1pix(int px, int py)                                           //  process one pixel
{
   int         ii, dist = 0;
   int         bright1, bright3;
   int         rgb, dx, dy, rad;
   uint16      *pix1, *pix3, *pixN;
   double      sumpix, kern, dold, dnew;
   
   if (sa_active) {                                                        //  select area active
      ii = py * Fww + px;
      dist = sa_pixisin[ii];                                               //  distance from edge
      if (! dist) return;                                                  //  outside pixel
   }

   rad = emboss_radius;

   if (px < rad || py < rad) return;
   if (px > E3ww-rad-1 || py > E3hh-rad-1) return;
   
   pix1 = PXMpix(E1pxm16,px,py);                                           //  input pixel
   pix3 = PXMpix(E3pxm16,px,py);                                           //  output pixel
   
   if (emboss_color) 
   {
      for (rgb = 0; rgb < 3; rgb++)
      {      
         sumpix = 0;
         
         for (dy = -rad; dy <= rad; dy++)                                  //  loop surrounding block of pixels
         for (dx = -rad; dx <= rad; dx++)
         {
            pixN = pix1 + (dy * E1ww + dx) * 3;
            kern = emboss_kernel[dx+rad][dy+rad];
            sumpix += kern * pixN[rgb];
      
            bright1 = pix1[rgb];
            bright3 = sumpix;
            if (bright3 < 0) bright3 = 0;
            if (bright3 > 65535) bright3 = 65535;

            if (sa_active && dist < sa_blend) {                            //  select area is active,
               dnew = 1.0 * dist / sa_blend;                               //    blend changes over sa_blend
               dold = 1.0 - dnew;
               bright3 = dnew * bright3 + dold * bright1;
            }

            pix3[rgb] = bright3;
         }
      }
   }
   
   else                                                                    //  use gray scale
   {
      sumpix = 0;
         
      for (dy = -rad; dy <= rad; dy++)                                     //  loop surrounding block of pixels
      for (dx = -rad; dx <= rad; dx++)
      {
         pixN = pix1 + (dy * E1ww + dx) * 3;
         kern = emboss_kernel[dx+rad][dy+rad];
         sumpix += kern * (pixN[0] + pixN[1] + pixN[2]);
      }
      
      bright1 = 0.3333 * (pix1[0] + pix1[1] + pix1[2]);
      bright3 = 0.3333 * sumpix;
      if (bright3 < 0) bright3 = 0;
      if (bright3 > 65535) bright3 = 65535;
      
      if (sa_active && dist < sa_blend) {                                  //  select area is active,
         dnew = 1.0 * dist / sa_blend;                                     //    blend changes over sa_blend
         dold = 1.0 - dnew;
         bright3 = dnew * bright3 + dold * bright1;
      }

      pix3[0] = pix3[1] = pix3[2] = bright3;
   }

   return;
}


/**************************************************************************/

//  convert image to simulate square tiles

int         tile_size, tile_gap;
uint16      *tile_pixmap = 0;


void m_tiles(GtkWidget *, cchar *)
{
   int    tile_dialog_event(zdialog *zd, cchar *event);
   void * tile_thread(void *);

   zfuncs::F1_help_topic = "tiles";                                        //  v.10.8

   if (! edit_setup("tiles",0,2)) return;                                  //  setup edit: no preview

   zdedit = zdialog_new(ZTX("Set Tile and Gap Size"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"label","labt","hb1",ZTX("tile size"),"space=5");
   zdialog_add_widget(zdedit,"spin","size","hb1","1|99|1|5","space=5");
   zdialog_add_widget(zdedit,"button","apply","hb1",Bapply,"space=10");
   zdialog_add_widget(zdedit,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"label","labg","hb2",ZTX("tile gap"),"space=5");
   zdialog_add_widget(zdedit,"spin","gap","hb2","0|9|1|1","space=5");

   zdialog_run(zdedit,tile_dialog_event);                                  //  start dialog

   tile_size = 5;
   tile_gap = 1;

   tile_pixmap = (uint16 *) zmalloc(E1ww*E1hh*6,"tile.pixmap");            //  set up pixel color map
   memset(tile_pixmap,0,E1ww*E1hh*6);

   start_thread(tile_thread,0);                                            //  start working thread
   return;
}


//  dialog event and completion callback function

int tile_dialog_event(zdialog * zd, cchar *event)
{
   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) edit_done();
      else edit_cancel();
      zfree(tile_pixmap);
      return 0;
   }

   if (strEqu(event,"undo")) edit_undo();                                  //  v.10.2
   if (strEqu(event,"redo")) edit_redo();                                  //  v.10.3
   if (strEqu(event,"blendwidth")) signal_thread();                        //  v.10.3

   if (strNeq(event,"apply")) return 0;

   zdialog_fetch(zd,"size",tile_size);                                     //  get tile size
   zdialog_fetch(zd,"gap",tile_gap);                                       //  get tile gap 

   if (tile_size < 2) {
      edit_reset();                                                        //  restore original image
      return 0;
   }
   
   signal_thread();                                                        //  trigger working thread
   return 1;
}


//  image tiles thread function

void * tile_thread(void *)
{
   int         sg, gg;
   int         sumpix, red, green, blue;
   int         ii, jj, px, py, qx, qy, dist;
   double      dnew, dold;
   uint16      *pix1, *pix3;

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      sg = tile_size + tile_gap;
      gg = tile_gap;

      for (py = 0; py < E1hh; py += sg)                                    //  initz. pixel color map for
      for (px = 0; px < E1ww; px += sg)                                    //    given pixel size
      {
         sumpix = red = green = blue = 0;

         for (qy = py + gg; qy < py + sg; qy++)                            //  get mean color for pixel block
         for (qx = px + gg; qx < px + sg; qx++)
         {
            if (qy > E1hh-1 || qx > E1ww-1) continue;

            pix1 = PXMpix(E1pxm16,qx,qy);
            red += pix1[0];
            green += pix1[1];
            blue += pix1[2];
            sumpix++;
         }

         if (sumpix) {
            red = (red / sumpix);
            green = (green / sumpix);
            blue = (blue / sumpix);
         }
         
         for (qy = py; qy < py + sg; qy++)                                 //  set color for pixels in block
         for (qx = px; qx < px + sg; qx++)
         {
            if (qy > E1hh-1 || qx > E1ww-1) continue;
            
            jj = (qy * E1ww + qx) * 3;

            if (qx-px < gg || qy-py < gg) {
               tile_pixmap[jj] = tile_pixmap[jj+1] = tile_pixmap[jj+2] = 0;
               continue;
            }

            tile_pixmap[jj] = red;
            tile_pixmap[jj+1] = green;
            tile_pixmap[jj+2] = blue;
         }
      }

      if (sa_active)                                                       //  process selected area
      {
         for (ii = 0; ii < Fww * Fhh; ii++)                                //  find pixels in select area
         {                                                                 //  v.9.6
            if (! sa_pixisin[ii]) continue;
            py = ii / Fww;
            px = ii - py * Fww;
            dist = sa_pixisin[ii];
            pix3 = PXMpix(E3pxm16,px,py);
            jj = (py * E3ww + px) * 3;

            if (dist >= sa_blend) {                                        //  blend changes over sa_blend   v.10.2
               pix3[0] = tile_pixmap[jj];
               pix3[1] = tile_pixmap[jj+1];                                //  apply block color to member pixels
               pix3[2] = tile_pixmap[jj+2];
            }
            else {
               dnew = 1.0 * dist / sa_blend;
               dold = 1.0 - dnew;
               pix1 = PXMpix(E1pxm16,px,py);
               pix3[0] = dnew * tile_pixmap[jj] + dold * pix1[0];
               pix3[1] = dnew * tile_pixmap[jj+1] + dold * pix1[1];
               pix3[2] = dnew * tile_pixmap[jj+2] + dold * pix1[2];
            }
         }
      }

      else                                                                 //  process entire image
      {
         for (py = 0; py < E3hh-1; py++)                                   //  loop all image pixels
         for (px = 0; px < E3ww-1; px++)
         {
            pix3 = PXMpix(E3pxm16,px,py);                                  //  target pixel
            jj = (py * E3ww + px) * 3;                                     //  color map for (px,py)
            pix3[0] = tile_pixmap[jj];
            pix3[1] = tile_pixmap[jj+1];
            pix3[2] = tile_pixmap[jj+2];
         }
      }

      Fmodified = 1;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}


/**************************************************************************/

//  convert image to simulate a painting
//  processing a 10 megapixel image needs 140 MB of main memory

namespace paint_names 
{
   int         color_depth;
   int         group_area;
   double      color_match;
   int         borders;

   typedef struct  {
      int16       px, py;
      char        direc;
   }  spixstack;

   int         Nstack;
   spixstack   *pixstack;                                                  //  pixel group search memory
   int         *pixgroup;                                                  //  maps (px,py) to pixel group no.
   int         *groupcount;                                                //  count of pixels in each group

   int         group;
   char        direc;
   uint16      gcolor[3];
}


void m_painting(GtkWidget *, cchar *)
{
   using namespace paint_names;

   int    painting_dialog_event(zdialog *zd, cchar *event);
   void * painting_thread(void *);

   zfuncs::F1_help_topic = "painting";                                     //  v.10.8

   if (! edit_setup("paint",0,2)) return;                                  //  setup edit: no preview

   zdedit = zdialog_new(ZTX("Simulate Painting"),mWin,Bdone,Bcancel,null);

   zdialog_add_widget(zdedit,"hbox","hbcd","dialog",0,"space=1");
   zdialog_add_widget(zdedit,"label","lab1","hbcd",ZTX("color depth"),"space=5");
   zdialog_add_widget(zdedit,"spin","colordepth","hbcd","1|5|1|3","space=5");

   zdialog_add_widget(zdedit,"hbox","hbts","dialog",0,"space=1");
   zdialog_add_widget(zdedit,"label","labts","hbts",ZTX("target group area"),"space=5");
   zdialog_add_widget(zdedit,"spin","grouparea","hbts","0|999|1|100","space=5");

   zdialog_add_widget(zdedit,"hbox","hbcm","dialog",0,"space=1");
   zdialog_add_widget(zdedit,"label","labcm","hbcm",ZTX("req. color match"),"space=5");
   zdialog_add_widget(zdedit,"spin","colormatch","hbcm","0|99|1|50","space=5");

   zdialog_add_widget(zdedit,"hbox","hbbd","dialog",0,"space=1");
   zdialog_add_widget(zdedit,"label","labbd","hbbd",ZTX("borders"),"space=5");
   zdialog_add_widget(zdedit,"check","borders","hbbd",0,"space=2");
   zdialog_add_widget(zdedit,"button","apply","hbbd",Bapply,"space=10");

   zdialog_run(zdedit,painting_dialog_event);                              //  run dialog - parallel

   start_thread(painting_thread,0);                                        //  start working thread
   return;
}


//  dialog event and completion callback function

int painting_dialog_event(zdialog *zd, cchar *event)
{
   using namespace paint_names;

   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) edit_done();
      else edit_cancel();
      return 0;
   }

   if (strEqu(event,"undo")) edit_undo();                                  //  v.10.2
   if (strEqu(event,"redo")) edit_redo();                                  //  v.10.3
   if (strEqu(event,"blendwidth")) signal_thread();                        //  v.10.3

   if (strEqu(event,"apply")) {                                            //  apply user settings
      zdialog_fetch(zd,"colordepth",color_depth);                          //  color depth
      zdialog_fetch(zd,"grouparea",group_area);                            //  target group area (pixels)
      zdialog_fetch(zd,"colormatch",color_match);                          //  req. color match to combine groups
      zdialog_fetch(zd,"borders",borders);                                 //  borders wanted
      color_match = 0.01 * color_match;                                    //  scale 0 to 1

      set_cursor(busycursor);                                              //  v.10.8
      signal_thread();
      wait_thread_idle();
      set_cursor(0);
   }
   
   return 0;
}


//  painting thread function

void * painting_thread(void *)
{
   void  paint_colordepth();
   void  paint_pixgroups();
   void  paint_mergegroups();
   void  paint_paintborders();
   void  paint_blend();

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      paint_colordepth();                                                  //  set new color depth
      paint_pixgroups();                                                   //  group pixel patches of a color
      paint_mergegroups();                                                 //  merge smaller into larger groups
      paint_paintborders();                                                //  add borders around groups
      paint_blend();                                                       //  blend edges of selected area

      Fmodified = 1;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}


//  set the specified color depth, 1-5 bits/color

void paint_colordepth()
{
   using namespace paint_names;

   int            ii, px, py, rgb;
   double         fmag;
   uint16         m1, m2, val1, val3;
   uint16         *pix1, *pix3;

   m1 = 0xFFFF << (16 - color_depth);                                      //  5 > 1111100000000000
   m2 = 0x8000 >> color_depth;                                             //  5 > 0000010000000000

   fmag = 65535.0 / m1;                                                    //  full brightness range

   if (sa_active)                                                          //  process select area
   {
      for (ii = 0; ii < Fww * Fhh; ii++)
      {                                                                    //  v.9.6
         if (! sa_pixisin[ii]) continue;
         py = ii / Fww;
         px = ii - py * Fww;

         pix1 = PXMpix(E1pxm16,px,py);                                     //  input pixel
         pix3 = PXMpix(E3pxm16,px,py);                                     //  output pixel

         for (rgb = 0; rgb < 3; rgb++)
         {
            val1 = pix1[rgb];
            if (val1 < m1) val3 = (val1 + m2) & m1;
            else val3 = m1;
            val3 = uint(val3 * fmag);
            pix3[rgb] = val3;
         }
      }
   }
   
   else                                                                    //  process entire image
   {
      for (py = 0; py < E3hh; py++)                                        //  loop all pixels
      for (px = 0; px < E3ww; px++)
      {
         pix1 = PXMpix(E1pxm16,px,py);                                     //  input pixel
         pix3 = PXMpix(E3pxm16,px,py);                                     //  output pixel
         
         for (rgb = 0; rgb < 3; rgb++)
         {
            val1 = pix1[rgb];
            if (val1 < m1) val3 = (val1 + m2) & m1;
            else val3 = m1;
            val3 = uint(val3 * fmag);
            pix3[rgb] = val3;
         }
      }
   }

   return;
}


//  find all groups of contiguous pixels with the same color

void paint_pixgroups()
{
   using namespace paint_names;

   void  paint_pushpix(int px, int py);

   int            cc1, cc2;
   int            kk, px, py;
   uint16         *pix3;
   
   cc1 = E3ww * E3hh;

   cc2 = cc1 * sizeof(int);
   pixgroup = (int *) zmalloc(cc2,"pixgroup");                             //  maps pixel to assigned group
   memset(pixgroup,0,cc2);
   
   if (sa_active) cc1 = sa_Npixel;

   cc2 = cc1 * sizeof(spixstack);
   pixstack = (spixstack *) zmalloc(cc2,"pixstack");                       //  memory stack for pixel search
   memset(pixstack,0,cc2);
   
   cc2 = cc1 * sizeof(int);
   groupcount = (int *) zmalloc(cc2,"groupcount");                         //  counts pixels per group
   memset(groupcount,0,cc2);
   
   group = 0;
   
   for (py = 0; py < E3hh; py++)                                           //  loop all pixels
   for (px = 0; px < E3ww; px++)
   {
      kk = py * E3ww + px;
      if (sa_active && ! sa_pixisin[kk]) continue;
      if (pixgroup[kk]) continue;                                          //  already assigned to group

      pixgroup[kk] = ++group;                                              //  assign next group
      ++groupcount[group];

      pix3 = PXMpix(E3pxm16,px,py);
      gcolor[0] = pix3[0];
      gcolor[1] = pix3[1];
      gcolor[2] = pix3[2];

      pixstack[0].px = px;                                                 //  put pixel into stack with
      pixstack[0].py = py;                                                 //    direction = right
      pixstack[0].direc = 'r';
      Nstack = 1;

      while (Nstack)
      {
         kk = Nstack - 1;                                                  //  get last pixel in stack
         px = pixstack[kk].px;
         py = pixstack[kk].py;
         direc = pixstack[kk].direc;
         
         if (direc == 'x') {
            Nstack--;
            continue;
         }
         
         if (direc == 'r') {                                               //  push next right pixel into stack
            paint_pushpix(px,py);                                          //   if no group assigned and if
            pixstack[kk].direc = 'l';                                      //    same color as group color
            continue;
         }

         if (direc == 'l') {                                               //  or next left pixel
            paint_pushpix(px,py);
            pixstack[kk].direc = 'a';
            continue;
         }

         if (direc == 'a') {                                               //  or next ahead pixel
            paint_pushpix(px,py);
            pixstack[kk].direc = 'x';
            continue;
         }
      }
   }
   
   return;
}      


//  push a pixel into the stack memory if it is not already assigned
//  to another group and it has the same color as the current group

void paint_pushpix(int px, int py)
{
   using namespace paint_names;

   int      kk, ppx, ppy, npx, npy;
   uint16   *pix3;

   if (Nstack > 1) {
      kk = Nstack - 2;                                                     //  get prior pixel in stack
      ppx = pixstack[kk].px;
      ppy = pixstack[kk].py;
   }
   else {
      ppx = px - 1;                                                        //  if only one, assume prior = left
      ppy = py;
   }
   
   if (direc == 'r') {                                                     //  get pixel in direction right
      npx = px + ppy - py;
      npy = py + px - ppx;
   }
   else if (direc == 'l') {                                                //  or left
      npx = px + py - ppy;
      npy = py + ppx - px;
   }
   else if (direc == 'a') {                                                //  or ahead
      npx = px + px - ppx;
      npy = py + py - ppy;
   }
   else npx = npy = -1;                                                    //  stop warning
   
   if (npx < 0 || npx >= E3ww) return;                                     //  pixel off the edge
   if (npy < 0 || npy >= E3hh) return;
   
   kk = npy * E3ww + npx;

   if (sa_active)
      if (! sa_pixisin[kk]) return;                                        //  pixel outside area

   if (pixgroup[kk]) return;                                               //  pixel already assigned

   pix3 = PXMpix(E3pxm16,npx,npy);
   if (pix3[0] != gcolor[0] || pix3[1] != gcolor[1]                        //  not same color as group
                            || pix3[2] != gcolor[2]) return;
   
   pixgroup[kk] = group;                                                   //  assign pixel to group
   ++groupcount[group];

   kk = Nstack++;                                                          //  put pixel into stack
   pixstack[kk].px = npx;
   pixstack[kk].py = npy;
   pixstack[kk].direc = 'r';                                               //  direction = right
   
   return;
}


//  merge small pixel groups into adjacent larger groups with best color match

void paint_mergegroups()
{
   using namespace paint_names;

   int         ii, jj, kk, px, py, npx, npy;
   int         nccc, mcount, group2;
   double      ff = 1.0 / 65536.0;
   double      fred, fgreen, fblue, match;
   int         nnpx[4] = {  0, -1, +1, 0 };
   int         nnpy[4] = { -1, 0,  0, +1 };
   uint16      *pix3, *pixN;

   typedef struct  {
      int         group;
      double      match;
      uint16      pixM[3];
   }  snewgroup;

   snewgroup      *newgroup;
   
   nccc = (group + 1) * sizeof(snewgroup);
   newgroup = (snewgroup *) zmalloc(nccc,"newgroup");
   
   if (sa_active)                                                          //  process select area
   {
      while (true)
      {
         memset(newgroup,0,nccc);

         for (ii = 0; ii < Fww * Fhh; ii++)                                //  find pixels in select area
         {                                                                 //  v.9.6
            if (! sa_pixisin[ii]) continue;
            py = ii / Fww;
            px = ii - py * Fww;

            kk = E3ww * py + px;                                           //  get assigned group
            group = pixgroup[kk];
            if (groupcount[group] >= group_area) continue;                 //  group count large enough

            pix3 = PXMpix(E3pxm16,px,py);

            for (jj = 0; jj < 4; jj++)                                     //  get 4 neighbor pixels
            {
               npx = px + nnpx[jj];
               npy = py + nnpy[jj];

               if (npx < 0 || npx >= E3ww) continue;                       //  off the edge
               if (npy < 0 || npy >= E3hh) continue;

               kk = E3ww * npy + npx;
               if (! sa_pixisin[kk]) continue;                             //  pixel outside area
               if (pixgroup[kk] == group) continue;                        //  already in same group
               
               pixN = PXMpix(E3pxm16,npx,npy);                             //  match color of group neighbor
               fred = ff * abs(pix3[0] - pixN[0]);                         //    to color of group
               fgreen = ff * abs(pix3[1] - pixN[1]);
               fblue = ff * abs(pix3[2] - pixN[2]);
               match = (1.0 - fred) * (1.0 - fgreen) * (1.0 - fblue);      //  color match, 0 to 1.0
               if (match < color_match) continue;

               if (match > newgroup[group].match) {
                  newgroup[group].match = match;                           //  remember best match
                  newgroup[group].group = pixgroup[kk];                    //  and corresp. group no.
                  newgroup[group].pixM[0] = pixN[0];                       //  and corresp. new color
                  newgroup[group].pixM[1] = pixN[1];
                  newgroup[group].pixM[2] = pixN[2];
               }
            }
         }

         mcount = 0;

         for (ii = 0; ii < Fww * Fhh; ii++)                                //  find pixels in select area
         {                                                                 //  v.9.6
            if (! sa_pixisin[ii]) continue;
            py = ii / Fww;
            px = ii - py * Fww;

            kk = E3ww * py + px;
            group = pixgroup[kk];                                          //  test for new group assignment
            group2 = newgroup[group].group;
            if (! group2) continue;
            
            if (groupcount[group] > groupcount[group2]) continue;          //  accept only bigger new group

            pixgroup[kk] = group2;                                         //  make new group assignment
            --groupcount[group];
            ++groupcount[group2];

            pix3 = PXMpix(E3pxm16,px,py);                                  //  make new color assignment
            pix3[0] = newgroup[group].pixM[0];
            pix3[1] = newgroup[group].pixM[1];
            pix3[2] = newgroup[group].pixM[2];

            mcount++;
         }
         
         if (mcount == 0) break;
      }
   }

   else                                                                    //  process entire image
   {
      while (true)
      {
         memset(newgroup,0,nccc);

         for (py = 0; py < E3hh; py++)                                     //  loop all pixels
         for (px = 0; px < E3ww; px++)
         {
            kk = E3ww * py + px;                                           //  get assigned group
            group = pixgroup[kk];
            if (groupcount[group] >= group_area) continue;                 //  group count large enough

            pix3 = PXMpix(E3pxm16,px,py);

            for (jj = 0; jj < 4; jj++)                                     //  get 4 neighbor pixels
            {
               npx = px + nnpx[jj];
               npy = py + nnpy[jj];

               if (npx < 0 || npx >= E3ww) continue;                       //  off the edge
               if (npy < 0 || npy >= E3hh) continue;
               
               kk = E3ww * npy + npx;
               if (pixgroup[kk] == group) continue;                        //  in same group

               pixN = PXMpix(E3pxm16,npx,npy);                             //  match color of group neighbor
               fred = ff * abs(pix3[0] - pixN[0]);                         //    to color of group
               fgreen = ff * abs(pix3[1] - pixN[1]);
               fblue = ff * abs(pix3[2] - pixN[2]);
               match = (1.0 - fred) * (1.0 - fgreen) * (1.0 - fblue);      //  color match, 0 to 1.0
               if (match < color_match) continue;

               if (match > newgroup[group].match) {
                  newgroup[group].match = match;                           //  remember best match
                  newgroup[group].group = pixgroup[kk];                    //  and corresp. group no.
                  newgroup[group].pixM[0] = pixN[0];                       //  and corresp. new color
                  newgroup[group].pixM[1] = pixN[1];
                  newgroup[group].pixM[2] = pixN[2];
               }
            }
         }

         mcount = 0;

         for (py = 0; py < E3hh; py++)                                     //  loop all pixels
         for (px = 0; px < E3ww; px++)
         {
            kk = E3ww * py + px;
            group = pixgroup[kk];                                          //  test for new group assignment
            group2 = newgroup[group].group;
            if (! group2) continue;
            
            if (groupcount[group] > groupcount[group2]) continue;          //  accept only bigger new group

            pixgroup[kk] = group2;                                         //  make new group assignment
            --groupcount[group];
            ++groupcount[group2];

            pix3 = PXMpix(E3pxm16,px,py);                                  //  make new color assignment
            pix3[0] = newgroup[group].pixM[0];
            pix3[1] = newgroup[group].pixM[1];
            pix3[2] = newgroup[group].pixM[2];

            mcount++;
         }
         
         if (mcount == 0) break;
      }
   }

   zfree(pixgroup);
   zfree(pixstack);
   zfree(groupcount);
   zfree(newgroup);

   return;
}


//  paint borders between the groups of contiguous pixels

void paint_paintborders()
{
   using namespace paint_names;

   int            ii, kk, px, py, cc;
   uint16         *pix3, *pixL, *pixA;
   
   if (! borders) return;
   
   cc = E3ww * E3hh;
   char * pixblack = zmalloc(cc,"pixblack");
   memset(pixblack,0,cc);

   if (sa_active)
   {
      for (ii = 0; ii < Fww * Fhh; ii++)                                   //  find pixels in select area
      {                                                                    //  v.9.6
         if (! sa_pixisin[ii]) continue;
         py = ii / Fww;
         px = ii - py * Fww;
         if (px < 1 || py < 1) continue;

         pix3 = PXMpix(E3pxm16,px,py);
         pixL = PXMpix(E3pxm16,px-1,py);
         pixA = PXMpix(E3pxm16,px,py-1);
         
         if (pix3[0] != pixL[0] || pix3[1] != pixL[1] || pix3[2] != pixL[2])
         {
            kk = ii - 1;
            if (pixblack[kk]) continue;
            kk += 1;
            pixblack[kk] = 1;
            continue;
         }

         if (pix3[0] != pixA[0] || pix3[1] != pixA[1] || pix3[2] != pixA[2])
         {
            kk = ii - E3ww;
            if (pixblack[kk]) continue;
            kk += E3ww;
            pixblack[kk] = 1;
         }
      }

      for (ii = 0; ii < Fww * Fhh; ii++)                                   //  find pixels in select area
      {                                                                    //  v.9.6
         if (! sa_pixisin[ii]) continue;
         py = ii / Fww;
         px = ii - py * Fww;
         if (px < 1 || py < 1) continue;

         if (! pixblack[ii]) continue;
         pix3 = PXMpix(E3pxm16,px,py);
         pix3[0] = pix3[1] = pix3[2] = 0;
      }
   }
         
   else
   {
      for (py = 1; py < E3hh; py++)                                        //  loop all pixels
      for (px = 1; px < E3ww; px++)                                        //  omit top and left
      {
         pix3 = PXMpix(E3pxm16,px,py);                                     //  output pixel
         pixL = PXMpix(E3pxm16,px-1,py);                                   //  pixel to left
         pixA = PXMpix(E3pxm16,px,py-1);                                   //  pixel above
         
         if (pix3[0] != pixL[0] || pix3[1] != pixL[1] || pix3[2] != pixL[2])
         {
            kk = E3ww * py + px-1;                                         //  have horiz. transition
            if (pixblack[kk]) continue;
            kk += 1;
            pixblack[kk] = 1;
            continue;
         }

         if (pix3[0] != pixA[0] || pix3[1] != pixA[1] || pix3[2] != pixA[2])
         {
            kk = E3ww * (py-1) + px;                                       //  have vertical transition
            if (pixblack[kk]) continue;
            kk += E3ww;
            pixblack[kk] = 1;
         }
      }

      for (py = 1; py < E3hh; py++)
      for (px = 1; px < E3ww; px++)
      {
         kk = E3ww * py + px;
         if (! pixblack[kk]) continue;
         pix3 = PXMpix(E3pxm16,px,py);
         pix3[0] = pix3[1] = pix3[2] = 0;
      }
   }
   
   zfree(pixblack);
   return;
}


//  blend edges of selected area

void paint_blend()
{
   int         ii, px, py, rgb, dist;
   uint16      *pix1, *pix3;
   double      f1, f2;
   
   if (sa_active && sa_blend > 0)
   {
      for (ii = 0; ii < Fww * Fhh; ii++)                                   //  find pixels in select area
      {                                                                    //  v.9.6
         dist = sa_pixisin[ii];
         if (! dist || dist >= sa_blend) continue;

         py = ii / Fww;
         px = ii - py * Fww;
         pix1 = PXMpix(E1pxm16,px,py);                                     //  input pixel
         pix3 = PXMpix(E3pxm16,px,py);                                     //  output pixel

         f2 = 1.0 * dist / sa_blend;                                       //  changes over distance sa_blend
         f1 = 1.0 - f2;

         for (rgb = 0; rgb < 3; rgb++)
            pix3[rgb] = int(f1 * pix1[rgb] + f2 * pix3[rgb]);
      }
   }

   return;
}


/**************************************************************************/

//  pixel edit function - edit individual pixels

#define pixed_undomaxmem (200 * mega)                                      //  pixel edit max. memory alloc.   v.10.10.1
#define pixed_undomaxpix (mega)                                            //  pixel edit max. pixel blocks

void  pixed_mousefunc();
void  pixed_dopixels(int px, int py);
void  pixed_saveundo(int px, int py);
void  pixed_undo1();
void  pixed_freeundo();

int      pixed_RGB[3];
int      pixed_mode;
int      pixed_suspend;
int      pixed_radius;
double   pixed_kernel[200][200];                                           //  radius <= 99

int      pixed_undototpix = 0;                                             //  total undo pixel blocks
int      pixed_undototmem = 0;                                             //  total undo memory allocated
int      pixed_undoseq = 0;                                                //  undo sequence no.
char     pixed_undomemmessage[100];                                        //  translated undo memory message

typedef struct {                                                           //  pixel block before edit
   int         seq;                                                        //  undo sequence no.
   uint16      npix;                                                       //  no. pixels in this block
   uint16      px, py;                                                     //  center pixel (radius org.)
   uint16      radius;                                                     //  radius of pixel block
   uint16      pixel[][3];                                                 //  array of pixel[npix][3] 
}  pixed_savepix;

pixed_savepix   **pixed_undopixmem = 0;                                    //  array of *pixed_savepix


void m_pixedit(GtkWidget *, cchar *)
{
   int   pixed_dialog_event(zdialog* zd, cchar *event);

   char        undomemmessage[100];

   zfuncs::F1_help_topic = "edit_pixels";                                  //  v.10.8

   if (! edit_setup("pixedit",0,1)) return;                                //  setup edit: no preview

   strncpy0(pixed_undomemmessage,ZTX("Undo Memory %d%c"),99);              //  translate undo memory message

   zdedit = zdialog_new(ZTX("Edit Pixels"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"hbox","hbc","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"label","labc","hbc",ZTX("color"),"space=8");
   zdialog_add_widget(zdedit,"colorbutt","color","hbc","100|100|100");
   zdialog_add_widget(zdedit,"label","space","hbc",0,"space=10");
   zdialog_add_widget(zdedit,"radio","radio1","hbc",ZTX("pick"),"space=3");
   zdialog_add_widget(zdedit,"radio","radio2","hbc",ZTX("paint"),"space=3");
   zdialog_add_widget(zdedit,"radio","radio3","hbc",ZTX("erase"),"space=3");
   zdialog_add_widget(zdedit,"hbox","hbbri","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"vbox","vbbr1","hbbri",0,"homog|space=3");
   zdialog_add_widget(zdedit,"vbox","vbbr2","hbbri",0,"homog|space=3");
   zdialog_add_widget(zdedit,"label","space","hbbri",0,"space=10");
   zdialog_add_widget(zdedit,"vbox","vbbr3","hbbri",0,"homog|space=3");
   zdialog_add_widget(zdedit,"hbox","hbrad","vbbr1",0,"space=3");
   zdialog_add_widget(zdedit,"label","space","hbrad",0,"expand");
   zdialog_add_widget(zdedit,"label","labbr","hbrad",ZTX("paintbrush radius"));
   zdialog_add_widget(zdedit,"label","labtc","vbbr1",ZTX("transparency center"));
   zdialog_add_widget(zdedit,"label","labte","vbbr1",ZTX("transparency edge"));
   zdialog_add_widget(zdedit,"spin","radius","vbbr2","1|99|1|2");
   zdialog_add_widget(zdedit,"spin","trcent","vbbr2","0|99|1|60");
   zdialog_add_widget(zdedit,"spin","tredge","vbbr2","0|99|1|99");
   zdialog_add_widget(zdedit,"button","susp-resm","vbbr3",Bsuspend);
   zdialog_add_widget(zdedit,"button","undlast","vbbr3",Bundolast);
   zdialog_add_widget(zdedit,"button","undall","vbbr3",Bundoall);
   zdialog_add_widget(zdedit,"label","labmem","dialog");

   zdialog_run(zdedit,pixed_dialog_event);                                 //  run dialog - parallel

   zdialog_send_event(zdedit,"radius");                                    //  get kernel initialized

   snprintf(undomemmessage,99,pixed_undomemmessage,0,'%');                 //  stuff undo memory status
   zdialog_stuff(zdedit,"labmem",undomemmessage);
   
   pixed_RGB[0] = pixed_RGB[1] = pixed_RGB[2] = 100;                       //  initialize color
   
   pixed_mode = 1;                                                         //  mode = pick color
   pixed_suspend = 0;                                                      //  not suspended

   pixed_undopixmem = 0;                                                   //  no undo data
   pixed_undototpix = 0;
   pixed_undototmem = 0;
   pixed_undoseq = 0;

   mouseCBfunc = pixed_mousefunc;                                          //  connect mouse function
   Mcapture = 1;                                                           //  capture mouse clicks
   set_cursor(drawcursor);                                                 //  set draw cursor
   return;
}


//  dialog event and completion callback function

int pixed_dialog_event(zdialog *zd, cchar *event)                          //  pixedit dialog event function
{
   char        color[20];
   cchar       *pp;
   int         radius, dx, dy, brad;
   double      rad, kern, trcent, tredge;
   
   if (zd->zstat) goto complete;

   if (strEqu(event,"undo")) edit_undo();                                  //  v.10.2
   if (strEqu(event,"redo")) edit_redo();                                  //  v.10.3

   zdialog_fetch(zd,"radio1",brad);                                        //  pick
   if (brad) pixed_mode = 1;
   zdialog_fetch(zd,"radio2",brad);                                        //  paint
   if (brad) pixed_mode = 2;
   zdialog_fetch(zd,"radio3",brad);                                        //  erase
   if (brad) pixed_mode = 3;
   
   if (strEqu(event,"color")) 
   {
      zdialog_fetch(zd,"color",color,19);                                  //  get color from color wheel
      pp = strField(color,"|",1);
      if (pp) pixed_RGB[0] = atoi(pp);
      pp = strField(color,"|",2);
      if (pp) pixed_RGB[1] = atoi(pp);
      pp = strField(color,"|",3);
      if (pp) pixed_RGB[2] = atoi(pp);
   }
   
   if (strstr("radius trcent tredge",event))                               //  get new brush attributes
   {
      zdialog_fetch(zd,"radius",radius);                                   //  radius
      zdialog_fetch(zd,"trcent",trcent);                                   //  center transparency
      zdialog_fetch(zd,"tredge",tredge);                                   //  edge transparency

      pixed_radius = radius;
      trcent = 0.01 * trcent;                                              //  scale 0 ... 1
      tredge = 0.01 * tredge;
      tredge = (1 - trcent) * (1 - tredge);
      tredge = 1 - tredge;
      trcent = sqrt(trcent);                                               //  speed up the curve
      tredge = sqrt(tredge);

      for (dy = -radius; dy <= radius; dy++)                               //  build kernel
      for (dx = -radius; dx <= radius; dx++)
      {
         rad = sqrt(dx*dx + dy*dy);
         kern = (radius - rad) / radius;                                   //  1 ... 0 
         kern = kern * (trcent - tredge) + tredge;                         //  trcent ... tredge
         if (rad > radius) kern = 1;
         if (kern < 0) kern = 0;
         if (kern > 1) kern = 1;
         pixed_kernel[dx+radius][dy+radius] = kern;
      }
   }
   
   if (strEqu(event,"undlast"))                                            //  undo last edit (click or drag)
      pixed_undo1();

   if (strEqu(event,"undall")) {                                           //  undo all edits
      edit_reset();                                                        //  v.10.3
      pixed_freeundo();
   }

   if (strEqu(event,"susp-resm"))                                          //  toggle suspend / resume 
   {
      if (pixed_suspend) {
         pixed_suspend = 0;
         mouseCBfunc = pixed_mousefunc;                                    //  connect mouse function
         Mcapture++;
         set_cursor(drawcursor);                                           //  set draw cursor
         zdialog_stuff(zd,"susp-resm",Bsuspend);
      }
      else  {
         pixed_suspend = 1;
         mouseCBfunc = 0;                                                  //  disconnect mouse function
         Mcapture = 0;
         set_cursor(0);                                                    //  restore normal cursor
         zdialog_stuff(zd,"susp-resm",Bresume);
      }
   }

   return 1;

complete:

   if (zd->zstat == 1) edit_done();                                        //  done
   else edit_cancel();                                                     //  cancel or destroy

   paint_toparc(2);                                                        //  remove brush outline
   Mcapture = 0;                                                           //  disconnect mouse
   mouseCBfunc = 0;
   set_cursor(0);                                                          //  restore normal cursor
   pixed_freeundo();                                                       //  free undo memory
   return 0;
}


//  pixel edit mouse function

void pixed_mousefunc()
{
   static int  pmxdown = 0, pmydown = 0;
   int         px, py;
   char        color[20];
   uint16      *ppix3;
   
   toparcx = Mxposn - pixed_radius;                                        //  define brush outline circle
   toparcy = Myposn - pixed_radius;
   toparcw = toparch = 2 * pixed_radius;
   if (pixed_mode == 1) Ftoparc = 0;
   else Ftoparc = 1;
   if (Ftoparc) paint_toparc(3);
   
   if (LMclick)                                                            //  left mouse click
   {
      LMclick = 0;
      px = Mxclick;
      py = Myclick;

      if (pixed_mode == 1)                                                 //  pick new color from image
      {
         ppix3 = PXMpix(E3pxm16,px,py);
         pixed_RGB[0] = ppix3[0] / 256;
         pixed_RGB[1] = ppix3[1] / 256;
         pixed_RGB[2] = ppix3[2] / 256;
         snprintf(color,19,"%d|%d|%d",pixed_RGB[0],pixed_RGB[1],pixed_RGB[2]);
         if (zdedit) zdialog_stuff(zdedit,"color",color);
      }
      else {                                                               //  paint or erase
         pixed_undoseq++;                                                  //  new undo seq. no.
         pixed_saveundo(px,py);                                            //  save for poss. undo
         pixed_dopixels(px,py);                                            //  do 1 block of pixels
      }
   }
   
   if (RMclick) 
   {
      pixed_undo1();                                                       //  undo last draw      v.10.11
      RMclick = 0;
   }
   
   if (Mxdrag || Mydrag)                                                   //  drag in progress
   {
      px = Mxdrag;
      py = Mydrag;
      Mxdrag = Mydrag = 0;

      if (Mxdown != pmxdown || Mydown != pmydown) {                        //  new drag
         pixed_undoseq++;                                                  //  new undo seq. no.
         pmxdown = Mxdown;
         pmydown = Mydown;
      }
      pixed_saveundo(px,py);                                               //  save for poss. undo
      pixed_dopixels(px,py);                                               //  do 1 block of pixels
   }
   
   return;
}


//  paint or erase 1 block of pixels within radius of px, py

void pixed_dopixels(int px, int py)
{
   uint16      *ppix1, *ppix3;
   int         radius, dx, dy, qx, qy, ii, ww, dist;
   int         red, green, blue;
   double      kern;

   edit_zapredo();                                                         //  delete redo copy    v.10.3

   radius = pixed_radius;

   red = 256 * pixed_RGB[0];
   green = 256 * pixed_RGB[1];
   blue = 256 * pixed_RGB[2];

   for (dy = -radius; dy <= radius; dy++)                                  //  loop surrounding block of pixels
   for (dx = -radius; dx <= radius; dx++)
   {
      qx = px + dx;
      qy = py + dy;
      
      if (qx < 0 || qx > E3ww-1) continue;
      if (qy < 0 || qy > E3hh-1) continue;
      
      if (sa_active) {                                                     //  select area active     v.10.11
         ii = qy * E3ww + qx;
         dist = sa_pixisin[ii];
         if (! dist) continue;                                             //  pixel is outside area
      }
      
      kern = pixed_kernel[dx+radius][dy+radius];
      ppix1 = PXMpix(E1pxm16,qx,qy);                                       //  original image pixel
      ppix3 = PXMpix(E3pxm16,qx,qy);                                       //  edited image pixel

      if (pixed_mode == 2)                                                 //  color pixels transparently
      {
         ppix3[0] = (1.0 - kern) * red   + kern * ppix3[0];
         ppix3[1] = (1.0 - kern) * green + kern * ppix3[1];
         ppix3[2] = (1.0 - kern) * blue  + kern * ppix3[2];
         Fmodified = 1;
      }

      if (pixed_mode == 3)                                                 //  restore org. pixels transparently
      {
         ppix3[0] = (1.0 - kern) * ppix1[0] + kern * ppix3[0];
         ppix3[1] = (1.0 - kern) * ppix1[1] + kern * ppix3[1];
         ppix3[2] = (1.0 - kern) * ppix1[2] + kern * ppix3[2];
      }
   }

   px = px - radius - 1;
   py = py - radius - 1;
   ww = 2 * radius + 3;
   mwpaint3(px,py,ww,ww);                                                  //  v.10.11
   paint_toparc(1);
   return;
}


//  save 1 block of pixels for possible undo

void pixed_saveundo(int px, int py)
{
   int            npix, radius, dx, dy;
   uint16         *ppix3;
   pixed_savepix  *ppixsave1;
   char           undomemmessage[100];
   int            mempercent;
   static int     ppercent = 0;

   if (! pixed_undopixmem)                                                 //  first call
   {
      pixed_undopixmem = (pixed_savepix **) zmalloc(pixed_undomaxpix * sizeof(void *),"pixed.undomem");
      pixed_undototpix = 0;
      pixed_undototmem = 0;
   }
   
   if (pixed_undototmem > pixed_undomaxmem) 
   {
      zmessageACK(MWIN,ZTX("Undo memory limit has been reached. \n"
                           "Save work with [done], then resume editing."));
      Mdrag = 0;
      return;
   }

   radius = pixed_radius;
   npix = 0;

   for (dy = -radius; dy <= radius; dy++)                                  //  count pixels in block
   for (dx = -radius; dx <= radius; dx++)
   {
      if (px + dx < 0 || px + dx > E3ww-1) continue;
      if (py + dy < 0 || py + dy > E3hh-1) continue;
      npix++;
   }
   
   ppixsave1 = (pixed_savepix *) zmalloc(npix*6+12,"pixed.pixsave");       //  allocate memory for block
   pixed_undopixmem[pixed_undototpix] = ppixsave1;
   pixed_undototpix += 1;
   pixed_undototmem += npix * 6 + 12;
   
   ppixsave1->seq = pixed_undoseq;                                         //  save pixel block poop
   ppixsave1->npix = npix;
   ppixsave1->px = px;
   ppixsave1->py = py;
   ppixsave1->radius = radius;

   npix = 0;

   for (dy = -radius; dy <= radius; dy++)                                  //  save pixels in block
   for (dx = -radius; dx <= radius; dx++)
   {
      if (px + dx < 0 || px + dx > E3ww-1) continue;
      if (py + dy < 0 || py + dy > E3hh-1) continue;
      ppix3 = PXMpix(E3pxm16,(px+dx),(py+dy));                             //  edited image pixel
      ppixsave1->pixel[npix][0] = ppix3[0];
      ppixsave1->pixel[npix][1] = ppix3[1];
      ppixsave1->pixel[npix][2] = ppix3[2];
      npix++;
   }

   mempercent = int(100.0 * pixed_undototmem / pixed_undomaxmem);          //  update undo memory status
   if (mempercent != ppercent) {
      ppercent = mempercent;
      snprintf(undomemmessage,99,pixed_undomemmessage,mempercent,'%');
      zdialog_stuff(zdedit,"labmem",undomemmessage);
   }

   return;
}


//  undo last undo sequence number

void pixed_undo1()
{
   int            pindex, npix, radius, px, py, dx, dy;
   uint16         *ppix3;
   pixed_savepix  *ppixsave1;
   char           undomemmessage[100];
   int            mempercent;
   
   pindex = pixed_undototpix;
   
   while (pindex > 0)
   {
      --pindex;
      ppixsave1 = pixed_undopixmem[pindex];
      if (ppixsave1->seq != pixed_undoseq) break;
      px = ppixsave1->px;
      py = ppixsave1->py;
      radius = ppixsave1->radius;

      npix = 0;

      for (dy = -radius; dy <= radius; dy++)
      for (dx = -radius; dx <= radius; dx++)
      {
         if (px + dx < 0 || px + dx > E3ww-1) continue;
         if (py + dy < 0 || py + dy > E3hh-1) continue;
         ppix3 = PXMpix(E3pxm16,(px+dx),(py+dy));
         ppix3[0] = ppixsave1->pixel[npix][0];
         ppix3[1] = ppixsave1->pixel[npix][1];
         ppix3[2] = ppixsave1->pixel[npix][2];
         npix++;
      }

      npix = ppixsave1->npix;
      zfree(ppixsave1);
      pixed_undopixmem[pindex] = 0;
      pixed_undototmem -= (npix * 6 + 12);
      --pixed_undototpix;
   }
   
   if (pixed_undoseq > 0) --pixed_undoseq;

   mempercent = int(100.0 * pixed_undototmem / pixed_undomaxmem);          //  update undo memory status
   snprintf(undomemmessage,99,pixed_undomemmessage,mempercent,'%');
   zdialog_stuff(zdedit,"labmem",undomemmessage);

   mwpaint2();
   return;
}


//  free all undo memory

void pixed_freeundo()
{
   int            pindex;
   pixed_savepix  *ppixsave1;
   char           undomemmessage[100];

   pindex = pixed_undototpix;
   
   while (pindex > 0)
   {
      --pindex;
      ppixsave1 = pixed_undopixmem[pindex];
      zfree(ppixsave1);
   }
   
   if (pixed_undopixmem) zfree(pixed_undopixmem);
   pixed_undopixmem = 0;
   
   pixed_undoseq = 0;
   pixed_undototpix = 0;
   pixed_undototmem = 0;

   if (zdedit) {
      snprintf(undomemmessage,99,pixed_undomemmessage,0,'%');              //  undo memory = 0%
      zdialog_stuff(zdedit,"labmem",undomemmessage);
   }

   return;
}


/**************************************************************************
   File scope variables and functions for composite images
      used by HDR, HDF, and Panorama.
   Overhauled for v.10.7
***************************************************************************/

int      cimNF;                                                   //  image count, <= 10
char     *cimFile[10];                                            //  image files
PXM      *cimPXMf[10];                                            //  original images
PXM      *cimPXMs[10];                                            //  alignment images, scaled and curved (pano)
PXM      *cimPXMw[10];                                            //  alignment images, warped

struct cimoffs  {                                                 //  image alignment offsets
   double   xf, yf, tf;                                           //  x, y, theta offsets
   double   wx[4], wy[4];                                         //  x/y corner warps, 0=NW, 1=NE, 2=SE, 3=SW
};
cimoffs  cimOffs[10];                                             //  image alignment data in E3 output image

double   cimScale;                                                //  alignment image size relative to full image
double   cimBlend;                                                //  image blend width at overlap, pixels (pano)
double   cimSearchRange;                                          //  alignment search range, pixels
double   cimSearchStep;                                           //  alignment search step, vpixels
double   cimWarpRange;                                            //  alignment corner warp range, pixels
double   cimWarpStep;                                             //  alignment corner warp step, vpixels
double   cimSampSize;                                             //  pixel sample size
int      cimOv1xlo, cimOv1xhi, cimOv1ylo, cimOv1yhi;              //  rectangle enclosing overlap area,
int      cimOv2xlo, cimOv2xhi, cimOv2ylo, cimOv2yhi;              //    image 1 and image2 coordinates
double   cimRGBmf1[3][65536];                                     //  RGB matching factors for pixel comparisons:
double   cimRGBmf2[3][65536];                                     //  cimRGBmf1[*][pix1[*]] == cimRGBmf2[*][pix2[*]]
char     *cimRedpix = 0;                                          //  maps high-contrast pixels for alignment
int      cimRedImage;                                             //  which image has red pixels
int      cimNsearch;                                              //  alignment search counter
int      cimShowIm1, cimShowIm2;                                  //  two images for cim_show_images()
int      cimShowAll;                                              //  if > 0, show all images
int      cimShrink;                                               //  image shrinkage from pano image curving
int      cimPano;                                                 //  pano mode flag for cim_warp_image()

int      cim_load_files();                                        //  load and check selected files
void     cim_fixblue(PXM *pxm);                                   //  replace blue = 0 pixels with blue = 1
void     cim_scale_image(int im, PXM **);                         //  scale image, 1.0 to cimScale (normally < 1)
void     cim_get_overlap(int im1, int im2, PXM **);               //  get overlap area for images (left, right)
void     cim_match_colors(int im1, int im2, PXM **);              //  match image RGB levels >> match data
void     cim_adjust_colors(PXM *pxm, int fwhich);                 //  adjust RGB levels from match data
void     cim_get_redpix(int im1);                                 //  find high-contrast pixels in overlap area
void     cim_curve_image(int im);                                 //  curve cimPXMs[im] using lens parameters
void     cim_warp_image(int im);                                  //  warp image corners: cimPXMs[im] >> cimPXMw[im]
void     cim_warp_image_pano(int im, int fblend);                 //  pano version, all / left side / blend stripe
void     cim_align_image(int im1, int im2);                       //  align image im2 (pano R) to im1 (pano L)
double   cim_match_images(int im1, int im2);                      //  compare overlapped images and evaluate match
void     cim_show_images(int fnew, int fblend);                   //  combine images >> E3pxm16 >> main window
void     cim_trim();                                              //  cut-off edges where all images do not overlap
void     cim_dump_offsets(cchar *text);                           //  diagnostic tool


//  load image file into pixmaps cimPXMf[*] and check for errors
//  returns 0 if error

int cim_load_files()                                                       //  v.10.7
{
   PXM      *pxm;

   for (int imx = 0; imx < cimNF; imx++)
   {
      if (cimPXMf[imx]) PXM_free(cimPXMf[imx]);
      cimPXMf[imx] = 0;

      pxm = f_load(cimFile[imx],16);                                       //  will diagnose errors
      if (! pxm) return 0;
      cimPXMf[imx] = pxm;
      
      cim_fixblue(pxm);
   }

   return 1;
}


//  replace blue = 0 pixels with blue = 1
//  (blue = 0 reserved for pixels voided by warp or overlay offsets)

void cim_fixblue(PXM *pxm)                                                 //  v.10.7
{
   int      ww, hh, px, py;
   uint16   *pixel;

   ww = pxm->ww;
   hh = pxm->hh;

   for (py = 0; py < hh; py++)
   for (px = 0; px < ww; px++) 
   {
      pixel = PXMpix(pxm,px,py);
      if (! pixel[2]) pixel[2] = 1;
   }

   return;
}


//  scale image from full size) to cimScale (normally < 1.0)

void cim_scale_image(int im, PXM** pxmout)                                 //  v.10.7
{
   int      ww, hh;
   
   ww = cimScale * cimPXMf[im]->ww;
   hh = cimScale * cimPXMf[im]->hh;

   if (pxmout[im]) PXM_free(pxmout[im]);   
   pxmout[im] = PXM_rescale(cimPXMf[im],ww,hh);
   
   cim_fixblue(pxmout[im]);

   return;
}


//  get overlap area for a pair of images im1 and im2
//  outputs are coordinates of overlap area in im1 and in im2

void cim_get_overlap(int im1, int im2, PXM **pxmx)                         //  v.10.7
{
   double      x1, y1, t1, x2, y2, t2;
   double      xoff, yoff, toff, costf, sintf;
   int         ww1, ww2, hh1, hh2, pxM;
   PXM         *pxm1, *pxm2;

   x1 = cimOffs[im1].xf;                                                   //  im1, im2 absolute offsets
   y1 = cimOffs[im1].yf;
   t1 = cimOffs[im1].tf;
   x2 = cimOffs[im2].xf;
   y2 = cimOffs[im2].yf;
   t2 = cimOffs[im2].tf;
   
   xoff = (x2 - x1) * cos(t1) + (y2 - y1) * sin(t1);                       //  offset of im2 relative to im1
   yoff = (y2 - y1) * cos(t1) - (x2 - x1) * sin(t1);
   toff = t2 - t1;

   costf = cos(toff);
   sintf = sin(toff);

   pxm1 = pxmx[im1];
   pxm2 = pxmx[im2];

   ww1 = pxm1->ww;
   hh1 = pxm1->hh;
   ww2 = pxm2->ww;
   hh2 = pxm2->hh;
   
   cimOv1xlo = 0;                                                          //  lowest x overlap
   if (xoff > 0) cimOv1xlo = xoff;

   cimOv1xhi = ww1-1;                                                      //  highest x overlap
   if (cimOv1xhi > xoff + ww2-1) cimOv1xhi = xoff + ww2-1;

   cimOv1ylo = 0;                                                          //  lowest y overlap
   if (yoff > 0) cimOv1ylo = yoff;

   cimOv1yhi = hh1-1;                                                      //  highest y overlap
   if (cimOv1yhi > yoff + hh2-1) cimOv1yhi = yoff + hh2-1;
   
   if (toff < 0) cimOv1xlo -= toff * (cimOv1yhi - cimOv1ylo);              //  reduce for theta offset
   if (toff < 0) cimOv1yhi += toff * (cimOv1xhi - cimOv1xlo);
   if (toff > 0) cimOv1xhi -= toff * (cimOv1yhi - cimOv1ylo);
   if (toff > 0) cimOv1ylo += toff * (cimOv1xhi - cimOv1xlo);
   
   cimOv1xlo += cimShrink + 2;                                             //  account for void areas from
   cimOv1xhi += - cimShrink - 2;                                           //    image shrinkage from 
   cimOv1ylo += cimShrink + 2;                                             //      cim_curve_image()
   cimOv1yhi += - cimShrink - 2;
   
   if (cimBlend > 0 && cimBlend < (cimOv1xhi - cimOv1xlo)) {               //  reduce x range to cimBlend   v.10.8
      pxM = (cimOv1xhi + cimOv1xlo) / 2;
      cimOv1xlo = pxM - cimBlend / 2;
      cimOv1xhi = pxM + cimBlend / 2;
   }
   
   cimOv2xlo = costf * (cimOv1xlo - xoff) + sintf * (cimOv1ylo - yoff);    //  overlap area in im2 coordinates
   cimOv2xhi = costf * (cimOv1xhi - xoff) + sintf * (cimOv1yhi - yoff);
   cimOv2ylo = costf * (cimOv1ylo - yoff) + sintf * (cimOv1xlo - xoff);
   cimOv2yhi = costf * (cimOv1yhi - yoff) + sintf * (cimOv1xhi - xoff);

   if (cimOv1xlo < 0) cimOv1xlo = 0;                                       //  take care of limits
   if (cimOv1ylo < 0) cimOv1ylo = 0;
   if (cimOv2xlo < 0) cimOv2xlo = 0;
   if (cimOv2ylo < 0) cimOv2ylo = 0;
   if (cimOv1xhi > ww1-1) cimOv1xhi = ww1-1;
   if (cimOv1yhi > hh1-1) cimOv1yhi = hh1-1;
   if (cimOv2xhi > ww2-1) cimOv2xhi = ww2-1;
   if (cimOv2yhi > hh2-1) cimOv2yhi = hh2-1;

   return;
}


//  Get the RGB brightness distribution in the overlap area for each image.
//  Compute matching factors to compare pixels within the overlap area.
//  compare  cimRGBmf1[rgb][pix1[rgb]]  to  cimRGBmf2[rgb][pix2[rgb]]

void cim_match_colors(int im1, int im2, PXM **pxmx)                        //  v.10.7
{
   double      Bratios1[3][256];                //  image2/image1 brightness ratio per color per level
   double      Bratios2[3][256];                //  image1/image2 brightness ratio per color per level

   uint16      *pix1, vpix2[3];
   int         vstat2, px1, py1;
   int         ii, jj, rgb;
   int         npix, npix1, npix2, npix3;
   int         brdist1[3][256], brdist2[3][256];
   double      x1, y1, t1, x2, y2, t2;
   double      xoff, yoff, toff, costf, sintf;
   double      px2, py2;
   double      brlev1[3][256], brlev2[3][256];
   double      a1, a2, b1, b2, bratio = 1;
   double      r256 = 1.0 / 256.0;
   PXM         *pxm1, *pxm2;
   
   pxm1 = pxmx[im1];
   pxm2 = pxmx[im2];
   
   x1 = cimOffs[im1].xf;                                                   //  im1, im2 absolute offsets
   y1 = cimOffs[im1].yf;
   t1 = cimOffs[im1].tf;
   x2 = cimOffs[im2].xf;
   y2 = cimOffs[im2].yf;
   t2 = cimOffs[im2].tf;
   
   xoff = (x2 - x1) * cos(t1) + (y2 - y1) * sin(t1);                       //  offset of im2 relative to im1
   yoff = (y2 - y1) * cos(t1) - (x2 - x1) * sin(t1);
   toff = t2 - t1;

   costf = cos(toff);
   sintf = sin(toff);

   for (rgb = 0; rgb < 3; rgb++)                                           //  clear distributions
   for (ii = 0; ii < 256; ii++)
      brdist1[rgb][ii] = brdist2[rgb][ii] = 0;

   npix = 0;

   for (py1 = cimOv1ylo; py1 < cimOv1yhi; py1++)                           //  loop image1/image2 rows
   for (px1 = cimOv1xlo; px1 < cimOv1xhi; px1++)                           //  loop overlapped columns
   {
      pix1 = PXMpix(pxm1,px1,py1);                                         //  image1 pixel
      if (! pix1[2]) continue;                                             //  ignore void pixels

      px2 = costf * (px1 - xoff) + sintf * (py1 - yoff);                   //  corresponding image2 pixel
      py2 = costf * (py1 - yoff) - sintf * (px1 - xoff);
      vstat2 = vpixel(pxm2,px2,py2,vpix2);
      if (! vstat2) continue;                                              //  does not exist

      ++npix;                                                              //  count overlapping pixels
      
      for (rgb = 0; rgb < 3; rgb++)                                        //  accumulate distributions
      {                                                                    //    by color in 256 bins
         ++brdist1[rgb][int(r256*pix1[rgb])];
         ++brdist2[rgb][int(r256*vpix2[rgb])];
      }
   }
   
   npix1 = npix / 256;                                                     //  1/256th of total pixels
   
   for (rgb = 0; rgb < 3; rgb++)                                           //  get brlev1[rgb][N] = mean bright
   for (ii = jj = 0; jj < 256; jj++)                                       //    for Nth group of image1 pixels
   {                                                                       //      for color rgb
      brlev1[rgb][jj] = 0;
      npix2 = npix1;                                                       //  1/256th of total pixels

      while (npix2 > 0 && ii < 256)                                        //  next 1/256th group from distr,
      {
         npix3 = brdist1[rgb][ii];
         if (npix3 == 0) { ++ii; continue; }
         if (npix3 > npix2) npix3 = npix2;
         brlev1[rgb][jj] += ii * npix3;                                    //  brightness * (pixels with)
         brdist1[rgb][ii] -= npix3;
         npix2 -= npix3;
      }

      brlev1[rgb][jj] = brlev1[rgb][jj] / npix1;                           //  mean brightness for group, 0-255
   }

   for (rgb = 0; rgb < 3; rgb++)                                           //  do same for image2
   for (ii = jj = 0; jj < 256; jj++)
   {
      brlev2[rgb][jj] = 0;
      npix2 = npix1;

      while (npix2 > 0 && ii < 256)
      {
         npix3 = brdist2[rgb][ii];
         if (npix3 == 0) { ++ii; continue; }
         if (npix3 > npix2) npix3 = npix2;
         brlev2[rgb][jj] += ii * npix3;
         brdist2[rgb][ii] -= npix3;
         npix2 -= npix3;
      }

      brlev2[rgb][jj] = brlev2[rgb][jj] / npix1;
   }

   for (rgb = 0; rgb < 3; rgb++)                                           //  color
   for (ii = jj = 0; ii < 256; ii++)                                       //  brlev1 brightness, 0 to 255
   {
      if (ii == 0) bratio = 1;
      while (ii > brlev2[rgb][jj] && jj < 256) ++jj;                       //  find matching brlev2 brightness
      a2 = brlev2[rgb][jj];                                                //  next higher value
      b2 = brlev1[rgb][jj];
      if (a2 > 0 && b2 > 0) {
         if (jj > 0) {
            a1 = brlev2[rgb][jj-1];                                        //  next lower value
            b1 = brlev1[rgb][jj-1];
         }
         else   a1 = b1 = 0;
         if (ii == 0)  bratio = b2 / a2;
         else   bratio = (b1 + (ii-a1)/(a2-a1) * (b2-b1)) / ii;            //  interpolate
      }

      if (bratio < 0.2) bratio = 0.2;                                      //  contain outliers
      if (bratio > 5) bratio = 5;
      Bratios2[rgb][ii] = bratio;
   }

   for (rgb = 0; rgb < 3; rgb++)                                           //  color
   for (ii = jj = 0; ii < 256; ii++)                                       //  brlev2 brightness, 0 to 255
   {
      if (ii == 0) bratio = 1;
      while (ii > brlev1[rgb][jj] && jj < 256) ++jj;                       //  find matching brlev1 brightness
      a2 = brlev1[rgb][jj];                                                //  next higher value
      b2 = brlev2[rgb][jj];
      if (a2 > 0 && b2 > 0) {
         if (jj > 0) {
            a1 = brlev1[rgb][jj-1];                                        //  next lower value
            b1 = brlev2[rgb][jj-1];
         }
         else   a1 = b1 = 0;
         if (ii == 0)  bratio = b2 / a2;
         else   bratio = (b1 + (ii-a1)/(a2-a1) * (b2-b1)) / ii;            //  interpolate
      }

      if (bratio < 0.2) bratio = 0.2;                                      //  contain outliers
      if (bratio > 5) bratio = 5;
      Bratios1[rgb][ii] = bratio;
   }
   
   for (ii = 0; ii < 65536; ii++)                                          //  convert brightness ratios into
   {                                                                       //    conversion factors
      jj = ii / 256;

      for (rgb = 0; rgb < 3; rgb++)
      {
         cimRGBmf1[rgb][ii] = sqrt(Bratios1[rgb][jj]) * ii;                //  use sqrt(ratio) so that adjustment
         cimRGBmf2[rgb][ii] = sqrt(Bratios2[rgb][jj]) * ii;                //    can be applied to both images
      }
   }
   
   return;
}


//  Use color match data from cim_match_colors() to 
//    modify images so the colors match.

void cim_adjust_colors(PXM *pxm, int fwhich)                               //  v.10.7
{
   int         ww, hh, px, py;
   int         red, green, blue, max;
   uint16      *pix;
   double      f1;
   
   ww = pxm->ww;
   hh = pxm->hh;

   for (py = 0; py < hh; py++)
   for (px = 0; px < ww; px++)
   {
      pix = PXMpix(pxm,px,py);
      red = pix[0];
      green = pix[1];
      blue = pix[2];
      if (! blue) continue;

      if (fwhich == 1) {
         red = cimRGBmf1[0][red];
         green = cimRGBmf1[1][green];
         blue = cimRGBmf1[2][blue];
      }

      if (fwhich == 2) {
         red = cimRGBmf2[0][red];
         green = cimRGBmf2[1][green];
         blue = cimRGBmf2[2][blue];
      }

      if (red > 65535 || green > 65535 || blue > 65535) {
         max = red;
         if (green > max) max = green;
         if (blue > max) max = blue;
         f1 = 65535.0 / max;
         red = red * f1;
         green = green * f1;
         blue = blue * f1;
      }
      
      if (! blue) blue = 1;                                                //  avoid 0    v.10.7

      pix[0] = red;
      pix[1] = green;
      pix[2] = blue;
   }

   return;
}


//  find pixels of greatest contrast within overlap area
//  flag high-contrast pixels to use in each image compare region

void cim_get_redpix(int im1)                                               //  v.10.7
{
   int         ww, hh, samp, xzone, yzone;
   int         pxL, pxH, pyL, pyH;
   int         px, py, ii, jj, npix;
   int         red1, green1, blue1, red2, green2, blue2, tcon;
   int         ov1xlo, ov1xhi, ov1ylo, ov1yhi;
   int         Hdist[256], Vdist[256], Hmin, Vmin;
   double      s8 = 1.0 / 770.0;
   double      zsamp[16] = { 4, 6, 6, 4, 6, 9, 9, 6, 6, 9, 9, 6, 4, 6, 6, 4 };    //  v.10.8
   uchar       *Hcon, *Vcon;
   uint16      *pix1, *pix2;
   PXM         *pxm;
   
   pxm = cimPXMw[im1];
   ww = pxm->ww;
   hh = pxm->hh;

   if (cimRedpix) zfree(cimRedpix);                                        //  clear prior
   cimRedpix = zmalloc(ww*hh,"cimRedpix");
   memset(cimRedpix,0,ww*hh);

   cimRedImage = im1;                                                      //  image with red pixels
   
   ov1xlo = cimOv1xlo + cimSearchRange;                                    //  stay within x/y search range
   ov1xhi = cimOv1xhi - cimSearchRange;                                    //    so that red pixels persist
   ov1ylo = cimOv1ylo + cimSearchRange;                                    //      over offset changes
   ov1yhi = cimOv1yhi - cimSearchRange;
   
   for (yzone = 0; yzone < 4; yzone++)                                     //  loop 16 zones       v.10.8
   for (xzone = 0; xzone < 4; xzone++)
   {
      pxL = ov1xlo + 0.25 * xzone     * (ov1xhi - ov1xlo);                 //  px and py zone limits
      pxH = ov1xlo + 0.25 * (xzone+1) * (ov1xhi - ov1xlo);
      pyL = ov1ylo + 0.25 * yzone     * (ov1yhi - ov1ylo);
      pyH = ov1ylo + 0.25 * (yzone+1) * (ov1yhi - ov1ylo);
      
      npix = (pxH - pxL) * (pyH - pyL);                                    //  zone pixels
      Hcon = (uchar *) zmalloc(npix,"redpix.hcon");                        //  horizontal pixel contrast 0-255
      Vcon = (uchar *) zmalloc(npix,"redpix.vcon");                        //  vertical pixel contrast 0-255
      
      ii = 4 * yzone + xzone;
      samp = cimSampSize * 0.01 * zsamp[ii];                               //  sample size for zone
      if (samp > 0.1 * npix) samp = 0.1 * npix;                            //  limit to 10% of zone pixels
      
      for (py = pyL; py < pyH; py++)                                       //  scan image pixels in zone
      for (px = pxL; px < pxH; px++)
      {
         ii = (py-pyL) * (pxH-pxL) + (px-pxL);
         Hcon[ii] = Vcon[ii] = 0;                                          //  horiz. = vert. contrast = 0

         if (py < 8 || py > hh-9) continue;                                //  keep away from image edges
         if (px < 8 || px > ww-9) continue;

         pix1 = PXMpix(pxm,px,py-6);                                       //  verify not near void areas
         if (! pix1[2]) continue;
         pix1 = PXMpix(pxm,px+6,py);
         if (! pix1[2]) continue;
         pix1 = PXMpix(pxm,px,py+6);
         if (! pix1[2]) continue;
         pix1 = PXMpix(pxm,px-6,py);
         if (! pix1[2]) continue;

         pix1 = PXMpix(pxm,px,py);                                         //  candidate red pixel
         red1 = pix1[0];
         green1 = pix1[1];
         blue1 = pix1[2];

         pix2 = PXMpix(pxm,px+2,py);                                       //  2 pixels to right
         red2 = pix2[0];
         green2 = pix2[1];
         blue2 = pix2[2];

         tcon = abs(red1-red2) + abs(green1-green2) + abs(blue1-blue2);    //  horizontal contrast
         Hcon[ii] = int(tcon * s8);                                        //  scale  0 - 255

         pix2 = PXMpix(pxm,px,py+2);                                       //  2 pixels below
         red2 = pix2[0];
         green2 = pix2[1];
         blue2 = pix2[2];

         tcon = abs(red1-red2) + abs(green1-green2) + abs(blue1-blue2);    //  vertical contrast
         Vcon[ii] = int(tcon * s8);
      }

      for (ii = 0; ii < 256; ii++) Hdist[ii] = Vdist[ii] = 0;              //  clear contrast distributions

      for (py = pyL; py < pyH; py++)                                       //  scan image pixels
      for (px = pxL; px < pxH; px++)
      {                                                                    //  build contrast distributions
         ii = (py-pyL) * (pxH-pxL) + (px-pxL);
         ++Hdist[Hcon[ii]];
         ++Vdist[Vcon[ii]];
      }
      
      for (npix = 0, ii = 255; ii > 0; ii--)                               //  find minimum contrast needed to get
      {                                                                    //    enough pixels for sample size
         npix += Hdist[ii];                                                //      (horizontal contrast pixels)
         if (npix > samp) break; 
      }
      Hmin = ii; 

      for (npix = 0, ii = 255; ii > 0; ii--)                               //  (verticle contrast pixels)
      {
         npix += Vdist[ii];
         if (npix > samp) break;
      }
      Vmin = ii;

      for (py = pyL; py < pyH; py++)                                       //  scan image pixels
      for (px = pxL; px < pxH; px++)
      {
         ii = (py-pyL) * (pxH-pxL) + (px-pxL);
         jj = py * ww + px;

         if (Hcon[ii] > Hmin) {
            cimRedpix[jj] = 1;                                             //  flag horizontal group of 3
            cimRedpix[jj+1] = 1;
            cimRedpix[jj+2] = 1;
         }

         if (Vcon[ii] > Vmin) {
            cimRedpix[jj] = 1;                                             //  flag verticle group of 3
            cimRedpix[jj+ww] = 1;
            cimRedpix[jj+2*ww] = 1;
         }
      }
      
      zfree(Hcon);
      zfree(Vcon);
   }

   return;
}


//  curve image based on lens parameters (pano)
//  replaces cimPXMs[im] with curved version

void cim_curve_image(int im)                                               //  v.10.7
{
   double      mmx[12] =  { 20, 22, 25,   26,  28,  32,   40,   48,   60,  80,   100, 200 };
   double      mmy[12] =  { 35, 19, 10.1, 8.7, 6.8, 5.26, 3.42, 2.32, 1.7, 1.35, 1.2, 1.1 };

   static double     curve = 0, bow = 0, R = 1;
   static double     prior_mm = 0, lens_curve;

   int         px, py, ww, hh, vstat;
   double      pxs, pys, xdisp, ydisp, xpull, ypull;
   double      ww2, hh2;
   PXM         *pxmin, *pxmout;
   uint16      vpix[3], *pix;

   pxmin = cimPXMs[im];                                                    //  input and output image
   ww = pxmin->ww;
   hh = pxmin->hh;
   
   pxmout = PXM_make(ww,hh,16);

   if (lens_mm != prior_mm) {                                              //  if lens_mm changed,
      prior_mm = lens_mm;                                                  //    recalculate lens_curve
      spline1(12,mmx,mmy);
      lens_curve = spline2(lens_mm);
   }

   curve = lens_curve * 0.01 * ww;                                         //  curve % to pixels
   bow = 0.01 * ww * lens_bow;                                             //  lens_bow % to pixels
   R = 1.0 * hh / ww;
   if (R > 1) curve = curve / R / R;                                       //  adjust for vertical format
   
   ww2 = 0.5 * ww;
   hh2 = 0.5 * hh;

   for (py = 0; py < hh; py++)
   for (px = 0; px < ww; px++)
   {
      xdisp = (px - ww2) / ww2;                                            //  -1 ... 0 ... +1
      ydisp = (py - hh2) / ww2;                                            //  -hh/ww ... 0 ... +hh/ww
      
      xpull = xdisp * xdisp * xdisp;
      ypull = 1.41 * ydisp * xdisp * xdisp;
      pxs = px + curve * xpull;                                            //  apply lens curve factor
      pys = py + curve * ypull;
      pxs -= bow * xdisp * ydisp * ydisp;                                  //  apply lens bow factor
   
      vstat = vpixel(pxmin,pxs,pys,vpix);                                  //  input virtual pixel
      pix = PXMpix(pxmout,px,py);                                          //  output real pixel
      if (vstat) {  
         pix[0] = vpix[0];
         pix[1] = vpix[1];
         pix[2] = vpix[2];
      }
      else pix[0] = pix[1] = pix[2] = 0;                                   //  voided pixels are (0,0,0)
   }

   for (px = 0; px < ww/3; px++) {                                         //  compute image shrinkage
      pix = PXMpix(pxmout,px,hh/2);
      if (pix[2]) break;
   }
   cimShrink = px-1;                                                       //  = 0 if no curve
   
   PXM_free(pxmin);                                                        //  replace input with output PXM
   cimPXMs[im] = pxmout;

   return;
}


//  Warp 4 image corners according to cimOffs[im].wx[ii] and .wy[ii]
//  corner = 0 = NW,  1 = NE,  2 = SE,  3 = SW
//  4 corners move by these pixel amounts and center does not move.
//  input: cimPXMs[im] (flat or curved) output: cimPXMw[im]

namespace cim_warp_image_names {
   PXM         *pxmin, *pxmout;
   double      ww, hh, wwi, hhi;
   double      wx0, wy0, wx1, wy1, wx2, wy2, wx3, wy3;
}

void cim_warp_image(int im)                                                //  caller function     v.10.8
{
   using namespace cim_warp_image_names;

   void * cim_warp_image_wthread(void *arg);
   
   pxmin = cimPXMs[im];                                                    //  input and output pixmaps
   pxmout = cimPXMw[im];
   
   ww = pxmin->ww;
   hh = pxmin->hh;
   wwi = 1.0 / ww;
   hhi = 1.0 / hh;

   wx0 = cimOffs[im].wx[0];                                                //  corner warps
   wy0 = cimOffs[im].wy[0];
   wx1 = cimOffs[im].wx[1];
   wy1 = cimOffs[im].wy[1];
   wx2 = cimOffs[im].wx[2];
   wy2 = cimOffs[im].wy[2];
   wx3 = cimOffs[im].wx[3];
   wy3 = cimOffs[im].wy[3];

   if (! pxmout) {                                                         //  create output if needed
      pxmout = PXM_copy(pxmin);
      cimPXMw[im] = pxmout;
   }
   
   if (pxmout->ww != ww || pxmout->hh != hh) {                             //  or if not my size
      PXM_free(pxmout);
      pxmout = PXM_copy(pxmin);
      cimPXMw[im] = pxmout;
   }
   
   for (int ii = 0; ii < Nwt; ii++)                                        //  start worker threads
      start_wt(cim_warp_image_wthread,&wtnx[ii]);
   wait_wts();                                                             //  wait for completion

   return;
}


void * cim_warp_image_wthread(void *arg)                                   //  worker thread function  v.10.8
{
   using namespace cim_warp_image_names;

   int         index = *((int *) arg);
   int         pxm, pym, vstat;
   uint16      vpix[3], *pixm;
   double      px, py, dx, dy, coeff;
   
   for (pym = index; pym < hh; pym += Nwt)                                 //  loop all pixels for this thread
   for (pxm = 0; pxm < ww; pxm++)
   {
      dx = dy = 0.0;

      coeff = (1.0 - pym * hhi - pxm * wwi);                               //  corner 0  NW
      if (coeff > 0) {
         dx += coeff * wx0;
         dy += coeff * wy0;
      }
      coeff = (1.0 - pym * hhi - (ww - pxm) * wwi);                        //  corner 1  NE
      if (coeff > 0) {
         dx += coeff * wx1;
         dy += coeff * wy1;
      }
      coeff = (1.0 - (hh - pym) * hhi - (ww - pxm) * wwi);                 //  corner 2  SE
      if (coeff > 0) {
         dx += coeff * wx2;
         dy += coeff * wy2;
      }
      coeff = (1.0 - (hh - pym) * hhi - pxm * wwi);                        //  corner 3  SW
      if (coeff > 0) {
         dx += coeff * wx3;
         dy += coeff * wy3;
      }

      px = pxm + dx;                                                       //  source pixel location
      py = pym + dy;

      vstat = vpixel(pxmin,px,py,vpix);                                    //  input virtual pixel
      pixm = PXMpix(pxmout,pxm,pym);                                       //  output real pixel

      if (vstat) {  
         pixm[0] = vpix[0];
         pixm[1] = vpix[1];
         pixm[2] = vpix[2];
      }
      else pixm[0] = pixm[1] = pixm[2] = 0;
   }
   
   exit_wt();
   return 0;                                                               //  not executed, avoid gcc warning
}


//  warp image for pano, left side corners only, reduced warp range
//  input: cimPXMs[im] (curved) 
//  output: cimPXMw[im] (warped)
//  fblend: 0 = process entire image
//          1 = process left half only
//          2 = process blend stripe only

void cim_warp_image_pano(int im, int fblend)                               //  v.10.8
{
   int         ww, hh, ww2, hh2, pxL, pxH;
   int         pxm, pym, vstat;
   uint16      vpix[3], *pixm;
   double      ww2i, hh2i, pxs, pys, xdisp, ydisp;
   double      wx0, wy0, wx3, wy3;
   PXM         *pxmin, *pxmout;

   pxmin = cimPXMs[im];                                                    //  input and output pixmaps
   pxmout = cimPXMw[im];
   
   ww = pxmin->ww;
   hh = pxmin->hh;

   if (! pxmout) {                                                         //  create output if needed
      pxmout = PXM_copy(pxmin);
      cimPXMw[im] = pxmout;
   }
   
   if (pxmout->ww != ww || pxmout->hh != hh) {                             //  or if not my size
      PXM_free(pxmout);
      pxmout = PXM_copy(pxmin);
      cimPXMw[im] = pxmout;
   }
   
   ww2 = ww / 2;
   hh2 = hh / 2;

   ww2i = 1.0 / ww2;                                                       //  v.10.8
   hh2i = 1.0 / hh2;

   wx0 = cimOffs[im].wx[0];                                                //  NW corner warp
   wy0 = cimOffs[im].wy[0];

   wx3 = cimOffs[im].wx[3];                                                //  SW corner warp
   wy3 = cimOffs[im].wy[3];

   pxL = 0;                                                                //  entire image        v.10.8
   pxH = ww;
   
   if (fblend == 1)                                                        //  left half
      pxH = ww2;

   if (fblend == 2) {
      pxL = cimOv2xlo;                                                     //  limit to overlap/blend width
      pxH = cimOv2xhi;
   }
   
   for (pym = 0; pym < hh; pym++)                                          //  loop all output pixels
   for (pxm = pxL; pxm < pxH; pxm++)
   {
      pixm = PXMpix(pxmout,pxm,pym);                                       //  output pixel
      
      xdisp = (pxm - ww2) * ww2i;                                          //  -1 ... 0 ... +1
      ydisp = (pym - hh2) * ww2i;                                          //  -hh/ww ... 0 ... +hh/ww
      
      if (xdisp > 0) {                                                     //  right half, no warp
         pxs = pxm;
         pys = pym;
      }
      else if (ydisp < 0) {                                                //  use NW corner warp
         pxs = pxm + wx0 * xdisp * ydisp;
         pys = pym + wy0 * xdisp * ydisp;
      }
      else {                                                               //  use SW corner warp
         pxs = pxm + wx3 * xdisp * ydisp;
         pys = pym + wy3 * xdisp * ydisp;
      }

      vstat = vpixel(pxmin,pxs,pys,vpix);                                  //  input virtual pixel

      if (vstat) {  
         pixm[0] = vpix[0];
         pixm[1] = vpix[1];
         pixm[2] = vpix[2];
      }
      else pixm[0] = pixm[1] = pixm[2] = 0;
   }

   for (pxm = 1; pxm < ww2; pxm++) {                                       //  compute image shrinkage
      pixm = PXMpix(pxmout,pxm,hh2);                                       //    used by cim_get_overlap()
      if (pixm[2]) break;
   }
   cimShrink = pxm-1;

   return;
}


//  fine-align a pair of images im1 and im2
//  cimPXMw[im2] is aligned with cimPXMw[im1]
//  inputs are cimOffs[im1] and cimOffs[im2] (x/y/t offsets)
//  output is adjusted offsets and corner warp values for im2
//  (im1 must be unwarped)

void cim_align_image(int im1, int im2)                                     //  v.10.7
{
   int         ii, pass;
   double      xyrange, xystep, trange, tstep, wrange, wstep;
   double      xfL, xfH, yfL, yfH, tfL, tfH;
   double      wxL, wxH, wyL, wyH;
   double      match, matchB;
   cimoffs     offsets0, offsetsB;
   
   offsets0 = cimOffs[im2];                                                //  initial offsets
   offsetsB = offsets0;                                                    //  = best offsets so far
   matchB = cim_match_images(im1,im2);                                     //  = best image match level

   cim_show_images(0,0);                                                   //  show with 50/50 blend

   for (pass = 1; pass <=2; pass++)                                        //  main pass and 2nd pass    v.10.8
   {
      xyrange = cimSearchRange;                                            //  x/y search range and step
      xystep = cimSearchStep;   

      trange = xyrange / (cimOv1yhi - cimOv1ylo);                          //  angle range, radians
      tstep = trange * xystep / xyrange;
      
      if (pass == 2) {
         xyrange = xyrange / 4;                                            //  2nd pass, 1/4 range and 1/2 step
         xystep = xystep / 2;                                              //                      v.10.8
         trange = trange / 4;
         tstep = tstep / 2;
      }
      
      //  search x/y/t range for best match

      xfL = cimOffs[im2].xf - xyrange;
      xfH = cimOffs[im2].xf + xyrange + 0.5 * xystep;
      yfL = cimOffs[im2].yf - xyrange;
      yfH = cimOffs[im2].yf + xyrange + 0.5 * xystep;
      tfL = cimOffs[im2].tf - trange;
      tfH = cimOffs[im2].tf + trange + 0.5 * tstep;
      
      for (cimOffs[im2].xf = xfL; cimOffs[im2].xf < xfH; cimOffs[im2].xf += xystep)
      for (cimOffs[im2].yf = yfL; cimOffs[im2].yf < yfH; cimOffs[im2].yf += xystep)
      for (cimOffs[im2].tf = tfL; cimOffs[im2].tf < tfH; cimOffs[im2].tf += tstep)
      {
         match = cim_match_images(im1,im2);                                //  get match level

         if (sigdiff(match,matchB,0.00001) > 0) {
            matchB = match;                                                //  save best match
            offsetsB = cimOffs[im2];
            cim_show_images(0,0);
         }

         sprintf(SB_text,"align: %d  match: %.5f",cimNsearch++,matchB);    //  update status bar
         SBupdate++;
      }
      
      cimOffs[im2] = offsetsB;                                             //  restore best match
      
      //  warp corners and search for best match

      wrange = cimWarpRange;                                               //  corner warp range and step
      wstep = cimWarpStep;
      if (! wrange) continue;

      if (pass == 2) {                                                     //  2nd pass, 1/4 range and 1/2 step
         wrange = wrange / 4;
         wstep = wstep / 2;
      }
      
      if (cimPano)                                                         //  pano mode, left side corners only
      {
         for (ii = 0; ii < 4; ii += 3)                                     //  corners 0 and 3 (NW, SW)
         {
            wxL = cimOffs[im2].wx[ii] - wrange;
            wxH = cimOffs[im2].wx[ii] + wrange + 0.5 * wstep;
            wyL = cimOffs[im2].wy[ii] - wrange;
            wyH = cimOffs[im2].wy[ii] + wrange + 0.5 * wstep;

            for (cimOffs[im2].wx[ii] = wxL; cimOffs[im2].wx[ii] < wxH; cimOffs[im2].wx[ii] += wstep)
            for (cimOffs[im2].wy[ii] = wyL; cimOffs[im2].wy[ii] < wyH; cimOffs[im2].wy[ii] += wstep)
            {
               cim_warp_image_pano(im2,2);                                 //  warp image using 2 corner offsets

               match = cim_match_images(im1,im2);                          //  get match level

               if (sigdiff(match,matchB,0.00001) > 0) {
                  matchB = match;                                          //  save best match
                  offsetsB = cimOffs[im2];
                  cim_show_images(0,0);
               }

               sprintf(SB_text,"warp: %d  match: %.5f",cimNsearch++,matchB);
               SBupdate++;
            }

            cimOffs[im2] = offsetsB;                                       //  restore best match
            cim_warp_image_pano(im2,1);
            cim_show_images(0,0);
         }
      }
      
      else                                                                 //  HDR / HDF mode, warp 4 corners
      {
         for (ii = 0; ii < 4; ii++)                                        //  modify one corner at a time
         {
            wxL = cimOffs[im2].wx[ii] - wrange;
            wxH = cimOffs[im2].wx[ii] + wrange + 0.5 * wstep;
            wyL = cimOffs[im2].wy[ii] - wrange;
            wyH = cimOffs[im2].wy[ii] + wrange + 0.5 * wstep;

            for (cimOffs[im2].wx[ii] = wxL; cimOffs[im2].wx[ii] < wxH; cimOffs[im2].wx[ii] += wstep)
            for (cimOffs[im2].wy[ii] = wyL; cimOffs[im2].wy[ii] < wyH; cimOffs[im2].wy[ii] += wstep)
            {
               cim_warp_image(im2);                                        //  warp image using all corner offsets

               match = cim_match_images(im1,im2);                          //  get match level

               if (sigdiff(match,matchB,0.00001) > 0) {
                  matchB = match;                                          //  save best match
                  offsetsB = cimOffs[im2];
                  cim_show_images(0,0);
               }

               sprintf(SB_text,"warp: %d  match: %.5f",cimNsearch++,matchB);
               SBupdate++;
            }

            cimOffs[im2] = offsetsB;                                       //  restore best match
            cim_warp_image(im2);
            cim_show_images(0,0);
         }
      }
   }

   return;
}


//  Compare 2 pixels using precalculated brightness ratios
//  1.0 = perfect match   0 = total mismatch (black/white)

inline double cim_match_pixels(uint16 *pix1, uint16 *pix2)                 //  v.10.7
{
   double      red1, green1, blue1, red2, green2, blue2;
   double      reddiff, greendiff, bluediff, match;
   double      ff = 1.0 / 65536.0;

   red1 = pix1[0];
   green1 = pix1[1];
   blue1 = pix1[2];

   red2 = pix2[0];
   green2 = pix2[1];
   blue2 = pix2[2];

   reddiff = ff * fabs(red1-red2);                                         //  0 = perfect match
   greendiff = ff * fabs(green1-green2);                                   //  1 = total mismatch
   bluediff = ff * fabs(blue1-blue2);
   
   match = (1.0 - reddiff) * (1.0 - greendiff) * (1.0 - bluediff);         //  1 = perfect match
   return match;
}


//  Compare two images in overlapping areas.
//  Use the high-contrast pixels from cim_get_redpix()
//  return: 1 = perfect match, 0 = total mismatch (black/white)

double cim_match_images(int im1, int im2)                                  //  v.10.7
{
   uint16      *pix1, vpix2[3];
   int         ww, px1, py1, ii, vstat;
   double      px2, py2;
   double      x1, y1, t1, x2, y2, t2;
   double      xoff, yoff, toff, costf, sintf;
   double      match, cmatch, maxcmatch;
   PXM         *pxm1, *pxm2;
   
   x1 = cimOffs[im1].xf;                                                   //  im1, im2 absolute offsets
   y1 = cimOffs[im1].yf;
   t1 = cimOffs[im1].tf;
   x2 = cimOffs[im2].xf;
   y2 = cimOffs[im2].yf;
   t2 = cimOffs[im2].tf;
   
   xoff = (x2 - x1) * cos(t1) + (y2 - y1) * sin(t1);                       //  offset of im2 relative to im1
   yoff = (y2 - y1) * cos(t1) - (x2 - x1) * sin(t1);
   toff = t2 - t1;

   costf = cos(toff);
   sintf = sin(toff);

   pxm1 = cimPXMw[im1];
   pxm2 = cimPXMw[im2];

   ww = pxm1->ww;   

   cmatch = 0;
   maxcmatch = 1;

   for (py1 = cimOv1ylo; py1 < cimOv1yhi; py1++)                           //  loop overlapping pixels
   for (px1 = cimOv1xlo; px1 < cimOv1xhi; px1++)
   {
      ii = py1 * ww + px1;                                                 //  skip low-contrast pixels
      if (! cimRedpix[ii]) continue;

      pix1 = PXMpix(pxm1,px1,py1);                                         //  image1 pixel
      if (! pix1[2]) continue;                                             //  ignore void pixels

      px2 = costf * (px1 - xoff) + sintf * (py1 - yoff);                   //  corresponding image2 pixel
      py2 = costf * (py1 - yoff) - sintf * (px1 - xoff);
      vstat = vpixel(pxm2,px2,py2,vpix2);
      if (! vstat) continue;
      
      match = cim_match_pixels(pix1,vpix2);                                //  compare brightness adjusted
      cmatch += match;                                                     //  accumulate total match
      maxcmatch += 1.0;
   }
   
   return cmatch / maxcmatch;
}


//  combine and show all images
//  fnew >> make new E3 output image and adjust x and y offsets
//  cimPXMw[*] >> E3pxm16 >> main window
//  fblend:  0 > 50/50 blend,  1 > gradual blend

namespace  cim_show_images_names {
   int         im1, im2, iminc, fblendd;
   int         wwlo[10], wwhi[10];
   double      costf[10], sintf[10];
}

void cim_show_images(int fnew, int fblend)                                 //  v.10.7
{
   using namespace cim_show_images_names;

   void * cim_show_images_wthread(void *arg);

   int         imx, pxr, pyr, ii, px3, py3;
   int         ww, hh, wwmax, hhmin, hhmax, bmid;
   double      xf, yf, tf;
   uint16      *pix3;
   
   mutex_lock(&pixmaps_lock);                                              //  stop window updates
   
   fblendd = fblend;                                                       //  blend 50/50 or gradual ramp

   im1 = cimShowIm1;                                                       //  two images to show
   im2 = cimShowIm2;
   iminc = im2 - im1;                                                      //  v.10.9
   
   if (cimShowAll) {                                                       //  show all images     v.10.9
      im1 = 0;
      im2 = cimNF-1;
      iminc = 1;
   }

   for (imx = 0; imx < cimNF; imx++) {                                     //  pre-calculate
      costf[imx] = cos(cimOffs[imx].tf);
      sintf[imx] = sin(cimOffs[imx].tf);
   }

   if (fnew) {
      PXM_free(E3pxm16);                                                   //  force new output pixmap
      E3pxm16 = 0;
   }
   
   if (! E3pxm16)                                                          //  allocate output pixmap
   {
      xf = cimOffs[im1].xf;                                                //  shift all images to left margin
      for (imx = im1; imx <= im2; imx += iminc)
         cimOffs[imx].xf -= xf;

      wwmax = cimPXMw[im2]->ww + cimOffs[im2].xf;                          //  ww extent of last image
      
      hhmin = cimOffs[im1].yf;
      hhmax = 0;

      for (imx = im1; imx <= im2; imx += iminc)                            //  find min and max hh extent
      {
         yf = cimOffs[imx].yf;
         tf = cimOffs[imx].tf;
         ww = cimPXMw[imx]->ww;
         hh = cimPXMw[imx]->hh;
         if (yf < hhmin) hhmin = yf;
         if (yf + tf * ww < hhmin) hhmin = yf + tf * ww;
         if (yf + hh > hhmax) hhmax = yf + hh;
         if (yf + hh + tf * ww > hhmax) hhmax = yf + hh + tf * ww;
      }

      for (imx = im1; imx <= im2; imx += iminc)                            //  align to top edge
         cimOffs[imx].yf -= hhmin;
      hhmax = hhmax - hhmin;

      E3pxm16 = PXM_make(wwmax,hhmax,16);                                  //  allocate output image
      E3ww = wwmax;
      E3hh = hhmax;
   }

   for (imx = im1; imx <= im2; imx += iminc)                               //  get ww range of each image
   {
      ww = cimPXMw[imx]->ww;
      hh = cimPXMw[imx]->hh;
      tf = cimOffs[imx].tf;
      wwlo[imx] = cimOffs[imx].xf;
      wwhi[imx] = wwlo[imx] + ww;
      wwlo[imx] -= 0.5 * tf * hh;                                          //  use midpoint of sloping edges
      wwhi[imx] -= 0.5 * tf * hh;
   }

   if (cimShowAll && cimBlend > -1) {                                      //  pano
      for (imx = im1; imx <= im2-1; imx++)                                 //  reduce for blend width
      {
         if (wwhi[imx] - wwlo[imx+1] > cimBlend) {
            bmid = (wwhi[imx] + wwlo[imx+1]) / 2;
            wwlo[imx+1] = bmid - cimBlend / 2;
            wwhi[imx] = bmid + cimBlend / 2;
         }
      }
   }

   for (ii = 0; ii < Nwt; ii++)                                            //  start worker threads   v.10.7
      start_wt(cim_show_images_wthread,&wtnx[ii]);
   wait_wts();                                                             //  wait for completion

   if (cimRedpix) 
   {
      imx = cimRedImage;                                                   //  paint red pixels for current image
      ww = cimPXMw[imx]->ww;                                               //    being aligned
      hh = cimPXMw[imx]->hh;

      for (ii = 0; ii < ww * hh; ii++)
      {
         if (cimRedpix[ii]) {
            pyr = ii / ww;                                                 //  red pixel
            pxr = ii - pyr * ww;
            px3 = cimOffs[imx].xf + pxr * costf[imx] - pyr * sintf[imx] + 0.5;
            py3 = cimOffs[imx].yf + pyr * costf[imx] + pxr * sintf[imx] + 0.5;
            pix3 = PXMpix(E3pxm16,px3,py3);
            pix3[0] = 65535; pix3[1] = pix3[2] = 1;
         }
      }
   }

   mutex_unlock(&pixmaps_lock);
   mwpaint2();                                                             //  update window
   zmainloop();

   return;
}


void * cim_show_images_wthread(void *arg)                                  //  working thread   v.10.7
{
   using namespace cim_show_images_names;

   int         index = *((int *) (arg));
   int         imx, imy;
   int         px3, py3;
   int         vstat, vstat1, vstat2;
   int         red1, green1, blue1;
   int         red2, green2, blue2;
   int         red3, green3, blue3;
   double      f1, f2, px, py;
   uint16      vpix[3], *pix3;
   
   red1 = green1 = blue1 = 0;

   f1 = f2 = 0.5;                                                          //  to use if no fblend flag

   for (py3 = index; py3 < E3hh; py3 += Nwt)                               //  loop E3 rows
   for (px3 = 0; px3 < E3ww; px3++)                                        //  loop E3 columns
   {
      vstat1 = vstat2 = 0;
      
      for (imx = imy = im1; imx <= im2; imx += iminc)                      //  find which images overlap this pixel
      {
         if (px3 < wwlo[imx] || px3 > wwhi[imx]) continue;
         px = costf[imx] * (px3 - cimOffs[imx].xf) + sintf[imx] * (py3 - cimOffs[imx].yf);
         py = costf[imx] * (py3 - cimOffs[imx].yf) - sintf[imx] * (px3 - cimOffs[imx].xf);
         vstat = vpixel(cimPXMw[imx],px,py,vpix);
         if (! vstat) continue;

         if (! vstat1) {                                                   //  first overlapping image
            vstat1 = 1;
            imy = imx;
            red1 = vpix[0];
            green1 = vpix[1];
            blue1 = vpix[2];
         }
         else {                                                            //  second image
            vstat2 = 1;
            red2 = vpix[0];
            green2 = vpix[1];
            blue2 = vpix[2];
            break;
         }
      }
      
      imx = imy;                                                           //  first of 1 or 2 overlapping images

      if (vstat1) {
         if (! vstat2) {
            red3 = red1;                                                   //  use image1 pixel
            green3 = green1;
            blue3 = blue1; 
         }
         else {                                                            //  use blended image1 + image2 pixels
            if (fblendd) {
               f1 = wwhi[imx] - px3;                                       //  gradual blend
               f2 = px3 - wwlo[imx+1];
               f1 = f1 / (f1 + f2);
               f2 = 1.0 - f1;
            }
            red3 = f1 * red1 + f2 * red2 + 0.5;
            green3 = f1 * green1 + f2 * green2 + 0.5;
            blue3 = f1 * blue1 + f2 * blue2 + 0.5;
         }
      }

      else red3 = green3 = blue3 = 0;                                      //  no overlapping image, use black pixel

      pix3 = PXMpix(E3pxm16,px3,py3);                                      //  output pixel
      pix3[0] = red3;
      pix3[1] = green3;
      pix3[2] = blue3;
   }

   exit_wt();
   return 0;                                                               //  not executed, avoid gcc warning
}


//  cut-off edges of output image where all input images do not overlap

void cim_trim()                                                            //  v.10.9
{
   int      edgex[8] =  {  0, 1,  2, 2,  2, 1,  0, 0 };                    //  4 corners and 4 midpoints of rectangle
   int      edgey[8] =  {  0, 0,  0, 1,  2, 2,  2, 1 };                    //  0 and 2 mark corners, 1 marks midpoints
   int      edgewx[4] = { +1, -1, -1, +1 };
   int      edgewy[4] = { +1, +1, -1, -1 };

   int      imx, ii, jj, ww, hh, px3, py3, px9, py9;
   int      wwmin, wwmax, hhmin, hhmax;
   double   xf, yf, tf, sintf, costf, px, py, wx, wy;
   uint16   *pix3, *pix9;

   wwmin = hhmin = 0;
   wwmax = E3ww;
   hhmax = E3hh;
   
   for (imx = 0; imx < cimNF; imx++)                                       //  loop all images
   {
      ww = cimPXMw[imx]->ww;                                               //  image size
      hh = cimPXMw[imx]->hh;
      xf = cimOffs[imx].xf;                                                //  alignment offsets
      yf = cimOffs[imx].yf;
      tf = cimOffs[imx].tf;
      sintf = sin(tf);
      costf = cos(tf);
      
      for (ii = 0; ii < 8; ii++)                                           //  8 points around image rectangle
      {
         px = ww * edgex[ii] / 2;                                          //  coordinates before warping
         py = hh * edgey[ii] / 2;
         
         if (edgex[ii] != 1 && edgey[ii] != 1) {                           //  if a corner
            jj = ii / 2;
            wx = cimOffs[imx].wx[jj];                                      //  corner warp
            wy = cimOffs[imx].wy[jj];
            if (edgewx[jj] > 0 && wx < 0) px -= wx;                        //  if warp direction inwards,
            if (edgewx[jj] < 0 && wx > 0) px -= wx;                        //    reduce px/py by warp
            if (edgewy[jj] > 0 && wy < 0) py -= wy;
            if (edgewy[jj] < 0 && wy > 0) py -= wy;
         }

         px3 = xf + px * costf - py * sintf;                               //  map px/py to output image px3/py3
         py3 = yf + py * costf + px * sintf;
         
         if (edgex[ii] != 1) {
            if (px3 < ww/2 && px3 > wwmin) wwmin = px3;                    //  remember px3/py3 extremes
            if (px3 > ww/2 && px3 < wwmax) wwmax = px3;
         }

         if (edgey[ii] != 1) {
            if (py3 < hh/2 && py3 > hhmin) hhmin = py3;
            if (py3 > hh/2 && py3 < hhmax) hhmax = py3;
         }
      }
   }
   
   wwmin += 2;                                                             //  compensate rounding
   wwmax -= 2;
   hhmin += 2;
   hhmax -= 2;

   ww = wwmax - wwmin;                                                     //  new image size
   hh = hhmax - hhmin;

   if (ww < 0.7 * E3ww) return;                                            //  sanity check
   if (hh < 0.7 * E3hh) return;
   
   E9pxm16 = PXM_make(ww,hh,16);

   for (py3 = hhmin; py3 < hhmax; py3++)                                   //  E9 = trimmed E3
   for (px3 = wwmin; px3 < wwmax; px3++)
   {
      px9 = px3 - wwmin;
      py9 = py3 - hhmin;
      pix3 = PXMpix(E3pxm16,px3,py3);
      pix9 = PXMpix(E9pxm16,px9,py9);
      pix9[0] = pix3[0];
      pix9[1] = pix3[1];
      pix9[2] = pix3[2];
   }

   PXM_free(E3pxm16);                                                      //  E3 = E9
   E3pxm16 = E9pxm16;
   E9pxm16 = 0;
   E3ww = ww;
   E3hh = hh;

   return;
}


//  dump offsets to stdout - diagnostic tool

void cim_dump_offsets(cchar *text)
{
   printf("\n offsets: %s \n",text);

   for (int imx = 0; imx < cimNF; imx++)
   {
      printf(" imx %d  x/y/t: %.1f %.1f %.4f  w0: %.1f %.1f  w1: %.1f %.1f  w2: %.1f %.1f  w3: %.1f %.1f \n",
          imx, cimOffs[imx].xf, cimOffs[imx].yf, cimOffs[imx].tf,
               cimOffs[imx].wx[0], cimOffs[imx].wy[0], cimOffs[imx].wx[1], cimOffs[imx].wy[1], 
               cimOffs[imx].wx[2], cimOffs[imx].wy[2], cimOffs[imx].wx[3], cimOffs[imx].wy[3]);
   }

   return;
}


/**************************************************************************

   Make an HDR (high dynamic range) image from several images of the same
   subject with different exposure levels. The composite image has better
   visibility of detail in both the brightest and darkest areas.

***************************************************************************/

int      HDRstat;                                                          //  1 = OK, 0 = failed or canceled
double   HDRinitAlignSize = 160;                                           //  initial align image size
double   HDRimageIncrease = 1.6;                                           //  image size increase per align cycle
double   HDRsampSize = 20000;                                              //  pixel sample size

double   HDRinitSearchRange = 8.0;                                         //  initial search range, +/- pixels
double   HDRinitSearchStep = 0.5;                                          //  initial search step, pixels 
double   HDRinitWarpRange = 2.0;                                           //  initial corner warp range, +/- pixels
double   HDRinitWarpStep = 0.5;                                            //  initial corner warp step, pixels 
double   HDRsearchRange = 2.0;                                             //  normal search range, +/- pixels
double   HDRsearchStep = 0.67;                                             //  normal search step, pixels 
double   HDRwarpRange = 1.0;                                               //  normal corner warp range, +/- pixels
double   HDRwarpStep = 0.67;                                               //  normal corner warp step, pixels

float    *HDRbright = 0;                                                   //  maps brightness per pixel
zdialog  *HDRzd = 0;                                                       //  tweak dialog
double   HDR_respfac[10][1000];                                            //  contribution / image / pixel brightness

void * HDR_align_thread(void *);                                           //  align 2 images
void   HDR_brightness();                                                   //  compute pixel brightness levels
void   HDR_tweak();                                                        //  adjust image contribution curves
void * HDR_combine_thread(void *);                                         //  combine images per contribution curves


//  menu function

void m_HDR(GtkWidget *, cchar *)                                           //  v.10.7
{
   char        **flist, *ftemp;
   int         imx, jj, err, px, py, ww, hh;
   double      diffw, diffh;
   double      fbright[10], btemp;
   double      pixsum, fnorm = 3.0 / 65536.0;
   uint16      *pixel;
   PXM         *pxmtemp;
   
   zfuncs::F1_help_topic = "HDR";                                          //  help topic

   if (mod_keep()) return;                                                 //  warn unsaved changes

   for (imx = 0; imx < 10; imx++)
   {                                                                       //  clear all file and PXM data
      cimFile[imx] = 0;
      cimPXMf[imx] = cimPXMs[imx] = cimPXMw[imx] = 0;
   }

   cimNF = 0;   
   HDRbright = 0;

   flist = zgetfileN(ZTX("Select 2 to 9 files"),"openN",image_file);       //  select images to combine
   if (! flist) return;

   for (imx = 0; flist[imx]; imx++);                                       //  count selected files
   if (imx < 2 || imx > 9) {
      zmessageACK(MWIN,ZTX("Select 2 to 9 files"));
      goto cleanup;
   }

   cimNF = imx;                                                            //  file count
   for (imx = 0; imx < cimNF; imx++)
      cimFile[imx] = strdupz(flist[imx],0,"HDR.file");                     //  set up file list
   
   if (! cim_load_files()) goto cleanup;                                   //  load and check all files

   ww = cimPXMf[0]->ww;
   hh = cimPXMf[0]->hh;
   
   for (imx = 1; imx < cimNF; imx++)                                       //  check image compatibility
   {
      diffw = abs(ww - cimPXMf[imx]->ww);
      diffw = diffw / ww;
      diffh = abs(hh - cimPXMf[imx]->hh);
      diffh = diffh / hh;

      if (diffw > 0.02 || diffh > 0.02) {
         zmessageACK(MWIN,ZTX("Images are not all the same size"));
         goto cleanup;
      }
   }
   
   free_resources();                                                       //  ready to commit

   err = f_open(cimFile[0],0);                                             //  image_file = 1st file in list
   if (err) goto cleanup;

   if (! edit_setup("HDR",0,0)) goto cleanup;                              //  setup edit, no preview

   for (imx = 0; imx < cimNF; imx++)                                       //  compute image brightness levels
   {
      pixsum = 0;
      for (py = 0; py < Fhh; py++)
      for (px = 0; px < Fww; px++)
      {
         pixel = PXMpix(cimPXMf[imx],px,py);
         pixsum += fnorm * (pixel[0] + pixel[1] + pixel[2]);
      }
      fbright[imx] = pixsum / (Fww * Fhh);
   }

   for (imx = 0; imx < cimNF; imx++)                                       //  sort file and pixmap lists
   for (jj = imx+1; jj < cimNF; jj++)                                      //    by decreasing brightness
   {
      if (fbright[jj] > fbright[imx]) {                                    //  bubble sort
         btemp = fbright[jj];
         fbright[jj] = fbright[imx];
         fbright[imx] = btemp;
         ftemp = cimFile[jj];
         cimFile[jj] = cimFile[imx];
         cimFile[imx] = ftemp;
         pxmtemp = cimPXMf[jj];
         cimPXMf[jj] = cimPXMf[imx];
         cimPXMf[imx] = pxmtemp;
      }
   }
   
   set_cursor(busycursor);
   start_thread(HDR_align_thread,0);                                       //  align each pair of images
   wrapup_thread(0);                                                       //  wait for completion
   set_cursor(0);
   if (HDRstat != 1) goto cancel;

   HDR_brightness();                                                       //  compute pixel brightness levels
   if (HDRstat != 1) goto cancel;
   
   HDR_tweak();                                                            //  combine images based on user inputs
   if (HDRstat != 1) goto cancel;

   Fmodified = 1;                                                          //  done
   edit_done();
   goto cleanup;

cancel:
   edit_cancel();

cleanup:

   if (flist) {
      for (imx = 0; flist[imx]; imx++)                                     //  free file list
         zfree(flist[imx]);
      zfree(flist);
   }

   for (imx = 0; imx < cimNF; imx++) {                                     //  free cim file and PXM data
      if (cimFile[imx]) zfree(cimFile[imx]);
      if (cimPXMf[imx]) PXM_free(cimPXMf[imx]);
      if (cimPXMs[imx]) PXM_free(cimPXMs[imx]);
      if (cimPXMw[imx]) PXM_free(cimPXMw[imx]);
   }
   
   if (HDRbright) zfree(HDRbright);
   *SB_text = 0;

   return;
}


//  HDR align each pair of input images, output combined image to E3pxm16
//  cimPXMf[*]  original image
//  cimPXMs[*]  scaled and color adjusted for pixel comparisons
//  cimPXMw[*]  warped for alignment

void * HDR_align_thread(void *)                                            //  v.10.7
{
   int         imx, im1, im2, ww, hh, ii, nn;
   double      R, maxtf, mintf, midtf;
   double      xoff, yoff, toff, dxoff, dyoff;
   cimoffs     offsets[10];                                                //  x/y/t offsets after alignment
   
   Fzoom = 0;                                                              //  fit to window if big
   Fblowup = 1;                                                            //  scale up to window if small

   for (imx = 0; imx < cimNF; imx++)                                       //  bugfix     v.10.8
      memset(&offsets[imx],0,sizeof(cimoffs));

   for (im1 = 0; im1 < cimNF-1; im1++)                                     //  loop each pair of images
   {
      im2 = im1 + 1;

      memset(&cimOffs[im1],0,sizeof(cimoffs));                             //  initial image offsets = 0
      memset(&cimOffs[im2],0,sizeof(cimoffs));

      ww = cimPXMf[im1]->ww;                                               //  image dimensions
      hh = cimPXMf[im1]->hh;

      nn = ww;                                                             //  use larger of ww, hh
      if (hh > ww) nn = hh;
      cimScale = HDRinitAlignSize / nn;                                    //  initial align image size
      if (cimScale > 1.0) cimScale = 1.0;

      cimBlend = -1;                                                       //  no blend width (use all)
      cim_get_overlap(im1,im2,cimPXMf);                                    //  get overlap area
      cim_match_colors(im1,im2,cimPXMf);                                   //  get color matching factors

      cimSearchRange = HDRinitSearchRange;                                 //  initial align search range
      cimSearchStep = HDRinitSearchStep;                                   //  initial align search step
      cimWarpRange = HDRinitWarpRange;                                     //  initial align corner warp range
      cimWarpStep = HDRinitWarpStep;                                       //  initial align corner warp step
      cimSampSize = HDRsampSize;                                           //  pixel sample size for align/compare
      cimNsearch = 0;                                                      //  reset align search counter
      cimShrink = 0;                                                       //  no warp shrinkage (pano)
      cimPano = 0;                                                         //  no pano mode

      while (true)                                                         //  loop, increasing image size
      {
         cim_scale_image(im1,cimPXMs);                                     //  scale images to cimScale
         cim_scale_image(im2,cimPXMs);

         cim_adjust_colors(cimPXMs[im1],1);                                //  apply color adjustments
         cim_adjust_colors(cimPXMs[im2],2);

         cim_warp_image(im1);                                              //  make alignment images cimPXMw[*]
         cim_warp_image(im2);      

         cimShowIm1 = im1;                                                 //  show two images with 50/50 blend
         cimShowIm2 = im2;
         cimShowAll = 0;
         cim_show_images(1,0);                                             //  (x/y offsets can change)

         cim_get_overlap(im1,im2,cimPXMw);                                 //  get overlap area
         cim_get_redpix(im1);                                              //  get high-contrast pixels

         cim_align_image(im1,im2);                                         //  align im2 to im1

         zfree(cimRedpix);                                                 //  clear red pixels
         cimRedpix = 0;

         if (cimScale == 1.0) break;                                       //  done

         R = HDRimageIncrease;                                             //  next larger image size
         cimScale = cimScale * R;
         if (cimScale > 0.85) {                                            //  if close to end, jump to end
            R = R / cimScale;
            cimScale = 1.0;
         }

         cimOffs[im1].xf *= R;                                             //  scale offsets for larger image
         cimOffs[im1].yf *= R;
         cimOffs[im2].xf *= R;
         cimOffs[im2].yf *= R;

         for (ii = 0; ii < 4; ii++) {
            cimOffs[im1].wx[ii] *= R;
            cimOffs[im1].wy[ii] *= R;
            cimOffs[im2].wx[ii] *= R;
            cimOffs[im2].wy[ii] *= R;
         }

         cimSearchRange = HDRsearchRange;                                  //  align search range
         cimSearchStep = HDRsearchStep;                                    //  align search step size
         cimWarpRange = HDRwarpRange;                                      //  align corner warp range
         cimWarpStep = HDRwarpStep;                                        //  align corner warp step size
      }

      offsets[im2].xf = cimOffs[im2].xf - cimOffs[im1].xf;                 //  save im2 offsets from im1
      offsets[im2].yf = cimOffs[im2].yf - cimOffs[im1].yf;
      offsets[im2].tf = cimOffs[im2].tf - cimOffs[im1].tf;
      
      for (ii = 0; ii < 4; ii++) {
         offsets[im2].wx[ii] = cimOffs[im2].wx[ii] - cimOffs[im1].wx[ii];
         offsets[im2].wy[ii] = cimOffs[im2].wy[ii] - cimOffs[im1].wy[ii];
      }
   }
   
   for (imx = 0; imx < cimNF; imx++)                                       //  offsets[*] >> cimOffs[*]
      cimOffs[imx] = offsets[imx];

   cimOffs[0].xf = cimOffs[0].yf = cimOffs[0].tf = 0;                      //  image 0 at (0,0,0)

   for (im1 = 0; im1 < cimNF-1; im1++)                                     //  absolute offsets for image 1 to last
   {
      im2 = im1 + 1;
      cimOffs[im2].xf += cimOffs[im1].xf;                                  //  x/y/t offsets are additive
      cimOffs[im2].yf += cimOffs[im1].yf;
      cimOffs[im2].tf += cimOffs[im1].tf;

      for (ii = 0; ii < 4; ii++) {                                         //  corner warps are additive
         cimOffs[im2].wx[ii] += cimOffs[im1].wx[ii];
         cimOffs[im2].wy[ii] += cimOffs[im1].wy[ii];
      }
   }

   for (imx = 1; imx < cimNF; imx++)                                       //  re-warp to absolute       v.10.8
      cim_warp_image(imx);

   toff = cimOffs[0].tf;                                                   //  balance +/- thetas
   maxtf = mintf = toff;
   for (imx = 1; imx < cimNF; imx++) {
      toff = cimOffs[imx].tf;
      if (toff > maxtf) maxtf = toff;
      if (toff < mintf) mintf = toff;
   }
   midtf = 0.5 * (maxtf + mintf);
   
   for (imx = 0; imx < cimNF; imx++)
      cimOffs[imx].tf -= midtf;

   for (im1 = 0; im1 < cimNF-1; im1++)                                     //  adjust x/y offsets for images after im1
   for (im2 = im1+1; im2 < cimNF; im2++)                                   //    due to im1 theta offset
   {
      toff = cimOffs[im1].tf;
      xoff = cimOffs[im2].xf - cimOffs[im1].xf;
      yoff = cimOffs[im2].yf - cimOffs[im1].yf;
      dxoff = yoff * sin(toff);
      dyoff = xoff * sin(toff);
      cimOffs[im2].xf -= dxoff;
      cimOffs[im2].yf += dyoff;
   }

   Fzoom = Fblowup = 0;
   HDRstat = 1;
   thread_exit();
   return 0;                                                               //  not executed
}


//  Compute mean image pixel brightness levels.
//  (basis for setting image contributions per brightness level)

void HDR_brightness()                                                      //  v.10.7
{
   int         px3, py3, ww, hh, imx, kk, vstat;
   double      px, py, red, green, blue;
   double      bright, maxbright, minbright;
   double      xoff, yoff, sintf[10], costf[10];
   double      norm, fnorm = 1.0 / 65536.0;
   uint16      vpix[3], *pix3;

   cimScale = 1.0;

   for (imx = 0; imx < cimNF; imx++)                                       //  replace alignment images
   {                                                                       //   (color adjusted for pixel matching)
      PXM_free(cimPXMs[imx]);                                              //    with the original images
      cimPXMs[imx] = PXM_copy(cimPXMf[imx]);
      cim_warp_image(imx);                                                 //  re-apply warps
   }

   for (imx = 0; imx < cimNF; imx++)                                       //  pre-calculate trig functions
   {
      sintf[imx] = sin(cimOffs[imx].tf);
      costf[imx] = cos(cimOffs[imx].tf);
   }
   
   ww = E3pxm16->ww;
   hh = E3pxm16->hh;

   HDRbright = (float *) zmalloc(ww*hh*sizeof(int),"hdr.brightmem");       //  get memory for brightness array
   
   minbright = 1.0;
   maxbright = 0.0;
   
   for (py3 = 0; py3 < hh; py3++)                                          //  step through all output pixels
   for (px3 = 0; px3 < ww; px3++)
   {
      red = green = blue = 0;
      vstat = 0;

      for (imx = 0; imx < cimNF; imx++)                                    //  step through all input images
      {
         xoff = cimOffs[imx].xf;
         yoff = cimOffs[imx].yf;

         px = costf[imx] * (px3 - xoff) + sintf[imx] * (py3 - yoff);       //  image N pixel, after offsets
         py = costf[imx] * (py3 - yoff) - sintf[imx] * (px3 - xoff);
         vstat = vpixel(cimPXMw[imx],px,py,vpix);
         if (! vstat) break;
         
         red += fnorm * vpix[0];                                           //  sum input pixels
         green += fnorm * vpix[1];
         blue += fnorm * vpix[2];
      }
      
      if (! vstat) {                                                       //  pixel outside some image
         pix3 = PXMpix(E3pxm16,px3,py3);                                   //  output pixel = black
         pix3[0] = pix3[1] = pix3[2] = 0;
         kk = py3 * ww + px3;
         HDRbright[kk] = 0;
         continue;
      }
      
      bright = (red + green + blue) / (3 * cimNF);                         //  mean pixel brightness, 0.0 to 1.0
      kk = py3 * ww + px3;
      HDRbright[kk] = bright;
      
      if (bright > maxbright) maxbright = bright;
      if (bright < minbright) minbright = bright;

      pix3 = PXMpix(E3pxm16,px3,py3);                                      //  output pixel
      pix3[0] = red * 65535.0 / cimNF;
      pix3[1] = green * 65535.0 / cimNF;
      pix3[2] = blue * 65535.0 / cimNF;
   }
   
   norm = 0.999 / (maxbright - minbright);                                 //  normalize to range 0.0 to 0.999   

   for (int ii = 0; ii < ww * hh; ii++)
      HDRbright[ii] = (HDRbright[ii] - minbright) * norm;

   mwpaint2();                                                             //  update window
   return;
}


//  Dialog for user to control the contributions of each input image
//  while watching the output image which is updated in real time.

void HDR_tweak()                                                           //  v.10.7
{
   using namespace splinecurve;

   int    HDR_tweak_event(zdialog *zd, cchar *event);
   void   HDR_curvedit(int);

   int         imx;
   double      cww = 1.0 / (cimNF-1);
   
   HDRzd = zdialog_new(ZTX("Adjust Image Contributions"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(HDRzd,"frame","brframe","dialog",0,"expand|space=2");
   zdialog_add_widget(HDRzd,"hbox","hb1","dialog",0);
   zdialog_add_widget(HDRzd,"label","lab11","hb1",ZTX("dark pixels"),"space=3");
   zdialog_add_widget(HDRzd,"label","lab12","hb1",0,"expand");
   zdialog_add_widget(HDRzd,"label","lab13","hb1",ZTX("light pixels"),"space=3");
   zdialog_add_widget(HDRzd,"hbox","hb2","dialog",0,"space=3");
   zdialog_add_widget(HDRzd,"label","labf1","hb2",ZTX("file:"),"space=3");
   zdialog_add_widget(HDRzd,"label","labf2","hb2","*");

   GtkWidget *brframe = zdialog_widget(HDRzd,"brframe");                   //  set up curve edit
   curve_init(brframe,HDR_curvedit);

   Nspc = cimNF;                                                           //  no. curves = no. files
   
   for (imx = 0; imx < cimNF; imx++)                                       //  set up initial response curve
   {                                                                       //    anchor points
      vert[imx] = 0;
      nap[imx] = 2;
      apx[imx][0] = 0.01;                                                  //  flatter curves, v.9.3
      apx[imx][1] = 0.99;
      apy[imx][0] = 0.9 - imx * 0.8 * cww;
      apy[imx][1] = 0.1 + imx * 0.8 * cww;
      curve_generate(imx);
   }
   
   start_thread(HDR_combine_thread,0);                                     //  start working thread
   signal_thread();

   zdialog_resize(HDRzd,400,300);                                          //  run dialog
   zdialog_run(HDRzd,HDR_tweak_event);
   zdialog_wait(HDRzd);                                                    //  wait for completion

   return;
}


//  dialog event and completion callback function

int HDR_tweak_event(zdialog *zd, cchar *event)
{
   if (zd->zstat)                                                          //  dialog complete
   {
      wrapup_thread(8);
      if (zd->zstat == 1) HDRstat = 1;
      else HDRstat = 0;
      zdialog_free(HDRzd);
      HDRzd = 0;
      if (HDRstat == 1) cim_trim();                                        //  cut-off edges       v.10.9
   }
   
   return 1;
}


//  this function is called when a curve is edited

void HDR_curvedit(int spc)
{
   cchar  *pp; 
   
   pp = strrchr(cimFile[spc],'/');
   zdialog_stuff(HDRzd,"labf2",pp+1);
   signal_thread();
   return;
}


//  Combine all input images >> E3pxm16 based on image response curves.

void * HDR_combine_thread(void *)
{
   using namespace splinecurve;

   void * HDR_combine_wthread(void *arg);

   int         imx, ii;
   double      xlo, xhi, xval, yval, sumrf;

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      for (imx = 0; imx < cimNF; imx++)                                    //  loop input images
      {
         ii = nap[imx];                                                    //  get low and high anchor points
         xlo = apx[imx][0];                                                //    for image response curve
         xhi = apx[imx][ii-1];
         if (xlo < 0.02) xlo = 0;                                          //  snap-to scale end points
         if (xhi > 0.98) xhi = 1;

         for (ii = 0; ii < 1000; ii++)                                     //  loop all brightness levels
         {
            HDR_respfac[imx][ii] = 0;
            xval = 0.001 * ii;
            if (xval < xlo || xval > xhi) continue;                        //  no influence for brightness level
            yval = curve_yval(imx,xval);                                   //  response curve value for brightness
            HDR_respfac[imx][ii] = yval;                                   //    = contribution of this input image
         }
      }

      for (ii = 0; ii < 1000; ii++)                                        //  normalize the factors so that
      {                                                                    //    they sum to 1.0
         sumrf = 0;
         for (imx = 0; imx < cimNF; imx++)
            sumrf += HDR_respfac[imx][ii];
         if (! sumrf) continue;
         for (imx = 0; imx < cimNF; imx++)
            HDR_respfac[imx][ii] = HDR_respfac[imx][ii] / sumrf;
      }

      mutex_lock(&pixmaps_lock);                                           //  stop window updates

      for (ii = 0; ii < Nwt; ii++)                                         //  start worker threads      v.10.7
         start_wt(HDR_combine_wthread,&wtnx[ii]);
      wait_wts();                                                          //  wait for completion
   
      mutex_unlock(&pixmaps_lock);
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed
}


void * HDR_combine_wthread(void *arg)                                      //  working thread
{
   int         index = *((int *) (arg));
   int         imx, ww, hh, ii, px3, py3, vstat;
   double      sintf[10], costf[10], xoff, yoff;
   double      px, py, red, green, blue, bright, factor;
   uint16      vpix[3], *pix3;

   for (imx = 0; imx < cimNF; imx++)                                       //  pre-calculate trig functions
   {
      sintf[imx] = sin(cimOffs[imx].tf);
      costf[imx] = cos(cimOffs[imx].tf);
   }

   ww = E3pxm16->ww;
   hh = E3pxm16->hh;

   for (py3 = index; py3 < hh; py3 += Nwt)                                 //  step through all output pixels
   for (px3 = 0; px3 < ww; px3++)
   {
      ii = py3 * ww + px3;
      bright = HDRbright[ii];                                              //  mean brightness, 0.0 to 1.0
      ii = 1000 * bright;
      
      red = green = blue = 0;
      
      for (imx = 0; imx < cimNF; imx++)                                    //  loop input images
      {
         factor = HDR_respfac[imx][ii];                                    //  image contribution to this pixel
         if (! factor) continue;                                           //  none
         
         xoff = cimOffs[imx].xf;
         yoff = cimOffs[imx].yf;

         px = costf[imx] * (px3 - xoff) + sintf[imx] * (py3 - yoff);       //  input virtual pixel mapping to
         py = costf[imx] * (py3 - yoff) - sintf[imx] * (px3 - xoff);       //    this output pixel

         vstat = vpixel(cimPXMw[imx],px,py,vpix);                          //  get input pixel
         if (! vstat) continue;

         red += factor * vpix[0];                                          //  accumulate brightness contribution
         green += factor * vpix[1];
         blue += factor * vpix[2];
      }
         
      pix3 = PXMpix(E3pxm16,px3,py3);                                      //  output pixel

      pix3[0] = red;                                                       //  = sum of input pixel contributions
      pix3[1] = green;
      pix3[2] = blue;
   }
   
   exit_wt();
   return 0;                                                               //  not executed
}


/**************************************************************************

   Make an HDF (high depth of field) image from several images of the same
   subject with different focus settings. Combine the images and allow the
   user to "paint" the output composite image using the mouse and choosing 
   the sharpest input image for each area of the output image. The result 
   is an image with a depth of field that exceeds the camera capability.
   
   The images are aligned at the center, but small differences in camera 
   position (hand-held photos) will cause parallax errors that prevent 
   perfect alignment of the images. Also, the images with nearer focus
   will be slightly larger than those with farther focus. These problems
   can be compensated by dragging and warping the images using the mouse.                 v.10.7 

**************************************************************************/

int      HDFstat;                                                          //  1 = OK, 0 = failed or canceled
double   HDFinitAlignSize = 160;                                           //  initial align image size
double   HDFimageIncrease = 1.6;                                           //  image size increase per align cycle
double   HDFsampSize = 20000;                                              //  pixel sample size

double   HDFinitSearchRange = 8.0;                                         //  initial search range, +/- pixels
double   HDFinitSearchStep = 1.0;                                          //  initial search step, pixels
double   HDFinitWarpRange = 4.0;                                           //  initial corner warp range
double   HDFinitWarpStep = 1.0;                                            //  initial corner warp step
double   HDFsearchRange = 2.0;                                             //  normal search range
double   HDFsearchStep = 1.0;                                              //  normal search step
double   HDFwarpRange = 1.0;                                               //  normal corner warp range
double   HDFwarpStep = 0.67;                                               //  normal corner warp step

void * HDF_align_thread(void *);
void   HDF_tweak();
void   HDF_mousefunc();
void * HDF_combine_thread(void *);


//  menu function

void m_HDF(GtkWidget *, cchar *)                                           //  v.10.7
{
   char        **flist;
   int         imx, err, ww, hh;
   double      diffw, diffh;
   
   zfuncs::F1_help_topic = "HDF";                                          //  help topic

   if (mod_keep()) return;                                                 //  warn unsaved changes

   for (imx = 0; imx < 10; imx++)
   {                                                                       //  clear all file and PXM data
      cimFile[imx] = 0;
      cimPXMf[imx] = cimPXMs[imx] = cimPXMw[imx] = 0;
   }

   cimNF = 0;   

   flist = zgetfileN(ZTX("Select 2 to 9 files"),"openN",image_file);       //  select images to combine
   if (! flist) return;

   for (imx = 0; flist[imx]; imx++);                                       //  count selected files
   if (imx < 2 || imx > 9) {
      zmessageACK(MWIN,ZTX("Select 2 to 9 files"));
      goto cleanup;
   }

   cimNF = imx;                                                            //  file count
   for (imx = 0; imx < cimNF; imx++)
      cimFile[imx] = strdupz(flist[imx],0,"HDF.file");                     //  set up file list
   
   if (! cim_load_files()) goto cleanup;                                   //  load and check all files

   ww = cimPXMf[0]->ww;
   hh = cimPXMf[0]->hh;
   
   for (imx = 1; imx < cimNF; imx++)                                       //  check image compatibility
   {
      diffw = abs(ww - cimPXMf[imx]->ww);
      diffw = diffw / ww;
      diffh = abs(hh - cimPXMf[imx]->hh);
      diffh = diffh / hh;

      if (diffw > 0.02 || diffh > 0.02) {
         zmessageACK(MWIN,ZTX("Images are not all the same size"));
         goto cleanup;
      }
   }

   free_resources();                                                       //  ready to commit

   err = f_open(cimFile[0],0);                                             //  image_file = 1st file in list
   if (err) goto cleanup;

   if (! edit_setup("HDF",0,0)) goto cleanup;                              //  setup edit, no preview

   set_cursor(busycursor);
   start_thread(HDF_align_thread,0);                                       //  align each pair of images
   wrapup_thread(0);                                                       //  wait for completion
   set_cursor(0);
   if (HDFstat != 1) goto cancel;

   HDF_tweak();                                                            //  combine images based on user inputs
   if (HDFstat != 1) goto cancel;

   Fmodified = 1;                                                          //  done
   edit_done();
   goto cleanup;

cancel:
   edit_cancel();

cleanup:

   if (flist) {
      for (imx = 0; flist[imx]; imx++)                                     //  free file list
         zfree(flist[imx]);
      zfree(flist);
   }

   for (imx = 0; imx < cimNF; imx++) {                                     //  free cim file and PXM data
      if (cimFile[imx]) zfree(cimFile[imx]);
      if (cimPXMf[imx]) PXM_free(cimPXMf[imx]);
      if (cimPXMs[imx]) PXM_free(cimPXMs[imx]);
      if (cimPXMw[imx]) PXM_free(cimPXMw[imx]);
   }
   
   *SB_text = 0;

   return;
}


//  HDF align each pair of input images, output combined image to E3pxm16
//  cimPXMf[*]  original image
//  cimPXMs[*]  scaled and color adjusted for pixel comparisons
//  cimPXMw[*]  warped for alignment

void * HDF_align_thread(void *)                                            //  v.10.7
{
   int         imx, im1, im2, ww, hh, ii, nn;
   double      R, maxtf, mintf, midtf;
   double      xoff, yoff, toff, dxoff, dyoff;
   cimoffs     offsets[10];                                                //  x/y/t offsets after alignment
   
   Fzoom = 0;                                                              //  fit to window if big
   Fblowup = 1;                                                            //  scale up to window if small
   
   for (imx = 0; imx < cimNF; imx++)                                       //  bugfix     v.10.8
      memset(&offsets[imx],0,sizeof(cimoffs));
   
   for (im1 = 0; im1 < cimNF-1; im1++)                                     //  loop each pair of images
   {
      im2 = im1 + 1;

      memset(&cimOffs[im1],0,sizeof(cimoffs));                             //  initial image offsets = 0
      memset(&cimOffs[im2],0,sizeof(cimoffs));
      
      ww = cimPXMf[im1]->ww;                                               //  image dimensions
      hh = cimPXMf[im1]->hh;

      nn = ww;                                                             //  use larger of ww, hh
      if (hh > ww) nn = hh;
      cimScale = HDFinitAlignSize / nn;                                    //  initial align image size
      if (cimScale > 1.0) cimScale = 1.0;

      cimBlend = -1;                                                       //  no blend width (use all)
      cim_get_overlap(im1,im2,cimPXMf);                                    //  get overlap area
      cim_match_colors(im1,im2,cimPXMf);                                   //  get color matching factors

      cimSearchRange = HDFinitSearchRange;                                 //  initial align search range
      cimSearchStep = HDFinitSearchStep;                                   //  initial align search step
      cimWarpRange = HDFinitWarpRange;                                     //  initial align corner warp range
      cimWarpStep = HDFinitWarpStep;                                       //  initial align corner warp step
      cimSampSize = HDFsampSize;                                           //  pixel sample size for align/compare
      cimNsearch = 0;                                                      //  reset align search counter
      cimShrink = 0;                                                       //  no warp shrinkage (pano)
      cimPano = 0;                                                         //  no pano mode

      while (true)                                                         //  loop, increasing image size
      {
         cim_scale_image(im1,cimPXMs);                                     //  scale images to cimScale
         cim_scale_image(im2,cimPXMs);

         cim_adjust_colors(cimPXMs[im1],1);                                //  apply color adjustments
         cim_adjust_colors(cimPXMs[im2],2);

         cim_warp_image(im1);                                              //  make alignment images
         cim_warp_image(im2);

         cimShowIm1 = im1;                                                 //  show these two images
         cimShowIm2 = im2;                                                 //    with 50/50 blend
         cimShowAll = 0;
         cim_show_images(1,0);                                             //  (y offset can change)

         cim_get_overlap(im1,im2,cimPXMw);                                 //  get overlap area
         cim_get_redpix(im1);                                              //  get high-contrast pixels

         cim_align_image(im1,im2);                                         //  align im2 to im1
         
         zfree(cimRedpix);                                                 //  clear red pixels
         cimRedpix = 0;

         if (cimScale == 1.0) break;                                       //  done

         R = HDFimageIncrease;                                             //  next larger image size
         cimScale = cimScale * R;
         if (cimScale > 0.85) {                                            //  if close to end, jump to end
            R = R / cimScale;
            cimScale = 1.0;
         }

         cimOffs[im1].xf *= R;                                             //  scale offsets for larger image
         cimOffs[im1].yf *= R;
         cimOffs[im2].xf *= R;
         cimOffs[im2].yf *= R;

         for (ii = 0; ii < 4; ii++) {
            cimOffs[im1].wx[ii] *= R;
            cimOffs[im1].wy[ii] *= R;
            cimOffs[im2].wx[ii] *= R;
            cimOffs[im2].wy[ii] *= R;
         }

         cimBlend *= R;                                                    //  scale blend width
         
         cimSearchRange = HDFsearchRange;                                  //  align search range
         cimSearchStep = HDFsearchStep;                                    //  align search step size
         cimWarpRange = HDFwarpRange;                                      //  align corner warp range
         cimWarpStep = HDFwarpStep;                                        //  align corner warp step size
      }
      
      offsets[im2].xf = cimOffs[im2].xf - cimOffs[im1].xf;                 //  save im2 offsets from im1
      offsets[im2].yf = cimOffs[im2].yf - cimOffs[im1].yf;
      offsets[im2].tf = cimOffs[im2].tf - cimOffs[im1].tf;

      for (ii = 0; ii < 4; ii++) {
         offsets[im2].wx[ii] = cimOffs[im2].wx[ii] - cimOffs[im1].wx[ii];
         offsets[im2].wy[ii] = cimOffs[im2].wy[ii] - cimOffs[im1].wy[ii];
      }
   }

   for (imx = 0; imx < cimNF; imx++)                                       //  offsets[*] >> cimOffs[*]
      cimOffs[imx] = offsets[imx];
   
   cimOffs[0].xf = cimOffs[0].yf = cimOffs[0].tf = 0;                      //  image 0 at (0,0,0)

   for (im1 = 0; im1 < cimNF-1; im1++)                                     //  absolute offsets for image 1 to last
   {
      im2 = im1 + 1;
      cimOffs[im2].xf += cimOffs[im1].xf;                                  //  x/y/t offsets are additive
      cimOffs[im2].yf += cimOffs[im1].yf;
      cimOffs[im2].tf += cimOffs[im1].tf;

      for (ii = 0; ii < 4; ii++) {                                         //  corner warps are additive
         cimOffs[im2].wx[ii] += cimOffs[im1].wx[ii];
         cimOffs[im2].wy[ii] += cimOffs[im1].wy[ii];
      }
   }

   for (imx = 1; imx < cimNF; imx++)                                       //  re-warp to absolute       v.10.8
      cim_warp_image(imx);

   toff = cimOffs[0].tf;                                                   //  balance +/- thetas
   maxtf = mintf = toff;
   for (imx = 1; imx < cimNF; imx++) {
      toff = cimOffs[imx].tf;
      if (toff > maxtf) maxtf = toff;
      if (toff < mintf) mintf = toff;
   }
   midtf = 0.5 * (maxtf + mintf);
   
   for (imx = 0; imx < cimNF; imx++)
      cimOffs[imx].tf -= midtf;

   for (im1 = 0; im1 < cimNF-1; im1++)                                     //  adjust x/y offsets for images after im1
   for (im2 = im1+1; im2 < cimNF; im2++)                                   //    due to im1 theta offset
   {
      toff = cimOffs[im1].tf;
      xoff = cimOffs[im2].xf - cimOffs[im1].xf;
      yoff = cimOffs[im2].yf - cimOffs[im1].yf;
      dxoff = yoff * sin(toff);
      dyoff = xoff * sin(toff);
      cimOffs[im2].xf -= dxoff;
      cimOffs[im2].yf += dyoff;
   }

   Fzoom = Fblowup = 0;
   HDFstat = 1;
   thread_exit();
   return 0;                                                               //  not executed
}


//  paint and warp output image

zdialog  *HDFzd = 0;                                                       //  paint dialog
int      HDFsuspend;                                                       //  suspend/resume status
int      HDFmode;                                                          //  mode: paint or warp
int      HDFimage;                                                         //  current image (0 based)
int      HDFradius;                                                        //  paint mode radius
char     *HDFpixmap = 0;                                                   //  map input image per output pixel
float    *HDFwarpx[10], *HDFwarpy[10];                                     //  warp memory, pixel displacements


void HDF_tweak()                                                           //  v.10.7
{
   char     imageN[8] = "imageN", labN[4] = "0";
   int      cc, imx, ww, hh;

   int HDF_tweak_dialog_event(zdialog *zd, cchar *event);

   //    image   (o) 1  (o) 2  (o) 3  ...
   //    (o) paint  radius [___]
   //    (o) warp [__]
   //    [suspend]

   HDFzd = zdialog_new(ZTX("Paint and Warp Image"),mWin,Bdone,Bcancel,null);

   zdialog_add_widget(HDFzd,"hbox","hbim","dialog",0,"space=3");
   zdialog_add_widget(HDFzd,"label","labim","hbim",ZTX("image"),"space=5");
   zdialog_add_widget(HDFzd,"hbox","hbpw","dialog",0,"space=3");
   zdialog_add_widget(HDFzd,"vbox","vbpw1","hbpw",0,"homog|space=5");
   zdialog_add_widget(HDFzd,"vbox","vbpw2","hbpw",0,"homog|space=5");
   zdialog_add_widget(HDFzd,"radio","paint","vbpw1",ZTX("paint"));
   zdialog_add_widget(HDFzd,"radio","warp","vbpw1",ZTX("warp"));
   zdialog_add_widget(HDFzd,"hbox","hbp","vbpw2");
   zdialog_add_widget(HDFzd,"label","labpr","hbp",Bradius,"space=5");
   zdialog_add_widget(HDFzd,"spin","radius","hbp","1|400|1|100");
   zdialog_add_widget(HDFzd,"label","space","vbpw2");
   zdialog_add_widget(HDFzd,"hbox","hbsr","dialog");
   zdialog_add_widget(HDFzd,"button","susp-resm","hbsr",Bsuspend,"space=5");

   for (imx = 0; imx < cimNF; imx++) {                                     //  add radio button for each image
      imageN[5] = '1' + imx;
      labN[0] = '1' + imx;
      zdialog_add_widget(HDFzd,"radio",imageN,"hbim",labN);
   }
   
   zdialog_stuff(HDFzd,"paint",1);                                         //  paint button on
   zdialog_stuff(HDFzd,"warp",0);                                          //  warp button off
   zdialog_stuff(HDFzd,"image1",1);                                        //  initial image = 1st

   HDFmode = 0;                                                            //  start in paint mode
   HDFimage = 0;                                                           //  initial image
   HDFradius = 50;                                                         //  paint radius
   HDFsuspend = 0;                                                         //  no suspend
   
   cc = E3ww * E3hh;                                                       //  allocate pixel map
   HDFpixmap = zmalloc(cc,"HDF.pixmap");
   memset(HDFpixmap,cimNF,cc);                                             //  initial state, blend all images

   for (imx = 0; imx < cimNF; imx++) {                                     //  allocate warp memory
      ww = cimPXMw[imx]->ww;
      hh = cimPXMw[imx]->hh;
      HDFwarpx[imx] = (float *) zmalloc(ww * hh * sizeof(float),"hdf.warp");
      HDFwarpy[imx] = (float *) zmalloc(ww * hh * sizeof(float),"hdf.warp");
   }
   
   start_thread(HDF_combine_thread,0);                                     //  start working thread
   signal_thread();

   mouseCBfunc = HDF_mousefunc;                                            //  connect mouse function
   Mcapture = 1;                                                           //  capture mouse clicks

   zdialog_run(HDFzd,HDF_tweak_dialog_event);                              //  run dialog, parallel
   zdialog_wait(HDFzd);                                                    //  wait for completion

   return;
}


//  dialog event and completion callback function

int HDF_tweak_dialog_event(zdialog *zd, cchar *event)                      //  v.10.7
{
   int      imx, nn;
   
   if (zd->zstat) {                                                        //  dialog finish
      signal_thread();
      wrapup_thread(8);
      if (zd->zstat == 1) HDFstat = 1;
      else HDFstat = 0;
      if (HDFstat == 1) cim_trim();                                        //  cut-off edges       v.10.9
      zdialog_free(HDFzd);
      HDFzd = 0;
      HDFmode = 0;
      paint_toparc(2);                                                     //  stop brush outline
      Mcapture = 0;                                                        //  disconnect mouse
      mouseCBfunc = 0;
      set_cursor(0);                                                       //  no drag cursor
      zfree(HDFpixmap);                                                    //  free pixel map
      for (imx = 0; imx < cimNF; imx++) {
         zfree(HDFwarpx[imx]);                                             //  free warp memory
         zfree(HDFwarpy[imx]);
      }
   }
   
   if (strEqu(event,"paint")) {                                            //  set paint mode
      zdialog_fetch(zd,"paint",nn);
      if (! nn) return 1;
      HDFmode = 0;
      set_cursor(0);                                                       //  no drag cursor
   }
   
   if (strEqu(event,"warp")) {                                             //  set warp mode
      zdialog_fetch(zd,"warp",nn);
      if (! nn) return 1;
      HDFmode = 1;
      paint_toparc(2);                                                     //  stop brush outline
      if (! HDFsuspend)                                                    //  bugfix  v.10.8
         set_cursor(dragcursor);                                           //  set drag cursor
   }
   
   if (strnEqu(event,"image",5)) {                                         //  image radio button
      nn = event[5] - '0';                                                 //  1 to cimNF
      if (nn > 0 && nn <= cimNF) 
         HDFimage = nn - 1;                                                //  0 to cimNF-1
      signal_thread();
   }

   if (strEqu(event,"radius"))                                             //  change paint radius
      zdialog_fetch(zd,"radius",HDFradius);

   if (strEqu(event,"susp-resm"))                                          //  toggle suspend / resume
   {
      if (HDFsuspend) {                                                    //  resume
         HDFsuspend = 0;
         if (HDFmode == 0) set_cursor(0);                                  //  paint mode, normal cursor
         if (HDFmode == 1) set_cursor(dragcursor);                         //  warp mode, drag cursor
         mouseCBfunc = HDF_mousefunc;                                      //  connect mouse function
         Mcapture++;
         zdialog_stuff(zd,"susp-resm",Bsuspend);
         signal_thread();
      }
      else  {                                                              //  suspend
         HDFsuspend = 1;
         paint_toparc(2);                                                  //  stop brush outline
         set_cursor(0);                                                    //  no drag cursor
         mouseCBfunc = 0;                                                  //  disconnect mouse function
         Mcapture = 0;
         zdialog_stuff(zd,"susp-resm",Bresume);
      }
   }

   return 1;
}


//  HDF dialog mouse function
//  paint:  during drag, selected image >> HDFpixmap (within paint radius) >> E3
//  warp:   for selected image, cimPXMs >> warp >> cimPXMw >> E3

void HDF_mousefunc()                                                       //  v.10.7
{
   uint16      vpix1[3], *pix2, *pix3;
   int         imx, radius, radius2, vstat1;
   int         mx, my, dx, dy, px3, py3;
   char        imageN[8] = "imageN";
   double      px1, py1;
   double      xoff, yoff, sintf[10], costf[10];
   int         ii, px, py, ww, hh;
   double      mag, dispx, dispy, d1, d2;
   PXM         *pxm1, *pxm2;
   
   if (HDFmode == 0) goto paint;
   if (HDFmode == 1) goto warp;
   return;

paint:

   radius = HDFradius;                                                     //  paintbrush radius
   radius2 = radius * radius;

   toparcx = Mxposn - radius;                                              //  paintbrush outline circle
   toparcy = Myposn - radius;
   toparcw = toparch = 2 * radius;
   Ftoparc = 1;
   paint_toparc(3);

   if (LMclick || RMclick) {                                               //  mouse click
      LMclick = RMclick = 0;
      return;                                                              //  ignore     v.10.8
   }

   else if (Mxdrag || Mydrag) {                                            //  drag in progress
      mx = Mxdrag;
      my = Mydrag;
   }
   
   else return;

   if (mx < 0 || mx > E3ww-1 || my < 0 || my > E3hh-1)                     //  mouse outside image area
      return;

   for (imx = 0; imx < cimNF; imx++)                                       //  pre-calculate trig funcs
   {
      sintf[imx] = sin(cimOffs[imx].tf);
      costf[imx] = cos(cimOffs[imx].tf);
   }

   for (dy = -radius; dy <= radius; dy++)                                  //  loop pixels around mouse
   for (dx = -radius; dx <= radius; dx++)
   {
      if (dx*dx + dy*dy > radius2) continue;                               //  outside radius

      px3 = mx + dx;                                                       //  output pixel
      py3 = my + dy;
      if (px3 < 0 || px3 > E3ww-1) continue;                               //  outside image
      if (py3 < 0 || py3 > E3hh-1) continue;

      pix3 = PXMpix(E3pxm16,px3,py3);                                      //  output pixel

      imx = py3 * E3ww + px3;                                              //  update pixmap to selected image
      HDFpixmap[imx] = HDFimage;
      
      imx = HDFimage;
      xoff = cimOffs[imx].xf;
      yoff = cimOffs[imx].yf;
      px1 = costf[imx] * (px3 - xoff) + sintf[imx] * (py3 - yoff);         //  input virtual pixel
      py1 = costf[imx] * (py3 - yoff) - sintf[imx] * (px3 - xoff);
      vstat1 = vpixel(cimPXMw[imx],px1,py1,vpix1);
      if (vstat1) {
         pix3[0] = vpix1[0];
         pix3[1] = vpix1[1];
         pix3[2] = vpix1[2];
      }
      else pix3[0] = pix3[1] = pix3[2] = 0;
   }

   mwpaint2();                                                             //  fast update window
   return;

warp:

   if (LMclick || RMclick) {                                               //  mouse click
      LMclick = RMclick = 0;                                               //  ignore     v.10.8
      return;
   }

   else if (Mxdrag || Mydrag) {                                            //  drag in progress
      mx = Mxdrag;
      my = Mydrag;
   }
   
   else return;

   if (mx < 0 || mx > E3ww-1 || my < 0 || my > E3hh-1)                     //  mouse outside image area
      return;

   imx = my * E3ww + mx;                                                   //  if pixel has been painted, 
   imx = HDFpixmap[imx];                                                   //    select corresp. image to warp
   if (imx == cimNF) return;                                               //  else no action      v.10.8
   
   if (imx != HDFimage) {
      HDFimage = imx;                                                      //  update selected image and
      imageN[5] = '1' + imx;                                               //    dialog radio button
      zdialog_stuff(HDFzd,imageN,1);
   }

   pxm1 = cimPXMs[imx];                                                    //  input image
   pxm2 = cimPXMw[imx];                                                    //  output image
   ww = pxm2->ww;
   hh = pxm2->hh;

   mx = Mxdown;                                                            //  drag origin, image coordinates
   my = Mydown;
   dx = Mxdrag - Mxdown;                                                   //  drag increment
   dy = Mydrag - Mydown;
   Mxdown = Mxdrag;                                                        //  next drag origin
   Mydown = Mydrag;

   d1 = ww * ww + hh * hh;
   
   for (py = 0; py < hh; py++)                                             //  process all output pixels
   for (px = 0; px < ww; px++)
   {
      d2 = (px-mx)*(px-mx) + (py-my)*(py-my);
      mag = (1.0 - d2 / d1);
      mag = mag * mag * mag * mag;
      mag = mag * mag * mag * mag;
      mag = mag * mag * mag * mag;

      dispx = -dx * mag;                                                   //  displacement = drag * mag
      dispy = -dy * mag;
      
      ii = py * ww + px;
      HDFwarpx[imx][ii] += dispx;                                          //  add this drag to prior sum
      HDFwarpy[imx][ii] += dispy;

      dispx = HDFwarpx[imx][ii];
      dispy = HDFwarpy[imx][ii];

      vstat1 = vpixel(pxm1,px+dispx,py+dispy,vpix1);                       //  input virtual pixel
      pix2 = PXMpix(pxm2,px,py);                                           //  output pixel
      if (vstat1) {
         pix2[0] = vpix1[0];
         pix2[1] = vpix1[1];
         pix2[2] = vpix1[2];
      }
      else pix2[0] = pix2[1] = pix2[2] = 0;
   }

   signal_thread();                                                        //  combine images >> E3 >> main window
   return;
}


//  Combine images in E3pxm16 (not reallocated). Update main window.

void * HDF_combine_thread(void *)                                          //  v.10.7
{
   void * HDF_combine_wthread(void *);
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      mutex_lock(&pixmaps_lock);                                           //  stop window updates

      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads   v.10.7
         start_wt(HDF_combine_wthread,&wtnx[ii]);
      wait_wts();                                                          //  wait for completion

      mutex_unlock(&pixmaps_lock);                                         //  update window
      mwpaint2();
   }

   return 0;                                                               //  not executed
}


void * HDF_combine_wthread(void *arg)                                      //  worker thread
{
   int         index = *((int *) (arg));                                   //  no more paint and warp modes  v.10.8
   int         px3, py3, vstat1;
   int         imx, red, green, blue;
   double      px, py;
   double      xoff, yoff, sintf[10], costf[10];
   uint16      vpix1[3], *pix3;

   for (imx = 0; imx < cimNF; imx++)                                       //  pre-calculate trig funcs
   {
      sintf[imx] = sin(cimOffs[imx].tf);
      costf[imx] = cos(cimOffs[imx].tf);
   }

   for (py3 = index+1; py3 < E3hh-1; py3 += Nwt)                           //  step through output pixels
   for (px3 = 1; px3 < E3ww-1; px3++)
   {
      pix3 = PXMpix(E3pxm16,px3,py3);
      
      imx = py3 * E3ww + px3;
      imx = HDFpixmap[imx];

      if (imx < cimNF)                                                     //  specific image maps to pixel
      {
         xoff = cimOffs[imx].xf;
         yoff = cimOffs[imx].yf;
         px = costf[imx] * (px3 - xoff) + sintf[imx] * (py3 - yoff);
         py = costf[imx] * (py3 - yoff) - sintf[imx] * (px3 - xoff);
         vstat1 = vpixel(cimPXMw[imx],px,py,vpix1);                        //  corresp. input vpixel
         if (vstat1) {
            pix3[0] = vpix1[0];
            pix3[1] = vpix1[1];
            pix3[2] = vpix1[2];
         }
         else pix3[0] = pix3[1] = pix3[2] = 0;
      }
      
      else                                                                 //  use blend of all images
      {
         red = green = blue = 0;

         for (imx = 0; imx < cimNF; imx++)
         {
            xoff = cimOffs[imx].xf;
            yoff = cimOffs[imx].yf;
            px = costf[imx] * (px3 - xoff) + sintf[imx] * (py3 - yoff);
            py = costf[imx] * (py3 - yoff) - sintf[imx] * (px3 - xoff);
            vstat1 = vpixel(cimPXMw[imx],px,py,vpix1);
            if (vstat1) {
               red += vpix1[0];
               green += vpix1[1];
               blue += vpix1[2];
            }
         }
         
         pix3[0] = red / cimNF;
         pix3[1] = green / cimNF;
         pix3[2] = blue / cimNF;
      }
   }

   exit_wt();
   return 0;                                                               //  not executed, avoid gcc warning
}


/**************************************************************************

   Combine multiple photos of the same subject and average the 
   pixels for noise reduction.

**************************************************************************/

double   stak_initAlignSize = 160;                                         //  initial align image size
double   stak_imageIncrease = 1.6;                                         //  image size increase per align cycle
double   stak_sampSize = 10000;                                            //  pixel sample size

double   stak_initSearchRange = 6.0;                                       //  initial search range, +/- pixels
double   stak_initSearchStep = 1.0;                                        //  initial search step, pixels
double   stak_initWarpRange = 3.0;                                         //  initial corner warp range
double   stak_initWarpStep = 1.0;                                          //  initial corner warp step
double   stak_searchRange = 2.0;                                           //  normal search range
double   stak_searchStep = 1.0;                                            //  normal search step
double   stak_warpRange = 1.0;                                             //  normal corner warp range
double   stak_warpStep = 0.67;                                             //  normal corner warp step

int      stak_stat;                                                        //  1 = OK, 0 = failed or canceled
int      stak_average = 1, stak_median = 0;                                //  use average/median of input pixels
int      stak_exlow = 0, stak_exhigh = 0;                                  //  exclude low/high pixel

void * stak_align_thread(void *);
void   stak_tweak();
void * stak_combine_thread(void *);


//  menu function

void m_stak(GtkWidget *, cchar *)                                          //  new v.10.9
{
   char        **flist;
   int         imx, err, ww, hh;
   double      diffw, diffh;
   
   zfuncs::F1_help_topic = "stack";                                        //  help topic

   if (mod_keep()) return;                                                 //  warn unsaved changes

   for (imx = 0; imx < 10; imx++)
   {                                                                       //  clear all file and PXM data
      cimFile[imx] = 0;
      cimPXMf[imx] = cimPXMs[imx] = cimPXMw[imx] = 0;
   }

   cimNF = 0;   

   flist = zgetfileN(ZTX("Select 2 to 9 files"),"openN",image_file);       //  select images to combine
   if (! flist) return;

   for (imx = 0; flist[imx]; imx++);                                       //  count selected files
   if (imx < 2 || imx > 9) {
      zmessageACK(MWIN,ZTX("Select 2 to 9 files"));
      goto cleanup;
   }

   cimNF = imx;                                                            //  file count
   for (imx = 0; imx < cimNF; imx++)
      cimFile[imx] = strdupz(flist[imx],0,"stak.file");                    //  set up file list
   
   if (! cim_load_files()) goto cleanup;                                   //  load and check all files

   ww = cimPXMf[0]->ww;
   hh = cimPXMf[0]->hh;
   
   for (imx = 1; imx < cimNF; imx++)                                       //  check image compatibility
   {
      diffw = abs(ww - cimPXMf[imx]->ww);
      diffw = diffw / ww;
      diffh = abs(hh - cimPXMf[imx]->hh);
      diffh = diffh / hh;

      if (diffw > 0.02 || diffh > 0.02) {
         zmessageACK(MWIN,ZTX("Images are not all the same size"));
         goto cleanup;
      }
   }

   free_resources();                                                       //  ready to commit

   err = f_open(cimFile[0],0);                                             //  image_file = 1st file in list
   if (err) goto cleanup;

   if (! edit_setup("stak",0,0)) goto cleanup;                             //  setup edit, no preview

   set_cursor(busycursor);
   start_thread(stak_align_thread,0);                                      //  align each pair of images
   wrapup_thread(0);                                                       //  wait for completion
   set_cursor(0);
   if (stak_stat != 1) goto cancel;

   stak_tweak();                                                           //  combine images based on user inputs
   if (stak_stat != 1) goto cancel;

   Fmodified = 1;                                                          //  done
   edit_done();
   goto cleanup;

cancel:
   edit_cancel();

cleanup:

   if (flist) {
      for (imx = 0; flist[imx]; imx++)                                     //  free file list
         zfree(flist[imx]);
      zfree(flist);
   }

   for (imx = 0; imx < cimNF; imx++) {                                     //  free cim file and PXM data
      if (cimFile[imx]) zfree(cimFile[imx]);
      if (cimPXMf[imx]) PXM_free(cimPXMf[imx]);
      if (cimPXMs[imx]) PXM_free(cimPXMs[imx]);
      if (cimPXMw[imx]) PXM_free(cimPXMw[imx]);
   }
   
   *SB_text = 0;

   return;
}


//  align each image 2nd-last to 1st image
//  cimPXMf[*]  original image
//  cimPXMs[*]  scaled and color adjusted for pixel comparisons
//  cimPXMw[*]  warped for alignment

void * stak_align_thread(void *)                                           //  v.10.9
{
   int         imx, im1, im2, ww, hh, ii, nn;
   double      R, maxtf, mintf, midtf;
   double      xoff, yoff, toff, dxoff, dyoff;
   
   Fzoom = 0;                                                              //  fit to window if big
   Fblowup = 1;                                                            //  scale up to window if small
   
   for (imx = 1; imx < cimNF; imx++)                                       //  loop 2nd to last image
   {
      im1 = 0;                                                             //  images to align
      im2 = imx;

      memset(&cimOffs[im1],0,sizeof(cimoffs));                             //  initial image offsets = 0
      memset(&cimOffs[im2],0,sizeof(cimoffs));
      
      ww = cimPXMf[im1]->ww;                                               //  image dimensions
      hh = cimPXMf[im1]->hh;

      nn = ww;                                                             //  use larger of ww, hh
      if (hh > ww) nn = hh;
      cimScale = stak_initAlignSize / nn;                                  //  initial align image size
      if (cimScale > 1.0) cimScale = 1.0;

      cimBlend = -1;                                                       //  no blend width (use all)
      cim_get_overlap(im1,im2,cimPXMf);                                    //  get overlap area
      cim_match_colors(im1,im2,cimPXMf);                                   //  get color matching factors

      cimSearchRange = stak_initSearchRange;                               //  initial align search range
      cimSearchStep = stak_initSearchStep;                                 //  initial align search step
      cimWarpRange = stak_initWarpRange;                                   //  initial align corner warp range
      cimWarpStep = stak_initWarpStep;                                     //  initial align corner warp step
      cimSampSize = stak_sampSize;                                         //  pixel sample size for align/compare
      cimNsearch = 0;                                                      //  reset align search counter
      cimShrink = 0;                                                       //  no warp shrinkage (pano)
      cimPano = 0;                                                         //  no pano mode

      while (true)                                                         //  loop, increasing image size
      {
         cim_scale_image(im1,cimPXMs);                                     //  scale images to cimScale
         cim_scale_image(im2,cimPXMs);

         cim_adjust_colors(cimPXMs[im1],1);                                //  apply color adjustments
         cim_adjust_colors(cimPXMs[im2],2);

         cim_warp_image(im1);                                              //  make alignment images
         cim_warp_image(im2);

         cimShowIm1 = im1;                                                 //  show these two images
         cimShowIm2 = im2;                                                 //    with 50/50 blend
         cimShowAll = 0;
         cim_show_images(1,0);                                             //  (y offset can change)

         cim_get_overlap(im1,im2,cimPXMw);                                 //  get overlap area
         cim_get_redpix(im1);                                              //  get high-contrast pixels

         cim_align_image(im1,im2);                                         //  align im2 to im1
         
         zfree(cimRedpix);                                                 //  clear red pixels
         cimRedpix = 0;

         if (cimScale == 1.0) break;                                       //  done

         R = stak_imageIncrease;                                           //  next larger image size
         cimScale = cimScale * R;
         if (cimScale > 0.85) {                                            //  if close to end, jump to end
            R = R / cimScale;
            cimScale = 1.0;
         }

         cimOffs[im1].xf *= R;                                             //  scale offsets for larger image
         cimOffs[im1].yf *= R;
         cimOffs[im2].xf *= R;
         cimOffs[im2].yf *= R;

         for (ii = 0; ii < 4; ii++) {
            cimOffs[im1].wx[ii] *= R;
            cimOffs[im1].wy[ii] *= R;
            cimOffs[im2].wx[ii] *= R;
            cimOffs[im2].wy[ii] *= R;
         }

         cimBlend *= R;                                                    //  scale blend width
         
         cimSearchRange = stak_searchRange;                                //  align search range
         cimSearchStep = stak_searchStep;                                  //  align search step size
         cimWarpRange = stak_warpRange;                                    //  align corner warp range
         cimWarpStep = stak_warpStep;                                      //  align corner warp step size
      }
   }

   toff = cimOffs[0].tf;                                                   //  balance +/- thetas
   maxtf = mintf = toff;
   for (imx = 1; imx < cimNF; imx++) {
      toff = cimOffs[imx].tf;
      if (toff > maxtf) maxtf = toff;
      if (toff < mintf) mintf = toff;
   }
   midtf = 0.5 * (maxtf + mintf);
   
   for (imx = 0; imx < cimNF; imx++)
      cimOffs[imx].tf -= midtf;

   for (im1 = 0; im1 < cimNF-1; im1++)                                     //  adjust x/y offsets for images after im1
   for (im2 = im1+1; im2 < cimNF; im2++)                                   //    due to im1 theta offset
   {
      toff = cimOffs[im1].tf;
      xoff = cimOffs[im2].xf - cimOffs[im1].xf;
      yoff = cimOffs[im2].yf - cimOffs[im1].yf;
      dxoff = yoff * sin(toff);
      dyoff = xoff * sin(toff);
      cimOffs[im2].xf -= dxoff;
      cimOffs[im2].yf += dyoff;
   }

   Fzoom = Fblowup = 0;
   stak_stat = 1;
   thread_exit();
   return 0;                                                               //  not executed
}


//  change pixel combination according to user input

void stak_tweak()                                                          //  v.10.9
{
   zdialog     *zd;

   int stak_tweak_dialog_event(zdialog *zd, cchar *event);

   //    Adjust Pixel Composition
   //
   //    (o) use average  (o) use median
   //    [x] omit lowest value
   //    [x] omit highest value

   zd = zdialog_new(ZTX("Adjust Pixel Composition"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=3");
   zdialog_add_widget(zd,"radio","average","hb1","use average","space=3");
   zdialog_add_widget(zd,"radio","median","hb1","use median","space=3");
   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=3");
   zdialog_add_widget(zd,"check","exlow","hb2","omit low pixel","space=3");
   zdialog_add_widget(zd,"check","exhigh","hb2","omit high pixel","space=3");
   
   zdialog_stuff(zd,"average",1);                                          //  default = average
   zdialog_stuff(zd,"median",0);
   zdialog_stuff(zd,"exlow",0);
   zdialog_stuff(zd,"exhigh",0);
   
   stak_average = 1;
   stak_median = 0;
   stak_exlow = 0;
   stak_exhigh = 0;
   
   start_thread(stak_combine_thread,0);                                    //  start working thread
   signal_thread();

   zdialog_resize(zd,250,0);
   zdialog_run(zd,stak_tweak_dialog_event);                                //  run dialog, parallel
   zdialog_wait(zd);                                                       //  wait for completion

   return;
}


//  dialog event and completion callback function

int stak_tweak_dialog_event(zdialog *zd, cchar *event)                     //  v.10.9
{
   if (zd->zstat) {                                                        //  dialog finish
      if (zd->zstat == 1) stak_stat = 1;
      else stak_stat = 0;
      wrapup_thread(8);
      zdialog_free(zd);
      if (stak_stat == 1) cim_trim();                                      //  trim edges    v.10.9
   }
   
   if (strEqu(event,"average")) {
      zdialog_fetch(zd,"average",stak_average);
      signal_thread();
   }
   
   if (strEqu(event,"median")) {
      zdialog_fetch(zd,"median",stak_median);
      signal_thread();
   }
   
   if (strEqu(event,"exlow")) {
      zdialog_fetch(zd,"exlow",stak_exlow);
      signal_thread();
   }
   
   if (strEqu(event,"exhigh")) {
      zdialog_fetch(zd,"exhigh",stak_exhigh);
      signal_thread();
   }
   
   return 1;
}


//  compute mean/median mix for each output pixel and update E3 image

void * stak_combine_thread(void *)                                         //  v.10.9
{
   void * stak_combine_wthread(void *arg);                                 //  worker thread

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wt(stak_combine_wthread,&wtnx[ii]);
      wait_wts();                                                          //  wait for completion

      Fmodified = 1;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed
}


//  worker thread

void * stak_combine_wthread(void *arg)                                     //  v.10.9
{
   int         index = *((int *) arg);
   int         imx, vstat, px3, py3;
   int         red, green, blue;
   int         ii, ns, ns1, ns2;
   int         Rlist[10], Glist[10], Blist[10];
   double      px, py;
   double      xoff, yoff, sintf[10], costf[10];
   uint16      *pix3, vpix[3];

   //    input layers     0      1      2      3      4      5      6      7      8      9     10
   int   nsx[11][2] = { {0,0}, {0,0}, {0,1}, {1,1}, {1,2}, {2,2}, {2,3}, {2,4}, {2,5}, {3,5}, {3,6} };
   
   for (imx = 0; imx < cimNF; imx++)                                       //  pre-calculate trig funcs
   {
      sintf[imx] = sin(cimOffs[imx].tf);
      costf[imx] = cos(cimOffs[imx].tf);
   }

   for (py3 = index+1; py3 < E3hh-1; py3 += Nwt)                           //  step through output pixels
   for (px3 = 1; px3 < E3ww-1; px3++)
   {
      for (imx = ns = 0; imx < cimNF; imx++)                               //  get aligned input pixels
      {
         xoff = cimOffs[imx].xf;
         yoff = cimOffs[imx].yf;
         px = costf[imx] * (px3 - xoff) + sintf[imx] * (py3 - yoff);
         py = costf[imx] * (py3 - yoff) - sintf[imx] * (px3 - xoff);

         vstat = vpixel(cimPXMw[imx],px,py,vpix);
         if (vstat) {
            Rlist[ns] = vpix[0];                                           //  add pixel RGB values to list
            Glist[ns] = vpix[1];
            Blist[ns] = vpix[2];
            ns++;
         }
      }
      
      if (! ns) continue;
      
      if (stak_exlow || stak_exhigh || stak_median) {                      //  RGB values must be sorted
         HeapSort(Rlist,ns);
         HeapSort(Glist,ns);
         HeapSort(Blist,ns);
      }

      red = green = blue = 0;

      if (stak_average)                                                    //  average the input pixels
      {
         ns1 = 0;                                                          //  low and high RGB values
         ns2 = ns - 1;
         
         if (stak_exlow) {                                                 //  exclude low
            ns1++;
            if (ns1 > ns2) ns1--;
         }

         if (stak_exhigh) {                                                //  exclude high
            ns2--;
            if (ns1 > ns2) ns2++;
         }
         
         for (ii = ns1; ii <= ns2; ii++)                                   //  sum remaining RGB levels
         {
            red += Rlist[ii];
            green += Glist[ii];
            blue += Blist[ii];
         }

         ns = ns2 - ns1 + 1;                                               //  sample count

         red = red / ns;                                                   //  output RGB = average
         green = green / ns;
         blue = blue / ns;
      }
      
      if (stak_median)                                                     //  use median input pixels
      {
         ns1 = nsx[ns][0];                                                 //  middle group of pixels
         ns2 = nsx[ns][1];
         
         for (ii = ns1; ii <= ns2; ii++)
         {
            red += Rlist[ii];
            green += Glist[ii];
            blue += Blist[ii];
         }

         ns = ns2 - ns1 + 1;                                               //  sample count

         red = red / ns;                                                   //  output RGB = average
         green = green / ns;
         blue = blue / ns;
      }

      pix3 = PXMpix(E3pxm16,px3,py3);                                      //  output pixel
      pix3[0] = red;
      pix3[1] = green;
      pix3[2] = blue;
   }

   exit_wt();
   return 0;                                                               //  not executed
}


/**************************************************************************
    New panorama function, join 2, 3, or 4 images.       v.10.7
***************************************************************************/

int      panStat;                                                          //  1 = OK, 0 = failed or canceled
zdialog  *pan4zd = 0;                                                      //  pre-align dialog

double   panPreAlignSize = 1000;                                           //  pre-align image size (ww)
double   panInitAlignSize = 200;                                           //  initial align image size
double   panImageIncrease = 1.6;                                           //  image size increase per align cycle
double   panSampSize = 20000;                                              //  pixel sample size

double   panPreAlignBlend = 0.30;                                          //  pre-align blend width * ww
double   panInitBlend = 0.20;                                              //  initial blend width during auto-align
double   panFinalBlend = 0.08;                                             //  final blend width * ww
double   panBlendDecrease = 0.8;                                           //  blend width reduction per align cycle

double   panInitSearchRange = 5.0;                                         //  initial search range, +/- pixels
double   panInitSearchStep = 0.5;                                          //  initial search step, pixels
double   panInitWarpRange = 3.0;                                           //  initial corner warp range, +/- pixels
double   panInitWarpStep = 1.0;                                            //  initial corner warp step, pixels
double   panSearchRange = 3.0;                                             //  normal search range, +/- pixels
double   panSearchStep = 1.0;                                              //  normal search step, pixels
double   panWarpRange = 2.0;                                               //  normal corner warp range, +/- pixels
double   panWarpStep = 1.0;                                                //  normal corner warp step, pixels

void  pan4_prealign();                                                     //  manual pre-align
void  pan4_align();                                                        //  auto fine-align
void  pan4_tweak();                                                        //  user color tweak


//  menu function

void m_pano4(GtkWidget *, cchar *)                                         //  v.10.7
{
   int      imx, err;
   char     **flist = 0;

   zfuncs::F1_help_topic = "panorama";                                     //  help topic

   if (mod_keep()) return;                                                 //  warn unsaved changes

   for (imx = 0; imx < 10; imx++)
   {                                                                       //  clear all file and PXM data
      cimFile[imx] = 0;
      cimPXMf[imx] = cimPXMs[imx] = cimPXMw[imx] = 0;
   }
   cimNF = 0;   

   flist = zgetfileN(ZTX("Select 2 to 4 files"),"openN",image_file);       //  select images to combine
   if (! flist) return;

   for (imx = 0; flist[imx]; imx++);                                       //  count selected files
   if (imx < 2 || imx > 4) {
      zmessageACK(MWIN,ZTX("Select 2 to 4 files"));
      goto cleanup;
   }

   cimNF = imx;                                                            //  file count
   for (imx = 0; imx < cimNF; imx++)
      cimFile[imx] = strdupz(flist[imx],0,"pano.file");                    //  set up file list
   
   if (! cim_load_files()) goto cleanup;                                   //  load and check all files

   free_resources();                                                       //  ready to commit

   err = f_open(cimFile[0],0);                                             //  image_file = 1st file in list
   if (err) goto cleanup;

   if (! edit_setup("pano4",0,0)) goto cleanup;                            //  setup edit, no preview

   cimShowAll = 1;                                                         //  for cim_show_images(), show all    v.10.9
   
   pan4_prealign();                                                        //  manual pre-alignment
   if (! panStat) goto cancel;

   pan4_align();                                                           //  auto full alignment
   if (! panStat) goto cancel;

   pan4_tweak();                                                           //  manual color adjustment
   if (! panStat) goto cancel;

   Fmodified = 1;                                                          //  done
   edit_done();
   goto cleanup;

cancel:                                                                    //  failed or canceled
   edit_cancel();

cleanup:

   if (flist) {
      for (imx = 0; flist[imx]; imx++)                                     //  free file list
         zfree(flist[imx]);
      zfree(flist);
   }

   for (imx = 0; imx < cimNF; imx++) {                                     //  free cim file and PXM data
      if (cimFile[imx]) zfree(cimFile[imx]);
      if (cimPXMf[imx]) PXM_free(cimPXMf[imx]);
      if (cimPXMs[imx]) PXM_free(cimPXMs[imx]);
      if (cimPXMw[imx]) PXM_free(cimPXMw[imx]);
   }
   
   *SB_text = 0;
   return;
}


//  perform manual pre-align of all images
//  returns alignment data in cimOffs[*]
//  lens_mm and lens_bow may also be altered

void pan4_prealign()                                                       //  v.10.7
{
   int    pan4_prealign_event(zdialog *zd, cchar *event);                  //  dialog event function
   void * pan4_prealign_thread(void *);                                    //  working thread

   int         imx, ww, err = 0;
   cchar       *exifkey = { exif_focal_length_key };
   char        lensname[40], **pp = 0;

   cchar  *align_mess = ZTX("Drag images into rough alignment.\n"
                            "To rotate, drag from lower edge.");
   cchar  *search_mess = ZTX("Search for lens mm and bow");

   pp = exif_get(image_file,&exifkey,1);                                   //  get lens mm from EXIF if available
   if (pp && *pp) {
      err = convSD(*pp, lens_mm, 20, 1000);                                //  leave lens_bow unchanged (no source)
      strcpy(lensname,"(EXIF)");                                           //  lens name = EXIF
   }

   if (! pp || ! *pp || err) {                                             //  not available
      lens_mm = lens4_mm[curr_lens];                                       //  get curr. lens mm, bow, name
      lens_bow = lens4_bow[curr_lens];
      *lensname = 0;
      strncatv(lensname,40,"(",lens4_name[curr_lens],")",null);
   }
   
   for (imx = 0; imx < 10; imx++)                                          //  set all alignment offsets = 0
      memset(&cimOffs[imx],0,sizeof(cimoffs));
      
   for (imx = ww = 0; imx < cimNF; imx++)                                  //  sum image widths
      ww += cimPXMf[imx]->ww;
   
   cimScale = 1.4 * panPreAlignSize / ww;                                  //  set alignment image scale
   if (cimScale > 1.0) cimScale = 1.0;                                     //  (* 0.7 after overlaps)

   for (imx = 0; imx < cimNF; imx++)                                       //  scale images > cimPXMs[*]
      cim_scale_image(imx,cimPXMs);

   for (imx = 0; imx < cimNF; imx++) {                                     //  curve images, cimPXMs[*] replaced
      cim_curve_image(imx);
      cimPXMw[imx] = PXM_copy(cimPXMs[imx]);                               //  copy to cimPXMw[*]
   }
   
   cimOffs[0].xf = cimOffs[0].yf = 0;                                      //  first image at (0,0)

   for (imx = 1; imx < cimNF; imx++)                                       //  position images with 30% overlap
   {                                                                       //    in horizontal row
      cimOffs[imx].xf = cimOffs[imx-1].xf + 0.7 * cimPXMw[imx-1]->ww;
      cimOffs[imx].yf = cimOffs[imx-1].yf;
   }
   
   cimBlend = panPreAlignBlend * cimPXMw[1]->ww;                           //  overlap in align window
   
   Fzoom = 0;                                                              //  scale image to fit window
   Fblowup = 1;                                                            //  magnify small image to window size

   cim_show_images(1,0);                                                   //  combine and show images in main window

   pan4zd = zdialog_new(ZTX("Pre-align Images"),mWin,Bproceed,Bcancel,null);     //  start pre-align dialog
   zdialog_add_widget(pan4zd,"label","lab1","dialog",align_mess,"space=5");
   zdialog_add_widget(pan4zd,"hbox","hb1","dialog",0,"space=2");
   zdialog_add_widget(pan4zd,"spin","spmm","hb1","22|200|0.1|35","space=5");     //  [ 35  ]   lens mm  (source)
   zdialog_add_widget(pan4zd,"label","labmm","hb1",ZTX("lens mm"));              //  [ 0.3 ]   lens bow
   zdialog_add_widget(pan4zd,"label","lablens","hb1","","space=5");              //  [resize]   resize window
   zdialog_add_widget(pan4zd,"hbox","hb2","dialog",0,"space=2");
   zdialog_add_widget(pan4zd,"spin","spbow","hb2","-9|9|0.01|0","space=5");
   zdialog_add_widget(pan4zd,"label","labbow","hb2",ZTX("lens bow"));
   zdialog_add_widget(pan4zd,"hbox","hb3","dialog",0,"space=2");
   zdialog_add_widget(pan4zd,"button","resize","hb3",ZTX("Resize"),"space=5");
   zdialog_add_widget(pan4zd,"label","labsiz","hb3",ZTX("resize window"),"space=5");
   zdialog_add_widget(pan4zd,"hbox","hb4","dialog",0,"space=2");
   zdialog_add_widget(pan4zd,"button","search","hb4",Bsearch,"space=5");
   zdialog_add_widget(pan4zd,"label","labsearch","hb4",search_mess,"space=5");

   zdialog_stuff(pan4zd,"spmm",lens_mm);                                   //  stuff lens data
   zdialog_stuff(pan4zd,"spbow",lens_bow);
   zdialog_stuff(pan4zd,"lablens",lensname);                               //  show source of lens data

   panStat = -1;                                                           //  busy status
   set_cursor(dragcursor);                                                 //  set drag cursor
   zdialog_run(pan4zd,pan4_prealign_event);                                //  start dialog
   start_thread(pan4_prealign_thread,0);                                   //  start working thread
   zdialog_wait(pan4zd);                                                   //  wait for dialog completion
   set_cursor(0);                                                          //  restore normal cursor
   Fzoom = Fblowup = 0;
   return;
}


//  pre-align dialog event function

int pan4_prealign_event(zdialog *zd, cchar *event)                         //  v.10.7
{
   int      imx, ww, overlap;

   if (strstr("spmm spbow",event)) {
      zdialog_fetch(zd,"spmm",lens_mm);                                    //  get revised lens data
      zdialog_fetch(zd,"spbow",lens_bow);
   }
   
   if (strEqu(event,"resize"))                                             //  allocate new E3 image
      cim_show_images(1,0);

   if (strEqu(event,"search")) {                                           //  search for optimal lens parms
      if (cimNF != 2)
         zmessageACK(MWIN,ZTX("use two images only"));
      else  panStat = 2;
      return 0;
   }

   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1)                                                  //  proceed
         panStat = 1;
      else                                                                 //  cancel or other
         panStat = 0;

      zdialog_free(zd);                                                    //  kill dialog
      pan4zd = 0;
      wrapup_thread(0);                                                    //  wait for thread
      
      if (! panStat) return 0;                                             //  cancel

      for (imx = 0; imx < cimNF-1; imx++)                                  //  check for enough overlap
      {
         cim_get_overlap(imx,imx+1,cimPXMw);
         overlap = cimOv1xhi - cimOv1xlo;
         ww = cimPXMw[imx+1]->ww;
         if (overlap < 0.1 * ww) {
            zmessageACK(MWIN,ZTX("Too little overlap, cannot align"));
            panStat = 0;
            return 0;
         }
      }
   }

   return 0;
}


//  pre-align working thread
//  convert mouse and KB events into image movements

void * pan4_prealign_thread(void *)                                        //  v.10.7
{
   void   pan4_autolens();

   PXM         *pxmtemp;
   char        *ftemp;
   int         im1, im2, imm, imx;
   int         mx0, my0, mx, my;                                           //  mouse drag origin, position
   int         xoff, yoff, lox, hix, midx;
   int         ww, hh, rotate;
   double      lens_mm0, lens_bow0;
   double      dx, dy, t1, t2, dt;
      
   imm = ww = hh = rotate = midx = xoff = yoff = 0;                        //  stop compiler warnings

   lens_mm0 = lens_mm;                                                     //  to detect changes
   lens_bow0 = lens_bow;

   mx0 = my0 = 0;                                                          //  no drag in progress
   Mcapture = KBcapture = 1;                                               //  capture mouse drag and KB keys

   while (true)                                                            //  loop and align until done
   {
      zsleep(0.05);                                                        //  logic simplified
      
      if (panStat == 2) {                                                  //  dialog search button
         panStat = -1;                                                     //  back to busy status
         pan4_autolens();
      }
      
      if (panStat != -1) break;                                            //  quit signal from dialog
      
      if (lens_mm != lens_mm0 || lens_bow != lens_bow0) {                  //  change in lens parameters
         lens_mm0 = lens_mm;
         lens_bow0 = lens_bow;

         for (imx = 0; imx < cimNF; imx++) {                               //  re-curve images
            cim_scale_image(imx,cimPXMs);
            cim_curve_image(imx);
            PXM_free(cimPXMw[imx]);
            cimPXMw[imx] = PXM_copy(cimPXMs[imx]);
         }

         cim_show_images(1,0);                                             //  combine and show images
         continue;
      }
      
      if (KBkey) {                                                         //  KB input
         if (KBkey == GDK_Left)  cimOffs[imm].xf -= 0.5;                   //  tweak alignment offsets
         if (KBkey == GDK_Right) cimOffs[imm].xf += 0.5;
         if (KBkey == GDK_Up)    cimOffs[imm].yf -= 0.5;
         if (KBkey == GDK_Down)  cimOffs[imm].yf += 0.5;
         if (KBkey == GDK_r)     cimOffs[imm].tf += 0.0005;
         if (KBkey == GDK_l)     cimOffs[imm].tf -= 0.0005;
         KBkey = 0;

         cim_show_images(0,0);                                             //  combine and show images
         continue;
      }

      if (! Mxdrag && ! Mydrag)                                            //  no drag underway
         mx0 = my0 = 0;                                                    //  reset drag origin

      if (Mxdrag || Mydrag)                                                //  mouse drag underway
      {
         mx = Mxdrag;                                                      //  mouse position in image
         my = Mydrag;

         if (! mx0 && ! my0)                                               //  new drag
         {
            mx0 = mx;                                                      //  set drag origin
            my0 = my;
            
            for (im2 = cimNF-1; im2 > 0; im2--)                            //  find image at mouse position
            {
               im1 = im2 - 1;                                              //  im1 = image to left
               hix = cimOffs[im1].xf + cimPXMw[im1]->ww;                   //  right margin of image im1
               lox = cimOffs[im2].xf;                                      //  left margin of image im2
               midx = (lox + hix) / 2;                                     //  midpoint of overlap 
               if (mx0 > midx) break;                                      //  mouse is right of overlap
            }
            
            imm = im2;                                                     //  image of mouse position
            xoff = cimOffs[imm].xf;
            yoff = cimOffs[imm].yf;
            ww = cimPXMw[imm]->ww;
            hh = cimPXMw[imm]->hh;

            rotate = 0;                                                    //  if drag at bottom edge,
            if (my > yoff + 0.85 * hh) rotate = 1;                         //    set rotate flag
         }
         
         if (mx != mx0 || my != my0)                                       //  drag is progressing
         {
            dx = mx - mx0;                                                 //  mouse movement
            dy = my - my0;
            
            if (rotate && my0 > yoff && my > yoff) {                       //  rotation
               t1 = atan(1.0 * (mx0-xoff) / (my0-yoff));
               t2 = atan(1.0 * (mx-xoff) / (my-yoff));
               dt = t1 - t2;                                               //  angle change
               dx = dt * (hh/2 + yoff);                                    //  pivot = middle of overlap on left
               dy = -dt * (midx-xoff);
            }
            else  dt = 0;                                                  //  x/y drag

            cimOffs[imm].xf += dx;                                         //  update image
            cimOffs[imm].yf += dy;
            cimOffs[imm].tf += dt;

            if (imm > 0) {
               for (imx = imm+1; imx < cimNF; imx++)                       //  update images to right
               {
                  cimOffs[imx].xf += dx;                                   //  same x, y, theta change
                  cimOffs[imx].yf += dy;
                  cimOffs[imx].tf += dt;
                  ww = cimOffs[imx].xf - cimOffs[imm].xf;                  //  y change from rotation
                  cimOffs[imx].yf += ww * dt;
               }
            }

            cim_show_images(0,0);                                          //  show combined images

            mx0 = mx;                                                      //  next drag origin = current mouse
            my0 = my;
         }
      }

      for (im1 = 0; im1 < cimNF-1; im1++)                                  //  drag too far >> re-order images
      {
         im2 = im1 + 1;
         if (cimOffs[im2].xf < cimOffs[im1].xf + 0.2 * cimPXMw[im1]->ww) 
         {
            ftemp = cimFile[im2];                                          //  switch filespecs
            cimFile[im2] = cimFile[im1];
            cimFile[im1] = ftemp;

            pxmtemp = cimPXMf[im2];                                        //  switch images
            cimPXMf[im2] = cimPXMf[im1];
            cimPXMf[im1] = pxmtemp;
            pxmtemp = cimPXMs[im2];                                        //  scaled images
            cimPXMs[im2] = cimPXMs[im1];
            cimPXMs[im1] = pxmtemp;
            pxmtemp = cimPXMw[im2];                                        //  warped images
            cimPXMw[im2] = cimPXMw[im1];
            cimPXMw[im1] = pxmtemp;

            cimOffs[0].xf = 0;                                             //  reposition all, 30% overlaps
            for (imx = 1; imx < cimNF; imx++)
               cimOffs[imx].xf = cimOffs[imx-1].xf + 0.7 * cimPXMw[imx-1]->ww;
            cim_show_images(1,0);
            break;
         }
      }
   }
   
   KBcapture = Mcapture = 0;
   thread_exit();
   return 0;                                                               //  never executed, stop g++ warning
}


//  optimize lens parameters
//  inputs and outputs:
//       pre-aligned images cimPXMw[0] and [1]
//       offsets in cimOffs[0] and [1]
//       lens_mm, lens_bow

void pan4_autolens()                                                       //  v.10.7
{
   double      mm_range, bow_range, xf_range, yf_range, tf_range;
   double      squeeze, xf_rfinal, rnum, matchB, matchlev;
   double      lens_mmB, lens_bowB;
   int         imx, randcount = 0;
   cimoffs     offsetsB;

   set_cursor(busycursor);

   cimSampSize = panSampSize;
   cimNsearch = 0;

   mm_range = 0.2 * lens_mm;                                               //  set initial search ranges
   bow_range = 0.5 * lens_bow;
   if (bow_range < 1) bow_range = 1;
   xf_range = 7;
   yf_range = 7;
   tf_range = 0.01;
   xf_rfinal = 0.3;                                                        //  final xf range - when to quit

   cim_get_overlap(0,1,cimPXMw);

   cim_match_colors(0,1,cimPXMw);                                          //  adjust colors for image matching
   cim_adjust_colors(cimPXMs[0],1);
   cim_adjust_colors(cimPXMw[0],1);
   cim_adjust_colors(cimPXMs[1],2);
   cim_adjust_colors(cimPXMw[1],2);

   lens_mmB = lens_mm;                                                     //  starting point
   lens_bowB = lens_bow;
   offsetsB = cimOffs[1];

   cimSearchRange = 7;                                                     //  insure red pixels inside range
   cim_get_redpix(0);                                                      //  get high-contrast pixels

   matchB = 0;

   while (true)
   {
      srand48(time(0) + randcount++);
      lens_mm = lens_mmB + mm_range * (drand48() - 0.5);                   //  new random lens factors
      lens_bow = lens_bowB + bow_range * (drand48() - 0.5);                //     within search range
      
      for (imx = 0; imx <= 1; imx++) {                                     //  re-curve images
         cim_scale_image(imx,cimPXMs);
         cim_curve_image(imx);
         PXM_free(cimPXMw[imx]);
         cimPXMw[imx] = PXM_copy(cimPXMs[imx]);
      }

      cim_show_images(0,0);                                                //  combine and show images

      squeeze = 0.97;                                                      //  search range reduction  v.10.7
         
      for (int ii = 0; ii < 1000; ii++)                                    //  loop random x/y/t alignments
      {                                                                    
         rnum = drand48();
         if (rnum < 0.33)                                                  //  random change some alignment offset 
            cimOffs[1].xf = offsetsB.xf + xf_range * (drand48() - 0.5);
         else if (rnum < 0.67)
            cimOffs[1].yf = offsetsB.yf + yf_range * (drand48() - 0.5);
         else
            cimOffs[1].tf = offsetsB.tf + tf_range * (drand48() - 0.5);

         matchlev = cim_match_images(0,1);                                 //  test quality of image alignment

         sprintf(SB_text,"align: %d  match: %.5f  lens: %.1f %.2f",        //  update status bar
                           ++cimNsearch, matchB, lens_mmB, lens_bowB);
         SBupdate++;

         if (sigdiff(matchlev,matchB,0.00001) > 0) {
            matchB = matchlev;                                             //  save new best fit
            lens_mmB = lens_mm;                                            //  alignment is better
            lens_bowB = lens_bow;
            offsetsB = cimOffs[1];
            cim_show_images(0,0);
            squeeze = 1;                                                   //  keep same search range as long
            break;                                                         //    as improvements are found
         }

         if (panStat != -1) goto done;                                     //  user kill
      }
      
      if (xf_range < xf_rfinal) goto done;

      sprintf(SB_text,"align: %d  match: %.5f  lens: %.1f %.2f",           //  update status bar
                        cimNsearch, matchB, lens_mmB, lens_bowB);
      SBupdate++;

      mm_range = squeeze * mm_range;                                       //  reduce search range if no 
      if (mm_range < 0.02 * lens_mmB) mm_range = 0.02 * lens_mmB;          //    improvements were found
      bow_range = squeeze * bow_range;
      if (bow_range < 0.1 * lens_bowB) bow_range = 0.1 * lens_bowB;
      if (bow_range < 0.2) bow_range = 0.2;
      xf_range = squeeze * xf_range;
      yf_range = squeeze * yf_range;
      tf_range = squeeze * tf_range;
   }

done:
   zfree(cimRedpix);
   cimRedpix = 0;

   lens_mm = lens_mmB;                                                     //  save best lens params found
   lens_bow = lens_bowB;
   if (panStat == -1 && pan4zd) {
      zdialog_stuff(pan4zd,"spmm",lens_mm);
      zdialog_stuff(pan4zd,"spbow",lens_bow);
   }

   set_cursor(0);
   cim_show_images(1,0);                                                   //  images are left color-matched
   return;
}


//  fine-alignment
//  start with very small image size
//  search around offset values for best match
//  increase image size and loop until full-size

void pan4_align()                                                          //  v.10.7
{
   int         imx, im1, im2, ww;
   double      R, dx, dy, dt;
   cimoffs     offsets0;
   
   Fzoom = 0;                                                              //  scale E3 to fit window
   Fblowup = 1;                                                            //  magnify small image to window size

   set_cursor(busycursor);

   for (imx = 0; imx < cimNF; imx++) {
      cimOffs[imx].xf = cimOffs[imx].xf / cimScale;                        //  scale x/y offsets for full-size images
      cimOffs[imx].yf = cimOffs[imx].yf / cimScale;
   }

   cimScale = 1.0;                                                         //  full-size

   for (imx = 0; imx < cimNF; imx++) {
      PXM_free(cimPXMs[imx]);
      cimPXMs[imx] = PXM_copy(cimPXMf[imx]);                               //  copy full-size images
      cim_curve_image(imx);                                                //  curve them
   }

   cimBlend = 0.3 * cimPXMs[0]->ww;
   cim_get_overlap(0,1,cimPXMs);                                           //  match images 0 & 1 in overlap area
   cim_match_colors(0,1,cimPXMs);
   cim_adjust_colors(cimPXMf[0],1);                                        //  image 0 << profile 1
   cim_adjust_colors(cimPXMf[1],2);                                        //  image 1 << profile 2

   if (cimNF > 2) {
      cimBlend = 0.3 * cimPXMs[1]->ww;
      cim_get_overlap(1,2,cimPXMs);
      cim_match_colors(1,2,cimPXMs);
      cim_adjust_colors(cimPXMf[0],1);
      cim_adjust_colors(cimPXMf[1],1);
      cim_adjust_colors(cimPXMf[2],2);
   }

   if (cimNF > 3) {
      cimBlend = 0.3 * cimPXMs[2]->ww;
      cim_get_overlap(2,3,cimPXMs);
      cim_match_colors(2,3,cimPXMs);
      cim_adjust_colors(cimPXMf[0],1);
      cim_adjust_colors(cimPXMf[1],1);
      cim_adjust_colors(cimPXMf[2],1);
      cim_adjust_colors(cimPXMf[3],2);
   }

   cimScale = panInitAlignSize / cimPXMf[1]->hh;                           //  initial align image scale
   if (cimScale > 1.0) cimScale = 1.0;

   for (imx = 0; imx < cimNF; imx++) {                                     //  scale offsets for image scale
      cimOffs[imx].xf = cimOffs[imx].xf * cimScale;
      cimOffs[imx].yf = cimOffs[imx].yf * cimScale;
   }
   
   cimSearchRange = panInitSearchRange;                                    //  initial align search range
   cimSearchStep = panInitSearchStep;                                      //  initial align search step
   cimWarpRange = panInitWarpRange;                                        //  initial align corner warp range
   cimWarpStep = panInitWarpStep;                                          //  initial align corner warp step
   ww = cimPXMf[0]->ww * cimScale;                                         //  initial align image width
   cimBlend = ww * panInitBlend;                                           //  initial align blend width
   cimSampSize = panSampSize;                                              //  pixel sample size for align/compare
   cimNsearch = 0;                                                         //  reset align search counter
   cimPano = 1;                                                            //  pano mode for cim_align_image()

   while (true)                                                            //  loop, increasing image size
   {
      for (imx = 0; imx < cimNF; imx++) {                                  //  prepare images
         cim_scale_image(imx,cimPXMs);                                     //  scale to new size
         cim_curve_image(imx);                                             //  curve based on lens params
         cim_warp_image_pano(imx,1);                                       //  apply corner warps
      }
      
      cim_show_images(1,0);                                                //  show with 50/50 blend

      for (im1 = 0; im1 < cimNF-1; im1++)                                  //  fine-align each image with left neighbor
      {
         im2 = im1 + 1;

         offsets0 = cimOffs[im2];                                          //  save initial alignment offsets
         cim_get_overlap(im1,im2,cimPXMw);                                 //  get overlap area
         cim_get_redpix(im1);                                              //  get high-contrast pixels
         
         cim_align_image(im1,im2);                                         //  search for best offsets and warps

         zfree(cimRedpix);                                                 //  clear red pixels
         cimRedpix = 0;

         dx = cimOffs[im2].xf - offsets0.xf;                               //  changes from initial offsets
         dy = cimOffs[im2].yf - offsets0.yf;
         dt = cimOffs[im2].tf - offsets0.tf;
         
         for (imx = im2+1; imx < cimNF; imx++)                             //  propagate to following images
         {
            cimOffs[imx].xf += dx;
            cimOffs[imx].yf += dy;
            cimOffs[imx].tf += dt;
            ww = cimOffs[imx].xf - cimOffs[im2].xf;
            cimOffs[imx].yf += ww * dt;
         }
      }

      if (cimScale == 1.0) break;                                          //  done

      R = panImageIncrease;                                                //  next larger image size
      cimScale = cimScale * R;
      if (cimScale > 0.85) {                                               //  if close to end, jump to end
         R = R / cimScale;
         cimScale = 1.0;
      }

      for (imx = 0; imx < cimNF; imx++)                                    //  scale offsets for new size
      {
         cimOffs[imx].xf *= R;
         cimOffs[imx].yf *= R;

         for (int ii = 0; ii < 4; ii++) {
            cimOffs[imx].wx[ii] *= R;
            cimOffs[imx].wy[ii] *= R;
         }
      }
      
      cimSearchRange = panSearchRange;                                     //  align search range
      cimSearchStep = panSearchStep;                                       //  align search step size
      cimWarpRange = panWarpRange;                                         //  align corner warp range
      cimWarpStep = panWarpStep;                                           //  align corner warp step size

      cimBlend = cimBlend * panBlendDecrease * R;                          //  blend width, reduced
      ww = cimPXMf[0]->ww * cimScale;
      if (cimBlend < panFinalBlend * ww) 
         cimBlend = panFinalBlend * ww;                                    //  stay above minimum
   }
   
   cimBlend = 1;                                                           //  no blend (add in tweak if wanted)
   Fzoom = Fblowup = 0;
   panStat = 1;
   cim_show_images(0,0);
   set_cursor(0);
   return;
}


//  get user inputs for RGB changes and blend width, update cimPXMw[*]

void pan4_tweak()                                                          //  v.10.7
{
   int    pan4_tweak_event(zdialog *zd, cchar *event);                     //  dialog event function
   
   cchar    *tweaktitle = ZTX("Match Brightness and Color");
   char     imageN[8] = "imageN";
   int      imx;
   
   cimBlend = 1;                                                           //  init. blend width
   
   pan4zd = zdialog_new(tweaktitle,mWin,Bdone,Bcancel,null);

   zdialog_add_widget(pan4zd,"hbox","hbim","dialog",0,"space=5");
   zdialog_add_widget(pan4zd,"label","labim","hbim",ZTX("image"),"space=5");     //  image  (o)  (o)  (o)  (o)
   zdialog_add_widget(pan4zd,"hbox","hbc1","dialog",0,"homog");                  //
   zdialog_add_widget(pan4zd,"label","labred","hbc1",Bred);                      //     red     green    blue
   zdialog_add_widget(pan4zd,"label","labgreen","hbc1",Bgreen);                  //   [_____]  [_____]  [_____]
   zdialog_add_widget(pan4zd,"label","labblue","hbc1",Bblue);                    //
   zdialog_add_widget(pan4zd,"hbox","hbc2","dialog",0,"homog");                  //   brightness [___]  [apply]
   zdialog_add_widget(pan4zd,"spin","red","hbc2","50|200|0.1|100","space=5");    //
   zdialog_add_widget(pan4zd,"spin","green","hbc2","50|200|0.1|100","space=5");  //   --------------------------
   zdialog_add_widget(pan4zd,"spin","blue","hbc2","50|200|0.1|100","space=5");   //
   zdialog_add_widget(pan4zd,"hbox","hbbri","dialog",0,"space=5");               //   [auto color]  [file color]
   zdialog_add_widget(pan4zd,"label","labbr","hbbri",Bbrightness,"space=5");     //
   zdialog_add_widget(pan4zd,"spin","bright","hbbri","50|200|0.1|100");          //   --------------------------
   zdialog_add_widget(pan4zd,"button","brapp","hbbri",Bapply,"space=10");        //
   zdialog_add_widget(pan4zd,"hsep","hsep","dialog",0,"space=5");                //   blend width [___] [apply]   
   zdialog_add_widget(pan4zd,"hbox","hbc3","dialog",0,"space=5");                //
   zdialog_add_widget(pan4zd,"button","auto","hbc3",ZTX("auto color"),"space=5");//          [done]  [cancel]
   zdialog_add_widget(pan4zd,"button","file","hbc3",ZTX("file color"),"space=5");
   zdialog_add_widget(pan4zd,"hsep","hsep","dialog",0,"space=5");
   zdialog_add_widget(pan4zd,"hbox","hbblen","dialog",0);
   zdialog_add_widget(pan4zd,"label","labbl","hbblen",Bblendwidth,"space=5");
   zdialog_add_widget(pan4zd,"spin","blend","hbblen","0|300|1|1");
   zdialog_add_widget(pan4zd,"button","blapp","hbblen",Bapply,"space=15");
   
   for (imx = 0; imx < cimNF; imx++) {                                     //  add radio button per image
      imageN[5] = '0' + imx;
      zdialog_add_widget(pan4zd,"radio",imageN,"hbim",0,"space=5");
   }
   
   zdialog_stuff(pan4zd,"image0",1);                                       //  pre-select 1st image
   zdialog_resize(pan4zd,300,0);

   panStat = -1;                                                           //  busy status
   zdialog_run(pan4zd,pan4_tweak_event);                                   //  run dialog, parallel
   zdialog_wait(pan4zd);                                                   //  wait for dialog completion
   return;
}


//  dialog event function

int pan4_tweak_event(zdialog *zd, cchar *event)                            //  v.10.7
{
   char        imageN[8] = "imageN";
   double      red, green, blue, bright, bright2;
   double      red1, green1, blue1;
   int         nn, im0, imx, im1, im2, ww, hh, px, py;
   uint16      *pixel;
   
   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) panStat = 1;                                     //  done
      if (zd->zstat == 2) panStat = 0;                                     //  cancel
      zdialog_free(zd);                                                    //  kill dialog
      pan4zd = 0;
      return 0;
   }

   for (im0 = 0; im0 < cimNF; im0++) {                                     //  get which image is selected
      imageN[5] = '0' + im0;                                               //    by the radio buttons
      zdialog_fetch(zd,imageN,nn);
      if (nn) break;
   }
   if (im0 == cimNF) return 1;

   zdialog_fetch(zd,"red",red);                                            //  get color adjustments
   zdialog_fetch(zd,"green",green);
   zdialog_fetch(zd,"blue",blue);
   zdialog_fetch(zd,"bright",bright);                                      //  brightness adjustment

   bright2 = (red + green + blue) / 3;                                     //  RGB brightness
   bright = bright / bright2;                                              //  bright setpoint / RGB brightness
   red = red * bright;                                                     //  adjust RGB brightness
   green = green * bright;
   blue = blue * bright;
   
   bright = (red + green + blue) / 3;
   zdialog_stuff(zd,"red",red);                                            //  force back into consistency
   zdialog_stuff(zd,"green",green);
   zdialog_stuff(zd,"blue",blue);
   zdialog_stuff(zd,"bright",bright);
   
   if (strEqu(event,"brapp"))                                              //  apply color & brightness changes
   {
      red = red / 100;                                                     //  normalize 0.5 ... 2.0
      green = green / 100;
      blue = blue / 100;
      
      cim_warp_image_pano(im0,0);                                          //  refresh cimPXMw from cimPXMs

      ww = cimPXMw[im0]->ww;
      hh = cimPXMw[im0]->hh;
      
      for (py = 0; py < hh; py++)                                          //  loop all image pixels
      for (px = 0; px < ww; px++)
      {
         pixel = PXMpix(cimPXMw[im0],px,py);
         red1 = red * pixel[0];                                            //  apply color factors
         green1 = green * pixel[1];
         blue1 = blue * pixel[2];
         if (! blue1) continue;

         if (red1 > 65535 || green1 > 65535 || blue1 > 65535) {
            bright = red1;                                                 //  avoid overflow
            if (green1 > bright) bright = green1;
            if (blue1 > bright) bright = blue1;
            bright = 65535.0 / bright;
            red1 = red1 * bright;
            green1 = green1 * bright;
            blue1 = blue1 * bright;
         }
         
         if (blue1 < 1) blue1 = 1;                                         //  avoid 0       v.10.7

         pixel[0] = red1;
         pixel[1] = green1;
         pixel[2] = blue1;
      }
      
      cimBlend = 1;
      cim_show_images(0,0);                                                //  combine and show with 50/50 blend
   }
   
   if (strEqu(event,"auto"))                                               //  auto match color of selected image
   {
      for (im1 = im0; im1 < cimNF-1; im1++)                                //  from selected image to last image
      {
         im2 = im1 + 1;
         cimBlend = 0.3 * cimPXMw[im2]->ww;
         cim_get_overlap(im1,im2,cimPXMw);                                 //  match images in overlap area
         cim_match_colors(im1,im2,cimPXMw);
         cim_adjust_colors(cimPXMw[im1],1);                                //  image im1 << profile 1
         cim_adjust_colors(cimPXMw[im2],2);                                //  image im2 << profile 2
         for (imx = im1-1; imx >= im0; imx--)
            cim_adjust_colors(cimPXMw[imx],1);
         cimBlend = 1;
         cim_show_images(0,0);
      }

      for (im1 = im0-1; im1 >= 0; im1--)                                   //  from selected image to 1st image
      {
         im2 = im1 + 1;
         cimBlend = 0.3 * cimPXMw[im2]->ww;
         cim_get_overlap(im1,im2,cimPXMw);                                 //  match images in overlap area
         cim_match_colors(im1,im2,cimPXMw);
         cim_adjust_colors(cimPXMw[im1],1);                                //  image im1 << profile 1
         cim_adjust_colors(cimPXMw[im2],2);                                //  image im2 << profile 2
         for (imx = im2+1; imx < cimNF; imx++)
            cim_adjust_colors(cimPXMw[imx],2);
         cimBlend = 1;
         cim_show_images(0,0);
      }
   }
   
   if (strEqu(event,"file"))                                               //  use original file colors
   {
      if (! cim_load_files()) return 1;

      for (imx = 0; imx < cimNF; imx++) {
         PXM_free(cimPXMs[imx]);
         cimPXMs[imx] = PXM_copy(cimPXMf[imx]);
         cim_curve_image(imx);                                             //  curve and warp
         cim_warp_image_pano(imx,0);
      }

      cimBlend = 1;
      cim_show_images(0,0);
   }
   
   if (strEqu(event,"blapp"))                                              //  apply new blend width
   {
      zdialog_fetch(zd,"blend",cimBlend);                                  //  can be zero
      cim_show_images(0,1);                                                //  show with gradual blend
   }

   return 1;
}


/**************************************************************************

   edit transaction and thread support functions

   edit transaction management
      edit_setup()                     start new edit - copy E3 > E1
      edit_cancel()                    cancel edit - E1 > E3, delete E1
      edit_done()                      commit edit - add to undo stack
      edit_undo()                      undo edit - E1 > E3
      edit_redo()                      redo edit - run thread again
      edit_fullsize()                  convert preview to full-size pixmaps

   main level thread management
      start_thread(func,arg)           start thread running
      signal_thread()                  signal thread that work is pending
      wait_thread_idle()               wait for pending work complete
      wrapup_thread(command)           wait for exit or command thread exit
      thread_working()                 return idle/working status

   thread function
      thread_idle_loop()               wait for pending work, exit if commanded
      thread_exit()                    exit thread unconditionally
      
   thread_status (thread ownership
      0     no thread is running
      1     thread is running and idle (no work)
      2     thread is working
      0     thread has exited

   thread_command (main program ownership)
      0     idle, no work pending
      8     exit when pending work is done
      9     exit now, unconditionally

   thread_pend       work requested counter
   thread_done       work done counter
   thread_hiwater    high water mark
   edit_action       done/cancel/undo/redo in progress

***************************************************************************/

int      thread_command = 0, thread_status = 0;
int      thread_pend = 0, thread_done = 0, thread_hiwater = 0;
int      edit_action = 0;

/**************************************************************************

  Setup for a new edit transaction
  Create E1 (edit input) and E3 (edit output) pixmaps from
  previous edit output (Fpxm16) or image file (new Fpxm16).

  uprev      0     edit full-size image
             1     edit preview image unless select area exists

  uarea      0     select_area is invalid and will be deleted (e.g. rotate)
             1     select_area not used but remains valid (e.g. red-eye)
             2     select_area is used and remains valid (e.g. flatten)

***************************************************************************/

int edit_setup(cchar *func, int uprev, int uarea)
{
   int      yn;
   PXM      *pxm;

   if (! image_file) return 0;                                             //  no image file
   if (! menulock(1)) return 0;                                            //  lock menu
   
   if (! Fexiftool && ! Fexifwarned) {
      zmessageACK(MWIN,ZTX("exiftool is not installed \n"                  //  warn if starting to edit
                           "edited images will lose EXIF data"));          //    and exiftool is missing  v.9.9
      Fexifwarned = 1;
   }

   if (Pundo > maxedits-2) {                                               //  v.10.2
      zmessageACK(MWIN,ZTX("Too many edits, please save image"));
      menulock(0);
      return 0;
   }

   if (Frotated) {
      mutex_lock(&pixmaps_lock);                                           //  reverse manual rotation  v.10.5
      pxm = PXM_rotate(Fpxm8,-Frotated);
      PXM_free(Fpxm8);
      Fpxm8 = pxm;
      Fww = pxm->ww;
      Fhh = pxm->hh;
      mutex_unlock(&pixmaps_lock);
      mwpaint();
      Frotated = 0;
   }
   
   if (uarea == 0 && sa_stat) {                                            //  select area will be lost, warn user
      yn = zmessageYN(MWIN,ZTX("Select area cannot be kept.\n"
                               "Continue?"));
      if (! yn) {
         menulock(0);
         return 0;
      }
      sa_delete();
      zdialog_free(zdsela);
      zdsela = 0;
   }
   
   if (uprev && uarea == 2 && sa_stat && ! sa_active) {                    //  select area exists and can be used,
      yn = zmessageYN(MWIN,ZTX("Select area not active.\n"                 //    but not active, ask user    v.10.1
                               "Continue?"));
      if (! yn) {
         menulock(0);
         return 0;
      }
   }
   
   Fpreview = 0;                                                           //  use preview image if supported
   if (uprev && ! (uarea == 2 && sa_active)) {                             //    and select area will not be used
      Fpreview = 1;
      if (Fzoom) {
         Fzoom = 0;                                                        //  bugfix, mpaint() req.    v.10.3.1
         mwpaint();
      }
   }

   if (Fpreview && Fshowarea) {                                            //  hide select area during preview mode edit
      sa_show(0);                                                          //    and bring it back when done   v.9.7
      Fshowarea = 1;
   }
   
   mutex_lock(&pixmaps_lock);                                              //  lock pixmaps

   if (! Fpxm16) Fpxm16 = f_load(image_file,16);                           //  create Fpxm16 if not already
   if (! Fpxm16) {
      mutex_unlock(&pixmaps_lock);                                         //  should not happen      v.10.9
      menulock(0);
      printf("f_load(file,16) failed \n");
      return 0;
   }

   PXM_free(E1pxm16);                                                      //  free prior edit pixmaps
   PXM_free(E3pxm16);

   if (Fpreview)                                                           //  edit pixmaps are window-size
      E1pxm16 = PXM_rescale(Fpxm16,dww,dhh);                               //  E1pxm16 = Fpxm16 scaled to window
   else E1pxm16 = PXM_copy(Fpxm16);                                        //  edit pixmaps are full-size
   E3pxm16 = PXM_copy(E1pxm16);                                            //  E1 >> E3

   E1ww = E3ww = E1pxm16->ww;
   E1hh = E3hh = E1pxm16->hh;
   
   if (Pundo == 0) {
      edit_function = "initial";                                           //  initial image >> undo stack    v.10.2
      save_undo();
   }

   edit_function = func;                                                   //  set current edit function    v.10.2
   Fmodified = 0;                                                          //  image not modified yet
   thread_command = thread_status = Fkillfunc = 0;                         //  no thread running
   thread_pend = thread_done = thread_hiwater = 0;                         //  no work pending or done

   mutex_unlock(&pixmaps_lock);
   mwpaint2();
   return 1;
}


/**************************************************************************/

//  process edit cancel

void edit_cancel()
{
   if (edit_action) return;
   edit_action++;

   Fkillfunc = 1;                                                          //  tell thread to quit now  v.9.1
   wrapup_thread(9);                                                       //  wait for exit
   Fkillfunc = 0;

   if (zdedit) {
      zdialog_free(zdedit);                                                //  kill dialog
      zdedit = null;
   }

   mutex_lock(&pixmaps_lock);
   PXM_free(E1pxm16);                                                      //  free edit pixmaps E1, E3
   PXM_free(E3pxm16);
   PXM_free(ERpxm16);                                                      //  free redo copy    v.10.3
   E1pxm16 = E3pxm16 = ERpxm16 = 0;
   E1ww = E3ww = ERww = 0;

   Fmodified = Fpreview = 0;                                               //  reset flags
   Ntoplines = Nptoplines;                                                 //  no overlay lines
   paint_toparc(2);                                                        //  no brush outline
   mutex_unlock(&pixmaps_lock);
   menulock(0);                                                            //  unlock menu
   mwpaint2();                                                             //  refresh window
   edit_action = 0;
   return;
}   


/**************************************************************************/

//  process edit dialog [done]  
//  E3pxm16 >> Fpxm16 >> Fpxm8

void edit_done()
{
   if (edit_action) return;
   edit_action++;

   wait_thread_idle();                                                     //  bugfix           v.10.2

   if (Fpreview && Fmodified) {
      Fzoom = 0;
      edit_fullsize();                                                     //  update full image
   }

   wrapup_thread(8);                                                       //  tell thread to exit

   if (zdedit) {
      zdialog_free(zdedit);                                                //  kill dialog
      zdedit = null;
   }
   
   mutex_lock(&pixmaps_lock);

   if (Fmodified) {
      PXM_free(Fpxm16);
      Fpxm16 = PXM_copy(E3pxm16);                                          //  E3 >> Fpxm16
      PXM_free(Fpxm8);
      Fpxm8 = PXM_convbpc(Fpxm16);                                         //  Fpxm16 >> Fpxm8
      Fww = Fpxm8->ww;
      Fhh = Fpxm8->hh;
      Pundo++;
      Pumax = Pundo;
      save_undo();                                                         //  save next undo state
   }

   PXM_free(E1pxm16);                                                      //  free edit pixmaps
   PXM_free(E3pxm16);
   PXM_free(ERpxm16);                                                      //  free redo copy    v.10.3
   E1pxm16 = E3pxm16 = ERpxm16 = 0;
   E1ww = E3ww = ERww = 0;
   
   Fmodified = Fpreview = 0;                                               //  reset flags
   Ntoplines = Nptoplines;                                                 //  no overlay lines
   paint_toparc(2);                                                        //  no brush outline
   mutex_unlock(&pixmaps_lock);
   menulock(0);                                                            //  unlock menu
   mwpaint2();                                                             //  update window
   edit_action = 0;
   return;
}


/**************************************************************************/

//  edit undo, redo, reset functions

void edit_undo()
{
   if (! Fmodified) return;
   if (thread_status == 2) return;                                         //  bugfix, thread busy      v.10.2
   if (edit_action) return;
   edit_action++;
   
   mutex_lock(&pixmaps_lock);
   PXM_free(ERpxm16);                                                      //  make redo copy    v.10.3
   ERpxm16 = E3pxm16;
   ERww = E3ww;
   ERhh = E3hh;
   E3pxm16 = PXM_copy(E1pxm16);                                            //  restore initial image status
   E3ww = E1ww;
   E3hh = E1hh;
   Fmodified = 0;                                                          //  reset image modified status
   mutex_unlock(&pixmaps_lock);

   mwpaint2();                                                             //  refresh window      v.9.8
   edit_action = 0;
   return;
}


void edit_redo()
{
   if (! ERpxm16) return;
   if (edit_action) return;
   edit_action++;
   PXM_free(E3pxm16);                                                      //  edit image = redo copy    v.10.3
   E3pxm16 = ERpxm16;
   E3ww = ERww;
   E3hh = ERhh;
   ERpxm16 = 0;
   Fmodified = 1;
   mwpaint2();
   edit_action = 0;
   return;
}


void edit_reset()                                                          //  new  v.10.3
{
   if (! Fmodified) return;
   if (thread_status == 2) return;
   if (edit_action) return;
   edit_action++;

   mutex_lock(&pixmaps_lock);
   PXM_free(ERpxm16);                                                      //  no redo copy
   ERpxm16 = 0;
   PXM_free(E3pxm16);
   E3pxm16 = PXM_copy(E1pxm16);                                            //  restore initial image
   E3ww = E1ww;
   E3hh = E1hh;
   Fmodified = 0;                                                          //  reset image modified status
   mutex_unlock(&pixmaps_lock);

   mwpaint2();                                                             //  refresh window      v.9.8
   edit_action = 0;
   return;
}


void edit_zapredo()
{
   PXM_free(ERpxm16);                                                      //  no redo copy
   ERpxm16 = 0;
   return;
}


/**************************************************************************/

//  Convert from preview mode (window-size pixmaps) to full-size pixmaps.

void edit_fullsize()
{
   if (! Fpreview) return;
   Fpreview = 0;

   mutex_lock(&pixmaps_lock);
   PXM_free(E1pxm16);                                                      //  free preview pixmaps
   PXM_free(E3pxm16);
   E1pxm16 = PXM_copy(Fpxm16);                                             //  make full-size pixmaps
   E3pxm16 = PXM_copy(Fpxm16);
   E1ww = E3ww = Fww;
   E1hh = E3hh = Fhh;
   mutex_unlock(&pixmaps_lock);

   signal_thread();                                                        //  signal thread to do edit
   wait_thread_idle();
   return;
}


/**************************************************************************
      undo / redo toolbar buttons
***************************************************************************/

//  [undo] menu function - reinstate previous edit in undo/redo stack

void m_undo(GtkWidget *, cchar *)
{
   if (zdedit) {                                                           //  v.10.2
      zdialog_send_event(zdedit,"undo");
      return;
   }
   if (Pundo == 0) return;
   if (! menulock(1)) return;
   Pundo--;
   load_undo();
   menulock(0);
   Frotated = 0;                                                           //  v.10.5
   return;
}


//  [redo] menu function - reinstate next edit in undo/redo stack

void m_redo(GtkWidget *, cchar *)
{
   if (zdedit) {                                                           //  v.10.2
      zdialog_send_event(zdedit,"redo");
      return;
   }
   if (Pundo == Pumax) return;
   if (! menulock(1)) return;
   Pundo++;
   load_undo();
   menulock(0);
   Frotated = 0;                                                           //  v.10.5
   return;
}


//  Save Fpxm16 to undo/redo file stack
//  stack position = Pundo

void save_undo()
{
   char     *pp, buff[24];
   int      fid, cc, cc2;
   
   pp = strstr(undo_files,"_undo_");
   if (! pp) zappcrash("undo/redo stack corrupted 1");
   snprintf(pp+6,3,"%02d",Pundo);
   
   fid = open(undo_files,O_WRONLY|O_CREAT|O_TRUNC,0640);
   if (! fid) zappcrash("undo/redo stack corrupted 2");

   snprintf(buff,24," %05d %05d fotoxx ",Fww,Fhh);
   cc = write(fid,buff,20);
   if (cc != 20) zappcrash("undo/redo stack corrupted 3");
   
   cc = Fww * Fhh * 6;
   cc2 = write(fid,Fpxm16->bmp,cc);
   if (cc2 != cc) zappcrash("undo/redo stack corrupted 4");

   close(fid);

   pvlist_replace(editlog,Pundo,edit_function);                            //  save log of edits done    v.10.2
   return;
}


//  Load Fpxm16 from undo/redo file stack
//  stack position = Pundo

void load_undo()
{
   char     *pp, buff[24], fotoxx[8];
   int      fid, ww, hh, cc, cc2;

   pp = strstr(undo_files,"_undo_");
   if (! pp) zappcrash("undo/redo stack corrupted 1");
   snprintf(pp+6,3,"%02d",Pundo);
   
   fid = open(undo_files,O_RDONLY);
   if (! fid) zappcrash("undo/redo stack corrupted 2");
   
   *fotoxx = 0;
   cc = read(fid,buff,20);
   sscanf(buff," %d %d %8s ",&ww, &hh, fotoxx);
   if (! strEqu(fotoxx,"fotoxx")) zappcrash("undo/redo stack corrupted 4");

   mutex_lock(&pixmaps_lock);

   PXM_free(Fpxm16);
   Fpxm16 = PXM_make(ww,hh,16);
   cc = ww * hh * 6;
   cc2 = read(fid,Fpxm16->bmp,cc);
   if (cc2 != cc) zappcrash("undo/redo stack corrupted 5");
   close(fid);
   
   PXM_free(Fpxm8);
   Fpxm8 = PXM_convbpc(Fpxm16);
   Fww = ww;
   Fhh = hh;
   
   edit_function = pvlist_get(editlog,Pundo);                              //  last edit func not un-done   v.10.2

   mutex_unlock(&pixmaps_lock);
   mwpaint2();
   return;
}


/**************************************************************************/

//  start thread that does the edit work

void start_thread(threadfunc func, void *arg)
{
   thread_status = 1;                                                      //  thread is running
   thread_command = thread_pend = thread_done = thread_hiwater = 0;        //  nothing pending
   start_detached_thread(func,arg);
   return;
}


//  signal thread that work is pending

void signal_thread()
{
   edit_zapredo();                                                         //  reset redo copy
   if (thread_status > 0) thread_pend++;
   return;
}


//  wait for edit thread to complete pending work and become idle

void wait_thread_idle()
{
   while (thread_status && thread_pend > thread_done)
   {
      zmainloop();
      zsleep(0.01);
   }
   
   return;
}


//  wait for thread exit or command thread exit
//  command = 0    wait for normal completion
//            8    finish pending work and exit
//            9    quit, exit now

void wrapup_thread(int command)
{
   thread_command = command;                                               //  tell thread to quit or finish

   while (thread_status > 0)                                               //  wait for thread to finish
   {                                                                       //    pending work and exit
      zmainloop();
      zsleep(0.01);
   }

   return;
}


//  return thread idle/working status

int thread_working()
{
   if (thread_status == 2) return 1;
   return 0;
}


//  called only from edit threads
//  idle loop - wait for work request or exit command

void thread_idle_loop()
{
   thread_status = 1;                                                      //  status = idle
   thread_done = thread_hiwater;                                           //  work done = high-water mark

   while (true)
   {
      if (thread_command == 9) thread_exit();                              //  quit now command
      if (thread_command == 8)                                             //  finish work and exit
         if (thread_pend <= thread_done) thread_exit();
      if (thread_pend > thread_done) break;                                //  wait for work request
      zsleep(0.01);
   }
   
   thread_hiwater = thread_pend;                                           //  set high-water mark
   thread_status = 2;                                                      //  thread is working
   return;                                                                 //  loop to thread
}


//  called only from edit threads
//  exit thread unconditionally

void thread_exit()
{
   thread_pend = thread_done = thread_hiwater = 0;
   thread_status = 0;
   pthread_exit(0);                                                        //  "return" cannot be used here
}


/**************************************************************************/

//  edit support functions for working threads (per processor core)

void start_wt(threadfunc func, void *arg)                                  //  start and increment busy count
{
   zadd_locked(wthreads_busy,+1);
   start_detached_thread(func,arg);
   return;
}


void exit_wt()                                                             //  decrement busy count and exit
{
   zadd_locked(wthreads_busy,-1);
   pthread_exit(0);                                                        //  "return" cannot be used here  v.9.4
}


void wait_wts()                                                            //  wait for all working threads done
{
   while (wthreads_busy) 
   {
      zmainloop();
      zsleep(0.01);
   }
   
   return;
}


/**************************************************************************
      other support functions
***************************************************************************/

//  help menu function

void m_help(GtkWidget *, cchar *menu)
{
   if (strEqu(menu,ZTX("About"))) 
      zmessageACK(MWIN,"%s \n%s \n%s \n\n%s \n\n%s \n\n%s",
        fversion,flicense,fhomepage,fcredits,ftranslators,fcontact);
      
   if (strEqu(menu,ZTX("User Guide"))) 
      showz_userguide();

   if (strEqu(menu,ZTX("Help")))                                           //  toolbar button   v.10.4
      showz_userguide();

   if (strEqu(menu,"README"))
      showz_readme();

   if (strEqu(menu,ZTX("Change Log")))
      showz_changelog();

   if (strEqu(menu,ZTX("Translate")))
      showz_translations();
      
   if (strEqu(menu,ZTX("Home Page")))
      showz_html(fhomepage);

   return;
}


/**************************************************************************/

//  restore state data from prior session
//  OBSOLETE from v.10.9

int load_old_fotoxx_state()
{
   int            ww, hh, ii, np, err;
   FILE           *fid;
   char           buff[maxfcc],text[100], *pp;
   float          parms[2];
   
   snprintf(buff,maxfcc,"%s/saved_state",get_zuserdir());                  //  open saved state file
   fid = fopen(buff,"r");
   if (! fid) return 0;
   
   pp = fgets_trim(buff,maxfcc,fid,1);                                     //  read last image file

   if (pp && *pp == '/') {
      if (image_file) zfree(image_file);                                   //  set current image file
      image_file = strdupz(pp,0,"image_file");
      if (image_dirk) zfree(image_dirk);
      image_dirk = strdupz(image_file,0,"imagedirk");                      //  and image directory    v.10.7
      pp = strrchr(image_dirk,'/');
      if (pp) *pp = 0;
      err = chdir(image_dirk);
   }

   pp = fgets_trim(buff,maxfcc,fid,1);                                     //  top directory
   if (pp && *pp == '/') topdirk = strdupz(pp,0,"topdirk");

   pp = fgets(buff,maxfcc,fid);                                            //  main window size
   if (pp) {
      ww = hh = 0;
      sscanf(buff," %d %d ",&ww,&hh);
      if (ww > 200 && ww < 3000) Dww = ww;
      if (hh > 200 && hh < 2000) Dhh = hh;
   }

   pp = fgets(buff,maxfcc,fid);                                            //  image gallery window size
   if (pp) {
      ww = hh = 0;
      sscanf(buff," %d %d ",&ww,&hh);
      if (ww > 200 && ww < 3000) image_navi::xwinW = ww;
      if (hh > 200 && hh < 2000) image_navi::xwinH = hh;
   }

   pp = fgets_trim(buff,maxfcc,fid,1);                                     //  thumbnail image size
   if (pp) {
      sscanf(buff," %d ",&ww);
      if (ww > 32 && ww < 256) image_navi::thumbsize = ww;
   }

   for (ii = 0; ii < 4; ii++)                                              //  4 sets of lens parameters
   {   
      pp = fgets_trim(buff,maxfcc,fid,1);
      if (! pp) break;
      np = sscanf(buff," %s %f %f ",text,&parms[0],&parms[1]);
      if (np != 3) break;
      strncpy0(lens4_name[ii],text,lens_cc);
      lens4_mm[ii] = parms[0];
      lens4_bow[ii] = parms[1];
   }
   
   pp = fgets_trim(buff,maxfcc,fid,1);                                     //  current lens
   if (pp) {
      sscanf(buff," %f ",&parms[0]);
      if (parms[0] >= 0 && parms[0] < 4) curr_lens = int(parms[0]);
   }

   pp = fgets_trim(buff,maxfcc,fid,1);                                     //  fotoxx prior version
   if (pp && strNeq(pp,fversion))                                          //  fix segfault     v.10.1.1
      printf("prior fotoxx version: %s \n",pp);

   fclose(fid);

   for (ii = 0; ii < Nrecentfiles; ii++)                                   //  recent image file list = empty
      recentfiles[ii] = 0;

   snprintf(buff,maxfcc,"%s/recent_files",get_zuserdir());                 //  open recent files file
   fid = fopen(buff,"r");
   if (! fid) return 0;
   
   for (ii = 0; ii < Nrecentfiles; ii++)                                   //  read list of recent files
   {
      pp = fgets_trim(buff,maxfcc,fid,1);
      if (! pp) break;
      if (*pp == '/') recentfiles[ii] = strdupz(buff,0,"recentfile");
   }

   fclose(fid);

   return 1;
}


/**************************************************************************/

//  table for loading and saving adjustable parameters between sessions

typedef struct {
   char     name[20];
   char     type[12];
   int      count;
   void     *location;
}  param;

#define Nparms 22

param paramTab[Nparms] = {
//     name                 type        count   location
{     "image file",        "char",        1,    &image_file                },
{     "top directory",     "char",        1,    &topdirk                   },
{     "window width",      "int",         1,    &Dww                       },
{     "window height",     "int",         1,    &Dhh                       },
{     "thumbsize",         "int",         1,    &image_navi::thumbsize     },
{     "lens name",         "char",        4,    &lens4_name                },
{     "lens mm",           "double",      4,    &lens4_mm                  },
{     "lens bow",          "double",      4,    &lens4_bow                 },
{     "current lens",      "int",         1,    &curr_lens                 },
{     "grid spacing",      "int",         2,    &gridspace                 },
{     "grid counts",       "int",         2,    &gridcount                 },
{     "trim size",         "int",         2,    &trimsize                  },
{     "trim buttons",      "char",        6,    &trimbuttons               },
{     "trim ratios",       "char",        6,    &trimratios                },
{     "edit resize",       "int",         2,    &editresize                },
{     "batch resize",      "int",         2,    &batchresize               },
{     "e-mail resize",     "int",         2,    &emailsize                 },
{     "annotate font",     "char",        1,    &annotate_font             },
{     "annotate text",     "char",        1,    &annotate_text             },
{     "annotate color",    "char",        2,    &annotate_color            },
{     "annotate trans",    "int",         2,    &annotate_trans            },
{     "annotate angle",    "double",      1,    &annotate_angle            }  };


//  save parameters to file /home/<user>/.fotoxx/parameters

void save_params()                                                         //  new v.10.9
{
   FILE        *fid;
   char        buff[1050], text[1000];                                     //  limit for character data cc
   char        *name, *type;
   int         count;
   void        *location;
   char        **charloc;
   int         *intloc;
   double      *doubleloc;
   
   gtk_window_get_size(MWIN,&Dww,&Dhh);                                    //  prevent shrinkage

   snprintf(buff,199,"%s/parameters",get_zuserdir());                      //  open output file
   fid = fopen(buff,"w");
   if (! fid) return;
   
   fprintf(fid,"%s \n",fversion);                                          //  write fotoxx version
   
   for (int ii = 0; ii < Nparms; ii++)                                     //  write table of state data
   {
      name = paramTab[ii].name;
      type = paramTab[ii].type;
      count = paramTab[ii].count;
      location = paramTab[ii].location;
      charloc = (char **) location;
      intloc = (int *) location;
      doubleloc = (double *) location;

      fprintf(fid,"%-20s  %-8s  %02d  ",name,type,count);                  //  write parm name, type, count

      for (int kk = 0; kk < count; kk++)                                   //  write "value" "value" ...
      {
         if (strEqu(type,"char")) {
            if (! *charloc) printf("bad param data %s %d \n",name,kk);     //  bugfix v.10.11.1
            if (! *charloc) break;
            repl_1str(*charloc++,text,"\n","\\n");                         //  replace newlines with "\n"   v.10.11
            fprintf(fid,"  \"%s\"",text);
         }
         if (strEqu(type,"int"))
            fprintf(fid,"  \"%d\"",*intloc++);

         if (strEqu(type,"double"))
            fprintf(fid,"  \"%.2f\"",*doubleloc++);
      }

      fprintf(fid,"\n");                                                   //  write EOR
   }
   
   fprintf(fid,"\n");
   fclose(fid);                                                            //  close file

   snprintf(buff,199,"%s/recent_files",get_zuserdir());                    //  open file for recent files list
   fid = fopen(buff,"w");
   if (! fid) return;
   
   for (int ii = 0; ii < Nrecentfiles; ii++)                               //  save list of recent files
      if (recentfiles[ii])
         fprintf(fid,"%s \n",recentfiles[ii]);
   
   fclose(fid);

   return;
}


//  load parameters from file /home/<user>/.fotoxx/parameters

void load_params()                                                         //  new v.10.9
{
   FILE        *fid;
   int         ii, err, pcount, Idata;
   double      Rdata;
   char        buff[1050], text[1000], *pp;
   char        name[20], type[12], count[8], data[1000];
   void        *location;
   char        **charloc;
   int         *intloc;
   double      *doubleloc;
   
   for (ii = 0; ii < Nparms; ii++)                                         //  set string parameters to "undefined"
   {                                                                       //  v.10.11
      if (strNeq(paramTab[ii].type,"char")) continue;
      charloc = (char **) paramTab[ii].location;
      pcount = paramTab[ii].count;
      for (int jj = 0; jj < pcount; jj++)
         *charloc++ = strdupz("undefined",20);
   }

   snprintf(buff,499,"%s/parameters",get_zuserdir());                      //  open parameters file
   fid = fopen(buff,"r");
   if (! fid) {
      printf("reading old state file \n");
      load_old_fotoxx_state();                                             //  OBSOLETE from v.10.9
      return;
   }

   pp = fgets_trim(buff,499,fid,1);                                        //  get last fotoxx version
   if (pp && strNeq(pp,fversion))                                          //  bugfix   v.10.11
      printf("version change: %s %s \n",pp,fversion);
   
   while (true)                                                            //  read parameters
   {
      pp = fgets_trim(buff,999,fid,1);
      if (! pp) break;
      if (*pp == '#') continue;                                            //  comment          v.10.10
      if (strlen(pp) < 40) continue;                                       //  rubbish          v.10.10

      err = 0;

      strncpy0(name,pp,20);                                                //  parm name
      strTrim2(name);
      
      strncpy0(type,pp+22,8);                                              //  parm type
      strTrim2(type);
      
      strncpy0(count,pp+32,4);                                             //  parm count
      strTrim2(count);
      err = convSI(count,pcount);
      
      strncpy0(data,pp+38,1000);                                           //  parm value(s)
      strTrim2(data);
      
      for (ii = 0; ii < Nparms; ii++)                                      //  match file record to param table
      {
         if (strNeq(name,paramTab[ii].name)) continue;                     //  parm name
         if (strNeq(type,paramTab[ii].type)) continue;                     //  parm type
         if (paramTab[ii].count != pcount) continue;                       //  parm count

         location = paramTab[ii].location;
         charloc = (char **) location;
         intloc = (int *) location;
         doubleloc = (double *) location;
         
         for (int kk = 1; kk <= pcount; kk++)
         {
            pp = (char *) strField(data,' ',kk);
            if (! pp) break;

            if (strEqu(type,"char")) {
               repl_1str(pp,text,"\\n","\n");                              //  replace "\n" with real newlines  v.10.11
               *charloc++ = strdupz(text);
            }

            if (strEqu(type,"int")) {
               err = convSI(pp,Idata);
               if (err) continue;
               *intloc++ = Idata;
            }

            if (strEqu(type,"double")) {
               err = convSD(pp,Rdata);
               if (err) continue;
               *doubleloc++ = Rdata;
            }
         }
      }
   }

   fclose(fid);

   for (ii = 0; ii < Nrecentfiles; ii++)                                   //  recent image file list = empty
      recentfiles[ii] = 0;

   snprintf(buff,499,"%s/recent_files",get_zuserdir());                    //  open recent files file
   fid = fopen(buff,"r");
   if (! fid) return;
   
   for (ii = 0; ii < Nrecentfiles; ii++)                                   //  read list of recent files
   {
      pp = fgets_trim(buff,499,fid,1);
      if (! pp) break;
      if (*pp == '/') recentfiles[ii] = strdupz(buff,0,"recentfile");
   }

   fclose(fid);

   return;
}


/**************************************************************************/

//  ask user if modified image should be kept or discarded
//  returns: 0 = no mods or OK to discard,  1 = do not discard mods

int mod_keep()
{
   if (! Fmodified && ! Pundo && ! Ftagschanged) return 0;                 //  no mods                v.10.5
   if (Fsaved == Pundo) return 0;                                          //  last mods were saved
   if (zmessageYN(MWIN,ZTX("Discard modifications?")))                     //  OK to discard
      return 0;
   return 1;                                                               //  not OK
}


/**************************************************************************/

//  menu lock/unlock - some functions must not run concurrently
//  returns 1 if success, else 0

int menulock(int lock)
{
   if (! lock && ! menu_lock) zappcrash("menu lock error");
   if (lock && menu_lock) {
      printf("menu locked \n");                                            //  v.10.8
      return 0;
   }
   if (lock) menu_lock++;
   else menu_lock--;
   SBupdate++;                                                             //  v.10.5
   return 1;
}


/**************************************************************************/

//  Compute the mean brightness of all pixel neighborhoods,                //  new v.9.6
//  using a Guassian or a flat distribution for the weightings. 
//  If a select area is active, only inside pixels are calculated.
//  The flat method is 10-100x faster.

int         brhood_radius;
double      brhood_kernel[200][200];                                       //  up to radius = 99
float       *brhood_brightness = 0;                                        //  neighborhood brightness per pixel
char        brhood_method;                                                 //  g = gaussian, f = flat distribution


void brhood_calc(int radius, char method)
{
   void * brhood_wthread(void *arg);

   int      rad, rad2, dx, dy, cc, ii;
   double   kern;
   
   brhood_radius = radius;
   brhood_method = method;

   if (brhood_method == 'g')
   {                                                                       //  compute Gaussian kernel
      rad = brhood_radius;
      rad2 = rad * rad;

      for (dy = -rad; dy <= rad; dy++)
      for (dx = -rad; dx <= rad; dx++)
      {
         if (dx*dx + dy*dy <= rad2)                                        //  cells within radius
            kern = exp( - (dx*dx + dy*dy) / rad2);
         else kern = 0;                                                    //  outside radius
         brhood_kernel[dy+rad][dx+rad] = kern;
      }
   }
   
   if (brhood_brightness) zfree(brhood_brightness);                        //  allocate memory for pixel map
   cc = E1ww * E1hh * sizeof(float);
   brhood_brightness = (float *) zmalloc(cc,"brhood");
   memset(brhood_brightness,0,cc);

   if (sa_active) SB_goal = sa_Npixel;                                     //  set up progress tracking
   else SB_goal = E1ww * E1hh;
   SB_done = 0;

   for (ii = 0; ii < Nwt; ii++)                                            //  start worker threads
      start_wt(brhood_wthread,&wtnx[ii]);
   wait_wts();                                                             //  wait for completion

   SB_goal = SB_done = 0;
   return;
}
 
 
//  worker thread function

void * brhood_wthread(void *arg)
{
   int      index = *((int *) arg);
   int      rad = brhood_radius;
   int      ii, px, py, qx, qy, Fstart;
   double   kern, bsum, bsamp, bmean;
   uint16   *pixel;

   if (brhood_method == 'g')                                               //  use round gaussian distribution
   {
      for (py = index; py < E1hh; py += Nwt)
      for (px = 0; px < E1ww; px++)
      {
         if (sa_active) {                                                  //  select area active,
            ii = py * E1ww + px;                                           //    use only inside pixels
            if (! sa_pixisin[ii]) continue;
         }
         
         bsum = bsamp = 0;

         for (qy = py-rad; qy <= py+rad; qy++)                             //  computed weighted sum of brightness
         for (qx = px-rad; qx <= px+rad; qx++)                             //    for pixels in neighborhood
         {
            if (qy < 0 || qy > E1hh-1) continue;
            if (qx < 0 || qx > E1ww-1) continue;
            kern = brhood_kernel[qy+rad-py][qx+rad-px];
            pixel = PXMpix(E1pxm16,qx,qy);
            bsum += pixbright(pixel) * kern;                               //  sum brightness * weight
            bsamp += kern;                                                 //  sum weights
         }

         bmean = bsum / bsamp;                                             //  mean brightness
         ii = py * E1ww + px;
         brhood_brightness[ii] = bmean;                                    //  pixel value

         SB_done++;
         if (drandz() < 0.0001) zsleep(0.001);                             //  trigger sorry kernel scheduler
      }
   }

   if (brhood_method == 'f')                                               //  use square flat distribution
   {
      Fstart = 1;
      bsum = bsamp = 0;

      for (py = index; py < E1hh; py += Nwt)
      for (px = 0; px < E1ww; px++)
      {
         if (sa_active) {                                                  //  select area active
            ii = py * E1ww + px;
            if (! sa_pixisin[ii]) {
               Fstart = 1;
               continue;
            }
         }
         
         if (px == 0) Fstart = 1;
         
         if (Fstart) 
         {
            Fstart = 0;
            bsum = bsamp = 0;

            for (qy = py-rad; qy <= py+rad; qy++)                          //  add up all columns
            for (qx = px-rad; qx <= px+rad; qx++)
            {
               if (qy < 0 || qy > E1hh-1) continue;
               if (qx < 0 || qx > E1ww-1) continue;
               pixel = PXMpix(E1pxm16,qx,qy);
               bsum += pixbright(pixel);
               bsamp += 1;
            }
         }
         else
         {
            qx = px-rad-1;                                                 //  subtract first-1 column
            if (qx >= 0) {
               for (qy = py-rad; qy <= py+rad; qy++)
               {
                  if (qy < 0 || qy > E1hh-1) continue;
                  pixel = PXMpix(E1pxm16,qx,qy);
                  bsum -= pixbright(pixel);
                  bsamp -= 1;
               }
            }
            qx = px+rad;                                                   //  add last column
            if (qx < E1ww) {
               for (qy = py-rad; qy <= py+rad; qy++)
               {
                  if (qy < 0 || qy > E1hh-1) continue;
                  pixel = PXMpix(E1pxm16,qx,qy);
                  bsum += pixbright(pixel);
                  bsamp += 1;
               }
            }
         }

         bmean = bsum / bsamp;                                             //  mean brightness
         ii = py * E1ww + px;
         brhood_brightness[ii] = bmean;

         SB_done++;
         if (drandz() < 0.0001) zsleep(0.001);                             //  trigger sorry kernel scheduler
      }
   }
         
   exit_wt();
   return 0;                                                               //  not executed, avoid gcc warning
}


//  get the neighborhood brightness for given pixel

float get_brhood(int px, int py)
{
   int ii = py * E1ww + px;
   return brhood_brightness[ii];
}


/**************************************************************************/

//  free all resources associated with the current image file

void free_resources()
{
   char        *pp;
   int         ignore;

   mutex_lock(&pixmaps_lock);                                              //  lock pixmaps
   
   strcpy(command,"rm -f ");                                               //  delete all undo files
   strcat(command,undo_files);
   pp = strstr(command,"_undo_");                                          //  clone edit
   strcpy(pp + 6,"*");
   ignore = system(command);

   Fmodified = Pundo = Pumax = Fsaved = 0;                                 //  reset undo/redo stack
   Ntoplines = Nptoplines;                                                 //  no image overlay lines
   paint_toparc(2);                                                        //  no brush outline
   
   if (Fshutdown) {                                                        //  stop here if shutdown mode
      mutex_unlock(&pixmaps_lock);
      return;
   }
   
   if (image_file) {
      sa_delete();                                                         //  delete select area
      zdialog_free(zdsela);                                                //  kill dialogs if active
      zdialog_free(zdRGB);
      zdsela = zdRGB = null;
      Mcapture = 0;                                                        //  kill mouse function if active
      mouseCBfunc = 0;
      set_cursor(0);                                                       //  restore normal cursor
      zfree(image_file);                                                   //  free image file
      image_file = 0;
      SB_goal = SB_done = 0;                                               //  v.9.2
      *SB_text = 0;                                                        //  v.10.7
      Frotated = 0;                                                        //  bugfix     v.10.11
   }
   
   if (brhood_brightness) zfree(brhood_brightness);                        //  free brightness map   v.9.6
   brhood_brightness = 0;

   PXM_free(Fpxm8);
   PXM_free(Dpxm8);
   Fpxm8 = Dpxm8 = 0;
   PXM_free(Fpxm16);
   PXM_free(E1pxm16);
   PXM_free(E3pxm16);
   PXM_free(ERpxm16);                                                      //  v.10.3
   Fpxm16 = E1pxm16 = E3pxm16 = ERpxm16 = 0;

   Fww = E1ww = E3ww = Dww = 0;                                            //  make unusable (crash)

   mutex_unlock(&pixmaps_lock);
   return;
}


//  set cursor for main/drawing window and track installed cursor

void set_cursor(GdkCursor *cursor)                                         //  v.10.1
{
   zlock();                                                                //  lock added    v.10.7
   gdk_window_set_cursor(drWin->window,cursor);
   zunlock();
   currentcursor = cursor;
   zmainloop();                                                            //  beware re-entrance bugs    v.10.8
   return;
}


//  Get a virtual pixel at location (px,py) (real) in an PXM-16 pixmap.
//  Get the overlapping real pixels and build a composite.

int vpixel(PXM *pxm, double px, double py, uint16 *vpix)
{
   int            ww, hh, px0, py0;
   uint16         *ppix, *pix0, *pix1, *pix2, *pix3;
   double         f0, f1, f2, f3;
   double         red, green, blue;
   
   ww = pxm->ww;
   hh = pxm->hh;
   ppix = (uint16 *) pxm->bmp;

   px0 = int(px);                                                          //  pixel containing (px,py)
   py0 = int(py);

   if (px0 < 1 || py0 < 1) return 0;                                       //  void edge pixels
   if (px0 > ww-3 || py0 > hh-3) return 0;
   
   pix0 = ppix + (py0 * ww + px0) * 3;                                     //  4 pixels based at (px0,py0)
   pix1 = pix0 + ww * 3;
   pix2 = pix0 + 3;
   pix3 = pix0 + ww * 3 + 3;

   f0 = (px0+1 - px) * (py0+1 - py);                                       //  overlap of (px,py)
   f1 = (px0+1 - px) * (py - py0);                                         //   in each of the 4 pixels
   f2 = (px - px0) * (py0+1 - py);
   f3 = (px - px0) * (py - py0);
   
   red =   f0 * pix0[0] + f1 * pix1[0] + f2 * pix2[0] + f3 * pix3[0];      //  sum the weighted inputs
   green = f0 * pix0[1] + f1 * pix1[1] + f2 * pix2[1] + f3 * pix3[1];
   blue =  f0 * pix0[2] + f1 * pix1[2] + f2 * pix2[2] + f3 * pix3[2];
   
   vpix[0] = red;
   vpix[1] = green;
   vpix[2] = blue;
   
   if (blue < 1) {                                                         //  v.10.7 
      if (blue < 0.9) return 0;                                            //  mostly void
      vpix[2] = 1;                                                         //  avoid 0.999 to integer 0
   }

   return 1;
}


//  compare two doubles for significant difference
//  return:  0  difference not significant
//          +1  d1 > d2
//          -1  d1 < d2

int sigdiff(double d1, double d2, double signf)
{
   double diff = fabs(d1-d2);
   if (diff == 0.0) return 0;
   diff = diff / (fabs(d1) + fabs(d2));
   if (diff < signf) return 0;
   if (d1 > d2) return 1;
   else return -1;
}


/**************************************************************************
   file read and write utilities
   PXM pixmap <--> file on disk
***************************************************************************/

//  Load an image file into an PXM pixmap of 8 or 16 bits/color
//  Also sets the following global variables:
//    file_MB = disk file megabytes
//    file_bpc = disk file bits per color (1, 8, 16)
//    file_type = "jpeg" "png" "tiff" or "other"

PXM * f_load(cchar *filespec, int bpc)                                     //  use pixbuf or tiff library   v.9.8
{
   int            err, angle;
   cchar          *pext;
   PXM            *pxm1, *pxm2;
   struct stat    fstat;
   char           **orientation;
   cchar          *orientationkey[1] = { exif_orientation_key };

   if (bpc != 8 && bpc != 16)                                              //  requested bpc must be 8 or 16
      zappcrash("image_load bpc: %d",bpc);

   err = stat(filespec,&fstat);
   if (err) return 0;                                                      //  file not found
   if (! S_ISREG(fstat.st_mode)) return 0;                                 //  not a regular file
   file_MB = 1.0 * fstat.st_size / mega;                                   //  disk file megabytes
   
   pext = strrchr(filespec,'/');
   if (! pext) pext = filespec;

   if (pext > filespec+12 && strnEqu(pext-12,"/.thumbnails/",13)) {        //  refuse thumbnail files    v.10.0
      zmessageACK(MWIN,ZTX("cannot open thumbnail file"));
      return 0;
   }

   pext = strrchr(pext,'.');
   if (! pext) pext = "";
   
   if (strstr(".jpg .jpeg .JPG .JPEG",pext)) strcpy(file_type,"jpeg");
   else if (strstr(".tiff .TIFF",pext)) strcpy(file_type,"tiff");
   else if (strstr(".png .PNG",pext)) strcpy(file_type,"png");
   else strcpy(file_type,"other");
   
   if (strEqu(file_type,"tiff"))                                           //  use tiff lib to read tiff file
      pxm1 = TIFFread(filespec);
   else  pxm1 = PXBread(filespec);                                         //  use pixbuf lib for others
   if (! pxm1) return 0;

   if (pxm1->bpc != bpc) {
      pxm2 = PXM_convbpc(pxm1);                                            //  convert to requested bpc
      PXM_free(pxm1);                                                      //    8 <--> 16
      pxm1 = pxm2;
   }

   angle = Fuprighted = 0;
   orientation = exif_get(filespec,orientationkey,1);                      //  upright turned image   v.10.5
   if (*orientation) {
      if (strstr(*orientation,"90")) angle = 90;
      if (strstr(*orientation,"180")) angle = 180;
      if (strstr(*orientation,"270")) angle = 270;
      zfree(*orientation);
   }
   
   while (angle >= 360) angle -= 360;
   while (angle <= -360) angle += 360;

   if (angle) {                                                            //  v.10.5
      pxm2 = PXM_rotate(pxm1,angle);
      PXM_free(pxm1);
      pxm1 = pxm2;
      Fuprighted = 1;                                                      //  flag for status bar
      SBupdate++;
   }
   
   return pxm1;
}


/**************************************************************************/

//  save current image to specified disk file (same or new).
//  update tags index file and gallery list.
//  return 0 if OK, else +N

int f_save(cchar *outfile, cchar *format)                                  //  use pixbuf or tiff library   v.9.8
{
   PXM            *pxm16;
   cchar          *exifkey[2] = { exif_orientation_key, exif_log_key };
   cchar          *exifdata[2];
   char           **ppv, funcslist[1000];
   char           *pp, *tempfile;
   int            nkeys, err = 1, cc1, cc2;
   struct stat    fstat;
   
   set_cursor(busycursor);                                                 //  set function busy cursor

   tempfile = strdupz(get_zuserdir(),24,"tempfile");                       //  use temp output file
   strcat(tempfile,"/temp_");                                              //    ~/.fotoxx/temp_ppppp.ext
   strcat(tempfile,PIDstring);
   strcat(tempfile,".");
   strncat(tempfile,format,4);

   if (strEqu(format,"jpeg")) {                                            //  save as JPEG file
      if (Fpxm16) err = PXBwrite(Fpxm16,tempfile);
      else  err = PXBwrite(Fpxm8,tempfile);
   }

   if (strEqu(format,"png")) {                                             //  save as PNG file
      if (Fpxm16) err = PXBwrite(Fpxm16,tempfile);
      else  err = PXBwrite(Fpxm8,tempfile);
   }

   if (strEqu(format,"tiff-8")) {                                          //  save as tiff-8 file
      if (Fpxm16) {
         PXM_free(Fpxm8);                                                  //  bugfix     v.10.8
         Fpxm8 = PXM_convbpc(Fpxm16);
      }
      err = TIFFwrite(Fpxm8,tempfile);
   }
   
   if (strEqu(format,"tiff-16")) {                                         //  save as tiff-16 file
      if (Fpxm16) err = TIFFwrite(Fpxm16,tempfile);                        //  edit file exists, use it
      else {
         if (file_bpc == 16) pxm16 = TIFFread(image_file);                 //  use original 16 bpc file
         else  pxm16 = PXM_convbpc(Fpxm8);                                 //  or convert 8 to 16 bpc
         if (! pxm16) err = 1;
         else {
            err = TIFFwrite(pxm16,tempfile);
            PXM_free(pxm16);
         }
      }
   }

   if (err) {
      snprintf(command,ccc,"rm \"%s\"",tempfile);                          //  failure, clean up
      err = system(command);
      zfree(tempfile);
      set_cursor(0);
      return 1;
   }

   exifdata[0] = "";                                                       //  EXIF orientation, upright
   nkeys = 1;                                                              //  remove N/A width, height        v.10.8

   if (Pundo > 0)                                                          //  if edits were made,
   {                                                                       //    update relevant EXIF key      v.10.2
      *funcslist = 0;
      ppv = exif_get(image_file,&exifkey[1],1);                            //  get existing list of edits
      if (ppv[0]) {
         strncpy0(funcslist,ppv[0],998);
         zfree(ppv[0]);
      }
      cc1 = strlen(funcslist);
      if (cc1 && funcslist[cc1-1] != ' ') {
         strcpy(funcslist+cc1," ");
         cc1++;
      }

      for (int ii = 1; ii <= Pundo; ii++)                                  //  append new list of edits
      {                                                                    //  (omit index 0 = initial)
         pp = pvlist_get(editlog,ii);
         cc2 = strlen(pp);
         if (cc1 + cc2 > 998) break;
         strcpy(funcslist+cc1,pp);
         strcpy(funcslist+cc1+cc2," ");
         cc1 += cc2 + 1;
      }
      
      exifdata[1] = funcslist;                                             //  EXIF log key, fotoxx edits done
      nkeys = 2;
   }

   err = exif_copy(image_file,tempfile,exifkey,exifdata,nkeys);            //  copy all EXIF data to temp file
   if (err) zmessageACK(MWIN,"Unable to copy EXIF");                       //    with above revisions
   
   snprintf(command,ccc,"cp -f \"%s\" \"%s\" ",tempfile,outfile);          //  copy temp file to output file
   err = system(command);
   if (err) zmessageACK(MWIN,"Unable to save image: %s",wstrerror(err));

   snprintf(command,ccc,"rm \"%s\"",tempfile);                             //  delete temp file
   err = system(command);
   zfree(tempfile);
   
   save_filetags(outfile);                                                 //  save tag changes if any
   if (strNeq(outfile,image_file))                                         //  if save to new file,         v.10.5
      update_tags_index(outfile);                                          //    update tags index file

   Fsaved = Pundo;                                                         //  note which mods are saved

   stat(outfile,&fstat);                                                   //  update file size
   file_MB = 1.0 * fstat.st_size / mega;
   SBupdate++;                                                             //  update status bar

   add_recent_file(outfile);                                               //  first in recent files list

   set_cursor(0);                                                          //  restore normal cursor
   return 0;
}


/**************************************************************************/

//  Read from TIFF file using TIFF library. 
//  Use native TIFF file bits/pixel.

PXM * TIFFread(cchar *filespec)                                            //  overhauled  v.10.8.1
{
   TIFF        *tiff;
   PXM         *pxm;
   char        *tiffbuff;
   uint8       *tiff8, *pxm8;
   uint16      *tiff16, *pxm16;
   uint16      bpc, nch, fmt; 
   uint        ww, hh, rps, stb, nst;
   uint        tiffstat, row, col, strip, cc;
   
   tiff = TIFFOpen(filespec,"r");
   if (! tiff) {
      zmessageACK(MWIN,ZTX("TIFF open failure"));
      return 0;
   }

   TIFFGetField(tiff, TIFFTAG_IMAGEWIDTH, &ww);                            //  width
   TIFFGetField(tiff, TIFFTAG_IMAGELENGTH, &hh);                           //  height
   TIFFGetField(tiff, TIFFTAG_BITSPERSAMPLE, &bpc);                        //  bits per color, 1/8/16
   TIFFGetField(tiff, TIFFTAG_SAMPLESPERPIXEL, &nch);                      //  no. channels (colors), 1/3
   TIFFGetField(tiff, TIFFTAG_ROWSPERSTRIP, &rps);                         //  rows per strip
   TIFFGetField(tiff, TIFFTAG_SAMPLEFORMAT, &fmt);                         //  data format
   stb = TIFFStripSize(tiff);                                              //  strip size
   nst = TIFFNumberOfStrips(tiff);                                         //  number of strips

// printf("ww %d  hh %d  nch %d  bpc %d  rps %d  stb %d  nst %d  fmt %d \n",ww,hh,nch,bpc,rps,stb,nst,fmt);
   
   if (! (bpc <= 8 || bpc == 16)) {                                        //  check for supported bits/color
      zmessageACK(MWIN,ZTX("TIFF bits/color=%d not supported"),bpc);
      TIFFClose(tiff);
      return 0;
   }

   if (bpc <= 8)                                                           //  use universal TIFF reader
   {                                                                       //    if bits/color <= 8
      tiffbuff = zmalloc(ww*hh*4,"tiffbuff");

      tiffstat = TIFFReadRGBAImage(tiff, ww, hh, (uint *) tiffbuff, 0);
      TIFFClose(tiff);

      if (tiffstat != 1) {
         zmessageACK(MWIN,ZTX("TIFF read failure"));
         zfree(tiffbuff);
         return 0;
      }

      pxm = PXM_make(ww,hh,8);

      tiff8 = (uint8 *) tiffbuff;

      for (row = 0; row < hh; row++)                                       //  convert RGBA to RGB
      {
         pxm8 = (uint8 *) pxm->bmp;
         pxm8 += (hh-1 - row) * ww * 3;

         for (col = 0; col < ww; col++)
         {
            pxm8[0] = tiff8[0];
            pxm8[1] = tiff8[1];
            pxm8[2] = tiff8[2];
            pxm8 += 3;
            tiff8 += 4;
         }
      }

      zfree(tiffbuff);
      return pxm;
   }
                                                                           //  16 bits per sample

   stb += 1000000;                                                         //  reduce risk of crash   v.10.8.2
   tiffbuff = zmalloc(stb,"tiffbuff");                                     //  read encoded strips 

   pxm = PXM_make(ww,hh,16);
   
   for (strip = 0; strip < nst; strip++)
   {
      cc = TIFFReadEncodedStrip(tiff,strip,tiffbuff,-1);
      if (cc < 0) {
         zmessageACK(MWIN,ZTX("TIFF read failure"));
         TIFFClose(tiff);
         zfree(tiffbuff);
         PXM_free(pxm);
         return 0;
      }
      
      if (cc == 0) break;
      
      tiff16 = (uint16 *) tiffbuff;
      pxm16 = (uint16 *) pxm->bmp;
      row = strip * rps;
      pxm16 += row * ww * 3;

      while (cc)
      {
         pxm16[0] = tiff16[0];
         pxm16[1] = tiff16[1];
         pxm16[2] = tiff16[2];
         pxm16 += 3;
         tiff16 += nch;
         cc -= nch * 2;
      }
   }

   TIFFClose(tiff);
   zfree(tiffbuff);
   return pxm;
}


//  Write to TIFF file using TIFF library. 
//  File bpc is taken from PXM (8 or 16).
//  returns 0 if OK, +N if error.

int TIFFwrite(PXM *pxm, cchar *filespec)                                   //  new v.9.8
{
   TIFF        *tiff;
   uint8       *tiff8, *pxm8;
   uint16      *tiff16, *pxm16;
   int         tiffstat = 0;
   uint        ww, hh, row, col, rowcc;
   uint16      bpc, nch, pm = 2, pc = 1, comp = 5; 
   char        *tiffbuff;
   
   tiff = TIFFOpen(filespec,"w");
   if (! tiff) {
      zmessageACK(MWIN,ZTX("TIFF open failure"));
      return 1;
   }
   
   ww = pxm->ww;
   hh = pxm->hh;
   bpc = pxm->bpc;
   nch = 3;                                                                //  alpha channel removed

   TIFFSetField(tiff, TIFFTAG_IMAGEWIDTH, ww);
   TIFFSetField(tiff, TIFFTAG_BITSPERSAMPLE, bpc);
   TIFFSetField(tiff, TIFFTAG_SAMPLESPERPIXEL, nch);
   TIFFSetField(tiff, TIFFTAG_PHOTOMETRIC, pm);                            //  RGB
   TIFFSetField(tiff, TIFFTAG_PLANARCONFIG, pc);
   TIFFSetField(tiff, TIFFTAG_COMPRESSION, comp);                          //  LZW
   
   rowcc = TIFFScanlineSize(tiff);
   tiffbuff = (char*) zmalloc(rowcc,"tiffbuff");

   for (row = 0; row < hh; row++)
   {
      if (bpc == 8)
      {
         tiff8 = (uint8 *) tiffbuff;
         pxm8 = (uint8 *) pxm->bmp + row * ww * 3;

         for (col = 0; col < ww; col++)
         {
            tiff8[0] = pxm8[0];
            tiff8[1] = pxm8[1];
            tiff8[2] = pxm8[2];
            pxm8 += 3;
            tiff8 += nch;
         }
      }

      if (bpc == 16) 
      {
         tiff16 = (uint16 *) tiffbuff;
         pxm16 = (uint16 *) pxm->bmp + row * ww * 3;

         for (col = 0; col < ww; col++)
         {
            tiff16[0] = pxm16[0];
            tiff16[1] = pxm16[1];
            tiff16[2] = pxm16[2];
            pxm16 += 3;
            tiff16 += nch;
         }
      }

      tiffstat = TIFFWriteScanline(tiff,tiffbuff,row,0);
      if (tiffstat != 1) break;
   }

   TIFFClose(tiff);
   zfree(tiffbuff);
   
   if (tiffstat == 1) return 0;
   zmessageACK(MWIN,ZTX("TIFF write failure"));
   return 2;
}


//  intercept TIFF warning messages (many)                                 //  new v.9.8

void tiffwarninghandler(cchar *module, cchar *format, va_list ap)
{
   return;                                                                 //  stop flood of crap
   char  message[200];
   vsnprintf(message,199,format,ap);
   printf("TIFF warning: %s %s \n",module,message);
   return;
}


//  Read from jpeg/png file using pixbuf library. bpc = 8.

PXM * PXBread(cchar *filespec)                                             //  new v.9.8
{
   GError      *gerror = 0;
   PXB         *pxb;
   PXM         *pxm;
   int         ww, hh, px, py, nch, rowst;
   uint8       *bmp1, *bmp2, *pix1, *pix2;
   
   pxb = gdk_pixbuf_new_from_file(filespec,&gerror);
   if (! pxb) {
      printf("%s \n",gerror->message);                                     //  v.10.8
      zmessageACK(MWIN,ZTX("file type not supported"));
      return 0;
   }

   ww = gdk_pixbuf_get_width(pxb);
   hh = gdk_pixbuf_get_height(pxb);
   nch = gdk_pixbuf_get_n_channels(pxb);
   rowst = gdk_pixbuf_get_rowstride(pxb);
   bmp1 = gdk_pixbuf_get_pixels(pxb);
   
   pxm = PXM_make(ww,hh,8);
   bmp2 = (uint8 *) pxm->bmp;

   for (py = 0; py < hh; py++)
   {
      pix1 = bmp1 + rowst * py;
      pix2 = bmp2 + py * ww * 3;

      for (px = 0; px < ww; px++)
      {
         pix2[0] = pix1[0];
         pix2[1] = pix1[1];
         pix2[2] = pix1[2];
         pix1 += nch;
         pix2 += 3;
      }
   }
   
   file_bpc = 8;                                                           //  set file bits per color
   g_object_unref(pxb);
   return pxm;
}


//  Write to jpeg/png file using pixbuf library. bpc = 8.
//  returns 0 if OK, +N if error.

int PXBwrite(PXM *pxm, cchar *filespec)                                    //  new v.9.8
{
   int         ww, hh, bpc, px, py, rowst;
   uint8       *bmp8, *pix8, *bmp2, *pix2;
   uint16      *bmp16, *pix16;
   PXB         *pxb;
   cchar       *pext;
   GError      *gerror = 0;
   int         pxbstat;

   ww = pxm->ww;
   hh = pxm->hh;
   bpc = pxm->bpc;
   
   pxb = gdk_pixbuf_new(COLORSPACE,0,8,ww,hh);
   if (! pxb) zappcrash("pixbuf allocation failure");

   bmp8 = (uint8 *) pxm->bmp;
   bmp16 = (uint16 *) pxm->bmp;
   bmp2 = gdk_pixbuf_get_pixels(pxb);
   rowst = gdk_pixbuf_get_rowstride(pxb);

   if (bpc == 8)
   {
      for (py = 0; py < hh; py++)
      {
         pix8 = bmp8 + py * ww * 3;
         pix2 = bmp2 + rowst * py;

         for (px = 0; px < ww; px++)
         {
            pix2[0] = pix8[0];
            pix2[1] = pix8[1];
            pix2[2] = pix8[2];
            pix2 += 3;
            pix8 += 3;
         }
      }
   }
   
   if (bpc == 16)
   {
      for (py = 0; py < hh; py++)
      {
         pix16 = bmp16 + py * ww * 3;
         pix2 = bmp2 + rowst * py;

         for (px = 0; px < ww; px++)
         {
            pix2[0] = pix16[0] >> 8;
            pix2[1] = pix16[1] >> 8;
            pix2[2] = pix16[2] >> 8;
            pix2 += 3;
            pix16 += 3;
         }
      }
   }
   
   pext = strrchr(filespec,'/');
   if (! pext) pext = filespec;
   pext = strrchr(pext,'.');
   if (! pext) pext = "";
   
   if (strstr(".png .PNG",pext))
      pxbstat = gdk_pixbuf_save(pxb,filespec,"png",&gerror,null);
   else pxbstat = gdk_pixbuf_save(pxb,filespec,"jpeg",&gerror,"quality",jpeg_quality,null);
   g_object_unref(pxb);
   if (pxbstat) return 0;

   printf("%s \n",gerror->message);                                        //  v.10.8
   zmessageACK(MWIN,ZTX("pixbuf write failure"));
   return 1;
}


/**************************************************************************
      PXM pixmap conversion and rescale functions
***************************************************************************/

//  initialize PXM pixmap - allocate memory

PXM * PXM_make(int ww, int hh, int bpc)
{
   if (ww < 1 || hh < 1 || (bpc != 8 && bpc != 16))
      zappcrash("PXM_make() %d %d %d",ww,hh,bpc);
   
   PXM *pxm = (PXM *) zmalloc(sizeof(PXM),"PXM_make");
   pxm->ww = ww;
   pxm->hh = hh;
   pxm->bpc = bpc;
   if (bpc == 8) pxm->bmp = zmalloc(ww*hh*3,"pxm.bmp");
   if (bpc == 16) pxm->bmp = zmalloc(ww*hh*6,"pxm.bmp");
   strcpy(pxm->wmi,"rgbrgb");
   return pxm;
}


//  free PXM pixmap

void PXM_free(PXM *pxm)
{
   if (! pxm) return;
   if (! strEqu(pxm->wmi,"rgbrgb")) 
      zappcrash("PXM_free(), bad PXM");
   strcpy(pxm->wmi,"xxxxxx");
   zfree(pxm->bmp);
   zfree(pxm);
   return;
}


//  create a copy of an PXM pixmap

PXM * PXM_copy(PXM *pxm1)
{
   int      cc;
   PXM      *pxm2;

   pxm2 = PXM_make(pxm1->ww, pxm1->hh, pxm1->bpc);
   cc = pxm1->ww * pxm1->hh * (pxm1->bpc / 8 * 3);
   memcpy(pxm2->bmp,pxm1->bmp,cc);
   return pxm2;
}


//  create a copy of an PXM area

PXM * PXM_copy_area(PXM *pxm1, int orgx, int orgy, int ww2, int hh2)
{
   uint8          *bmp1, *pix1, *bmp2, *pix2;
   uint16         *bmp3, *pix3, *bmp4, *pix4;
   PXM            *pxm2 = 0;
   int            ww1, hh1, bpc, px1, py1, px2, py2;

   ww1 = pxm1->ww;
   hh1 = pxm1->hh;
   bpc = pxm1->bpc;

   if (bpc == 8)
   {
      pxm2 = PXM_make(ww2,hh2,8);
      bmp1 = (uint8 *) pxm1->bmp;
      bmp2 = (uint8 *) pxm2->bmp;
     
      for (py1 = orgy, py2 = 0; py2 < hh2; py1++, py2++) 
      {
         for (px1 = orgx, px2 = 0; px2 < ww2; px1++, px2++)
         {
            pix1 = bmp1 + (py1 * ww1 + px1) * 3;
            pix2 = bmp2 + (py2 * ww2 + px2) * 3;

            pix2[0] = pix1[0];
            pix2[1] = pix1[1];
            pix2[2] = pix1[2];
            pix1 += 3;
            pix2 += 3;
         }
      }
   }

   if (bpc == 16)
   {
      pxm2 = PXM_make(ww2,hh2,16);
      bmp3 = (uint16 *) pxm1->bmp;
      bmp4 = (uint16 *) pxm2->bmp;
     
      for (py1 = orgy, py2 = 0; py2 < hh2; py1++, py2++) 
      {
         for (px1 = orgx, px2 = 0; px2 < ww2; px1++, px2++)
         {
            pix3 = bmp3 + (py1 * ww1 + px1) * 3;
            pix4 = bmp4 + (py2 * ww2 + px2) * 3;

            pix4[0] = pix3[0];
            pix4[1] = pix3[1];
            pix4[2] = pix3[2];
            pix3 += 3;
            pix4 += 3;
         }
      }
   }
   
   return pxm2;
}


//   convert PXM pixmap from 8/16 to 16/8 bits per color

PXM * PXM_convbpc(PXM *pxm1)
{
   uint8          *bmp8, *pix8;
   uint16         *bmp16, *pix16;
   PXM            *pxm2 = 0;
   int            ww, hh, bpc, px, py;

   ww = pxm1->ww;
   hh = pxm1->hh;
   bpc = pxm1->bpc;

   if (bpc == 8)                                                           //  8 > 16
   {
      pxm2 = PXM_make(ww,hh,16);
      bmp8 = (uint8 *) pxm1->bmp;
      bmp16 = (uint16 *) pxm2->bmp;
     
      for (py = 0; py < hh; py++) 
      {
         pix8 = bmp8 + py * ww * 3;
         pix16 = bmp16 + py * ww * 3;

         for (px = 0; px < ww; px++)
         {
            pix16[0] = pix8[0] << 8;
            pix16[1] = pix8[1] << 8;
            pix16[2] = pix8[2] << 8;
            pix8 += 3;
            pix16 += 3;
         }
      }
   }

   if (bpc == 16)                                                          //  16 > 8
   {
      pxm2 = PXM_make(ww,hh,8);
      bmp8 = (uint8 *) pxm2->bmp;
      bmp16 = (uint16 *) pxm1->bmp;
     
      for (py = 0; py < hh; py++) 
      {
         pix8 = bmp8 + py * ww * 3;
         pix16 = bmp16 + py * ww * 3;

         for (px = 0; px < ww; px++)
         {
            pix8[0] = pix16[0] >> 8;
            pix8[1] = pix16[1] >> 8;
            pix8[2] = pix16[2] >> 8;
            pix8 += 3;
            pix16 += 3;
         }
      }
   }

   return pxm2;
}


//  rescale PXM pixmap to size ww2 x hh2

PXM * PXM_rescale(PXM *pxm1, int ww2, int hh2)
{
   void bmp8_rescale(uint8*, uint8*, int, int, int, int);
   void bmp16_rescale(uint16*, uint16*, int, int, int, int);

   PXM      *pxm2;
   int      ww1, hh1, bpc;
   uint8    *bmp1, *bmp2;
   uint16   *bmp3, *bmp4;

   ww1 = pxm1->ww;
   hh1 = pxm1->hh;
   bpc = pxm1->bpc;
   
   pxm2 = PXM_make(ww2,hh2,bpc);
   
   if (bpc == 8) {
      bmp1 = (uint8 *) pxm1->bmp;
      bmp2 = (uint8 *) pxm2->bmp;
      bmp8_rescale(bmp1,bmp2,ww1,hh1,ww2,hh2);
   }

   if (bpc == 16) {
      bmp3 = (uint16 *) pxm1->bmp;
      bmp4 = (uint16 *) pxm2->bmp;
      bmp16_rescale(bmp3,bmp4,ww1,hh1,ww2,hh2);
   }
   
   return pxm2;
}


//  Copy and rescale a modified area within a PXM-16 image into the
//  corresponding area of a PXM-8 image previously rescaled from the 
//  PXM-16 image. Keep the same mapping of input to output pixels, so 
//  that the modified area fits seamlessly into the PXM-8 image. Used 
//  when a section of an image is edited and the window image is updated.
//   
//  pxm1           PXM-16 image with changed area to rescale and copy
//  pxm2           existing rescaled PXM-8 copy of pxm1
//  org1x, org1y   pxm1 origin of area to copy (top left corner)
//  ww1a, hh1a     width and height of area to copy                        //  new v.10.11

void PXM_update(PXM *pxm1, PXM *pxm2, int org1x, int org1y, int ww1a, int hh1a)
{
   uint16      *bmp1, *pix1;
   uint8       *bmp2, *pix2;
   int         ww1, hh1, ww2, hh2;
   int         px1, py1, px2, py2;
   int         pxl, pyl, pxm, pym, ii;
   int         *px1L, *py1L;
   int         maxmapx, maxmapy;
   int         org2x, org2y, ww2a, hh2a;
   float       scalex, scaley;
   float       px1a, py1a, px1b, py1b;
   float       fx, fy, ftot;
   float       red, green, blue;
   float       *pxmap, *pymap;

   bmp1 = (uint16 *) pxm1->bmp;   
   bmp2 = (uint8 *) pxm2->bmp;   
   ww1 = pxm1->ww;
   hh1 = pxm1->hh;
   ww2 = pxm2->ww;
   hh2 = pxm2->hh;

   scalex = 1.0 * ww1 / ww2;                                               //  compute x and y scales
   scaley = 1.0 * hh1 / hh2;
   
   if (scalex <= 1) maxmapx = 2;                                           //  compute max input pixels
   else maxmapx = int(scalex + 2);                                         //    mapping into output pixels
   maxmapx += 1;                                                           //      for both dimensions
   if (scaley <= 1) maxmapy = 2;                                           //  (pixels may not be square)
   else maxmapy = int(scaley + 2);
   maxmapy += 1;
   
   pymap = (float *) zmalloc(hh2 * maxmapy * sizeof(float));               //  maps overlap of < maxmap input
   pxmap = (float *) zmalloc(ww2 * maxmapx * sizeof(float));               //    pixels per output pixel

   py1L = (int *) zmalloc(hh2 * sizeof(int));                              //  maps first (lowest) input pixel
   px1L = (int *) zmalloc(ww2 * sizeof(int));                              //    per output pixel

   for (py2 = 0; py2 < hh2; py2++)                                         //  loop output y-pixels
   {
      py1a = py2 * scaley;                                                 //  corresponding input y-pixels
      py1b = py1a + scaley;
      if (py1b >= hh1) py1b = hh1 - 0.001;                                 //  fix precision limitation
      pyl = int(py1a);
      py1L[py2] = pyl;                                                     //  1st overlapping input pixel

      for (py1 = pyl, pym = 0; py1 < py1b; py1++, pym++)                   //  loop overlapping input pixels
      {
         if (py1 < py1a) {                                                 //  compute amount of overlap
            if (py1+1 < py1b) fy = py1+1 - py1a;                           //    0.0 to 1.0 
            else fy = scaley;
         }
         else if (py1+1 > py1b) fy = py1b - py1;
         else fy = 1;

         ii = py2 * maxmapy + pym;                                         //  save it
         pymap[ii] = 0.9999 * fy / scaley;
      }
      ii = py2 * maxmapy + pym;                                            //  set an end marker after
      pymap[ii] = -1;                                                      //    last overlapping pixel
   }
   
   for (px2 = 0; px2 < ww2; px2++)                                         //  do same for x-pixels
   {
      px1a = px2 * scalex;
      px1b = px1a + scalex;
      if (px1b >= ww1) px1b = ww1 - 0.001;
      pxl = int(px1a);
      px1L[px2] = pxl;

      for (px1 = pxl, pxm = 0; px1 < px1b; px1++, pxm++)
      {
         if (px1 < px1a) {
            if (px1+1 < px1b) fx = px1+1 - px1a;
            else fx = scalex;
         }
         else if (px1+1 > px1b) fx = px1b - px1;
         else fx = 1;

         ii = px2 * maxmapx + pxm;
         pxmap[ii] = 0.9999 * fx / scalex;
      }
      ii = px2 * maxmapx + pxm;
      pxmap[ii] = -1;
   }
   
   org2x = org1x / scalex - 1;                                             //  compute rectangle in output image
   if (org2x < 0) org2x = 0;
   ww2a = ww1a / scalex + 2;
   if (org2x + ww2a > ww2) ww2a = ww2 - org2x;

   org2y = org1y / scaley - 1;
   if (org2y < 0) org2y = 0;
   hh2a = hh1a / scaley + 2;
   if (org2y + hh2a > hh2) hh2a = hh2 - org2y;

   for (py2 = org2y; py2 < org2y + hh2a; py2++)                            //  loop output y-pixels
   {
      pyl = py1L[py2];                                                     //  corresp. 1st input y-pixel

      for (px2 = org2x; px2 < org2x + ww2a; px2++)                         //  loop output x-pixels
      {
         pxl = px1L[px2];                                                  //  corresp. 1st input x-pixel

         red = green = blue = 0;                                           //  initz. output pixel

         for (py1 = pyl, pym = 0; ; py1++, pym++)                          //  loop overlapping input y-pixels
         {
            ii = py2 * maxmapy + pym;                                      //  get y-overlap
            fy = pymap[ii];
            if (fy < 0) break;                                             //  no more pixels

            for (px1 = pxl, pxm = 0; ; px1++, pxm++)                       //  loop overlapping input x-pixels
            {
               ii = px2 * maxmapx + pxm;                                   //  get x-overlap
               fx = pxmap[ii];
               if (fx < 0) break;                                          //  no more pixels

               ftot = fx * fy;                                             //  area overlap = x * y overlap
               pix1 = bmp1 + (py1 * ww1 + px1) * 3;
               red += pix1[0] * ftot;                                      //  add input pixel * overlap
               green += pix1[1] * ftot;
               blue += pix1[2] * ftot;
            }

            pix2 = bmp2 + (py2 * ww2 + px2) * 3;                           //  save output pixel
            pix2[0] = int(red) >> 8;
            pix2[1] = int(green) >> 8;
            pix2[2] = int(blue) >> 8;
         }
      }
   }

   zfree(px1L);
   zfree(py1L);
   zfree(pxmap);
   zfree(pymap);
   return;
}


//  rotate PXM pixmap through given angle in degrees (+ = clockwise)

PXM * PXM_rotate(PXM *pxm1, double angle)
{
   PXM * PXM_rotate8(PXM *, double);
   PXM * PXM_rotate16(PXM *, double);

   PXM      *pxm2 = 0;
   int      bpc;
   
   bpc = pxm1->bpc;
   if (bpc == 8) pxm2 = PXM_rotate8(pxm1,angle);
   if (bpc == 16) pxm2 = PXM_rotate16(pxm1,angle);
   return pxm2;
}


/**************************************************************************

   Rescale 8 bpc image (3 x 8 bits per color) to new width and height.
   The scale ratios may be different for width and height.

   Method: 
   The input and output images are overlayed, stretching or shrinking the
   output pixels as needed. The contribution of each input pixel overlapping
   an output pixel is proportional to the area of the output pixel covered by
   the input pixel. The contributions of all overlaping input pixels are added.
   The work is spread among Nwt threads to reduce the elapsed time on modern 
   computers having multiple SMP processors.

   Example: if the output image is 40% of the input image, then:
     outpix[0,0] = 0.16 * inpix[0,0] + 0.16 * inpix[1,0] + 0.08 * inpix[2,0]
                 + 0.16 * inpix[0,1] + 0.16 * inpix[1,1] + 0.08 * inpix[2,1]
                 + 0.08 * inpix[0,2] + 0.08 * inpix[1,2] + 0.04 * inpix[2,2]

*********/

namespace bmp8rescale {                                                    //  data for threads
   uint8    *pixmap1;
   uint8    *pixmap2;
   int      ww1;
   int      hh1;
   int      ww2;
   int      hh2;
   int      *px1L;
   int      *py1L;
   float    *pxmap;
   float    *pymap;
   int      maxmapx;
   int      maxmapy;
   int      busy[max_threads];
}


void bmp8_rescale(uint8 *pixmap1x, uint8 *pixmap2x, int ww1x, int hh1x, int ww2x, int hh2x)
{
   using namespace bmp8rescale;

   void * bmp8_rescale_thread(void *arg);

   int         px1, py1, px2, py2;
   int         pxl, pyl, pxm, pym, ii;
   float       scalex, scaley;
   float       px1a, py1a, px1b, py1b;
   float       fx, fy;
   
   pixmap1 = pixmap1x;
   pixmap2 = pixmap2x;
   ww1 = ww1x;
   hh1 = hh1x;
   ww2 = ww2x;
   hh2 = hh2x;

   memset(pixmap2, 0, ww2 * hh2 * 3 * sizeof(uint8));                      //  clear output pixmap

   scalex = 1.0 * ww1 / ww2;                                               //  compute x and y scales
   scaley = 1.0 * hh1 / hh2;
   
   if (scalex <= 1) maxmapx = 2;                                           //  compute max input pixels
   else maxmapx = int(scalex + 2);                                         //    mapping into output pixels
   maxmapx += 1;                                                           //      for both dimensions
   if (scaley <= 1) maxmapy = 2;                                           //  (pixels may not be square)
   else maxmapy = int(scaley + 2);
   maxmapy += 1;
   
   pymap = (float *) zmalloc(hh2 * maxmapy * sizeof(float));               //  maps overlap of < maxmap input
   pxmap = (float *) zmalloc(ww2 * maxmapx * sizeof(float));               //    pixels per output pixel

   py1L = (int *) zmalloc(hh2 * sizeof(int));                              //  maps first (lowest) input pixel
   px1L = (int *) zmalloc(ww2 * sizeof(int));                              //    per output pixel

   for (py2 = 0; py2 < hh2; py2++)                                         //  loop output y-pixels
   {
      py1a = py2 * scaley;                                                 //  corresponding input y-pixels
      py1b = py1a + scaley;
      if (py1b >= hh1) py1b = hh1 - 0.001;                                 //  fix precision limitation
      pyl = int(py1a);
      py1L[py2] = pyl;                                                     //  1st overlapping input pixel

      for (py1 = pyl, pym = 0; py1 < py1b; py1++, pym++)                   //  loop overlapping input pixels
      {
         if (py1 < py1a) {                                                 //  compute amount of overlap
            if (py1+1 < py1b) fy = py1+1 - py1a;                           //    0.0 to 1.0 
            else fy = scaley;
         }
         else if (py1+1 > py1b) fy = py1b - py1;
         else fy = 1;

         ii = py2 * maxmapy + pym;                                         //  save it
         pymap[ii] = 0.9999 * fy / scaley;
      }
      ii = py2 * maxmapy + pym;                                            //  set an end marker after
      pymap[ii] = -1;                                                      //    last overlapping pixel
   }
   
   for (px2 = 0; px2 < ww2; px2++)                                         //  do same for x-pixels
   {
      px1a = px2 * scalex;
      px1b = px1a + scalex;
      if (px1b >= ww1) px1b = ww1 - 0.001;
      pxl = int(px1a);
      px1L[px2] = pxl;

      for (px1 = pxl, pxm = 0; px1 < px1b; px1++, pxm++)
      {
         if (px1 < px1a) {
            if (px1+1 < px1b) fx = px1+1 - px1a;
            else fx = scalex;
         }
         else if (px1+1 > px1b) fx = px1b - px1;
         else fx = 1;

         ii = px2 * maxmapx + pxm;
         pxmap[ii] = 0.9999 * fx / scalex;
      }
      ii = px2 * maxmapx + pxm;
      pxmap[ii] = -1;
   }

   for (ii = 0; ii < Nwt; ii++) {                                          //  start working threads
      busy[ii] = 1;
      start_detached_thread(bmp8_rescale_thread,&wtnx[ii]);
   }
   
   for (ii = 0; ii < Nwt; ii++)                                            //  wait for all done
      while (busy[ii]) zsleep(0.004);

   zfree(px1L);
   zfree(py1L);
   zfree(pxmap);
   zfree(pymap);
   return;
}


void * bmp8_rescale_thread(void *arg)                                      //  worker thread function
{
   using namespace bmp8rescale;

   int         index = *((int *) arg);
   int         px1, py1, px2, py2;
   int         pxl, pyl, pxm, pym, ii;
   uint8       *pixel1, *pixel2;
   float       fx, fy, ftot;
   float       red, green, blue;

   for (py2 = index; py2 < hh2; py2 += Nwt)                                //  loop output y-pixels
   {
      pyl = py1L[py2];                                                     //  corresp. 1st input y-pixel

      for (px2 = 0; px2 < ww2; px2++)                                      //  loop output x-pixels
      {
         pxl = px1L[px2];                                                  //  corresp. 1st input x-pixel

         red = green = blue = 0;                                           //  initz. output pixel

         for (py1 = pyl, pym = 0; ; py1++, pym++)                          //  loop overlapping input y-pixels
         {
            ii = py2 * maxmapy + pym;                                      //  get y-overlap
            fy = pymap[ii];
            if (fy < 0) break;                                             //  no more pixels

            for (px1 = pxl, pxm = 0; ; px1++, pxm++)                       //  loop overlapping input x-pixels
            {
               ii = px2 * maxmapx + pxm;                                   //  get x-overlap
               fx = pxmap[ii];
               if (fx < 0) break;                                          //  no more pixels

               ftot = fx * fy;                                             //  area overlap = x * y overlap
               pixel1 = pixmap1 + (py1 * ww1 + px1) * 3;
               red += pixel1[0] * ftot;                                    //  add input pixel * overlap
               green += pixel1[1] * ftot;
               blue += pixel1[2] * ftot;
            }

            pixel2 = pixmap2 + (py2 * ww2 + px2) * 3;                      //  save output pixel
            pixel2[0] = int(red);
            pixel2[1] = int(green);
            pixel2[2] = int(blue);
         }
      }
   }

   busy[index] = 0;
   return 0;
}


/**************************************************************************

   Rescale 16 bpc image (3 x 16 bits per color) to new width and height.
   Identical to bmp8_rescale except for the following: 
      uint8 >> uint16
      xxx8 >> xxx16

*******/

namespace bmp16rescale {                                                   //  data for threads
   uint16   *pixmap1;
   uint16   *pixmap2;
   int      ww1;
   int      hh1;
   int      ww2;
   int      hh2;
   int      *px1L;
   int      *py1L;
   float    *pxmap;
   float    *pymap;
   int      maxmapx;
   int      maxmapy;
   int      busy[max_threads];
}


void bmp16_rescale(uint16 *pixmap1x, uint16 *pixmap2x, int ww1x, int hh1x, int ww2x, int hh2x)
{
   using namespace bmp16rescale;

   void * bmp16_rescale_thread(void *arg);

   int         px1, py1, px2, py2;
   int         pxl, pyl, pxm, pym, ii;
   float       scalex, scaley;
   float       px1a, py1a, px1b, py1b;
   float       fx, fy;
   
   pixmap1 = pixmap1x;
   pixmap2 = pixmap2x;
   ww1 = ww1x;
   hh1 = hh1x;
   ww2 = ww2x;
   hh2 = hh2x;

   memset(pixmap2, 0, ww2 * hh2 * 3 * sizeof(uint16));                     //  clear output pixmap

   scalex = 1.0 * ww1 / ww2;                                               //  compute x and y scales
   scaley = 1.0 * hh1 / hh2;
   
   if (scalex <= 1) maxmapx = 2;                                           //  compute max input pixels
   else maxmapx = int(scalex + 2);                                         //    mapping into output pixels
   maxmapx += 1;                                                           //      for both dimensions
   if (scaley <= 1) maxmapy = 2;                                           //  (pixels may not be square)
   else maxmapy = int(scaley + 2);
   maxmapy += 1;
   
   pymap = (float *) zmalloc(hh2 * maxmapy * sizeof(float));               //  maps overlap of < maxmap input
   pxmap = (float *) zmalloc(ww2 * maxmapx * sizeof(float));               //    pixels per output pixel

   py1L = (int *) zmalloc(hh2 * sizeof(int));                              //  maps first (lowest) input pixel
   px1L = (int *) zmalloc(ww2 * sizeof(int));                              //    per output pixel

   for (py2 = 0; py2 < hh2; py2++)                                         //  loop output y-pixels
   {
      py1a = py2 * scaley;                                                 //  corresponding input y-pixels
      py1b = py1a + scaley;
      if (py1b >= hh1) py1b = hh1 - 0.001;                                 //  fix precision limitation
      pyl = int(py1a);
      py1L[py2] = pyl;                                                     //  1st overlapping input pixel

      for (py1 = pyl, pym = 0; py1 < py1b; py1++, pym++)                   //  loop overlapping input pixels
      {
         if (py1 < py1a) {                                                 //  compute amount of overlap
            if (py1+1 < py1b) fy = py1+1 - py1a;                           //    0.0 to 1.0 
            else fy = scaley;
         }
         else if (py1+1 > py1b) fy = py1b - py1;
         else fy = 1;

         ii = py2 * maxmapy + pym;                                         //  save it
         pymap[ii] = 0.9999 * fy / scaley;
      }
      ii = py2 * maxmapy + pym;                                            //  set an end marker after
      pymap[ii] = -1;                                                      //    last overlapping pixel
   }
   
   for (px2 = 0; px2 < ww2; px2++)                                         //  do same for x-pixels
   {
      px1a = px2 * scalex;
      px1b = px1a + scalex;
      if (px1b >= ww1) px1b = ww1 - 0.001;
      pxl = int(px1a);
      px1L[px2] = pxl;

      for (px1 = pxl, pxm = 0; px1 < px1b; px1++, pxm++)
      {
         if (px1 < px1a) {
            if (px1+1 < px1b) fx = px1+1 - px1a;
            else fx = scalex;
         }
         else if (px1+1 > px1b) fx = px1b - px1;
         else fx = 1;

         ii = px2 * maxmapx + pxm;
         pxmap[ii] = 0.9999 * fx / scalex;
      }
      ii = px2 * maxmapx + pxm;
      pxmap[ii] = -1;
   }

   for (ii = 0; ii < Nwt; ii++) {                                          //  start working threads
      busy[ii] = 1;
      start_detached_thread(bmp16_rescale_thread,&wtnx[ii]);
   }
   
   for (ii = 0; ii < Nwt; ii++)                                            //  wait for all done
      while (busy[ii]) zsleep(0.004);

   zfree(px1L);
   zfree(py1L);
   zfree(pxmap);
   zfree(pymap);
   return;
}


void * bmp16_rescale_thread(void *arg)                                     //  worker thread function
{
   using namespace bmp16rescale;

   int         index = *((int *) arg);
   int         px1, py1, px2, py2;
   int         pxl, pyl, pxm, pym, ii;
   uint16      *pixel1, *pixel2;
   float       fx, fy, ftot;
   float       red, green, blue;

   for (py2 = index; py2 < hh2; py2 += Nwt)                                //  loop output y-pixels
   {
      pyl = py1L[py2];                                                     //  corresp. 1st input y-pixel

      for (px2 = 0; px2 < ww2; px2++)                                      //  loop output x-pixels
      {
         pxl = px1L[px2];                                                  //  corresp. 1st input x-pixel

         red = green = blue = 0;                                           //  initz. output pixel

         for (py1 = pyl, pym = 0; ; py1++, pym++)                          //  loop overlapping input y-pixels
         {
            ii = py2 * maxmapy + pym;                                      //  get y-overlap
            fy = pymap[ii];
            if (fy < 0) break;                                             //  no more pixels

            for (px1 = pxl, pxm = 0; ; px1++, pxm++)                       //  loop overlapping input x-pixels
            {
               ii = px2 * maxmapx + pxm;                                   //  get x-overlap
               fx = pxmap[ii];
               if (fx < 0) break;                                          //  no more pixels

               ftot = fx * fy;                                             //  area overlap = x * y overlap
               pixel1 = pixmap1 + (py1 * ww1 + px1) * 3;
               red += pixel1[0] * ftot;                                    //  add input pixel * overlap
               green += pixel1[1] * ftot;
               blue += pixel1[2] * ftot;
            }

            pixel2 = pixmap2 + (py2 * ww2 + px2) * 3;                      //  save output pixel
            pixel2[0] = int(red);
            pixel2[1] = int(green);
            pixel2[2] = int(blue);
         }
      }
   }

   busy[index] = 0;
   return 0;
}


/**************************************************************************

      PXM *pxm2 = PXM_rotate8(PXM *pxm1, double angle)

      Rotate PXM-8 pixmap through an arbitrary angle (degrees).

      The returned image has the same size as the original, but the
      pixmap size is increased to accomodate the rotated image.
      (e.g. a 100x100 image rotated 45 deg. needs a 142x142 pixmap).
      The parameters ww and hh are the dimensions of the input
      pixmap, and are updated to the dimensions of the output pixmap.

      The space added around the rotated image is black (RGB 0,0,0).
      Angle is in degrees. Positive direction is clockwise.
      Speed is about 3 million pixels/sec/thread for a 2.4 GHz CPU.
      Loss of resolution is less than 1 pixel.
      
      Work is divided among Nwt threads to gain speed.
      
      v.9.3: affine transform instead of trig functions, for speed

***************************************************************************/

namespace rotpxm8 {
   int      busy = 0;
   uint8    *pixmap1;
   uint8    *pixmap2;
   int      ww1;
   int      hh1;
   int      ww2;
   int      hh2;
   double   angle;
}


PXM * PXM_rotate8(PXM *pxm1, double anglex)
{
   using namespace rotpxm8;

   void     *PXM_rotate8_thread(void *);

   int      cc, ii;
   PXM      *pxm2;
   
   ww1 = pxm1->ww;
   hh1 = pxm1->hh;
   pixmap1 = (uint8 *) pxm1->bmp;
   angle = anglex;

   while (angle < -180) angle += 360;                                      //  normalize, -180 to +180
   while (angle > 180) angle -= 360;
   angle = angle * pi / 180;                                               //  radians, -pi to +pi
   
   if (fabs(angle) < 0.001) {                                              //  angle = 0 within my precision
      pxm2 = PXM_make(ww1,hh1,8);                                          //  return a copy of the input
      pixmap2 = (uint8 *) pxm2->bmp;
      cc = ww1 * hh1 * 3 * sizeof(uint8);
      memcpy(pixmap2,pixmap1,cc);
      return pxm2;
   }

   ww2 = int(ww1*fabs(cos(angle)) + hh1*fabs(sin(angle)));                 //  rectangle containing rotated image
   hh2 = int(ww1*fabs(sin(angle)) + hh1*fabs(cos(angle)));
   
   pxm2 = PXM_make(ww2,hh2,8);
   pixmap2 = (uint8 *) pxm2->bmp;

   for (ii = 0; ii < Nwt; ii++)                                            //  start worker threads
      start_detached_thread(PXM_rotate8_thread,&wtnx[ii]);
   zadd_locked(busy,+Nwt);

   while (busy) zsleep(0.004);                                             //  wait for completion
   return pxm2;
}


void * PXM_rotate8_thread(void *arg)
{
   using namespace rotpxm8;

   int      index = *((int *) (arg));
   int      px2, py2, px0, py0;
   uint8    *pix0, *pix1, *pix2, *pix3;
   double   px1, py1;
   double   f0, f1, f2, f3, red, green, blue;
   double   a, b, d, e, ww15, hh15, ww25, hh25;

   ww15 = 0.5 * ww1;
   hh15 = 0.5 * hh1;
   ww25 = 0.5 * ww2;
   hh25 = 0.5 * hh2;

   a = cos(angle);
   b = sin(angle);
   d = - sin(angle);
   e = cos(angle);
   
   for (py2 = index; py2 < hh2; py2 += Nwt)                                //  loop through output pixels
   for (px2 = 0; px2 < ww2; px2++)                                         //  outer loop y
   {
      px1 = a * (px2 - ww25) + b * (py2 - hh25) + ww15;                    //  (px1,py1) = corresponding    v.9.3
      py1 = d * (px2 - ww25) + e * (py2 - hh25) + hh15;                    //    point within input pixels

      px0 = int(px1);                                                      //  pixel containing (px1,py1)
      py0 = int(py1);
      
      if (px1 < 0 || px0 >= ww1-1 || py1 < 0 || py0 >= hh1-1) {            //  if outside input pixel array
         pix2 = pixmap2 + (py2 * ww2 + px2) * 3;                           //    output is black
         pix2[0] = pix2[1] = pix2[2] = 0;
         continue;
      }

      pix0 = pixmap1 + (py0 * ww1 + px0) * 3;                              //  4 input pixels based at (px0,py0)
      pix1 = pix0 + ww1 * 3;
      pix2 = pix0 + 3;
      pix3 = pix1 + 3;

      f0 = (px0+1 - px1) * (py0+1 - py1);                                  //  overlap of (px1,py1)
      f1 = (px0+1 - px1) * (py1 - py0);                                    //    in each of the 4 pixels
      f2 = (px1 - px0) * (py0+1 - py1);
      f3 = (px1 - px0) * (py1 - py0);
   
      red =   f0 * pix0[0] + f1 * pix1[0] + f2 * pix2[0] + f3 * pix3[0];   //  sum the weighted inputs
      green = f0 * pix0[1] + f1 * pix1[1] + f2 * pix2[1] + f3 * pix3[1];
      blue =  f0 * pix0[2] + f1 * pix1[2] + f2 * pix2[2] + f3 * pix3[2];
      
      pix2 = pixmap2 + (py2 * ww2 + px2) * 3;                              //  output pixel
      pix2[0] = int(red);
      pix2[1] = int(green);
      pix2[2] = int(blue);
   }
   
   zadd_locked(busy,-1);
   return 0;
}


/**************************************************************************

   PXM *pxm2 = PXM_rotate16(PXM *pxm1, double angle)
   Rotate PXM-16 pixmap through an arbitrary angle (degrees).
   Identical to PXM_rotate8() except for:
      uint8 >> uint16
      rotpxm8 >> rotpxm16
      8 >> 16   

**********/

namespace rotpxm16 {
   int      busy = 0;
   uint16   *pixmap1;
   uint16   *pixmap2;
   int      ww1;
   int      hh1;
   int      ww2;
   int      hh2;
   double   angle;
}


PXM * PXM_rotate16(PXM *pxm1, double anglex)
{
   using namespace rotpxm16;

   void     *PXM_rotate16_thread(void *);

   int      cc, ii;
   PXM      *pxm2;
   
   ww1 = pxm1->ww;
   hh1 = pxm1->hh;
   pixmap1 = (uint16 *) pxm1->bmp;
   angle = anglex;

   while (angle < -180) angle += 360;                                      //  normalize, -180 to +180
   while (angle > 180) angle -= 360;
   angle = angle * pi / 180;                                               //  radians, -pi to +pi
   
   if (fabs(angle) < 0.001) {                                              //  angle = 0 within my precision
      pxm2 = PXM_make(ww1,hh1,16);                                         //  return a copy of the input
      pixmap2 = (uint16 *) pxm2->bmp;
      cc = ww1 * hh1 * 3 * sizeof(uint16);
      memcpy(pixmap2,pixmap1,cc);
      return pxm2;
   }

   ww2 = int(ww1*fabs(cos(angle)) + hh1*fabs(sin(angle)));                 //  rectangle containing rotated image
   hh2 = int(ww1*fabs(sin(angle)) + hh1*fabs(cos(angle)));
   
   pxm2 = PXM_make(ww2,hh2,16);
   pixmap2 = (uint16 *) pxm2->bmp;

   for (ii = 0; ii < Nwt; ii++)                                            //  start worker threads
      start_detached_thread(PXM_rotate16_thread,&wtnx[ii]);
   zadd_locked(busy,+Nwt);

   while (busy) zsleep(0.004);                                             //  wait for completion
   return pxm2;
}


void * PXM_rotate16_thread(void *arg)
{
   using namespace rotpxm16;

   int      index = *((int *) (arg));
   int      px2, py2, px0, py0;
   uint16   *pix0, *pix1, *pix2, *pix3;
   double   px1, py1;
   double   f0, f1, f2, f3, red, green, blue;
   double   a, b, d, e, ww15, hh15, ww25, hh25;

   ww15 = 0.5 * ww1;
   hh15 = 0.5 * hh1;
   ww25 = 0.5 * ww2;
   hh25 = 0.5 * hh2;

   a = cos(angle);
   b = sin(angle);
   d = - sin(angle);
   e = cos(angle);
   
   for (py2 = index; py2 < hh2; py2 += Nwt)                                //  loop through output pixels
   for (px2 = 0; px2 < ww2; px2++)                                         //  outer loop y
   {
      px1 = a * (px2 - ww25) + b * (py2 - hh25) + ww15;                    //  (px1,py1) = corresponding    v.9.3
      py1 = d * (px2 - ww25) + e * (py2 - hh25) + hh15;                    //    point within input pixels

      px0 = int(px1);                                                      //  pixel containing (px1,py1)
      py0 = int(py1);
      
      if (px1 < 0 || px0 >= ww1-1 || py1 < 0 || py0 >= hh1-1) {            //  if outside input pixel array
         pix2 = pixmap2 + (py2 * ww2 + px2) * 3;                           //    output is black
         pix2[0] = pix2[1] = pix2[2] = 0;
         continue;
      }

      pix0 = pixmap1 + (py0 * ww1 + px0) * 3;                              //  4 input pixels based at (px0,py0)
      pix1 = pix0 + ww1 * 3;
      pix2 = pix0 + 3;
      pix3 = pix1 + 3;

      f0 = (px0+1 - px1) * (py0+1 - py1);                                  //  overlap of (px1,py1)
      f1 = (px0+1 - px1) * (py1 - py0);                                    //    in each of the 4 pixels
      f2 = (px1 - px0) * (py0+1 - py1);
      f3 = (px1 - px0) * (py1 - py0);
   
      red =   f0 * pix0[0] + f1 * pix1[0] + f2 * pix2[0] + f3 * pix3[0];   //  sum the weighted inputs
      green = f0 * pix0[1] + f1 * pix1[1] + f2 * pix2[1] + f3 * pix3[1];
      blue =  f0 * pix0[2] + f1 * pix1[2] + f2 * pix2[2] + f3 * pix3[2];
      
      pix2 = pixmap2 + (py2 * ww2 + px2) * 3;                              //  output pixel
      pix2[0] = int(red);
      pix2[1] = int(green);
      pix2[2] = int(blue);
   }
   
   zadd_locked(busy,-1);
   return 0;
}



